#ifndef _H_RPMSPEC_
#define _H_RPMSPEC_

/** \ingroup rpmbuild
 * \file build/rpmspec.h
 *  The Spec and Package data structures used during build.
 */

#include <stdio.h>
#include <rpmdefs.h>
#include <rpmiotypes.h>
#include <rpmmacro.h>
#include <rpmio.h>
#include <rpmtag.h>
#include <rpmtypes.h>

#include <rpmevr.h>
#include <rpmfi.h>	/* XXX rpmfi typedef */
#include <rpmcli.h>	/* XXX QVA_t typedef */

/** \ingroup rpmbuild
 */
typedef struct Package_s * Package;

/** \ingroup rpmbuild
 */
typedef struct Source * SpecSource;

/** \ingroup rpmbuild
 */
struct TriggerFileEntry {
    int index;
    char * fileName;
    char * script;
    char * prog;
    struct TriggerFileEntry * next;
};

#define RPMBUILD_DEFAULT_LANG "C"

/** \ingroup rpmbuild
 */
struct Source {
    const char * fullSource;
    const char * source;	/* Pointer into fullSource */
    int flags;
    rpmuint32_t num;
    struct Source * next;
};

/** \ingroup rpmbuild
 */
typedef struct ReadLevelEntry {
    int reading;
    struct ReadLevelEntry * next;
} RLE_t;

/** \ingroup rpmbuild
 */
typedef struct OpenFileInfo {
    const char * fileName;
    FD_t fd;
    int lineNum;
    char readBuf[BUFSIZ];
    char * readPtr;
    struct OpenFileInfo * next;
} OFI_t;

/** \ingroup rpmbuild
 */
typedef struct spectag_s {
    int t_tag;
    int t_startx;
    int t_nlines;
    const char * t_lang;
    const char * t_msgid;
} * spectag;

/** \ingroup rpmbuild
 */
typedef struct spectags_s {
    spectag st_t;
    int st_nalloc;
    int st_ntags;
} * spectags;

/** \ingroup rpmbuild
 */
typedef struct speclines_s {
    char **sl_lines;
    int sl_nalloc;
    int sl_nlines;
} * speclines;

/** \ingroup rpmbuild
 * The structure used to store values parsed from a spec file.
 */
struct Spec_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    const char * specFile;	/*!< Name of the spec file. */
    const char * buildSubdir;
    const char * rootURL;

    speclines sl;
    spectags st;

    struct OpenFileInfo * fileStack;
    char *lbuf;
    size_t lbuf_len;
    char *lbufPtr;
    char nextpeekc;
    char * nextline;
    char * line;
    int lineNum;

    struct ReadLevelEntry * readStack;

    Spec * BASpecs;
    const char ** BANames;
    int BACount;
    int recursing;		/*!< parse is recursive? */
    int toplevel;

    int force;
    int anyarch;

    char * passPhrase;
    int timeCheck;
    const char * cookie;

    struct Source * sources;
    int numSources;
    int noSource;

    const char * sourceRpmName;
    unsigned char * sourcePkgId;
    Header sourceHeader;
    rpmfi fi;
    int sourceHdrInit;

    MacroContext macros;

    rpmRC (*_parseRCPOT) (Spec spec, Package pkg, const char *field, rpmTag tagN,
               rpmuint32_t index, rpmsenseFlags tagflags);

    rpmuint32_t sstates[RPMSCRIPT_MAX];		/*!< scriptlet states. */
    rpmuint32_t smetrics[RPMSCRIPT_MAX];	/*!< scriptlet time metrics. */

    rpmiob prep;		/*!< %prep scriptlet. */
    rpmiob build;		/*!< %build scriptlet. */
    rpmiob install;		/*!< %install scriptlet. */
    rpmiob check;		/*!< %check scriptlet. */
    rpmiob clean;		/*!< %clean scriptlet. */

    size_t nfoo;
    tagStore_t foo;

    void * dig;			/*!< Auto-signing keypair. */

    Package packages;		/*!< Package list. */
};

/** \ingroup rpmbuild
 * The structure used to store values for a package.
 */
struct Package_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    Header header;
    rpmds ds;			/*!< Requires: N = EVR */
    rpmfi fi;

    int autoReq;
    int autoProv;
    int noarch;

    const char * preInFile;	/*!< %pre scriptlet. */
    const char * postInFile;	/*!< %post scriptlet. */
    const char * preUnFile;	/*!< %preun scriptlet. */
    const char * postUnFile;	/*!< %postun scriptlet. */
    const char * preTransFile;	/*!< %pretrans scriptlet. */
    const char * postTransFile;	/*!< %posttrans scriptlet. */
    const char * verifyFile;	/*!< %verifyscript scriptlet. */
    const char * sanityCheckFile;/*!< %sanitycheck scriptlet. */

    rpmiob specialDoc;

    struct TriggerFileEntry * triggerFiles;

    const char * fileFile;
    rpmiob fileList;		/* If NULL, package will not be written */

    Package next;
};

#ifdef __cplusplus
extern "C" {
#endif

/** \ingroup rpmbuild
 * Destroy a spec file control structure.
 * @todo Rename to specFree.
 * @param spec		spec file control structure
 * @return		NULL on last dereference
 */
Spec freeSpec(Spec spec);
#define	freeSpec(_spec)	\
    ((Spec)rpmioFreePoolItem((rpmioItem)(_spec), __FUNCTION__, __FILE__, __LINE__))

/** \ingroup rpmbuild
 * Create and initialize Spec structure.
 * @return spec		spec file control structure
 */
Spec newSpec(void);

/** \ingroup rpmbuild
 * Function to query spec file(s).
 * @param ts		transaction set
 * @param qva		parsed query/verify options
 * @param arg		query argument
 * @return		0 on success, else no. of failures
 */
int rpmspecQuery(rpmts ts, QVA_t qva, const char * arg);

/** \ingroup rpmbuild
 */
struct OpenFileInfo * newOpenFileInfo(void);

/** \ingroup rpmbuild
 * stashSt.
 * @param spec		spec file control structure
 * @param h		header
 * @param tag		tag
 * @param lang		locale
 * @return		ptr to saved entry
 */
spectag stashSt(Spec spec, Header h, rpmTag tag, const char * lang);

/** \ingroup rpmbuild
 * addSource.
 * @param spec		spec file control structure
 * @param pkg		package control
 * @param field		field to parse
 * @param tag		tag
 * @return		0 on success
 */
int addSource(Spec spec, Package pkg, const char * field, rpmTag tag);

/** \ingroup rpmbuild
 * parseNoSource.
 * @param spec		spec file control structure
 * @param field		field to parse
 * @param tag		tag
 * @return		0 on success
 */
int parseNoSource(Spec spec, const char * field, rpmTag tag);

/** \ingroup rpmbuild
 * Return the count of source set in specfile
 * @param spec      spec file control structure
 * @return  the count of source
 */
int SpecSourceCount(Spec spec)
	RPM_GNUC_PURE;

/** \ingroup rpmbuild
 * Return a source control structure
 * @param spec      spec file control structure
 * @param num       the number of the wanted source (starting from 0)
 * @return          a SpecSource structure, NULL if not found
 */
SpecSource getSource(Spec spec, int num)
	RPM_GNUC_PURE;

/** \ingroup rpmbuild
 * Return a ptr to the source file name
 * @param source    SpecSource control structure
 * @return          ptr to filename
 */
const char * specSourceName(SpecSource source)
	RPM_GNUC_PURE;

/** \ingroup rpmbuild
 * Return a ptr to the full url of the source
 * @param source    SpecSource control structure
 * @return          ptr to url
 */
const char * specFullSourceName(SpecSource source)
	RPM_GNUC_PURE;

/** \ingroup rpmbuild
 * Return the spec or source patch number
 * @param source    SpecSource control structure
 * @return          the number of the source
 */
int specSourceNum(SpecSource source)
	RPM_GNUC_PURE;

/** \ingroup rpmbuild
 * Return flags set for the source
 * @param source    SpecSource control structure
 * @return          flags
 */
int specSourceFlags(SpecSource source)
	RPM_GNUC_PURE;

/** \ingroup rpmbuild
 * Return the macro directory location from source file flags
 * @param attr      rpmfileAttrs from source
 * @return          string containings macros about location, NULL on failure
 */
#if defined(RPM_VENDOR_OPENPKG) /* splitted-source-directory */
const char * getSourceDir(rpmfileAttrs attr, const char *filename)
#else
const char * getSourceDir(rpmfileAttrs attr)
#endif
	RPM_GNUC_CONST;

#ifdef __cplusplus
}
#endif

#endif /* _H_SPEC_ */
