/** \ingroup js_c
 * \file js/rpmsp-js.c
 */

#include "system.h"

#include "rpmsp-js.h"
#include "rpmjs-debug.h"

#if defined(WITH_SEPOL)
#include <sepol/sepol.h>
#endif
#define	_RPMSP_INTERNAL
#include <rpmsp.h>

#include "debug.h"

/*@unchecked@*/
static int _debug = 0;

/* Required JSClass vectors */
#define	rpmsp_addprop		JS_PropertyStub
#define	rpmsp_delprop		JS_PropertyStub
#define	rpmsp_convert		JS_ConvertStub

/* Optional JSClass vectors */
#define	rpmsp_getobjectops	NULL
#define	rpmsp_checkaccess	NULL
#define	rpmsp_call		rpmsp_call
#define	rpmsp_construct		rpmsp_ctor
#define	rpmsp_xdrobject		NULL
#define	rpmsp_hasinstance	NULL
#define	rpmsp_mark		NULL
#define	rpmsp_reserveslots	NULL

/* Extended JSClass vectors */
#define rpmsp_equality		NULL
#define rpmsp_outerobject	NULL
#define rpmsp_innerobject	NULL
#define rpmsp_iteratorobject	NULL
#define rpmsp_wrappedobject	NULL

/* --- helpers */

/* --- Object methods */

static JSFunctionSpec rpmsp_funcs[] = {
    JS_FS_END
};

/* --- Object properties */
enum rpmsp_tinyid {
    _DEBUG	= -2,
};

static JSPropertySpec rpmsp_props[] = {
    {"debug",	_DEBUG,		JSPROP_ENUMERATE,	NULL,	NULL},
    {NULL, 0, 0, NULL, NULL}
};

#define	_GET_STR(_str)	\
    ((_str) ? STRING_TO_JSVAL(JS_NewStringCopyZ(cx, (_str))) : JSVAL_VOID)
#define	_GET_CON(_test)	((_test) ? _GET_STR(con) : JSVAL_VOID)

static bool
rpmsp_getprop(JSContext *cx, JSObject *obj, jsid id, jsval *vp)
{
    void * ptr = JS_GetInstancePrivate(cx, obj, &rpmspClass, nullptr);
    jsint tiny = JSVAL_TO_INT(id);

    /* XXX the class has ptr == NULL, instances have ptr != NULL. */
    if (ptr == NULL)
	return true;

    switch (tiny) {
    case _DEBUG:
	*vp = INT_TO_JSVAL(_debug);
	break;
    default:
	break;
    }

    return true;
}

#define	_PUT_CON(_put)	(JSVAL_IS_STRING(*vp) && !(_put) \
	? JSVAL_TRUE : JSVAL_FALSE)
#define	_PUT_INT(_put)	(JSVAL_IS_INT(*vp) && !(_put) \
	? JSVAL_TRUE : JSVAL_FALSE)

static bool
rpmsp_setprop(JSContext *cx, JSObject *obj, jsid id, bool strict, jsval *vp)
{
    void * ptr = JS_GetInstancePrivate(cx, obj, &rpmspClass, nullptr);
    jsint tiny = JSVAL_TO_INT(id);

    /* XXX the class has ptr == NULL, instances have ptr != NULL. */
    if (ptr == NULL)
	return true;

    switch (tiny) {
    case _DEBUG:
	if (!JS_ValueToInt32(cx, *vp, &_debug))
	    break;
	break;
    default:
	break;
    }

    return true;
}

static bool
rpmsp_resolve(JSContext *cx, JSObject *obj, jsid id, uintN flags,
	JSObject **objp)
{
    void * ptr = JS_GetInstancePrivate(cx, obj, &rpmspClass, nullptr);

//_RESOLVE_DEBUG_ENTRY(_debug < 0);

    if ((flags & JSRESOLVE_ASSIGNING)
     || (ptr == NULL)) { /* don't resolve to parent prototypes objects. */
	*objp = NULL;
	goto exit;
    }

    *objp = obj;	/* XXX always resolve in this object. */

exit:
    return true;
}

static bool
rpmsp_enumerate(JSContext *cx, JSObject *obj, JSIterateOp op,
		  jsval *statep, jsid *idp)
{

//_ENUMERATE_DEBUG_ENTRY(_debug < 0);

    switch (op) {
    case JSENUMERATE_INIT:
    case JSENUMERATE_INIT_ALL:
	*statep = JSVAL_VOID;
        if (idp)
            *idp = JSVAL_ZERO;
        break;
    case JSENUMERATE_NEXT:
	*statep = JSVAL_VOID;
	if (!JSID_IS_VOID(*idp))
            break;
        /*@fallthrough@*/
    case JSENUMERATE_DESTROY:
	*statep = JSVAL_NULL;
        break;
    }
    return true;
}

/* --- Object ctors/dtors */
static rpmsp
rpmsp_init(JSContext *cx, JSObject *obj)
{
    const char * _fn = NULL;
    int _flags = 0;
    rpmsp sp = rpmspNew(_fn, _flags);

if (_debug)
fprintf(stderr, "==> %s(%p,%p) sp %p\n", __FUNCTION__, cx, obj, sp);

    if (!JS_SetPrivate(cx, obj, (void *)sp)) {
	/* XXX error msg */
	return NULL;
    }
    return sp;
}

static void
rpmsp_dtor(JSContext *cx, JSObject *obj)
{
    void * ptr = JS_GetInstancePrivate(cx, obj, &rpmspClass, nullptr);
    rpmsp sp = (rpmsp) ptr;

_DTOR_DEBUG_ENTRY(_debug);

    (void) rpmspFree(sp);
}

static bool
rpmsp_ctor(JSContext *cx, unsigned argc, Value* vp)
{
    CallArgs args = CallArgsFromVp(argc, vp);
    JSObject *obj = JS_NewObjectForConstructor(cx, vp);
    bool ok = false;

_CTOR_DEBUG_ENTRY(_debug);

    if (JS_IsConstructing(cx, vp)) {
	(void) rpmsp_init(cx, obj);
    } else {
	if ((obj = JS_NewObject(cx, &rpmspClass, NULL, NULL)) == NULL)
	    goto exit;
	*vp = OBJECT_TO_JSVAL(obj);
    }
    ok = true;

exit:
    return ok;
}

static bool
rpmsp_call(JSContext *cx, unsigned argc, Value* vp)
{
    CallArgs args = CallArgsFromVp(argc, vp);
    /* XXX obj is the global object so lookup "this" object. */
    JSObject * o = JSVAL_TO_OBJECT(args[-2]);
    void * ptr = JS_GetInstancePrivate(cx, o, &rpmspClass, NULL);
    bool ok = false;
#ifdef	FIXME
    rpmsp sp = (rpmsp) ptr;
    const char *_fn = NULL;
    const char * _con = NULL;

    if (!(ok = JS_ConvertArguments(cx, args, "s", &_fn)))
        goto exit;

    *vp = (sp && _fn && (_con = rpmspLgetfilecon(sp, _fn)) != NULL)
	? STRING_TO_JSVAL(JS_NewStringCopyZ(cx, _con)) : JSVAL_VOID;
    _con = _free(_con);

    ok = true;

exit:
#endif

if (_debug)
fprintf(stderr, "<== %s(%p,%p[%u],%p) o %p ptr %p\n", __FUNCTION__, cx, args, (unsigned)argc, vp, o, ptr);

    return ok;
}

/* --- Class initialization */
JSClass rpmspClass = {
    "Sp", JSCLASS_NEW_RESOLVE | JSCLASS_NEW_ENUMERATE | JSCLASS_HAS_PRIVATE,
    rpmsp_addprop,   rpmsp_delprop, rpmsp_getprop, rpmsp_setprop,
    (JSEnumerateOp)rpmsp_enumerate, (JSResolveOp)rpmsp_resolve,
    rpmsp_convert,	rpmsp_dtor,

    rpmsp_getobjectops,	rpmsp_checkaccess,
    rpmsp_call,		rpmsp_construct,
    rpmsp_xdrobject,	rpmsp_hasinstance,
    rpmsp_mark,		rpmsp_reserveslots,
};

JSObject *
rpmjs_InitSpClass(JSContext *cx, JSObject* obj)
{
    JSObject *proto;

if (_debug)
fprintf(stderr, "==> %s(%p,%p)\n", __FUNCTION__, cx, obj);

    proto = JS_InitClass(cx, obj, NULL, &rpmspClass, rpmsp_ctor, 1,
		rpmsp_props, rpmsp_funcs, NULL, NULL);
assert(proto != NULL);
    return proto;
}

JSObject *
rpmjs_NewSpObject(JSContext *cx)
{
    JSObject *obj;
    rpmsp sp;

    if ((obj = JS_NewObject(cx, &rpmspClass, NULL, NULL)) == NULL) {
	/* XXX error msg */
	return NULL;
    }
    if ((sp = rpmsp_init(cx, obj)) == NULL) {
	/* XXX error msg */
	return NULL;
    }
    return obj;
}

GPSEE_MODULE_WRAP(rpmsp, Sp, true)
