#ifndef H_RPMDS
#define H_RPMDS

/** \ingroup rpmds
 * \file lib/rpmds.h
 * Structure(s) used for dependency tag sets.
 */

#include <stdio.h>
#include <time.h>
#include <rpmdefs.h>
#include <rpmiotypes.h>
#include <rpmtypes.h>
#include <rpmtag.h>

#include <rpmevr.h>
#define	_RPMNS_INTERNAL
#include <rpmns.h>
#include <rpmps.h>

/** \ingroup rpmds
 */
extern int _rpmds_debug;

/** \ingroup rpmds
 */
extern const char *_sysinfo_path;

/** \ingroup rpmds
 */
extern int _rpmds_nopromote;

#if defined(_RPMDS_INTERNAL)
#include <mire.h>

/** \ingroup rpmds
 * A dependency set.
 */
struct rpmds_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    const char * Type;		/*!< Tag name. */
    const char * DNEVR;		/*!< Formatted dependency string. */
    Header h;			/*!< Header for dependency set (or NULL) */
    const char ** N;		/*!< Name. */
    const char ** EVR;		/*!< Epoch-Version-Release. */
    evrFlags * Flags;		/*!< Bit(s) identifying context/comparison. */
    rpmuint32_t * Color;	/*!< Bit(s) calculated from file color(s). */
    rpmuint32_t * Refs;		/*!< No. of file refs. */
    rpmint32_t * Result;	/*!< Dependency check result. */
    int (*EVRparse) (const char *evrstr, EVR_t evr);	 /* EVR parsing. */
    int (*EVRcmp) (const char *a, const char *b);	 /* EVR comparison. */
    struct rpmns_s ns;		/*!< Name (split). */
    miRE exclude;		/*!< Iterator exclude patterns. */
    int nexclude;		/*!< No. of exclude patterns. */
    miRE include;		/*!< Iterator include patterns. */
    int ninclude;		/*!< No. of include patterns. */
    const char * A;		/*!< Arch (from containing package). */
    rpmuint32_t BT;		/*!< Package build time tie breaker. */
    rpmTag tagN;		/*!< Header tag. */
    rpmuint32_t Count;		/*!< No. of elements */
    int i;			/*!< Element index. */
    unsigned l;			/*!< Low element (bsearch). */
    unsigned u;			/*!< High element (bsearch). */
    int nopromote;		/*!< Don't promote Epoch: in rpmdsCompare()? */
};
#endif	/* _RPMDS_INTERNAL */

#if defined(_RPMPRCO_INTERNAL)
/** \ingroup rpmds
 * Container for provides/requires/conflicts/obsoletes dependency set(s).
 */
struct rpmPRCO_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    rpmds * Pdsp;		/*!< Provides: collector. */
    rpmds * Rdsp;		/*!< Requires: collector. */
    rpmds * Cdsp;		/*!< Conflicts: collector. */
    rpmds * Odsp;		/*!< Obsoletes: collector. */
    rpmds * Tdsp;		/*!< Triggers collector. */
    rpmds * Ddsp;		/*!< Dirnames collector. */
    rpmds * Ldsp;		/*!< Linktos collector. */
    rpmds my;			/*!< N = EVR */
    rpmds P;			/*!< Provides: */
    rpmds R;			/*!< Requires: */
    rpmds C;			/*!< Conflicts: */
    rpmds O;			/*!< Obsoletes: */
    rpmds T;			/*!< Triggers */
    rpmds D;			/*!< Dirnames */
    rpmds L;			/*!< Linktos */
};
#endif	/* _RPMPRCO_INTERNAL */

#ifdef __cplusplus
extern "C" {
#endif

/** \name RPMDS */
/*@{*/

/** \ingroup rpmds
 * Unreference a dependency set instance.
 * @param ds		dependency set
 * @param msg
 * @return		NULL on last dereference
 */
rpmds rpmdsUnlink (rpmds ds, const char * msg);
#define	rpmdsUnlink(_ds, _msg)	\
    ((rpmds)rpmioUnlinkPoolItem((rpmioItem)(_ds), _msg, __FILE__, __LINE__))

/** \ingroup rpmds
 * Reference a dependency set instance.
 * @param ds		dependency set
 * @param msg
 * @return		new dependency set reference
 */
rpmds rpmdsLink (rpmds ds, const char * msg);
#define	rpmdsLink(_ds, _msg)	\
    ((rpmds)rpmioLinkPoolItem((rpmioItem)(_ds), _msg, __FILE__, __LINE__))

/** \ingroup rpmds
 * Destroy a dependency set.
 * @param ds		dependency set
 * @return		NULL on last dereference
 */
rpmds rpmdsFree(rpmds ds);
#define	rpmdsFree(_ds)	\
    ((rpmds)rpmioFreePoolItem((rpmioItem)(_ds), __FUNCTION__, __FILE__, __LINE__))

/** \ingroup rpmds
 * Create and load a dependency set.
 * @param h		header
 * @param tagN		type of dependency
 * @param flags		scareMem(0x1), nofilter(0x2)
 * @return		new dependency set
 */
rpmds rpmdsNew(Header h, rpmTag tagN, int flags);

/** \ingroup rpmds
 * Return N string, expanded if necessary.
 * @param ds		dependency set
 * @return		new N string (malloc'ed)
 */
const char * rpmdsNewN(rpmds ds);

/** \ingroup rpmds
 * Return new formatted dependency string.
 * @param dspfx		formatted dependency string prefix
 * @param ds		dependency set
 * @return		new formatted dependency (malloc'ed)
 */
char * rpmdsNewDNEVR(const char * dspfx, rpmds ds);

/** \ingroup rpmds
 * Create, load and initialize a dependency for this header. 
 * @param h		header
 * @param tagN		type of dependency
 * @param Flags		comparison flags
 * @return		new dependency set
 */
rpmds rpmdsThis(Header h, rpmTag tagN, evrFlags Flags);

/** \ingroup rpmds
 * Create, load and initialize a dependency set of size 1.
 * @param tagN		type of dependency
 * @param N		name
 * @param EVR		epoch:version-release
 * @param Flags		comparison/context flags
 * @return		new dependency set
 */
rpmds rpmdsSingle(rpmTag tagN, const char * N, const char * EVR, evrFlags Flags);

/** \ingroup rpmds
 * Return dependency set count.
 * @param ds		dependency set
 * @return		current count
 */
int rpmdsCount(const rpmds ds)
	RPM_GNUC_PURE;

/** \ingroup rpmds
 * Return dependency set index.
 * @param ds		dependency set
 * @return		current index
 */
int rpmdsIx(const rpmds ds)
	RPM_GNUC_PURE;

/** \ingroup rpmds
 * Set dependency set index.
 * @param ds		dependency set
 * @param ix		new index
 * @return		current index
 */
int rpmdsSetIx(rpmds ds, int ix);

/** \ingroup rpmds
 * Return current formatted dependency string.
 * @param ds		dependency set
 * @return		current dependency DNEVR, NULL on invalid
 */
extern const char * rpmdsDNEVR(const rpmds ds)
	RPM_GNUC_PURE;

/** \ingroup rpmds
 * Return current dependency name.
 * @param ds		dependency set
 * @return		current dependency name, NULL on invalid
 */
extern const char * rpmdsN(rpmds ds);

/** \ingroup rpmds
 * Return current dependency epoch-version-release.
 * @param ds		dependency set
 * @return		current dependency EVR, NULL on invalid
 */
extern const char * rpmdsEVR(const rpmds ds)
	RPM_GNUC_PURE;

/** \ingroup rpmds
 * Return current dependency flags.
 * @param ds		dependency set
 * @return		current dependency flags, 0 on invalid
 */
evrFlags rpmdsFlags(const rpmds ds)
	RPM_GNUC_PURE;

/** \ingroup rpmds
 * Return current dependency type.
 * @param ds		dependency set
 * @return		current dependency type, 0 on invalid
 */
rpmTag rpmdsTagN(const rpmds ds)
	RPM_GNUC_PURE;

/** \ingroup rpmds
 * Return current dependency arch.
 * @param ds		dependency set
 * @return		current dependency arch, NULL on invalid
 */
extern const char * rpmdsA(const rpmds ds)
	RPM_GNUC_PURE;

/** \ingroup rpmds
 * Return dependency build time.
 * @param ds		dependency set
 * @return		dependency build time, 0 on invalid
 */
time_t rpmdsBT(const rpmds ds)
	RPM_GNUC_PURE;

/** \ingroup rpmds
 * Set dependency build time.
 * @param ds		dependency set
 * @param BT		build time
 * @return		dependency build time, 0 on invalid
 */
time_t rpmdsSetBT(const rpmds ds, time_t BT);

/** \ingroup rpmds
 * Return dependency class type.
 * @param ds		dependency set
 * @return		dependency class type
 */
nsType rpmdsNSType(const rpmds ds)
	RPM_GNUC_PURE;

/** \ingroup rpmds
 * Return current "Don't promote Epoch:" flag.
 *
 * This flag controls for Epoch: promotion when a dependency set is
 * compared. If the flag is set (for already installed packages), then
 * an unspecified value will be treated as Epoch: 0. Otherwise (for added
 * packages), the Epoch: portion of the comparison is skipped if the value
 * is not specified, i.e. an unspecified Epoch: is assumed to be equal
 * in dependency comparisons.
 *
 * @param ds		dependency set
 * @return		current "Don't promote Epoch:" flag
 */
int rpmdsNoPromote(const rpmds ds)
	RPM_GNUC_PURE;

/** \ingroup rpmds
 * Set "Don't promote Epoch:" flag.
 * @param ds		dependency set
 * @param nopromote	Should an unspecified Epoch: be treated as Epoch: 0?
 * @return		previous "Don't promote Epoch:" flag
 */
int rpmdsSetNoPromote(rpmds ds, int nopromote);

/** \ingroup rpmds
 * Set EVR parsing function.
 * @param ds		dependency set
 * @param EVRparse	EVR parsing function (NULL uses default)
 * @return		previous EVR parsing function
 */
void * rpmdsSetEVRparse(rpmds ds,
		int (*EVRparse)(const char *everstr, EVR_t evr));

/** \ingroup rpmds
 * Set EVR comparison function.
 * @param ds		dependency set
 * @param EVRcmp	EVR comparison function (NULL uses default)
 * @return		previous EVR comparison function
 */
void * rpmdsSetEVRcmp(rpmds ds,
		int (*EVRcmp)(const char *a, const char *b));

/** \ingroup rpmds
 * Return current dependency color.
 * @param ds		dependency set
 * @return		current dependency color (0 if not set)
 */
rpmuint32_t rpmdsColor(const rpmds ds)
	RPM_GNUC_PURE;

/** \ingroup rpmds
 * Set current dependency color.
 * @param ds		dependency set
 * @param color		new dependency color
 * @return		previous dependency color
 */
rpmuint32_t rpmdsSetColor(const rpmds ds, rpmuint32_t color);

/** \ingroup rpmds
 * Return dependency exclude patterns.
 * @param ds		dependency set
 * @return		dependency exclude patterns (NULL if not set)
 */
void * rpmdsExclude(const rpmds ds)
	RPM_GNUC_PURE;

/** \ingroup rpmds
 * Return no. of dependency exclude patterns.
 * @param ds		dependency set
 * @return		dependency exclude patterns (0 if not set)
 */
int rpmdsNExclude(const rpmds ds)
	RPM_GNUC_PURE;

/** \ingroup rpmds
 * Return dependency include patterns.
 * @param ds		dependency set
 * @return		dependency include patterns (NULL if not set)
 */
void * rpmdsInclude(const rpmds ds)
	RPM_GNUC_PURE;

/** \ingroup rpmds
 * Return no. of dependency include patterns.
 * @param ds		dependency set
 * @return		dependency include patterns (0 if not set)
 */
int rpmdsNInclude(const rpmds ds)
	RPM_GNUC_PURE;

/** \ingroup rpmds
 * Return current dependency file refs.
 * @param ds		dependency set
 * @return		current dependency file refs (0 if not set)
 */
rpmuint32_t rpmdsRefs(const rpmds ds)
	RPM_GNUC_PURE;

/** \ingroup rpmds
 * Set current dependency file refs.
 * @param ds		dependency set
 * @param refs		new dependency refs
 * @return		previous dependency refs
 */
rpmuint32_t rpmdsSetRefs(const rpmds ds, rpmuint32_t refs);

/** \ingroup rpmds
 * Return current dependency comparison result.
 * @param ds		dependency set
 * @return		current dependency result (0 if not set)
 */
rpmint32_t rpmdsResult(const rpmds ds)
	RPM_GNUC_PURE;

/** \ingroup rpmds
 * Set current dependency comparison result.
 * @param ds		dependency set
 * @param result	new dependency result
 * @return		previous dependency result
 */
rpmint32_t rpmdsSetResult(const rpmds ds, rpmint32_t result);

/** \ingroup rpmds
 * Notify of results of dependency match.
 * @param ds		dependency set
 * @param where		where dependency was resolved (or NULL)
 * @param rc		0 == YES, otherwise NO
 */
void rpmdsNotify(rpmds ds, const char * where, int rc);

/** \ingroup rpmds
 * Return next dependency set iterator index.
 * @param ds		dependency set
 * @return		dependency set iterator index, -1 on termination
 */
int rpmdsNext(rpmds ds);

/** \ingroup rpmds
 * Initialize dependency set iterator.
 * @param ds		dependency set
 * @return		dependency set
 */
rpmds rpmdsInit(rpmds ds);

/** \ingroup rpmds
 * Find a dependency set element using binary search.
 * @param ds		dependency set to search
 * @param ods		dependency set element to find.
 * @return		dependency index (or -1 if not found)
 */
int rpmdsFind(rpmds ds, const rpmds ods);

/** \ingroup rpmds
 * Merge a dependency set maintaining (N,EVR,Flags) sorted order.
 * @retval *dsp		(merged) dependency set
 * @param ods		dependency set to merge
 * @return		0 on success
 */
int rpmdsMerge(rpmds * dsp, rpmds ods);

/** \ingroup rpmds
 * Search a sorted dependency set for an element that overlaps.
 * A boolean result is saved (if allocated) and accessible through
 * rpmdsResult(ods) afterwards.
 * @param ds		dependency set to search
 * @param ods		dependency set element to find.
 * @return		dependency index (or -1 if not found)
 */
int rpmdsSearch(rpmds ds, rpmds ods);

/**
 */
extern const char * _cpuinfo_path;

/** \ingroup rpmds
 * Load /proc/cpuinfo provides into a dependency set.
 * @retval *dsp		(loaded) depedency set
 * @param fn		path to file (NULL uses /proc/cpuinfo)
 * @return		0 on success
 */
int rpmdsCpuinfo(rpmds * dsp, const char * fn);

/** \ingroup rpmds
 * Load rpmlib provides into a dependency set.
 * @retval *dsp		(loaded) depedency set
 * @param tblp		rpmlib provides table (NULL uses internal table)
 * @return		0 on success
 */
int rpmdsRpmlib(rpmds * dsp, void * tblp);

/** \ingroup rpmds
 * Load sysinfo dependencies into a dependency set.
 * @retval *PRCO	provides/requires/conflicts/obsoletes depedency set(s)
 * @param fn		path to file (NULL uses /etc/rpm/sysinfo)
 * @return		0 on success
 */
int rpmdsSysinfo(rpmPRCO PRCO, const char * fn);

/** \ingroup rpmds
 * Load getconf provides into a dependency set.
 * @retval *dsp		(loaded) depedency set
 * @param path		getconf path (NULL uses /)
 * @return		0 on success
 */
int rpmdsGetconf(rpmds * dsp, const char * path);

/** \ingroup rpmds
 * Merge provides/requires/conflicts/obsoletes dependencies.
 * @param context	merge dependency set(s) container
 * @param ds		dependency set to merge
 * @return		0 on success
 */
int rpmdsMergePRCO(void * context, rpmds ds);

/** \ingroup rpmds
 * Free dependency set(s) container.
 * @param PRCO		dependency set(s) container
 * @return		NULL
 */
rpmPRCO rpmdsFreePRCO(rpmPRCO PRCO);

/** \ingroup rpmds
 * Create dependency set(s) container.
 * @param h		header
 * @return		0 on success
 */
rpmPRCO rpmdsNewPRCO(Header h);

/** \ingroup rpmds
 * Retrieve a dependency set from container.
 * @param PRCO		dependency set(s) container
 * @param tagN		type of dependency set
 * @return		dependency set (or NULL)
 */
rpmds rpmdsFromPRCO(rpmPRCO PRCO, rpmTag tagN)
	RPM_GNUC_PURE;

/** \ingroup rpmds
 * Extract ELF dependencies from a file.
 * @param fn		file name
 * @param flags		1: skip provides 2: skip requires
 * @param add		*add add(arg, ds) saves next provide/require elf dependency.
 * @param context	add() callback context
 * @return		0 on success
 */
int rpmdsELF(const char * fn, int flags,
		int (*add) (void * context, rpmds ds), void * context);
#define RPMELF_FLAG_SKIPPROVIDES	0x1	/* rpmdsELF: skip provides */
#define RPMELF_FLAG_SKIPREQUIRES	0x2	/* rpmdsELF: skip requires */

/** \ingroup rpmds
 * Load /etc/ld.so.cache provides into a dependency set.
 * @todo Add dependency colors, and attach to file.
 * @retval PRCO		*PRCO provides/requires/conflicts/obsoletes depedency set(s)
 * @param fn		cache file name (NULL uses /etc/ld.so.cache)
 * @return		0 on success
 */
int rpmdsLdconfig(rpmPRCO PRCO, const char * fn);

#if defined(__sun)
/** \ingroup rpmds
 * Given a colon-separated list of directories, search each of those
 * directories for (ELF or ELF64) shared objects, and load the provided
 * shared objects into a dependency set.
 * @todo Add dependency colors, and attach to file.
 * @retval PRCO		*PRCO provides/requires/conflicts/obsoletes depedency set(s)
 * @param rldp		:-sep string of dirs (NULL uses /lib:/usr/lib)
 * @return		0 on success
 */
int rpmdsRldpath(rpmPRCO PRCO, const char * rldp);

/** \ingroup rpmds
 * Use Solaris' crle(1) command to find the ELF (or ELF64) loader path.
 * calls rpmdsRldpath once it has the loader path.
 * @todo Add dependency colors, and attach to file.
 * @retval PRCO		*PRCO provides/requires/conflicts/obsoletes depedency set(s)
 * @param fn		unused.
 * @return		0 on success
 */
int rpmdsCrle(rpmPRCO PRCO, const char * fn);
#endif

struct utsname;
/** \ingroup rpmds
 * Load uname(2) provides into a dependency set.
 * @retval dsp		*dsp (loaded) depedency set
 * @param un		utsname struct (NULL calls uname(2))
 * @return		0 on success
 */
int rpmdsUname(rpmds * dsp, const struct utsname * un);

/** \ingroup rpmds
 * Load provides from a pipe into a dependency set.
 * @retval dsp		*dsp (loaded) depedency set
 * @param tagN		rpmds tag (0 uses RPMTAG_PROVIDENAME).
 * @param cmd		popen cmd to run (NULL loads perl provides)
 * @return		0 on success
 */
int rpmdsPipe(rpmds * dsp, rpmTag tagN, const char * cmd);

/** \ingroup rpmds
 * Compare two versioned dependency ranges, looking for overlap.
 * @param A		1st dependency
 * @param B		2nd dependency
 * @return		1 if dependencies overlap, 0 otherwise
 */
int rpmdsCompare(const rpmds A, const rpmds B);

/** \ingroup rpmds
 * Compare A against every member of B, looking for 1st match.
 * @param A		1st dependency
 * @param B		2nd dependency
 * @return		1 if some dependency overlaps, 0 otherwise
 */
int rpmdsMatch(const rpmds A, rpmds B);

/** \ingroup rpmds
 * Report a Requires: or Conflicts: dependency problem.
 * @param ps		transaction set problems
 * @param pkgNEVR	package name/epoch/version/release
 * @param ds		dependency set
 * @param suggestedKeys	filename or python object address
 * @param adding	dependency problem is from added package set?
 */
void rpmdsProblem(rpmps ps, const char * pkgNEVR, const rpmds ds,
		const fnpyKey * suggestedKeys, int adding);

/** \ingroup rpmds
 * Compare package provides dependencies from header with a single dependency.
 * @param h		header
 * @param req		dependency set
 * @param nopromote	Don't promote Epoch: in comparison?
 * @return		1 if any dependency overlaps, 0 otherwise
 */
int rpmdsAnyMatchesDep (const Header h, const rpmds req, int nopromote);

/** \ingroup rpmds
 * Compare package name-version-release from header with a single dependency.
 * @param h		header
 * @param req		dependency set
 * @param nopromote	Don't promote Epoch: in comparison?
 * @return		1 if dependency overlaps, 0 otherwise
 */
int rpmdsNVRMatchesDep(const Header h, const rpmds req, int nopromote);

/** \ingroup rpmds
 * Negate return code for negated comparisons.
 * @param ds		dependency set
 * @param rc		positive return code
 * @return		return code
 */
int rpmdsNegateRC(const rpmds ds, int rc)
	RPM_GNUC_PURE;

/** \ingroup rpmds
 * Return current dependency type name.
 * @param ds		dependency set
 * @return		current dependency type name
 */
const char * rpmdsType(const rpmds ds);

#if !defined(SWIG)
/** \ingroup rpmds
 * Print current dependency set contents.
 * @param ds		dependency set
 * @param fp		file handle (NULL uses stderr)
 * @return		0 always
 */
static inline
int rpmdsPrint(rpmds ds, FILE * fp)
{
    if (fp == NULL)
	fp = stderr;
    ds = rpmdsInit(ds);
    while (rpmdsNext(ds) >= 0)
        fprintf(fp, "%6d\t%s: %s\n", rpmdsIx(ds), rpmdsType(ds), rpmdsDNEVR(ds)+2);
    return 0;
}

/** \ingroup rpmds
 * Print current dependency set results.
 * @param ds		dependency set
 * @param fp		file handle (NULL uses stderr)
 * @return		0 always
 */
static inline
int rpmdsPrintResults(rpmds ds, FILE * fp)
{
    if (fp == NULL)
	fp = stderr;
    ds = rpmdsInit(ds);
    while (rpmdsNext(ds) >= 0) {
	rpmint32_t rc = rpmdsResult(ds);
	if (rc > 0)
	    continue;
        fprintf(fp, "%6d\t%s: %s\n", rpmdsIx(ds), rpmdsType(ds), rpmdsDNEVR(ds)+2);
    }
    return 0;
}

/** \ingroup rpmds
 * Check Provides: against Requires: and print closure results.
 * @param P		Provides: dependency set
 * @param R		Requires: dependency set
 * @param fp		file handle (NULL uses stderr)
 * @return		0 always
 */
static inline
int rpmdsPrintClosure(rpmds P, rpmds R, FILE * fp)
{
    int rc;

    /* Allocate the R results array (to be filled in by rpmdsSearch). */
    (void) rpmdsSetResult(R, 0);	/* allocate result array. */

    /* Collect the rpmdsSearch results (in the R dependency set). */
    R = rpmdsInit(R);
    while (rpmdsNext(R) >= 0)
        rc = rpmdsSearch(P, R);

    return rpmdsPrintResults(R, fp);
}
#endif
/*@}*/

#ifdef __cplusplus
}
#endif

#endif	/* H_RPMDS */
