#ifndef H_RPMFI
#define H_RPMFI

/** \ingroup rpmfi
 * \file lib/rpmfi.h
 * Structure(s) used for file info tag sets.
 */

#include <stddef.h>
#include <rpmdefs.h>
#include <rpmiotypes.h>
#include <rpmtypes.h>
#include <rpmtag.h>

extern int _rpmfi_debug;

/** \ingroup rpmfi
 * File types.
 * These are the file types used internally by rpm. The file
 * type is determined by applying stat(2) macros like S_ISDIR to
 * the file mode tag from a header. The values are arbitrary,
 * but are identical to the linux stat(2) file types.
 */
typedef enum rpmFileTypes_e {
    PIPE	=  1,	/*!< pipe/fifo */
    CDEV	=  2,	/*!< character device */
    XDIR	=  4,	/*!< directory */
    BDEV	=  6,	/*!< block device */
    REG		=  8,	/*!< regular file */
    LINK	= 10,	/*!< hard link */
    SOCK	= 12	/*!< socket */
} rpmFileTypes;

/**
 * File States (when installed).
 */
typedef enum rpmfileState_e {
    RPMFILE_STATE_NORMAL 	= 0,
    RPMFILE_STATE_REPLACED 	= 1,
    RPMFILE_STATE_NOTINSTALLED	= 2,
    RPMFILE_STATE_NETSHARED	= 3,
    RPMFILE_STATE_WRONGCOLOR	= 4
} rpmfileState;
#define	RPMFILE_STATE_MISSING	-1	/* XXX used for unavailable data */

/**
 * File Attributes.
 */
typedef	enum rpmfileAttrs_e {
    RPMFILE_NONE	= 0,
    RPMFILE_CONFIG	= (1 <<  0),	/*!< from %%config */
    RPMFILE_DOC		= (1 <<  1),	/*!< from %%doc */
    RPMFILE_ICON	= (1 <<  2),	/*!< from Icon: */
    RPMFILE_MISSINGOK	= (1 <<  3),	/*!< from %%config(missingok) */
    RPMFILE_NOREPLACE	= (1 <<  4),	/*!< from %%config(noreplace) */
    RPMFILE_SPECFILE	= (1 <<  5),	/*!< the specfile (srpm only). */
    RPMFILE_GHOST	= (1 <<  6),	/*!< from %%ghost */
    RPMFILE_LICENSE	= (1 <<  7),	/*!< from %%license */
    RPMFILE_README	= (1 <<  8),	/*!< from %%readme */
    RPMFILE_EXCLUDE	= (1 <<  9),	/*!< from %%exclude, internal */
    RPMFILE_UNPATCHED	= (1 << 10),	/*!< (deprecated) placeholder (SuSE) */
    RPMFILE_PUBKEY	= (1 << 11),	/*!< from %%pubkey */
    RPMFILE_POLICY	= (1 << 12),	/*!< from %%policy */
    RPMFILE_EXISTS	= (1 << 13),	/*!< did lstat(fn, st) succeed? */
    RPMFILE_SPARSE	= (1 << 14),	/*!< was ((512*st->st_blocks) < st->st_size) ? */
    RPMFILE_TYPED	= (1 << 15),	/*!< (unimplemented) from %%spook */
    RPMFILE_SOURCE	= (1 << 16),	/*!< from SourceN: (srpm only). */
    RPMFILE_PATCH	= (1 << 17),	/*!< from PatchN: (srpm only). */
    RPMFILE_OPTIONAL	= (1 << 18),	/*!< from %%optional. */
    RPMFILE_REMOVE	= (1 << 19)	/*!< remove file (after %post). */
} rpmfileAttrs;

#define	RPMFILE_SPOOK	(RPMFILE_GHOST|RPMFILE_TYPED)
#define	RPMFILE_ALL	~(RPMFILE_NONE)

/** \ingroup rpmfi
 * File info tag sets from a header, so that a header can be discarded early.
 */
typedef struct rpmfi_s * rpmfi;

#if defined(_RPMFI_INTERNAL)

#include "mire.h"

/** \ingroup rpmfi
 * A package filename set.
 */
struct rpmfi_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    int i;			/*!< Current file index. */
    int j;			/*!< Current directory index. */

    const char * Type;		/*!< Tag name. */

    rpmTag tagN;		/*!< Header tag. */
    Header h;			/*!< Header for file info set (or NULL) */

    const char ** bnl;		/*!< Base name(s) (from header) */
    const char ** dnl;		/*!< Directory name(s) (from header) */

    const char ** fdigests;	/*!< File digest(s) (from header) */
    rpmuint32_t * fdigestalgos;	/*!< File digest algorithm(s) (from header) */
    const char ** flinks;	/*!< File link(s) (from header) */
    const char ** flangs;	/*!< File lang(s) (from header) */

          rpmuint32_t * dil;	/*!< Directory indice(s) (from header) */
    const rpmuint32_t * fflags;	/*!< File flag(s) (from header) */
    const rpmuint32_t * fsizes;	/*!< File size(s) (from header) */
    const rpmuint32_t * fmtimes;/*!< File modification time(s) (from header) */
          rpmuint16_t * fmodes;	/*!< File mode(s) (from header) */
    const rpmuint16_t * frdevs;	/*!< File rdev(s) (from header) */
    const rpmuint32_t * finodes;/*!< File inodes(s) (from header) */

    const char ** fuser;	/*!< File owner(s) (from header) */
    const char ** fgroup;	/*!< File group(s) (from header) */

    rpmuint8_t * fstates;	/*!< File state(s) (from header) */

    const rpmuint32_t * fcolors;/*!< File color bits (header) */

    const char ** fcaps;	/*! File capabilities. */

    const char ** fcontexts;	/*! File security contexts. */

    const char ** cdict;	/*!< File class dictionary (header) */
    rpmuint32_t ncdict;		/*!< No. of class entries. */
    const rpmuint32_t * fcdictx;/*!< File class dictionary index (header) */

    const rpmuint32_t * ddict;	/*!< File depends dictionary (header) */
    rpmuint32_t nddict;		/*!< No. of depends entries. */
    const rpmuint32_t * fddictx;/*!< File depends dictionary start (header) */
    const rpmuint32_t * fddictn;/*!< File depends dictionary count (header) */

    const rpmuint32_t * vflags;	/*!< File verify flag(s) (from header) */

    rpmuint32_t dc;		/*!< No. of directories. */
    rpmuint32_t fc;		/*!< No. of files. */

/*=============================*/
    void * te;

    void * _fnbf;		/*!< File paths Bloom filter. */
    miRE exclude;		/*!< Iterator exclude patterns. */
    int nexclude;		/*!< No. of exclude patterns. */
    miRE include;		/*!< Iterator include patterns. */
    int ninclude;		/*!< No. of include patterns. */

/*-----------------------------*/
    uid_t uid;			/*!< File uid (default). */
    gid_t gid;			/*!< File gid (default). */
    rpmuint32_t flags;		/*!< File flags (default). */
    int action;			/*!< File disposition (default). */
    int * actions;		/*!< File disposition(s). */
    struct fingerPrint_s * fps;	/*!< File fingerprint(s). */
    const char ** obnl;		/*!< Original basename(s) (from header) */
    const char ** odnl;		/*!< Original dirname(s) (from header) */
    rpmuint32_t * odil;		/*!< Original dirindex(s) (from header) */

    unsigned char * digests;	/*!< File digest(s) in binary. */
    rpmuint32_t digestalgo;	/*!< File digest algorithm. */
    rpmuint32_t digestlen;	/*!< No. bytes in binary digest. */

    const char * pretrans;
    const char * pretransprog;
    const char * posttrans;
    const char * posttransprog;
    const char * verifyscript;
    const char * verifyscriptprog;

    char * fn;			/*!< File name buffer, fnlen + 1 bytes. */
    size_t fnlen;		/*!< Maximum file name length (without NUL). */

    size_t astriplen;
    size_t striplen;
    unsigned long long archivePos;
    unsigned long long archiveSize;
    mode_t dperms;		/*!< Directory perms (0755) if not mapped. */
    mode_t fperms;		/*!< File perms (0644) if not mapped. */
    const char ** apath;
    int mapflags;
    int * fmapflags;
    void * fsm;			/*!< File state machine data. */
    rpmuint32_t color;		/*!< Color bit(s) from file color union. */

    int isSource;		/*!< Is this a SRPM? */

    rpmuint32_t * replacedSizes;/*!< (TR_ADDED) */

    unsigned int record;	/*!< (TR_REMOVED) */
    int magic;
#define	RPMFIMAGIC	0x09697923
/*=============================*/
};

#endif	/* _RPMFI_INTERNAL */

#ifdef __cplusplus
extern "C" {
#endif

#if !defined(_RPMFI_NOMETHODS)
/** \name RPMFI */
/*@{*/

/**
 * Unreference a file info set instance.
 * @param fi		file info set
 * @param msg
 * @return		NULL on last dereference
 */
rpmfi rpmfiUnlink (rpmfi fi, const char * msg);
#define	rpmfiUnlink(_fi, _msg)	\
    ((rpmfi) rpmioUnlinkPoolItem((rpmioItem)(_fi), _msg, __FILE__, __LINE__))

/**
 * Reference a file info set instance.
 * @param fi		file info set
 * @param msg
 * @return		new file info set reference
 */
rpmfi rpmfiLink (rpmfi fi, const char * msg);
#define	rpmfiLink(_fi, _msg)	\
    ((rpmfi) rpmioLinkPoolItem((rpmioItem)(_fi), _msg, __FILE__, __LINE__))

/**
 * Return file count from file info set.
 * @param fi		file info set
 * @return		current file count
 */
int rpmfiFC(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Return current file index from file info set.
 * @param fi		file info set
 * @return		current file index
 */
int rpmfiFX(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Set current file index in file info set.
 * @param fi		file info set
 * @param fx		new file index
 * @return		current file index
 */
int rpmfiSetFX(rpmfi fi, int fx);

/**
 * Return directory count from file info set.
 * @param fi		file info set
 * @return		current directory count
 */
int rpmfiDC(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Return current directory index from file info set.
 * @param fi		file info set
 * @return		current directory index
 */
int rpmfiDX(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Set current directory index in file info set.
 * @param fi		file info set
 * @param dx		new directory index
 * @return		current directory index
 */
int rpmfiSetDX(rpmfi fi, int dx);

/**
 * Return source rpm marker from file info set.
 * @param fi		file info set
 * @return		source rpm?
 */
int rpmfiIsSource(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Return current base name from file info set.
 * @param fi		file info set
 * @return		current base name, NULL on invalid
 */
extern const char * rpmfiBN(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Return current directory name from file info set.
 * @param fi		file info set
 * @return		current directory, NULL on invalid
 */
extern const char * rpmfiDN(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Return current file name from file info set.
 * @param fi		file info set
 * @return		current file name
 */
extern const char * rpmfiFN(rpmfi fi);

/**
 * Return maximum file name length from file info set.
 * @param fi		file info set
 * @return		maximum file name length (not including NUL)
 */
extern size_t rpmfiFNMaxLen(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Return current file flags from file info set.
 * @param fi		file info set
 * @return		current file flags, 0 on invalid
 */
rpmuint32_t rpmfiFFlags(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Set current file flags in file info set.
 * @param fi		file info set
 * @param FFlags	new file flags
 * @return		previous file flags, 0 on invalid
 */
rpmuint32_t rpmfiSetFFlags(rpmfi fi, rpmuint32_t FFlags);

/**
 * Return current file verify flags from file info set.
 * @param fi		file info set
 * @return		current file verify flags, 0 on invalid
 */
rpmuint32_t rpmfiVFlags(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Set current file verify flags in file info set.
 * @param fi		file info set
 * @param VFlags	new file verify flags
 * @return		previous file verify flags, 0 on invalid
 */
rpmuint32_t rpmfiSetVFlags(rpmfi fi, rpmuint32_t VFlags);

/**
 * Return current file mode from file info set.
 * @param fi		file info set
 * @return		current file mode, 0 on invalid
 */
rpmuint16_t rpmfiFMode(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Return current file state from file info set.
 * @param fi		file info set
 * @return		current file state, 0 on invalid
 */
rpmfileState rpmfiFState(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Set current file state in file info set.
 * @param fi		file info set
 * @param fstate	new file state
 * @return		previous file state, 0 on invalid
 */
rpmfileState rpmfiSetFState(rpmfi fi, rpmfileState fstate);

/**
 * Return current file (binary) digest from file info set.
 * @param fi		file info set
 * @retval *algop	digest algorithm
 * @retval *lenp	digest length (in bytes)
 * @return		current file digest, NULL on invalid
 */
extern const unsigned char * rpmfiDigest(rpmfi fi,
		int * algop, size_t * lenp);

/**
 * Return current file linkto (i.e. symlink(2) target) from file info set.
 * @param fi		file info set
 * @return		current file linkto, NULL on invalid
 */
extern const char * rpmfiFLink(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Return current file size from file info set.
 * @param fi		file info set
 * @return		current file size, 0 on invalid
 */
rpmuint32_t rpmfiFSize(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Return current file rdev from file info set.
 * @param fi		file info set
 * @return		current file rdev, 0 on invalid
 */
rpmuint16_t rpmfiFRdev(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Return current file inode from file info set.
 * @param fi		file info set
 * @return		current file inode, 0 on invalid
 */
rpmuint32_t rpmfiFInode(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Return union of all file color bits from file info set.
 * @param fi		file info set
 * @return		current color
 */
rpmuint32_t rpmfiColor(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Return current file color bits from file info set.
 * @param fi		file info set
 * @return		current file color
 */
rpmuint32_t rpmfiFColor(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Return current file class from file info set.
 * @param fi		file info set
 * @return		current file class, 0 on invalid
 */
extern const char * rpmfiFClass(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Return current file security context from file info set.
 * @param fi		file info set
 * @return		current file context, 0 on invalid
 */
extern const char * rpmfiFContext(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Return current file depends dictionary from file info set.
 * @param fi		file info set
 * @retval *fddictp	file depends dictionary array (or NULL)
 * @return		no. of file depends entries, 0 on invalid
 */
rpmuint32_t rpmfiFDepends(rpmfi fi, const rpmuint32_t ** fddictp);

/**
 * Return (calculated) current file nlink count from file info set.
 * @param fi		file info set
 * @return		current file nlink count, 0 on invalid
 */
rpmuint32_t rpmfiFNlink(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Return current file modify time from file info set.
 * @param fi		file info set
 * @return		current file modify time, 0 on invalid
 */
rpmuint32_t rpmfiFMtime(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Return current file owner from file info set.
 * @param fi		file info set
 * @return		current file owner, NULL on invalid
 */
extern const char * rpmfiFUser(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Return current file group from file info set.
 * @param fi		file info set
 * @return		current file group, NULL on invalid
 */
extern const char * rpmfiFGroup(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Return current file cpabilities from file info set.
 * @param fi		file info set
 * @return		current file cpabilities, "" for none, NULL on invalid
 */
extern const char * rpmfiFCaps(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Return current file language(s) from file info set.
 * @param fi		file info set
 * @return		current file langs, NULL on invalid
 */
extern const char * rpmfiFLangs(rpmfi fi)
	RPM_GNUC_PURE;

/** \ingroup rpmfi
 * Return file path Bloom filter.
 * @param fi		file info set
 * @return		file path Bloom filter (NULL if not set)
 */
void * rpmfiBloomFN(const rpmfi fi)
	RPM_GNUC_PURE;

/** \ingroup rpmfi
 * Return file info exclude patterns.
 * @param fi		file info set
 * @return		file info exclude patterns (NULL if not set)
 */
void * rpmfiExclude(const rpmfi fi)
	RPM_GNUC_PURE;

/** \ingroup rpmfi
 * Return no. of file info exclude patterns.
 * @param fi		file info set
 * @return		file info exclude patterns (0 if not set)
 */
int rpmfiNExclude(const rpmfi fi)
	RPM_GNUC_PURE;

/** \ingroup rpmfi
 * Return file info include patterns.
 * @param fi		file info set
 * @return		file info include patterns (NULL if not set)
 */
void * rpmfiInclude(const rpmfi fi)
	RPM_GNUC_PURE;

/** \ingroup rpmfi
 * Return no. of file info include patterns.
 * @param fi		file info set
 * @return		file info include patterns (0 if not set)
 */
int rpmfiNInclude(const rpmfi fi)
	RPM_GNUC_PURE;

#if defined(_RPMFI_INTERNAL)
#include "fprint.h"
/**
 * Return fingerprint from file info set.
 * @param fi		file info set
 * @param ix
 * @return		fingerprint, NULL on invalid
 */
struct fingerPrint_s * rpmfiFpsIndex(rpmfi fi, int ix)
	RPM_GNUC_PURE;

void rpmfiFpLookup(rpmfi fi, fingerPrintCache fpc);
#endif

/**
 * Return FN Bloom filter from file info set.
 * @param fi		file info set
 * @return		FN Bloom filter, NULL on invalid
 */
void * rpmfiFNBF(rpmfi fi);

/**
 * Return next file iterator index.
 * @param fi		file info set
 * @return		file iterator index, -1 on termination
 */
int rpmfiNext(rpmfi fi);

/**
 * Initialize file iterator index.
 * @param fi		file info set
 * @param fx		file iterator index
 * @return		file info set
 */
rpmfi rpmfiInit(rpmfi fi, int fx);

/**
 * Return next directory iterator index.
 * @param fi		file info set
 * @return		directory iterator index, -1 on termination
 */
int rpmfiNextD(rpmfi fi);

/**
 * Initialize directory iterator index.
 * @param fi		file info set
 * @param dx		directory iterator index
 * @return		file info set, NULL if dx is out of range
 */
rpmfi rpmfiInitD(rpmfi fi, int dx);

/**
 * Link a header to a file info set.
 * @param fi		file info set
 * @param h		header
 * @return		0 always
 */
int rpmfiSetHeader(rpmfi fi, Header h);

/**
 * Destroy a file info set.
 * @param fi		file info set
 * @return		NULL on last dereference
 */
rpmfi rpmfiFree(rpmfi fi);
#define	rpmfiFree(_fi)	\
    ((rpmfi) rpmioFreePoolItem((rpmioItem)(_fi), __FUNCTION__, __FILE__, __LINE__))

/**
 * Create and load a file info set.
 * @param _ts		transaction set (NULL skips path relocation)
 * @param h		header
 * @param tagN		RPMTAG_BASENAMES
 * @param flags		scareMem(0x1), nofilter(0x2)
 * @return		new file info set
 */
rpmfi rpmfiNew(const void * _ts, Header h, rpmTag tagN, int flags);

/**
 * Return current stat(2) data from file info set.
 * @param fi		file info set
 * @retval *st		stat(2) data
 * @return		-1 on error, 0 on success
 */
struct stat;
int rpmfiFStat(rpmfi fi, struct stat * st);

/**
 * Return lstat(2) data of path from file info set.
 * @param fi		file info set
 * @param path		file path
 * @retval *st		stat(2) data
 * @return		-1 on error, 0 on success
 */
int rpmfiStat(rpmfi fi, const char * path, struct stat * st);

/**
 * Return directory stream onto file info set.
 * @param fi		file info set
 * @param name		directory path
 * @return		NULL on error
 */
void * rpmfiOpendir(rpmfi fi, const char * name);

/**
 * Retrieve file classes from header.
 *
 * This function is used to retrieve file classes from the header.
 * 
 * @param h		header
 * @retval *fclassp	array of file classes
 * @retval *fcp		number of files
 */
void rpmfiBuildFClasses(Header h,
		const char *** fclassp, rpmuint32_t * fcp);

#ifdef	DYING
/**
 * Retrieve file security contexts from header.
 *
 * This function is used to retrieve file contexts from the header.
 * 
 * @param h		header
 * @retval *fcontextp	array of file contexts
 * @retval *fcp		number of files
 */
void rpmfiBuildFContexts(Header h,
		const char *** fcontextp, rpmuint32_t * fcp);

/**
 * Retrieve file security contexts from file system.
 *
 * This function is used to retrieve file contexts from the file system.
 * 
 * @param h		header
 * @retval *fcontextp	array of file contexts
 * @retval *fcp		number of files
 */
void rpmfiBuildFSContexts(Header h,
		const char *** fcontextp, rpmuint32_t * fcp);

/**
 * Retrieve file security contexts from policy RE's.
 *
 * This function is used to retrieve file contexts from policy RE's.
 * 
 * @param h		header
 * @retval *fcontextp	array of file contexts
 * @retval *fcp		number of files
 */
void rpmfiBuildREContexts(Header h,
		const char *** fcontextp, rpmuint32_t * fcp);
#endif

/**
 * Retrieve per-file dependencies from header.
 *
 * This function is used to retrieve per-file dependencies from the header.
 * 
 * @param h		header
 * @param tagN		RPMTAG_PROVIDENAME | RPMTAG_REQUIRENAME
 * @retval *fdepsp	array of file dependencies
 * @retval *fcp		number of files
 */
void rpmfiBuildFDeps(Header h, rpmTag tagN,
		const char *** fdepsp, rpmuint32_t * fcp);

/**
 * Return file info comparison.
 * @param afi		1st file info
 * @param bfi		2nd file info
 * @return		0 if identical
 */
int rpmfiCompare(const rpmfi afi, const rpmfi bfi);

/**
 * Return file disposition.
 * @param ofi		old file info
 * @param nfi		new file info
 * @param skipMissing	OK to skip missing files?
 * @return		file disposition
 */
int rpmfiDecideFate(const rpmfi ofi, rpmfi nfi, int skipMissing);

/**
 * Return formatted string representation of package disposition.
 * @param fi		file info set
 * @return		formatted string
 */
const char * rpmfiTypeString(rpmfi fi)
	RPM_GNUC_PURE;

/**
 * Free relocation array.
 * @param relocs	relocation array
 * @return		NULL always
 */
rpmRelocation rpmfiFreeRelocations(rpmRelocation relocs);

/**
 * Duplicate a relocation array.
 * @param relocs	relocation array
 * @retval *nrelocsp	no. of elements
 * @return		duplicate relocation array
 */
rpmRelocation rpmfiDupeRelocations(rpmRelocation relocs, int * nrelocsp);

/**
 * Add relocation element to array.
 * @retval *relp	relocation array
 * @retval *nrelp	no. of elements
 * @param oldPath	old path
 * @param newPath	new path
 * @return		0 on success
 */
int rpmfiAddRelocation(rpmRelocation * relp, int * nrelp,
		const char * oldPath, const char * newPath);

/*@}*/
#endif	/* _RPMFI_NOMETHODS */

#ifdef __cplusplus
}
#endif

#endif	/* H_RPMFI */
