#ifndef	H_RPMGI
#define	H_RPMGI

/** \ingroup rpmgi
 * \file lib/rpmgi.h
 */

#include <stddef.h>
#include <rpmdefs.h>
#include <rpmiotypes.h>
#include <argv.h>

#include <rpmtypes.h>
#include <rpmtag.h>

#include <fts.h>
#include <rpmds.h>
#include <rpmte.h>
#include <rpmts.h>

/**
 */
extern int _rpmgi_debug;

/**
 */
typedef enum rpmgiFlags_e {
    RPMGI_NONE		= 0,
    RPMGI_TSADD		= (1 << 0),
    RPMGI_TSORDER	= (1 << 1),
    RPMGI_NOGLOB	= (1 << 2),
    RPMGI_NOMANIFEST	= (1 << 3),
    RPMGI_NOHEADER	= (1 << 4),
    RPMGI_ERASING	= (1 << 5)
} rpmgiFlags;

/**
 */
extern rpmgiFlags giFlags;

#if defined(_RPMGI_INTERNAL)
/** \ingroup rpmgi
 */
struct rpmgi_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    rpmts ts;			/*!< Iterator transaction set. */
    int (*tsOrder) (rpmts ts);	/*!< Iterator transaction ordering. */
    rpmTag tag;			/*!< Iterator type. */
    const void * keyp;		/*!< Iterator key. */
    size_t keylen;		/*!< Iterator key length. */

    rpmgiFlags flags;		/*!< Iterator control bits. */
    int active;			/*!< Iterator is active? */
    int i;			/*!< Element index. */
    const char * hdrPath;	/*!< Path to current iterator header. */
    Header h;			/*!< Current iterator header. */
    int rc;			/*!< Current exit code, specific to item. */

    rpmtsi tsi;

    rpmmi mi;

    FD_t fd;

    ARGV_t argv;
    int argc;

    int ftsOpts;
    FTS * ftsp;
    FTSENT * fts;
    rpmRC (*walkPathFilter) (rpmgi gi);
    rpmRC (*stash) (rpmgi gi, Header h);

};
#endif

#ifdef __cplusplus
extern "C" {
#endif

/** \name RPMGI */
/*@{*/

/**
 * Unreference a generalized iterator instance.
 * @param gi		generalized iterator
 * @param msg
 * @return		NULL on last dereference
 */
rpmgi rpmgiUnlink (rpmgi gi, const char * msg);
#define	rpmgiUnlink(_gi, _msg)	\
    ((rpmgi)rpmioUnlinkPoolItem((rpmioItem)(_gi), _msg, __FILE__, __LINE__))

/**
 * Reference a generalized iterator instance.
 * @param gi		generalized iterator
 * @param msg
 * @return		new generalized iterator reference
 */
rpmgi rpmgiLink (rpmgi gi, const char * msg);
#define	rpmgiLink(_gi, _msg)	\
    ((rpmgi)rpmioLinkPoolItem((rpmioItem)(_gi), _msg, __FILE__, __LINE__))

/** Destroy a generalized iterator.
 * @param gi		generalized iterator
 * @return		NULL on last dereference
 */
rpmgi rpmgiFree(rpmgi gi);
#define	rpmgiFree(_gi)	\
    ((rpmgi)rpmioFreePoolItem((rpmioItem)(_gi), __FUNCTION__, __FILE__, __LINE__))

/**
 * Return a generalized iterator.
 * @param ts		transaction set
 * @param tag		rpm tag
 * @param keyp		key data (NULL for sequential access)
 * @param keylen	key data length (0 will use strlen(keyp))
 * @return		new iterator
 */
rpmgi rpmgiNew(rpmts ts, int tag, const void * keyp, size_t keylen);

/**
 * Perform next iteration step.
 * @param gi		generalized iterator
 * @return		RPMRC_OK on success, RPMRC_NOTFOUND on EOI
 */
rpmRC rpmgiNext(rpmgi gi);

/**
 * Return current iteration flags.
 * @param gi		generalized iterator
 * @return		flags
 */
rpmgiFlags rpmgiGetFlags(rpmgi gi)
	RPM_GNUC_PURE;

/**
 * Return current header path.
 * @param gi		generalized iterator
 * @return		header path
 */
const char * rpmgiHdrPath(rpmgi gi)
	RPM_GNUC_PURE;

/**
 * Return current iteration header.
 * @param gi		generalized iterator
 * @return		header
 */
Header rpmgiHeader(rpmgi gi)
	RPM_GNUC_PURE;

/**
 * Return current iteration transaction set.
 * @param gi		generalized iterator
 * @return		transaction set
 */
rpmts rpmgiTs(rpmgi gi)
	RPM_GNUC_PURE;

/**
 * Return current iteration item(s) exit code.
 * @param gi		generalized iterator
 * @return		transaction set
 */
int rpmgiRc(rpmgi gi)
	RPM_GNUC_PURE;

/**
 * Escape isspace(3) characters in string.
 * @param s		string
 * @return		escaped string
 */
const char * rpmgiEscapeSpaces(const char * s);

/**
 * Load iterator args.
 * @param gi		generalized iterator
 * @param argv		arg list
 * @param ftsOpts	fts(3) flags
 * @param flags		iterator flags
 * @return		RPMRC_OK on success
 */
rpmRC rpmgiSetArgs(rpmgi gi, ARGV_t argv, int ftsOpts, rpmgiFlags flags);

/**
 * Return header from package.
 * @param gi		generalized iterator
 * @param path		file path
 * @return		header (NULL on failure)
 */
Header rpmgiReadHeader(rpmgi gi, const char * path);

/*@}*/

#ifdef __cplusplus
}
#endif

#endif	/* H_RPMGI */
