#ifndef H_RPMDB
#define H_RPMDB

/** \ingroup rpmdb dbi db1 db3
 * \file rpmdb/rpmdb.h
 * Access RPM indices using Berkeley DB interface(s).
 */

#include <assert.h>	/* XXX assert() within wrapper */

#include <mire.h>
#include <stddef.h>
#include <stdint.h>
#include <sys/types.h>

#include <rpmtypes.h>
#include <rpmtag.h>	/* XXX Header typedef */

#if defined(_RPMDB_INTERNAL)
#if defined(WITH_DB)
#include <db.h>
#else
#include "db_emu.h"
#endif
#endif

#if defined(_RPMDB_INTERNAL)
#define DBT_INIT {0} 	/* -Wno-missing-field-initializers */
#endif

extern int _rpmdb_debug;

extern int _rpmmi_debug;

#ifdef	NOTYET
/** \ingroup rpmdb
 * Database of headers and tag value indices.
 */
typedef struct rpmdb_s * rpmdb;

/** \ingroup rpmdb
 * Database iterator.
 */
typedef struct rpmmi_s * rpmmi;
#endif

/**
 */
typedef struct _dbiIndexItem * rpmdbiIndexItem;

/** \ingroup rpmdb
 * A single element (i.e. inverted list from tag values) of a database.
 */
typedef struct _dbiIndexSet * rpmdbiIndexSet;

/**
 */
typedef struct rpmdbi_s * rpmdbi;

#if defined(_RPMDB_INTERNAL)
#include <rpmio.h>
#include <rpmbf.h>
#include <rpmsw.h>

#if !defined(SWIG)	/* XXX inline rpmdbiFoo() need */
/** \ingroup rpmdbi
 * A single item from an index database (i.e. the "data returned").
 */
struct _dbiIndexItem {
    uint32_t hdrNum;			/*!< header instance in db */
    uint32_t tagNum;			/*!< tag index in header */
};

/** \ingroup rpmdbi
 * Items retrieved from the index database.
 */
struct _dbiIndexSet {
    struct _dbiIndexItem * recs;	/*!< array of records */
    unsigned int count;			/*!< number of records */
};

/** \ingroup rpmdbi
 * Private methods for accessing an index database.
 */
struct _dbiVec {
    const char * dbv_version;		/*!< DB version string */
    int dbv_major;			/*!< DB version major */
    int dbv_minor;			/*!< DB version minor */
    int dbv_patch;			/*!< DB version patch */

/** \ingroup rpmdbi
 * Return handle for an index database.
 * @param rpmdb		rpm database
 * @param tag		rpm tag
 * @return		0 on success
 */
    int (*open) (rpmdb rpmdb, rpmTag tag, rpmdbi * dbip);

/** \ingroup rpmdbi
 * Close index database, and destroy database handle.
 * @param dbi		index database handle
 * @param flags		(unused)
 * @return		0 on success
 */
    int (*close) (rpmdbi dbi, unsigned int flags);

/** \ingroup rpmdbi
 * Flush pending operations to disk.
 * @param dbi		index database handle
 * @param flags		(unused)
 * @return		0 on success
 */
    int (*sync) (rpmdbi dbi, unsigned int flags);

/** \ingroup rpmdbi
 * Associate secondary database with primary.
 * @param dbi		index database handle
 * @param dbisecondary	secondary index database handle
 * @param callback	create secondary key from primary (NULL if DB_RDONLY)
 * @param flags		DB_CREATE or 0
 * @return		0 on success
 */
    int (*associate) (rpmdbi dbi, rpmdbi dbisecondary,
                int (*callback) (DB *, const DBT *, const DBT *, DBT *),
                unsigned int flags);

/** \ingroup rpmdbi
 * Associate foreign secondary database with primary.
 * @param dbi		index database handle
 * @param dbisecondary	secondary index database handle
 * @param callback	create secondary key from primary (NULL if DB_RDONLY)
 * @param flags		DB_CREATE or 0
 * @return		0 on success
 */
    int (*associate_foreign) (rpmdbi dbi, rpmdbi dbisecondary,
                int (*callback) (DB *, const DBT *, DBT *, const DBT *, int *),
                unsigned int flags);

/** \ingroup rpmdbi
 * Return join cursor for list of cursors.
 * @param dbi		index database handle
 * @param curslist	NULL terminated list of database cursors
 * @retval dbcp		address of join database cursor
 * @param flags		DB_JOIN_NOSORT or 0
 * @return		0 on success
 */
    int (*join) (rpmdbi dbi, DBC ** curslist, DBC ** dbcp,
                unsigned int flags);

/** \ingroup rpmdbi
 * Return whether key exists in a database.
 * @param dbi		index database handle
 * @param key		retrieve key value/length/flags
 * @param flags		usually 0
 * @return		0 if key exists, DB_NOTFOUND if not, else error
 */
    int (*exists) (rpmdbi dbi, DBT * key, unsigned int flags);

/** \ingroup rpmdbi
 * Return next sequence number.
 * @param dbi		index database handle (with attached sequence)
 * @retval *seqnop	IN: delta (0 does seqno++) OUT: returned 64bit seqno
 * @param flags		usually 0
 * @return		0 on success
 */
    int (*seqno) (rpmdbi dbi, int64_t * seqnop, unsigned int flags);

/** \ingroup rpmdbi
 * Open database cursor.
 * @param dbi		index database handle
 * @param txnid		database transaction handle
 * @retval dbcp		address of new database cursor
 * @param dbiflags	DB_WRITECURSOR or 0
 * @return		0 on success
 */
    int (*copen) (rpmdbi dbi, DB_TXN * txnid,
			DBC ** dbcp, unsigned int dbiflags);

/** \ingroup rpmdbi
 * Close database cursor.
 * @param dbi		index database handle
 * @param dbcursor	database cursor
 * @param flags		(unused)
 * @return		0 on success
 */
    int (*cclose) (rpmdbi dbi, DBC * dbcursor, unsigned int flags);

/** \ingroup rpmdbi
 * Duplicate a database cursor.
 * @param dbi		index database handle
 * @param dbcursor	database cursor
 * @retval dbcp		address of new database cursor
 * @param flags		DB_POSITION for same position, 0 for uninitialized
 * @return		0 on success
 */
    int (*cdup) (rpmdbi dbi, DBC * dbcursor, DBC ** dbcp, unsigned int flags);

/** \ingroup rpmdbi
 * Delete (key,data) pair(s) using db->del or dbcursor->c_del.
 * @param dbi		index database handle
 * @param dbcursor	database cursor (NULL will use db->del)
 * @param key		delete key value/length/flags
 * @param data		delete data value/length/flags
 * @param flags		(unused)
 * @return		0 on success
 */
    int (*cdel) (rpmdbi dbi, DBC * dbcursor, DBT * key, DBT * data,
			unsigned int flags);

/** \ingroup rpmdbi
 * Retrieve (key,data) pair using db->get or dbcursor->c_get.
 * @param dbi		index database handle
 * @param dbcursor	database cursor (NULL will use db->get)
 * @param key		retrieve key value/length/flags
 * @param data		retrieve data value/length/flags
 * @param flags		(unused)
 * @return		0 on success
 */
    int (*cget) (rpmdbi dbi, DBC * dbcursor, DBT * key, DBT * data,
			unsigned int flags);

/** \ingroup rpmdbi
 * Retrieve (key,data) pair using dbcursor->c_pget.
 * @param dbi		index database handle
 * @param dbcursor	database cursor
 * @param key		secondary retrieve key value/length/flags
 * @param pkey		primary retrieve key value/length/flags
 * @param data		primary retrieve data value/length/flags
 * @param flags		DB_NEXT, DB_SET, or 0
 * @return		0 on success
 */
    int (*cpget) (rpmdbi dbi, DBC * dbcursor,
		DBT * key, DBT * pkey, DBT * data, unsigned int flags);

/** \ingroup rpmdbi
 * Store (key,data) pair using db->put or dbcursor->c_put.
 * @param dbi		index database handle
 * @param dbcursor	database cursor (NULL will use db->put)
 * @param key		store key value/length/flags
 * @param data		store data value/length/flags
 * @param flags		(unused)
 * @return		0 on success
 */
    int (*cput) (rpmdbi dbi, DBC * dbcursor, DBT * key, DBT * data,
			unsigned int flags);

/** \ingroup rpmdbi
 * Retrieve count of (possible) duplicate items using dbcursor->c_count.
 * @param dbi		index database handle
 * @param dbcursor	database cursor
 * @param countp	address of count
 * @param flags		(unused)
 * @return		0 on success
 */
    int (*ccount) (rpmdbi dbi, DBC * dbcursor, unsigned int * countp,
			unsigned int flags);

/** \ingroup rpmdbi
 * Is database byte swapped?
 * @param dbi		index database handle
 * @return		0 no
 */
    int (*byteswapped) (rpmdbi dbi);

/** \ingroup rpmdbi
 * Save statistics in database handle.
 * @param dbi		index database handle
 * @param flags		retrieve statistics that don't require traversal?
 * @return		0 on success
 */
    int (*stat) (rpmdbi dbi, unsigned int flags);
};

/** \ingroup rpmdbi
 * Describes an index database (implemented on Berkeley db3 functionality).
 */
struct rpmdbi_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    const char * dbi_root;	/*!< chroot(2) component of path */
    const char * dbi_home;	/*!< directory component of path */
    const char * dbi_file;	/*!< file component of path */
    const char * dbi_subfile;
    const char * dbi_tmpdir;	/*!< temporary directory */

    int	dbi_ecflags;		/*!< db_env_create flags */
    int	dbi_cflags;		/*!< db_create flags */
    int	dbi_oeflags;		/*!< common (db,dbenv)->open flags */
    int	dbi_eflags;		/*!< dbenv->open flags */
    int	dbi_oflags;		/*!< db->open flags */
    int	dbi_tflags;		/*!< dbenv->txn_begin flags */

    int	dbi_type;		/*!< db index type */
    unsigned dbi_mode;		/*!< mode to use on open */
    int	dbi_perms;		/*!< file permission used when creating */
    long dbi_shmkey;		/*!< shared memory base key */
    int	dbi_api;		/*!< Berkeley API type */

    int	dbi_use_dbenv;		/*!< use db environment? */
    int	dbi_no_fsync;		/*!< no-op fsync for db */
    int	dbi_no_dbsync;		/*!< don't call dbiSync */
    int	dbi_lockdbfd;		/*!< do fcntl lock on db fd */
    int	dbi_temporary;		/*!< non-persistent index/table */
    int	dbi_debug;

    rpmbf dbi_bf;

    char * dbi_host;
    unsigned long dbi_cl_timeout;
    unsigned long dbi_sv_timeout;

	/* dbenv parameters */
    int	dbi_lorder;
    void (*db_errcall) (void * dbenv, const char *db_errpfx, char *buffer);
    FILE *	dbi_errfile;
    const char * dbi_errpfx;
    int	dbi_region_init;
    unsigned int dbi_thread_count;

	/* locking sub-system parameters */
	/* logging sub-system parameters */
	/* mpool sub-system parameters */
	/* mutex sub-system parameters */
	/* replication sub-system parameters */

	/* sequences sub-system parameters */
    const char * dbi_seq_id;
    unsigned int dbi_seq_cachesize;
    unsigned int dbi_seq_flags;
    int64_t dbi_seq_initial;
    int64_t dbi_seq_min;
    int64_t dbi_seq_max;
    int64_t dbi_seqno;

	/* transaction sub-system parameters */
#if 0
    int	(*dbi_tx_recover) (DB_ENV *dbenv, DBT *log_rec,
				DB_LSN *lsnp, int redo, void *info);
#endif
	/* dbinfo parameters */
    int	dbi_pagesize;		/*!< (fs blksize) */
    int	dbi_heapsize;		/*!< DB_HEAP */
    int	dbi_heap_regionsize;	/*!< DB_HEAP */
    void * (*dbi_malloc) (size_t nbytes);
	/* hash access parameters */
    unsigned int dbi_h_ffactor;	/*!< */
    unsigned int (*dbi_h_hash_fcn) (DB *, const void *bytes,
				unsigned int length);
    unsigned int dbi_h_nelem;	/*!< */
    unsigned int dbi_h_flags;	/*!< DB_DUP, DB_DUPSORT */
    int (*dbi_h_dup_compare_fcn) (DB *, const DBT *, const DBT *, size_t *);
	/* btree access parameters */
    int	dbi_bt_flags;
    int	dbi_bt_minkey;
    int	(*dbi_bt_compare_fcn) (DB *, const DBT *, const DBT *, size_t *);
    int	(*dbi_bt_dup_compare_fcn) (DB *, const DBT *, const DBT *, size_t *);
    size_t (*dbi_bt_prefix_fcn) (DB *, const DBT *, const DBT *);
	/* recno access parameters */
    int	dbi_re_flags;
    int	dbi_re_delim;
    unsigned int dbi_re_len;
    int	dbi_re_pad;
    const char * dbi_re_source;
	/* queue access parameters */
    unsigned int dbi_q_extentsize;

    const char * dbi_primary;	/*!< Primary table for secondary index. */
    const char * dbi_foreign;

    rpmdb dbi_rpmdb;		/*!< the parent rpm database */
    rpmTag dbi_rpmtag;		/*!< rpm tag used for index */
    uint32_t dbi_hdrnum;	/*!< current hdrnum (rebuilddb) */
    size_t dbi_jlen;		/*!< size of join key */

    int dbi_table_nkeys;	/*!< Sqlite3: COUNT(key) */
    int dbi_table_exists;	/*!< Sqlite3: table exists? */

    void * dbi_seq;		/*!< Berkeley DB_SEQUENCE handle */
    void * dbi_db;		/*!< Berkeley DB handle */
    void * dbi_txnid;		/*!< Berkeley DB_TXN handle */
    void * dbi_stats;		/*!< Berkeley DB statistics */

    const struct _dbiVec * dbi_vec;	/*!< private methods */

};
#endif	/* !defined(SWIG) */

/** \ingroup rpmdb
 * Describes the collection of index databases used by rpm.
 */
struct rpmdb_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    const char * db_root;	/*!< rpmdb path prefix */
    const char * db_home;	/*!< rpmdb directory path */
    int		db_flags;
    int		db_mode;	/*!< rpmdb open mode */
    int		db_perms;	/*!< rpmdb open permissions */
    int		db_api;		/*!< Berkeley API type */
    const char * db_errpfx;	/*!< Berkeley DB error msg prefix. */

    int		db_remove_env;	/*!< Discard dbenv on close? */
    uint32_t	db_maxkey;	/*!< Max. primary key. */

    int		db_chrootDone;	/*!< If chroot(2) done, ignore db_root. */
    void (*db_errcall) (const char * db_errpfx, char * buffer);
    FILE *	db_errfile;	/*!< Berkeley DB stderr clone. */
    void * (*db_malloc) (size_t nbytes);
    void * (*db_realloc) (void * ptr, size_t nbytes);
    void (*db_free) (void * ptr);

    int	(*db_export) (rpmdb db, Header h, int adding);

    Header db_h;		/*!< Currently active header */

    rpmdb	db_next;	/*!< Chain of rpmdbOpen'ed rpmdb's. */
    int		db_opens;	/*!< No. of opens for this rpmdb. */

    void *	db_dbenv;	/*!< Berkeley DB_ENV handle. */
    void *	db_txn;		/*!< Berkeley DB_TXN handle */
    void *	db_logc;	/*!< Berkeley DB_LOGC handle */
    void *	db_mpf;		/*!< Berkeley DB_MPOOLFILE handle */

    tagStore_t	db_tags;	/*!< Tag name/value mappings. */
    size_t	db_ndbi;	/*!< No. of tag indices. */
    rpmdbi * _dbi;		/*!< Tag indices. */

    struct rpmop_s db_getops;	/*!< dbiGet statistics. */
    struct rpmop_s db_putops;	/*!< dbiPut statistics. */
    struct rpmop_s db_delops;	/*!< dbiDel statistics. */

};
#endif	/* defined(_RPMDB_INTERNAL) */

#ifdef __cplusplus
extern "C" {
#endif

#if defined(_RPMDB_INTERNAL)
#if defined(WITH_DB) || defined(WITH_SQLITE)
/** \ingroup db3
 * Return new configured index database handle instance.
 * @param rpmdb		rpm database
 * @param tag		rpm tag
 * @return		index database handle
 */
rpmdbi db3New(rpmdb rpmdb, rpmTag tag);

/** \ingroup db3
 * Destroy index database handle instance.
 * @param dbi		index database handle
 * @return		NULL always
 */
rpmdbi db3Free(rpmdbi dbi);
#define	db3Free(_dbi)	\
    ((rpmdbi)rpmioFreePoolItem((rpmioItem)(_dbi), __FUNCTION__, __FILE__, __LINE__))

/** \ingroup db3
 * Format db3 open flags for debugging print.
 * @param dbflags		db open flags
 * @param print_dbenv_flags	format db env flags instead?
 * @return			formatted flags (static buffer)
 */
extern const char * prDbiOpenFlags(int dbflags, int print_dbenv_flags);
#endif

/** \ingroup rpmdbi
 * Return handle for an index database.
 * @param db		rpm database
 * @param tag		rpm tag
 * @param flags		(unused)
 * @return		index database handle
 */
rpmdbi rpmdbiOpen(rpmdb db, rpmTag tag, unsigned int flags);

/**
 * Return dbiStats accumulator structure.
 * @param dbi		index database handle
 * @param opx		per-rpmdb accumulator index (aka rpmtsOpX)
 * @return		per-rpmdb accumulator pointer
 */
void * rpmdbiStatsAccumulator(rpmdbi dbi, int opx)
	RPM_GNUC_PURE;

#if !defined(SWIG)
/** \ingroup rpmdbi
 * Open a database cursor.
 * @param dbi		index database handle
 * @param txnid		database transaction handle
 * @retval dbcp		returned database cursor
 * @param flags		DB_WRITECURSOR if writing, or 0
 * @return		0 on success
 */
static inline
int rpmdbiCopen(rpmdbi dbi, DB_TXN * txnid, DBC ** dbcp, unsigned int flags)
{
    return (*dbi->dbi_vec->copen) (dbi, txnid, dbcp, flags);
}

/** \ingroup rpmdbi
 * Close a database cursor.
 * @param dbi		index database handle
 * @param dbcursor	database cursor
 * @param flags		(unused)
 * @return		0 on success
 */
static inline
int rpmdbiCclose(rpmdbi dbi, DBC * dbcursor, unsigned int flags)
{
    return (*dbi->dbi_vec->cclose) (dbi, dbcursor, flags);
}

/** \ingroup rpmdbi
 * Duplicate a database cursor.
 * @param dbi		index database handle
 * @param dbcursor	database cursor
 * @retval dbcp		address of new database cursor
 * @param flags		DB_POSITION for same position, 0 for uninitialized
 * @return		0 on success
 */
static inline
int rpmdbiCdup(rpmdbi dbi, DBC * dbcursor, DBC ** dbcp, unsigned int flags)
{
    return (*dbi->dbi_vec->cdup) (dbi, dbcursor, dbcp, flags);
}

/** \ingroup rpmdbi
 * Delete (key,data) pair(s) from index database.
 * @param dbi		index database handle
 * @param dbcursor	database cursor (NULL will use db->del)
 * @param key		delete key value/length/flags
 * @param data		delete data value/length/flags
 * @param flags		(unused)
 * @return		0 on success
 */
static inline
int rpmdbiDel(rpmdbi dbi, DBC * dbcursor, DBT * key, DBT * data,
		unsigned int flags)
{
    rpmop sw = (rpmop)rpmdbiStatsAccumulator(dbi, 16);	/* RPMTS_OP_DBDEL */
    int rc;
    assert(key->data != NULL && key->size > 0);
    (void) rpmswEnter(sw, 0);
    rc = (dbi->dbi_vec->cdel) (dbi, dbcursor, key, data, flags);
    (void) rpmswExit(sw, data->size);
    return rc;
}

/** \ingroup rpmdbi
 * Retrieve (key,data) pair from index database.
 * @param dbi		index database handle
 * @param dbcursor	database cursor (NULL will use db->get)
 * @param key		retrieve key value/length/flags
 * @param data		retrieve data value/length/flags
 * @param flags		(unused)
 * @return		0 on success
 */
static inline
int rpmdbiGet(rpmdbi dbi, DBC * dbcursor, DBT * key, DBT * data,
		unsigned int flags)
{
    rpmop sw = (rpmop)rpmdbiStatsAccumulator(dbi, 14);	/* RPMTS_OP_DBGET */
    int rc;
    (void) rpmswEnter(sw, 0);
    rc = (dbi->dbi_vec->cget) (dbi, dbcursor, key, data, flags);
    (void) rpmswExit(sw, data->size);
    return rc;
}

/** \ingroup rpmdbi
 * Retrieve (key,data) pair using dbcursor->c_pget.
 * @param dbi		index database handle
 * @param dbcursor	database cursor (NULL will use db->get)
 * @param key		secondary retrieve key value/length/flags
 * @param pkey		primary retrieve key value/length/flags
 * @param data		primary retrieve data value/length/flags
 * @param flags		DB_NEXT, DB_SET, or 0
 * @return		0 on success
 */
static inline
int rpmdbiPget(rpmdbi dbi, DBC * dbcursor,
		DBT * key, DBT * pkey, DBT * data, unsigned int flags)
{
    rpmop sw = (rpmop)rpmdbiStatsAccumulator(dbi, 14);	/* RPMTS_OP_DBGET */
    int rc;
    (void) rpmswEnter(sw, 0);
    rc = (dbi->dbi_vec->cpget) (dbi, dbcursor, key, pkey, data, flags);
    (void) rpmswExit(sw, data->size);
    return rc;
}

/** \ingroup rpmdbi
 * Store (key,data) pair in index database.
 * @param dbi		index database handle
 * @param dbcursor	database cursor (NULL will use db->put)
 * @param key		store key value/length/flags
 * @param data		store data value/length/flags
 * @param flags		(unused)
 * @return		0 on success
 */
static inline
int rpmdbiPut(rpmdbi dbi, DBC * dbcursor, DBT * key, DBT * data,
		unsigned int flags)
{
    rpmop sw = (rpmop)rpmdbiStatsAccumulator(dbi, 15);	/* RPMTS_OP_DBPUT */
    int rc;
    assert(key->data != NULL && key->size > 0 && data->data != NULL && data->size > 0);
    (void) rpmswEnter(sw, 0);
    rc = (dbi->dbi_vec->cput) (dbi, dbcursor, key, data, flags);
    (void) rpmswExit(sw, data->size);
    return rc;
}

/** \ingroup rpmdbi
 * Retrieve count of (possible) duplicate items.
 * @param dbi		index database handle
 * @param dbcursor	database cursor
 * @param countp	address of count
 * @param flags		(unused)
 * @return		0 on success
 */
static inline
int rpmdbiCount(rpmdbi dbi, DBC * dbcursor, unsigned int * countp,
		unsigned int flags)
{
    return (*dbi->dbi_vec->ccount) (dbi, dbcursor, countp, flags);
}

/** \ingroup rpmdbi
 * Close index database.
 * @param dbi		index database handle
 * @param flags		(unused)
 * @return		0 on success
 */
static inline
int rpmdbiClose(rpmdbi dbi, unsigned int flags)
{
    return (*dbi->dbi_vec->close) (dbi, flags);
}

/** \ingroup rpmdbi
 * Flush pending operations to disk.
 * @param dbi		index database handle
 * @param flags		(unused)
 * @return		0 on success
 */
static inline
int rpmdbiSync (rpmdbi dbi, unsigned int flags)
{
    return (*dbi->dbi_vec->sync) (dbi, flags);
}

/** \ingroup rpmdbi
 * Return whether key exists in a database.
 * @param dbi		index database handle
 * @param key		retrieve key value/length/flags
 * @param flags		usually 0
 * @return		0 if key exists, DB_NOTFOUND if not, else error
 */
static inline
int rpmdbiExists(rpmdbi dbi, DBT * key, unsigned int flags)
{
    return (*dbi->dbi_vec->exists) (dbi, key, flags);
}

/** \ingroup rpmdbi
 * Return next sequence number.
 * @param dbi		index database handle (with attached sequence)
 * @retval *seqnop	IN: delta (0 does seqno++) OUT: returned 64bit seqno
 * @param flags		usually 0
 * @return		0 on success
 */
static inline
int rpmdbiSeqno(rpmdbi dbi, int64_t * seqnop, unsigned int flags)
{
    return (*dbi->dbi_vec->seqno) (dbi, seqnop, flags);
}

/** \ingroup rpmdbi
 * Associate secondary database with primary.
 * @param dbi		index database handle
 * @param dbisecondary	secondary index database handle
 * @param callback	create secondary key from primary (NULL if DB_RDONLY)
 * @param flags		DB_CREATE or 0
 * @return		0 on success
 */
static inline
int rpmdbiAssociate(rpmdbi dbi, rpmdbi dbisecondary,
                int (*callback) (DB *, const DBT *, const DBT *, DBT *),
                unsigned int flags)
{
    return (*dbi->dbi_vec->associate) (dbi, dbisecondary, callback, flags);
}

/** \ingroup rpmdbi
 * Return join cursor for list of cursors.
 * @param dbi		index database handle
 * @param curslist	NULL terminated list of database cursors
 * @retval dbcp		address of join database cursor
 * @param flags		DB_JOIN_NOSORT or 0
 * @return		0 on success
 */
static inline
int rpmdbiJoin(rpmdbi dbi, DBC ** curslist, DBC ** dbcp,
                unsigned int flags)
{
    return (*dbi->dbi_vec->join) (dbi, curslist, dbcp, flags);
}

/** \ingroup rpmdbi
 * Is database byte swapped?
 * @param dbi		index database handle
 * @return		0 same order, 1 swapped order
 */
static inline
int rpmdbiByteSwapped(rpmdbi dbi)
{
    return (*dbi->dbi_vec->byteswapped) (dbi);
}

/** \ingroup rpmdbi
 * Return dbi statistics.
 * @param dbi		index database handle
 * @param flags		DB_FAST_STAT or 0
 * @return		0 on success
 */
static inline
int rpmdbiStat(rpmdbi dbi, unsigned int flags)
{
    return (*dbi->dbi_vec->stat) (dbi, flags);
}

/** \ingroup rpmdbi
 * Return dbi transaction id.
 * @param dbi		index database handle
 * @return		transaction id
 */
static inline
DB_TXN * rpmdbiTxnid(rpmdbi dbi)
{
    rpmdb rpmdb = (dbi ? dbi->dbi_rpmdb : NULL);
    DB_TXN * _txn = (DB_TXN*)(rpmdb ? rpmdb->db_txn : NULL);
    return _txn;
}

#if defined(_RPMDB_INTERNAL)
#if defined(WITH_DB)
static inline
int rpmlkId(rpmdb rpmdb, uint32_t *_idp)
{
    DB_ENV * dbenv = (DB_ENV *) rpmdb->db_dbenv;
    int rc = (rpmdb->_dbi[0]->dbi_eflags & DB_INIT_LOCK)
	? dbenv->lock_id(dbenv, _idp) : ENOTSUP;
if (_rpmdb_debug)
fprintf(stderr, "<-- %s(%p,%p) id 0x%x rc %d\n", "dbenv->lock_id", dbenv, _idp, *_idp, rc);
    return rc;
}

static inline
int rpmlkIdFree(rpmdb rpmdb, uint32_t _id)
{
    DB_ENV * dbenv = (DB_ENV *)rpmdb->db_dbenv;
    int rc = (rpmdb->_dbi[0]->dbi_eflags & DB_INIT_LOCK)
	? dbenv->lock_id_free(dbenv, _id) : ENOTSUP;
if (_rpmdb_debug)
fprintf(stderr, "<-- %s(%p,%u) rc %d\n", "dbenv->lock_id_free", dbenv, (unsigned)_id, rc);
    return rc;
}

static inline
int rpmlkGet(rpmdb rpmdb, DBT * _object, uint32_t _lockmode, void * _lock)
{
    DB_ENV * dbenv = (DB_ENV *)rpmdb->db_dbenv;
    uint32_t _locker = 0x12344321;
    uint32_t _flags = 0;
    int rc = (rpmdb->_dbi[0]->dbi_eflags & DB_INIT_LOCK)
	? dbenv->lock_get(dbenv, _locker, _flags, _object, (db_lockmode_t)_lockmode, (DB_LOCK*)_lock)
	: ENOTSUP;
if (_rpmdb_debug)
fprintf(stderr, "<-- %s(%p,0x%x,0x%x,%p,0x%x,%p) rc %d\n", "dbenv->lock_get", dbenv, _locker, _flags, _object, _lockmode, _lock, rc);
    return rc;
}

static inline
int rpmlkPut(rpmdb rpmdb, void * _lock)
{
    DB_ENV * dbenv = (DB_ENV *)rpmdb->db_dbenv;
    int rc = (rpmdb->_dbi[0]->dbi_eflags & DB_INIT_LOCK)
	? dbenv->lock_put(dbenv, (DB_LOCK*)_lock)
	: ENOTSUP;
if (_rpmdb_debug)
fprintf(stderr, "<-- %s(%p,%p) rc %d\n", "dbenv->lock_put", dbenv, _lock, rc);
    return rc;
}

static inline
int rpmlkDetect(rpmdb rpmdb, uint32_t _flags, uint32_t _atype, int *_rejected)
{
    DB_ENV * dbenv = (DB_ENV *)rpmdb->db_dbenv;
    int rejected = -1;
    int rc = (rpmdb->_dbi[0]->dbi_eflags & DB_INIT_LOCK)
	? dbenv->lock_detect(dbenv, _flags, _atype, &rejected)
	: ENOTSUP;
if (_rpmdb_debug)
fprintf(stderr, "<-- %s(%p,0x%x,0x%x,%p) #rejected %d rc %d\n", "dbenv->lock_detect", dbenv, _flags, _atype, _rejected, rejected, rc);
    if (_rejected)
	*_rejected = rejected;
    return rc;
}

static inline
int rpmlgcOpen(rpmdb rpmdb)
{
    DB_ENV * dbenv = (DB_ENV *)rpmdb->db_dbenv;
    DB_LOGC * _logc = NULL;
    uint32_t _flags = 0;
    int rc = (rpmdb->_dbi[0]->dbi_eflags & DB_INIT_LOG)
	? dbenv->log_cursor(dbenv, &_logc, _flags) : ENOTSUP;
    rpmdb->db_logc = (!rc ? _logc : NULL);
if (_rpmdb_debug)
fprintf(stderr, "<-- %s(%p,%p,0x%x) logc %p rc %d\n", "dbenv->log_cursor", dbenv, &_logc, _flags, _logc, rc);
    return rc;
}

static inline
int rpmlgcGet(rpmdb rpmdb, DB_LSN * _lsn, DBT * data, uint32_t flags)
{
    DB_LOGC * _logc = rpmdb->db_logc;
    int rc = (_logc ? _logc->get(_logc, _lsn, data, flags) : ENOTSUP);
if (_rpmdb_debug)
fprintf(stderr, "<-- %s(%p,%p,%p,0x%x) rc %d\n", "logc->get", _logc, _lsn, data, flags, rc);
    return rc;
}

static inline
int rpmlgcClose(rpmdb rpmdb)
{
    DB_LOGC * _logc = rpmdb->db_logc;
    uint32_t _flags = 0;
    int rc = (_logc ? _logc->close(_logc, _flags) : ENOTSUP);
    rpmdb->db_logc = NULL;
if (_rpmdb_debug)
fprintf(stderr, "<-- %s(%p) rc %d\n", "logc->close", _logc, rc);
    return rc;
}

static inline
int rpmlgcFile(rpmdb rpmdb, const DB_LSN * _lsn, char * name, size_t len)
{
    DB_ENV * dbenv = (DB_ENV*)rpmdb->db_dbenv;
    int rc = dbenv->log_file(dbenv, _lsn, name, len);
if (_rpmdb_debug)
fprintf(stderr, "<-- %s(%p,%p,%p[%u]) \"%s\" rc %d\n", "dbenv->log_file", dbenv, _lsn, name, (unsigned)len, name, rc);
    return rc;
}

static inline
int rpmlgcFlush(rpmdb rpmdb, const DB_LSN * _lsn)
{
    DB_ENV * dbenv = (DB_ENV*)rpmdb->db_dbenv;
    int rc = dbenv->log_flush(dbenv, _lsn);
if (_rpmdb_debug)
fprintf(stderr, "<-- %s(%p,%p) rc %d\n", "dbenv->log_flush", dbenv, _lsn, rc);
    return rc;
}

static inline
int rpmlgcPrintf(rpmdb rpmdb, const char * fmt, void *_A1, void *_A2, void *_A3, void *_A4, void *_A5)
{
    DB_ENV * dbenv = (DB_ENV*)rpmdb->db_dbenv;
    DB_TXN * _txnid = (DB_TXN*)rpmdb->db_txn;
    int rc = dbenv->log_printf(dbenv, _txnid, fmt, _A1, _A2, _A3, _A4, _A5);
if (_rpmdb_debug)
fprintf(stderr, "<-- %s(%p,%p,\"%s\", ...) rc %d\n", "dbenv->log_printf", dbenv, _txnid, fmt, rc);
    return rc;
}

static inline
int rpmlgcPut(rpmdb rpmdb, DB_LSN * _lsn, const DBT * data, uint32_t flags)
{
    DB_ENV * dbenv = (DB_ENV*)rpmdb->db_dbenv;
    int rc = dbenv->log_put(dbenv, _lsn, data, flags);
if (_rpmdb_debug)
fprintf(stderr, "<-- %s(%p,%p,%p,0x%x) rc %d\n", "dbenv->log_put", dbenv, _lsn, data, flags, rc);
    return rc;
}

static inline
int rpmmpfSyncAll(rpmdb rpmdb)
{
    DB_ENV * dbenv = (DB_ENV*)rpmdb->db_dbenv;
    DB_LSN * _lsn = NULL;
    int rc = (rpmdb->_dbi[0]->dbi_eflags & DB_INIT_MPOOL)
	? dbenv->memp_sync(dbenv, _lsn) : ENOTSUP;
if (_rpmdb_debug)
fprintf(stderr, "<-- %s(%p,%p) rc %d\n", "dbenv->memp_sync", dbenv, _lsn, rc);
    return rc;
}

static inline
int rpmmpfTrickle(rpmdb rpmdb)
{
    DB_ENV * dbenv = (DB_ENV*)rpmdb->db_dbenv;
    int _percent = 20;
    int _nwrote = 0;
    int rc = (rpmdb->_dbi[0]->dbi_eflags & DB_INIT_MPOOL)
	? dbenv->memp_trickle(dbenv, _percent, &_nwrote) : ENOTSUP;
if (_rpmdb_debug)
fprintf(stderr, "<-- %s(%p,%d) nwrote %d rc %d\n", "dbenv->memp_trickle", dbenv, _percent, _nwrote, rc);
    return rc;
}

static inline
int rpmmpfClose(rpmdb rpmdb)
{
    DB_MPOOLFILE * mpf = rpmdb->db_mpf;
    uint32_t _flags = 0;
    int rc = (mpf ? mpf->close(mpf, _flags) : ENOTSUP);
    rpmdb->db_mpf = NULL;
if (_rpmdb_debug)
fprintf(stderr, "<-- %s(%p) rc %d\n", "mpf->close", mpf, rc);
    return rc;
}

static inline
int rpmmpfGet(rpmdb rpmdb, uint32_t * _pgnop, uint32_t _flags, void ** _pagep)
{
    DB_MPOOLFILE * mpf = rpmdb->db_mpf;
    DB_TXN * _txnid = (DB_TXN*)rpmdb->db_txn;
    int rc = mpf->get(mpf, _pgnop, _txnid, _flags, _pagep);
if (_rpmdb_debug)
fprintf(stderr, "<-- %s(%p,%p,%p,0x%x,%p) rc %d\n", "mpf->get", mpf, _pgnop, _txnid, _flags, _pagep, rc);
    return rc;
}

static inline
int rpmmpfOpen(rpmdb rpmdb, const char * fn, uint32_t flags)
{
    DB_ENV * dbenv = (DB_ENV*)rpmdb->db_dbenv;
    DB_MPOOLFILE * mpf = NULL;
    int _perms = rpmdb->_dbi[0]->dbi_perms;
    size_t _pagesize = BUFSIZ;
    uint32_t _fcreate_flags = 0;
    int rc = (rpmdb->_dbi[0]->dbi_eflags & DB_INIT_MPOOL)
	? dbenv->memp_fcreate(dbenv, &mpf, _fcreate_flags) : ENOTSUP;

if (_rpmdb_debug)
fprintf(stderr, "<-- %s(%p,&mpf,0x%x) mpf %p rc %d\n", "dbenv->memp_fcreate", dbenv, _fcreate_flags, mpf, rc);
    if (rc) goto exit;
    rc = mpf->open(mpf, fn, flags, _perms, _pagesize);
    rpmdb->db_mpf = (!rc ? mpf : NULL);		/* FIXME: mpf memleak */
if (_rpmdb_debug)
fprintf(stderr, "<-- %s(%p,%s,0x%x) rc %d\n", "mpf->open", mpf, fn, flags, rc);

exit:
    return rc;
}

static inline
int rpmmpfPut(rpmdb rpmdb, void * _page, uint32_t flags)
{
    DB_MPOOLFILE * mpf = rpmdb->db_mpf;
    uint32_t _priority = DB_PRIORITY_DEFAULT;
    int rc = mpf->put(mpf, _page, (DB_CACHE_PRIORITY)_priority, flags);
if (_rpmdb_debug)
fprintf(stderr, "<-- %s(%p,%p,0x%x,0x%x) rc %d\n", "mpf->put", mpf, _page, _priority, flags, rc);
    return rc;
}

static inline
int rpmmpfSync(rpmdb rpmdb)
{
    DB_MPOOLFILE * mpf = rpmdb->db_mpf;
    int rc = (mpf ? mpf->sync(mpf) : ENOTSUP);
if (_rpmdb_debug)
fprintf(stderr, "<-- %s(%p) rc %d\n", "mpf->close", mpf, rc);
    return rc;
}
#endif	/* defined(WITH_DB) */
#endif	/* _RPMDB_INTERNAL */
#endif	/* !defined(SWIG) */

/** \ingroup rpmdbi
 * Destroy set of index database items.
 * @param set	set of index database items
 * @return	NULL always
 */
rpmdbiIndexSet rpmdbiFreeIndexSet(rpmdbiIndexSet set);

/** \ingroup rpmdbi
 * Count items in index database set.
 * @param set	set of index database items
 * @return	number of items
 */
unsigned int rpmdbiIndexSetCount(rpmdbiIndexSet set)
	RPM_GNUC_PURE;

/** \ingroup rpmdbi
 * Return record offset of header from element in index database set.
 * @param set	set of index database items
 * @param recno	index of item in set
 * @return	record offset of header
 */
uint32_t rpmdbiIndexRecordOffset(rpmdbiIndexSet set, unsigned int recno)
	RPM_GNUC_PURE;

/** \ingroup rpmdbi
 * Return file index from element in index database set.
 * @param set	set of index database items
 * @param recno	index of item in set
 * @return	file index
 */
uint32_t rpmdbiIndexRecordFileNumber(rpmdbiIndexSet set, unsigned int recno)
	RPM_GNUC_PURE;
#endif	/* defined(_RPMDB_INTERNAL) */

/** \ingroup rpmdb
 * Unreference a database instance.
 * @param db		rpm database
 * @param msg
 * @return		NULL always
 */
rpmdb rpmdbUnlink (rpmdb db, const char * msg);
#define	rpmdbUnlink(_db, _msg)	\
	((rpmdb)rpmioUnlinkPoolItem((rpmioItem)(_db), _msg, __FILE__, __LINE__))

/** \ingroup rpmdb
 * Reference a database instance.
 * @param db		rpm database
 * @param msg
 * @return		new rpm database reference
 */
rpmdb rpmdbLink (rpmdb db, const char * msg);
#define	rpmdbLink(_db, _msg)	\
	((void *)rpmioLinkPoolItem((rpmioItem)(_db), _msg, __FILE__, __LINE__))

/** @todo document rpmdbNew
 */
rpmdb rpmdbNew(const char * root, const char * home,
		int mode, mode_t perms, int flags);

/** \ingroup rpmdb
 * Open rpm database.
 * @param prefix	path to top of install tree
 * @retval dbp		address of rpm database
 * @param mode		open(2) flags:  O_RDWR or O_RDONLY (O_CREAT also)
 * @param perms		database permissions
 * @return		0 on success
 */
int rpmdbOpen(const char * prefix, rpmdb * dbp, int mode, mode_t perms);

/**
 * Block access to a single database index.
 * @param db		rpm database
 * @param tag		rpm tag (negative to block)
 * @return              0 on success
 */
int rpmdbBlockDBI(rpmdb db, int tag);

/**
 * Close a single database index.
 * @param db		rpm database
 * @param tag		rpm tag
 * @return              0 on success
 */
int rpmdbCloseDBI(rpmdb db, int tag);

/** \ingroup rpmdb
 * Close all database indices and free rpmdb.
 * @param db		rpm database
 * @return		0 on success
 */
int rpmdbClose (rpmdb db);

/** \ingroup rpmdb
 * Open all database indices.
 * @param db		rpm database
 * @return		0 on success
 */
int rpmdbOpenAll (rpmdb db);

/** \ingroup rpmdb
 * Return number of instances of key in a tag index.
 * @param db		rpm database
 * @param tag		rpm tag
 * @param keyp		key data
 * @param keylen	key data length (0 will use strlen(keyp))
 * @return		number of instances
 */
int rpmdbCount(rpmdb db, rpmTag tag, const void * keyp, size_t keylen);

/** \ingroup rpmdb
 * Return number of instances of package in Name index.
 * @param db		rpm database
 * @param N		rpm package name
 * @return		number of instances
 */
int rpmdbCountPackages(rpmdb db, const char * N);

/** \ingroup rpmdb
 * Return header instance for current position of rpmdb iterator.
 * @param mi		rpm database iterator
 * @return		current header instance
 */
uint32_t rpmmiInstance(rpmmi mi);

/** \ingroup rpmdb
 * Return basename tag for current position of rpmdb iterator.
 * @param mi		rpm database iterator
 * @return		current basename tag
 */
uint32_t rpmmiBNTag(rpmmi mi);

/** \ingroup rpmdb
 * Return number of elements in rpm database iterator.
 * @param mi		rpm database iterator
 * @return		number of elements
 */
unsigned int rpmmiCount(rpmmi mi);

/** \ingroup rpmdb
 * Append items to set of package instances to iterate.
 * @param mi		rpm database iterator
 * @param hdrNums	array of package instances
 * @param nHdrNums	number of elements in array
 * @return		0 on success, 1 on failure (bad args)
 */
int rpmmiGrow(rpmmi mi, const uint32_t * hdrNums, int nHdrNums);

/** \ingroup rpmdb
 * Append packages containing common basename to iterator.
 * @param mi		rpm database iterator
 * @param bn		basename key
 * @return		0 on success, 1 on failure (bad args)
 */
int rpmmiGrowBasename(rpmmi mi, const char * bn);

/** \ingroup rpmdb
 * Sort iterator instances.
 * @param mi		rpm database iterator
 * @return		0 on success
 */
int rpmmiSort(rpmmi mi);

/** \ingroup rpmdb
 * Remove items from set of package instances to iterate.
 * @note Sorted hdrNums are always passed in rpmlib.
 * @param mi		rpm database iterator
 * @param hdrNums	array of package instances
 * @param nHdrNums	number of elements in array
 * @param sorted	is the array sorted? (array will be sorted on return)
 * @return		0 on success, 1 on failure (bad args)
 */
int rpmmiPrune(rpmmi mi, uint32_t * hdrNums, int nHdrNums, int sorted);

/** \ingroup rpmdb
 * Add pattern to iterator selector.
 * @param mi		rpm database iterator
 * @param tag		rpm tag
 * @param mode		type of pattern match
 * @param pattern	pattern to match
 * @return		0 on success
 */
int rpmmiAddPattern(rpmmi mi, rpmTag tag,
		rpmMireMode mode, const char * pattern);

/** \ingroup rpmdb
 * Prepare iterator for lazy writes.
 * @note Must be called before rpmmiNext() with CDB model database.
 * @param mi		rpm database iterator
 * @param rewrite	new value of rewrite
 * @return		previous value
 */
int rpmmiSetRewrite(rpmmi mi, int rewrite);

/** \ingroup rpmdb
 * Modify iterator to mark header for lazy write on release.
 * @param mi		rpm database iterator
 * @param modified	new value of modified
 * @return		previous value
 */
int rpmmiSetModified(rpmmi mi, int modified);

/** \ingroup rpmdb
 * Return database iterator.
 * @param db		rpm database
 * @param tag		rpm tag
 * @param keyp		key data (NULL for sequential access)
 * @param keylen	key data length (0 will use strlen(keyp))
 * @return		NULL on failure
 */
rpmmi rpmmiInit(rpmdb db, rpmTag tag, const void * keyp, size_t keylen);

/** \ingroup rpmdb
 * Return next package header from iteration.
 * @param mi		rpm database iterator
 * @return		NULL on end of iteration.
 */
Header rpmmiNext(rpmmi mi);

/** \ingroup rpmdb
 * Check rpmdb signal handler for trapped signal and/or requested exit.
 * Clean up any open iterators and databases on termination condition.
 * On non-zero exit any open references to rpmdb are invalid and cannot
 * be accessed anymore, calling process should terminate immediately.
 *
 * @param terminate	0 to only check for signals, 1 to terminate anyway
 * @return		0 to continue, 1 if termination cleanup was done.
 */
int rpmdbCheckTerminate(int terminate);

/** \ingroup rpmdb
 * Check for and exit on termination signals.
 */
int rpmdbCheckSignals(void);

/** \ingroup rpmdb
 * Unreference a rpm database iterator.
 * @param mi		rpm database iterator
 * @return		NULL on last dereference
 */
rpmmi rpmmiUnlink(rpmmi mi);
#define rpmmiUnlink(_mi)  \
    ((rpmmi)rpmioUnlinkPoolItem((rpmioItem)(_mi), __FUNCTION__, __FILE__, __LINE__))

/** \ingroup rpmdb
 * Reference a rpm database iterator.
 * @param mi		rpm database iterator
 * @return		NULL on last dereference
 */
rpmmi rpmmiLink(rpmmi mi);
#define rpmmiLink(_mi)  \
    ((rpmmi)rpmioLinkPoolItem((rpmioItem)(_mi), __FUNCTION__, __FILE__, __LINE__))

/** \ingroup rpmdb
 * Destroy rpm database iterator.
 * @param mi		rpm database iterator
 * @return		NULL always
 */
rpmmi rpmmiFree(rpmmi mi);
#define rpmmiFree(_mi)  \
    ((rpmmi)rpmioFreePoolItem((rpmioItem)(_mi), __FUNCTION__, __FILE__, __LINE__))

/** \ingroup rpmdb
 * Return array of keys matching a pattern.
 * @param db		rpm database
 * @param tag		rpm tag
 * @param mode		type of pattern match
 * @param pat		pattern to match
 * @retval *argvp	array of keys that match
 * @return		0 on success
 */
int rpmdbMireApply(rpmdb db, rpmTag tag, rpmMireMode mode, const char * pat,
		const char *** argvp);

/** \ingroup rpmdb
 * Add package header to rpm database and indices.
 * @param db		rpm database
 * @param iid		install transaction id (iid = 0 or -1 to skip)
 * @param h		header
 * @param ts		(unused) transaction set (or NULL)
 * @return		0 on success
 */
int rpmdbAdd(rpmdb db, int iid, Header h, rpmts ts);

/** \ingroup rpmdb
 * Remove package header from rpm database and indices.
 * @param db		rpm database
 * @param rid		(unused) remove transaction id (rid = 0 or -1 to skip)
 * @param hdrNum	package instance number in database
 * @param ts		(unused) transaction set (or NULL)
 * @return		0 on success
 */
int rpmdbRemove(rpmdb db, int rid, uint32_t hdrNum, rpmts ts);

/**
 * Mergesort, same arguments as qsort(2).
 */
int rpm_mergesort(void *base, size_t nmemb, size_t size,
                int (*cmp) (const void *, const void *));

#ifdef __cplusplus
}
#endif

#endif	/* H_RPMDB */
