/** \ingroup rpmio
 * \file rpmio/bzdio.c
 * Support for BZIP2 compression library.
 */

#include "system.h"
#include "rpmio_internal.h"
#include <rpmmacro.h>
#include <rpmcb.h>

#if defined(WITH_BZIP2)

#define	_RPMBZ_INTERNAL
#include "rpmbz.h"

#include "debug.h"

#define	BZDONLY(fd)	assert(fdGetIo(fd) == bzdio)

const char * rpmbzStrerror(rpmbz bz)
{
    return BZ2_bzerror(bz->bzfile, &bz->bzerr);
}

void rpmbzClose(rpmbz bz, int abort, const char ** errmsg)
{
    if (bz->bzfile != NULL) {
	if ((bz->oflags & O_ACCMODE) == O_RDONLY)
	    BZ2_bzReadClose(&bz->bzerr, bz->bzfile);
	else
	    BZ2_bzWriteClose(&bz->bzerr, bz->bzfile, abort,
		&bz->nbytes_in, &bz->nbytes_out);
	if (bz->bzerr != BZ_OK && errmsg)
	    *errmsg = rpmbzStrerror(bz);
    }
    bz->bzfile = NULL;
}

rpmbz rpmbzFree(rpmbz bz, int abort)
{
    rpmbzClose(bz, abort, NULL);
    if (bz->fp != NULL) {
	(void) fclose(bz->fp);
	bz->fp = NULL;
    }
    return rpmbzFini(bz);
}

rpmbz rpmbzNew(const char * path, const char * fmode, int fdno)
{
    rpmbz bz;
    int level = -1;	/* XXX use _bzdB default */
    int flags = 0;
    mode_t omode = 0644;
    int small = -1;	/* XXX use _bzdS default */
    int verbosity = -1;	/* XXX use _bzdV default */
    const char * s = fmode;
    char stdio[20];
    char *t = stdio;
    char *te = t + sizeof(stdio) - 2;
    int c;

assert(fmode != NULL);		/* XXX return NULL instead? */

    switch ((c = *s++)) {
    case 'a':
	*t++ = (char)c;
	flags &= ~O_ACCMODE;
	flags |= O_WRONLY | O_CREAT | O_APPEND;
	break;
    case 'w':
	*t++ = (char)c;
	flags &= ~O_ACCMODE;
	flags |= O_WRONLY | O_CREAT | O_APPEND;
	break;
    case 'r':
	*t++ = (char)c;
	flags &= ~O_ACCMODE;
	flags |= O_RDONLY;
	break;
    }
	
    while ((c = *s++) != 0) {
    switch (c) {
    case '.':
	break;
    case '+':
    case 'x':
    case 'm':
    case 'c':
    case 'b':
	if (t < te) *t++ = c;
	break;
    case 's':
	if (small < 0) small = 0;
	small++;
	break;
    case 'q':
	verbosity = 0;
	break;
    case 'v':
	if (verbosity < 0) verbosity = 0;
	if (verbosity < 4) verbosity++;
	break;
    default:
	if (c >= (int)'0' && c <= (int)'9')
	    level = c - (int)'0';
	break;
    }}
    *t = '\0';

    bz = rpmbzInit(level, small, verbosity, (flags & _rpmioOMask), omode);

    if (fdno >= 0) {
	if ((bz->fp = fdopen(fdno, stdio)) != NULL)
	    bz->bzfile = ((bz->oflags & O_ACCMODE) == O_RDONLY)
		? BZ2_bzReadOpen(&bz->bzerr, bz->fp, bz->V, bz->S, NULL, 0)
		: BZ2_bzWriteOpen(&bz->bzerr, bz->fp, bz->B, bz->V, bz->W);
    } else if (path != NULL) {
	if ((bz->fp = fopen(path, stdio)) != NULL)
	    bz->bzfile = ((bz->oflags & O_ACCMODE) == O_RDONLY)
		? BZ2_bzReadOpen(&bz->bzerr, bz->fp, bz->V, bz->S, NULL, 0)
		: BZ2_bzWriteOpen(&bz->bzerr, bz->fp, bz->B, bz->V, bz->W);
    }

    return (bz->bzfile != NULL ? bz : rpmbzFree(bz, 0));
}

#ifdef	NOTYET
static void rpmbzCompress(rpmbz bz, rpmzJob job)
{
    bz_stream *bzstrm = &bz->strm;
    size_t len;			/* remaining bytes to compress/check */
    int ret;

    /* initialize the deflate stream for this block */
    bzstrm->bzfree = NULL;
    bzstrm->bzalloc = NULL;
    bzstrm->opaque = NULL;
    if ((ret = BZ2_bzCompressInit(bzstrm, bz->B, bz->V, bz->W)) != BZ_OK)
	bail("not enough memory", "BZ2_bzCompressInit");

    bzstrm->next_in = job->in->buf;
    bzstrm->next_out = job->out->buf;
    bzstrm->avail_out = job->out->len;

    /* run _PIGZMAX-sized amounts of input through deflate -- this loop is
    * needed for those cases where the integer type is smaller than the
    * size_t type, or when len is close to the limit of the size_t type */
    len = job->in->len;
    while (len > _PIGZMAX) {
	bzstrm->avail_in = _PIGZMAX;
	if ((ret = BZ2_bzCompress(bzstrm, BZ_RUN)) != BZ_RUN_OK)
	    fprintf(stderr, "*** BZ2_bzCompress(%d): %d\n", BZ_RUN, ret);
assert(bzstrm->avail_in == 0 && bzstrm->avail_out != 0);
	len -= _PIGZMAX;
    }

    /* run the last piece through deflate -- terminate with a sync marker,
    * or finish deflate stream if this is the last block */
    bzstrm->avail_in = (unsigned)len;
    ret = BZ2_bzCompress(bzstrm, BZ_FINISH);
    if (!(ret == BZ_FINISH_OK || ret == BZ_STREAM_END))
	fprintf(stderr, "*** BZ2_bzCompress(%d): %d\n", BZ_FINISH, ret);
    if ((ret = BZ2_bzCompressEnd(bzstrm)) != BZ_OK)
	fprintf(stderr, "*** BZ2_bzCompressEnd: %d\n", ret);
#ifdef	NOTYET
assert(bzstrm->avail_in == 0 && bzstrm->avail_out != 0);
#endif

}

static void rpmbzDecompress(rpmbz bz, rpmzJob job)
{
    bz_stream *bzstrm = &bz->strm;
    int ret;

    /* initialize the inflate stream for this block */
    bzstrm->bzfree = NULL;
    bzstrm->bzalloc = NULL;
    bzstrm->opaque = NULL;
    if ((ret = BZ2_bzDecompressInit(bzstrm, bz->V, bz->S)) != BZ_OK)
	bail("not enough memory", "BZ2_bzDecompressInit");

    bzstrm->next_in = job->in->buf;
    bzstrm->avail_in = job->in->len;
    bzstrm->next_out = job->out->buf;
    bzstrm->avail_out = job->out->len;

    if ((ret = BZ2_bzDecompress(bzstrm)) != BZ_RUN_OK)
	fprintf(stderr, "*** BZ2_bzDecompress: %d\n", ret);

    job->out->len -= bzstrm->avail_out;

    if ((ret = BZ2_bzDecompressEnd(bzstrm)) != BZ_OK)
	fprintf(stderr, "*** BZ2_bzDecompressEnd: %d\n", ret);

}
#endif	/* NOTYET */

ssize_t rpmbzRead(rpmbz bz, char * buf, size_t count,
		const char ** errmsg)
{
    ssize_t rc = 0;

#ifdef	NOTYET	/* XXX hmmm, read after close needs to return EOF. */
assert(bz->bzfile != NULL);
#else
    if (bz->bzfile == NULL) return 0;
#endif
    rc = BZ2_bzRead(&bz->bzerr, bz->bzfile, buf, (int)count);
    switch (bz->bzerr) {
    case BZ_STREAM_END: {
	void * unused = NULL;
	int nUnused = 0;

	BZ2_bzReadGetUnused(&bz->bzerr, bz->bzfile, &unused, &nUnused);
	if (unused != NULL && nUnused > 0)
	    unused = memcpy(xmalloc(nUnused), unused, nUnused);
	else {
	    unused = NULL;
	    nUnused = 0;
	}
	rpmbzClose(bz, 0, NULL);
	bz->bzfile = BZ2_bzReadOpen(&bz->bzerr, bz->fp, bz->V, bz->S,
			unused, nUnused);
	unused = _free(unused);
    }   /*@fallthrough@*/
    case BZ_OK:
assert(rc >= 0);
	break;
    default:
	rc = -1;
	if (errmsg != NULL)
	    *errmsg = rpmbzStrerror(bz);
	rpmbzClose(bz, 1, NULL);
	break;
    }
    return rc;
}

ssize_t rpmbzWrite(rpmbz bz, const char * buf, size_t count,
		const char ** errmsg)
{
    ssize_t rc;

assert(bz->bzfile != NULL);	/* XXX TODO: lazy open? */
    BZ2_bzWrite(&bz->bzerr, bz->bzfile, (void *)buf, (int)count);
    switch (bz->bzerr) {
    case BZ_OK:
	rc = count;
	break;
    default:
	if (errmsg != NULL)
	    *errmsg = rpmbzStrerror(bz);
	rpmbzClose(bz, 1, NULL);
	rc = -1;
	break;
    }
    return rc;
}

static int rpmbzSeek(void * _bz, _libio_pos_t pos,
			int whence)
{
    return -2;
}

static rpmbz rpmbzOpen(const char * path, const char * fmode)
{
    return  rpmbzNew(path, fmode, -1);
}

static rpmbz rpmbzFdopen(void * _fdno, const char * fmode)
{
    int fdno = (int)((long)_fdno);	/* XXX hack */
    return  rpmbzNew(NULL, fmode, fdno);
}

static int rpmbzFlush(void * _bz)
{
    rpmbz bz = (rpmbz) _bz;
    return BZ2_bzflush(bz->bzfile);
}

/* =============================================================== */
RPM_GNUC_PURE
static inline void * bzdFileno(FD_t fd)
{
    void * rc = NULL;
    int i;

    FDSANE(fd);
    for (i = fd->nfps; i >= 0; i--) {
	FDSTACK_t * fps = &fd->fps[i];
	if (fps->io != bzdio)
	    continue;
	rc = fps->fp;
	break;
    }

    return rc;
}

static FD_t bzdOpen(const char * path, const char * fmode)
{
    FD_t fd;

    if (path == NULL) return NULL;
    if (fmode == NULL) return NULL;

    /* Create the private FILE handle. */
    rpmbz bz = rpmbzOpen(path, fmode);
    if (bz == NULL)
	return NULL;

    /* Create the new FD_t */
    fd = fdNew("open (bzdOpen)");
#ifdef	NOTYET		/* XXX persistent URI cache prevents fileno(bz->fp) */
    fdPop(fd); fdPush(fd, bzdio, bz, fileno(bz->fp));
#else
    fdPop(fd); fdPush(fd, bzdio, bz, -1);
#endif
    fdSetOpen(fd, path, bz->oflags, bz->omode);

    DBGIO(fd, (stderr, "==>\tbzdOpen(\"%s\", \"%s\") fd %p %s\n", path, fmode, (fd ? fd : NULL), fdbg(fd)));
    return fdLink(fd, __FUNCTION__);
}

static FD_t bzdFdopen(void * cookie, const char * fmode)
{
    FD_t fd = c2f(cookie);
    int fdno = fdFileno(fd);

    /* Create the private FILE handle. */
    rpmbz bz = rpmbzFdopen((void *)((long)fdno), fmode);
    if (bz == NULL)
	return NULL;

    /* Pull forward the ufdio path. */
    const char *fn = fdGetOPath(fd);
    if (fn) {
#ifdef	NOTYET	/* XXX add path to rpmbz? */
	if (bz->path)
	    bz->path = _free(bz->path);
	bz->path = xstrdup(fn);
#endif
    }

    /* Replace ufdio with bzdio. */
#ifdef	NOTYET		/* XXX persistent URI cache prevents pop */
    fdPop(fd); fdPush(fd, bzdio, bz, fileno(bz->fp));
#else
    (void) fdSetFdno(fd, -1);		/* XXX skip the fdio close */
    fdPush(fd, bzdio, bz, fdno);		/* Push bzdio onto stack */
#endif
    fdSetOpen(fd, fn, bz->oflags, bz->omode);	/* XXX add path to rpmbz? */

    return fdLink(fd, __FUNCTION__);
}

static int bzdFlush(void * cookie)
{
    FD_t fd = c2f(cookie);
    rpmbz bz = (rpmbz) bzdFileno(fd);
    return rpmbzFlush(bz);
}

static ssize_t bzdRead(void * cookie, char * buf, size_t count)
{
    FD_t fd = c2f(cookie);
    rpmbz bz = (rpmbz) bzdFileno(fd);
    ssize_t rc = 0;

assert(bz != NULL);
    if (fd->bytesRemain == 0) return 0;	/* XXX simulate EOF */
    fdstat_enter(fd, FDSTAT_READ);
    rc = rpmbzRead(bz, buf, count, (const char **)&fd->errcookie);
    if (rc >= 0) {
	fdstat_exit(fd, FDSTAT_READ, rc);
	if (fd->ndigests && rc > 0) fdUpdateDigests(fd, (const unsigned char *)buf, rc);
    }
    return rc;
}

static ssize_t bzdWrite(void * cookie, const char * buf, size_t count)
{
    FD_t fd = c2f(cookie);
    rpmbz bz = (rpmbz) bzdFileno(fd);
    ssize_t rc;

assert(bz != NULL);
    if (fd->bytesRemain == 0) return 0;	/* XXX simulate EOF */

    if (fd->ndigests && count > 0) fdUpdateDigests(fd, (const unsigned char *)buf, count);

    fdstat_enter(fd, FDSTAT_WRITE);
    rc = rpmbzWrite(bz, buf, count, (const char **)&fd->errcookie);
    if (rc >= 0)
	fdstat_exit(fd, FDSTAT_WRITE, rc);
    return rc;
}

RPM_GNUC_PURE
static int bzdSeek(void * cookie, _libio_pos_t pos, int whence)
{
    FD_t fd = c2f(cookie);
    rpmbz bz = (rpmbz) bzdFileno(fd);

assert(bz != NULL);
    BZDONLY(fd);
    return rpmbzSeek(bz, pos, whence);
}

static int bzdClose(void * cookie)
{
    FD_t fd = c2f(cookie);
    rpmbz bz = (rpmbz) bzdFileno(fd);
    int rc;

assert(bz != NULL);
#ifdef	DYING
    if (bz->bzfile == NULL)	/* XXX memory leak w errors? */
	return -2;
#endif

    fdstat_enter(fd, FDSTAT_CLOSE);
    rpmbzClose(bz, 0, (const char **)&fd->errcookie);
    rc = 0;	/* XXX FIXME */

    /* XXX TODO: preserve fd if errors */

    if (fd)
    if (rc >= 0)
	fdstat_exit(fd, FDSTAT_CLOSE, rc);

DBGIO(fd, (stderr, "==>\tbzdClose(%p) rc %lx %s\n", cookie, (unsigned long)rc, fdbg(fd)));

    if (_rpmio_debug || rpmIsDebug()) fdstat_print(fd, "BZDIO", stderr);

    if (rc == 0) {
	bz = rpmbzFree(bz, 0);
	fd = fdFree(fd, "open (bzdClose)");
    }
    return rc;
}

static struct FDIO_s bzdio_s = {
  bzdRead, bzdWrite, bzdSeek, bzdClose,	bzdOpen, bzdFdopen, bzdFlush,
};

FDIO_t bzdio = &bzdio_s;

#endif

