/*
 *   File:   fanotify-example.c
 *   Date:   Fri Nov 15 14:55:49 2013
 *   Author: Aleksander Morgado <aleksander@lanedo.com>
 *
 *   A simple tester of fanotify in the Linux kernel.
 *
 *   This program is released in the Public Domain.
 *
 *   Compile with:
 *     $> gcc -o fanotify-example fanotify-example.c
 *
 *   Run as:
 *     $> ./fanotify-example /path/to/monitor /another/path/to/monitor ...
 */

/* Define _GNU_SOURCE, Otherwise we don't get O_LARGEFILE */
#if !defined(_GNU_SOURCE)
#define _GNU_SOURCE
#endif

#include "system.h"

#include <sys/signalfd.h>
#include <sys/fanotify.h>

#include <rpmio.h>
#include <argv.h>
#include <poptIO.h>

#include "debug.h"

static int _debug;

/* Structure to keep track of monitored directories */
typedef struct {
    char *path;			/* Path of the directory */
} monitored_t;

/* Enumerate list of FDs to poll */
enum {
    FD_POLL_SIGNAL = 0,
    FD_POLL_FANOTIFY,
    FD_POLL_MAX
};

/* Setup fanotify notifications (FAN) mask. All these defined in fanotify.h. */
static uint64_t event_mask = (
    FAN_ACCESS |		/* File accessed */
    FAN_MODIFY |		/* File modified */
    FAN_CLOSE_WRITE |		/* Writtable file closed */
    FAN_CLOSE_NOWRITE |		/* Unwrittable file closed */
    FAN_OPEN |			/* File was opened */
    FAN_ONDIR |			/* We want events in the directory */
    FAN_EVENT_ON_CHILD |	/* We want events in files of the directory */
    0);

/* Array of directories being monitored */
static monitored_t *monitors;
static int n_monitors;

static char *get_program_name_from_pid(int pid, char *b, size_t nb)
{
    int fdno;
    ssize_t len;
    char *aux;

    /* Try to get program name by PID */
    snprintf(b, nb, "/proc/%d/cmdline", pid);
    fdno = open(b, O_RDONLY);
    if (fdno < 0)
	return NULL;

    /* Read file contents into buffer */
    len = read(fdno, b, nb - 1);
    close(fdno);
    if (len <= 0)
	return NULL;

    b[len] = '\0';
    aux = strstr(b, "^@");
    if (aux)
	*aux = '\0';

    return b;
}

static char *get_file_path_from_fd(int fdno, char *b, size_t nb)
{
    ssize_t len;

    if (fdno <= 0)
	return NULL;

    snprintf(b, nb, "/proc/self/fd/%d", fdno);
    if ((len = readlink(b, b, nb - 1)) < 0)
	return NULL;

    b[len] = '\0';
    return b;
}

static void event_process(struct fanotify_event_metadata *event)
{
    char b[PATH_MAX];
    size_t nb = sizeof(b);

    printf("Received event in path '%s'",
	   get_file_path_from_fd(event->fd, b, nb) ? b : "unknown");
    printf(" pid=%d (%s): \n", event->pid,
	   (get_program_name_from_pid(event->pid, b, nb) ? b : "unknown"));

    if (event->mask & FAN_OPEN)
	printf("\tFAN_OPEN\n");
    if (event->mask & FAN_ACCESS)
	printf("\tFAN_ACCESS\n");
    if (event->mask & FAN_MODIFY)
	printf("\tFAN_MODIFY\n");
    if (event->mask & FAN_CLOSE_WRITE)
	printf("\tFAN_CLOSE_WRITE\n");
    if (event->mask & FAN_CLOSE_NOWRITE)
	printf("\tFAN_CLOSE_NOWRITE\n");
    fflush(stdout);

    close(event->fd);
}

static void shutdown_fanotify(int fdno)
{
    int i;

    for (i = 0; i < n_monitors; ++i) {
	char * path = monitors[i].path;
	monitors[i].path = NULL;
	/* Remove the mark, using same event mask as when creating it */
	fanotify_mark(fdno, FAN_MARK_REMOVE, event_mask, AT_FDCWD, path);
	free(path);
    }
    free(monitors);
    close(fdno);
}

static int initialize_fanotify(int argc, char *argv[])
{
    int fdno;
    int rc;

    /* Create new fanotify device */
    fdno = fanotify_init(FAN_CLOEXEC, O_RDONLY | O_CLOEXEC | O_LARGEFILE);
    if (fdno < 0) {
	fprintf(stderr, "Couldn't setup new fanotify device: %s\n",
		strerror(errno));
	goto exit;
    }

    /* Allocate array of monitor setups */
    n_monitors = argc - 1;
    monitors = malloc(n_monitors * sizeof(monitored_t));

    /* Loop all input directories, setting up marks */
    for (int i = 0; i < n_monitors; ++i) {
	char * path = strdup(argv[i+1]);
	monitors[i].path = path;
	/* Add new fanotify mark */
	rc = fanotify_mark(fdno, FAN_MARK_ADD, event_mask, AT_FDCWD, path);
	if (rc  < 0) {
	    fprintf(stderr, "Couldn't add monitor in directory '%s': '%s'\n",
		    path, strerror(errno));
	    close(fdno);
	    fdno = -1;
	    goto exit;
	}

	printf("Started monitoring directory '%s'...\n", path);
    }

exit:
    return fdno;
}

static void shutdown_signals(int signal_fd)
{
    close(signal_fd);
}

static int initialize_signals(void)
{
    int signal_fd;
    sigset_t sigmask;

    /* Handle SIGINT and SIGTERM in the signal_fd, so we block them. */
    sigemptyset(&sigmask);
    sigaddset(&sigmask, SIGINT);
    sigaddset(&sigmask, SIGTERM);

    if (sigprocmask(SIG_BLOCK, &sigmask, NULL) < 0) {
	fprintf(stderr, "Couldn't block signals: '%s'\n", strerror(errno));
	return -1;
    }

    /* Get new FD to read signals from */
    if ((signal_fd = signalfd(-1, &sigmask, 0)) < 0) {
	fprintf(stderr, "Couldn't setup signal FD: '%s'\n", strerror(errno));
	return -1;
    }

    return signal_fd;
}

/*==============================================================*/
static const char *comm;
static int current_mount;
static const char *output;
static const char *filter;
static int timeout;
static const char **pids;
static int timestamp;

static struct poptOption rpmfanOptionsTable[] = {
 { "debug", 'd', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN,  &_debug, 1,
	NULL, NULL },

 { "current-mount", 'c', POPT_ARG_VAL,  &current_mount, 1,
	N_("Only record events on partition/mount of current directory."), NULL },
 { "output", 'o', POPT_ARG_STRING, 	 &output, 1,
	N_("Write events to FILE instead of standard output."), N_("FILE") },
 { "seconds", 's', POPT_ARG_INT, 	 &timeout, 1,
	N_("Stop after SECONDS."), N_("SECONDS") },
 { "timestamp", 't', POPT_ARG_VAL, 	 &timestamp, 1,
	N_("Add timestamp to events. Give twice for seconds since the epoch."), NULL },
 { "ignore-pid", 'p', POPT_ARG_ARGV, 	 &pids, 1,
	N_("Ignore events for process PID. Can be specified multiple times."), N_("PID") },
 { "filter", 'f', POPT_ARG_STRING, 	 &filter, 1,
	N_("Show only the given event types; choose from C, R, O, or W, e. g. --filter=OC."), N_("TYPES") },
 { "command", 'C', POPT_ARG_STRING, 	 &comm, 1,
	N_("Show only events for CMD."), N_("CMD") },

 { NULL, '\0', POPT_ARG_INCLUDE_TABLE, rpmioAllPoptTable, 0,
	N_(" Common options for all rpmio executables:"), NULL },

  POPT_AUTOALIAS
  POPT_AUTOHELP
  POPT_TABLEEND
};

int main(int argc, char *argv[])
{
    poptContext con = rpmioInit(argc, argv, rpmfanOptionsTable);
    int signal_fd = -1;
    int fanotify_fd = -1;
    struct pollfd fds[FD_POLL_MAX];
    int ec = EXIT_FAILURE;	/* assume failure */

    /* Input arguments... */
    if (argc < 2) {
	fprintf(stderr, "Usage: %s directory1 [directory2 ...]\n", argv[0]);
	goto exit;
    }

    /* Initialize signals FD */
    if ((signal_fd = initialize_signals()) < 0) {
	fprintf(stderr, "Couldn't initialize signals\n");
	goto exit;
    }

    /* Initialize fanotify FD and the marks */
    if ((fanotify_fd = initialize_fanotify(argc, argv)) < 0) {
	fprintf(stderr, "Couldn't initialize fanotify\n");
	goto exit;
    }

    /* Setup polling */
    fds[FD_POLL_SIGNAL].fd = signal_fd;
    fds[FD_POLL_SIGNAL].events = POLLIN;
    fds[FD_POLL_FANOTIFY].fd = fanotify_fd;
    fds[FD_POLL_FANOTIFY].events = POLLIN;

    /* Now loop */
    for (;;) {
	/* Block until there is something to be read */
	if (poll(fds, FD_POLL_MAX, -1) < 0) {
	    fprintf(stderr, "Couldn't poll(): '%s'\n", strerror(errno));
	    goto exit;
	}

	/* Signal received? */
	if (fds[FD_POLL_SIGNAL].revents & POLLIN) {
	    int fdno = fds[FD_POLL_SIGNAL].fd;
	    struct signalfd_siginfo fdsi;

	    if (read(fdno, &fdsi, sizeof(fdsi)) != sizeof(fdsi)) {
		fprintf(stderr, "Couldn't read signal, wrong size read\n");
		goto exit;
	    }

	    /* Break loop if we got the expected signal */
	    if (fdsi.ssi_signo == SIGINT || fdsi.ssi_signo == SIGTERM)
		break;

	    fprintf(stderr, "Received unexpected signal\n");
	}

	/* fanotify event received? */
	if (fds[FD_POLL_FANOTIFY].revents & POLLIN) {
	    int fdno = fds[FD_POLL_FANOTIFY].fd;
	    char b[BUFSIZ];
	    size_t nb = sizeof(b);
	    ssize_t length;

	    /* Read from the FD. It will read all events available up to
	     * the given buffer size. */
	    if ((length = read(fdno, b, nb)) > 0) {
		struct fanotify_event_metadata *metadata;

		metadata = (struct fanotify_event_metadata *) b;
		while (FAN_EVENT_OK(metadata, length)) {
		    event_process(metadata);
		    if (metadata->fd > 0)
			close(metadata->fd);
		    metadata = FAN_EVENT_NEXT(metadata, length);
		}
	    }
	}
    }
    ec = EXIT_SUCCESS;

exit:
    if (fanotify_fd >= 0)
	shutdown_fanotify(fanotify_fd);
    if (signal_fd >= 0)
	shutdown_signals(signal_fd);

    con = rpmioFini(con);

    return ec;
}
