/** \ingroup rpmio
 * \file rpmio/msqio.c
 * Support for SysV message queues.
 */

#include "system.h"

#if defined(HAVE_SYS_IPC_H)
# include <sys/ipc.h>
#endif

#if defined(WITH_MQ)
#if defined(HAVE_MQUEUE_H)
# include <mqueue.h>
#endif
#else	/* WITH_MQ */
struct mq_attr
{
    long mq_flags;		/* Message queue flags.  */
    long mq_maxmsg;		/* Maximum number of messages.  */
    long mq_msgsize;		/* Maximum message size.  */
    long mq_curmsgs;		/* Number of messages currently queued.  */
    long __pad[4];
};
#endif	/* WITH_MQ */

#if defined(WITH_MSQ)
#if defined(HAVE_SYS_MSG_H)
# include <sys/msg.h>
#endif

#else	/* WITH_MSQ */
/* XXX stub in enough to use struct mq_attr */
typedef int mqd_t;
#endif	/* WITH_MSQ */

#include "rpmio_internal.h"
#include <rpmlog.h>
#include <rpmmacro.h>
#include <rpmcb.h>		/* XXX rpmIsDebug() */
#include <rpmzlog.h>		/* XXX rpmzLog type */
#include <yarn.h>

#if defined(WITH_AIO)

#if defined(HAVE_AIO_H)
# include <aio.h>
#endif

#else	/* WITH_AIO */

/* XXX stub in enough to use struct aiocb_s pool */
enum {
    LIO_READ,
    LIO_WRITE,
    LIO_NOP,
};

struct aiocb
{
  int aio_fildes;		/* File desriptor.  */
  int aio_lio_opcode;		/* Operation to be performed.  */
  int aio_reqprio;		/* Request priority offset.  */
  volatile void *aio_buf;	/* Location of buffer.  */
  size_t aio_nbytes;		/* Length of transfer.  */
  struct sigevent aio_sigevent;	/* Signal number and value.  */

  /* Internal members.  */
  struct aiocb *__next_prio;
  int __abs_prio;
  int __policy;
  int __error_code;
  __ssize_t __return_value;

#ifndef __USE_FILE_OFFSET64
  __off_t aio_offset;		/* File offset.  */
  char __pad[sizeof (__off64_t) - sizeof (__off_t)];
#else
  __off64_t aio_offset;		/* File offset.  */
#endif
  char __glibc_reserved[32];
};

static ssize_t
aio_return (struct aiocb *aiocbp)
{
    return aiocbp->__return_value;
}

static int
aio_error (const struct aiocb *aiocbp)
{
    return aiocbp->__error_code;
}

#endif	/* WITH_AIO */

#define	_RPMAIO_INTERNAL
#include "rpmaio.h"

typedef	struct AIOQ_s * AIOQ_t;
struct AIOQ_s {
    yarnLock have;
    void * pool;
    rpmioItem head;             /*!< AIO cb queue head. */
    rpmioItem * tail;           /*!< AIO cb queue tail. */
};

#define	_RPMMSQ_INTERNAL
#include "rpmmsq.h"

#include "debug.h"

int _rpmmsq_debug;

#define F_ISSET(_f, _FLAG) ((_f) & RPMMSQ_FLAGS_##_FLAG)
#define MSQF_ISSET(_FLAG) F_ISSET(msq->flags, _FLAG)

#define SPEW(_fmt, ...) \
    if ((msq && MSQF_ISSET(DEBUGIO)) || _rpmmsq_debug || _rpmio_debug) \
	fprintf(stderr, _fmt, __VA_ARGS__)

#define	MSQONLY(fd)	assert(fdGetIo(fd) == msqio)

#define Z(_rc)  assert((_rc) == 0)

#ifdef __cplusplus
GENfree(rpmmsq)
#endif	/* __cplusplus */

/* XXX TODO: read system settings. */
/* /proc/sys/fs/mqueue/msg_default */
#define	MSQ_MSGMAX	10
/* /proc/sys/fs/mqueue/msgsize_default */
#define	MSQ_MSGSIZE	BUFSIZ

/* =============================================================== */
static int _lockdebug = 0;
static int _conddebug = -1;
#define PEEK(_bolt)	yarnPeekLock(_bolt)
#define	POSSESS(_bolt) \
    {	yarnPossess(_bolt); \
	if (zlog && _lockdebug) \
	    rpmzLogAdd(zlog, "***\t  POSSESS(%p)\t%ld", _bolt, PEEK(_bolt)); \
    }
#define	RELEASE(_bolt) \
    {	if (zlog && _lockdebug) \
	    rpmzLogAdd(zlog, "***\t  RELEASE(%p)\t%ld", _bolt, PEEK(_bolt)); \
	yarnRelease(_bolt); \
    }
#define	TWIST(_bolt, _op, _val) \
    {	if (zlog && _conddebug) \
	    rpmzLogAdd(zlog, "***\t    TWIST(%p, %d, %d)\t%ld", _bolt, _op, _val, PEEK(_bolt)); \
	yarnTwist(_bolt, _op, _val); \
    }
#define	WAITFOR(_bolt, _op, _val) \
    {	if (zlog && _conddebug) \
	    rpmzLogAdd(zlog, "***\t  WAITFOR(%p, %d, %d)\t%ld", _bolt, _op, _val, PEEK(_bolt));\
	yarnWaitFor(_bolt, _op, _val); \
    }

static inline
void aioqInit(AIOQ_t aioq)
{
    rpmmsq msq = NULL;
    RPM_GNUC_TM_ATOMIC {
	aioq->have = yarnNewLock(0);
	aioq->pool = NULL;
	aioq->head = NULL;
	aioq->tail = &aioq->head;
    }
SPEW("<--\t%s(%p)\n", __FUNCTION__, aioq);
}

static inline
rpmioItem aioqGetHead(AIOQ_t aioq)
{
    rpmmsq msq = NULL;
    rpmioItem item;
    RPM_GNUC_TM_ATOMIC {
	if ((item = aioq->head) != NULL) {
	    aioq->head = item->next;
	    item->next = NULL;
	    if (aioq->head == NULL)
		aioq->tail = &aioq->head;
	}
    }
if (item)
SPEW("<--\t%s(%p) %p\n", __FUNCTION__, aioq, item);
    return item;
}

static inline
void aioqPutTail(AIOQ_t aioq, rpmioItem item)
{
    rpmmsq msq = NULL;
    RPM_GNUC_TM_ATOMIC {
	*aioq->tail = item;
	aioq->tail = (rpmioItem *) &item->next;
    }
SPEW("<--\t%s(%p, %p)\n", __FUNCTION__, aioq, item);
}

static inline
ssize_t aioqPutWait(rpmmsq msq, unsigned long *priop,
		int op, int prio, void *b, size_t nb,
		struct sigevent *sigev, off_t offset)
{
    AIOQ_t aioq = &msq->aioq;
    rpmzLog zlog = msq->zlog;
    ssize_t rc;

    rpmaio aio = rpmaioNew(msq->qid, op, prio, b, nb, sigev, offset);
    aio->ix = ++msq->nqueued;
    rpmioItem item = (rpmioItem) aio;

    switch (op) {
    case LIO_WRITE:
	/* XXX Asynchronous writes? */
	if ((msq->oflags & O_NONBLOCK) && nb <= sizeof(aio->buf)) {
	    memcpy(aio->buf, b, nb);
	    b = aio->buf;
	    break;
	}
	/*@fallthrough@*/
    default:
	item = rpmioLinkPoolItem(item,
			__FUNCTION__, __FILE__, __LINE__);
	break;
    }


    /* Producer monitor. */
    POSSESS(aioq->have);
    WAITFOR(aioq->have, NOT_TO_BE, msq->msgmax);
    aioqPutTail(aioq, item);
    TWIST(aioq->have, BY, 1);

    if (b != aio->buf) {
	/* Wait for AIO completion. */
	POSSESS(item->use);
	WAITFOR(item->use, TO_BE, 1);
	RELEASE(item->use);

	rc = aio_return(&aio->aiocb);
	if (rc >= 0 && priop)
	    *priop = aio->aiocb.aio_reqprio;	/* XXX prio */
	if (rc < 0)
	    errno = aio_error(&aio->aiocb);	/* XXX W2DO? */

	aio = rpmaioFree(aio);
    } else {
	/* XXX Asynchronous writes. */
	rc = 0;
	errno = EINPROGRESS;			/* XXX */
    }

SPEW("<--\t%s(%p, %p) rc %ld\n", __FUNCTION__, aioq, item, rc);
    return rc;
}

/* =============================================================== */
static char * rpmmsqDbug(void *_msq, char *b, size_t nb)
{
    rpmmsq msq = (rpmmsq) _msq;
    size_t len = strlen(b);
    char * be = b + len;
    rpmioItem item = (rpmioItem) msq;
    long use;

    if (msq && (use = PEEK(item->use)) > 0) {
	int colorize = isatty(fileno(stderr));
#define ANSI_BRIGHT_BLUE	"\x1b[34;1m"
#define ANSI_RESET		"\x1b[0m"
	*be++ = '\n';
	if (colorize) be = stpcpy(be, ANSI_BRIGHT_BLUE);
	be += sprintf(be, "========================== msq(%p) use %ld\n",
		msq, use);
#define PRINT(_fmt, _foo) \
    {	be += sprintf(be, "%25s: %"#_fmt"\n", #_foo, msq->_foo); }
#define PRINT_ATTRS(_fmt, _foo) \
    {	be += sprintf(be, "%25s: %"#_fmt"\n", #_foo, msq->oattrs._foo); }
#define PRINT_SIGEV(_fmt, _foo) \
    {	be += sprintf(be, "%25s: %"#_fmt"\n", #_foo, msq->sigev._foo); }
	PRINT(s, qname);
	PRINT(s, fmode);
	PRINT(d, fdno);
	PRINT(x, flags);
	PRINT(x, qid);
	PRINT(#o, oflags);
	PRINT(0o, omode);
	PRINT(x, key);
	PRINT(ld, mtype);
	PRINT(d, msgmax);
	PRINT(d, msgsize);
	PRINT(d, nsent);
	PRINT(d, nrecv);
	PRINT(d, ntimeout);
	PRINT(d, nagain);
	PRINT_ATTRS(ld, mq_flags);
	PRINT_ATTRS(ld, mq_maxmsg);
	PRINT_ATTRS(ld, mq_msgsize);
	PRINT_ATTRS(ld, mq_curmsgs);
	switch (msq->sigev.sigev_notify) {
	default:
	    break;
	case SIGEV_THREAD:
	    PRINT_SIGEV(p, sigev_value.sival_ptr);
	    PRINT_SIGEV(d, sigev_signo);
	    PRINT_SIGEV(d, sigev_notify);
	    PRINT_SIGEV(p, sigev_notify_function);
	    PRINT_SIGEV(p, sigev_notify_attributes);
	    break;
	}
	PRINT(p, aioq.head);
	PRINT(p, aioq.tail);
	PRINT(d, nqueued);

#undef	PRINT_SIGEV
#undef	PRINT_ATTRS
#undef	PRINT
	be--;
	if (colorize) be = stpcpy(be, ANSI_RESET);
	*be = '\0';
    }
    return b;
}

static void rpmmsqInit(void *_msq)
{
}

static void rpmmsqFini(void *_msq)
{
    rpmmsq msq = (rpmmsq) _msq;

    if (msq) {
SPEW("%s: queued %d sent %d recv %d timeout %d again %d\n", __FUNCTION__, msq->nqueued, msq->nsent, msq->nrecv, msq->ntimeout, msq->nagain);
	msq->msgmax = MSQ_MSGMAX;
	msq->msgsize = MSQ_MSGSIZE;

	msq->nsent = 0;
	msq->nrecv = 0;
	msq->ntimeout = 0;
	msq->nagain = 0;

	msq->flags = 0;
	msq->qname = _free(msq->qname);
	msq->fmode = _free(msq->fmode);
	msq->fdno = -1;
	msq->qid = -1;
	msq->oflags = 0;
	msq->omode = 0;
	msq->key = 0;
	msq->mtype = 0;

	AIOQ_t aioq = &msq->aioq;
	rpmzLog zlog = msq->zlog;
	rpmioItem item;

	/* Drain the queue. */
	POSSESS(aioq->have);			/* XXX enter */
	while (1) {
	    if (PEEK(aioq->have) == 0)
		break;
	    item = aioqGetHead(&msq->aioq);
	    TWIST(aioq->have, BY, -1);		/* XXX exit */
	    while (item)			/* XXX deadlock w msq ref */
		item = rpmioFreePoolItem(item,
			__FUNCTION__, __FILE__, __LINE__);
	    POSSESS(aioq->have);		/* XXX re-enter */
	}
	RELEASE(aioq->have);			/* XXX exit */
	aioq->have = yarnFreeLock(aioq->have);
	msq->nqueued = 0;

	msq->st = _free(msq->st);
	msq->ds = _free(msq->ds);
	msq->mi = _free(msq->mi);

	if (msq->zlog)
	    msq->zlog = rpmzLogDump(msq->zlog, NULL);
    }
}

RPMIOPOOL_MODULE(msq)

rpmmsq rpmmsqNew(const char * path, const char * fmode, int fdno, unsigned flags)
{
    rpmmsq msq = rpmmsqGetPool(_rpmmsqPool);

assert(fmode != NULL);		/* XXX return NULL instead? */

    const char * s = fmode;
    int c;
    int omode = 0664;

    switch ((c = *s++)) {
    case 'a':
	flags &= ~O_ACCMODE;
	flags |= O_WRONLY | O_CREAT | O_APPEND;
	omode |= IPC_CREAT;
	break;
    case 'w':
	flags &= ~O_ACCMODE;
	flags |= O_WRONLY | O_CREAT | O_TRUNC;
	omode |= IPC_CREAT;
	break;
    case 'r':
	flags &= ~O_ACCMODE;
	flags |= O_RDONLY;
	break;
    }
	
    while ((c = *s++) != 0) {
	switch (c) {
	case '.':
	    break;
	case '+':
	    flags &= ~O_ACCMODE;
	    flags |= O_RDWR;
	    continue;
	    break;
	case 'c':	/* XXX no cancel */
	    continue;
	    break;
	case 'm':	/* XXX mmap */
	    continue;
	    break;
	case 'e':	/* O_CLOEXEC */
	    /* XXX always set with mq_open */
	    flags |= O_CLOEXEC;
	    continue;
	    break;
	case 'n':	/* XXX O_NONBLOCK */
	    flags |= O_NONBLOCK;
	    continue;
	    break;
	case 't':	/* XXX O_TRUNC */
	    continue;
	    break;
	case 'x':
	    flags |= O_EXCL;
	    omode |= IPC_EXCL;
	    continue;
	    break;
	case 'D':	/* XXX collides with rpmio O_DIRECT */
	    flags |= RPMMSQ_FLAGS_DELETE;
	    continue;
	    break;
	case 'R':	/* XXX */
	    flags |= RPMMSQ_FLAGS_RESET;
	    continue;
	    break;
	case 'I':	/* XXX collides with rpmio fallocate(2) */
	    flags |= RPMMSQ_FLAGS_INFO;
	    continue;
	    break;
	case 'P':	/* XXX collides with rpmio syncfs(2) */
	    flags &= ~(RPMMSQ_TYPE_MASK << RPMMSQ_FLAGS_TYPE_SHIFT);
	    flags |= RPMMSQ_FLAGS_POSIX;
	    continue;
	    break;
	case 'S':	/* XXX collides with rpmio sync(2) */
	    flags &= ~(RPMMSQ_TYPE_MASK << RPMMSQ_FLAGS_TYPE_SHIFT);
	    flags |= RPMMSQ_FLAGS_SYSV;
	    continue;
	    break;
	case '?':
	    flags |= RPMMSQ_FLAGS_DEBUGREFS;
	    flags |= RPMMSQ_FLAGS_DEBUGIO;
	    continue;
	    break;
	default:
	    continue;
	    break;
	}
	break;
    }

    msq->fmode = xstrdup(fmode);
    msq->fdno = fdno;
    msq->flags = flags;

    msq->qid = -1;
    msq->oflags = (flags & _rpmioOMask);
    msq->omode = omode & 0777;	/* XXX mask IPC_CREAT | IPC_EXCL */
    msq->mtype = getpid();	/* XXX SysV message queues only. */

    /* XXX TODO: configure non-default settings. */
    msq->msgmax = MSQ_MSGMAX;
    msq->msgsize = MSQ_MSGSIZE;

    /* XXX TODO: use POSIX shared mutexes? */
    aioqInit(&msq->aioq);
    msq->nqueued = 0;
    msq->nsent = 0;
    msq->nrecv = 0;
    msq->ntimeout = 0;
    msq->nagain = 0;
    msq->st = NULL;
    msq->ds = NULL;
    msq->mi = NULL;
    msq->zlog = rpmzLogNew(NULL);

/* XXX TODO: fix rpmmsqFdopen() entry. */
assert(path != NULL);
assert(fdno == -1);
    if (path == NULL && fdno != -1) {		/* rpmmsqFdopen */
	path = "/rpm";	/* XXX */
	msq->qid = fdno;
    }

    /* XXX TODO: append ftok(3) keys to path. */
    msq->qname = rpmGetPath("/", path, NULL);

    key_t key = 0;
    if (!strcmp(msq->qname, "/private")) {
	key = IPC_PRIVATE;
	msq->flags |= RPMMSQ_FLAGS_DELETE;	/* XXX delete on close. */
    } else
    if (!strcmp(msq->qname, "/program")) {
	key = __progname_key;
    } else {
	const char * lpath = msq->qname;
	int projid = __progname_projid;
	if (!strcmp(lpath, "/rpm"))
	     lpath = "/usr/lib/rpm";
	key = ftok(lpath, projid);
	if (key == -1) {
	    int lvl = RPMLOG_WARNING;
	    rpmlog(lvl, "%s: ftok(%s,0x%02x) failed: %m\n",
		__FUNCTION__, path, (projid & 0xff));
	    key = __progname_key;
	    rpmlog(lvl, "Using program key 0x%x\n", key);
	}
    }
    msq->key = key;

    /* rpmmsqOpen: Reset the queue by removing. */
    if (fdno == -1 && MSQF_ISSET(RESET))
    switch (RPMMSQ_FLAGS_TYPE(msq->flags)) {
    case RPMMSQ_TYPE_DEFAULT:
    case RPMMSQ_TYPE_POSIX:
    {
#if defined(WITH_MQ)
	(void) mq_unlink(msq->qname);
#endif	/* WITH_MQ */
    }	break;
    case RPMMSQ_TYPE_SYSV:
    {
#if defined(WITH_MSQ)
	int qid = msgget(msq->key, (omode & 0777));
	if (qid != -1) {
	    (void) msgctl(qid, IPC_RMID, NULL);
	}
#endif	/* WITH_MSQ */
    }	break;
    default:
	break;
    }

    /* Set the QID. */
    if (msq->qid == -1)
    switch (RPMMSQ_FLAGS_TYPE(msq->flags)) {
    case RPMMSQ_TYPE_DEFAULT:
    case RPMMSQ_TYPE_POSIX:
    {
#if defined(WITH_MQ)
	/* XXX put *open in a monitor. */
	msq->oattrs.mq_flags = (flags & O_NONBLOCK);
	msq->oattrs.mq_maxmsg = msq->msgmax;
	msq->oattrs.mq_msgsize = msq->msgsize;
	msq->oattrs.mq_curmsgs = 0;
	msq->qid = mq_open(msq->qname, msq->flags, msq->omode, &msq->oattrs);

	/* XXX FIXME: modern glibc supplies pthread_getattr_default_np() */
	Z(pthread_getattr_default_np(&msq->attr));
	Z(pthread_attr_setdetachstate(&msq->attr, PTHREAD_CREATE_DETACHED));
	msq->sigev.sigev_notify = SIGEV_THREAD;
	msq->sigev.sigev_notify_function = rpmmsqReader;
	msq->sigev.sigev_notify_attributes = &msq->attr;
	msq->sigev.sigev_value.sival_ptr = msq;
	if (msq->qid != -1)
	    Z(rpmmsqNotify(msq, &msq->sigev));
#endif	/* WITH_MQ */
    }	break;
    case RPMMSQ_TYPE_SYSV:
    {
#if defined(WITH_MSQ)
	msq->qid = msgget(msq->key, omode);
#endif	/* WITH_MSQ */
    }	break;
    default:
	break;
    }

SPEW("<== %s(%s,%s,%d,0x%x) qid %d\n", __FUNCTION__, path, fmode, fdno, flags, (msq ? msq->qid : -1));

    return rpmmsqLink(msq);
}

/* =============================================================== */
ssize_t rpmmsqRecv(rpmmsq msq, char * buf, size_t count, unsigned long *priop)
{
    ssize_t rc = -1;	/* assume failure */

    if (msq)
    switch (RPMMSQ_FLAGS_TYPE(msq->flags)) {
    case RPMMSQ_TYPE_DEFAULT:
    case RPMMSQ_TYPE_POSIX:
    {
#if defined(WITH_MQ)
	/* Queue the read. */
	rc = aioqPutWait(msq, priop,
			LIO_READ, 0, buf, count, NULL, 0);
#endif	/* WITH_MQ */
    }	break;

    case RPMMSQ_TYPE_SYSV:
    {
#if defined(WITH_MSQ)
	struct msgbuf * msgp = xmalloc(sizeof(*msgp) + count);
	size_t msgsz = count;
	long msgtyp = msq->mtype;
	int msgflg = 0;	/* IPC_NOWAIT, MSG_COPY, MSG_EXCEPT, MSG_NOERROR */

	rc = msgrcv(msq->qid, msgp, msgsz, msgtyp, msgflg);
	if (rc >= 0) {
	    if (priop)
		*priop = msgp->mtype;
	    if (rc > 0)
		memcpy(buf, msgp->mtext, rc);
	}
	msgp = _free(msgp);
#endif	/* WITH_MSQ */
    }	break;
    default:
	break;
    }

SPEW("<==\t%s(%p,%p[%lu],%p) qid %d rc %ld *priop %lu\n", __FUNCTION__, msq, buf, count, priop, (msq ? msq->qid : -1), rc, (priop ? *priop : 0));
    return rc;
}

ssize_t rpmmsqSend(rpmmsq msq, const char * buf, size_t count, unsigned long prio)
{
    ssize_t rc = -1;	/* assume failure */

    if (msq)
    switch (RPMMSQ_FLAGS_TYPE(msq->flags)) {
    case RPMMSQ_TYPE_DEFAULT:
    case RPMMSQ_TYPE_POSIX:
    {
#if defined(WITH_MQ)
	/* Send the first message to trigger the reader thread startup. */
	if (msq->nsent == 0) {
	    rc = mq_send(msq->qid, buf, count, prio);
	    if (rc == 0)
		msq->nsent++;

	    /* Yield the CPU. */
	    sched_yield();
	    struct timespec ts = { 0, 100*1000 };
	    Z(nanosleep(&ts, NULL));
	} else {
	    /* Queue the write. */
	    rc = aioqPutWait(msq, NULL,
			LIO_WRITE, 0, (char *)buf, count, NULL, 0);
	}
	if (rc == 0)	/* XXX remap to write(2) return */
	    rc = count;
#endif	/* WITH_MQ */
    }	break;
    case RPMMSQ_TYPE_SYSV:
    {
#if defined(WITH_MSQ)
	struct msgbuf * msgp = xmalloc(sizeof(*msgp) + count);
	size_t msgsz = count;
	int msgflg = 0;

	msgp->mtype = msq->mtype;
	if (count)
	    memcpy(msgp->mtext, buf, count);
	rc = msgsnd(msq->qid, msgp, msgsz, msgflg);
	if (rc == 0)
	    msq->nsent++;

	msgp = _free(msgp);
	if (rc == 0)	/* XXX remap to write(2) return */
	    rc = count;
#endif	/* WITH_MSQ */
    }	break;
    default:
	break;
    }

SPEW("<==\t%s(%p,%p[%lu],%lu) qid %d rc %ld\n", __FUNCTION__, msq, buf, count, prio, (msq ? msq->qid : -1), rc);
    return rc;
}

static char *SEEK_[] = { "SET", "CUR", "END", "???" };

static int rpmmsqSeek(rpmmsq msq, _libio_pos_t pos,
			int whence)
{
#ifdef USE_COOKIE_SEEK_POINTER
    _IO_off64_t p = *pos;
#else
    off_t p = pos;
#endif
    int rc = -2;	/* assume failure */
SPEW("<==\t%s(%p, %ld, SEEK_%s) qid %d rc %d\n", __FUNCTION__, msq, p, SEEK_[whence&0x3], (msq ? msq->qid : -1), rc);
    return rc;
}

int rpmmsqClose(rpmmsq msq, int delete)
{
    int rc = -2;	/* assume failure */

SPEW("==>\t%s(%p,%d) qid %d\n", __FUNCTION__, msq, delete, (msq ? msq->qid : -1));
    if (msq)
    switch (RPMMSQ_FLAGS_TYPE(msq->flags)) {
    case RPMMSQ_TYPE_DEFAULT:
    case RPMMSQ_TYPE_POSIX:
    {
#if defined(WITH_MQ)
	if (msq->qid != -1) {

	    if (MSQF_ISSET(INFO) || _rpmmsq_debug)
		rpmmsqDump(NULL, msq, NULL);

	    /* Reader thread never started: disable notify and close. */
	    if (msq->nsent == 0)  {
		rc = rpmmsqNotify(msq, NULL);
		rc = mq_close(msq->qid);
		if (rc == 0)
		    msq->qid = -1;
		break;
	    } else {
		/* Queue the close. */
		rc = aioqPutWait(msq, NULL,
			LIO_CLOSE, 0, NULL, 0, NULL, 0);
	    }

	    /* Yield the CPU. */
	    sched_yield();
	    struct timespec ts = { 0, 10*1000*1000 };
	    Z(nanosleep(&ts, NULL));
	}

	if (!rc && (delete || MSQF_ISSET(DELETE)))
	    rc = mq_unlink(msq->qname);		/* XXX rpmmsqReset? */
#endif	/* WITH_MQ */
    }	break;

    case RPMMSQ_TYPE_SYSV:
    {
#if defined(WITH_MSQ)
	if (msq->qid != -1) {
	    if (MSQF_ISSET(INFO) || _rpmmsq_debug)
		rpmmsqDump(__FUNCTION__, msq, NULL);
	}
	msq->qid = -1;
	rc = 0;			/* XXX */
	if (delete || MSQF_ISSET(DELETE))
	    rc = msgctl(msq->qid, IPC_RMID, NULL);
#endif	/* WITH_MSQ */
    }	break;
    default:
	break;
    }
SPEW("<==\t%s(%p,%d) qid %d rc %d\n", __FUNCTION__, msq, delete, (msq ? msq->qid : -1), rc);
    return rc;
}

rpmmsq rpmmsqOpen(const char * path, const char * fmode)
{
    rpmmsq msq = NULL;
if (strchr(fmode, '?') || _rpmmsq_debug || _rpmio_debug)
fprintf(stderr, "==> %s(%s,%s)\n", __FUNCTION__, path, fmode);
    msq = rpmmsqNew(path, fmode, -1, 0);
SPEW("<== %s(%s,%s) qid %d\n", __FUNCTION__, path, fmode, (msq ? msq->qid : -1));
    return msq;
}

rpmmsq rpmmsqFdopen(void * _fdno, const char * fmode)
{
    rpmmsq msq = NULL;
    int fdno = (int)((long)_fdno);	/* XXX hack */
if (strchr(fmode, '?') || _rpmmsq_debug || _rpmio_debug)
fprintf(stderr, "==> %s(%d,%s)\n", __FUNCTION__, fdno, fmode);
    msq = rpmmsqNew(NULL, fmode, fdno, 0);
SPEW("<== %s(%d,%s) qid %d\n", __FUNCTION__, fdno, fmode, (msq ? msq->qid : -1));
    return msq;
}

int rpmmsqFlush(void * _msq)
{
    int rc = -2;	/* assume failure */
    rpmmsq msq = (rpmmsq) _msq;
    /* XXX TODO: drain the message queue? */
SPEW("<== %s(%p) qid %d rc %d\n", __FUNCTION__, _msq, (msq ? msq->qid : -1), rc);
    return rc;
}

/* =============================================================== */
int rpmmsqCtl(void * _msq, int cmd, void *buf)
{
    rpmmsq msq = (rpmmsq) _msq;
    int rc = -2;	/* assume failure */

    if (msq)
    switch (RPMMSQ_FLAGS_TYPE(msq->flags)) {
    case RPMMSQ_TYPE_DEFAULT:
    case RPMMSQ_TYPE_POSIX:
    {
    }	break;

    case RPMMSQ_TYPE_SYSV:
    {
#if defined(WITH_MSQ)
	if (msq)
	switch (cmd) {
	default:
	    break;
#if defined(IPC_INFO)
	case IPC_INFO:
	    /* XXX rc is index of highest used message queue. */
#endif
#if defined(MSG_INFO)
	case MSG_INFO:
	    /* XXX rc is index of highest used message queue. */
#endif
#if defined(MSG_STAT)
	case MSG_STAT:
	    /* XXX 1st arg is kernel index, not qid. */
	    /* XXX rc is the qid. */
#endif
	case IPC_STAT:
	case IPC_SET:
	case IPC_RMID:
	    rc = msgctl(msq->qid, cmd, buf);
	    break;
	}
#endif	/* WITH_MSQ */
    }	break;
    default:
	break;
    }

SPEW("<== %s(%p,%d,%p) qid %d rc %d\n", __FUNCTION__, msq, cmd, buf, (msq ? msq->qid : -1), rc);
    return rc;
}

int rpmmsqNotify(rpmmsq msq, const void *_sevp)
{
    int rc = -2;	/* assume failure */

    if (msq)
    switch (RPMMSQ_FLAGS_TYPE(msq->flags)) {
    case RPMMSQ_TYPE_DEFAULT:
    case RPMMSQ_TYPE_POSIX:
    {
#if defined(WITH_MQ)
	rc = mq_notify(msq->qid, _sevp);
#endif	/* WITH_MQ */
    }	break;

    case RPMMSQ_TYPE_SYSV:
    {
    }	break;
    default:
	break;
    }

SPEW("<== %s(%p,%p) qid %d rc %d\n", __FUNCTION__, msq, _sevp, (msq ? msq->qid : -1), rc);
    return rc;
}

int rpmmsqDeliver(rpmmsq msq, char *b, size_t nb, unsigned long prio)
{
    int rc = -2;	/* assume failure */

    rc = 0;

SPEW("<== %s(%p,%p[%u],%lu) qid %d rc %d\n", __FUNCTION__, msq, b, (unsigned)nb, prio, (msq ? msq->qid : -1), rc);

    return rc;
}

static ssize_t rpmmsqReaderLoop(rpmmsq msq, int op, char *b, size_t nb, unsigned *priop)
{
    ssize_t rc = -2;	/* assume error */

#if defined(WITH_MQ)
    struct timespec ts = { 0, 1000 };	/* XXX move into msq for persistence? */
    errno = 0;
    do {
	    
	switch (op) {
	case LIO_READ:
	    rc = mq_receive(msq->qid, b, nb, priop);
	    break;
	case LIO_WRITE:
	    rc = mq_send(msq->qid, (const char *)b, nb, *priop);
	    break;
	default:
	    rc = -1;
	    errno = EINVAL;	/* XXX W2DO? */
	    break;
	}
	/* Retry operation on timeouts and queue full. */
#define	NSEC	(1000*1000*1000)
	if (rc < 0) {
	    switch (errno) {
	    case ETIMEDOUT:	/* XXX mq_timedreceive() */
		msq->ntimeout++;
		Z(nanosleep(&ts, NULL));
		ts.tv_nsec <<= 1;
		ts.tv_sec += ts.tv_nsec / NSEC;
		ts.tv_nsec %= NSEC;
		continue;
	    case EAGAIN:	/* XXX O_NONBLOCK */
		msq->nagain++;
		Z(nanosleep(&ts, NULL));
		ts.tv_nsec <<= 1;
		ts.tv_sec += ts.tv_nsec / NSEC;
		ts.tv_nsec %= NSEC;
		continue;
	    }
	}
#undef	NSEC
	break;
    } while (rc < 0);
#endif	/* WITH_MQ */

    return rc;
}

void rpmmsqReader(union sigval sv)
{
#if defined(WITH_MQ)
    rpmmsq msq = (rpmmsq) sv.sival_ptr;
assert(msq = rpmmsqLink(msq));

SPEW("==> %s(%p) qid %d\n", __FUNCTION__, msq, msq->qid);
    ssize_t rc;

    if (msq->qid != -1)
    do {
	AIOQ_t aioq = &msq->aioq;
	rpmzLog zlog = msq->zlog;
	rpmioItem item;

	/* Consumer monitor. */
	POSSESS(aioq->have);
	WAITFOR(aioq->have, NOT_TO_BE, 0);
	item = aioqGetHead(&msq->aioq);
	TWIST(aioq->have, BY, -1);
assert(item);

	/* Process the request. */
	rpmaio aio = (rpmaio) item;
	int op = aio->aiocb.aio_lio_opcode;
	char *b = (char *)aio->aiocb.aio_buf;
	size_t nb = aio->aiocb.aio_nbytes;
	unsigned int prio = aio->aiocb.aio_reqprio;	/* XXX prio */

	rc = -1;
	switch (op) {
	case LIO_READ:
	{   rc = rpmmsqReaderLoop(msq, op, b, nb, &prio);
	    if (rc >= 0) {
		msq->nrecv++;
		aio->aiocb.aio_reqprio = prio;		/* XXX prio */
	    }
	}   break;
	case LIO_WRITE:
	{   rc = rpmmsqReaderLoop(msq, op, b, nb, &prio);
	    if (rc == 0)
		msq->nsent++;
	}   break;
	case LIO_CLOSE:
	{
	    if (b == NULL && nb == 0) {
		rc = mq_close(msq->qid);
		if (rc == 0)
		    msq->qid = -1;
	    }
	}	break;
	case LIO_NOP:
	case LIO_DSYNC:
	case LIO_SYNC:
	case LIO_READ64:
	case LIO_WRITE64:
	default:
	    break;
	}

	aio->aiocb.__error_code = errno;
	aio->aiocb.__return_value = rc;
	aio = rpmaioFree(aio);
    } while (msq->qid != -1);

SPEW("<== %s(%p) qid %d rc %ld\n", __FUNCTION__, msq, msq->qid, rc);

    msq = rpmmsqFree(msq);

#endif	/* WITH_MQ */
}

static void rpmmsqDumpST(const char * msg, void * _ptr, FILE * fp)
{
    struct stat *st = (struct stat *) _ptr;
    char b[BUFSIZ];
    char * be = b;
    int colorize = isatty(fileno(stderr));

    if (fp == NULL)	fp = stderr;
#define ANSI_BRIGHT_BLUE	"\x1b[34;1m"
#define ANSI_RESET		"\x1b[0m"
    if (colorize)	be = stpcpy(be, ANSI_BRIGHT_BLUE);
    if (msg)	be += sprintf(be, "========================== %s\n", msg);

    if (st) {
#define PRINT_STAT(_fmt, _foo) \
    {	be += sprintf(be, "%25s: %"#_fmt"\n", #_foo, st->st_##_foo); }
	PRINT_STAT(lx, dev);
	PRINT_STAT(lu, ino);
	PRINT_STAT( o, mode);
	PRINT_STAT(lu, nlink);
	PRINT_STAT( d, uid);
	PRINT_STAT( d, gid);
	PRINT_STAT(lx, rdev);
	PRINT_STAT(zu, size);
	PRINT_STAT(zu, blksize);
	PRINT_STAT(zu, blocks);
	PRINT_STAT(lu, atime);
	PRINT_STAT(lu, mtime);
	PRINT_STAT(lu, ctime);
#undef	PRINT_STAT
	*(--be) = '\0';
	if (colorize) be = stpcpy(be, ANSI_RESET);
	fprintf(fp, "%s\n", b);
    }
}

static void rpmmsqDumpDS(const char * msg, void * _ptr, FILE * fp)
{
    if (fp == NULL)	fp = stderr;
    if (msg)		fprintf(fp, "================ %s\n", msg);

#if defined(WITH_MSQ)
    struct msqid_ds *ds = (struct msqid_ds *) _ptr;
    if (ds) {
#if defined(__linux__)
	fprintf(fp, "\t    key: 0x%08x\n", ds->msg_perm.__key);
#endif
	fprintf(fp, "\t    uid: %d\n", ds->msg_perm.uid);
	fprintf(fp, "\t    gid: %d\n", ds->msg_perm.gid);
	fprintf(fp, "\t   cuid: %d\n", ds->msg_perm.cuid);
	fprintf(fp, "\t   cgid: %d\n", ds->msg_perm.cgid);
	fprintf(fp, "\t   mode: 0%o\n", ds->msg_perm.mode);
#if defined(__linux__)
	fprintf(fp, "\t    seq: %d\n", ds->msg_perm.__seq);
#endif
	fprintf(fp, "\t  stime: %lu\n", ds->msg_stime);
	fprintf(fp, "\t  rtime: %lu\n", ds->msg_rtime);
	fprintf(fp, "\t  ctime: %lu\n", ds->msg_ctime);
#if defined(__linux__)
	fprintf(fp, "\t cbytes: %lu b in queue\n", ds->__msg_cbytes);
#endif
	fprintf(fp, "\t   qnum: %lu msgs in queue\n", ds->msg_qnum);
	fprintf(fp, "\t qbytes: %lu\n", ds->msg_qbytes);
	fprintf(fp, "\t  lspid: %d pid of last send\n", ds->msg_lspid);
	fprintf(fp, "\t  lrpid: %d pid of last recv\n", ds->msg_lrpid);
    }
#endif	/* WITH_MSQ */
}

static void rpmmsqDumpMI(const char * msg, void * _ptr, FILE * fp)
{
    if (fp == NULL)	fp = stderr;
    if (msg)		fprintf(fp, "================ %s\n", msg);

#if defined(WITH_MSQ) && defined(IPC_INFO)
    struct msginfo * mi = (struct msginfo *) _ptr;
    if (mi) {
	fprintf(fp, "\tmsgpool: %7d Kib buffer pool\n", mi->msgpool);
	fprintf(fp, "\t msgmap: %7d entries in msg map\n", mi->msgmap);
	fprintf(fp, "\t msgmax: %7d b max msg size\n", mi->msgmax);
	fprintf(fp, "\t msgmnb: %7d b max in queue\n", mi->msgmnb);
	fprintf(fp, "\t msgmni: %7d max no. of queues\n", mi->msgmni);
	fprintf(fp, "\t msgssz: %7d msg segment size\n", mi->msgssz);
	fprintf(fp, "\t msgtql: %7d max msgs in system\n", mi->msgtql);
	fprintf(fp, "\t msgseg: %7d max no. of segments\n", mi->msgseg);
    }
#endif	/* WITH_MSQ */
}

int rpmmsqDump(const char * msg, void *_msq, FILE *fp)
{
    rpmmsq msq = (rpmmsq) _msq;
    int rc = -2;	/* assume failure */

    if (fp == NULL)	fp = stderr;
    if (msg)		fprintf(fp, "================ %s\n", msg);

    if (msq)
    switch (RPMMSQ_FLAGS_TYPE(msq->flags)) {
    case RPMMSQ_TYPE_DEFAULT:
    case RPMMSQ_TYPE_POSIX:
    {
#if defined(WITH_MQ)
#if defined(__linux__)
	/* XXX DELETE: don't bother if buggy. */
	if (!(msq->qid != -1 && msq->qname && *msq->qname))
	    break;

	const char * lpath = rpmGetPath("/dev/mqueue/", msq->qname, NULL);

	if (msq->st)
	    free(msq->st);
	msq->st = xmalloc(sizeof(struct stat));

	rc = Stat(lpath, msq->st);
	if (!rc)
	    rpmmsqDumpST(lpath, msq->st, NULL);

	/* XXX .fdio avoids select/poll issues on /dev/mqueue with .ufdio. */
	{   FD_t fd = Fopen(lpath, "r.fdio");
	    if (fd) {
		unsigned char b[BUFSIZ];
		size_t nb = sizeof(b);
		size_t nr = Fread(b, 1, nb-1, fd);
		b[nr] = '\0';
		fprintf(stderr, "\t%.*s", (int)nr, b);
		(void) Fclose(fd);
	    }
	}

	lpath = _free(lpath);
#endif	/* __linux__ */
#endif	/* WITH_MQ */
    }	break;

    case RPMMSQ_TYPE_SYSV:
    {
#if defined(WITH_MSQ)
	if (msq->ds)
	    free(msq->ds);
	msq->ds = xmalloc(sizeof(struct msqid_ds));

	rc = rpmmsqCtl(msq, IPC_STAT, msq->ds);
	if (!rc)
	    rpmmsqDumpDS(NULL, msq->ds, NULL);

#if defined(IPC_INFO)	/* XXX linux-only? */
	if (msq->mi)
	    free(msq->mi);
	msq->mi = xmalloc(sizeof(struct msginfo));

	rc = rpmmsqCtl(msq, IPC_INFO, (struct msqid_ds *)&msq->mi);
	if (!rc)
	    rpmmsqDumpMI(NULL, &msq->mi, NULL);
#endif
#endif	/* WITH_MSQ */
    }	break;
    default:
	break;
    }

SPEW("<== %s(%p,%p) qid %d rc %d\n", __FUNCTION__, msg, msq, (msq ? msq->qid : -1), rc);
    return rc;
}

/* =============================================================== */
static
RPM_GNUC_PURE
void * msqFileno(FD_t fd)
{
    void * rc = NULL;
    int i;

    FDSANE(fd);
    for (i = fd->nfps; i >= 0; i--) {
	FDSTACK_t * fps = &fd->fps[i];
	if (fps->io != msqio)
	    continue;
	rc = fps->fp;
	break;
    }

    return rc;
}

static
FD_t msqOpen(const char * path, const char * fmode)
{
    FD_t fd;
    rpmmsq msq = rpmmsqOpen(path, fmode);

    if (msq == NULL)
	return NULL;
    fd = fdNew(__FUNCTION__);
#ifdef	NOTYET
    fdPop(fd); fdPush(fd, msqio, msq, fileno(msq->fp));
#else
    fdPop(fd); fdPush(fd, msqio, msq, -1);
#endif
    fdSetOpen(fd, path, msq->oflags, msq->omode);
#if defined(WITH_MQ)
    (void) fdSetFdno(fd, msq->qid);
#endif
    return fd;
}

static
FD_t msqFdopen(void * cookie, const char * fmode)
{
    FD_t fd = c2f(cookie);
    int fdno = fdFileno(fd);
    rpmmsq msq = rpmmsqFdopen((void *)((long)fdno), fmode);

    if (msq == NULL)
	return NULL;

    if (msq)
    switch (RPMMSQ_FLAGS_TYPE(msq->flags)) {
    case RPMMSQ_TYPE_DEFAULT:
    case RPMMSQ_TYPE_POSIX:
    {
#if defined(WITH_MQ)
	fdno = -1;
	/* XXX dup(msq->qid) ??? */
	if (msq->qid != -1) {
	    fdno = dup(msq->qid);
	    msq->qid = fdno;
	}
#endif	/* WITH_MQ */
    }	break;

    case RPMMSQ_TYPE_SYSV:
    {
#if defined(WITH_MSQ)
	fdno = -1;
#endif	/* WITH_MSQ */
    }	break;
    default:
	/* discard previous msqio? */
	fdno = -1;
	break;
    }

    fdPush(fd, msqio, msq, fdno);		/* Push msqio onto stack */
    return fdLink(fd, __FUNCTION__);
}

static
RPM_GNUC_PURE
int msqFlush(void * cookie)
{
    FD_t fd = c2f(cookie);
    rpmmsq msq = (rpmmsq) msqFileno(fd);
    return rpmmsqFlush(msq);
}

static
ssize_t msqRead(void * cookie, char * buf, size_t count)
{
    FD_t fd = c2f(cookie);
    rpmmsq msq = (rpmmsq) msqFileno(fd);
    ssize_t rc = 0;

assert(msq != NULL);
    if (fd->bytesRemain == 0) return 0;	/* XXX simulate EOF */
    fdstat_enter(fd, FDSTAT_READ);
    rc = rpmmsqRecv(msq, buf, count, NULL);
    if (rc >= 0) {
	fdstat_exit(fd, FDSTAT_READ, rc);
	if (fd->ndigests && rc > 0) fdUpdateDigests(fd, (const unsigned char *)buf, rc);
    }
    return rc;
}

static ssize_t msqWrite(void * cookie, const char * buf, size_t count)
{
    FD_t fd = c2f(cookie);
    rpmmsq msq = (rpmmsq) msqFileno(fd);
    ssize_t rc;

assert(msq != NULL);
    if (fd->bytesRemain == 0) return 0;	/* XXX simulate EOF */

    if (fd->ndigests && count > 0) fdUpdateDigests(fd, (const unsigned char *)buf, count);

    fdstat_enter(fd, FDSTAT_WRITE);
    rc = rpmmsqSend(msq, buf, count, 0);
    if (rc >= 0)
	fdstat_exit(fd, FDSTAT_WRITE, rc);
    return rc;
}

static
RPM_GNUC_PURE
int msqSeek(void * cookie, _libio_pos_t pos, int whence)
{
    FD_t fd = c2f(cookie);
    rpmmsq msq = (rpmmsq) msqFileno(fd);
#ifdef USE_COOKIE_SEEK_POINTER
    _IO_off64_t p = *pos;
#else
    off_t p = pos;
#endif

SPEW("==> %s(%p, %ld, SEEK_%s)\n", __FUNCTION__, fd, p, SEEK_[whence&0x3]);
assert(msq != NULL);
    return rpmmsqSeek(msq, pos, whence);
}

static
int msqClose(void * cookie)
{
    FD_t fd = c2f(cookie);
    rpmmsq msq = (rpmmsq) msqFileno(fd);
    int rc;

assert(msq != NULL);

    fdstat_enter(fd, FDSTAT_CLOSE);
    rc = rpmmsqClose(msq, 0);
    fdstat_exit(fd, FDSTAT_CLOSE, rc);
    fdSetFdno(fd, -1);

DBGIO(fd, (stderr, "<==\tmsqClose(%p) rc %lx %s\n", cookie, (unsigned long)rc, fdbg(fd)));

    if (_rpmio_debug || rpmIsDebug()) fdstat_print(fd, "MSQIO", stderr);

    if (rc == 0) {
	msq = rpmmsqFree(msq);
	fd = fdFree(fd, "open (msqClose)");
    }
    return rc;
}

static struct FDIO_s msqio_s = {
  msqRead, msqWrite, msqSeek, msqClose,	msqOpen, msqFdopen, msqFlush,
};

FDIO_t msqio = &msqio_s;
