/** \ingroup rpmio
 * \file rpmio/msqio.c
 * Support for SysV message queues.
 */

#include "system.h"

#include <rpmio.h>
#include <yarn.h>

#if defined(WITH_AIO)
#if defined(HAVE_AIO_H)
# include <aio.h>
#endif

#else	/* WITH_AIO */

/* XXX stub in enough to use struct aiocb_s pool */
enum {
    LIO_READ,
    LIO_WRITE,
    LIO_NOP,
};

struct aiocb
{
  int aio_fildes;		/* File desriptor.  */
  int aio_lio_opcode;		/* Operation to be performed.  */
  int aio_reqprio;		/* Request priority offset.  */
  volatile void *aio_buf;	/* Location of buffer.  */
  size_t aio_nbytes;		/* Length of transfer.  */
  struct sigevent aio_sigevent;	/* Signal number and value.  */

  /* Internal members.  */
  struct aiocb *__next_prio;
  int __abs_prio;
  int __policy;
  int __error_code;
  __ssize_t __return_value;

#ifndef __USE_FILE_OFFSET64
  __off_t aio_offset;		/* File offset.  */
  char __pad[sizeof (__off64_t) - sizeof (__off_t)];
#else
  __off64_t aio_offset;		/* File offset.  */
#endif
  char __glibc_reserved[32];
};

#endif	/* WITH_AIO */

#define	_RPMAIO_INTERNAL
#include "rpmaio.h"

#include "debug.h"

int _rpmaio_debug;

#define SPEW(_fmt, ...) \
    if (_rpmaio_debug || _rpmio_debug) \
	fprintf(stderr, _fmt, __VA_ARGS__)

#define Z(_rc)  assert((_rc) == 0)

#define PEEK(_bolt)	yarnPeekLock(_bolt)

#ifdef __cplusplus
GENfree(rpmaio)
#endif	/* __cplusplus */

static char *LIO_[] = { "READ", "WRITE", "NOP", "DSYNC", "SYNC", "CLOSE" };

static char * rpmaioDbug(void *_aio, char *b, size_t nb)
{
    rpmaio aio = (rpmaio) _aio;
    size_t len = strlen(b);
    char * be = b + len;
    rpmioItem item = (rpmioItem) aio;
    long use;

    if (aio && (use = PEEK(item->use)) > 1) {
	int colorize = isatty(fileno(stderr));
#define ANSI_BRIGHT_BLUE	"\x1b[34;1m"
#define ANSI_RESET		"\x1b[0m"
	*be++ = '\n';
	if (colorize) be = stpcpy(be, ANSI_BRIGHT_BLUE);
	be += sprintf(be, "========================== aio(%p) use %ld ix %d\n",
		aio, use, aio->ix);
#define PRINT_AIO(_fmt, _foo) \
    {	be += sprintf(be, "%25s: %"#_fmt"\n", #_foo, aio->aiocb._foo); }
#define PRINT_AIO_SIGEV(_fmt, _foo) \
    {	be += sprintf(be, "%25s: %"#_fmt"\n", #_foo, aio->aiocb.aio_sigevent._foo); }
	PRINT_AIO(d, aio_fildes);
	be += sprintf(be, "%25s: %s\n", "aio_lio_opcode",
		LIO_[aio->aiocb.aio_lio_opcode % (sizeof(LIO_)/sizeof(LIO_[0]))]);
	if (aio->aiocb.aio_reqprio)
	    PRINT_AIO(d, aio_reqprio);
	PRINT_AIO(p, aio_buf);
	PRINT_AIO(zd, aio_nbytes);
	switch (aio->aiocb.aio_sigevent.sigev_notify) {
	default:
	    break;
	case SIGEV_THREAD:
	    PRINT_AIO_SIGEV(p, sigev_value.sival_ptr);
	    PRINT_AIO_SIGEV(d, sigev_signo);
	    PRINT_AIO_SIGEV(d, sigev_notify);
	    PRINT_AIO_SIGEV(p, sigev_notify_function);
	    PRINT_AIO_SIGEV(p, sigev_notify_attributes);
	    break;
	}
#undef	PRINT_AIO_SIGEV
#undef	PRINT_AIO
	be--;
	if (colorize) be = stpcpy(be, ANSI_RESET);
	*be = '\0';
    }
    return b;
}

static void rpmaioInit(void *_aio)
{
    rpmaio aio = (rpmaio) _aio;

    if (aio) {
	memset(&aio->aiocb, 0, sizeof(aio->aiocb));
	aio->ix = -1;
    }
}

static void rpmaioFini(void *_aio)
{
    rpmaio aio = (rpmaio) _aio;

    if (aio) {
	memset(&aio->aiocb, 0, sizeof(aio->aiocb));
	aio->ix = -1;
    }
}

RPMIOPOOL_MODULE(aio)

rpmaio rpmaioNew(int fdno, int op, int prio, void *b, size_t nb,
		void * _sigev, off_t offset)
{
    rpmaio aio = rpmaioGetPool(_rpmaioPool);
    struct aiocb * aiocb = &aio->aiocb;
    aiocb->aio_fildes = fdno;
    aiocb->aio_lio_opcode = op;
    aiocb->aio_reqprio = prio;
    aiocb->aio_buf = b;
    aiocb->aio_nbytes = nb;
    if (_sigev)
	aiocb->aio_sigevent = *(struct sigevent *) _sigev;

    aiocb->__next_prio = NULL;
    aiocb->__abs_prio = 0;
    aiocb->__policy = 0;
    aiocb->__error_code = EINPROGRESS;	/* XXX */
    aiocb->__return_value = 0;

    aiocb->aio_offset = offset;

    aio->ix = -1;

    return rpmaioLink(aio);
}
