/** \ingroup rpmio
 * \file rpmio/rpmcap.c
 */

#include "system.h"

/* XXX permit building --without-cap */
#if !defined(WITH_CAP)
#undef	HAVE_SYS_CAPABILITY_H
#endif

#if defined(HAVE_SYS_CAPABILITY_H)
# include <sys/capability.h>
#endif
#if defined(WITH_CAPNG) && defined(HAVE_CAP_NG_H)
#include <cap-ng.h>
#endif

#if defined(HAVE_SYS_PRCTL_H)
# include <sys/prctl.h>
#endif
#if defined(HAVE_SYS_SYSCALL_H)
# include <sys/syscall.h>
#endif
#if defined(HAVE_LINUX_SECUREBITS_H)
#include <linux/securebits.h>
#endif

#include <rpmlog.h>
#include <rpmmacro.h>
#include <argv.h>

#include <rpmcap.h>

#include "debug.h"

int _rpmcap_debug = -1;

/* Retrofit cap_compare() if not available. */
#if defined(WITH_CAP) && !defined(HAVE_CAP_COMPARE)
static int cap_compare(cap_t acap, cap_t bcap)
{
    int rc = 0;
    size_t asize = cap_size(acap);
    size_t bsize = cap_size(bcap);

    if (asize != bsize) {
	rc = 1;
    } else {
	char *abuf = xcalloc(asize, sizeof(*abuf));
	char *bbuf = xcalloc(bsize, sizeof(*bbuf));
	cap_copy_ext(abuf, acap, asize);
	cap_copy_ext(bbuf, bcap, bsize);
	rc = memcmp(abuf, bbuf, asize);
	free(abuf);
	free(bbuf);
    }
    return rc;
}
#endif	/* WITH_CAP */

char * rpmcapValidate(const char *s)
{
    char * text = NULL;
#if defined(WITH_CAP)
    cap_t fcaps = cap_from_text(s);
    if (fcaps) {
	char *t = cap_to_text(fcaps, NULL);
	text = xstrdup(t);
	cap_free(t);
	cap_free(fcaps);
    } else
	text = xstrdup("");
#endif	/* WITH_CAP */
if (_rpmcap_debug)
fprintf(stderr, "<-- %s(%s) text %s\n", __FUNCTION__, s, text);
    return text;
}

char * rpmcapGetFD(int fdno)
{
    char * text = NULL;
#if defined(WITH_CAP)
    cap_t fcaps = cap_get_fd(fdno);
    if (fcaps) {
	char *t = cap_to_text(fcaps, NULL);
	text = xstrdup(t);
	cap_free(t);
	cap_free(fcaps);
    }
#endif	/* WITH_CAP */
if (_rpmcap_debug)
fprintf(stderr, "<-- %s(%d) text %s\n", __FUNCTION__, fdno, text);
    return text;
}

char * rpmcapGetFN(const char *fn)
{
    char * text = NULL;
#if defined(WITH_CAP)
    cap_t fcaps = cap_get_file(fn);
    if (fcaps) {
	char *t = cap_to_text(fcaps, NULL);
	text = xstrdup(t);
	cap_free(t);
	cap_free(fcaps);
    }
#endif	/* WITH_CAP */
if (_rpmcap_debug)
fprintf(stderr, "<-- %s(%s) text %s\n", __FUNCTION__, fn, text);
    return text;
}

int rpmcapSetFD(int fdno, const char *s)
{
    int rc = -1;		/* assume failure */
#if defined(WITH_CAP)
    cap_t fcaps = NULL;
    if (s && *s != '\0') {
        fcaps = cap_from_text(s);
	if (fcaps == NULL
         || cap_set_fd(fdno, fcaps) != 0)
	    goto exit;
    }
    rc = 0;
exit:
    if (fcaps)
        cap_free(fcaps);
#endif	/* WITH_CAP */
if (_rpmcap_debug)
fprintf(stderr, "<-- %s(%d,%s) rc %d\n", __FUNCTION__, fdno, s, rc);
    return rc;
}

int rpmcapSetFN(const char *fn, const char *s)
{
    int rc = -1;		/* assume failure */
#if defined(WITH_CAP)
    cap_t fcaps = NULL;
    if (s && *s != '\0') {
        fcaps = cap_from_text(s);
	if (fcaps == NULL
         || cap_set_file(fn, fcaps) != 0)
	    goto exit;
    }
    rc = 0;
exit:
    if (fcaps)
        cap_free(fcaps);
#endif	/* WITH_CAP */
if (_rpmcap_debug)
fprintf(stderr, "<-- %s(%s,%s) rc %d\n", __FUNCTION__, fn, s, rc);
    return rc;
}

int rpmcapVerifyFD(int fdno, const char *s)
{
    int rc = -1;		/* assume failure */
#if defined(WITH_CAP)
    cap_t fcap = cap_get_fd(fdno);
    cap_t cap = cap_from_text(s);
    /*
     * Empty capability set ("=") is not exactly the same as no
     * capabilities at all but suffices for now...
     */
    if (cap == NULL)
	cap = cap_from_text("=");
    if (fcap == NULL)
	fcap = cap_from_text("=");
	
    rc = (cap_compare(cap, fcap) != 0);

    cap_free(fcap);
    cap_free(cap);
#endif	/* WITH_CAP */
if (_rpmcap_debug)
fprintf(stderr, "<-- %s(%d,%s) rc %d\n", __FUNCTION__, fdno, s, rc);
    return rc;
}

int rpmcapVerifyFN(const char *fn, const char *s)
{
    int rc = -1;		/* assume failure */
#if defined(WITH_CAP)
    cap_t fcap = cap_get_file(fn);
    cap_t cap = cap_from_text(s);
    /*
     * Empty capability set ("=") is not exactly the same as no
     * capabilities at all but suffices for now...
     */
    if (cap == NULL)
	cap = cap_from_text("=");
    if (fcap == NULL)
	fcap = cap_from_text("=");
	
    rc = (cap_compare(cap, fcap) != 0);

    cap_free(fcap);
    cap_free(cap);
#endif	/* WITH_CAP */
if (_rpmcap_debug)
fprintf(stderr, "<-- %s(%s,%s) rc %d\n", __FUNCTION__, fn, s, rc);
    return rc;
}

char * rpmcapGetP(pid_t pid)
{
    char * text = NULL;
#if defined(WITH_CAP)
    cap_t pcaps = (pid > 0 ? cap_get_pid(pid) : cap_get_proc());
    if (pcaps) {
	char *t = cap_to_text(pcaps, NULL);
	text = xstrdup(t);
	cap_free(t);
	cap_free(pcaps);
    }
#endif	/* WITH_CAP */
if (_rpmcap_debug)
fprintf(stderr, "<-- %s(%d) text %s\n", __FUNCTION__, pid, text);
    return text;
}

int rpmcapSetP(pid_t pid, const char *s)
{
    int rc = -1;		/* assume failure */
#if defined(WITH_CAP)
    cap_t pcaps = NULL;
    if (pid <= 0)		/* XXX use capsetp() ? */
	goto exit;
    if (s && *s != '\0') {
        pcaps = cap_from_text(s);
	if (pcaps == NULL
         || cap_set_proc(pcaps) != 0)
	    goto exit;
    }
    rc = 0;
exit:
    if (pcaps)
        cap_free(pcaps);
#endif	/* WITH_CAP */
if (_rpmcap_debug)
fprintf(stderr, "<-- %s(%d,%s) rc %d\n", __FUNCTION__, pid, s, rc);
    return rc;
}

int Prctl(int option, unsigned long arg2, unsigned long arg3,
                 unsigned long arg4, unsigned long arg5)
{
    int rc = -2;
#if defined(HAVE_PRCTL)
    rc = prctl(option, arg2, arg3, arg4, arg5);
#elif defined(__linux__) && defined(SYS_prctl)
    rc = syscall(SYS_prctl, option, arg2, arg3, arg4, arg5);
#else
#error No prctl(2) avaliable.
    errno = ENOSYS;
#endif
    if (rc < 0)
	rpmlog(RPMLOG_ERR, _("%s(%d,%lu,%lu,%lu,%lu) failed: %m\n"),
		__FUNCTION__, option, arg2, arg3, arg4, arg5);
if (_rpmio_debug)
fprintf(stderr, _("<-- %s(%d,%lu,%lu,%lu,%lu) rc %d\n"), __FUNCTION__, option, arg2, arg3, arg4, arg5, rc);
    return rc;
}

int rpmcapDump(int lvl, const char * msg)
{
    int rc = -1;
#if defined(WITH_CAPNG)
    if (msg)
	rpmlog(lvl, "--- %s\n", msg);

    if (capng_get_caps_process())
	goto exit;

    if (capng_have_capabilities(CAPNG_SELECT_CAPS) == CAPNG_NONE)
	rpmlog(lvl, "Current capabilities: none\n");
    else {
	char * cs =
		capng_print_caps_numeric(CAPNG_PRINT_BUFFER, CAPNG_SELECT_BOTH);
	ARGV_t av = NULL;
	(void) argvSplit(&av, cs, "\n");
	int ac = argvCount(av);
	for (int i = 0; i < ac; i++)
	    rpmlog(lvl, "%s\n", av[i]);
	av = argvFree(av);
	cs = _free(cs);
    }

#ifdef PR_SET_SECUREBITS
    const char *ncomma = "";
    char b[BUFSIZ];
    char * be;

    be = b;
    be = stpcpy(be, "securebits flags: ");
    rc = prctl(PR_GET_SECUREBITS, 1 << SECURE_NOROOT);
    if (rc & (1 << SECURE_NOROOT)) {
	be = stpcpy(stpcpy(be, ncomma), "NOROOT");
	ncomma = ", ";
    }
    rc = prctl(PR_GET_SECUREBITS, 1 << SECURE_NOROOT_LOCKED);
    if (rc & (1 << SECURE_NOROOT_LOCKED)) {
	be = stpcpy(stpcpy(be, ncomma), "NOROOT_LOCKED");
	ncomma = ", ";
    }
    rc = prctl(PR_GET_SECUREBITS, 1 << SECURE_NO_SETUID_FIXUP);
    if (rc & (1 << SECURE_NO_SETUID_FIXUP)) {
	be = stpcpy(stpcpy(be, ncomma), "NO_SETUID_FIXUP");
	ncomma = ", ";
    }
    rc = prctl(PR_GET_SECUREBITS, 1 << SECURE_NO_SETUID_FIXUP_LOCKED);
    if (rc & (1 << SECURE_NO_SETUID_FIXUP_LOCKED)) {
	be = stpcpy(stpcpy(be, ncomma), "NO_SETUID_FIXUP_LOCKED");
	ncomma = ", ";
    }
    if (*ncomma != ',')
	be = stpcpy(be, "none");
    be = stpcpy(be, "\n");
    rpmlog(lvl, "%s", b);
#endif
    rc = 0;
exit:
#endif	/* WITH_CAPNG */
    return rc;
}

int rpmcapTest(const char * s)
{
    int rc = -1;
#if defined(WITH_CAP)
    cap_value_t pcap = 0;
    if (cap_from_name(s, &pcap) != -1)
	rc = cap_get_bound(pcap);
#endif	/* WITH_CAP */
#if defined(WITH_CAPNG)
    int _lvl = RPMLOG_NOTICE;
    const char *name = s;
    if (!strncasecmp(name, "CAP_", sizeof("CAP_")-1))
	name += sizeof("CAP_")-1;
    int cap = capng_name_to_capability(name);
    if (cap >= 0) {
	rpmcapDump(_lvl, "rpmcapTest");
    }
#endif	/* WITH_CAPNG */
if (_rpmcap_debug)
fprintf(stderr, "<-- %s(%s) rc %d\n", __FUNCTION__, s, rc);
    return rc;
}

int rpmcapDrop(const char * s)
{
    int rc = -1;
#if defined(WITH_CAP_XXX)
    cap_value_t pcap = 0;
    if (cap_from_name(s, &pcap) != -1)
	rc = cap_drop_bound(pcap);
#endif	/* WITH_CAP */
#if defined(WITH_CAPNG)
    int _lvl = RPMLOG_NOTICE;
    const char *name = s;
    if (!strncasecmp(name, "CAP_", sizeof("CAP_")-1))
	name += sizeof("CAP_")-1;
    int cap = capng_name_to_capability(name);
    if (cap >= 0) {
	rpmcapDump(_lvl, "rpmcapDrop before");

	capng_clear(CAPNG_SELECT_BOTH);
	capng_type_t type = CAPNG_EFFECTIVE|CAPNG_PERMITTED;
	rc = capng_update(CAPNG_DROP, type, cap);
	rc = capng_apply(CAPNG_SELECT_BOTH);

	rpmcapDump(_lvl, "rpmcapDrop  after");
    }
#endif	/* WITH_CAP */
if (_rpmcap_debug)
fprintf(stderr, "<-- %s(%s) rc %d\n", __FUNCTION__, s, rc);
    return rc;
}
