/** \ingroup rpmio
 * \file rpmio/rpmio.c
 */

#include "system.h"
#include <stdarg.h>

#if defined(HAVE_SYS_FILE_H)
#include <sys/file.h>
#endif

#if defined(HAVE_MACHINE_TYPES_H)
# include <machine/types.h>
#endif

#if defined(HAVE_SYS_SOCKET_H)
# include <sys/socket.h>
#endif

#if defined(HAVE_SYS_SYSCALL_H)
# include <sys/syscall.h>
#endif

#if defined(HAVE_SYS_XATTR_H)
# include <sys/xattr.h>
#endif

#ifndef NI_MAXHOST
#define NI_MAXHOST      1025
#endif

#include <netdb.h>		/* XXX getaddrinfo et al */

#include <netinet/in.h>
#include <arpa/inet.h>		/* XXX for inet_aton and HP-UX */

#if defined(HAVE_NETINET_IN_SYSTM_H)
# include <sys/types.h>
# include <netinet/in_systm.h>
#endif

#include <rpmiotypes.h>
#include <rpmmacro.h>		/* XXX rpmioAccess needs rpmCleanPath() */

#include <rpmgfs.h>

#include <rpmaug.h>
#include <rpmct.h>
#include <rpmdate.h>
#include <rpmficl.h>
#include <rpmgit.h>
#include <rpmjni.h>
#include <rpmjss.h>
#include <rpmlua.h>		/* XXX rpmioClean() calls rpmluaFree() */
#include <rpmmqtt.h>
#include <rpmmrb.h>
#include <rpmnix.h>
#include <rpmodbc.h>
#include <rpmperl.h>
#include <rpmpython.h>
#include <rpmruby.h>
#include <rpmsed.h>
#include <rpmsql.h>
#include <rpmsquirrel.h>
#include <rpmtcl.h>

#define	_RPMHKP_INTERNAL	/* XXX awol/crl bloom filters */
#include <rpmhkp.h>

#include <rpmsm.h>
#include <rpmsp.h>
#include <rpmsx.h>

#if defined(HAVE_FOPENCOOKIE) && defined(HAVE_LIBIO_H) && defined(_G_IO_IO_FILE_VERSION)
#define	_USE_LIBIO	1
#elif defined(HAVE_FUNOPEN)
#define	_USE_LIBIO	1
#else
#define	_USE_LIBIO	0
#endif

/* XXX HP-UX w/o -D_XOPEN_SOURCE needs */
#if !defined(HAVE_HERRNO) && (defined(hpux) || defined(__hpux))
extern int h_errno;
#endif

#ifndef IPPORT_FTP
#define IPPORT_FTP	21
#endif
#ifndef	IPPORT_HTTP
#define	IPPORT_HTTP	80
#endif

#if !defined(HAVE_INET_ATON)
#define inet_aton(cp,inp) rpm_inet_aton(cp,inp)
static int rpm_inet_aton(const char *cp, struct in_addr *inp)
{
    long addr;

    addr = inet_addr(cp);
    if (addr == ((long) -1)) return 0;

    memcpy(inp, &addr, sizeof(addr));
    return 1;
}
#endif

#if defined(USE_ALT_DNS) && USE_ALT_DNS
#include "dns.h"
#endif

#include <rpmio_internal.h>
#undef	fdFileno
#undef	fdOpen
#define	fdOpen	__fdOpen
#undef	fdRead
#define	fdRead	__fdRead
#undef	fdWrite
#define	fdWrite	__fdWrite
#undef	fdClose
#define	fdClose	__fdClose

#include <ugid.h>
#include <rpmcb.h>
#include <rpmdav.h>

#include "debug.h"

#define FDTO(fd)	(fd ? ((FD_t)fd)->rd_timeoutsecs : -99)
#define FDCPIOPOS(fd)	(fd ? ((FD_t)fd)->fd_cpioPos : -99)

#define	FDONLY(fd)	assert(fdGetIo(fd) == fdio)

#define	UFDONLY(fd)	/* assert(fdGetIo(fd) == ufdio) */

#define	fdGetFILE(_fd)	((FILE *)fdGetFp(_fd))

/**
 */
#if _USE_LIBIO		/* XXX FIXME: inverted logic */
int noLibio = 0;
#else
int noLibio = 1;
#endif

#define TIMEOUT_SECS 60

/**
 */
static int ftpTimeoutSecs = TIMEOUT_SECS;

/**
 */
int _rpmio_debug = 0;

/**
 */
int _av_debug = 0;

/**
 */
int _ftp_debug = 0;

/**
 */
int _dav_debug = 0;

/* =============================================================== */

const char * fdbg(FD_t fd)
{
    static char buf[BUFSIZ];
    char *be = buf;
    int i;

    buf[0] = '\0';
    if (fd == NULL)
	return buf;

#ifdef DYING
    sprintf(be, "fd %p", fd);	be += strlen(be);
    if (fd->rd_timeoutsecs >= 0) {
	sprintf(be, " secs %d", fd->rd_timeoutsecs);
	be += strlen(be);
    }
#endif
    if (fd->bytesRemain != -1) {
	sprintf(be, " clen %d", (int)fd->bytesRemain);
	be += strlen(be);
    }
    if (fd->wr_chunked) {
	strcpy(be, " chunked");
	be += strlen(be);
    }
    *be++ = '\t';
    for (i = fd->nfps; i >= 0; i--) {
	FDSTACK_t * fps = &fd->fps[i];
	if (i != fd->nfps)
	    *be++ = ' ';
	*be++ = '|';
	*be++ = ' ';
	if (fps->io == fdio) {
	    sprintf(be, "FD %d fp %p", fps->fdno, fps->fp);
	} else if (fps->io == ufdio) {
	    sprintf(be, "UFD %d fp %p", fps->fdno, fps->fp);
#if defined(WITH_ZLIB)
	} else if (fps->io == gzdio) {
	    sprintf(be, "GZD %p fdno %d", fps->fp, fps->fdno);
#endif
#if defined(WITH_BZIP2)
	} else if (fps->io == bzdio) {
	    sprintf(be, "BZD %p fdno %d", fps->fp, fps->fdno);
#endif
#if defined(WITH_XZ)
	} else if (fps->io == lzdio) {
	    sprintf(be, "LZD %p fdno %d", fps->fp, fps->fdno);
	} else if (fps->io == xzdio) {
	    sprintf(be, "XZD %p fdno %d", fps->fp, fps->fdno);
#endif
#if defined(WITH_ZSTD)
	} else if (fps->io == zstdio) {
	    sprintf(be, "ZSTD %p fdno %d", fps->fp, fps->fdno);
#endif
#if defined(WITH_MQ) || defined(WITH_MSQ)
	} else if (fps->io == msqio) {
	    sprintf(be, "MSQ %p fdno %d", fps->fp, fps->fdno);
#endif
	} else if (fps->io == fpio) {
	    sprintf(be, "%s %p(%d) fdno %d",
		(fps->fdno < 0 ? "LIBIO" : "FP"),
		fps->fp, fileno(((FILE *)fps->fp)), fps->fdno);
	} else {
	    sprintf(be, "??? io %p fp %p fdno %d ???",
		fps->io, fps->fp, fps->fdno);
	}
	be += strlen(be);
	*be = '\0';
    }
    return buf;
}

/* =============================================================== */
off_t fdSize(FD_t fd)
{
    struct stat sb;
    off_t rc = -1;

    FDSANE(fd);
    if (fstat(Fileno(fd), &sb) == 0)
        rc = sb.st_size;
    return rc;
}

FD_t fdDup(int fdno)
{
    FD_t fd;
    int nfdno;

    if ((nfdno = dup(fdno)) < 0)
	return NULL;
    if (fcntl(nfdno, F_SETFD, FD_CLOEXEC)) {
	(void) close(nfdno);
	return NULL;
    }
    fd = fdNew("open (fdDup)");
    fdSetOpen(fd, "fdDup", 0, 0);	/* XXX bogus */
    (void) fdSetFdno(fd, nfdno);
DBGIO(fd, (stderr, "<-- fdDup(%d) fd %p %s\n", fdno, (fd ? fd : NULL), fdbg(fd)));
    return fd;
}

static char *SEEK_[] = { "SET", "CUR", "END", "???" };

static inline
int fdSeekNot(void * cookie, _libio_pos_t pos, int whence)
{
    FD_t fd = c2f(cookie);
    FDSANE(fd);		/* XXX keep gcc quiet */
    return -2;
}

/* =============================================================== */

static void fdFini(void * _fd)
{
    FD_t fd = (FD_t) _fd;
    int i;

assert(fd != NULL);
    fd->opath = _free(fd->opath);
    if (fd->stats)
	free(fd->stats);
    fd->stats = NULL;
    if (fd->ndigests > 0)
    for (i = fd->ndigests - 1; i >= 0; i--) {
	DIGEST_CTX ctx = fd->digests[i];
	if (ctx == NULL)
	    continue;
	(void) rpmDigestFinal(ctx, NULL, NULL, 0);
	fd->digests[i] = NULL;
    }
    fd->nio = 0;
    if (fd->io)	free(fd->io);
    fd->io = NULL;
    fd->digests = _free(fd->digests);
    fd->ndigests = 0;
    fd->contentType = _free(fd->contentType);
    fd->contentDisposition = _free(fd->contentDisposition);
    fd->xar = rpmxarFree(fd->xar, "fdFini");
#ifdef WITH_NEON
#ifndef	NOTYET
if (fd->req != NULL && fd->req != (void *)-1)
fprintf(stderr, "*** %s: fd->req %p\n", __FUNCTION__, fd->req);
fd->req = NULL;
#else
assert(fd->req == NULL);
#endif
#endif
    fd->dig = pgpDigFree(fd->dig);
}

rpmioPool _fdPool;

static char * fdDbug(void * _fd, char *b, size_t nb)
{
    FD_t fd = (FD_t) _fd;
    size_t len = strlen(b);
    char * be = b + len;
    *be++ = ' ';
    be = stpncpy(be, fdbg(fd), nb-len-1);
    return b;
}

static void fdInit(void * _fd)
{
}

static FD_t fdGetPool(rpmioPool pool)
{
    FD_t fd;

    if (_fdPool == NULL) {
	_fdPool = rpmioNewPool("fd", sizeof(*fd), -1, _rpmio_debug,
		fdDbug, fdInit, fdFini);
	pool = _fdPool;
    }
    fd = (FD_t) rpmioGetPool(pool, sizeof(*fd));
    memset(((char *)fd)+sizeof(fd->_item), 0, sizeof(*fd)-sizeof(fd->_item));
    return fd;
}

FD_t XfdNew(const char * msg, const char * fn, unsigned ln)
{
    FD_t fd = fdGetPool(_fdPool);
    if (fd == NULL) /* XXX xmalloc never returns NULL */
	return NULL;
    fd->flags = 0;
    fd->magic = FDMAGIC;

    fd->nfps = 0;
    memset(fd->fps, 0, sizeof(fd->fps));

    fd->fps[0].io = ufdio;
    fd->fps[0].fp = NULL;
    fd->fps[0].fdno = -1;

    fd->u = NULL;
    fd->req = NULL;
    fd->rd_timeoutsecs = 1;	/* XXX default value used to be -1 */
    fd->bytesRemain = -1;
    fd->contentLength = -1;
    fd->persist = 0;
    fd->wr_chunked = 0;

    fd->syserrno = 0;
    fd->errcookie = NULL;

    fd->opath = NULL;
    fd->oflags = 0;
    fd->omode = 0;

    fd->nio = 0;
    fd->io = NULL;
    if (fd->nio > 0) {
#if defined(HAVE_POSIX_MEMALIGN)
	assert(posix_memalign((void **)&fd->io, BUFSIZ, fd->nio) == 0);
#else
	fd->io = xmalloc(fd->nio);
#endif
    }

    fd->xar = NULL;
    fd->dig = NULL;
    fd->stats = (FDSTAT_t) xcalloc(1, sizeof(*fd->stats));
    fd->ndigests = 0;
    fd->digests = NULL;

    fd->contentType = NULL;
    fd->contentDisposition = NULL;
    fd->lastModified = 0;
    fd->ftpFileDoneNeeded = 0;
    fd->fd_cpioPos = 0;

    return (FD_t)rpmioLinkPoolItem((rpmioItem)fd, msg, fn, ln);
}

static int fdChkAlign(FD_t fd, const void *buf, size_t count)
{
    static unsigned long m = (512 - 1);
    unsigned long slop = ((unsigned long)buf & m) + (count & m);
    int rc = (fd->flags & O_DIRECT) && slop ? 0 : 1;
DBGIO(fd, (stderr, "<--\tfdChkAlign(%p,%p,%lx) rc %d\n", fd, buf, (long)count, rc));
    return rc;
}

static ssize_t fdRead(void * cookie, char * buf, size_t count)
{
    static const int opx = FDSTAT_READ;
    FD_t fd = c2f(cookie);
    ssize_t rc = -1;	/* assume failure */

    /* XXX handle fd->req similar to ufdRead() */
    if (fd->bytesRemain == 0) return 0;	/* XXX simulate EOF */

    fdstat_enter(fd, opx);

    if (!fdChkAlign(fd, buf, count)) {
	errno = EINVAL;
    } else
#if defined(WITH_NEON)
    if (fd->req != NULL) {	/* HACK: flimsy wiring for davRead */
	if (fd->req != (void *)-1)
	    rc = davRead(fd, buf, (count > (size_t)fd->bytesRemain ? (size_t)fd->bytesRemain : count));
	/* XXX Chunked davRead EOF. */
	if (rc == 0)
	    fd->bytesRemain = 0;
    } else
#endif	/* WITH_NEON */
#if defined(WITH_XAR)
    if (fd->xar != NULL) {
	rc = xarRead(fd, buf, (count > (size_t)fd->bytesRemain ? (size_t)fd->bytesRemain : count));
    } else
#endif	/* WITH_XAR */
	rc = read(fdFileno(fd), buf, (count > (size_t)fd->bytesRemain ? (size_t)fd->bytesRemain : count));

    fdstat_exit(fd, opx, rc);

    if (fd->ndigests > 0 && rc > 0) fdUpdateDigests(fd, (const unsigned char *)buf, rc);

DBGIO(fd, (stderr, "<--\tfdRead(%p,%p,%ld) rc %ld %s\n", cookie, buf, (long)count, (long)rc, fdbg(fd)));

    return rc;
}

static ssize_t fdWrite(void * cookie, const char * buf, size_t count)
{
    static const int opx = FDSTAT_WRITE;
    FD_t fd = c2f(cookie);
    int fdno = fdFileno(fd);
    /* XXX Round up last write to block boundary. */
    static size_t blksize = 512;
    /* XXX O_DIRECT requires length padding. */
    size_t pad = (fd->flags & O_DIRECT) && (count % blksize)
	? 512 - (count % blksize)
	: 0;
    ssize_t rc = -1;	/* assume failure */

    if (fd->bytesRemain == 0) return 0;	/* XXX simulate EOF */

    if (fd->ndigests > 0 && count > 0) fdUpdateDigests(fd, (const unsigned char *)buf, count);

    if (count == 0) return 0;

    fdstat_enter(fd, opx);

    if (!fdChkAlign(fd, buf, count+pad)) {
	errno = EINVAL;
    } else
#if defined(WITH_NEON)
    if (fd->req != NULL)		/* HACK: flimsy wiring for davWrite */
	if (fd->req != (void *)-1)
	    rc = davWrite(fd, buf, (count > (size_t)fd->bytesRemain ? (size_t)fd->bytesRemain : count));
	else
	rc = -1;
    else
#endif	/* WITH_NEON */
    {	/* XXX O_DIRECT requires length padding. */
	size_t nw = count;
	if (pad) nw += pad;
	rc = write(fdno, buf, (nw > (size_t)fd->bytesRemain ? (size_t)fd->bytesRemain : nw));
	if (pad) rc -= pad;
    }

    fdstat_exit(fd, opx, rc);

DBGIO(fd, (stderr, "<--\tfdWrite(%p,%p,%ld) rc %ld %s\n", cookie, buf, (long)count, (long)rc, fdbg(fd)));

    return rc;
}

static int fdSeek(void * cookie, _libio_pos_t pos, int whence)
{
    static const int opx = FDSTAT_SEEK;
#ifdef USE_COOKIE_SEEK_POINTER
    _IO_off64_t p = *pos;
#else
    off_t p = pos;
#endif
    FD_t fd = c2f(cookie);
    off_t rc;

    assert(fd->bytesRemain == -1);	/* XXX FIXME fdio only for now */
    fdstat_enter(fd, opx);
    rc = lseek(fdFileno(fd), p, whence);
    fdstat_exit(fd, opx, rc);

DBGIO(fd, (stderr, "<--\tfdSeek(%p, %ld, SEEK_%s) rc %lx %s\n", cookie, (long)p, SEEK_[whence&0x3], (unsigned long)rc, fdbg(fd)));

    return (rc >= 0 ? 0 : rc);
}

static int fdSync(FD_t fd)
{
    int rc = -2;
    int fdno = fdFileno(fd);
    int nincore = 0;

    if (fdno < 0) {				/* XXX needed? */
	errno = EBADF;
	goto exit;
    }

    if (RPMFD_ISSET(fd, DEBUGIO))
	nincore = Fincore(fd, nincore);
    if (RPMFD_ISSET(fd, FDATASYNC)) {
	rc = Fdatasync(fd);
	if (rc < 0 && errno != ENOSYS)		/* XXX best effort */
	    goto exit;
    }
    if (RPMFD_ISSET(fd, FADVISE)) {
#if defined(POSIX_FADV_DONTNEED)
	rc = Fadvise(fd, 0, 0, POSIX_FADV_DONTNEED);
#endif  
	if (rc < 0 && errno != ENOSYS)		/* XXX best effort */
	    goto exit;
    }
    if (RPMFD_ISSET(fd, FSYNC)) {
	rc = Fsync(fd);
	if (rc < 0 && errno != ENOSYS)		/* XXX best effort */
	    goto exit;
    }
    if (RPMFD_ISSET(fd, SYNCFS)) {
	rc = Syncfs(fd);
	if (rc < 0 && errno != ENOSYS)		/* XXX best effort */
	    goto exit;
    }
    if (RPMFD_ISSET(fd, DEBUGIO))
	nincore -= Fincore(fd, nincore);
    rc = 0;

exit:
    return rc;
}

static int fdClose(void * cookie)
{
    static const int opx = FDSTAT_CLOSE;
    FD_t fd;
    int fdno;
    int rc;

    if (cookie == NULL) return -2;
    fd = c2f(cookie);

    /* Perform sync-on-close (if requested). */
    if (RPMFD_ISSET(fd, FSYNC)
     || RPMFD_ISSET(fd, SYNCFS))
	rc = fdSync(fd);

    fdno = fdSetFdno(fd, -1);

    fdstat_enter(fd, opx);
    /* HACK: flimsy wiring for davClose */
    if (fd->req != NULL) {
#ifdef WITH_NEON
	rc = davClose(fd);
#else
	rc = -1;
#endif
    } else
	rc = ((fdno >= 0) ? close(fdno) : -2);
    fdstat_exit(fd, opx, rc);

DBGIO(fd, (stderr, "<--\tfdClose(%p) rc %lx %s\n", (fd ? fd : NULL), (unsigned long)rc, fdbg(fd)));
    if (!rc && (_rpmio_debug || rpmIsDebug())) fdstat_print(fd, " FDIO", stderr);

    fd = fdFree(fd, "open (fdClose)");
    return rc;
}

int _rpmioOMask = (O_ACCMODE
	/* -- linux */
#ifdef	O_APPEND
	|O_APPEND
#endif
#ifdef	O_ASYNC
	|O_ASYNC
#endif
#ifdef	O_CLOEXEC
	|O_CLOEXEC
#endif
#ifdef	O_CREAT
	|O_CREAT
#endif
#ifdef	O_DIRECT
	|O_DIRECT
#endif
#ifdef	O_DSYNC
	|O_DSYNC
#endif
#ifdef	O_EXCL
	|O_EXCL
#endif
#ifdef	O_LARGEFILE
	|O_LARGEFILE
#endif
#ifdef	O_NOATIME
	|O_NOATIME
#endif
#ifdef	O_NOCTTY
	|O_NOCTTY
#endif
#ifdef	O_NOFOLLOW
	|O_NOFOLLOW
#endif
#ifdef	O_NONBLOCK
	|O_NONBLOCK
#endif
#ifdef	O_NDELAY
	|O_NDELAY
#endif

#ifdef	O_DIRECTORY
	|O_DIRECTORY
#endif
#ifdef	O_SYNC
	|O_SYNC
#endif
#ifdef	O_PATH
	|O_PATH
#endif

#ifdef	O_TMPFILE
	|O_TMPFILE
#endif

#ifdef	O_TRUNC
	|O_TRUNC
#endif
	/* -- OS X */
#ifdef	O_SHLOCK
	|O_SHLOCK
#endif
#ifdef	O_EXLOCK
	|O_EXLOCK
#endif
#ifdef	O_SYMLINK
	|O_SYMLINK
#endif
#ifdef	O_EVTONLY
	|O_EVTONLY
#endif
    );

static FD_t fdOpen(const char *path, int flags, mode_t mode)
{
    FD_t fd;
    int fdno;
    int oflags = (flags & _rpmioOMask);
    /* XXX 23 bits for linux, 9 bits to rpm */
#if defined(O_TMPFILE)
    oflags &= ((O_TMPFILE << 1) -1);
#else
    oflags &= 0x07ffffff;
#endif

    {
	fdno = open(path, oflags, mode);
    }
    if (fdno < 0) return NULL;
    if (fcntl(fdno, F_SETFD, FD_CLOEXEC)) {
	(void) close(fdno);
	return NULL;
    }
    fd = fdNew("open (fdOpen)");
    fdSetOpen(fd, path, oflags, mode);
    (void) fdSetFdno(fd, fdno);
assert(fd != NULL);
    fd->flags = flags;
DBGIO(fd, (stderr, "<--\tfdOpen(\"%s\",%x,0%o) %s\n", path, (unsigned)flags, (unsigned)mode, fdbg(fd)));
    return fd;
}

#ifdef NOTUSED
FILE *fdFdopen(void * cookie, const char *fmode)
{
    FD_t fd = c2f(cookie);
    int fdno;
    FILE * fp;

    if (fmode == NULL) return NULL;
    fdno = fdFileno(fd);
    if (fdno < 0) return NULL;
    fp = fdopen(fdno, fmode);
DBGIO(fd, (stderr, "<-- fdFdopen(%p,\"%s\") fdno %d -> fp %p fdno %d\n", cookie, fmode, fdno, fp, fileno(fp)));
    fd = fdFree(fd, "open (fdFdopen)");
    return fp;
}
#endif

static struct FDIO_s fdio_s = {
  fdRead, fdWrite, fdSeek, fdClose, NULL, NULL, NULL,
};

FDIO_t fdio = &fdio_s;

int fdWritable(FD_t fd, int secs)
{
    int fdno;
    int rc;
#if defined(HAVE_POLL_H)
    int msecs = (secs >= 0 ? (1000 * secs) : -1);
    struct pollfd wrfds;
#else
    struct timeval timeout, *tvp = (secs >= 0 ? &timeout : NULL);
    fd_set wrfds;
    FD_ZERO(&wrfds);
#endif
	
    /* HACK: flimsy wiring for davWrite */
    if (fd->req != NULL)
	return (fd->req == (void *)-1 ? -1 : 1);

    if ((fdno = fdFileno(fd)) < 0)
	return -1;	/* XXX W2DO? */
	
    do {
#if defined(HAVE_POLL_H)
	wrfds.fd = fdno;
	wrfds.events = POLLOUT;
	wrfds.revents = 0;
	rc = poll(&wrfds, 1, msecs);
#else
	if (tvp) {
	    tvp->tv_sec = secs;
	    tvp->tv_usec = 0;
	}
	FD_SET(fdno, &wrfds);
	rc = select(fdno + 1, NULL, &wrfds, NULL, tvp);
#endif

	/* HACK: EBADF on PUT chunked termination from ufdClose. */
if (_rpmio_debug && !(rc == 1 && errno == 0))
fprintf(stderr, "*** fdWritable fdno %d rc %d %s\n", fdno, rc, strerror(errno));
	if (rc < 0) {
	    switch (errno) {
	    case EINTR:
		continue;
		break;
	    default:
		return rc;
		break;
	    }
	}
	return rc;
    } while (1);
}

int fdReadable(FD_t fd, int secs)
{
    int fdno;
    int rc;
#if defined(HAVE_POLL_H)
    int msecs = (secs >= 0 ? (1000 * secs) : -1);
    struct pollfd rdfds;
#else
    struct timeval timeout, *tvp = (secs >= 0 ? &timeout : NULL);
    fd_set rdfds;
    FD_ZERO(&rdfds);
#endif

    /* HACK: flimsy wiring for davRead */
    if (fd->req != NULL)
	return (fd->req == (void *)-1 ? -1 : 1);

    if ((fdno = fdFileno(fd)) < 0)
	return -1;	/* XXX W2DO? */
	
    do {
#if defined(HAVE_POLL_H)
	rdfds.fd = fdno;
	rdfds.events = POLLIN;
	rdfds.revents = 0;
	rc = poll(&rdfds, 1, msecs);
#else
	if (tvp) {
	    tvp->tv_sec = secs;
	    tvp->tv_usec = 0;
	}
	FD_SET(fdno, &rdfds);
	rc = select(fdno + 1, &rdfds, NULL, NULL, tvp);
#endif

	if (rc < 0) {
	    switch (errno) {
	    case EINTR:
		continue;
		break;
	    default:
		return rc;
		break;
	    }
	}
	return rc;
    } while (1);
}

int fdFgets(FD_t fd, char * buf, size_t len)
{
    int fdno;
    int secs = fd->rd_timeoutsecs;
    size_t nb = 0;
    int ec = 0;
    char lastchar = '\0';

    if ((fdno = fdFileno(fd)) < 0)
	return 0;	/* XXX W2DO? */
	
    do {
	int rc;

	/* Is there data to read? */
	rc = fdReadable(fd, secs);

	switch (rc) {
	case -1:	/* error */
	    ec = -1;
	    continue;
	    break;
	case  0:	/* timeout */
	    ec = -1;
	    continue;
	    break;
	default:	/* data to read */
	    break;
	}

	errno = 0;
#ifdef	NOISY
	rc = fdRead(fd, buf + nb, 1);
#else
	rc = (int)read(fdFileno(fd), buf + nb, 1);
#endif
	if (rc < 0) {
	    fd->syserrno = errno;
	    switch (errno) {
	    case EWOULDBLOCK:
		continue;
		break;
	    default:
		break;
	    }
if (_rpmio_debug)
fprintf(stderr, "*** read: fd %p rc %d errno %d %s \"%s\"\n", fd, rc, errno, strerror(errno), buf);
	    ec = -1;
	    break;
	} else if (rc == 0) {
if (_rpmio_debug)
fprintf(stderr, "*** read: fd %p rc %d EOF errno %d %s \"%s\"\n", fd, rc, errno, strerror(errno), buf);
	    break;
	} else {
	    nb += rc;
	    buf[nb] = '\0';
	    lastchar = buf[nb - 1];
	}
    } while (ec == 0 && nb < len && lastchar != '\n');

    return (ec >= 0 ? (int)nb : ec);
}

/* =============================================================== */
/* Support for FTP/HTTP I/O.
 */
const char * ftpStrerror(int errorNumber)
{
    switch (errorNumber) {
    /* HACK error impediance match, coalesce and rename. */
    case FTPERR_NE_OK:
	return ("OK");
    case FTPERR_NE_ERROR:
	return ("NE_ERROR: Generic error.");
    case FTPERR_NE_LOOKUP:
	return ("NE_LOOKUP: Hostname lookup failed.");
    case FTPERR_NE_AUTH:
	return ("NE_AUTH: Server authentication failed.");
    case FTPERR_NE_PROXYAUTH:
	return ("NE_PROXYAUTH: Proxy authentication failed.");
    case FTPERR_NE_CONNECT:
	return ("NE_CONNECT: Could not connect to server.");
    case FTPERR_NE_TIMEOUT:
	return ("NE_TIMEOUT: Connection timed out.");
    case FTPERR_NE_FAILED:
	return ("NE_FAILED: The precondition failed.");
    case FTPERR_NE_RETRY:
	return ("NE_RETRY: Retry request.");
    case FTPERR_NE_REDIRECT:
	return ("NE_REDIRECT: Redirect received.");

    case FTPERR_BAD_SERVER_RESPONSE:
	return _("Bad server response");
    case FTPERR_SERVER_IO_ERROR:
	return _("Server I/O error");
    case FTPERR_SERVER_TIMEOUT:
	return _("Server timeout");
    case FTPERR_BAD_HOST_ADDR:
	return _("Unable to lookup server host address");
    case FTPERR_BAD_HOSTNAME:
	return _("Unable to lookup server host name");
    case FTPERR_FAILED_CONNECT:
	return _("Failed to connect to server");
    case FTPERR_FAILED_DATA_CONNECT:
	return _("Failed to establish data connection to server");
    case FTPERR_FILE_IO_ERROR:
	return _("I/O error to local file");
    case FTPERR_PASSIVE_ERROR:
	return _("Error setting remote server to passive mode");
    case FTPERR_FILE_NOT_FOUND:
	return _("File not found on server");
    case FTPERR_NIC_ABORT_IN_PROGRESS:
	return _("Abort in progress");

    case FTPERR_UNKNOWN:
    default:
	return _("Unknown or unexpected error");
    }
}

const char *urlStrerror(const char *url)
{
    const char *retstr;
    switch (urlIsURL(url)) {
    case URL_IS_HKP:
    case URL_IS_FTP:
    case URL_IS_HTTP:
    case URL_IS_HTTPS:
    {	urlinfo u;
/* XXX This only works for httpReq/ftpLogin/ftpReq failures */
	if (urlSplit(url, &u) == 0)
	    retstr = ftpStrerror(u->openError);
	else
	    retstr = _("Malformed URL");
    }	break;
    case URL_IS_MONGO:	/* XXX FIXME */
    default:
	retstr = strerror(errno);
	break;
    }
    return retstr;
}

#if !defined(HAVE_GETADDRINFO)
#if !defined(USE_ALT_DNS) || !USE_ALT_DNS
static int mygethostbyname(const char * host, struct in_addr * address)
{
    struct hostent * hostinfo;

    hostinfo = gethostbyname(host);
    if (!hostinfo) return 1;

    memcpy(address, hostinfo->h_addr_list[0], sizeof(*address));
    return 0;
}
#endif

static int getHostAddress(const char * host, struct in_addr * address)
{
#if 0	/* XXX workaround nss_foo module hand-off using valgrind. */
    if (!strcmp(host, "localhost")) {
	if (!inet_aton("127.0.0.1", address))
	    return FTPERR_BAD_HOST_ADDR;
    } else
#endif
    if (xisdigit(host[0])) {
	if (!inet_aton(host, address))
	    return FTPERR_BAD_HOST_ADDR;
    } else {
	if (mygethostbyname(host, address)) {
	    errno = h_errno;
	    return FTPERR_BAD_HOSTNAME;
	}
    }

    return 0;
}
#endif	/* HAVE_GETADDRINFO */

static int tcpConnect(FD_t ctrl, const char * host, int port)
{
    int fdno = -1;
    int rc;
#ifdef	HAVE_GETADDRINFO
    struct addrinfo hints, *res, *res0;
#ifndef	NI_MAXSERV
#define	NI_MAXSERV	32
#endif
    char pbuf[NI_MAXSERV];
    int xx;

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    sprintf(pbuf, "%d", port);
    pbuf[sizeof(pbuf)-1] = '\0';
    rc = FTPERR_FAILED_CONNECT;
    if (getaddrinfo(host, pbuf, &hints, &res0) == 0) {
	for (res = res0; res != NULL; res = res->ai_next) {
	    if ((fdno = socket(res->ai_family, res->ai_socktype, res->ai_protocol)) < 0)
		continue;
	    if (connect(fdno, res->ai_addr, (int)res->ai_addrlen) < 0) {
		xx = close(fdno);
		fdno = -1;	/* XXX coverity #1036724 */
		continue;
	    }
	    /* success */
	    rc = 0;
	    if (_ftp_debug) {
		char hbuf[NI_MAXHOST];
		hbuf[0] = '\0';
		xx = getnameinfo(res->ai_addr, res->ai_addrlen, hbuf, sizeof(hbuf),
				NULL, 0, NI_NUMERICHOST);
		fprintf(stderr,"++ connect [%s]:%d on fdno %d\n",
				hbuf, port, fdno);
	    }
	    break;
	}
	freeaddrinfo(res0);
    }
    if (rc < 0)
	goto errxit;
#else	/* HAVE_GETADDRINFO */
    struct sockaddr_in sin;

    memset(&sin, 0, sizeof(sin));
    sin.sin_family = AF_INET;
    sin.sin_port = htons(port);
    sin.sin_addr.s_addr = INADDR_ANY;

  do {
    if ((rc = getHostAddress(host, &sin.sin_addr)) < 0)
	break;

    if ((fdno = socket(sin.sin_family, SOCK_STREAM, IPPROTO_IP)) < 0) {
	rc = FTPERR_FAILED_CONNECT;
	break;
    }

    if (connect(fdno, (struct sockaddr *) &sin, sizeof(sin))) {
	rc = FTPERR_FAILED_CONNECT;
	break;
    }
  } while (0);

    if (rc < 0)
	goto errxit;

if (_ftp_debug)
fprintf(stderr,"++ connect %s:%d on fdno %d\n", inet_ntoa(sin.sin_addr) (int)ntohs(sin.sin_port), fdno);
#endif	/* HAVE_GETADDRINFO */

    (void) fdSetFdno(ctrl, (fdno >= 0 ? fdno : -1));
    return 0;

errxit:
    fdSetSyserrno(ctrl, errno, ftpStrerror(rc));
    if (fdno >= 0)
	(void) close(fdno);
    return rc;
}

static int checkResponse(void * _u, FD_t ctrl, int *ecp, char ** str)
{
    urlinfo u = (urlinfo) _u;
    char *buf;
    size_t bufAlloced;
    int bufLength = 0;
    const char *s;
    char *se;
    int ec = 0;
    int moretodo = 1;
    char errorCode[4];

    URLSANE(u);
    if (u->bufAlloced == 0 || u->buf == NULL) {
	u->bufAlloced = _url_iobuf_size;
	u->buf = (char *) xcalloc(u->bufAlloced, sizeof(u->buf[0]));
    }
    buf = u->buf;
    bufAlloced = u->bufAlloced;
    *buf = '\0';

    errorCode[0] = '\0';

    do {
	int rc;

	/*
	 * Read next line from server.
	 */
	se = buf + bufLength;
	*se = '\0';
	rc = fdFgets(ctrl, se, (bufAlloced - bufLength));
	if (rc < 0) {
	    ec = FTPERR_BAD_SERVER_RESPONSE;
	    continue;
	} else if (rc == 0 || fdWritable(ctrl, 0) < 1)
	    moretodo = 0;

	/*
	 * Process next line from server.
	 */
	for (s = se; *s != '\0'; s = se) {
		const char *e;

		while (*se && *se != '\n') se++;

		if (se > s && se[-1] == '\r')
		   se[-1] = '\0';
		if (*se == '\0')
		    break;

if (_ftp_debug)
fprintf(stderr, "<- %s\n", s);

		/* HTTP: header termination on empty line */
		if (*s == '\0') {
		    moretodo = 0;
		    break;
		}
		*se++ = '\0';

		/* HTTP: look for "HTTP/1.1 123 ..." */
		if (!strncmp(s, "HTTP", sizeof("HTTP")-1)) {
		    ctrl->contentLength = -1;
		    if ((e = strchr(s, '.')) != NULL) {
			e++;
			u->httpVersion = (int)(*e - '0');
			if (u->httpVersion < 1 || u->httpVersion > 2)
			    ctrl->persist = u->httpVersion = 0;
			else
			    ctrl->persist = 1;
		    }
		    if ((e = strchr(s, ' ')) != NULL) {
			e++;
			if (strchr("0123456789", *e))
			    strncpy(errorCode, e, 3);
			errorCode[3] = '\0';
		    }
		    continue;
		}

		/* HTTP: look for "token: ..." */
		for (e = s; *e && !(*e == ' ' || *e == ':'); e++)
		    {};
		if (e > s && *e++ == ':') {
		    size_t ne = (e - s);
		    while (*e && *e == ' ') e++;
#if 0
		    if (!strncmp(s, "Date:", ne)) {
		    } else
		    if (!strncmp(s, "Server:", ne)) {
		    } else
		    if (!strncmp(s, "Last-Modified:", ne)) {
		    } else
		    if (!strncmp(s, "ETag:", ne)) {
		    } else
#endif
		    if (!strncmp(s, "Accept-Ranges:", ne)) {
			if (!strcmp(e, "bytes"))
			    u->caps |= RPMURL_SERVER_HASRANGE;
			if (!strcmp(e, "none"))
			    u->caps &= ~RPMURL_SERVER_HASRANGE;
		    } else
		    if (!strncmp(s, "Content-Length:", ne)) {
			if (strchr("0123456789", *e))
			    ctrl->contentLength = atol(e);
		    } else
		    if (!strncmp(s, "Connection:", ne)) {
			if (!strcmp(e, "close"))
			    ctrl->persist = 0;
		    }
#if 0
		    else
		    if (!strncmp(s, "Content-Type:", ne)) {
		    } else
		    if (!strncmp(s, "Transfer-Encoding:", ne)) {
			if (!strcmp(e, "chunked"))
			    ctrl->wr_chunked = 1;
			else
			    ctrl->wr_chunked = 0;
		    } else
		    if (!strncmp(s, "Allow:", ne)) {
		    }
#endif
		    continue;
		}

		/* HTTP: look for "<TITLE>501 ... </TITLE>" */
		if (!strncmp(s, "<TITLE>", sizeof("<TITLE>")-1))
		    s += sizeof("<TITLE>") - 1;

		/* FTP: look for "123-" and/or "123 " */
		if (strchr("0123456789", *s)) {
		    if (errorCode[0] != '\0') {
			if (!strncmp(s, errorCode, sizeof("123")-1) && s[3] == ' ')
			    moretodo = 0;
		    } else {
			strncpy(errorCode, s, sizeof("123")-1);
			errorCode[3] = '\0';
			if (s[3] != '-')
			    moretodo = 0;
		    }
		}
	}

	if (moretodo && se > s) {
	    bufLength = se - s - 1;
	    if (s != buf)
		memmove(buf, s, bufLength);
	} else {
	    bufLength = 0;
	}
    } while (moretodo && ec == 0);

    if (str)	*str = buf;
    if (ecp)	*ecp = atoi(errorCode);

    return ec;
}

static int ftpCheckResponse(urlinfo u, char ** str)
{
    int ec = 0;
    int rc;

    URLSANE(u);
    rc = checkResponse(u, u->ctrl, &ec, str);

    switch (ec) {
    case 550:
	return FTPERR_FILE_NOT_FOUND;
	break;
    case 552:
	return FTPERR_NIC_ABORT_IN_PROGRESS;
	break;
    default:
	if (ec >= 400 && ec <= 599) {
	    return FTPERR_BAD_SERVER_RESPONSE;
	}
	break;
    }
    return rc;
}

static int ftpCommand(urlinfo u, char ** str, ...)
{
    va_list ap;
    int len = 0;
    const char * s, * t;
    char * te;
    int rc;

    URLSANE(u);
    va_start(ap, str);
    while ((s = va_arg(ap, const char *)) != NULL) {
	if (len) len++;
	len += strlen(s);
    }
    len += sizeof("\r\n")-1;
    va_end(ap);

    t = te = (char *) alloca(len + 1);

    va_start(ap, str);
    while ((s = va_arg(ap, const char *)) != NULL) {
	if (te > t) *te++ = ' ';
	te = stpcpy(te, s);
    }
    te = stpcpy(te, "\r\n");
    va_end(ap);

if (_ftp_debug)
fprintf(stderr, "-> %s", t);
    if (fdWrite(u->ctrl, t, (te-t)) != (te-t))
	return FTPERR_SERVER_IO_ERROR;

    rc = ftpCheckResponse(u, str);
    return rc;
}

static int ftpLogin(urlinfo u)
{
    const char * host;
    const char * user;
    const char * password;
    int port;
    int rc;

    URLSANE(u);
    u->ctrl = fdLink(u->ctrl, "open ctrl");

    if (((host = (u->proxyh ? u->proxyh : u->host)) == NULL)) {
	rc = FTPERR_BAD_HOSTNAME;
	goto errxit;
    }

    if ((port = (u->proxyp > 0 ? u->proxyp : u->port)) < 0) port = IPPORT_FTP;

    if ((user = (u->proxyu ? u->proxyu : u->user)) == NULL)
	user = "anonymous";

    if ((password = u->password) == NULL) {
 	uid_t uid = getuid();
	struct passwd * pw;
	if (uid && (pw = getpwuid(uid)) != NULL) {
	    char *myp = (char *) alloca(strlen(pw->pw_name) + sizeof("@"));
	    strcpy(myp, pw->pw_name);
	    strcat(myp, "@");
	    password = myp;
	} else {
	    password = "root@";
	}
    }

    if (fdFileno(u->ctrl) >= 0 && fdWritable(u->ctrl, 0) < 1)
	(void) fdClose(u->ctrl);

    if (fdFileno(u->ctrl) < 0) {
	rc = tcpConnect(u->ctrl, host, port);
	if (rc < 0)
	    goto errxit2;
    }

    if ((rc = ftpCheckResponse(u, NULL)))
	goto errxit;

    if ((rc = ftpCommand(u, NULL, "USER", user, NULL)))
	goto errxit;

    if ((rc = ftpCommand(u, NULL, "PASS", password, NULL)))
	goto errxit;

    if ((rc = ftpCommand(u, NULL, "TYPE", "I", NULL)))
	goto errxit;

    return 0;

errxit:
    fdSetSyserrno(u->ctrl, errno, ftpStrerror(rc));
errxit2:
    if (fdFileno(u->ctrl) >= 0)
	(void) fdClose(u->ctrl);
    return rc;
}

int ftpReq(FD_t data, const char * ftpCmd, const char * ftpArg)
{
    urlinfo u = (urlinfo) data->u;
#if !defined(HAVE_GETADDRINFO)
    struct sockaddr_in dataAddress;
#endif	/* HAVE_GETADDRINFO */
    char remoteIP[NI_MAXHOST];
    char * cmd;
    size_t cmdlen;
    char * passReply;
    char * chptr;
    int rc;
    int epsv;
    int port;

    remoteIP[0] = '\0';
    URLSANE(u);
    if (ftpCmd == NULL)
	return FTPERR_UNKNOWN;	/* XXX W2DO? */

    cmdlen = strlen(ftpCmd) + (ftpArg ? 1+strlen(ftpArg) : 0) + sizeof("\r\n");
    cmd = (char *) alloca(cmdlen);	/* XXX coverity #1035968 */
    chptr = stpcpy(cmd, ftpCmd);
    if (ftpArg) {
	*chptr++ = ' ';
	chptr = stpcpy(chptr, ftpArg);
    }
    chptr = stpcpy(chptr, "\r\n");
    cmdlen = chptr - cmd;

/*
 * Get the ftp version of the Content-Length.
 */
    if (!strncmp(cmd, "RETR", 4)) {
	unsigned cl;

	passReply = NULL;
	rc = ftpCommand(u, &passReply, "SIZE", ftpArg, NULL);
	if (rc)
	    goto errxit;
	if (sscanf(passReply, "%d %u", &rc, &cl) != 2) {
	    rc = FTPERR_BAD_SERVER_RESPONSE;
	    goto errxit;
	}
	rc = 0;
	data->contentLength = cl;
    }

    epsv = 0;
    passReply = NULL;
#ifdef HAVE_GETNAMEINFO
    rc = ftpCommand(u, &passReply, "EPSV", NULL);
    if (rc == 0) {
#ifdef HAVE_GETADDRINFO
	struct sockaddr_storage ss;
#else /* HAVE_GETADDRINFO */
	struct sockaddr_in ss;
#endif /* HAVE_GETADDRINFO */
	socklen_t sslen = sizeof(ss);

	/* we need to know IP of remote host */
	if ((getpeername(fdFileno(c2f(u->ctrl)), (struct sockaddr *)&ss, &sslen) == 0)
	 && (getnameinfo((struct sockaddr *)&ss, sslen,
			remoteIP, sizeof(remoteIP),
			NULL, 0, NI_NUMERICHOST) == 0))
	{
		epsv++;
	} else {
		/* abort EPSV and fall back to PASV */
		rc = ftpCommand(u, &passReply, "ABOR", NULL);
		if (rc) {
		    rc = FTPERR_PASSIVE_ERROR;
		    goto errxit;
		}
	}
    }
   if (epsv == 0)
#endif /* HAVE_GETNAMEINFO */
	rc = ftpCommand(u, &passReply, "PASV", NULL);
    if (rc) {
	rc = FTPERR_PASSIVE_ERROR;
	goto errxit;
    }

    chptr = passReply;
assert(chptr != NULL);
    while (*chptr && *chptr != '(') chptr++;
    if (*chptr != '(') return FTPERR_PASSIVE_ERROR;
    chptr++;
    passReply = chptr;
    while (*chptr && *chptr != ')') chptr++;
    if (*chptr != ')') return FTPERR_PASSIVE_ERROR;
    *chptr-- = '\0';

  if (epsv) {
	int i;
	if(sscanf(passReply,"%*c%*c%*c%d%*c",&i) != 1) {
	   rc = FTPERR_PASSIVE_ERROR;
	   goto errxit;
	}
	port = i;
  } else {

    while (*chptr && *chptr != ',') chptr--;
    if (*chptr != ',') return FTPERR_PASSIVE_ERROR;
    chptr--;
    while (*chptr && *chptr != ',') chptr--;
    if (*chptr != ',') return FTPERR_PASSIVE_ERROR;
    *chptr++ = '\0';

    /* now passReply points to the IP portion, and chptr points to the
       port number portion */

    {	int i, j;
	if (sscanf(chptr, "%d,%d", &i, &j) != 2) {
	    rc = FTPERR_PASSIVE_ERROR;
	    goto errxit;
	}
	port = (((unsigned)i) << 8) + j;
    }

    chptr = passReply;
    while (*chptr++ != '\0') {
	if (*chptr == ',') *chptr = '.';
    }
    sprintf(remoteIP, "%s", passReply);
  } /* if (epsv) */

#ifdef HAVE_GETADDRINFO
    {
	struct addrinfo hints, *res, *res0;
	char pbuf[NI_MAXSERV];
	int xx;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_NUMERICHOST;
#if defined(AI_IDN)
	hints.ai_flags |= AI_IDN;
#endif
	sprintf(pbuf, "%d", port);
	pbuf[sizeof(pbuf)-1] = '\0';
	if (getaddrinfo(remoteIP, pbuf, &hints, &res0)) {
	    rc = FTPERR_PASSIVE_ERROR;
	    goto errxit;
	}

	for (res = res0; res != NULL; res = res->ai_next) {
	    rc = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
	    (void) fdSetFdno(data, (rc >= 0 ? rc : -1));
	    if (rc < 0) {
	        if (res->ai_next)
		    continue;
		else {
		    rc = FTPERR_FAILED_CONNECT;
		    freeaddrinfo(res0);
		    goto errxit;
		}
	    }
	    data = fdLink(data, "open data (ftpReq)");

	    /* XXX setsockopt SO_LINGER */
	    /* XXX setsockopt SO_KEEPALIVE */
	    /* XXX setsockopt SO_TOS IPTOS_THROUGHPUT */

	    {
		int criterr = 0;
	        while (connect(fdFileno(data), res->ai_addr, (int)res->ai_addrlen) < 0) {
	            if (errno == EINTR)
		        continue;
		    criterr++;
		}
		if (criterr) {
		    if (res->ai_addr) {
		        xx = fdClose(data);
		        continue;
		    } else {
		        rc = FTPERR_PASSIVE_ERROR;
		        freeaddrinfo(res0);
		        goto errxit;
		    }
		}
	    }
	    /* success */
	    rc = 0;
	    break;
	}
	freeaddrinfo(res0);
    }
#else /* HAVE_GETADDRINFO */
    memset(&dataAddress, 0, sizeof(dataAddress));
    dataAddress.sin_family = AF_INET;
    dataAddress.sin_port = htons(port);

    if (!inet_aton(remoteIP, &dataAddress.sin_addr)) {
	rc = FTPERR_PASSIVE_ERROR;
	goto errxit;
    }

    rc = socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
    (void) fdSetFdno(data, (rc >= 0 ? rc : -1));
    if (rc < 0) {
	rc = FTPERR_FAILED_CONNECT;
	goto errxit;
    }
    data = fdLink(data, "open data (ftpReq)");

    /* XXX setsockopt SO_LINGER */
    /* XXX setsockopt SO_KEEPALIVE */
    /* XXX setsockopt SO_TOS IPTOS_THROUGHPUT */

    while (connect(fdFileno(data), (struct sockaddr *) &dataAddress,
	        sizeof(dataAddress)) < 0)
    {
	if (errno == EINTR)
	    continue;
	rc = FTPERR_FAILED_DATA_CONNECT;
	goto errxit;
    }
#endif /* HAVE_GETADDRINFO */

if (_ftp_debug)
fprintf(stderr, "-> %s", cmd);
    if ((size_t)fdWrite(u->ctrl, cmd, cmdlen) != cmdlen) {
	rc = FTPERR_SERVER_IO_ERROR;
	goto errxit;
    }

    if ((rc = ftpCheckResponse(u, NULL))) {
	goto errxit;
    }

    data->ftpFileDoneNeeded = 1;
    u->ctrl = fdLink(u->ctrl, "grab data (ftpReq)");
    u->ctrl = fdLink(u->ctrl, "open data (ftpReq)");
    return 0;

errxit:
    fdSetSyserrno(u->ctrl, errno, ftpStrerror(rc));
    if (fdFileno(data) >= 0)
	(void) fdClose(data);
    return rc;
}

#ifdef	DYING
static rpmCallbackFunction	_urlNotify = NULL;

static void *			_urlNotifyData = NULL;

static int			_urlNotifyCount = -1;

static void urlSetCallback(rpmCallbackFunction notify, void *notifyData, int notifyCount) {
    _urlNotify = notify;
    _urlNotifyData = notifyData;
    _urlNotifyCount = (notifyCount >= 0) ? notifyCount : 4096;
}
#endif

int ufdCopy(FD_t sfd, FD_t tfd)
{
    char buf[BUFSIZ];
    int itemsRead;
    int itemsCopied = 0;
    int rc = 0;
#ifdef	DYING
    int notifier = -1;

    if (_urlNotify) {
	(void) (*_urlNotify) (NULL, RPMCALLBACK_INST_OPEN_FILE,
		0, 0, NULL, _urlNotifyData);
    }
#endif

    while (1) {
	rc = (int) Fread(buf, sizeof(buf[0]), sizeof(buf), sfd);
	if (rc < 0)	/* XXX never happens Fread returns size_t */
	    break;
	else if (rc == 0) {
	    rc = itemsCopied;
	    break;
	}
	itemsRead = rc;
	rc = (int) Fwrite(buf, sizeof(buf[0]), itemsRead, tfd);
	if (rc < 0)	/* XXX never happens Fwrite returns size_t */
	    break;
 	if (rc != itemsRead) {
	    rc = FTPERR_FILE_IO_ERROR;
	    break;
	}

	itemsCopied += itemsRead;
#ifdef	DYING
	if (_urlNotify && _urlNotifyCount > 0) {
	    int n = itemsCopied/_urlNotifyCount;
	    if (n != notifier) {
		(void) (*_urlNotify) (NULL, RPMCALLBACK_INST_PROGRESS,
			itemsCopied, 0, NULL, _urlNotifyData);
		notifier = n;
	    }
	}
#endif
    }

    DBGIO(sfd, (stderr, "++ copied %d bytes: %s\n", itemsCopied,
	ftpStrerror(rc)));

#ifdef	DYING
    if (_urlNotify) {
	(void) (*_urlNotify) (NULL, RPMCALLBACK_INST_OPEN_FILE,
		itemsCopied, itemsCopied, NULL, _urlNotifyData);
    }
#endif

    return rc;
}

static int urlConnect(const char * url, urlinfo * uret)
{
    urlinfo u;
    int rc = 0;

    if (urlSplit(url, &u) < 0)
	return -1;

    if (urlType(u) == URL_IS_FTP) {
	FD_t fd;

	if ((fd = u->ctrl) == NULL) {
	    fd = u->ctrl = fdNew("persist ctrl (urlConnect FTP)");
	    fdSetOpen(u->ctrl, url, 0, 0);
	    fdSetIo(u->ctrl, ufdio);
	}
	
assert(fd != NULL);
	fd->rd_timeoutsecs = ftpTimeoutSecs;
	fd->contentLength = fd->bytesRemain = -1;
	fd->u = NULL;		/* XXX FTP ctrl has not */
	fd->ftpFileDoneNeeded = 0;
	fd = fdLink(fd, "grab ctrl (urlConnect FTP)");

	if (fdFileno(u->ctrl) < 0) {
	    rpmlog(RPMLOG_DEBUG, D_("logging into %s as %s, pw %s\n"),
			u->host ? u->host : "???",
			u->user ? u->user : "ftp",
			u->password ? u->password : "(username)");

	    if ((rc = ftpLogin(u)) < 0) {	/* XXX save ftpLogin error */
		u->ctrl = fdFree(fd, "grab ctrl (urlConnect FTP)");
		u->openError = rc;
	    }
	}
    }

    if (uret != NULL)
	*uret = urlLink(u, "urlConnect");
    u = urlFree(u, "urlSplit (urlConnect)");	

    return rc;
}

int ufdGetFile(FD_t sfd, FD_t tfd)
{
    int rc;

    FDSANE(sfd);
    FDSANE(tfd);
    rc = ufdCopy(sfd, tfd);
    (void) Fclose(sfd);
    if (rc > 0)		/* XXX ufdCopy now returns no. bytes copied */
	rc = 0;
    return rc;
}

int ftpCmd(const char * cmd, const char * url, const char * arg2)
{
    urlinfo u;
    int rc;
    const char * path;

    if (urlConnect(url, &u) < 0)
	return -1;

    (void) urlPath(url, &path);

    rc = ftpCommand(u, NULL, cmd, path, arg2, NULL);
    u->ctrl = fdFree(u->ctrl, "grab ctrl (ftpCmd)");
    return rc;
}

/* XXX these aren't worth the pain of including correctly */
#if !defined(IAC)
#define	IAC	((unsigned char)255)	/* interpret as command: */
#endif
#if !defined(IP)
#define	IP	((unsigned char)244)	/* interrupt process--permanently */
#endif
#if !defined(DM)
#define	DM	((unsigned char)242)	/* data mark--for connect. cleaning */
#endif
#if !defined(SHUT_RDWR)
#define	SHUT_RDWR	1+1
#endif

static int ftpAbort(urlinfo u, FD_t data)
{
    static unsigned char ipbuf[3] = { IAC, IP, IAC };
    FD_t ctrl;
    int rc;
    int tosecs;

    URLSANE(u);

    if (data != NULL) {
	data->ftpFileDoneNeeded = 0;
	if (fdFileno(data) >= 0)
	    u->ctrl = fdFree(u->ctrl, "open data (ftpAbort)");
	u->ctrl = fdFree(u->ctrl, "grab data (ftpAbort)");
    }
    ctrl = u->ctrl;

    DBGIO(0, (stderr, "-> ABOR\n"));

    if (send(fdFileno(ctrl), ipbuf, sizeof(ipbuf), MSG_OOB) != sizeof(ipbuf)) {
	(void) fdClose(ctrl);
	return FTPERR_SERVER_IO_ERROR;
    }

    sprintf(u->buf, "%cABOR\r\n",(char) DM);
    if (fdWrite(ctrl, u->buf, 7) != 7) {
	(void) fdClose(ctrl);
	return FTPERR_SERVER_IO_ERROR;
    }

    if (data && fdFileno(data) >= 0) {
	/* XXX shorten data drain time wait */
	tosecs = data->rd_timeoutsecs;
	data->rd_timeoutsecs = 10;
	if (fdReadable(data, data->rd_timeoutsecs) > 0) {
	    while ((ufdio->read)(data, u->buf, u->bufAlloced) > 0)
		u->buf[0] = '\0';
	}
	data->rd_timeoutsecs = tosecs;
	/* XXX ftp abort needs to close the data channel to receive status */
	(void) shutdown(fdFileno(data), SHUT_RDWR);
	(void) close(fdFileno(data));
	data->fps[0].fdno = -1;	/* XXX WRONG but expedient */
    }

    /* XXX shorten ctrl drain time wait */
assert(u->ctrl != NULL);
    tosecs = u->ctrl->rd_timeoutsecs;
    u->ctrl->rd_timeoutsecs = 10;
    if ((rc = ftpCheckResponse(u, NULL)) == FTPERR_NIC_ABORT_IN_PROGRESS) {
	rc = ftpCheckResponse(u, NULL);
    }
    rc = ftpCheckResponse(u, NULL);
    u->ctrl->rd_timeoutsecs = tosecs;

    return rc;
}

static int ftpFileDone(urlinfo u, FD_t data)
{
    int rc = 0;

    URLSANE(u);
    assert(data->ftpFileDoneNeeded);

    if (data->ftpFileDoneNeeded) {
	data->ftpFileDoneNeeded = 0;
	u->ctrl = fdFree(u->ctrl, "open data (ftpFileDone)");
	u->ctrl = fdFree(u->ctrl, "grab data (ftpFileDone)");
	rc = ftpCheckResponse(u, NULL);
    }
    return rc;
}

#ifndef WITH_NEON
static int httpResp(urlinfo u, FD_t ctrl, char ** str)
{
    int ec = 0;
    int rc;

    URLSANE(u);
    rc = checkResponse(u, ctrl, &ec, str);

if (_ftp_debug && !(rc == 0 && (ec == 200 || ec == 201)))
fprintf(stderr, "*** httpResp: rc %d ec %d\n", rc, ec);

    switch (ec) {
    case 200:
    case 201:			/* 201 Created. */
	break;
    case 204:			/* HACK: if overwriting, 204 No Content. */
    case 403:			/* 403 Forbidden. */
	ctrl->syserrno = EACCES;	/* HACK */
	rc = FTPERR_UNKNOWN;
	break;
    default:
	rc = FTPERR_FILE_NOT_FOUND;
	break;
    }
    return rc;
}

static int httpReq(FD_t ctrl, const char * httpCmd, const char * httpArg)
{
    urlinfo u;
    const char * host;
    const char * path;
    char hthost[NI_MAXHOST];
    int port;
    int rc;
    char * req;
    size_t len;
    int retrying = 0;

assert(ctrl != NULL);
    u = ctrl->u;
    URLSANE(u);

    if (((host = (u->proxyh ? u->proxyh : u->host)) == NULL))
	return FTPERR_BAD_HOSTNAME;
    if (strchr(host, ':'))
	sprintf(hthost, "[%s]", host);
    else
	strcpy(hthost, host);

    if ((port = (u->proxyp > 0 ? u->proxyp : u->port)) < 0) port = 80;
    path = (u->proxyh || u->proxyp > 0) ? u->url : httpArg;
    if (path == NULL) path = "";

reopen:
    if (fdFileno(ctrl) >= 0 && (rc = fdWritable(ctrl, 0)) < 1) {
	(void) fdClose(ctrl);
    }

    if (fdFileno(ctrl) < 0) {
	rc = tcpConnect(ctrl, host, port);
	if (rc < 0)
	    goto errxit2;
	ctrl = fdLink(ctrl, "open ctrl (httpReq)");
    }

    len = sizeof("\
req x HTTP/1.0\r\n\
User-Agent: rpm/3.0.4\r\n\
Host: y:z\r\n\
Accept: text/plain\r\n\
Transfer-Encoding: chunked\r\n\
\r\n\
") + strlen(httpCmd) + strlen(path) + sizeof(VERSION) + strlen(hthost) + 20;

    req = alloca(len);
    *req = '\0';

  if (!strcmp(httpCmd, "PUT")) {
    sprintf(req, "\
%s %s HTTP/1.%d\r\n\
User-Agent: rpm/%s\r\n\
Host: %s:%d\r\n\
Accept: text/plain\r\n\
Transfer-Encoding: chunked\r\n\
\r\n\
",	httpCmd, path, (u->httpVersion ? 1 : 0), VERSION, hthost, port);
} else {
    sprintf(req, "\
%s %s HTTP/1.%d\r\n\
User-Agent: rpm/%s\r\n\
Host: %s:%d\r\n\
Accept: text/plain\r\n\
\r\n\
",	httpCmd, path, (u->httpVersion ? 1 : 0), VERSION, hthost, port);
}

if (_ftp_debug)
fprintf(stderr, "-> %s", req);

    len = strlen(req);
    if (fdWrite(ctrl, req, len) != (ssize_t)len) {
	rc = FTPERR_SERVER_IO_ERROR;
	goto errxit;
    }

    if (!strcmp(httpCmd, "PUT")) {
	ctrl->wr_chunked = 1;
    } else {

	rc = httpResp(u, ctrl, NULL);

	if (rc) {
	    if (!retrying) {	/* not HTTP_OK */
		retrying = 1;
		(void) fdClose(ctrl);
		goto reopen;
	    }
	    goto errxit;
	}
    }

    ctrl = fdLink(ctrl, "open data (httpReq)");
    return 0;

errxit:
    fdSetSyserrno(ctrl, errno, ftpStrerror(rc));
errxit2:
    if (fdFileno(ctrl) >= 0)
	(void) fdClose(ctrl);
    return rc;
}
#endif /* WITH_NEON */

/* XXX DYING: unused */
void * ufdGetUrlinfo(FD_t fd)
{
    FDSANE(fd);
    if (fd->u == NULL)
	return NULL;
    return urlLink(fd->u, "ufdGetUrlinfo");
}

/* =============================================================== */
static ssize_t ufdRead(void * cookie, char * buf, size_t count)
{
    FD_t fd = c2f(cookie);
    size_t bytesRead;
    size_t total;

    if (fdGetIo(fd) == fdio) {
	struct stat sb;
	int fdno = fdFileno(fd);
	(void) fstat(fdno, &sb);
	if (S_ISREG(sb.st_mode))
	    return fdRead(fd, buf, count);
    }

    UFDONLY(fd);
    assert(fd->rd_timeoutsecs >= 0);

    for (total = 0; total < count; total += bytesRead) {

	int rc;

	bytesRead = 0;

	/* Is there data to read? */
	if (fd->bytesRemain == 0) return (ssize_t) total; /* XXX simulate EOF */
	rc = fdReadable(fd, fd->rd_timeoutsecs);

	switch (rc) {
	case -1:	/* error */
	case  0:	/* timeout */
	    return (ssize_t) total;
	    break;
	default:	/* data to read */
	    break;
	}

	rc = (int) fdRead(fd, buf + total, count - total);

	if (rc < 0) {
	    switch (errno) {
	    case EWOULDBLOCK:
		continue;
		break;
	    default:
		break;
	    }
if (_rpmio_debug)
fprintf(stderr, "*** read: rc %d errno %d %s \"%s\"\n", rc, errno, strerror(errno), buf);
	    return rc;
	    break;
	} else if (rc == 0) {
	    return (ssize_t) total;
	    break;
	}
	bytesRead = (size_t) rc;
    }

    return (ssize_t) count;
}

static ssize_t ufdWrite(void * cookie, const char * buf, size_t count)
{
    FD_t fd = c2f(cookie);
    size_t bytesWritten;
    size_t total = 0;

#ifdef	NOTYET
    if (fdGetIo(fd) == fdio) {
	struct stat sb;
	(void) fstat(fdGetFdno(fd), &sb);
	if (S_ISREG(sb.st_mode))
	    return fdWrite(fd, buf, count);
    }
#endif

    UFDONLY(fd);

    for (total = 0; total < count; total += bytesWritten) {

	int rc;

	bytesWritten = 0;

	/* Is there room to write data? */
	if (fd->bytesRemain == 0) {
fprintf(stderr, "*** ufdWrite fd %p WRITE PAST END OF CONTENT\n", fd);
	    return (ssize_t) total;	/* XXX simulate EOF */
	}
	rc = fdWritable(fd, 2);		/* XXX configurable? */

	switch (rc) {
	case -1:	/* error */
	case  0:	/* timeout */
	    return (ssize_t) total;
	    break;
	default:	/* data to write */
	    break;
	}

	rc = (int) fdWrite(fd, buf + total, count - total);

	if (rc < 0) {
	    switch (errno) {
	    case EWOULDBLOCK:
		continue;
		break;
	    default:
		break;
	    }
if (_rpmio_debug)
fprintf(stderr, "*** write: rc %d errno %d %s \"%s\"\n", rc, errno, strerror(errno), buf);
	    return rc;
	    break;
	} else if (rc == 0) {
	    return (ssize_t) total;
	    break;
	}
	bytesWritten = (size_t) rc;
    }

    return (ssize_t) count;
}

static int ufdSeek(void * cookie, _libio_pos_t pos, int whence)
{
    FD_t fd = c2f(cookie);

    switch (urlType(fd->u)) {
    case URL_IS_UNKNOWN:
    case URL_IS_PATH:
	break;
    case URL_IS_DASH:
    case URL_IS_HKP:
    case URL_IS_FTP:
    case URL_IS_HTTP:
    case URL_IS_HTTPS:
    case URL_IS_MONGO:	/* XXX FIXME */
    default:
	return -2;
	break;
    }
    return fdSeek(cookie, pos, whence);
}

int ufdClose(void * cookie)
{
    FD_t fd = c2f(cookie);

    UFDONLY(fd);

    if (fd->u) {
	urlinfo u = (urlinfo) fd->u;

	if (fd == u->data)
	    fd = u->data = fdFree(fd, "grab data (ufdClose persist)");
	else
	    fd = fdFree(fd, "grab data (ufdClose)");
assert(fd != NULL);
	(void) urlFree(fd->u, "url (ufdClose)");
	fd->u = NULL;
	u->ctrl = fdFree(u->ctrl, "grab ctrl (ufdClose)");

	if (urlType(u) == URL_IS_FTP) {

	    /* XXX if not using libio, lose the fp from fpio */
	    {   FILE * fp;
		fp = fdGetFILE(fd);
		if (noLibio && fp)
		    fdSetFp(fd, NULL);
	    }

	    /*
	     * Non-error FTP has 4 refs on the data fd:
	     *	"persist data (ufdOpen FTP)"		rpmio.c:888
	     *	"grab data (ufdOpen FTP)"		rpmio.c:892
	     *	"open data (ftpReq)"			ftp.c:633
	     *	"fopencookie"				rpmio.c:1507
	     *
	     * Non-error FTP has 5 refs on the ctrl fd:
	     *	"persist ctrl"				url.c:176
	     *	"grab ctrl (urlConnect FTP)"		rpmio.c:404
	     *	"open ctrl"				ftp.c:504
	     *	"grab data (ftpReq)"			ftp.c:661
	     *	"open data (ftpReq)"			ftp.c:662
	     */
	    if (fd->bytesRemain > 0) {
		if (fd->ftpFileDoneNeeded) {
		    if (fdReadable(u->ctrl, 0) > 0)
			(void) ftpFileDone(u, fd);
		    else
			(void) ftpAbort(u, fd);
		}
	    } else {
		int rc;
		/* XXX STOR et al require close before ftpFileDone */
		rc = fdClose(fd);
#if 0	/* XXX error exit from ufdOpen does not have this set */
		assert(fd->ftpFileDoneNeeded != 0);
#endif
		if (fd->ftpFileDoneNeeded)
		    (void) ftpFileDone(u, fd);
		return rc;
	    }
	}

	/* XXX Why not (u->urltype == URL_IS_HTTP) ??? */
	/* XXX Why not (u->urltype == URL_IS_HTTPS) ??? */
	/* XXX Why not (u->urltype == URL_IS_HKP) ??? */
	if (u->scheme != NULL
	 && (!strncmp(u->scheme, "http", sizeof("http")-1) || !strncmp(u->scheme, "hkp", sizeof("hkp")-1)))
	{
	    /*
	     * HTTP has 4 (or 5 if persistent malloc) refs on the fd:
	     *	"persist ctrl"				url.c:177
	     *	"grab ctrl (ufdOpen HTTP)"		rpmio.c:924
	     *	"grab data (ufdOpen HTTP)"		rpmio.c:928
	     *	"open ctrl (httpReq)"			ftp.c:382
	     *	"open data (httpReq)"			ftp.c:435
	     */

	    if (fd == u->ctrl)
		fd = u->ctrl = fdFree(fd, "open data (ufdClose HTTP persist ctrl)");
	    else if (fd == u->data)
		fd = u->data = fdFree(fd, "open data (ufdClose HTTP persist data)");
	    else
		fd = fdFree(fd, "open data (ufdClose HTTP)");

	    /* XXX if not using libio, lose the fp from fpio */
	    {   FILE * fp;
		fp = fdGetFILE(fd);
		if (noLibio && fp)
		    fdSetFp(fd, NULL);
	    }

	    /* If content remains, then don't persist. */
assert(fd != NULL);
	    if (fd->bytesRemain > 0) {
		fd->persist = 0;
	    }
	    fd->contentLength = fd->bytesRemain = -1;

	    /* If persisting, then Fclose will juggle refcounts. */
	    if (fd->persist && (fd == u->ctrl || fd == u->data))
		return 0;
	}
    }
    return fdClose(fd);
}

FD_t ftpOpen(const char *url, int flags, mode_t mode, urlinfo *uret)
{
    urlinfo u = NULL;
    FD_t fd = NULL;

#if 0	/* XXX makeTempFile() heartburn */
    assert(!(flags & O_RDWR));
#endif
    if (urlConnect(url, &u) < 0)
	goto exit;

    if (u->data == NULL)
	u->data = fdNew("persist data (ftpOpen)");

assert(u->data != NULL);
    if (u->data->u == NULL)
	fd = u->data = fdLink(u->data, "grab data (ftpOpen persist data)");
    else
	fd = fdNew("grab data (ftpOpen)");

    if (fd != NULL) {
	fdSetOpen(fd, url, flags, mode);
	fdSetIo(fd, ufdio);
	fd->ftpFileDoneNeeded = 0;
	fd->rd_timeoutsecs = ftpTimeoutSecs;
	fd->contentLength = fd->bytesRemain = -1;
	fd->u = urlLink(u, "url (ufdOpen FTP)");
    }

exit:
    if (uret)
	*uret = u;
    return fd;
}

static FD_t ufdOpen(const char * url, int flags, mode_t mode)
{
    FD_t fd = NULL;
    const char * cmd;
    urlinfo u;
    const char * path;
    urltype ut = urlPath(url, &path);

if (_rpmio_debug)
fprintf(stderr, "*** ufdOpen(%s,0x%x,0%o)\n", url, (unsigned)flags, (unsigned)mode);

    switch (ut) {
    case URL_IS_FTP:
	fd = ftpOpen(url, flags, mode, &u);
	if (fd == NULL || u == NULL)
	    break;

	/* XXX W2DO? use STOU rather than STOR to prevent clobbering */
	cmd = ((flags & O_WRONLY)
		?  ((flags & O_APPEND) ? "APPE" :
		   ((flags & O_CREAT) ? "STOR" : "STOR"))
		:  ((flags & O_CREAT) ? "STOR" : "RETR"));
	u->openError = ftpReq(fd, cmd, path);
	if (u->openError < 0) {
	    /* XXX make sure that we can exit through ufdClose */
	    fd = fdLink(fd, "error data (ufdOpen FTP)");
	} else {
	    fd->bytesRemain = ((!strcmp(cmd, "RETR"))
		?  fd->contentLength : -1);
	    fd->wr_chunked = 0;
	}
	break;
    case URL_IS_HKP:
    case URL_IS_HTTP:
    case URL_IS_HTTPS:
#ifdef WITH_NEON
	fd = davOpen(url, flags, mode, &u);
#else
	fd = httpOpen(url, flags, mode, &u);
#endif
	if (fd == NULL || u == NULL)
	    break;

	cmd = ((flags & O_WRONLY)
		?  ((flags & O_APPEND) ? "PUT" :
		   ((flags & O_CREAT) ? "PUT" : "PUT"))
		: "GET");
#ifdef WITH_NEON
	u->openError = davReq(fd, cmd, path);
#else
	u->openError = httpReq(fd, cmd, path);
#endif
	if (u->openError < 0) {
	    /* XXX make sure that we can exit through ufdClose */
	    fd = fdLink(fd, "error ctrl (ufdOpen HTTP)");
	    fd = fdLink(fd, "error data (ufdOpen HTTP)");
	} else {
	    fd->bytesRemain = ((!strcmp(cmd, "GET"))
		?  fd->contentLength : -1);
	    fd->wr_chunked = ((!strcmp(cmd, "PUT"))
		?  fd->wr_chunked : 0);
	}
	break;
    case URL_IS_MONGO:	/* XXX FIXME */
	break;
    case URL_IS_DASH:
	assert(!(flags & O_RDWR));
	fd = fdDup( ((flags & O_WRONLY) ? STDOUT_FILENO : STDIN_FILENO) );
	if (fd) {
	    fdSetOpen(fd, url, flags, mode);
	    fdSetIo(fd, ufdio);
	    fd->rd_timeoutsecs = 600;	/* XXX W2DO? 10 mins? */
	    fd->contentLength = fd->bytesRemain = -1;
	}
	break;
    case URL_IS_PATH:
    case URL_IS_UNKNOWN:
    default:
	fd = fdOpen(path, flags, mode);
	if (fd) {
	    fdSetIo(fd, ufdio);
	    fd->rd_timeoutsecs = 1;
	    fd->contentLength = fd->bytesRemain = -1;
	}
	break;
    }

    if (fd == NULL) return NULL;
    if (Fileno(fd) < 0) {
	(void) ufdClose(fd);
	return NULL;
    }
DBGIO(fd, (stderr, "<--\tufdOpen(\"%s\",%x,0%o) %s\n", url, (unsigned)flags, (unsigned)mode, fdbg(fd)));
    return fd;
}

static struct FDIO_s ufdio_s = {
  ufdRead, ufdWrite, ufdSeek, ufdClose,	NULL, NULL, NULL,
};

FDIO_t ufdio = &ufdio_s;

/* =============================================================== */
static struct FDIO_s fpio_s = {
  ufdRead, ufdWrite, fdSeek, ufdClose, NULL, NULL, NULL,
};

FDIO_t fpio = &fpio_s;

/* =============================================================== */
static const char * getFdErrstr (FD_t fd)
{
    const char *errstr = NULL;

#if defined(WITH_ZLIB)
    if (fdGetIo(fd) == gzdio) {
	errstr = (const char *)fd->errcookie;
    } else
#endif	/* WITH_ZLIB */

#if defined(WITH_BZIP2)
    if (fdGetIo(fd) == bzdio) {
	errstr = (const char *)fd->errcookie;
    } else
#endif

#if defined(WITH_XZ)
    if (fdGetIo(fd) == lzdio) {
	errstr = (const char *)fd->errcookie;
    } else
    if (fdGetIo(fd) == xzdio) {
	errstr = (const char *)fd->errcookie;
    } else
#endif

#if defined(WITH_ZSTD)
    if (fdGetIo(fd) == zstdio) {
	errstr = (const char *)fd->errcookie;
    } else
#endif

    {
	errstr = (fd->syserrno ? strerror(fd->syserrno) : "");
    }

    return errstr;
}

/* =============================================================== */

const char *Fstrerror(FD_t fd)
{
    if (fd == NULL)
	return (errno ? strerror(errno) : "");
    FDSANE(fd);
    return getFdErrstr(fd);
}

#define	FDIOVEC(_fd, _vec)	\
  ((fdGetIo(_fd) && fdGetIo(_fd)->_vec) ? fdGetIo(_fd)->_vec : NULL)

size_t Fread(void *buf, size_t size, size_t nmemb, FD_t fd) {
    fdio_read_function_t _read;
    ssize_t rc;

    if (buf == NULL || size == 0 || nmemb == 0 || fd == NULL)
	return 0;

    FDSANE(fd);
DBGIO(fd, (stderr, "==> Fread(%p,%zu,%zu,%p)\n", buf, size, nmemb, (fd ? fd : NULL)));

    if (fdGetIo(fd) == fpio) {
	rc = fread(buf, size, nmemb, fdGetFILE(fd));
#ifdef	NOISY
DBGIO(fd, (stderr, "<== Fread(%p,%zu,%zu,%p) rc %zd\n", buf, size, nmemb, (fd ? fd : NULL), rc));
#endif
	return rc;
    }

    _read = FDIOVEC(fd, read);

    rc = (_read ? (*_read) (fd, (char *)buf, size * nmemb) : -2);
#ifdef	NOISY
DBGIO(fd, (stderr, "<== Fread(%p,%zu,%zu,%p) rc %zd\n", buf, size, nmemb, (fd ? fd : NULL), rc));
#endif
    return (rc >= 0 ? (rc/size) : 0);
}

size_t Fwrite(const void *buf, size_t size, size_t nmemb, FD_t fd)
{
    fdio_write_function_t _write;
    ssize_t rc;

    if (buf == NULL || size == 0 || nmemb == 0 || fd == NULL)
	return 0;

    FDSANE(fd);
DBGIO(fd, (stderr, "==> Fwrite(%p,%zu,%zu,%p)\n", buf, size, nmemb, (fd ? fd : NULL)));

    if (fdGetIo(fd) == fpio) {
	rc = fwrite(buf, size, nmemb, fdGetFILE(fd));
#ifdef	NOIST
DBGIO(fd, (stderr, "<== Fwrite(%p,%zu,%zu,%p) rc %zd\n", buf, size, nmemb, (fd ? fd : NULL), rc));
#endif
	return rc;
    }

    _write = FDIOVEC(fd, write);

    rc = (_write ? _write(fd, (const char *)buf, size * nmemb) : -2);
#ifdef	NOISY
DBGIO(fd, (stderr, "<== Fwrite(%p,%zu,%zu,%p) rc %zd\n", buf, size, nmemb, (fd ? fd : NULL), (rc >= 0 ? (rc/size) : 0)));
#endif
    return (rc >= 0 ? (rc/size) : 0);
}

int Fseek(FD_t fd, _libio_off_t offset, int whence)
{
    fdio_seek_function_t _seek;
#ifdef USE_COOKIE_SEEK_POINTER
    _IO_off64_t o64 = offset;
    _libio_pos_t pos = &o64;
#else
    _libio_pos_t pos = offset;
#endif

    ssize_t rc;

    FDSANE(fd);
#ifdef	NOISY
DBGIO(fd, (stderr, "==> Fseek(%p, %ld, SEEK_%s) %s\n", fd, (long)offset, SEEK_[whence&0x3], fdbg(fd)));
#endif

    if (fdGetIo(fd) == fpio) {
	rc = fseek(fdGetFILE(fd), (long)offset, whence);
DBGIO(fd, (stderr, "<== Fseek(%p, %ld, SEEK_%s) rc %zd %s\n", fd, (long)offset, SEEK_[whence&0x3], rc, fdbg(fd)));
	return (int) rc;
    }

    _seek = FDIOVEC(fd, seek);

    rc = (_seek ? _seek(fd, pos, whence) : -2);
DBGIO(fd, (stderr, "<== Fseek(%p, %ld, SEEK_%s) rc %zd %s\n", fd, (long)offset, SEEK_[whence&0x3], rc, fdbg(fd)));
    return (rc == 0 ? rc : -1);
}

long Ftell(FD_t fd)
{
    long rc = -2;

    FDSANE(fd);
#ifdef	NOISY
DBGIO(fd, (stderr, "==> Ftell(%p) %s\n", fd, fdbg(fd)));
#endif

    if (fdGetIo(fd) == fpio)
	rc = ftell(fdGetFILE(fd));
    else {
#ifdef USE_COOKIE_SEEK_POINTER
	_IO_off64_t o64 = 0;
	_libio_pos_t pos = &o64;
#else
	_libio_pos_t pos = 0;
#endif
	fdio_seek_function_t _seek = FDIOVEC(fd, seek);
	rc = (_seek ? _seek(fd, pos, SEEK_CUR) : -2);
    }
DBGIO(fd, (stderr, "<== Ftell(%p) rc %ld %s\n", fd, rc, fdbg(fd)));
    return (rc >= 0 ? rc : -1);
}

void Rewind(FD_t fd)
{
    FDSANE(fd);
DBGIO(fd, (stderr, "==> Rewind(%p) %s\n", fd, fdbg(fd)));

    if (fdGetIo(fd) == fpio)
	rewind(fdGetFILE(fd));
    else {
#ifdef USE_COOKIE_SEEK_POINTER
	_IO_off64_t o64 = 0;
	_libio_pos_t pos = &o64;
#else
	_libio_pos_t pos = 0;
#endif
	fdio_seek_function_t _seek = FDIOVEC(fd, seek);
	long rc = (_seek ? _seek(fd, pos, SEEK_SET) : -2);
DBGIO(fd, (stderr, "*** Rewind(%p) rc %ld %s\n", fd, rc, fdbg(fd)));
    }
}

int Fgetpos(FD_t fd, fpos_t *pos)
{
    int rc = -2;

    FDSANE(fd);

    if (fdGetIo(fd) == fpio)
	rc = fgetpos(fdGetFILE(fd), pos);
    else {
	errno = EBADF;	/* XXX identify mbstate_t internals. */
    }
DBGIO(fd, (stderr, "<== Fgetpos(%p,%p) rc %d %s\n", fd, pos, rc, fdbg(fd)));
    return (rc == 0 ? rc : -1);
}

int Fsetpos(FD_t fd, fpos_t *pos)
{
    int rc = -2;

    FDSANE(fd);

    if (fdGetIo(fd) == fpio)
	rc = fgetpos(fdGetFILE(fd), pos);
    else {
	errno = EBADF;	/* XXX identify mbstate_t internals. */
    }
DBGIO(fd, (stderr, "<== Fsetpos(%p,%p) rc %d %s\n", fd, pos, rc, fdbg(fd)));
    return (rc == 0 ? rc : -1);
}

int Fclose(FD_t fd)
{
    int rc = 0;
    int ec = 0;

    if (fd == NULL)
	return -1;

    FDSANE(fd);
DBGIO(fd, (stderr, "==> Fclose(%p) %s\n", fd, fdbg(fd)));

    fd = fdLink(fd, "Fclose");
    if (fd != NULL)
    while (fd->nfps >= 0) {
	FDSTACK_t * fps = &fd->fps[fd->nfps];
	
	if (fps->io == fpio) {
	    FILE *fp;
	    int fpno;

	    fp = fdGetFILE(fd);
	    fpno = fileno(fp);
	/* XXX persistent HTTP/1.1 returns the previously opened fp */
	    if (fd->nfps > 0 && fpno == -1 &&
		fd->fps[fd->nfps-1].io == ufdio &&
		fd->fps[fd->nfps-1].fp == fp &&
		(fd->fps[fd->nfps-1].fdno >= 0 || fd->req != NULL))
	    {
		int hadreqpersist = (fd->req != NULL);

		if (fp)
		    rc = fflush(fp);
		fd->nfps--;
		rc = ufdClose(fd);
		if (fdGetFdno(fd) >= 0)
		    break;
		if (!fd->persist)
		    hadreqpersist = 0;
		fdSetFp(fd, NULL);
		fd->nfps++;
		if (fp) {
		    /* HACK: flimsy Keepalive wiring. */
		    if (hadreqpersist) {
#ifdef	NOTYET	/* XXX not quite right yet. */
			(void) davDisconnect(fd);
			fd->req = NULL;
#endif
			fd->nfps--;
			fdSetFp(fd, fp);
			(void) fdClose(fd);
			fdSetFp(fd, NULL);
			fd->nfps++;
			(void) fdClose(fd);
		    } else
			rc = fclose(fp);
		}
		fdPop(fd);
		if (noLibio)
		    fdSetFp(fd, NULL);
	    } else {
		if (fp)
		    rc = fclose(fp);
		if (fpno == -1) {
		    fd = fdFree(fd, "fopencookie (Fclose)");
		    fdPop(fd);
		}
	    }
	} else {
	    fdio_close_function_t _close = FDIOVEC(fd, close);
	    rc = _close(fd);
	}
	if (fd == NULL || fd->nfps == 0)	/* XXX fd != NULL ever */
	    break;
	if (ec == 0 && rc)
	    ec = rc;
	fdPop(fd);
    }
    fd = fdFree(fd, "Fclose");
    return ec;
}

/**
 * Convert stdio fmode to open(2) mode, filtering out zlib/bzlib flags.
 *	returns stdio[0] = NUL on error.
 *
 * @todo glibc also supports ",ccs="
 *
 * - glibc:	'c' no cancel
 * - glibc:	'e' close on exec (O_CLOEXEC)
 * - glibc:	'm' use mmap'd input
 * - glibc:	'x' don't clobber (O_EXCL)
 * - gzopen:	[0-9] is compression level
 * - gzopen:	'f' is filtered (Z_FILTERED)
 * - gzopen:	'h' is Huffman encoding (Z_HUFFMAN_ONLY)
 * - bzopen:	[1-9] is block size (in 100K units)
 * - bzopen:	's' is smallmode
 * - bzopen:	'q' sets verbosity to 0
 * - bzopen:	'v' does verbosity++ (up to 4)
 * - HACK:	'.' terminates, rest is type of I/O
 * - HACK:	'n' non-blocking (O_NONBLOCK)
 * - HACK:	't' truncate (O_TRUNC)
 * - HACK:	'D' sync (O_DIRECT)
 * - HACK:	'S' sync (O_SYNC)
 * - HACK:	'I' fallocate(2)
 * - HACK:	'O' fdatasync(2)
 * - HACK:	'N' fadvise(2)
 * - HACK:	'F' fsync(2)
 * - HACK:	'P' syncfs(2)
 * - HACK:	'T' tempfile (O_TMPFILE)
 * - HACK:	'?' debug I/O + refcnt
 */
static inline void cvtfmode (const char *m,
				char *stdio, size_t nstdio,
				char *other, size_t nother,
				const char **end, int * f)
{
    int flags = 0;
    char c;

    switch (*m) {
    case 'a':
	flags &= ~O_ACCMODE;
	flags |= O_WRONLY | O_CREAT | O_APPEND;
	if (--nstdio > 0) *stdio++ = *m;
	break;
    case 'w':
	flags &= ~O_ACCMODE;
	flags |= O_WRONLY | O_CREAT | O_TRUNC;
	if (--nstdio > 0) *stdio++ = *m;
	break;
    case 'r':
	flags &= ~O_ACCMODE;
	flags |= O_RDONLY;
	if (--nstdio > 0) *stdio++ = *m;
	break;
    default:
	*stdio = '\0';
	return;
	break;
    }
    m++;

    while ((c = *m++) != '\0') {
	switch (c) {
	case '.':
	    break;
	case '+':
	    flags &= ~O_ACCMODE;
	    flags |= O_RDWR;
	    if (--nstdio > 0) *stdio++ = c;
	    continue;
	    break;
	case 'x':	/* glibc: don't clobber */
	    flags |= O_EXCL;
	    /*@fallthrough@*/
	case 'c':	/* glibc: no cancel */
	case 'e':	/* glibc: close on exec */
	case 'm':	/* glibc: mmap'd reads */
#if defined(__GLIBC__) && __GLIBC__ == 2 && __GLIBC_MINOR__ >= 3
	    if (--nstdio > 0) *stdio++ = c;
#endif
	    continue;
	    break;
	case 'b':
	    if (--nstdio > 0) *stdio++ = c;
	    continue;
	    break;
	case 'n':
	    flags |= O_NONBLOCK;
	    goto other;
	case 'D':
	    flags |= O_DIRECT;
	    RPMFD_SET(flags, FTRUNCATE);
	    goto other;
	case 'S':
	    flags |= O_SYNC;	/* XXX O_DSYNC? */
	    goto other;
	case 'T':
#if defined(O_TMPFILE)
	    flags |= O_TMPFILE;
#endif
	    goto other;
	break;
	case 'I':
	    RPMFD_SET(flags, FALLOCATE);
	    goto other;
	case 'O':
	    RPMFD_SET(flags, FDATASYNC);
	    goto other;
	case 'N':		/* XXX O_NONBLOCK? */
	    RPMFD_SET(flags, FADVISE);
	    goto other;
	case 'F':
	    RPMFD_SET(flags, FSYNC);
	    goto other;
	case 'P':
	    RPMFD_SET(flags, SYNCFS);
	    goto other;
	case '?':
	    RPMFD_SET(flags, DEBUGIO);
	    RPMFD_SET(flags, DEBUGREFS);
	    goto other;
	default:
	  other:
	    if (--nother > 0) *other++ = c;
	    continue;
	}
	break;
    }
    if (c == '\0') 	m--;	/* one too far */

    *stdio = *other = '\0';
    if (end != NULL)
	*end = (*m != '\0' ? m : NULL);
    if (f != NULL)
	*f = flags;
}

#if defined(HAVE_FOPENCOOKIE)
#if defined(__GLIBC__) && __GLIBC__ == 2 && __GLIBC_MINOR__ == 0
/* XXX retrofit glibc-2.1.x typedef on glibc-2.0.x systems */
typedef _IO_cookie_io_functions_t cookie_io_functions_t;
#endif
#endif

FD_t Fdopen(FD_t ofd, const char *fmode)
{
    char stdio[64], other[64], zfmode[64+1];
    const char *end = NULL;
    FDIO_t iof = NULL;
    FD_t fd = ofd;

if (_rpmio_debug)
fprintf(stderr, "*** Fdopen(%p,%s) %s\n", fd, fmode, fdbg(fd));
    FDSANE(fd);

    if (fmode == NULL)
	return NULL;

    cvtfmode(fmode, stdio, sizeof(stdio), other, sizeof(other), &end, NULL);
    if (stdio[0] == '\0')
	return NULL;
    zfmode[0] = '\0';
    (void) stpcpy( stpcpy(zfmode, stdio), other);

    if (end == NULL && other[0] == '\0')
	return fd;

    if (end && *end) {
	if (!strcmp(end, "fdio")) {
	    iof = fdio;
#if defined(WITH_ZLIB)
	} else if (!strcmp(end, "gzdio")) {
	    iof = gzdio;
	    fd = iof->_fdopen(fd, zfmode);
#endif
#if defined(WITH_BZIP2)
	} else if (!strcmp(end, "bzdio")) {
	    iof = bzdio;
	    fd = iof->_fdopen(fd, zfmode);
#endif
#if defined(WITH_XZ)
	} else if (!strcmp(end, "lzdio")) {
	    iof = lzdio;
	    fd = iof->_fdopen(fd, zfmode);
	} else if (!strcmp(end, "xzdio")) {
	    iof = xzdio;
	    fd = iof->_fdopen(fd, zfmode);
#endif
#if defined(WITH_ZSTD)
	} else if (!strcmp(end, "zstdio")) {
	    iof = zstdio;
	    fd = iof->_fdopen(fd, zfmode);
#endif
#if defined(WITH_MQ) || defined(WITH_MSQ)
	} else if (!strcmp(end, "msqio")) {
	    iof = msqio;
	    /* XXX already opened in Fopen */
#endif
	} else if (!strcmp(end, "ufdio")) {
	    iof = ufdio;
	} else if (!strcmp(end, "fpio")) {
	    iof = fpio;
	    if (noLibio) {
		int fdno = Fileno(fd);
		FILE * fp = fdopen(fdno, stdio);
if (_rpmio_debug)
fprintf(stderr, "*** Fdopen fpio fp %p\n", (void *)fp);
		if (fp == NULL)
		    return NULL;
		/* XXX gzdio/bzdio/zstdio use fp for private data */
		if (fdGetFp(fd) == NULL)
		    fdSetFp(fd, fp);
		fdPush(fd, fpio, fp, fdno);	/* Push fpio onto stack */
	    }
	}
    } else if (other[0] != '\0') {
	for (end = other; *end && strchr("0123456789fh", *end); end++)
	    {};
	if (*end == '\0') {
#if defined(WITH_ZLIB)
	    iof = gzdio;
	    fd = iof->_fdopen(fd, zfmode);
#endif
	}
    }
    if (iof == NULL)
	return fd;

    if (!noLibio) {
	FILE * fp = NULL;

#if _USE_LIBIO
#if defined(HAVE_FOPENCOOKIE) && defined(linux)
	{   cookie_io_functions_t ciof;
	    ciof.read = iof->read;
	    ciof.write = iof->write;
	    ciof.seek = iof->seek;
	    ciof.close = iof->close;
	    fp = fopencookie(fd, stdio, ciof);
DBGIO(fd, (stderr, "<-- fopencookie(%p,\"%s\",*%p) returns fp %p\n", fd, stdio, iof, fp));
	}
#elif defined(HAVE_FUNOPEN)
	{   void * cookie = (void *) fd;
	    int    (*readfn)  (void *cookie, char *, size_t) =
			(int (*) (void *, char *, size_t)) iof->read;
	    int    (*writefn) (void *cookie, const char *, size_t) =
			(int (*) (void *, const char *, size_t)) iof->write;
	    fpos_t (*seekfn)  (void *cookie, fpos_t, int) =
			(fpos_t (*) (void *, fpos_t, int)) iof->seek;
	    int    (*closefn) (void *cookie) =
			(int (*) (void *)) iof->close;
	    /* XXX FIXME: read/write/seek/close vectors based on stdio fmode. */
	    fp = funopen(cookie, readfn, writefn, seekfn, closefn);
DBGIO(fd, (stderr, "<-- funopen(%p,...) mode %s returns fp %p\n", fd, stdio, fp));
	}
#else
#error	_USE_LIBIO needs either fopencookie(3) or funopen(3).
#endif	/* FOPENCOOKIE || FUNOPEN */
#endif	/* _USE_LIBIO */

	if (fp) {

	    if (fd->flags & O_DIRECT)
		assert(setvbuf(fp, NULL, _IONBF, 0) == 0);

	    /* XXX gzdio/bzdio use fp for private data */
	    if (fdGetFp(fd) == NULL)
		fdSetFp(fd, fp);
	    fdPush(fd, fpio, fp, fileno(fp));	/* Push fpio onto stack */
	    fd = fdLink(fd, "fopencookie");
	}
    }

DBGIO(fd, (stderr, "<== Fdopen(%p,\"%s\") returns fd %p %s\n", ofd, fmode, (fd ? fd : NULL), fdbg(fd)));
    return fd;
}

FD_t Fopen(const char *path, const char *_fmode)
{
    const char * fmode = NULL;
    char stdio[64], other[64];
    const char *end = NULL;
    mode_t perms = 0666;
    int flags = 0;
    FD_t fd = NULL;

    if (path == NULL || _fmode == NULL)
	goto exit;
    fmode = rpmExpand(_fmode, NULL);

if (_rpmio_debug)
fprintf(stderr, "==> Fopen(%s, %s)\n", path, fmode);

    stdio[0] = '\0';
    cvtfmode(fmode, stdio, sizeof(stdio), other, sizeof(other), &end, &flags);
    if (stdio[0] == '\0')
	goto exit;

#if defined(WITH_MQ) || defined(WITH_MSQ)
    if (end && !strcmp(end, "msqio")) {
	fd = msqio->_fopen(path, fmode);
	if (fd) {
	    fd->flags = flags;
	    fd = Fdopen(fd, fmode);
	}
	goto exit;
    } else
#endif
    if (end && !strcmp(end, "fdio")) {
	fd = fdOpen(path, flags, perms);
	if (fdFileno(fd) < 0) {
	    if (fd) (void) fdClose(fd);
	    fd = NULL;
	    goto exit;
	}
    } else {
	FILE *fp;
	int fdno;
	int isHTTP = 0;

	/* XXX gzdio/bzdio/lzdio/xzdio through here too */

	switch (urlIsURL(path)) {
	case URL_IS_HKP:
	case URL_IS_HTTP:
	case URL_IS_HTTPS:
	    isHTTP = 1;
	    /*@fallthrough@*/
	case URL_IS_PATH:
	case URL_IS_DASH:
	case URL_IS_FTP:
	case URL_IS_UNKNOWN:
	    fd = ufdOpen(path, flags, perms);
	    if (fd == NULL || !(fdFileno(fd) >= 0 || fd->req != NULL)) {
		if (fd) (void) fdClose(fd);
		fd = NULL;
		goto exit;
	    }
	    break;
	case URL_IS_MONGO:	/* XXX FIXME */
	default:
	    goto exit;
	    break;
	}

	/* XXX persistent HTTP/1.1 returns the previously opened fp */
	if (isHTTP && ((fp = (FILE *) fdGetFp(fd)) != NULL)
	 && ((fdno = fdGetFdno(fd)) >= 0 || fd->req != NULL))
	{
	    fdPush(fd, fpio, fp, fileno(fp));	/* Push fpio onto stack */
	    goto exit;
	}
    }

    if (fd)
	fd = Fdopen(fd, fmode);
exit:

if (_rpmio_debug)
fprintf(stderr, "<== Fopen(%s, %s) fd %p\n", path, fmode, fd);
    fmode = _free(fmode);
    return fd;
}

int Fflush(FD_t fd)
{
    void * vh;
    if (fd == NULL) return -1;
    if (fdGetIo(fd) == fpio)
	return fflush(fdGetFILE(fd));

    vh = fdGetFp(fd);
#if defined(WITH_ZLIB)
    if (vh && fdGetIo(fd) == gzdio && gzdio->_flush != NULL)
	return (*gzdio->_flush) ((void *)fd);
#endif
#if defined(WITH_BZIP2)
    if (vh && fdGetIo(fd) == bzdio && bzdio->_flush != NULL)
	return (*bzdio->_flush) ((void *)fd);
#endif
#if defined(WITH_XZ)
    if (vh && fdGetIo(fd) == lzdio && lzdio->_flush != NULL)
	return (*lzdio->_flush) ((void *)fd);
    if (vh && fdGetIo(fd) == xzdio && xzdio->_flush != NULL)
	return (*xzdio->_flush) ((void *)fd);
#endif
#if defined(WITH_ZSTD)
    if (vh && fdGetIo(fd) == zstdio && zstdio->_flush != NULL)
	return (*bzdio->_flush) ((void *)fd);
#endif
#if defined(WITH_MQ) || defined(WITH_MSQ)
    if (vh && fdGetIo(fd) == msqio && msqio->_flush != NULL)
	return (*msqio->_flush) ((void *)fd);
#endif

    return 0;
}

int Ferror(FD_t fd)
{
    int i, rc = 0;

    if (fd == NULL) return -1;
    if (fd->req != NULL) {
	/* HACK: flimsy wiring for neon errors. */
	rc = (fd->syserrno  || fd->errcookie != NULL) ? -1 : 0;
    } else
    for (i = fd->nfps; rc == 0 && i >= 0; i--) {
	FDSTACK_t * fps = &fd->fps[i];
	int ec;
	
	if (fps->io == fpio) {
	    ec = ferror(fdGetFILE(fd));
#if defined(WITH_ZLIB)
	} else if (fps->io == gzdio) {
	    ec = (fd->syserrno || fd->errcookie != NULL) ? -1 : 0;
	    i--;	/* XXX fdio under gzdio always has fdno == -1 */
#endif
#if defined(WITH_BZIP2)
	} else if (fps->io == bzdio) {
	    ec = (fd->syserrno  || fd->errcookie != NULL) ? -1 : 0;
	    i--;	/* XXX fdio under bzdio always has fdno == -1 */
#endif
#if defined(WITH_XZ)
	} else if (fps->io == lzdio) {
	    ec = (fd->syserrno  || fd->errcookie != NULL) ? -1 : 0;
	    i--;	/* XXX fdio under lzdio always has fdno == -1 */
	} else if (fps->io == xzdio) {
	    ec = (fd->syserrno  || fd->errcookie != NULL) ? -1 : 0;
	    i--;	/* XXX fdio under xzdio always has fdno == -1 */
#endif
#if defined(WITH_ZSTD)
	} else if (fps->io == zstdio) {
	    ec = (fd->syserrno  || fd->errcookie != NULL) ? -1 : 0;
	    i--;	/* XXX fdio under zstdio always has fdno == -1 */
#endif
#if defined(WITH_MQ) || defined(WITH_MSQ)
	} else if (fps->io == msqio) {
	    ec = (fd->syserrno  || fd->errcookie != NULL) ? -1 : 0;
	    i--;	/* XXX fdio under bzdio always has fdno == -1 */
#endif
	} else {
	/* XXX need to check ufdio/gzdio/bzdio/fdio errors correctly. */
	    ec = (fdFileno(fd) < 0 ? -1 : 0);
	}

	if (rc == 0 && ec)
	    rc = ec;
    }
DBGIO(fd, (stderr, "<== Ferror(%p) rc %d %s\n", fd, rc, fdbg(fd)));
    return rc;
}

int Fileno(FD_t fd)
{
    int rc = -1;

    if (fd == NULL)
        return rc;
    if (fd->req != NULL)
	rc = 123456789;	/* HACK: https has no steenkin fileno. */
    else
    for (int i = fd->nfps ; rc == -1 && i >= 0; i--) {
	rc = fd->fps[i].fdno;
    }

#ifdef	NOISY
DBGIO(fd, (stderr, "<== Fileno(%p) rc %d %s\n", (fd ? fd : NULL), rc, fdbg(fd)));
#endif
    return rc;
}

/* =============================================================== */
#define	fdSyscall(_syscall)	(fdno >= 0 ? (_syscall) : -2)

int Fcntl(FD_t fd, int cmd, ...)
{
    static const int opx = FDSTAT_FCNTL;
    int fdno = fdSyscallPreFdno(__FUNCTION__, fd, 1, NULL);
    va_list ap;
    void * arg = NULL;
    int rc = -2;

    va_start(ap, cmd);
    arg = va_arg(ap, void *);
    va_end(ap);

    fdstat_enter(fd, opx);
    rc = fdSyscall( fcntl(fdno, cmd, arg) );
    fdstat_exit(fd, opx, 0);

    return fdSyscallPostFdno(__FUNCTION__, fd, fdno, rc);
}

int Flock(FD_t fd, int op)
{
    static const int opx = FDSTAT_FLOCK;
    int fdno = fdSyscallPreFdno(__FUNCTION__, fd, 1, NULL);
    fdstat_enter(fd, opx);
    int rc = fdSyscall( flock(fdno, op) );
    fdstat_exit(fd, opx, 0);
    return fdSyscallPostFdno(__FUNCTION__, fd, fdno, rc);
}

int Fchdir(FD_t fd)
{
    static const int opx = FDSTAT_FCHDIR;
    int fdno = fdSyscallPreFdno(__FUNCTION__, fd, 1, NULL);
    fdstat_enter(fd, opx);
    int rc = fdSyscall( fchdir(fdno) );
    fdstat_exit(fd, opx, 0);
    return fdSyscallPostFdno(__FUNCTION__, fd, fdno, rc);
}

int Fchmod(FD_t fd, mode_t mode)
{
    static const int opx = FDSTAT_FCHMOD;
    int fdno = fdSyscallPreFdno(__FUNCTION__, fd, 1, NULL);
    fdstat_enter(fd, opx);
    int rc = fdSyscall( fchmod(fdno, mode) );
    fdstat_exit(fd, opx, 0);
    return fdSyscallPostFdno(__FUNCTION__, fd, fdno, rc);
}

int Fchown(FD_t fd, uid_t uid, gid_t gid)
{
    static const int opx = FDSTAT_FCHOWN;
    int fdno = fdSyscallPreFdno(__FUNCTION__, fd, 1, NULL);
    fdstat_enter(fd, opx);
    int rc = fdSyscall( fchown(fdno, uid, gid) );
    fdstat_exit(fd, opx, 0);
    return fdSyscallPostFdno(__FUNCTION__, fd, fdno, rc);
}

int Futimes(FD_t fd, const struct timeval tv[2])
{
    static const int opx = FDSTAT_FUTIMES;
    int fdno = fdSyscallPreFdno(__FUNCTION__, fd, 1, NULL);
    fdstat_enter(fd, opx);
#if defined(HAVE_FUTIMENS)
    struct timespec times[2];
    times[0].tv_sec = tv[0].tv_sec;
    times[0].tv_nsec = tv[0].tv_usec * 1000;
    times[1].tv_sec = tv[1].tv_sec;
    times[1].tv_nsec = tv[1].tv_usec * 1000;
    int rc = fdSyscall( futimens(fdno, times) );
#elif defined(HAVE_FUTIMES)
    int rc = fdSyscall( futimes(fdno, tv) );
#else
    errno = ENOSYS;
#endif
    fdstat_exit(fd, opx, 0);
    return fdSyscallPostFdno(__FUNCTION__, fd, fdno, rc);
}

int Futimens(FD_t fd, const struct timespec times[2])
{
    static const int opx = FDSTAT_FUTIMENS;
    int fdno = fdSyscallPreFdno(__FUNCTION__, fd, 1, NULL);
    fdstat_enter(fd, opx);
#if defined(HAVE_FUTIMENS)
    int rc = fdSyscall( futimens(fdno, times) );
#elif defined(HAVE_FUTIMES)
    struct timeval tv[2];
    tv[0].tv_sec = times[0].tv_sec;
    tv[0].tv_usec = times[0].tv_nsec / 1000;
    tv[1].tv_sec = times[1].tv_sec;
    tv[1].tv_usec = times[1].tv_nsec / 1000;
    int rc = fdSyscall( futimes(fdno, tv) );
#else
    errno = ENOSYS;
#endif
    fdstat_exit(fd, opx, 0);
    return fdSyscallPostFdno(__FUNCTION__, fd, fdno, rc);
}

ssize_t Fgetxattr(FD_t fd, const char *name, void *value, size_t size)
{
    static const int opx = FDSTAT_FGETXATTR;
    int fdno = fdSyscallPreFdno(__FUNCTION__, fd, 1, NULL);
    ssize_t rc = -2;

    fdstat_enter(fd, opx);
#if defined(HAVE_FGETXATTR)
#if defined(linux)
    rc = fdSyscall( fgetxattr(fdno, name, value, size) );
#elif defined(__APPLE__) || defined(__FreeBSD__) || defined(__NetBSD__)
    uint32_t _position = 0;
    int _options = 0;	/* XXX XATTR_{NOFOLLOW,SHOWCOMPRESSION} */
    rc = fdSyscall( fgetxattr(fdno, name, value, size, _position, _options) );
#else
    errno = ENOSYS;
#endif
#endif
    fdstat_exit(fd, opx, (rc >= 0 ? rc : 0));
    return fdSyscallPostFdno(__FUNCTION__, fd, fdno, rc);
}

int Fsetxattr(FD_t fd, const char *name, const void *value, size_t size,
		int flags)
{
    static const int opx = FDSTAT_FSETXATTR;
    int fdno = fdSyscallPreFdno(__FUNCTION__, fd, 1, NULL);
    ssize_t rc = -2;

    fdstat_enter(fd, opx);
#if defined(HAVE_FSETXATTR)
#if defined(linux)
    rc = fdSyscall( fsetxattr(fdno, name, value, size, flags) );
#elif defined(__APPLE__) || defined(__FreeBSD__) || defined(__NetBSD__)
    uint32_t _position = 0;
    int _options = 0;	/* XXX XATTR_{NOFOLLOW,CREATE,REPLACE} */
    rc = fdSyscall( fsetxattr(fdno, name, value, size, _position, _options) );
#else
    errno = ENOSYS;
#endif
#endif
    fdstat_exit(fd, opx, (rc >= 0 ? size : 0));
    return fdSyscallPostFdno(__FUNCTION__, fd, fdno, rc);
}

ssize_t Flistxattr(FD_t fd, char *list, size_t size)
{
    static const int opx = FDSTAT_FLISTXATTR;
    int fdno = fdSyscallPreFdno(__FUNCTION__, fd, 1, NULL);
    ssize_t rc = -2;

    fdstat_enter(fd, opx);
#if defined(HAVE_FGETXATTR)
#if defined(linux)
    rc = fdSyscall( flistxattr(fdno, list, size) );
#elif defined(__APPLE__) || defined(__FreeBSD__) || defined(__NetBSD__)
    int _options = 0;	/* XXX XATTR_{NOFOLLOW,SHOWCOMPRESSION} */
    rc = fdSyscall( flistxattr(fdno, list, size, _options) );
#else
    errno = ENOSYS;
#endif
#endif
    fdstat_exit(fd, opx, (rc >= 0 ? rc : 0));
    return fdSyscallPostFdno(__FUNCTION__, fd, fdno, rc);
}

int Fremovexattr(FD_t fd, char *name)
{
    static const int opx = FDSTAT_FREMOVEXATTR;
    int fdno = fdSyscallPreFdno(__FUNCTION__, fd, 1, NULL);
    ssize_t rc = -2;

    fdstat_enter(fd, opx);
#if defined(HAVE_FGETXATTR)
#if defined(linux)
    rc = fdSyscall( fremovexattr(fdno, name) );
#elif defined(__APPLE__) || defined(__FreeBSD__) || defined(__NetBSD__)
    int _options = 0;	/* XXX XATTR_{NOFOLLOW,SHOWCOMPRESSION} */
    rc = fdSyscall( fremovexattr(fdno, name, _options) );
#else
    errno = ENOSYS;
#endif
#endif
    fdstat_exit(fd, opx, (rc >= 0 ? rc : 0));
    return fdSyscallPostFdno(__FUNCTION__, fd, fdno, rc);
}

int Fchflags(FD_t fd, unsigned int flags)
{
    static const int opx = FDSTAT_FCHFLAGS;
    int fdno = fdSyscallPreFdno(__FUNCTION__, fd, 1, NULL);
    int rc = -2;

    fdstat_enter(fd, opx);
#if defined(HAVE_FCHFLAGS)
    rc = fdSyscall( fchflags(fdno, flags) );
#else
    errno = ENOSYS;
#endif
    fdstat_exit(fd, opx, 0);
    return fdSyscallPostFdno(__FUNCTION__, fd, fdno, rc);
}

int Fadvise(FD_t fd, off_t offset, off_t len, int advice)
{
    static const int opx = FDSTAT_FADVISE;
    int fdno = fdSyscallPreFdno(__FUNCTION__, fd, 1, NULL);
    int rc = -2;

    if (fdno < 0) {
	/* XXX remap non-local errno */
	if (errno == EINVAL)
	    errno = ESPIPE;
    }
    fdstat_enter(fd, opx);
    switch (advice) {
#if defined(HAVE_POSIX_FADVISE)
    case POSIX_FADV_NORMAL:
    case POSIX_FADV_SEQUENTIAL:
    case POSIX_FADV_RANDOM:
    case POSIX_FADV_NOREUSE:
    case POSIX_FADV_WILLNEED:
    case POSIX_FADV_DONTNEED:
	rc = fdSyscall( posix_fadvise(fdno, offset, len, advice) );
#else
	errno = ENOSYS;
#endif
	break;
    default:
	errno = EINVAL;
	goto exit;
	break;
    }
    fdstat_exit(fd, opx, len);

exit:
    return fdSyscallPostFdno(__FUNCTION__, fd, fdno, rc);
}

#undef HAVE_FALLOCATE  /* XXX hmmm, fallocate64 is AWOL in F11. */
int Fallocate(FD_t fd, off_t offset, off_t len)
{
    static const int opx = FDSTAT_FALLOCATE;
    int fdno = fdSyscallPreFdno(__FUNCTION__, fd, 1, NULL);
    int rc = -2;

    fdstat_enter(fd, opx);
#if defined(HAVE_POSIX_FALLOCATE)
    rc = fdSyscall( posix_fallocate(fdno, offset, len) );
#elif defined(HAVE_FALLOCATE)
    /* XXX linux FALLOC_FL_KEEP_SIZE zeroes allocated blocks */
    rc = fdSyscall( fallocate(fdno, 0, (loff_t)offset, (loff_t)len) );
#else
    errno = ENOSYS;
#endif
    fdstat_exit(fd, opx, len);

    return fdSyscallPostFdno(__FUNCTION__, fd, fdno, rc);
}

int Ftruncate(FD_t fd, off_t len)
{
    static const int opx = FDSTAT_FTRUNCATE;
    int fdno = fdSyscallPreFdno(__FUNCTION__, fd, 1, NULL);
    int rc = -2;

    fdstat_enter(fd, opx);
#if defined(HAVE_FTRUNCATE)
    rc = fdSyscall( ftruncate(fdno, len) );
#else
    errno = ENOSYS;
#endif
    fdstat_exit(fd, opx, len);

    /* XXX doesn't display len */
    return fdSyscallPostFdno(__FUNCTION__, fd, fdno, rc);
}

int Fdatasync(FD_t fd)
{
    static const int opx = FDSTAT_FDATASYNC;
    int fdno = fdSyscallPreFdno(__FUNCTION__, fd, 1, NULL);
    int rc = -2;

    fdstat_enter(fd, opx);
#if defined(HAVE_FDATASYNC) || defined(HAVE_RAW_DECL_FDATASYNC)
    rc = fdSyscall( fdatasync(fdno) );
#else
    errno = ENOSYS;
#endif
    fdstat_exit(fd, opx, 0);	/* XXX #bytes? */

    return fdSyscallPostFdno(__FUNCTION__, fd, fdno, rc);
}

int Fsync(FD_t fd)
{
    static const int opx = FDSTAT_FSYNC;
    int fdno = fdSyscallPreFdno(__FUNCTION__, fd, 1, NULL);
    int rc = -2;

    fdstat_enter(fd, opx);
#if defined(HAVE_FSYNC) || defined(HAVE_RAW_DECL_FSYNC)
    rc = fdSyscall( fsync(fdno) );
#else
    errno = ENOSYS;
#endif
    fdstat_exit(fd, opx, 0);		/* XXX #bytes? */

    return fdSyscallPostFdno(__FUNCTION__, fd, fdno, rc);
}

int Syncfs(FD_t fd)
{
    static const int opx = FDSTAT_SYNCFS;
    int fdno = fdSyscallPreFdno(__FUNCTION__, fd, 1, NULL);
    int rc = -2;

    fdstat_enter(fd, opx);
#if defined(HAVE_SYNCFS)
    rc = fdSyscall( syncfs(fdno) );
#else
    errno = ENOSYS;
#endif
    fdstat_exit(fd, opx, 0);		/* XXX #bytes? */

    return fdSyscallPostFdno(__FUNCTION__, fd, fdno, rc);
}

int Fincore(FD_t fd, int nincore)
{
    static const int opx = FDSTAT_FINCORE;
    int fdno = fdSyscallPreFdno(__FUNCTION__, fd, 1, NULL);
    void * mapped = MAP_FAILED;
    size_t nmapped = 0;
    size_t pagesize = sysconf(_SC_PAGESIZE);
    struct stat sb;
    size_t npages;
    unsigned char * vec;
    int rc = -2;
  
    fdstat_enter(fd, opx);
    if (fdno < 0)
	goto exit;

    rc = fstat(fdno, &sb);
    if (rc)
	goto exit;
    if (sb.st_size == 0)
	goto exit;
  
    nmapped = sb.st_size;
    mapped = Mmap(NULL, nmapped, PROT_NONE, MAP_PRIVATE, fd, 0);
    if (mapped == MAP_FAILED) {
	rc = -1;
	goto exit;
    }

    npages = (nmapped + pagesize-1)/pagesize;
    vec = alloca(npages+1);
    rc = Mincore(mapped, nmapped, vec);
    if (rc)
	goto exit;
  
    rc = 0;
    for (size_t i = 0; i < npages; i++) {
	if (!vec[i] & 0x1)
	    continue;
	rc++;
    }
  
exit:
    if (mapped != MAP_FAILED)
	(void) Munmap(mapped, nmapped);
    if (nincore > 0 && rc >= 0)
	nincore -= rc;
    fdstat_exit(fd, opx, nincore * pagesize);
    return fdSyscallPostFdno(__FUNCTION__, fd, fdno, rc);
}

#undef	fdSyscall

/* =============================================================== */
/* Helper routines that may be generally useful.
 */
int rpmioMkpath(const char * path, mode_t mode, uid_t uid, gid_t gid)
{
    char * d, * de;
    int created = 0;
    int rc;

    if (path == NULL || *path == '\0')
	return -1;
    d = (char *) alloca(strlen(path)+2);
    de = stpcpy(d, path);
    de[1] = '\0';
    for (de = d; *de != '\0'; de++) {
	struct stat st;
	char savec;

	while (*de && *de != '/') de++;
	savec = de[1];
	de[1] = '\0';

	rc = Stat(d, &st);
	if (rc) {
	    switch(errno) {
	    default:
		return errno;
		break;
	    case ENOENT:
		break;
	    }
	    rc = Mkdir(d, mode);
	    if (rc)
		return errno;
	    created = 1;
	    if (!(uid == (uid_t) -1 && gid == (gid_t) -1)) {
		rc = Chown(d, uid, gid);
		if (rc)
		    return errno;
	    }
	} else if (!S_ISDIR(st.st_mode)) {
	    return ENOTDIR;
	}
	de[1] = savec;
    }
    rc = 0;
    if (created)
	rpmlog(RPMLOG_DEBUG, D_("created directory(s) %s mode 0%o\n"),
			path, (unsigned)mode);
    return rc;
}

#define	_PATH	"/bin:/usr/bin:/sbin:/usr/sbin"
static const char *_path = _PATH;

#define alloca_strdup(_s)       strcpy((char *)alloca(strlen(_s)+1), (_s))

int rpmioAccess(const char * FN, const char * path, int mode)
{
    char fn[4096];
    char * bn;
    char * r, * re;
    char * t, * te;
    int negate = 0;
    int rc = 0;

    /* Empty paths are always accessible. */
    if (FN == NULL || *FN == '\0')
	return 0;

    if (mode == 0)
	mode = X_OK;

    /* Strip filename out of its name space wrapper. */
    bn = (char *) alloca_strdup(FN);
    for (t = bn; t && *t; t++) {
	if (*t != '(')
	    continue;
	*t++ = '\0';

	/* Permit negation on name space tests. */
	if (*bn == '!') {
	    negate = 1;
	    bn++;
	}

	/* Set access flags from name space marker. */
	if (strlen(bn) == 3
	 && strchr("Rr_", bn[0]) != NULL
	 && strchr("Ww_", bn[1]) != NULL
	 && strchr("Xx_", bn[2]) != NULL) {
	    mode = 0;
	    if (strchr("Rr", bn[0]) != NULL)
		mode |= R_OK;
	    if (strchr("Ww", bn[1]) != NULL)
		mode |= W_OK;
	    if (strchr("Xx", bn[2]) != NULL)
		mode |= X_OK;
	    if (mode == 0)
		mode = F_OK;
	} else if (!strcmp(bn, "exists"))
	    mode = F_OK;
	else if (!strcmp(bn, "executable"))
	    mode = X_OK;
	else if (!strcmp(bn, "readable"))
	    mode = R_OK;
	else if (!strcmp(bn, "writable"))
	    mode = W_OK;

	bn = t;
	te = bn + strlen(t) - 1;
	if (*te != ')')		/* XXX syntax error, never exists */
	    return 1;
	*te = '\0';
	break;
    }

    /* Empty paths are always accessible. */
    if (*bn == '\0')
	goto exit;

    /* Check absolute path for access. */
    if (*bn == '/') {
	rc = (Access(bn, mode) != 0 ? 1 : 0);
if (_rpmio_debug)
fprintf(stderr, "*** rpmioAccess(\"%s\", 0x%x) rc %d\n", bn, mode, rc);
	goto exit;
    }

    /* Find path to search. */
    if (path == NULL)
	path = getenv("PATH");
    if (path == NULL)
	path = _path;
    if (path == NULL) {
	rc = 1;
	goto exit;
    }

    /* Look for relative basename on PATH. */
    for (r = alloca_strdup(path); r != NULL && *r != '\0'; r = re) {

	/* Find next element, terminate current element. */
	for (re = r; (re = strchr(re, ':')) != NULL; re++) {
	    if (!(re[1] == '/' && re[2] == '/'))
		break;
	}
	if (re && *re == ':')
	    *re++ = '\0';
	else
	    re = r + strlen(r);

	/* Expand ~/ to $HOME/ */
	fn[0] = '\0';
	t = fn;
	*t = '\0';	/* XXX redundant. */
	if (r[0] == '~' && r[1] == '/') {
	    const char * home = getenv("HOME");
	    if (home == NULL)	/* XXX No HOME? */
		continue;
	    if (strlen(home) > (sizeof(fn) - strlen(r))) /* XXX too big */
		continue;
	    t = stpcpy(t, home);
	    r++;	/* skip ~ */
	}
	t = stpcpy(t, r);
	if (t[-1] != '/' && *bn != '/')
	    *t++ = '/';
	t = stpcpy(t, bn);
	t = rpmCleanPath(fn);
	if (t == NULL)	/* XXX can't happen */
	    continue;

	/* Check absolute path for access. */
	rc = (Access(t, mode) != 0 ? 1 : 0);
if (_rpmio_debug)
fprintf(stderr, "*** rpmioAccess(\"%s\", 0x%x) rc %d\n", t, mode, rc);
	if (rc == 0)
	    goto exit;
    }

    rc = 1;

exit:
    if (negate)
	rc ^= 1;
    return rc;
}

#if defined(WITH_NSS)	/* XXX TODO: add nssDestroy */
#ifdef __cplusplus
extern "C" {
#endif
#ifdef	HAVE_NSS_INITCONTEXT
extern void NSS_ShutdownContext(void * _nss_context);
#else
extern void NSS_Shutdown(void);
#endif
extern void PL_ArenaFinish(void);
extern void PR_Cleanup(void);
#ifdef __cplusplus
}
#endif

int _rpmnss_init = 0;

void * _rpmnss_context = NULL;
#endif

void rpmioClean(void)
{
    extern rpmioPool _odbcPool;

#if defined(WITH_LUA)
    (void) rpmluaFree(NULL);
#endif
#if defined(WITH_NEON)
    davDestroy();
#endif
#if defined(WITH_NSS)
    if (_rpmnss_init) {
#ifdef	HAVE_NSS_INITCONTEXT
	if (_rpmnss_context != NULL)
	    (void) NSS_ShutdownContext(_rpmnss_context);
#else
	(void) NSS_Shutdown();
#endif
	PL_ArenaFinish();
	PR_Cleanup();
	_rpmnss_context = NULL;
	_rpmnss_init = 0;
    }
#endif
    urlFreeCache();

    RPMIOPOOL_INTERP_FREE(tcl)
    RPMIOPOOL_INTERP_FREE(squirrel)
    RPMIOPOOL_INTERP_FREE(sql)
    RPMIOPOOL_INTERP_FREE(ruby)
    RPMIOPOOL_INTERP_FREE(python)
    RPMIOPOOL_INTERP_FREE(perl)
    RPMIOPOOL_INTERP_FREE(mrb)
    RPMIOPOOL_INTERP_FREE(jss)
    RPMIOPOOL_INTERP_FREE(jni)
    RPMIOPOOL_INTERP_FREE(ficl)

    RPMIOPOOL_INTERP_FREE(git)

    RPMIOPOOL_INTERP_FREE(aug)
    RPMIOPOOL_INTERP_FREE(gfs)
    RPMIOPOOL_INTERP_FREE(mqtt)

#ifdef	NOTYET
    RPMIOPOOL_FREE(mgo)
    RPMIOPOOL_FREE(jsmn)
#endif
    RPMIOPOOL_FREE(asn)
    RPMIOPOOL_FREE(bag)
    RPMIOPOOL_FREE(ct)
    RPMIOPOOL_FREE(cudf)
    RPMIOPOOL_FREE(cvs)
    RPMIOPOOL_FREE(date)

    _odbcPool = rpmioFreePool(_odbcPool);
    RPMIOPOOL_FREE(sed)
    RPMIOPOOL_FREE(set)
    RPMIOPOOL_FREE(svn)
    RPMIOPOOL_FREE(tpm)

#ifdef	NOTYET	/* XXX FIXME: dig out the recursion deadlock. */
    RPMIOPOOL_INTERP_FREE(nix)
#endif

    RPMIOPOOL_FREE(luav)
    RPMIOPOOL_FREE(lua)

    RPMIOPOOL_INTERP_FREE(hkp)
    _rpmhkp_awol.bf = rpmbfFree(_rpmhkp_awol.bf);
    _rpmhkp_crl.bf = rpmbfFree(_rpmhkp_crl.bf);

    RPMIOPOOL_FREE(vc)
    RPMIOPOOL_FREE(vt)

    RPMIOPOOL_INTERP_FREE(sm)
    RPMIOPOOL_FREE(sp)
    RPMIOPOOL_INTERP_FREE(sx)

    RPMIOPOOL_FREE(aio)
    RPMIOPOOL_FREE(eio)
    RPMIOPOOL_FREE(ev)
    RPMIOPOOL_FREE(msq)
    RPMIOPOOL_FREE(w)
    RPMIOPOOL_FREE(zstd)

    RPMIOPOOL_FREE(html)
    RPMIOPOOL_FREE(mire)
    RPMIOPOOL_FREE(mg)
    RPMIOPOOL_FREE(bf)
    RPMIOPOOL_FREE(ht)
    RPMIOPOOL_FREE(cph)
    RPMIOPOOL_FREE(ctx)
    RPMIOPOOL_FREE(syck)
    RPMIOPOOL_FREE(iob)
    RPMIOPOOL_FREE(dig)
    RPMIOPOOL_FREE(xar)
    RPMIOPOOL_FREE(avx)
    RPMIOPOOL_FREE(url)
    _fdPool = rpmioFreePool(_fdPool);

    RPMIOPOOL_FREE(me)
    RPMIOPOOL_INTERP_FREE(mc)

    rpmlogClose();
}
