#ifndef	H_RPMIO
#define	H_RPMIO

/** \ingroup rpmio
 * \file rpmio/rpmio.h
 *
 */

#include <sys/types.h>
#include <features.h>
#include <stdio.h>

#include <rpmiotypes.h>
#include <rpmdefs.h>

/** \ingroup rpmio
 * Hide libio API lossage.
 * The libio interface changed after glibc-2.1.3 to pass the seek offset
 * argument as a pointer rather than as an off_t. The snarl below defines
 * typedefs to isolate the lossage.
 */
/*@{*/
#if !defined(__UCLIBC__) && defined(__GLIBC__) && \
	(__GLIBC__ > 2 || (__GLIBC__ == 2 && __GLIBC_MINOR__ >= 2)) && \
	!defined(__UCLIBC__)
#define USE_COOKIE_SEEK_POINTER 1
typedef _IO_off64_t 	_libio_off_t;
typedef _libio_off_t *	_libio_pos_t;
#else
typedef off_t 		_libio_off_t;
typedef off_t 		_libio_pos_t;
#endif
/*@}*/

/** \ingroup rpmio
 */
extern int _rpmio_debug;
extern const char * __progname;
extern int __progname_key;
extern int __progname_projid;
extern int _rpmioOMask;

/** \ingroup rpmio
 */
typedef	struct _FD_s * FD_t;
typedef	struct _FDSTAT_s * FDSTAT_t;

/** \ingroup rpmio
 */
typedef struct FDIO_s * FDIO_t;

#ifdef __cplusplus
extern "C" {
#endif

/** \ingroup rpmio
 * \name RPMIO Vectors.
 */
/*@{*/

/**
 */
typedef ssize_t (*fdio_read_function_t) (void *cookie, char *buf, size_t nbytes);

/**
 */
typedef ssize_t (*fdio_write_function_t) (void *cookie, const char *buf, size_t nbytes);

/**
 */
typedef int (*fdio_seek_function_t) (void *cookie, _libio_pos_t pos, int whence);

/**
 */
typedef int (*fdio_close_function_t) (void *cookie);

/**
 */
typedef FD_t (*fdio_fopen_function_t) (const char * path, const char * fmode);

/**
 */
typedef FD_t (*fdio_fdopen_function_t) (void * cookie, const char * fmode);

/**
 */
typedef int (*fdio_flush_function_t) (void * cookie);

/*@}*/

/** \ingroup rpmio
 */
struct FDIO_s {
    fdio_read_function_t		read;
    fdio_write_function_t		write;
    fdio_seek_function_t		seek;
    fdio_close_function_t		close;
    fdio_fopen_function_t		_fopen;
    fdio_fdopen_function_t		_fdopen;
    fdio_flush_function_t		_flush;
};


/** \ingroup rpmio
 * \name RPMIO Interface.
 */
/*@{*/

/**
 * strerror(3) clone.
 */
const char * Fstrerror(FD_t fd);

/**
 * fread(3) clone.
 */
size_t Fread(void * buf, size_t size, size_t nmemb, FD_t fd);

/**
 * fwrite(3) clone.
 */
size_t Fwrite(const void * buf, size_t size, size_t nmemb, FD_t fd);

/**
 * fseek(3) clone.
 */
int Fseek(FD_t fd, _libio_off_t offset, int whence);

long Ftell(FD_t fd);
void Rewind(FD_t fd);
int Fgetpos(FD_t fd, fpos_t *pos);
int Fsetpos(FD_t fd, fpos_t *pos);

/**
 * fclose(3) clone.
 */
int Fclose(FD_t fd);

/**
 */
FD_t	Fdopen(FD_t ofd, const char * fmode);

/**
 * fopen(3) clone.
 */
FD_t	Fopen(const char * path, const char * fmode);

/**
 * fflush(3) clone.
 */
int Fflush(FD_t fd);

/**
 * ferror(3) clone.
 */
int Ferror(FD_t fd);

/**
 * fileno(3) clone.
 */
RPM_GNUC_PURE
int Fileno(FD_t fd);

/**
 * fcntl(2) clone.
 */
int Fcntl(FD_t fd, int cmd, ...);

/**
 * flock(2) clone.
 */
int Flock(FD_t fd, int op);

/**
 * fchdir(2) clone.
 */
int Fchdir(FD_t fd);

/**
 * fchmod(2) clone.
 */
int Fchmod(FD_t fd, mode_t mode);

/**
 * fchown(2) clone.
 */
int Fchown(FD_t fd, uid_t uid, gid_t gid);

/**
 * futimes(2) clone.
 */
int Futimes(FD_t fd, const struct timeval tv[2]);

/**
 * futimens(2) clone.
 */
int Futimens(FD_t fd, const struct timespec times[2]);

/**
 * fgetxattr(2) clone.
 */
ssize_t Fgetxattr(FD_t fd, const char *name, void *value, size_t size);

/**
 * fsetxattr(2) clone.
 */
int Fsetxattr(FD_t fd, const char *name,
		const void *value, size_t size, int flags);

/**
 * flistxattr(2) clone.
 */
ssize_t Flistxattr(FD_t fd, char *list, size_t size);

/**
 * fremovexattr(2) clone.
 */
int Fremovexattr(FD_t fd, char *name);

/**
 * fchflags(2) clone.
 */
int Fchflags(FD_t fd, unsigned int flags);

/**
 * posix_fadvise(2) clone.
 */
int Fadvise(FD_t fd, off_t offset, off_t length, int advice);

/**
 * posix_fallocate(3)/fallocate(2) clone.
 */
int Fallocate(FD_t fd, off_t offset, off_t length);

/**
 * ftruncate(2) clone.
 */
int Ftruncate(FD_t fd, off_t len);

/**
 * fdatasync(2) clone.
 */
int Fdatasync(FD_t fd);

/**
 * fsync(2) clone.
 */
int Fsync(FD_t fd);

/**
 * syncfs(2) clone.
 */
int Syncfs(FD_t fd);

/**
 * Return no. of pages in cache (using mmap(2) and mincore(2)).
 * @param nincore	expected no. pages in core
 * @return		no. of pages in core, -1 on error
 */
int Fincore(FD_t fd, int nincore);

/*@}*/

/** \ingroup rpmrpc
 * \name RPMRPC Interface.
 */
/*@{*/

/**
 * mkdir(2) clone.
 */
int Mkdir(const char * path, mode_t mode);

/**
 * chdir(2) clone.
 */
int Chdir(const char * path);

/**
 * rmdir(2) clone.
 */
int Rmdir(const char * path);

extern const char * _chroot_prefix;

/**
 * chroot(2) clone.
 * @todo Implement remotely.
 */
int Chroot(const char * path);

/**
 * open(2) clone.
 * @todo Implement remotely.
 */
int Open(const char * path, int flags, mode_t mode);

/**
 * rename(2) clone.
 */
int Rename(const char * oldpath, const char * newpath);

/**
 * link(2) clone.
 */
int Link(const char * oldpath, const char * newpath);

/**
 * unlink(2) clone.
 */
int Unlink(const char * path);

/**
 * stat(2) clone.
 */
struct stat;
int Stat(const char * path, struct stat * st);

/**
 * lstat(2) clone.
 */
int Lstat(const char * path, struct stat * st);

/**
 * fstat(2) clone.
 */
int Fstat(FD_t fd, struct stat * st);

/**
 * mmap(2) clone.
 */
void *Mmap(void *addr, size_t len, int prot, int flags, FD_t fd, off_t offset);

/**
 * munmap(2) clone.
 */
int Munmap(void * addr, size_t len);

/**
 * msync(2) clone.
 */
int Msync(void * addr, size_t len, int flags);

/**
 * mremap(2) clone.
 */
void *Mremap(void * oaddr, size_t olen, size_t nlen, int flags, ...);

/**
 * mprotect(2) clone.
 */
int Mprotect(void * addr, size_t len, int prot);

/**
 * madvise(2) clone.
 */
int Madvise(void *addr, size_t len, int advice);

/**
 * mincore(2) clone.
 */
int Mincore(void * addr, size_t len, unsigned char * vec);

/**
 * splice(2) clone.
 */
ssize_t Splice(int fd_in, loff_t *off_in, int fd_out, loff_t *off_out,
                size_t len, unsigned int flags);

/**
 * tee(2) clone.
 */
ssize_t Tee(int fd_in, int fd_out, size_t len, unsigned int flags);

/**
 * vmsplice(2) clone.
 */
ssize_t Vmsplice(int fd, const struct iovec *iov,
                unsigned long nr_segs, unsigned int flags);

/**
 * getauxval(2) clone.
 */
unsigned long Getauxval(unsigned long type);

/**
 * personality(2) clone.
 */
int Personality(unsigned long persona);

/**
 * prctl(2) clone.
 */
int Prctl(int option, unsigned long arg2, unsigned long arg3,
                   unsigned long arg4, unsigned long arg5);

/**
 * chown(2) clone.
 * @todo Implement remotely.
 */
int Chown(const char * path, uid_t owner, gid_t group);

/**
 * lchown(2) clone.
 * @todo Implement remotely.
 */
int Lchown(const char * path, uid_t owner, gid_t group);

/**
 * chmod(2) clone.
 * @todo Implement remotely.
 */
int Chmod(const char * path, mode_t mode);

/**
 * lchmod(2) clone.
 * @todo Implement remotely.
 */
int Lchmod(const char * path, mode_t mode);

/**
 * chflags(2) clone.
 * @todo Implement remotely.
 */
int Chflags(const char * path, unsigned int flags);

/**
 * lchflags(2) clone.
 * @todo Implement remotely.
 */
int Lchflags(const char * path, unsigned int flags);

/**
 * mkfifo(3) clone.
 * @todo Implement remotely.
 */
int Mkfifo(const char * path, mode_t mode);

/**
 * mknod(3) clone.
 * @todo Implement remotely.
 */
int Mknod(const char * path, mode_t mode, dev_t dev);

/**
 * utime(2) clone.
 * @todo Implement remotely.
 */
struct utimbuf;
int Utime(const char * path, const struct utimbuf * buf);

/**
 * utimes(2) clone.
 * @todo Implement remotely.
 */
struct timeval;
int Utimes(const char * path, const struct timeval * times);

/**
 * lutimes(2) clone.
 * @todo Implement remotely.
 */
int Lutimes(const char * path, const struct timeval * times);

/**
 * symlink(3) clone.
 * @todo Implement remotely.
 */
int Symlink(const char * oldpath, const char * newpath);

/**
 * readlink(2) clone.
 * @todo Implement remotely.
 */
int Readlink(const char * path, char * buf, size_t bufsiz);

/**
 * access(2) clone.
 * @todo Implement remotely.
 */
int Access(const char * path, int amode);

#if defined(__linux__)
/**
 * mount(2) clone.
 */
int Mount(const char *source, const char *target,
		const char *filesystemtype, unsigned long mountflags,
		const void *data);

/**
 * umount(2) clone.
 */
int Umount(const char *target);

/**
 * umount2(2) clone.
 */
int Umount2(const char *target, int flags);
#endif

/**
 * glob_pattern_p(3) clone.
 */
int Glob_pattern_p (const char *pattern, int quote);

/**
 * glob_error(3) clone.
 */
int Glob_error(const char * epath, int eerrno)
	RPM_GNUC_CONST;

/**
 * glob(3) clone.
 */
int Glob(const char * pattern, int flags,
		int errfunc(const char * epath, int eerrno),
		void * _pglob);

/**
 * globfree(3) clone.
 */
void Globfree(void * _pglob);

/**
 * realpath(3) clone.
 */
char * Realpath(const char * path, char * resolved_path);

/**
 * lseek(2) clone.
 * @todo Implement SEEK_HOLE/SEEK_DATA.
 */
off_t Lseek(int fdno, off_t offset, int whence);

/*@}*/


/** \ingroup rpmio
 * \name RPMIO Utilities.
 */
/*@{*/

/**
 */
off_t   fdSize(FD_t fd);

/**
 */
FD_t fdDup(int fdno);

/**
 */
ssize_t fdRead(void * cookie, char * buf, size_t count);
#define	fdRead(_fd, _buf, _count)	fdio->read((_fd), (_buf), (_count))

/**
 */
ssize_t	fdWrite(void * cookie, const char * buf, size_t count);
#define	fdWrite(_fd, _buf, _count)	fdio->write((_fd), (_buf), (_count))

/**
 */
int fdClose(void * cookie);
#define	fdClose(_fd)		fdio->close(_fd)

/**
 */
FD_t fdOpen(const char *path, int flags, mode_t mode);
#define	fdOpen(_path, _flags, _mode)	fdio->_open((_path), (_flags), (_mode))

/**
 */
FD_t fdLink (void * cookie, const char * msg);
#define	fdLink(_fd, _msg)	\
	((FD_t)rpmioLinkPoolItem((rpmioItem)(_fd), _msg, __FILE__, __LINE__))

/**
 */
FD_t fdFree(FD_t fd, const char * msg);
#define	fdFree(_fd, _msg)	\
	((FD_t)rpmioFreePoolItem((rpmioItem)(_fd), _msg, __FILE__, __LINE__))

/**
 */
FD_t fdNew (const char * msg);
FD_t XfdNew (const char * msg, const char * fn, unsigned ln);
#define	fdNew(_msg)		XfdNew(_msg, __FILE__, __LINE__)

/**
 */
int fdWritable(FD_t fd, int secs);

/**
 */
int fdReadable(FD_t fd, int secs);

/**
 * Insure that directories in path exist, creating as needed.
 * @param path		directory path
 * @param mode		directory mode (if created)
 * @param uid		directory uid (if created), or -1 to skip
 * @param gid		directory uid (if created), or -1 to skip
 * @return		0 on success, errno (or -1) on error
 */
int rpmioMkpath(const char * path, mode_t mode, uid_t uid, gid_t gid);

/**
 * Check FN access, expanding relative paths and twiddles.
 * @param FN		file path to check
 * @param path		colon separated search path (NULL uses $PATH)
 * @param mode		type of access(2) to check (0 uses X_OK)
 * @return		0 if accessible
 */
int rpmioAccess(const char *FN, const char * path, int mode);

extern char * (*Getpass) (const char * prompt);

/**
 * Return a password.
 * @param prompt	prompt string
 * @return		password
 */
char * _GetPass (const char * prompt);

/**
 * Return a password.
 * @param prompt	prompt string
 * @return		password
 */
char * _RequestPass (const char * prompt);

/**
 * FTP and HTTP error codes.
 */
typedef enum ftperrCode_e {
    FTPERR_NE_OK		=  0,
    FTPERR_NE_ERROR		= -1,	/*!< Generic error. */
    FTPERR_NE_LOOKUP		= -2,	/*!< Hostname lookup failed. */
    FTPERR_NE_AUTH		= -3,	/*!< Server authentication failed. */
    FTPERR_NE_PROXYAUTH		= -4,	/*!< Proxy authentication failed. */
    FTPERR_NE_CONNECT		= -5,	/*!< Could not connect to server. */
    FTPERR_NE_TIMEOUT		= -6,	/*!< Connection timed out. */
    FTPERR_NE_FAILED		= -7,	/*!< The precondition failed. */
    FTPERR_NE_RETRY		= -8,	/*!< Retry request. */
    FTPERR_NE_REDIRECT		= -9,	/*!< Redirect received. */

    FTPERR_BAD_SERVER_RESPONSE	= -81,	/*!< Bad server response */
    FTPERR_SERVER_IO_ERROR	= -82,	/*!< Server I/O error */
    FTPERR_SERVER_TIMEOUT	= -83,	/*!< Server timeout */
    FTPERR_BAD_HOST_ADDR	= -84,	/*!< Unable to lookup server host address */
    FTPERR_BAD_HOSTNAME		= -85,	/*!< Unable to lookup server host name */
    FTPERR_FAILED_CONNECT	= -86,	/*!< Failed to connect to server */
    FTPERR_FILE_IO_ERROR	= -87,	/*!< Failed to establish data connection to server */
    FTPERR_PASSIVE_ERROR	= -88,	/*!< I/O error to local file */
    FTPERR_FAILED_DATA_CONNECT	= -89,	/*!< Error setting remote server to passive mode */
    FTPERR_FILE_NOT_FOUND	= -90,	/*!< File not found on server */
    FTPERR_NIC_ABORT_IN_PROGRESS= -91,	/*!< Abort in progress */
    FTPERR_UNKNOWN		= -100	/*!< Unknown or unexpected error */
} ftperrCode;

/**
 */
const char * ftpStrerror(int errorNumber);

/**
 */
void * ufdGetUrlinfo(FD_t fd);

/**
 */
const char * urlStrerror(const char * url);

/**
 */
int ufdCopy(FD_t sfd, FD_t tfd);

/**
 */
int ufdGetFile(FD_t sfd, FD_t tfd);

/**
 */
extern FDIO_t fdio;

/**
 */
extern FDIO_t fpio;

/**
 */
extern FDIO_t ufdio;

/**
 */
extern FDIO_t gzdio;

/**
 */
extern FDIO_t bzdio;

/**
 */
extern FDIO_t lzdio;

/**
 */
extern FDIO_t xzdio;

/**
 */
extern FDIO_t zstdio;

/**
 */
extern FDIO_t msqio;

/*@}*/

extern rpmioPool _fdPool;

/**
 * Free all memory allocated by rpmio usage.
 */
void rpmioClean(void);

#define	RPMIOPOOL_MODULE(XY)			\
rpmioPool _rpm##XY##Pool;			\
static rpm##XY rpm##XY##GetPool(rpmioPool pool)	\
{   rpm##XY XY; 				\
    if (_rpm##XY##Pool == NULL) {		\
	_rpm##XY##Pool = 			\
		rpmioNewPool(#XY, sizeof(*XY), -1, \
			_rpm##XY##_debug,	\
			rpm##XY##Dbug,		\
			rpm##XY##Init,		\
			rpm##XY##Fini);		\
	pool = _rpm##XY##Pool;			\
    }						\
    XY = (rpm##XY) rpmioGetPool(pool, sizeof(*XY)); \
    memset(((char *)XY)+sizeof(XY->_item), 0, sizeof(*XY)-sizeof(XY->_item));\
    return XY;					\
}

#define	RPMIOPOOL_INTERP_MODULE(XY)		\
RPMIOPOOL_MODULE(XY)				\
rpm##XY _rpm##XY##I;				\
static rpm##XY rpm##XY##I(void)			\
{						\
    if (_rpm##XY##I == NULL)			\
	_rpm##XY##I = rpm##XY##New(NULL, 0);	\
    return _rpm##XY##I;				\
}

#define	RPMIOPOOL_FREE(XY)			\
{   extern rpmioPool _rpm##XY##Pool;		\
    _rpm##XY##Pool = rpmioFreePool(_rpm##XY##Pool); \
}

#define	RPMIOPOOL_INTERP_FREE(XY)		\
{   extern rpm##XY _rpm##XY##I;			\
    extern rpmioPool _rpm##XY##Pool;		\
    _rpm##XY##I = rpm##XY##Free(_rpm##XY##I);	\
    _rpm##XY##Pool = rpmioFreePool(_rpm##XY##Pool); \
}

/**
 * Reclaim memory pool items.
 * @param pool		memory pool (NULL uses global rpmio pool)
 * @return		NULL always
 */
rpmioPool rpmioFreePool(rpmioPool pool);

/**
 * Create a memory pool.
 * @param name		pool name
 * @param size		item size
 * @param limit		no. of items permitted (-1 for unlimited)
 * @param flags		debugging flags
 * @param dbg		(*dbg)() generate string for Unlink/Link/Free debugging
 * @param init		(*init)() create item contents
 * @param fini		(*fini)() destroy item contents
 * @return		memory pool
 */
rpmioPool rpmioNewPool(const char * name,
		size_t size, int limit, int flags,
		char * (*dbg) (void *item, char *b, size_t nb),
		void (*init) (void *item),
		void (*fini) (void *item));

/**
 * Decrement a pool item refcount.
 * @param item		pool item
 * @param msg		debugging msg (NULL disables debugging)
 * @param fn		usually __FILE__
 * @param ln		usually __LINE__
 * @return		pool item (NULL on last dereference)
 */
rpmioItem rpmioUnlinkPoolItem(rpmioItem item,
		const char * msg, const char * fn, unsigned ln);

/**
 * Increment a pool item refcount.
 * @param item		pool item
 * @param msg		debugging msg (NULL disables debugging)
 * @param fn		usually __FILE__
 * @param ln		usually __LINE__
 * @return		pool item
 */
rpmioItem rpmioLinkPoolItem(rpmioItem item,
		const char * msg, const char * fn, unsigned ln);

/**
 * Free a pool item.
 * @param item		pool item
 * @param msg		debugging msg (NULL disables debugging)
 * @param fn		usually __FILE__
 * @param ln		usually __LINE__
 * @return		pool item (NULL on last dereference)
 */
void * rpmioFreePoolItem(rpmioItem item,
		const char * msg, const char * fn, unsigned ln);

/**
 * Get unused item from pool, or alloc a new item.
 * @param pool		memory pool (NULL will always alloc a new item)
 * @param size		item size
 * @return		new item
 */
rpmioItem rpmioGetPool(rpmioPool pool, size_t size);

/**
 * Put unused item into pool (or free).
 * @param item		unused item
 * @return		NULL always
 */
rpmioItem rpmioPutPool(rpmioItem item);

int rpmioMcheck(void (*abortfunc)(int));
int rpmioMcheckPedantic(void (*abortfunc)(int));
void rpmioMcheckCheckAll(void);
int rpmioMprobe(void *ptr);

/**
 * Print/set malloc options.
 * @param		param (0 will print all parameters on stderr)
 * @val			value
 * @return		0 on success
 */
int rpmioMallopt(int param, int val);

/**
 * Log malloc info.
 * @param msg		message
 */
void rpmioMallocInfo(const char *msg)
	RPM_GNUC_CONST;

/**
 * Print malloc stats on stderr.
 */
void rpmioMallocStats(void);

int rpmioMallocTrim(size_t pad);
size_t rpmioMallocUsableSize(void *ptr);
void * rpmioMallocGetState(void);
int rpmioMallocSetState(void *ptr);

/**
 * Trace/untrace malloc (only if MALLOC_TRACE=mtrace-output-file).
 */
void rpmioMtrace(void);
void rpmioMuntrace(void);

#ifdef __cplusplus
}
#endif

#endif	/* H_RPMIO */
