// following code might be needed in some cases
// #define __STDC_LIMIT_MACROS
// #include <stdint.h>

#pragma GCC diagnostic ignored "-Winvalid-offsetof"

#include "system.h"

#include "jsapi.h"
#include "jsprf.h"

#include "jsfriendapi.h"

#include "jswrapper.h"

using namespace JS;

#define	_RPMJSS_INTERNAL
#include <rpmjss.h>

#include "debug.h"

/* The class of the global object. */
static JSClass global_class = {
    "global",
    JSCLASS_GLOBAL_FLAGS,
    JS_PropertyStub,
    JS_PropertyStub,
    JS_PropertyStub,
    JS_StrictPropertyStub,
    JS_EnumerateStub,
    JS_ResolveStub,
    JS_ConvertStub,
};

static int rpmjss_nopens;

int _rpmjss_debug;
#define SPEW(_fmt, ...) \
    if (_rpmjss_debug) \
	fprintf(stderr, _fmt, __VA_ARGS__)

/*==============================================================*/
typedef struct JSI_s * JSI_t;
struct JSI_s {
    JSRuntime	*rt;
    JSContext	*cx;
    JSObject	*global;
};

static void
rpmjssReportError(JSContext *cx, const char *message, JSErrorReport *report)
{
    fprintf(stderr, "%s:%u:%s\n",
	report->filename ? report->filename : "<no filename=\"filename\">",
	(unsigned int) report->lineno, message);
}

/*==============================================================*/
typedef enum JSShellExitCode {
    EXITCODE_RUNTIME_ERROR	= 3,
    EXITCODE_FILE_NOT_FOUND	= 4,
    EXITCODE_OUT_OF_MEMORY	= 5,
    EXITCODE_TIMEOUT		= 6
} JSShellExitCode;

static JSRuntime *gRuntime = NULL;

int gExitCode = 0;
bool gQuitting = false;
bool gGotError = false;

static bool reportWarnings = true;
static bool compileOnly = false;

/*
 * JSContext option name to flag map. The option names are in alphabetical
 * order for better reporting.
 */
static const struct JSOption {
    const char  *name;
    uint32_t    flag;
} js_options[] = {
    {"atline",          JSOPTION_ATLINE},
    {"methodjit",       JSOPTION_METHODJIT},
    {"methodjit_always",JSOPTION_METHODJIT_ALWAYS},
    {"relimit",         JSOPTION_RELIMIT},
    {"strict",          JSOPTION_STRICT},
    {"typeinfer",       JSOPTION_TYPE_INFERENCE},
    {"werror",          JSOPTION_WERROR},
    {"allow_xml",       JSOPTION_ALLOW_XML},
    {"moar_xml",        JSOPTION_MOAR_XML},
    {"strict_mode",     JSOPTION_STRICT_MODE},
};

static uint32_t
MapContextOptionNameToFlag(JSContext* cx, const char* name)
{
    for (size_t i = 0; i < mozilla::ArrayLength(js_options); ++i) {
        if (strcmp(name, js_options[i].name) == 0)
            return js_options[i].flag;
    }

    char* msg = JS_sprintf_append(NULL,
                                  "unknown option name '%s'."
                                  " The valid names are ", name);
    for (size_t i = 0; i < mozilla::ArrayLength(js_options); ++i) {
        if (!msg)
            break;
        msg = JS_sprintf_append(msg, "%s%s", js_options[i].name,
                                (i + 2 < mozilla::ArrayLength(js_options)
                                 ? ", "
                                 : i + 2 == mozilla::ArrayLength(js_options)
                                 ? " and "
                                 : "."));
    }
    if (!msg) {
        JS_ReportOutOfMemory(cx);
    } else {
        JS_ReportError(cx, msg);
        free(msg);
    }
    return 0;
}

static JSBool
Version(JSContext *cx, unsigned argc, jsval *vp)
{
    jsval *argv = JS_ARGV(cx, vp);
    if (argc == 0 || JSVAL_IS_VOID(argv[0])) {
        /* Get version. */
        *vp = INT_TO_JSVAL(JS_GetVersion(cx));
    } else {
        /* Set version. */
        int32_t v = -1;
        if (JSVAL_IS_INT(argv[0])) {
            v = JSVAL_TO_INT(argv[0]);
        } else if (JSVAL_IS_DOUBLE(argv[0])) {
            double fv = JSVAL_TO_DOUBLE(argv[0]);
            if (int32_t(fv) == fv)
                v = int32_t(fv);
        }
        if (v < 0 || v > JSVERSION_LATEST) {
#ifdef	FIXME
            JS_ReportErrorNumber(cx, my_GetErrorMessage, NULL, JSSMSG_INVALID_ARGS, "version");
#endif
            return false;
        }
        *vp = INT_TO_JSVAL(JS_SetVersion(cx, JSVersion(v)));
    }
    return true;
}

#ifdef	NOTYET
static JSBool
RevertVersion(JSContext *cx, unsigned argc, jsval *vp)
{
    js_RevertVersion(cx);
    JS_SET_RVAL(cx, vp, JSVAL_VOID);
    return true;
}
#endif

static JSBool
Options(JSContext *cx, unsigned argc, jsval *vp)
{
    uint32_t optset, flag;
    JSString *str;
    char *names;
    bool found;

    optset = 0;
    jsval *argv = JS_ARGV(cx, vp);
    for (unsigned i = 0; i < argc; i++) {
        str = JS_ValueToString(cx, argv[i]);
        if (!str)
            return false;
        argv[i] = STRING_TO_JSVAL(str);
        JSAutoByteString opt(cx, str);
        if (!opt)
            return false;
        flag = MapContextOptionNameToFlag(cx, opt.ptr());
        if (!flag)
            return false;
        optset |= flag;
    }
    optset = JS_ToggleOptions(cx, optset);

    names = NULL;
    found = false;
    for (size_t i = 0; i < mozilla::ArrayLength(js_options); i++) {
        if (js_options[i].flag & optset) {
            found = true;
            names = JS_sprintf_append(names, "%s%s",
                                      names ? "," : "", js_options[i].name);
            if (!names)
                break;
        }
    }
    if (!found)
        names = strdup("");
    if (!names) {
        JS_ReportOutOfMemory(cx);
        return false;
    }
    str = JS_NewStringCopyZ(cx, names);
    free(names);
    if (!str)
        return false;
    *vp = STRING_TO_JSVAL(str);
    return true;
}

#ifdef	NOTYET
static JSBool
Load(JSContext *cx, unsigned argc, jsval *vp)
{
    RootedObject thisobj(cx, JS_THIS_OBJECT(cx, vp));
    if (!thisobj)
        return false;

    jsval *argv = JS_ARGV(cx, vp);
    for (unsigned i = 0; i < argc; i++) {
        JSString *str = JS_ValueToString(cx, argv[i]);
        if (!str)
            return false;
        argv[i] = STRING_TO_JSVAL(str);
        JSAutoByteString filename(cx, str);
        if (!filename)
            return false;
        errno = 0;
        CompileOptions opts(cx);
        opts.setCompileAndGo(true).setNoScriptRval(true);
        if ((compileOnly && !Compile(cx, thisobj, opts, filename.ptr())) ||
            !Evaluate(cx, thisobj, opts, filename.ptr(), NULL))
        {
            return false;
        }
    }

    JS_SET_RVAL(cx, vp, JSVAL_VOID);
    return true;
}
#endif

#ifdef	NOTYET
class AutoNewContext
{
  private:
    JSContext *oldcx;
    JSContext *newcx;
    Maybe<JSAutoSuspendRequest> suspension;
    Maybe<JSAutoRequest> newRequest;

    AutoNewContext(const AutoNewContext &) MOZ_DELETE;

  public:
    AutoNewContext() : oldcx(NULL), newcx(NULL) {}

    bool enter(JSContext *cx) {
        JS_ASSERT(!JS_IsExceptionPending(cx));
        oldcx = cx;
        newcx = NewContext(JS_GetRuntime(cx));
        if (!newcx)
            return false;
        JS_SetOptions(newcx, JS_GetOptions(newcx) | JSOPTION_DONT_REPORT_UNCAUGHT);
        JS_SetGlobalObject(newcx, JS_GetGlobalForScopeChain(cx));

        suspension.construct(cx);
        newRequest.construct(newcx);
        return true;
    }

    JSContext *get() { return newcx; }

    ~AutoNewContext() {
        if (newcx) {
            RootedValue exc(oldcx);
            bool throwing = JS_IsExceptionPending(newcx);
            if (throwing)
                JS_GetPendingException(newcx, exc.address());
            newRequest.destroy();
            suspension.destroy();
            if (throwing)
                JS_SetPendingException(oldcx, exc);
            JS_DestroyContextNoGC(newcx);
        }
    }
};
#endif

#ifdef	NOTYET
static JSBool
Evaluate(JSContext *cx, unsigned argc, jsval *vp)
{
    CallArgs args = CallArgsFromVp(argc, vp);

    if (argc < 1 || argc > 2) {
#ifdef	FIXME
        JS_ReportErrorNumber(cx, my_GetErrorMessage, NULL,
                             argc < 1 ? JSSMSG_NOT_ENOUGH_ARGS : JSSMSG_TOO_MANY_ARGS,
                             "evaluate");
#endif
        return false;
    }
    if (!args[0].isString() || (argc == 2 && args[1].isPrimitive())) {
#ifdef	FIXME
        JS_ReportErrorNumber(cx, my_GetErrorMessage, NULL, JSSMSG_INVALID_ARGS, "evaluate");
#endif
        return false;
    }

    bool newContext = false;
    bool compileAndGo = true;
    bool noScriptRval = false;
    const char *fileName = "@evaluate";
    JSAutoByteString fileNameBytes;
    jschar *sourceMapURL = NULL;
    unsigned lineNumber = 1;
    RootedObject global(cx, NULL);

    global = JS_GetGlobalForObject(cx, &args.callee());
    if (!global)
        return false;

    if (argc == 2) {
        RootedObject options(cx, &args[1].toObject());
        jsval v;

        if (!JS_GetProperty(cx, options, "newContext", &v))
            return false;
        if (!JSVAL_IS_VOID(v)) {
            JSBool b;
            if (!JS_ValueToBoolean(cx, v, &b))
                return false;
            newContext = b;
        }

        if (!JS_GetProperty(cx, options, "compileAndGo", &v))
            return false;
        if (!JSVAL_IS_VOID(v)) {
            JSBool b;
            if (!JS_ValueToBoolean(cx, v, &b))
                return false;
            compileAndGo = b;
        }

        if (!JS_GetProperty(cx, options, "noScriptRval", &v))
            return false;
        if (!JSVAL_IS_VOID(v)) {
            JSBool b;
            if (!JS_ValueToBoolean(cx, v, &b))
                return false;
            noScriptRval = b;
        }

        if (!JS_GetProperty(cx, options, "fileName", &v))
            return false;
        if (JSVAL_IS_NULL(v)) {
            fileName = NULL;
        } else if (!JSVAL_IS_VOID(v)) {
            JSString *s = JS_ValueToString(cx, v);
            if (!s)
                return false;
            fileName = fileNameBytes.encode(cx, s);
            if (!fileName)
                return false;
        }

        if (!JS_GetProperty(cx, options, "sourceMapURL", &v))
            return false;
        if (!JSVAL_IS_VOID(v)) {
            JSString *s = JS_ValueToString(cx, v);
            if (!s)
                return false;
            const jschar* smurl = s->getCharsZ(cx);
            if (!smurl)
                return false;
            sourceMapURL = js_strdup(cx, smurl);
        }

        if (!JS_GetProperty(cx, options, "lineNumber", &v))
            return false;
        if (!JSVAL_IS_VOID(v)) {
            uint32_t u;
            if (!JS_ValueToECMAUint32(cx, v, &u))
                return false;
            lineNumber = u;
        }

        if (!JS_GetProperty(cx, options, "global", &v))
            return false;
        if (!JSVAL_IS_VOID(v)) {
            global = JSVAL_IS_PRIMITIVE(v) ? NULL : JSVAL_TO_OBJECT(v);
            if (global) {
                global = JS_UnwrapObject(global);
                if (!global)
                    return false;
            }
            if (!global || !(JS_GetClass(global)->flags & JSCLASS_IS_GLOBAL)) {
#ifdef	FIXME
                JS_ReportErrorNumber(cx, js_GetErrorMessage, NULL, JSMSG_UNEXPECTED_TYPE,
                                     "\"global\" passed to evaluate()", "not a global object");
#endif
                return false;
            }
        }
    }

    RootedString code(cx, args[0].toString());

    size_t codeLength;
    const jschar *codeChars = JS_GetStringCharsAndLength(cx, code, &codeLength);
    if (!codeChars)
        return false;

    AutoNewContext ancx;
    if (newContext) {
        if (!ancx.enter(cx))
            return false;
        cx = ancx.get();
    }

    {
        JSAutoCompartment ac(cx, global);
        uint32_t saved = JS_GetOptions(cx);
        uint32_t options = saved & ~(JSOPTION_COMPILE_N_GO | JSOPTION_NO_SCRIPT_RVAL);
        if (compileAndGo)
            options |= JSOPTION_COMPILE_N_GO;
        if (noScriptRval)
            options |= JSOPTION_NO_SCRIPT_RVAL;
        JS_SetOptions(cx, options);

        JSScript *script = JS_CompileUCScript(cx, global, codeChars, codeLength, fileName, lineNumber);
        if (!script)
            return false;
        if (sourceMapURL && !script->scriptSource()->hasSourceMap()) {
            if (!script->scriptSource()->setSourceMap(cx, sourceMapURL, script->filename))
                return false;
        }
        JS_SetOptions(cx, saved);
        if (!JS_ExecuteScript(cx, global, script, vp))
            return false;
    }

    return JS_WrapValue(cx, vp);
}
#endif

#ifdef	NOTYET
static JSString *
FileAsString(JSContext *cx, const char *pathname)
{
    FILE *file;
    JSString *str = NULL;
    size_t len, cc;
    char *buf;

    file = fopen(pathname, "rb");
    if (!file) {
        JS_ReportError(cx, "can't open %s: %s", pathname, strerror(errno));
        return NULL;
    }

    if (fseek(file, 0, SEEK_END) != 0) {
        JS_ReportError(cx, "can't seek end of %s", pathname);
    } else {
        len = ftell(file);
        if (fseek(file, 0, SEEK_SET) != 0) {
            JS_ReportError(cx, "can't seek start of %s", pathname);
        } else {
            buf = (char*) JS_malloc(cx, len + 1);
            if (buf) {
                cc = fread(buf, 1, len, file);
                if (cc != len) {
                    JS_ReportError(cx, "can't read %s: %s", pathname,
                                   (ptrdiff_t(cc) < 0) ? strerror(errno) : "short read");
                } else {
                    jschar *ucbuf;
                    size_t uclen;

                    len = (size_t)cc;

                    if (!JS_DecodeUTF8(cx, buf, len, NULL, &uclen)) {
                        JS_ReportError(cx, "Invalid UTF-8 in file '%s'", pathname);
                        gExitCode = EXITCODE_RUNTIME_ERROR;
                        return NULL;
                    }

                    ucbuf = (jschar*)malloc(uclen * sizeof(jschar));
                    JS_DecodeUTF8(cx, buf, len, ucbuf, &uclen);
                    str = JS_NewUCStringCopyN(cx, ucbuf, uclen);
                    free(ucbuf);
                }
                JS_free(cx, buf);
            }
        }
    }
    fclose(file);

    return str;
}
#endif

#ifdef	NOTYET
static JSObject *
FileAsTypedArray(JSContext *cx, const char *pathname)
{
    FILE *file = fopen(pathname, "rb");
    if (!file) {
        JS_ReportError(cx, "can't open %s: %s", pathname, strerror(errno));
        return NULL;
    }

    JSObject *obj = NULL;
    if (fseek(file, 0, SEEK_END) != 0) {
        JS_ReportError(cx, "can't seek end of %s", pathname);
    } else {
        size_t len = ftell(file);
        if (fseek(file, 0, SEEK_SET) != 0) {
            JS_ReportError(cx, "can't seek start of %s", pathname);
        } else {
            obj = JS_NewUint8Array(cx, len);
            if (!obj)
                return NULL;
            char *buf = (char *) TypedArray::viewData(obj);
            size_t cc = fread(buf, 1, len, file);
            if (cc != len) {
                JS_ReportError(cx, "can't read %s: %s", pathname,
                               (ptrdiff_t(cc) < 0) ? strerror(errno) : "short read");
                obj = NULL;
            }
        }
    }
    fclose(file);
    return obj;
}
#endif

/*
 * Function to run scripts and return compilation + execution time. Semantics
 * are closely modelled after the equivalent function in WebKit, as this is used
 * to produce benchmark timings by SunSpider.
 */
#ifdef	NOTYET
static JSBool
Run(JSContext *cx, unsigned argc, jsval *vp)
{
    if (argc != 1) {
#ifdef	FIXME
        JS_ReportErrorNumber(cx, my_GetErrorMessage, NULL, JSSMSG_INVALID_ARGS, "run");
#endif
        return false;
    }

    RootedObject thisobj(cx, JS_THIS_OBJECT(cx, vp));
    if (!thisobj)
        return false;

    jsval *argv = JS_ARGV(cx, vp);
    JSString *str = JS_ValueToString(cx, argv[0]);
    if (!str)
        return false;
    argv[0] = STRING_TO_JSVAL(str);
    JSAutoByteString filename(cx, str);
    if (!filename)
        return false;

    const jschar *ucbuf = NULL;
    size_t buflen;
    str = FileAsString(cx, filename.ptr());
    if (str)
        ucbuf = JS_GetStringCharsAndLength(cx, str, &buflen);
    if (!ucbuf)
        return false;

    JS::Anchor<JSString *> a_str(str);
    uint32_t oldopts = JS_GetOptions(cx);
    JS_SetOptions(cx, oldopts | JSOPTION_COMPILE_N_GO | JSOPTION_NO_SCRIPT_RVAL);

    int64_t startClock = PRMJ_Now();
    JSScript *script = JS_CompileUCScript(cx, thisobj, ucbuf, buflen, filename.ptr(), 1);
    JS_SetOptions(cx, oldopts);
    if (!script || !JS_ExecuteScript(cx, thisobj, script, NULL))
        return false;

    int64_t endClock = PRMJ_Now();
    JS_SET_RVAL(cx, vp, DOUBLE_TO_JSVAL((endClock - startClock) / double(PRMJ_USEC_PER_MSEC)));
    return true;
}
#endif

/*
 * function readline()
 * Provides a hook for scripts to read a line from stdin.
 */
#ifdef	NOTYET
static JSBool
ReadLine(JSContext *cx, unsigned argc, jsval *vp)
{
#define BUFSIZE 256
    FILE *from;
    char *buf, *tmp;
    size_t bufsize, buflength, gotlength;
    bool sawNewline;
    JSString *str;

    from = stdin;
    buflength = 0;
    bufsize = BUFSIZE;
    buf = (char *) JS_malloc(cx, bufsize);
    if (!buf)
        return false;

    sawNewline = false;
    while ((gotlength =
            js_fgets(buf + buflength, bufsize - buflength, from)) > 0) {
        buflength += gotlength;

        /* Are we done? */
        if (buf[buflength - 1] == '\n') {
            buf[buflength - 1] = '\0';
            sawNewline = true;
            break;
        } else if (buflength < bufsize - 1) {
            break;
        }

        /* Else, grow our buffer for another pass. */
        bufsize *= 2;
        if (bufsize > buflength) {
            tmp = (char *) JS_realloc(cx, buf, bufsize);
        } else {
            JS_ReportOutOfMemory(cx);
            tmp = NULL;
        }

        if (!tmp) {
            JS_free(cx, buf);
            return false;
        }

        buf = tmp;
    }

    /* Treat the empty string specially. */
    if (buflength == 0) {
        *vp = feof(from) ? JSVAL_NULL : JS_GetEmptyStringValue(cx);
        JS_free(cx, buf);
        return true;
    }

    /* Shrink the buffer to the real size. */
    tmp = (char *) JS_realloc(cx, buf, buflength);
    if (!tmp) {
        JS_free(cx, buf);
        return false;
    }

    buf = tmp;

    /*
     * Turn buf into a JSString. Note that buflength includes the trailing null
     * character.
     */
    str = JS_NewStringCopyN(cx, buf, sawNewline ? buflength - 1 : buflength);
    JS_free(cx, buf);
    if (!str)
        return false;

    *vp = STRING_TO_JSVAL(str);
    return true;
}
#endif

static JSBool
PutStr(JSContext *cx, unsigned argc, jsval *vp)
{
    FILE *gOutFile = stdout;
    jsval *argv;
    JSString *str;
    char *bytes;

    if (argc != 0) {
        argv = JS_ARGV(cx, vp);
        str = JS_ValueToString(cx, argv[0]);
        if (!str)
            return false;
        bytes = JS_EncodeString(cx, str);
        if (!bytes)
            return false;
        fputs(bytes, gOutFile);
        JS_free(cx, bytes);
        fflush(gOutFile);
    }

    JS_SET_RVAL(cx, vp, JSVAL_VOID);
    return true;
}

static JSBool
Now(JSContext *cx, unsigned argc, jsval *vp)
{
#ifdef	FIXME
    double now = PRMJ_Now() / double(PRMJ_USEC_PER_MSEC);
#else
    struct timeval tv;
    (void) gettimeofday(&tv, NULL);
    double now = (tv.tv_sec * 1000 * 1000 + tv.tv_usec)/1000.;
#endif
    JS_SET_RVAL(cx, vp, DOUBLE_TO_JSVAL(now));
    return true;
}

static JSBool
PrintInternal(JSContext *cx, unsigned argc, jsval *vp, FILE *file)
{
    jsval *argv;
    unsigned i;
    JSString *str;
    char *bytes;

    argv = JS_ARGV(cx, vp);
    for (i = 0; i < argc; i++) {
        str = JS_ValueToString(cx, argv[i]);
        if (!str)
            return false;
        bytes = JS_EncodeString(cx, str);
        if (!bytes)
            return false;
        fprintf(file, "%s%s", i ? " " : "", bytes);
        JS_free(cx, bytes);
    }

    fputc('\n', file);
    fflush(file);

    JS_SET_RVAL(cx, vp, JSVAL_VOID);
    return true;
}

static JSBool
Print(JSContext *cx, unsigned argc, jsval *vp)
{
    FILE *gOutFile = stdout;
    return PrintInternal(cx, argc, vp, gOutFile);
}

static JSBool
PrintErr(JSContext *cx, unsigned argc, jsval *vp)
{
    FILE *gErrFile = stderr;
    return PrintInternal(cx, argc, vp, gErrFile);
}

#ifdef	NOTYET
static JSBool
Help(JSContext *cx, unsigned argc, jsval *vp);
#endif

static JSBool
Quit(JSContext *cx, unsigned argc, jsval *vp)
{
    JS_ConvertArguments(cx, argc, JS_ARGV(cx, vp), "/ i", &gExitCode);

    gQuitting = true;
    return false;
}

static JSFunctionSpec global_functions[] = {
    JS_FN("version",		Version,	0,0),
#ifdef	NOTYET
    JS_FN("revertVersion",	RevertVersion,	0,0),
#endif	/* NOTYET */
    JS_FN("options",		Options,	0,0),
#ifdef	NOTYET
    JS_FN("load",		Load,		1,0),
    JS_FN("evaluate",		Evaluate,	2,0),
    JS_FN("run",		Run,		1,0),
    JS_FN("readline",		ReadLine,	0,0),
#endif	/* NOTYET */
    JS_FN("print",		Print,		0,0),
    JS_FN("printErr",		PrintErr,	0,0),
    JS_FN("putstr",		PutStr,		0,0),
    JS_FN("dateNow",		Now,		0,0),
#ifdef	NOTYET
    JS_FN("help",		Help,		0,0),
#endif	/* NOTYET */
    JS_FN("quit",		Quit,		0,0),
#ifdef	NOTYET
    JS_FS("require",		Require,	0,0),
#endif	/* NOTYET */
    JS_FS_END
};

/*==============================================================*/
static void mozFini(rpmjss jss)
{
    JSI_t I = (JSI_t) jss->I;

    if (I->cx)
	JS_DestroyContext(I->cx);
    I->cx = NULL;

    if (I->rt)
	JS_DestroyRuntime(I->rt);
    I->rt = NULL;

    if (--rpmjss_nopens <= 0)  {
	JS_ShutDown();
	rpmjss_nopens = 0;
    }
    free(I);
}

static void * mozInit(rpmjss jss)
{
    JSI_t I = NULL;
    uint32_t flags = jss->flags;

    static uint32_t _maxbytes = 8L * 1024L * 1024L;
    static size_t _stackChunkSize = 8192;
    JSClass * _clasp = (JSClass *)&global_class;
    JSFunctionSpec * _functions = global_functions;
    JSPrincipals * _principals = NULL;
    JSBool ok;

    if (rpmjss_nopens++ == 0) {
#ifdef	NOTYET		/* XXX collides with ancient JS_NewRuntime */
	JS_Init();
#endif
    }

    I = (JSI_t) calloc(1, sizeof(*I));
assert(I);

    I->rt = JS_NewRuntime(_maxbytes);
assert(I->rt);
    JS_SetRuntimePrivate(I->rt, (void *)jss);

    I->cx = JS_NewContext(I->rt, _stackChunkSize);
assert(I->cx);
    JS_SetContextPrivate(I->cx, (void *)jss);

    JS_SetOptions(I->cx,
		JSOPTION_VAROBJFIX | JSOPTION_METHODJIT);
    JS_SetVersion(I->cx, JSVERSION_LATEST);
    JS_SetErrorReporter(I->cx, rpmjssReportError);

    JS::RootedObject glob(I->cx,
		JS_NewGlobalObject(I->cx, _clasp, _principals));
    I->global = glob;
assert(I->global);
    JS_SetGlobalObject(I->cx, I->global);

    ok = JS_InitStandardClasses(I->cx, I->global);
assert(ok);
#ifdef  JS_HAS_CTYPES
    ok = JS_InitCTypesClass(I->cx, I->global);
assert(ok);
#endif

    ok = JS_DefineFunctions(I->cx, I->global, _functions);
assert(ok);

    return I;
}

static int mozRun(rpmjss jss, const char ** resultp,
		const char * script, const char * fn, unsigned ln)
{
    JSI_t I = (JSI_t) jss->I;
    JSContext* cx = I->cx;
    JSObject* global = I->global;

    JS::RootedObject glob(cx, global);
    JS::RootedValue rval(cx);
    {	// Scope for JSAutoCompartment
	JSAutoCompartment ac(cx, glob);

	JSBool ok = JS_EvaluateScript(cx, glob,
			script, strlen(script), fn, ln,
			rval.address());
	if (!ok)
	    return 1;
    }

    char b[128];
    size_t nb = sizeof(b);
    char * t = NULL;

    if (rval.isNull()) {
	t = xstrdup("");
    } else
    if (rval.isUndefined()) {
	t = xstrdup("");
    } else
    if (rval.isBoolean()) {
	snprintf(b, nb, "%c", (rval.toBoolean() ? 'T' : 'F'));
	t = xstrdup(b);
    } else
    if (rval.isInt32()) {
	snprintf(b, nb, "%d", rval.toInt32());
	t = xstrdup(b);
    } else
    if (rval.isDouble()) {
	snprintf(b, nb, "%.16g", rval.toDouble());
	t = xstrdup(b);
    } else
    if (rval.isNumber()) {
	snprintf(b, nb, "%.16g", rval.toNumber());
	t = xstrdup(b);
    } else
    if (rval.isString()) {
    	t = JS_EncodeString(cx, rval.toString());
    } else
	t = xstrdup("");

    if (resultp)
	*resultp = t;
    else
	t = _free(t);

    return 0;
}

static struct JSIO_s _mozjs = { mozFini, mozInit, mozRun };
JSIO_t mozjs17 = &_mozjs;

#if defined(RPMJSS_SELF_TEST)
#include <rpmio.h>
#include <argv.h>

struct rpmjss_s _jss;
rpmjss jss = &_jss;

/*==============================================================*/
int main(int argc, const char *argv[])
{
    int rc = 0;

    JSI_t I = (JSI_t) mozInit(jss);
    jss->I = I;

    const char * _fn = "rpmjss.inp";
    FD_t fd = Fopen(_fn, "r.fpio");
    ARGV_t av = NULL;
    rc = argvFgets(&av, fd, 0);
    rc = Fclose(fd);
    int ac = argvCount(av);

    for (int i = 0; i < ac; i++) {
	const char * script = av[i];
	const char * result = NULL;
	rc = mozRun(jss, &result, script, _fn, i);
	fprintf(stderr, "<== result |%s| from |%s|\n", result, script);
	result = _free(result);
    }

    av = argvFree(av);

    mozFini(jss);

    return rc;
}
#endif	/* RPMJSS_SELF_TEST */
