// following code might be needed in some cases
// #define __STDC_LIMIT_MACROS
// #include <stdint.h>

#pragma GCC diagnostic ignored "-Winvalid-offsetof"

#include "system.h"

#include "jsapi.h"

#define	_RPMJSS_INTERNAL
#include <rpmjss.h>

#include "debug.h"

/* The class of the global object. */
static JSClass global_class = {
    "global",
    JSCLASS_GLOBAL_FLAGS,
    JS_PropertyStub,
    JS_PropertyStub,
    JS_PropertyStub,
    JS_StrictPropertyStub,
    JS_EnumerateStub,
    JS_ResolveStub,
    JS_ConvertStub,
};

static int rpmjss_nopens;

static int _rpmjss185_debug;
#define SPEW(_fmt, ...) \
    if (_rpmjss185_debug) \
	fprintf(stderr, _fmt, __VA_ARGS__)

/*==============================================================*/
typedef struct JSI_s * JSI_t;
struct JSI_s {
    JSRuntime	*rt;
    JSContext	*cx;
    JSObject	*global;
};

static void
rpmjssReportError(JSContext *cx, const char *message, JSErrorReport *report)
{
    fprintf(stderr, "%s:%u:%s\n",
	report->filename ? report->filename : "<no filename=\"filename\">",
	(unsigned int) report->lineno, message);
}

/*==============================================================*/
static JSBool compileOnly = JS_FALSE;

static JSBool
Version(JSContext *cx, uintN argc, jsval *vp)
{
    jsval *argv = JS_ARGV(cx, vp);
    if (argc > 0 && JSVAL_IS_INT(argv[0]))
        *vp = INT_TO_JSVAL(JS_SetVersion(cx, (JSVersion) JSVAL_TO_INT(argv[0])));
    else
        *vp = INT_TO_JSVAL(JS_GetVersion(cx));
    return JS_TRUE;
}

static JSBool
Load(JSContext *cx, uintN argc, jsval *vp)
{
    jsval *argv = JS_ARGV(cx, vp);
    JSObject *thisobj = JS_THIS_OBJECT(cx, vp);
    if (!thisobj)
        return JS_FALSE;

    for (uintN i = 0; i < argc; i++) {
        JSString *str = JS_ValueToString(cx, argv[i]);
        if (!str)
            return false;
        argv[i] = STRING_TO_JSVAL(str);
        JSAutoByteString filename(cx, str);
        if (!filename)
            return JS_FALSE;
        errno = 0;
        uint32 oldopts = JS_GetOptions(cx);
        JS_SetOptions(cx, oldopts | JSOPTION_COMPILE_N_GO | JSOPTION_NO_SCRIPT_RVAL);
        JSObject *scriptObj = JS_CompileFile(cx, thisobj, filename.ptr());
        JS_SetOptions(cx, oldopts);
        if (!scriptObj)
            return false;

        if (!compileOnly && !JS_ExecuteScript(cx, thisobj, scriptObj, NULL))
            return false;
    }

    return true;
}

static JSBool
Evaluate(JSContext *cx, uintN argc, jsval *vp)
{
    if (argc != 1 || !JSVAL_IS_STRING(JS_ARGV(cx, vp)[0])) {
#ifdef	FIXME
        JS_ReportErrorNumber(cx, my_GetErrorMessage, NULL,
                             (argc != 1) ? JSSMSG_NOT_ENOUGH_ARGS : JSSMSG_INVALID_ARGS,
                             "evaluate");
#endif
        return false;
    }

    JSString *code = JSVAL_TO_STRING(JS_ARGV(cx, vp)[0]);

    size_t codeLength;
    const jschar *codeChars = JS_GetStringCharsAndLength(cx, code, &codeLength);
    if (!codeChars)
        return false;

    JSObject *thisobj = JS_THIS_OBJECT(cx, vp);
    if (!thisobj)
        return false;

    if ((JS_GET_CLASS(cx, thisobj)->flags & JSCLASS_IS_GLOBAL) != JSCLASS_IS_GLOBAL) {
#ifdef	FIXME
        JS_ReportErrorNumber(cx, js_GetErrorMessage, NULL, JSMSG_UNEXPECTED_TYPE,
                             "this-value passed to evaluate()", "not a global object");
#endif
        return false;
    }

    JS_SET_RVAL(cx, vp, JSVAL_VOID);
    return JS_EvaluateUCScript(cx, thisobj, codeChars, codeLength, "@evaluate", 0, NULL);
}

static JSString *
FileAsString(JSContext *cx, const char *pathname)
{
    FILE *file;
    JSString *str = NULL;
    size_t len, cc;
    char *buf;

    file = fopen(pathname, "rb");
    if (!file) {
        JS_ReportError(cx, "can't open %s: %s", pathname, strerror(errno));
        return NULL;
    }

    if (fseek(file, 0, SEEK_END) == EOF) {
        JS_ReportError(cx, "can't seek end of %s", pathname);
    } else {
        len = ftell(file);
        if (fseek(file, 0, SEEK_SET) == EOF) {
            JS_ReportError(cx, "can't seek start of %s", pathname);
        } else {
            buf = (char*) JS_malloc(cx, len + 1);
            if (buf) {
                cc = fread(buf, 1, len, file);
                if (cc != len) {
                    JS_ReportError(cx, "can't read %s: %s", pathname,
                                   (ptrdiff_t(cc) < 0) ? strerror(errno) : "short read");
                } else {
                    len = (size_t)cc;
                    str = JS_NewStringCopyN(cx, buf, len);
                }
                JS_free(cx, buf);
            }
        }
    }
    fclose(file);

    return str;
}

/*
 * Function to run scripts and return compilation + execution time. Semantics
 * are closely modelled after the equivalent function in WebKit, as this is used
 * to produce benchmark timings by SunSpider.
 */
static JSBool
Run(JSContext *cx, uintN argc, jsval *vp)
{
    if (argc != 1) {
#ifdef	FIXME
        JS_ReportErrorNumber(cx, my_GetErrorMessage, NULL, JSSMSG_INVALID_ARGS, "run");
#endif
        return false;
    }

    JSObject *thisobj = JS_THIS_OBJECT(cx, vp);
    if (!thisobj)
        return false;

    jsval *argv = JS_ARGV(cx, vp);
    JSString *str = JS_ValueToString(cx, argv[0]);
    if (!str)
        return false;
    argv[0] = STRING_TO_JSVAL(str);
    JSAutoByteString filename(cx, str);
    if (!filename)
        return false;

    const jschar *ucbuf = NULL;
    size_t buflen;
    str = FileAsString(cx, filename.ptr());
    if (str)
        ucbuf = JS_GetStringCharsAndLength(cx, str, &buflen);
    if (!ucbuf)
        return false;

    JS::Anchor<JSString *> a_str(str);
    uint32 oldopts = JS_GetOptions(cx);
    JS_SetOptions(cx, oldopts | JSOPTION_COMPILE_N_GO | JSOPTION_NO_SCRIPT_RVAL);

#ifdef	FIXME
    int64 startClock = PRMJ_Now();
#endif
    JSObject *scriptObj = JS_CompileUCScript(cx, thisobj, ucbuf, buflen, filename.ptr(), 1);
    JS_SetOptions(cx, oldopts);
    if (!scriptObj || !JS_ExecuteScript(cx, thisobj, scriptObj, NULL))
        return false;

#ifdef	FIXME
    int64 endClock = PRMJ_Now();
    JS_SET_RVAL(cx, vp, DOUBLE_TO_JSVAL((endClock - startClock) / double(PRMJ_USEC_PER_MSEC)));
#endif
    return true;
}

#ifdef	NOTYET
/*
 * function readline()
 * Provides a hook for scripts to read a line from stdin.
 */
static JSBool
ReadLine(JSContext *cx, uintN argc, jsval *vp)
{
#define BUFSIZE 256
    FILE *from;
    char *buf, *tmp;
    size_t bufsize, buflength, gotlength;
    JSBool sawNewline;
    JSString *str;

    from = stdin;
    buflength = 0;
    bufsize = BUFSIZE;
    buf = (char *) JS_malloc(cx, bufsize);
    if (!buf)
        return JS_FALSE;

    sawNewline = JS_FALSE;
    while ((gotlength =
            js_fgets(buf + buflength, bufsize - buflength, from)) > 0) {
        buflength += gotlength;

        /* Are we done? */
        if (buf[buflength - 1] == '\n') {
            buf[buflength - 1] = '\0';
            sawNewline = JS_TRUE;
            break;
        } else if (buflength < bufsize - 1) {
            break;
        }

        /* Else, grow our buffer for another pass. */
        bufsize *= 2;
        if (bufsize > buflength) {
            tmp = (char *) JS_realloc(cx, buf, bufsize);
        } else {
            JS_ReportOutOfMemory(cx);
            tmp = NULL;
        }

        if (!tmp) {
            JS_free(cx, buf);
            return JS_FALSE;
        }

        buf = tmp;
    }

    /* Treat the empty string specially. */
    if (buflength == 0) {
        *vp = feof(from) ? JSVAL_NULL : JS_GetEmptyStringValue(cx);
        JS_free(cx, buf);
        return JS_TRUE;
    }

    /* Shrink the buffer to the real size. */
    tmp = (char *) JS_realloc(cx, buf, buflength);
    if (!tmp) {
        JS_free(cx, buf);
        return JS_FALSE;
    }

    buf = tmp;

    /*
     * Turn buf into a JSString. Note that buflength includes the trailing null
     * character.
     */
    str = JS_NewStringCopyN(cx, buf, sawNewline ? buflength - 1 : buflength);
    JS_free(cx, buf);
    if (!str)
        return JS_FALSE;

    *vp = STRING_TO_JSVAL(str);
    return JS_TRUE;
}
#endif

static JSBool
PutStr(JSContext *cx, uintN argc, jsval *vp)
{
    FILE * gOutFile = stdout;
    jsval * argv = JS_ARGV(cx, vp);
    JSString *str;
    char *bytes;

    if (argc != 0) {
        str = JS_ValueToString(cx, argv[0]);
        if (!str)
            return JS_FALSE;
        bytes = JS_EncodeString(cx, str);
        if (!bytes)
            return JS_FALSE;
        fputs(bytes, gOutFile);
        JS_free(cx, bytes);
        fflush(gOutFile);
    }

    JS_SET_RVAL(cx, vp, JSVAL_VOID);
    return JS_TRUE;
}

static JSBool
Now(JSContext *cx, uintN argc, jsval *vp)
{
#ifdef	FIXME
    jsdouble now = PRMJ_Now() / double(PRMJ_USEC_PER_MSEC);
#else
    jsdouble now = time(NULL) * 1000;
#endif
    JS_SET_RVAL(cx, vp, DOUBLE_TO_JSVAL(now));
    return true;
}

static JSBool
Print(JSContext *cx, uintN argc, jsval *vp)
{
    FILE * gOutFile = stdout;
    jsval * argv = JS_ARGV(cx, vp);
    uintN i;
    JSString *str;
    char *bytes;

    for (i = 0; i < argc; i++) {
        str = JS_ValueToString(cx, argv[i]);
        if (!str)
            return JS_FALSE;
        bytes = JS_EncodeString(cx, str);
        if (!bytes)
            return JS_FALSE;
        fprintf(gOutFile, "%s%s", i ? " " : "", bytes);
        JS_free(cx, bytes);
    }

    fputc('\n', gOutFile);
    fflush(gOutFile);

    JS_SET_RVAL(cx, vp, JSVAL_VOID);
    return JS_TRUE;
}

static JSBool
Require(JSContext *cx, uintN argc, jsval *vp)
{
    jsval * argv = JS_ARGV(cx, vp);
SPEW("==> %s(%p,%p[%u],%p)\n", __FUNCTION__, cx, argv, (unsigned)argc, vp);

    JS_SET_RVAL(cx, vp, JSVAL_VOID);
    return JS_TRUE;
}

static JSFunctionSpec global_functions[] = {
    JS_FN("version",	Version,	0,0),
    JS_FN("load",	Load,		1,0),
    JS_FN("evaluate",	Evaluate,	1,0),
    JS_FN("run",	Run,		1,0),
#ifdef	NOTYET
    JS_FN("readline",	ReadLine,	0,0),
#endif
    JS_FN("print",	Print,		0,0),
    JS_FN("putstr",	PutStr,		0,0),
    JS_FN("dateNow",	Now,		0,0),
    JS_FS("require",	Require,	0,0),
    JS_FS_END
};

/*==============================================================*/
static void mozFini(rpmjss jss)
{
    JSI_t I = (JSI_t) jss->I;

    if (I->cx)
	JS_DestroyContext(I->cx);
    I->cx = NULL;

    if (I->rt)
	JS_DestroyRuntime(I->rt);
    I->rt = NULL;

    if (--rpmjss_nopens <= 0)  {
	JS_ShutDown();
	rpmjss_nopens = 0;
    }
    free(I);
}

static void * mozInit(rpmjss jss)
{
    JSI_t I = NULL;
    uint32_t flags = jss->flags;

    static uint32 _maxbytes = 8L * 1024L * 1024L;
    static size_t _stackChunkSize = 8192;
    JSClass * _clasp = (JSClass *)&global_class;
    JSFunctionSpec * _functions = global_functions;
    JSPrincipals * _principals = NULL;
    JSBool ok;

    if (rpmjss_nopens++ == 0) {
#ifdef	NOTYET		/* XXX collides with ancient JS_NewRuntime */
	JS_Init();
#endif
    }

    I = (JSI_t) calloc(1, sizeof(*I));
assert(I);

    I->rt = JS_NewRuntime(_maxbytes);
assert(I->rt);
    JS_SetRuntimePrivate(I->rt, (void *)jss);

    I->cx = JS_NewContext(I->rt, _stackChunkSize);
assert(I->cx);
    JS_SetContextPrivate(I->cx, (void *)jss);

    JS_SetOptions(I->cx,
		JSOPTION_VAROBJFIX | JSOPTION_JIT | JSOPTION_METHODJIT);
    JS_SetVersion(I->cx, JSVERSION_LATEST);
    JS_SetErrorReporter(I->cx, rpmjssReportError);

    I->global = JS_NewCompartmentAndGlobalObject(I->cx,
			_clasp, _principals);
assert(I->global);

    ok = JS_InitStandardClasses(I->cx, I->global);
assert(ok);
#ifdef  JS_HAS_CTYPES
    ok = JS_InitCTypesClass(I->cx, I->global);
assert(ok);
#endif

    ok = JS_DefineFunctions(I->cx, I->global, _functions);
assert(ok);

    return I;
}

static int mozRun(rpmjss jss, const char ** resultp,
		const char * script, const char * fn, unsigned ln)
{
    JSI_t I = (JSI_t) jss->I;
    JSContext* cx = I->cx;
    JSObject* global = I->global;

    jsval rval;
    {	// Scope for JSAutoCompartment

#ifdef	NOTYET
	JSAutoEnterCompartment ac;
        if (!ac.enter(cx, global))
	    return 1;
#endif

	JSBool ok = JS_EvaluateScript(cx, global,
			script, strlen(script), fn, ln,
			&rval);
	if (!ok)
	    return 1;
    }

    char b[128];
    size_t nb = sizeof(b);
    char * t = NULL;

    if (JSVAL_IS_NULL(rval)) {
	t = xstrdup("");
    } else
    if (JSVAL_IS_VOID(rval)) {
	t = xstrdup("");
    } else
    if (JSVAL_IS_BOOLEAN(rval)) {
	snprintf(b, nb, "%c", (JSVAL_TO_BOOLEAN(rval) ? 'T' : 'F'));
	t = xstrdup(b);
    } else
    if (JSVAL_IS_INT(rval)) {	/* int32_t */
	snprintf(b, nb, "%d", JSVAL_TO_INT(rval));
	t = xstrdup(b);
    } else
    if (JSVAL_IS_DOUBLE(rval)) {
	snprintf(b, nb, "%.16g", JSVAL_TO_DOUBLE(rval));
	t = xstrdup(b);
    } else
#ifdef	FIXME
    if (JSVAL_IS_NUMBER(rval)) {
	snprintf(b, nb, "%.16g", JSVAL_TO_NUMBER(rval));
	t = xstrdup(b);
    } else
#endif
    if (JSVAL_IS_STRING(rval)) {
    	t = JS_EncodeString(cx, JSVAL_TO_STRING(rval));
    } else
	t = xstrdup("");

    if (resultp)
	*resultp = t;
    else
	t = _free(t);

    return 0;
}

static struct JSIO_s _mozjs185 = { mozFini, mozInit, mozRun };
JSIO_t mozjs185 = &_mozjs185;

#if defined(RPMJSS_SELF_TEST)
#include <rpmio.h>
#include <argv.h>

struct rpmjss_s _jss;
rpmjss jss = &_jss;

/*==============================================================*/
int main(int argc, const char *argv[])
{
    int rc = 0;

    JSI_t I = (JSI_t) mozInit(jss);
    jss->I = I;

    const char * _fn = "rpmjss.inp";
    FD_t fd = Fopen(_fn, "r.fpio");
    ARGV_t av = NULL;
    rc = argvFgets(&av, fd, 0);
    rc = Fclose(fd);
    int ac = argvCount(av);

    for (int i = 0; i < ac; i++) {
	const char *script = av[i];
	const char * result = NULL;
	rc = mozRun(jss, &result, script, _fn, i);
	fprintf(stderr, "<== result |%s| from |%s|\n", result, script);
	result = _free(result);
    }

    av = argvFree(av);

    mozFini(jss);

    return rc;
}
#endif	/* RPMJSS_SELF_TEST */
