// following code might be needed in some cases
// #define __STDC_LIMIT_MACROS
// #include <stdint.h>

#pragma GCC diagnostic ignored "-Winvalid-offsetof"

#include "system.h"

#include "jsapi.h"
#include "jsprf.h"

#include "jsfriendapi.h"

#include "jswrapper.h"

using namespace JS;

#define	_RPMJSS_INTERNAL
#include <rpmjss.h>

#include "debug.h"

/* The class of the global object. */
static JSClass global_class = {
    "global",
    JSCLASS_GLOBAL_FLAGS,
    JS_PropertyStub,
    JS_DeletePropertyStub,
    JS_PropertyStub,
    JS_StrictPropertyStub,
    JS_EnumerateStub,
    JS_ResolveStub,
    JS_ConvertStub,
    nullptr,
    nullptr,
    nullptr,
    nullptr,
    JS_GlobalObjectTraceHook
};

static int rpmjss_nopens;

int rpmjss_debug;
#define SPEW(_fmt, ...) \
    if (_rpmjss_debug) \
	fprintf(stderr, _fmt, __VA_ARGS__)

/*==============================================================*/
enum JSShellExitCode {
    EXITCODE_RUNTIME_ERROR	= 3,
    EXITCODE_FILE_NOT_FOUND	= 4,
    EXITCODE_OUT_OF_MEMORY	= 5,
    EXITCODE_TIMEOUT		= 6
};

enum PathResolutionMode {
    RootRelative,
    ScriptRelative
};

static JSRuntime *gRuntime = nullptr;

static int gExitCode = 0;
static bool gQuitting = false;
static bool gGotError = false;

static bool reportWarnings = true;
static bool compileOnly = false;
static bool fuzzingSafe = false;

static bool
Version(JSContext *cx, unsigned argc, jsval *vp)
{
    CallArgs args = CallArgsFromVp(argc, vp);
    JSVersion origVersion = JS_GetVersion(cx);
    if (args.length() == 0 || JSVAL_IS_VOID(args[0])) {
        /* Get version. */
        args.rval().setInt32(origVersion);
    } else {
        /* Set version. */
        int32_t v = -1;
        if (args[0].isInt32()) {
            v = args[0].toInt32();
        } else if (args[0].isDouble()) {
            double fv = args[0].toDouble();
            int32_t fvi;
            if (mozilla::NumberEqualsInt32(fv, &fvi))
                v = fvi;
        }
        if (v < 0 || v > JSVERSION_LATEST) {
#ifdef	FIXME
            JS_ReportErrorNumber(cx, my_GetErrorMessage, nullptr, JSSMSG_INVALID_ARGS, "version");
#endif
            return false;
        }
        JS_SetVersionForCompartment(js::GetContextCompartment(cx), JSVersion(v));
        args.rval().setInt32(origVersion);
    }
    return true;
}

/*
 * Resolve a (possibly) relative filename to an absolute path. If
 * |scriptRelative| is true, then the result will be relative to the directory
 * containing the currently-running script, or the current working directory if
 * the currently-running script is "-e" (namely, you're using it from the
 * command line.) Otherwise, it will be relative to the current working
 * directory.
 */
#ifdef	NOTYET
static JSString *
ResolvePath(JSContext *cx, HandleString filenameStr, PathResolutionMode resolveMode)
{
    JSAutoByteString filename(cx, filenameStr);
    if (!filename)
        return nullptr;

    const char *pathname = filename.ptr();
    if (pathname[0] == '/')
        return filenameStr;
#ifdef XP_WIN
    // Various forms of absolute paths per http://msdn.microsoft.com/en-us/library/windows/desktop/aa365247%28v=vs.85%29.aspx
    // "\..."
    if (pathname[0] == '\\')
        return filenameStr;
    // "C:\..."
    if (strlen(pathname) > 3 && isalpha(pathname[0]) && pathname[1] == ':' && pathname[2] == '\\')
        return filenameStr;
    // "\\..."
    if (strlen(pathname) > 2 && pathname[1] == '\\' && pathname[2] == '\\')
        return filenameStr;
#endif

    /* Get the currently executing script's name. */
    JS::AutoFilename scriptFilename;
    if (!DescribeScriptedCaller(cx, &scriptFilename))
        return nullptr;

    if (!scriptFilename.get())
        return nullptr;

    if (strcmp(scriptFilename.get(), "-e") == 0 || strcmp(scriptFilename.get(), "typein") == 0)
        resolveMode = RootRelative;

    static char buffer[PATH_MAX+1];
    if (resolveMode == ScriptRelative) {
#ifdef XP_WIN
        // The docs say it can return EINVAL, but the compiler says it's void
        _splitpath(scriptFilename.get(), nullptr, buffer, nullptr, nullptr);
#else
        strncpy(buffer, scriptFilename.get(), PATH_MAX+1);
        if (buffer[PATH_MAX] != '\0')
            return nullptr;

        // dirname(buffer) might return buffer, or it might return a
        // statically-allocated string
        memmove(buffer, dirname(buffer), strlen(buffer) + 1);
#endif
    } else {
        const char *cwd = getcwd(buffer, PATH_MAX);
        if (!cwd)
            return nullptr;
    }

    size_t len = strlen(buffer);
    buffer[len] = '/';
    strncpy(buffer + len + 1, pathname, sizeof(buffer) - (len+1));
    if (buffer[PATH_MAX] != '\0')
        return nullptr;

    return JS_NewStringCopyZ(cx, buffer);
}
#endif

#ifdef	NOTYET
static bool
CreateMappedArrayBuffer(JSContext *cx, unsigned argc, Value *vp)
{
    CallArgs args = CallArgsFromVp(argc, vp);

    if (args.length() < 1 || args.length() > 3) {
#ifdef	FIXME
        JS_ReportErrorNumber(cx, my_GetErrorMessage, nullptr,
                             args.length() < 1 ? JSSMSG_NOT_ENOUGH_ARGS : JSSMSG_TOO_MANY_ARGS,
                             "createMappedArrayBuffer");
#endif
        return false;
    }

    RootedString rawFilenameStr(cx, JS::ToString(cx, args[0]));
    if (!rawFilenameStr)
        return false;
    // It's a little bizarre to resolve relative to the script, but for testing
    // I need a file at a known location, and the only good way I know of to do
    // that right now is to include it in the repo alongside the test script.
    // Bug 944164 would introduce an alternative.
    JSString *filenameStr = ResolvePath(cx, rawFilenameStr, ScriptRelative);
    if (!filenameStr)
        return false;
    JSAutoByteString filename(cx, filenameStr);
    if (!filename)
        return false;

    uint32_t offset = 0;
    if (args.length() >= 2) {
        if (!JS::ToUint32(cx, args[1], &offset))
            return false;
    }

    bool sizeGiven = false;
    uint32_t size;
    if (args.length() >= 3) {
        if (!JS::ToUint32(cx, args[2], &size))
            return false;
        sizeGiven = true;
        if (offset > size) {
#ifdef	FIXME
            JS_ReportErrorNumber(cx, js_GetErrorMessage, nullptr,
                                 JSMSG_ARG_INDEX_OUT_OF_RANGE, "2");
#endif
            return false;
        }
    }

    FILE *file = fopen(filename.ptr(), "r");
    if (!file) {
#ifdef	FIXME
        JS_ReportErrorNumber(cx, my_GetErrorMessage, nullptr,
                             JSSMSG_CANT_OPEN, filename.ptr(), strerror(errno));
#endif
        return false;
    }
    AutoCloseInputFile autoClose(file);

    if (!sizeGiven) {
        struct stat st;
        if (fstat(fileno(file), &st) < 0) {
            JS_ReportError(cx, "Unable to stat file");
            return false;
        }
        if (st.st_size < offset) {
#ifdef	FIXME
            JS_ReportErrorNumber(cx, js_GetErrorMessage, nullptr,
                                 JSMSG_ARG_INDEX_OUT_OF_RANGE, "2");
#endif
            return false;
        }
        size = st.st_size - offset;
    }

    void *contents = JS_CreateMappedArrayBufferContents(fileno(file), offset, size);
    if (!contents) {
        JS_ReportError(cx, "failed to allocate mapped array buffer contents (possibly due to bad alignment)");
        return false;
    }

    RootedObject obj(cx, JS_NewMappedArrayBufferWithContents(cx, size, contents));
    if (!obj)
        return false;

    args.rval().setObject(*obj);
    return true;
}
#endif

static bool
Options(JSContext *cx, unsigned argc, jsval *vp)
{
    CallArgs args = CallArgsFromVp(argc, vp);

    JS::ContextOptions oldContextOptions = JS::ContextOptionsRef(cx);
    for (unsigned i = 0; i < args.length(); i++) {
        JSString *str = JS::ToString(cx, args[i]);
        if (!str)
            return false;
        args[i].setString(str);

        JSAutoByteString opt(cx, str);
        if (!opt)
            return false;

        if (strcmp(opt.ptr(), "strict") == 0)
            JS::ContextOptionsRef(cx).toggleExtraWarnings();
        else if (strcmp(opt.ptr(), "werror") == 0)
            JS::ContextOptionsRef(cx).toggleWerror();
        else if (strcmp(opt.ptr(), "strict_mode") == 0)
            JS::ContextOptionsRef(cx).toggleStrictMode();
        else {
            char* msg = JS_sprintf_append(nullptr,
                                          "unknown option name '%s'."
                                          " The valid names are strict,"
                                          " werror, and strict_mode.",
                                          opt.ptr());
            if (!msg) {
                JS_ReportOutOfMemory(cx);
                return false;
            }

            JS_ReportError(cx, msg);
            free(msg);
            return false;
        }
    }

    char *names = strdup("");
    bool found = false;
    if (!names && oldContextOptions.extraWarnings()) {
        names = JS_sprintf_append(names, "%s%s", found ? "," : "", "strict");
        found = true;
    }
    if (!names && oldContextOptions.werror()) {
        names = JS_sprintf_append(names, "%s%s", found ? "," : "", "werror");
        found = true;
    }
    if (!names && oldContextOptions.strictMode()) {
        names = JS_sprintf_append(names, "%s%s", found ? "," : "", "strict_mode");
        found = true;
    }
    if (!names) {
        JS_ReportOutOfMemory(cx);
        return false;
    }

    JSString *str = JS_NewStringCopyZ(cx, names);
    free(names);
    if (!str)
        return false;
    args.rval().setString(str);
    return true;
}

#ifdef	NOTYET
static bool
LoadScript(JSContext *cx, unsigned argc, jsval *vp, bool scriptRelative)
{
    CallArgs args = CallArgsFromVp(argc, vp);
    RootedObject thisobj(cx, JS_THIS_OBJECT(cx, vp));
    if (!thisobj)
        return false;

    RootedString str(cx);
    for (unsigned i = 0; i < args.length(); i++) {
        str = JS::ToString(cx, args[i]);
        if (!str) {
#ifdef	FIXME
            JS_ReportErrorNumber(cx, my_GetErrorMessage, nullptr, JSSMSG_INVALID_ARGS, "load");
#endif
            return false;
        }
        str = ResolvePath(cx, str, scriptRelative ? ScriptRelative : RootRelative);
        if (!str) {
            JS_ReportError(cx, "unable to resolve path");
            return false;
        }
        JSAutoByteString filename(cx, str);
        if (!filename)
            return false;
        errno = 0;
        CompileOptions opts(cx);
        opts.setIntroductionType("js shell load")
            .setUTF8(true)
            .setCompileAndGo(true)
            .setNoScriptRval(true);
        if ((compileOnly && !Compile(cx, thisobj, opts, filename.ptr())) ||
            !Evaluate(cx, thisobj, opts, filename.ptr()))
        {
            return false;
        }
    }

    args.rval().setUndefined();
    return true;
}
#endif

#ifdef	NOTYET
static bool
Load(JSContext *cx, unsigned argc, jsval *vp)
{
    return LoadScript(cx, argc, vp, false);
}
#endif

#ifdef	NOTYET
static bool
LoadScriptRelativeToScript(JSContext *cx, unsigned argc, jsval *vp)
{
    return LoadScript(cx, argc, vp, true);
}
#endif

// Populate |options| with the options given by |opts|'s properties. If we
// need to convert a filename to a C string, let fileNameBytes own the
// bytes.
#ifdef	NOTYET
static bool
ParseCompileOptions(JSContext *cx, CompileOptions &options, HandleObject opts,
                    JSAutoByteString &fileNameBytes)
{
    RootedValue v(cx);
    RootedString s(cx);

    if (!JS_GetProperty(cx, opts, "compileAndGo", &v))
        return false;
    if (!v.isUndefined())
        options.setCompileAndGo(ToBoolean(v));

    if (!JS_GetProperty(cx, opts, "noScriptRval", &v))
        return false;
    if (!v.isUndefined())
        options.setNoScriptRval(ToBoolean(v));

    if (!JS_GetProperty(cx, opts, "fileName", &v))
        return false;
    if (v.isNull()) {
        options.setFile(nullptr);
    } else if (!v.isUndefined()) {
        s = ToString(cx, v);
        if (!s)
            return false;
        char *fileName = fileNameBytes.encodeLatin1(cx, s);
        if (!fileName)
            return false;
        options.setFile(fileName);
    }

    if (!JS_GetProperty(cx, opts, "element", &v))
        return false;
    if (v.isObject())
        options.setElement(&v.toObject());

    if (!JS_GetProperty(cx, opts, "elementAttributeName", &v))
        return false;
    if (!v.isUndefined()) {
        s = ToString(cx, v);
        if (!s)
            return false;
        options.setElementAttributeName(s);
    }

    if (!JS_GetProperty(cx, opts, "lineNumber", &v))
        return false;
    if (!v.isUndefined()) {
        uint32_t u;
        if (!ToUint32(cx, v, &u))
            return false;
        options.setLine(u);
    }

    if (!JS_GetProperty(cx, opts, "sourceIsLazy", &v))
        return false;
    if (v.isBoolean())
        options.setSourceIsLazy(v.toBoolean());

    return true;
}
#endif

#ifdef	NOTYET
class AutoNewContext
{
  private:
    JSContext *oldcx;
    JSContext *newcx;
    Maybe<JSAutoRequest> newRequest;
    Maybe<AutoCompartment> newCompartment;

    AutoNewContext(const AutoNewContext &) MOZ_DELETE;

  public:
    AutoNewContext() : oldcx(nullptr), newcx(nullptr) {}

    bool enter(JSContext *cx) {
        JS_ASSERT(!JS_IsExceptionPending(cx));
        oldcx = cx;
        newcx = NewContext(JS_GetRuntime(cx));
        if (!newcx)
            return false;
        JS::ContextOptionsRef(newcx).setDontReportUncaught(true);
        js::SetDefaultObjectForContext(newcx, JS::CurrentGlobalOrNull(cx));

        newRequest.construct(newcx);
        newCompartment.construct(newcx, JS::CurrentGlobalOrNull(cx));
        return true;
    }

    JSContext *get() { return newcx; }

    ~AutoNewContext() {
        if (newcx) {
            RootedValue exc(oldcx);
            bool throwing = JS_IsExceptionPending(newcx);
            if (throwing)
                JS_GetPendingException(newcx, &exc);
            newCompartment.destroy();
            newRequest.destroy();
            if (throwing)
                JS_SetPendingException(oldcx, exc);
            DestroyContext(newcx, false);
        }
    }
};
#endif

static const uint32_t CacheEntry_SOURCE = 0;
static const uint32_t CacheEntry_BYTECODE = 1;

static const JSClass CacheEntry_class = {
    "CacheEntryObject", JSCLASS_HAS_RESERVED_SLOTS(2),
    JS_PropertyStub,       /* addProperty */
    JS_DeletePropertyStub, /* delProperty */
    JS_PropertyStub,       /* getProperty */
    JS_StrictPropertyStub, /* setProperty */
    JS_EnumerateStub,
    JS_ResolveStub,
    JS_ConvertStub,
    nullptr,               /* finalize */
    nullptr,               /* call */
    nullptr,               /* hasInstance */
    nullptr,               /* construct */
    nullptr,               /* trace */
    JSCLASS_NO_INTERNAL_MEMBERS
};

#ifdef	NOTYET
static bool
CacheEntry(JSContext* cx, unsigned argc, JS::Value *vp)
{
    CallArgs args = CallArgsFromVp(argc, vp);

    if (args.length() != 1 || !args[0].isString()) {
#ifdef	FIXME
        JS_ReportErrorNumber(cx, my_GetErrorMessage, nullptr, JSSMSG_INVALID_ARGS, "CacheEntry");
#endif
        return false;
    }

    RootedObject obj(cx, JS_NewObject(cx, &CacheEntry_class, JS::NullPtr(), JS::NullPtr()));
    if (!obj)
        return false;

    SetReservedSlot(obj, CacheEntry_SOURCE, args[0]);
    SetReservedSlot(obj, CacheEntry_BYTECODE, UndefinedValue());
    args.rval().setObject(*obj);
    return true;
}

static bool
CacheEntry_isCacheEntry(JSObject *cache)
{
    return JS_GetClass(cache) == &CacheEntry_class;
}

static JSString *
CacheEntry_getSource(HandleObject cache)
{
    JS_ASSERT(CacheEntry_isCacheEntry(cache));
    Value v = JS_GetReservedSlot(cache, CacheEntry_SOURCE);
    if (!v.isString())
        return nullptr;

    return v.toString();
}

static uint8_t *
CacheEntry_getBytecode(HandleObject cache, uint32_t *length)
{
    JS_ASSERT(CacheEntry_isCacheEntry(cache));
    Value v = JS_GetReservedSlot(cache, CacheEntry_BYTECODE);
    if (!v.isObject() || !v.toObject().is<ArrayBufferObject>())
        return nullptr;

    ArrayBufferObject *arrayBuffer = &v.toObject().as<ArrayBufferObject>();
    *length = arrayBuffer->byteLength();
    return arrayBuffer->dataPointer();
}

static bool
CacheEntry_setBytecode(JSContext *cx, HandleObject cache, uint8_t *buffer, uint32_t length)
{
    JS_ASSERT(CacheEntry_isCacheEntry(cache));
    Rooted<ArrayBufferObject*> arrayBuffer(cx, ArrayBufferObject::create(cx, length, buffer));

    if (!arrayBuffer || !ArrayBufferObject::ensureNonInline(cx, arrayBuffer))
        return false;

    SetReservedSlot(cache, CacheEntry_BYTECODE, OBJECT_TO_JSVAL(arrayBuffer));
    return true;
}
#endif

#ifdef	NOTYET
class AutoSaveFrameChain
{
    JSContext *cx_;
    bool saved_;

  public:
    AutoSaveFrameChain(JSContext *cx)
      : cx_(cx),
        saved_(false)
    {}

    bool save() {
        if (!JS_SaveFrameChain(cx_))
            return false;
        saved_ = true;
        return true;
    }

    ~AutoSaveFrameChain() {
        if (saved_)
            JS_RestoreFrameChain(cx_);
    }
};
#endif

#ifdef	NOTYET
static bool
Evaluate(JSContext *cx, unsigned argc, jsval *vp)
{
    CallArgs args = CallArgsFromVp(argc, vp);

    if (args.length() < 1 || args.length() > 2) {
#ifdef	FIXME
        JS_ReportErrorNumber(cx, my_GetErrorMessage, nullptr,
                             args.length() < 1 ? JSSMSG_NOT_ENOUGH_ARGS : JSSMSG_TOO_MANY_ARGS,
                             "evaluate");
#endif
        return false;
    }

    RootedString code(cx, nullptr);
    RootedObject cacheEntry(cx, nullptr);
    if (args[0].isString()) {
        code = args[0].toString();
    } else if (args[0].isObject() && CacheEntry_isCacheEntry(&args[0].toObject())) {
        cacheEntry = &args[0].toObject();
        code = CacheEntry_getSource(cacheEntry);
    }

    if (!code || (args.length() == 2 && args[1].isPrimitive())) {
#ifdef	FIXME
        JS_ReportErrorNumber(cx, my_GetErrorMessage, nullptr, JSSMSG_INVALID_ARGS, "evaluate");
#endif
        return false;
    }

    CompileOptions options(cx);
    JSAutoByteString fileNameBytes;
    bool newContext = false;
    RootedString displayURL(cx);
    RootedString sourceMapURL(cx);
    RootedObject global(cx, nullptr);
    bool catchTermination = false;
    bool saveFrameChain = false;
    bool loadBytecode = false;
    bool saveBytecode = false;
    bool assertEqBytecode = false;
    RootedObject callerGlobal(cx, cx->global());

    options.setIntroductionType("js shell evaluate")
           .setFileAndLine("@evaluate", 1);

    global = JS_GetGlobalForObject(cx, &args.callee());
    if (!global)
        return false;

    if (args.length() == 2) {
        RootedObject opts(cx, &args[1].toObject());
        RootedValue v(cx);

        if (!ParseCompileOptions(cx, options, opts, fileNameBytes))
            return false;

        if (!JS_GetProperty(cx, opts, "newContext", &v))
            return false;
        if (!v.isUndefined())
            newContext = ToBoolean(v);

        if (!JS_GetProperty(cx, opts, "displayURL", &v))
            return false;
        if (!v.isUndefined()) {
            displayURL = ToString(cx, v);
            if (!displayURL)
                return false;
        }

        if (!JS_GetProperty(cx, opts, "sourceMapURL", &v))
            return false;
        if (!v.isUndefined()) {
            sourceMapURL = ToString(cx, v);
            if (!sourceMapURL)
                return false;
        }

        if (!JS_GetProperty(cx, opts, "global", &v))
            return false;
        if (!v.isUndefined()) {
            if (v.isObject()) {
                global = js::UncheckedUnwrap(&v.toObject());
                if (!global)
                    return false;
            }
            if (!global || !(JS_GetClass(global)->flags & JSCLASS_IS_GLOBAL)) {
#ifdef	FIXME
                JS_ReportErrorNumber(cx, js_GetErrorMessage, nullptr, JSMSG_UNEXPECTED_TYPE,
                                     "\"global\" passed to evaluate()", "not a global object");
#endif
                return false;
            }
        }

        if (!JS_GetProperty(cx, opts, "catchTermination", &v))
            return false;
        if (!v.isUndefined())
            catchTermination = ToBoolean(v);

        if (!JS_GetProperty(cx, opts, "saveFrameChain", &v))
            return false;
        if (!v.isUndefined())
            saveFrameChain = ToBoolean(v);

        if (!JS_GetProperty(cx, opts, "loadBytecode", &v))
            return false;
        if (!v.isUndefined())
            loadBytecode = ToBoolean(v);

        if (!JS_GetProperty(cx, opts, "saveBytecode", &v))
            return false;
        if (!v.isUndefined())
            saveBytecode = ToBoolean(v);

        if (!JS_GetProperty(cx, opts, "assertEqBytecode", &v))
            return false;
        if (!v.isUndefined())
            assertEqBytecode = ToBoolean(v);

        // We cannot load or save the bytecode if we have no object where the
        // bytecode cache is stored.
        if (loadBytecode || saveBytecode) {
            if (!cacheEntry) {
#ifdef	FIXME
                JS_ReportErrorNumber(cx, my_GetErrorMessage, nullptr, JSSMSG_INVALID_ARGS,
                                     "evaluate");
#endif
                return false;
            }
        }
    }

    size_t codeLength;
    const jschar *codeChars = JS_GetStringCharsAndLength(cx, code, &codeLength);
    if (!codeChars)
        return false;

    AutoNewContext ancx;
    if (newContext) {
        if (!ancx.enter(cx))
            return false;
        cx = ancx.get();
    }

    uint32_t loadLength = 0;
    uint8_t *loadBuffer = nullptr;
    uint32_t saveLength = 0;
    ScopedJSFreePtr<uint8_t> saveBuffer;

    if (loadBytecode) {
        loadBuffer = CacheEntry_getBytecode(cacheEntry, &loadLength);
        if (!loadBuffer)
            return false;
    }

    {
        AutoSaveFrameChain asfc(cx);
        if (saveFrameChain && !asfc.save())
            return false;

        JSAutoCompartment ac(cx, global);
        RootedScript script(cx);

        if (!options.wrap(cx, cx->compartment()))
            return false;

        {
            JS::AutoSaveContextOptions asco(cx);
            JS::ContextOptionsRef(cx).setNoScriptRval(options.noScriptRval);
            if (saveBytecode) {
                if (!JS::CompartmentOptionsRef(cx).getSingletonsAsTemplates()) {
#ifdef	FIXME
                    JS_ReportErrorNumber(cx, my_GetErrorMessage, nullptr,
                                         JSSMSG_CACHE_SINGLETON_FAILED);
#endif
                    return false;
                }
                JS::CompartmentOptionsRef(cx).cloneSingletonsOverride().set(true);
            }

            if (loadBytecode) {
                script = JS_DecodeScript(cx, loadBuffer, loadLength, options.originPrincipals(cx));
            } else {
                script = JS::Compile(cx, global, options, codeChars, codeLength);
            }

            if (!script)
                return false;
        }

        if (displayURL && !script->scriptSource()->hasDisplayURL()) {
            const jschar *durl = JS_GetStringCharsZ(cx, displayURL);
            if (!durl)
                return false;
            if (!script->scriptSource()->setDisplayURL(cx, durl))
                return false;
        }
        if (sourceMapURL && !script->scriptSource()->hasSourceMapURL()) {
            const jschar *smurl = JS_GetStringCharsZ(cx, sourceMapURL);
            if (!smurl)
                return false;
            if (!script->scriptSource()->setSourceMapURL(cx, smurl))
                return false;
        }
        if (!JS_ExecuteScript(cx, global, script, args.rval())) {
            if (catchTermination && !JS_IsExceptionPending(cx)) {
                JSAutoCompartment ac1(cx, callerGlobal);
                JSString *str = JS_NewStringCopyZ(cx, "terminated");
                if (!str)
                    return false;
                args.rval().setString(str);
                return true;
            }
            return false;
        }

        if (saveBytecode) {
            saveBuffer = reinterpret_cast<uint8_t *>(JS_EncodeScript(cx, script, &saveLength));
            if (!saveBuffer)
                return false;
        }
    }

    if (saveBytecode) {
        // If we are both loading and saving, we assert that we are going to
        // replace the current bytecode by the same stream of bytes.
        if (loadBytecode && assertEqBytecode) {
            if (saveLength != loadLength) {
#ifdef	FIXME
                JS_ReportErrorNumber(cx, my_GetErrorMessage, nullptr, JSSMSG_CACHE_EQ_SIZE_FAILED,
                                     loadLength, saveLength);
#endif
            } else if (!mozilla::PodEqual(loadBuffer, saveBuffer.get(), loadLength)) {
#ifdef	FIXME
                JS_ReportErrorNumber(cx, my_GetErrorMessage, nullptr,
                                     JSSMSG_CACHE_EQ_CONTENT_FAILED);
#endif
            }
        }

        if (!CacheEntry_setBytecode(cx, cacheEntry, saveBuffer, saveLength))
            return false;

        saveBuffer.forget();
    }

    return JS_WrapValue(cx, args.rval());
}
#endif

#ifdef	NOTYET
static JSString *
FileAsString(JSContext *cx, const char *pathname)
{
    FILE *file;
    RootedString str(cx);
    size_t len, cc;
    char *buf;

    file = fopen(pathname, "rb");
    if (!file) {
        JS_ReportError(cx, "can't open %s: %s", pathname, strerror(errno));
        return nullptr;
    }
    AutoCloseInputFile autoClose(file);

    if (fseek(file, 0, SEEK_END) != 0) {
        JS_ReportError(cx, "can't seek end of %s", pathname);
    } else {
        len = ftell(file);
        if (fseek(file, 0, SEEK_SET) != 0) {
            JS_ReportError(cx, "can't seek start of %s", pathname);
        } else {
            buf = (char*) JS_malloc(cx, len + 1);
            if (buf) {
                cc = fread(buf, 1, len, file);
                if (cc != len) {
                    JS_ReportError(cx, "can't read %s: %s", pathname,
                                   (ptrdiff_t(cc) < 0) ? strerror(errno) : "short read");
                } else {
                    jschar *ucbuf =
                        JS::UTF8CharsToNewTwoByteCharsZ(cx, JS::UTF8Chars(buf, len), &len).get();
                    if (!ucbuf) {
                        JS_ReportError(cx, "Invalid UTF-8 in file '%s'", pathname);
                        gExitCode = EXITCODE_RUNTIME_ERROR;
                        return nullptr;
                    }
                    str = JS_NewUCStringCopyN(cx, ucbuf, len);
                    free(ucbuf);
                }
                JS_free(cx, buf);
            }
        }
    }

    return str;
}
#endif

#ifdef	NOTYET
static JSObject *
FileAsTypedArray(JSContext *cx, const char *pathname)
{
    FILE *file = fopen(pathname, "rb");
    if (!file) {
        JS_ReportError(cx, "can't open %s: %s", pathname, strerror(errno));
        return nullptr;
    }
    AutoCloseInputFile autoClose(file);

    RootedObject obj(cx);
    if (fseek(file, 0, SEEK_END) != 0) {
        JS_ReportError(cx, "can't seek end of %s", pathname);
    } else {
        size_t len = ftell(file);
        if (fseek(file, 0, SEEK_SET) != 0) {
            JS_ReportError(cx, "can't seek start of %s", pathname);
        } else {
            obj = JS_NewUint8Array(cx, len);
            if (!obj)
                return nullptr;
            char *buf = (char *) obj->as<TypedArrayObject>().viewData();
            size_t cc = fread(buf, 1, len, file);
            if (cc != len) {
                JS_ReportError(cx, "can't read %s: %s", pathname,
                               (ptrdiff_t(cc) < 0) ? strerror(errno) : "short read");
                obj = nullptr;
            }
        }
    }
    return obj;
}
#endif

/*
 * Function to run scripts and return compilation + execution time. Semantics
 * are closely modelled after the equivalent function in WebKit, as this is used
 * to produce benchmark timings by SunSpider.
 */
#ifdef	NOTYET
static bool
Run(JSContext *cx, unsigned argc, jsval *vp)
{
    CallArgs args = CallArgsFromVp(argc, vp);
    if (args.length() != 1) {
#ifdef	FIXME
        JS_ReportErrorNumber(cx, my_GetErrorMessage, nullptr, JSSMSG_INVALID_ARGS, "run");
#endif
        return false;
    }

    RootedObject thisobj(cx, JS_THIS_OBJECT(cx, vp));
    if (!thisobj)
        return false;

    JSString *str = JS::ToString(cx, args[0]);
    if (!str)
        return false;
    args[0].setString(str);
    JSAutoByteString filename(cx, str);
    if (!filename)
        return false;

    const jschar *ucbuf = nullptr;
    size_t buflen;
    str = FileAsString(cx, filename.ptr());
    if (str)
        ucbuf = JS_GetStringCharsAndLength(cx, str, &buflen);
    if (!ucbuf)
        return false;

    JS::Anchor<JSString *> a_str(str);

    RootedScript script(cx);
#ifdef	FIXME
    int64_t startClock = PRMJ_Now();
#else
    struct timeval tv;
    (void) gettimeofday(&tv, NULL);
    int64_t startClock = (tv.tv_sec * 1000 * 1000 + tv.tv_usec);
#endif
    {
        JS::AutoSaveContextOptions asco(cx);
        JS::ContextOptionsRef(cx).setNoScriptRval(true);

        JS::CompileOptions options(cx);
        options.setIntroductionType("js shell run")
               .setFileAndLine(filename.ptr(), 1)
               .setCompileAndGo(true);
        script = JS_CompileUCScript(cx, thisobj, ucbuf, buflen, options);
        if (!script)
            return false;
    }

    if (!JS_ExecuteScript(cx, thisobj, script))
        return false;

#ifdef	FIXME
    int64_t endClock = PRMJ_Now();
    args.rval().setDouble((endClock - startClock) / double(PRMJ_USEC_PER_MSEC));
#else
    (void) gettimeofday(&tv, NULL);
    int64_t endClock = (tv.tv_sec * 1000 * 1000 + tv.tv_usec);
    args.rval().setDouble((endClock - startClock) / 1000.);
#endif
    return true;
}
#endif

/*
 * function readline()
 * Provides a hook for scripts to read a line from stdin.
 */
#ifdef	NOTYET
static bool
ReadLine(JSContext *cx, unsigned argc, jsval *vp)
{
    CallArgs args = CallArgsFromVp(argc, vp);

#define BUFSIZE 256
    FILE *from = stdin;
    size_t buflength = 0;
    size_t bufsize = BUFSIZE;
    char *buf = (char *) JS_malloc(cx, bufsize);
    if (!buf)
        return false;

    bool sawNewline = false;
    size_t gotlength;
    while ((gotlength = js_fgets(buf + buflength, bufsize - buflength, from)) > 0) {
        buflength += gotlength;

        /* Are we done? */
        if (buf[buflength - 1] == '\n') {
            buf[buflength - 1] = '\0';
            sawNewline = true;
            break;
        } else if (buflength < bufsize - 1) {
            break;
        }

        /* Else, grow our buffer for another pass. */
        char *tmp;
        bufsize *= 2;
        if (bufsize > buflength) {
            tmp = (char *) JS_realloc(cx, buf, bufsize);
        } else {
            JS_ReportOutOfMemory(cx);
            tmp = nullptr;
        }

        if (!tmp) {
            JS_free(cx, buf);
            return false;
        }

        buf = tmp;
    }

    /* Treat the empty string specially. */
    if (buflength == 0) {
        args.rval().set(feof(from) ? NullValue() : JS_GetEmptyStringValue(cx));
        JS_free(cx, buf);
        return true;
    }

    /* Shrink the buffer to the real size. */
    char *tmp = static_cast<char*>(JS_realloc(cx, buf, buflength));
    if (!tmp) {
        JS_free(cx, buf);
        return false;
    }

    buf = tmp;

    /*
     * Turn buf into a JSString. Note that buflength includes the trailing null
     * character.
     */
    JSString *str = JS_NewStringCopyN(cx, buf, sawNewline ? buflength - 1 : buflength);
    JS_free(cx, buf);
    if (!str)
        return false;

    args.rval().setString(str);
    return true;
}
#endif

static bool
PutStr(JSContext *cx, unsigned argc, jsval *vp)
{
    FILE *gOutFile = stdout;
    CallArgs args = CallArgsFromVp(argc, vp);

    if (args.length() != 0) {
        JSString *str = JS::ToString(cx, args[0]);
        if (!str)
            return false;
#ifdef	FIXME
        char *bytes = JSStringToUTF8(cx, str);
#else
	char *bytes = JS_EncodeString(cx, str);
#endif
        if (!bytes)
            return false;
        fputs(bytes, gOutFile);
        JS_free(cx, bytes);
        fflush(gOutFile);
    }

    args.rval().setUndefined();
    return true;
}

static bool
Now(JSContext *cx, unsigned argc, jsval *vp)
{
    CallArgs args = CallArgsFromVp(argc, vp);
#ifdef	FIXME
    double now = PRMJ_Now() / double(PRMJ_USEC_PER_MSEC);
#else
    struct timeval tv;
    (void) gettimeofday(&tv, NULL);
    double now = (tv.tv_sec * 1000 * 1000 + tv.tv_usec)/1000.;
#endif
    args.rval().setDouble(now);
    return true;
}

static bool
PrintInternal(JSContext *cx, const CallArgs &args, FILE *file)
{
    for (unsigned i = 0; i < args.length(); i++) {
        JSString *str = JS::ToString(cx, args[i]);
        if (!str)
            return false;
#ifdef	FIXME
        char *bytes = JSStringToUTF8(cx, str);
#else
	char *bytes = JS_EncodeString(cx, str);
#endif
        if (!bytes)
            return false;
        fprintf(file, "%s%s", i ? " " : "", bytes);
        JS_free(cx, bytes);
    }

    fputc('\n', file);
    fflush(file);

    args.rval().setUndefined();
    return true;
}

static bool
Print(JSContext *cx, unsigned argc, jsval *vp)
{
    FILE *gOutFile = stdout;
    CallArgs args = CallArgsFromVp(argc, vp);
    return PrintInternal(cx, args, gOutFile);
}

static bool
PrintErr(JSContext *cx, unsigned argc, jsval *vp)
{
    FILE *gErrFile = stderr;
    CallArgs args = CallArgsFromVp(argc, vp);
    return PrintInternal(cx, args, gErrFile);
}

#ifdef	NOTYET
static bool
Help(JSContext *cx, unsigned argc, jsval *vp);
#endif

static bool
Quit(JSContext *cx, unsigned argc, jsval *vp)
{
#ifdef JS_MORE_DETERMINISTIC
    // Print a message to stderr in more-deterministic builds to help jsfunfuzz
    // find uncatchable-exception bugs.
    fprintf(stderr, "quit called\n");
#endif

    CallArgs args = CallArgsFromVp(argc, vp);
    JS_ConvertArguments(cx, args, "/ i", &gExitCode);

    gQuitting = true;
    return false;
}

/*==============================================================*/
typedef struct JSI_s * JSI_t;
struct JSI_s {
    JSRuntime	*rt;
    JSContext	*cx;
    JSObject	*global;
};

static void
rpmjssReportError(JSContext *cx, const char *message, JSErrorReport *report)
{
    fprintf(stderr, "%s:%u:%s\n",
	report->filename ? report->filename : "<no filename=\"filename\">",
	(unsigned int) report->lineno, message);
}

static JSFunctionSpec global_functions[] = {
    JS_FN("version",			Version,			0,0),
    JS_FN("options",			Options,			0,0),
#ifdef	NOTYET
    JS_FN("load",			Load,				1,0),
    JS_FN("loadRelativeToScript",	LoadScriptRelativeToScript,	1,0),
    JS_FN("evaluate",			Evaluate,			2,0),
    JS_FN("run",			Run,				1,0),
    JS_FN("readline",			ReadLine,			0,0),
#endif	/* NOTYET */
    JS_FN("print",			Print,				0,0),
    JS_FN("printErr",			PrintErr,			0,0),
    JS_FN("putstr",			PutStr,				0,0),
    JS_FN("dateNow",			Now,				0,0),
#ifdef	NOTYET
    JS_FN("help",			Help,				0,0),
#endif	/* NOTYET */
    JS_FN("quit",			Quit,				0,0),
#ifdef	NOTYET
    JS_FS("require",			Require,			0,0),
#endif	/* NOTYET */
    JS_FS_END
};

/*==============================================================*/
static void mozFini(rpmjss jss)
{
    JSI_t I = (JSI_t) jss->I;

    if (I->cx)
	JS_DestroyContext(I->cx);
    I->cx = NULL;

    if (I->rt)
	JS_DestroyRuntime(I->rt);
    I->rt = NULL;

    if (--rpmjss_nopens <= 0)  {
	JS_ShutDown();
	rpmjss_nopens = 0;
    }
    free(I);
}

static void * mozInit(rpmjss jss)
{
    JSI_t I = NULL;
    uint32_t flags = jss->flags;

    static uint32_t _maxbytes = 8L * 1024L * 1024L;
    static size_t _stackChunkSize = 8192;
    JSClass * _clasp = (JSClass *)&global_class;
    const JSFunctionSpec * _functions = global_functions;
    JSPrincipals * _principals = NULL;
    bool ok;

    if (rpmjss_nopens++ == 0)
	JS_Init();

    I = (JSI_t) calloc(1, sizeof(*I));
assert(I);

    I->rt = JS_NewRuntime(_maxbytes, JS_USE_HELPER_THREADS);
assert(I->rt);
    JS_SetRuntimePrivate(I->rt, (void *)jss);

    I->cx = JS_NewContext(I->rt, _stackChunkSize);
assert(I->cx);
    JS_SetContextPrivate(I->cx, (void *)jss);

    JS_SetErrorReporter(I->cx, rpmjssReportError);

    {
	JS::CompartmentOptions options;
	options.setVersion(JSVERSION_DEFAULT);
	JS::RootedObject glob(I->cx,
		JS_NewGlobalObject(I->cx, _clasp, _principals, JS::DontFireOnNewGlobalHook, options));
	I->global = glob;
assert(I->global);
	{
	    JSAutoCompartment ac(I->cx, I->global);
	    ok = JS_InitStandardClasses(I->cx, glob);
assert(ok);
#ifdef JS_HAS_CTYPES
	    ok = JS_InitCTypesClass(cx, glob);
assert(ok);
#endif
	    ok = JS_DefineFunctions(I->cx, glob, _functions);
assert(ok);
	}
	JS_FireOnNewGlobalObject(I->cx, glob);
    }

    return I;
}

static int mozRun(rpmjss jss, const char ** resultp,
		const char * script, const char * fn, unsigned ln)
{
    JSI_t I = (JSI_t) jss->I;
    JSContext* cx = I->cx;
    JSObject* global = I->global;

    JS::RootedObject glob(cx, global);
    JS::RootedValue rval(cx);
    {	// Scope for JSAutoCompartment
	JSAutoCompartment ac(cx, glob);

	bool ok = JS_EvaluateScript(cx, glob,
			script, strlen(script), fn, ln,
			&rval);
	if (!ok)
	    return 1;
    }

    char b[128];
    size_t nb = sizeof(b);
    char * t = NULL;

    if (rval.isNull()) {
	t = xstrdup("");
    } else
    if (rval.isUndefined()) {
	t = xstrdup("");
    } else
    if (rval.isBoolean()) {
	snprintf(b, nb, "%c", (rval.toBoolean() ? 'T' : 'F'));
	t = xstrdup(b);
    } else
    if (rval.isInt32()) {
	snprintf(b, nb, "%d", rval.toInt32());
	t = xstrdup(b);
    } else
    if (rval.isDouble()) {
	snprintf(b, nb, "%.16g", rval.toDouble());
	t = xstrdup(b);
    } else
    if (rval.isNumber()) {
	snprintf(b, nb, "%.16g", rval.toNumber());
	t = xstrdup(b);
    } else
    if (rval.isString()) {
    	t = JS_EncodeString(cx, rval.toString());
    } else
	t = xstrdup("");

    if (resultp)
	*resultp = t;
    else
	t = _free(t);

    return 0;
}

static struct JSIO_s _mozjs = { mozFini, mozInit, mozRun };
JSIO_t mozjs31 = &_mozjs;

#if defined(RPMJSS_SELF_TEST)
#include <rpmio.h>
#include <argv.h>

struct rpmjss_s _jss;
rpmjss jss = &_jss;

/*==============================================================*/
int main(int argc, const char *argv[])
{
    int rc = 0;

    JSI_t I = (JSI_t) mozInit(jss);
    jss->I = I;

    const char * _fn = "rpmjss.inp";
    FD_t fd = Fopen(_fn, "r.fpio");
    ARGV_t av = NULL;
    rc = argvFgets(&av, fd, 0);
    rc = Fclose(fd);
    int ac = argvCount(av);

    for (int i = 0; i < ac; i++) {
	const char * script = av[i];
	const char * result = NULL;
	rc = mozRun(jss, &result, script, _fn, i);
	fprintf(stderr, "<== result |%s| from |%s|\n", result, script);
	result = _free(result);
    }

    av = argvFree(av);

    mozFini(jss);

    return rc;
}
#endif	/* RPMJSS_SELF_TEST */
