/** \ingroup rpmio
 * \file rpmio/rpmku.c
 */

#include "system.h"

#if defined(HAVE_SYS_SYSCALL_H)
#include <sys/syscall.h>
#endif

#define	_RPMIOB_INTERNAL
#include <rpmio_internal.h>

#if defined(HAVE_KEYUTILS_H)
#include <rpmmacro.h>
#include <argv.h>

#ifdef __cplusplus
extern "C" {
#endif
#include <keyutils.h>
#ifdef __cplusplus
}
#endif

#define _RPMPGP_INTERNAL
#include <rpmpgp.h>
#endif	/* HAVE_KEYUTILS_H */

#include <rpmku.h>

#include "debug.h"

#if defined(HAVE_KEYUTILS_H)
rpmint32_t _kuKeyring;

static int _kuCache = 1;

typedef struct _kuItem_s {
    const char *name;
    key_serial_t val;
} * _kuItem;

/* NB: the following table must be sorted lexically for bsearch(3). */
static struct _kuItem_s kuTable[] = {
    { "group",		KEY_SPEC_GROUP_KEYRING },
    { "process",	KEY_SPEC_PROCESS_KEYRING },
    { "reqkey_auth",	KEY_SPEC_REQKEY_AUTH_KEY },
    { "session",	KEY_SPEC_SESSION_KEYRING },
    { "thread",		KEY_SPEC_THREAD_KEYRING },
    { "user",		KEY_SPEC_USER_KEYRING },
    { "user_session",	KEY_SPEC_USER_SESSION_KEYRING },
};
static size_t nkuTable = sizeof(kuTable) / sizeof(kuTable[0]);

static int
kuCmp(const void * a, const void * b)
{
    return strcmp(((_kuItem)a)->name, ((_kuItem)b)->name);
}

static key_serial_t
kuValue(const char * name)
{
    _kuItem k = NULL;

    if (name != NULL && *name != '\0') {
	_kuItem tmp = (_kuItem) memset(alloca(sizeof(*tmp)), 0, sizeof(*tmp));
	tmp->name = name;
	k = (_kuItem)bsearch(tmp, kuTable, nkuTable, sizeof(kuTable[0]), kuCmp);
    }
    return (k != NULL ? k->val :  0);
}
#endif	/* HAVE_KEYUTILS_H */

char * _GetPass(const char * prompt)
{
    char * pw;

    pw = getpass( prompt ? prompt : "" );

#if defined(HAVE_KEYUTILS_H)
    if (_kuKeyring == 0) {
	const char * _keyutils_keyring
		= rpmExpand("%{?_keyutils_keyring}", NULL);
	_kuKeyring = (rpmuint32_t) kuValue(_keyutils_keyring);
	if (_kuKeyring == 0)
	    _kuKeyring = KEY_SPEC_PROCESS_KEYRING;
	_keyutils_keyring = _free(_keyutils_keyring);
    }

    if (pw && *pw) {
	key_serial_t keyring = (key_serial_t) _kuKeyring;
	size_t npw = strlen(pw);
	(void) add_key("user", "rpm:passwd", pw, npw, keyring);
	(void) memset(pw, 0, npw);	/* burn the password */
	pw = (char *) "@u user rpm:passwd";
    }
#endif	/* HAVE_KEYUTILS_H */

assert(pw != NULL);
    return pw;
}

char * _RequestPass(const char * prompt)
{
    static char * password = NULL;
#if defined(HAVE_KEYUTILS_H)
    const char * foo = "user rpm:yyyy spoon";
    ARGV_t av = NULL;
    int xx = argvSplit(&av, foo, NULL);
    key_serial_t dest = 0;
    key_serial_t key = 0;

    if (password != NULL) {
	free(password);
	password = NULL;
    }
assert(av != NULL);
assert(av[0] != NULL);
assert(av[1] != NULL);
assert(av[2] != NULL);
    key = request_key(av[0], av[1], av[2], dest);
    av = argvFree(av);

    xx = keyctl_read_alloc(key, (void **)&password);
    if (password == NULL)
	password = (char *) "";
#endif	/* HAVE_KEYUTILS_H */

    return password;
}

char * (*Getpass) (const char * prompt) = _GetPass;

rpmRC rpmkuFindPubkey(pgpDigParams sigp, rpmiob * iobp)
{
    if (iobp != NULL)
	*iobp = NULL;

#if defined(HAVE_KEYUTILS_H)
    if (_kuCache) {
	static const char krprefix[] = "rpm:gpg:pubkey:";
	key_serial_t keyring = (key_serial_t) _kuKeyring;
	char krfp[32];
	char * krn = (char *) alloca(strlen(krprefix) + sizeof("12345678"));
	long key;
	int xx;

	(void) snprintf(krfp, sizeof(krfp), "%08X", pgpGrab(sigp->signid+4, 4));
	krfp[sizeof(krfp)-1] = '\0';
	*krn = '\0';
	(void) stpcpy( stpcpy(krn, krprefix), krfp);

	key = keyctl_search(keyring, "user", krn, 0);
	xx = keyctl_read(key, NULL, 0);
	if (xx > 0) {
	    rpmiob iob = rpmiobNew(xx);
	    xx = keyctl_read(key, (char *)iob->b, iob->blen);
	    if (xx > 0) {
#ifdef	NOTYET
		pubkeysource = xstrdup(krn);
		_kuCache = 0;	/* XXX don't bother caching. */
#endif
	    } else
		iob = rpmiobFree(iob);

	    if (iob != NULL && iobp != NULL) {
		*iobp = iob;
		return RPMRC_OK;
	    } else {
		iob = rpmiobFree(iob);
		return RPMRC_NOTFOUND;
	    }
	} else
	    return RPMRC_NOTFOUND;
    } else
#endif	/* HAVE_KEYUTILS_H */
    return RPMRC_NOTFOUND;
}

rpmRC rpmkuStorePubkey(pgpDigParams sigp, rpmiob iob)
{
#if defined(HAVE_KEYUTILS_H)
    if (_kuCache) {
	static const char krprefix[] = "rpm:gpg:pubkey:";
	key_serial_t keyring = (key_serial_t) _kuKeyring;
	char krfp[32];
	char * krn = (char *) alloca(strlen(krprefix) + sizeof("12345678"));

	(void) snprintf(krfp, sizeof(krfp), "%08X", pgpGrab(sigp->signid+4, 4));
	krfp[sizeof(krfp)-1] = '\0';
	*krn = '\0';
	(void) stpcpy( stpcpy(krn, krprefix), krfp);
	(void) add_key("user", krn, iob->b, iob->blen, keyring);
    }
#endif	/* HAVE_KEYUTILS_H */
    iob = rpmiobFree(iob);
    return RPMRC_OK;
}

const char * rpmkuPassPhrase(const char * passPhrase)
{
    const char * pw;

#if defined(HAVE_KEYUTILS_H)
    if (passPhrase && !strcmp(passPhrase, "@u user rpm:passwd")) {
	key_serial_t keyring = (key_serial_t) _kuKeyring;
	long key;
	int xx;

	key = keyctl_search(keyring, "user", "rpm:passwd", 0);
	pw = NULL;
	xx = keyctl_read_alloc(key, (void **)&pw);
	if (xx < 0)
	    pw = NULL;
    } else
#endif	/* HAVE_KEYUTILS_H */
	pw = xstrdup(passPhrase);
    return pw;
}

#if defined(HAVE_KEYUTILS_H)
#include <keyutils.h>

#else	/* HAVE_KEYUTILS_H */

typedef	int32_t key_serial_t;

#if defined(HAVE_ADD_KEY)
extern	key_serial_t add_key(const char *type, const char *description,
                            const void *payload, size_t plen,
                            key_serial_t ringid);
#else	/* defined(HAVE_ADD_KEY) */
static inline key_serial_t add_key(const char *type, const char *description,
                            const void *payload, size_t plen,
                            key_serial_t ringid)
{
    key_serial_t rc = -2;
#if defined(__linux__) && defined(SYS_add_key)
    rc = syscall(SYS_add_key, type, description, payload, plen, ringid);
#else
    errno = ENOSYS;
#endif
    return rc;
}
#endif	/* defined(HAVE_ADD_KEY) */

#if defined(HAVE_REQUEST_KEY)
extern key_serial_t request_key(const char *type,
                                const char *description,
                                const char *callout_info,
                                key_serial_t destringid);
#else	/* defined(HAVE_REQUEST_KEY) */
static inline key_serial_t request_key(const char *type,
                                const char *description,
                                const char *callout_info,
                                key_serial_t destringid)
{
    key_serial_t rc = -2;
#if defined(__linux__) && defined(SYS_request_key)
    rc = syscall(SYS_request_key, type, description, callout_info, destringid);
#else
    errno = ENOSYS;
#endif
    return rc;
}
#endif	/* defined(HAVE_REQUEST_KEY) */

#if defined(HAVE_KEYCTL)
extern long keyctl(int cmd, ...);
#else	/* defined(HAVE_KEYCTL) */
static inline long __keyctl(int cmd,
		unsigned long arg2,
		unsigned long arg3,
		unsigned long arg4,
		unsigned long arg5)
{
    long rc = -2;
#if defined(__linux__) && defined(SYS_keyctl)
    rc = syscall(SYS_keyctl, cmd, arg2, arg3, arg4, arg5);
#else
    errno = ENOSYS;
#endif
    return rc;
}

static inline long keyctl(int cmd, ...)
{
    va_list ap;
    va_start(ap, cmd);
    unsigned long arg2 = va_arg(ap, unsigned long);
    unsigned long arg3 = va_arg(ap, unsigned long);
    unsigned long arg4 = va_arg(ap, unsigned long);
    unsigned long arg5 = va_arg(ap, unsigned long);
    va_end(ap);
    return __keyctl(cmd, arg2, arg3, arg4, arg5);
}

#endif	/* defined(HAVE_KEYCTL) */

#endif	/* defined(HAVE_KEYUTILS_H) */

key_serial_t Add_key(const char *type, const char *description,
                            const void *payload, size_t plen,
                            key_serial_t ringid);
key_serial_t Add_key(const char *type, const char *description,
                            const void *payload, size_t plen,
                            key_serial_t ringid)
{
    key_serial_t rc = add_key(type, description, payload, plen, ringid);
    if (rc < 0)
	rpmlog(RPMLOG_ERR, _("%s(%s,%s,%p[%u],%d) failed: %m\n"),
		__FUNCTION__, type, description, payload, (unsigned)plen, (int)ringid);
if (_rpmio_debug)
fprintf(stderr, _("<-- %s(%s,%s,%p[%u],%d) rc %d\n"), __FUNCTION__, type, description, payload, (unsigned)plen, (int)ringid, (int)rc);
    return rc;
}

key_serial_t Request_key(const char *type,
                                const char *description,
                                const char *callout_info,
                                key_serial_t destringid);
key_serial_t Request_key(const char *type,
                                const char *description,
                                const char *callout_info,
                                key_serial_t destringid)
{
    key_serial_t rc = request_key(type, description, callout_info, destringid);
    if (rc < 0)
	rpmlog(RPMLOG_ERR, _("%s(%s,%s,%s,%d) failed: %m\n"),
		__FUNCTION__, type, description, callout_info, (int)destringid);
if (_rpmio_debug)
fprintf(stderr, _("<-- %s(%s,%s,%s,%d) rc %d\n"), __FUNCTION__, type, description, callout_info, (int)destringid, (int)rc);
    return rc;
}

static inline long vkeyctl(int cmd, va_list ap)
{
    long rc = -2;
    va_list apc;
    va_copy(apc, ap);
    unsigned long arg2 = va_arg(apc, unsigned long);
    unsigned long arg3 = va_arg(apc, unsigned long);
    unsigned long arg4 = va_arg(apc, unsigned long);
    unsigned long arg5 = va_arg(apc, unsigned long);
    va_end(apc);

#if defined(HAVE_KEYCTL)
    rc = keyctl(cmd, arg2, arg3, arg4, arg5);
#else
    rc = __keyctl(cmd, arg2, arg3, arg4, arg5);
#endif
    return rc;
}

long Keyctl(int cmd, ...);
long Keyctl(int cmd, ...)
{
    va_list ap;
    va_start(ap, cmd);
    long rc = vkeyctl(cmd, ap);
    va_end(ap);

    if (rc < 0)
	rpmlog(RPMLOG_ERR, _("%s(%d) failed: %m\n"),
		__FUNCTION__, cmd);
if (_rpmio_debug)
fprintf(stderr, _("<-- %s(%d,...) rc %ld\n"), __FUNCTION__, cmd, (long)rc);
    return rc;
}
