#ifndef _H_MACRO_
#define	_H_MACRO_

/** \ingroup rpmio
 * \file rpmio/rpmmacro.h
 */
#include <stdio.h>

#include <rpmdefs.h>

typedef struct MacroEntry_s * MacroEntry;
typedef struct MacroContext_s * rpmme;
typedef struct MacroContext_s * MacroContext;
typedef struct MacroContext_s * rpmmc;

#if defined(_MACRO_INTERNAL)
#include <rpmiotypes.h>
/*! The structure used to store a macro. */
struct MacroEntry_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    struct MacroEntry_s *prev;	/*!< Macro entry stack. */
    const char *name;		/*!< Macro name. */
    const char *opts;		/*!< Macro parameters (a la getopt) */
    const char *body;		/*!< Macro body. */
    int	used;			/*!< No. of expansions. */
    short level;		/*!< Scoping level. */
    unsigned short flags;	/*!< Flags. */
};

/*! The structure used to store the set of macros in a context. */
struct MacroContext_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    MacroEntry *macroTable;	/*!< Macro entry table for context. */
    int	macrosAllocated;	/*!< No. of allocated macros. */
    int	firstFree;		/*!< No. of macros. */
    uint32_t flags;
};
#endif

extern MacroContext rpmGlobalMacroContext;

extern MacroContext rpmCLIMacroContext;

/** \ingroup rpmrc
 * List of macro files to read when configuring rpm.
 * This is a colon separated list of files. URI's are permitted as well,
 * identified by the token '://', so file paths must not begin with '//'.
 */
extern const char * rpmMacrofiles;

/**
 * Markers for sources of macros added throughout rpm.
 */
#define	RMIL_DEFAULT	-15
#define	RMIL_MACROFILES	-13
#define	RMIL_RPMRC	-11

#define	RMIL_CMDLINE	-7
#define	RMIL_TARBALL	-5
#define	RMIL_SPEC	-3
#define	RMIL_OLDSPEC	-1
#define	RMIL_GLOBAL	0

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Print macros to file stream.
 * @param mc		macro context (NULL uses global context).
 * @param fp		file stream (NULL uses stderr).
 */
void rpmDumpMacroTable(MacroContext mc, FILE * fp);

/**
 * Return macro entries as string array.
 * @param mc		macro context (NULL uses global context)
 * @param _mire		pattern to match (NULL disables)
 * @param used		macro usage (<0 all, =0 unused, >=1 used count)
 * @retval *avp		macro definitions
 * @return		no. of entries
 */
int
rpmGetMacroEntries(MacroContext mc, void * _mire,
		int used, const char *** avp);

/**
 * Check whether configuration file is moderately secure to load.
 * @param filename	filename to check
 * @return		1 on success, 0 on failure
 * XXX FIXME: rpmRC return?
 */
int rpmSecuritySaneFile(const char *filename);

/**
 * Return URL path(s) from a (URL prefixed) pattern glob.
 * @param patterns	glob pattern
 * @retval *argcPtr	no. of paths
 * @retval *argvPtr	array of paths (malloc'd contiguous blob)
 * @return		0 on success
 * XXX FIXME: rpmRC return?
 */
int rpmGlob(const char * patterns, int * argcPtr, const char *** argvPtr);

/**
 * Expand macro into buffer.
 * @deprecated Use rpmExpand().
 * @todo Eliminate from API.
 * @param spec		cookie (unused)
 * @param mc		macro context (NULL uses global context).
 * @retval sbuf		input macro to expand, output expansion
 * @param slen		size of buffer
 * @return		0 on success
 */
int expandMacros(void * spec, MacroContext mc, char * sbuf, size_t slen);

/**
 * Add macro to context.
 * @deprecated Use rpmDefineMacro().
 * @param mc		macro context (NULL uses global context).
 * @param n		macro name
 * @param o		macro paramaters
 * @param b		macro body
 * @param level		macro recursion level (0 is entry API)
 */
void addMacro(MacroContext mc, const char * n,
		const char * o, const char * b, int level);

/**
 * Delete macro from context.
 * @deprecated Use rpmUndefineMacro().
 * @param mc		macro context (NULL uses global context).
 * @param n		macro name
 */
void delMacro(MacroContext mc, const char * n);

/**
 * Define macro in context.
 * @param mc		macro context (NULL uses global context).
 * @param macro		macro name, options, body
 * @param level		macro recursion level (0 is entry API)
 * @return		0 on success
 */
int rpmDefineMacro(MacroContext mc, const char * macro, int level);

/**
 * Undefine macro in context.
 * @param mc		macro context (NULL uses global context).
 * @param macro		macro name
 * @return		0 on success
 */
int rpmUndefineMacro(MacroContext mc, const char * macro);

/**
 * Load macros from specific context into global context.
 * @param mc		macro context (NULL does nothing).
 * @param level		macro recursion level (0 is entry API)
 */
void rpmLoadMacros(MacroContext mc, int level);

/**
 * Load macro context from a macro file.
 * @param mc		(unused)
 * @param fn		macro file name
 * @param nesting	max load recursion depth, 0 disables.
 */
int rpmLoadMacroFile(MacroContext mc, const char * fn, int nesting);

/**
 * Initialize macro context from set of macrofile(s).
 * @param mc		macro context
 * @param macrofiles	colon separated list of macro files (NULL does nothing)
 */
void rpmInitMacros(MacroContext mc, const char * macrofiles);

/**
 * Destroy macro context.
 * @param mc		macro context (NULL uses global context).
 */
void rpmFreeMacros(MacroContext mc);

typedef enum rpmCompressedMagic_e {
    COMPRESSED_NOT		=  0,	/*!< not compressed */
    COMPRESSED_OTHER		=  1,	/*!< gzip can handle */
    COMPRESSED_BZIP2		=  2,	/*!< bzip2 can handle */
    COMPRESSED_ZIP		=  3,	/*!< unzip can handle */
    COMPRESSED_LZOP		=  4,	/*!< lzop can handle */
    COMPRESSED_LZMA		=  5,	/*!< lzma can handle */
    COMPRESSED_XZ		=  6,	/*!< xz can handle */
    COMPRESSED_LRZIP		=  7,	/*!< lrzip can handle */
    COMPRESSED_LZIP		=  8,	/*!< lzip can handle */
    COMPRESSED_7ZIP		=  9,	/*!< 7zip can handle */
    COMPRESSED_ZSTD		= 10,	/*!< 7zip can handle */
} rpmCompressedMagic;

/**
 * Return type of compression used in file.
 * @param file		name of file
 * @retval compressed	address of compression type
 * @return		0 on success, 1 on I/O error
 */
int isCompressed(const char * file, rpmCompressedMagic * compressed);

/**
 * Return (malloc'ed) concatenated macro expansion(s) in a context.
 * @param mc		macro context
 * @param arg		macro(s) to expand (NULL terminates list)
 * @return		macro expansion (malloc'ed)
 */
char * vrpmMCExpand(MacroContext mc, const char * arg, va_list ap);
char * rpmMCExpand(MacroContext mc, const char * arg, ...)
	RPM_GNUC_NULL_TERMINATED;

/**
 * Return (malloc'ed) concatenated macro expansion(s).
 * @param arg		macro(s) to expand (NULL terminates list)
 * @return		macro expansion (malloc'ed)
 */
char * rpmExpand(const char * arg, ...)
	RPM_GNUC_NULL_TERMINATED;

/**
 * Canonicalize file path.
 * @note Path canonicalization always removes the pesky trailing '/'.
 * @param path		path to canonicalize (in-place)
 * @return		pointer to path
 */
char * rpmCleanPath(char * path);

/**
 * Return (malloc'ed) expanded, canonicalized, file path.
 * @note A final "/" argument will force append a pesky trailing '/'.
 * @param path		macro(s) to expand (NULL terminates list)
 * @return		canonicalized path (malloc'ed)
 */
char * rpmGetPath(const char * path, ...)
	RPM_GNUC_NULL_TERMINATED;

/**
 * Merge 3 args into path, any or all of which may be a url.
 * The leading part of the first URL encountered is used
 * for the result, other URL prefixes are discarded, permitting
 * a primitive form of URL inheiritance.
 * @param urlroot	root URL (often path to chroot, or NULL)
 * @param urlmdir	directory URL (often a directory, or NULL)
 * @param urlfile	file URL (often a file, or NULL)
 * @return		expanded, merged, canonicalized path (malloc'ed)
 */
const char * rpmGenPath(const char * urlroot,
			const char * urlmdir,
			const char * urlfile);

/**
 * Return macro expansion as a numeric value.
 * Boolean values ('Y' or 'y' returns 1, 'N' or 'n' returns 0)
 * are permitted as well. An undefined macro returns 0.
 * @param arg		macro to expand
 * @return		numeric value
 */
int rpmExpandNumeric (const char * arg);

int rpmMCExpandNumeric(MacroContext mc, const char *arg);

/**
 * Unreference a me wrapper instance.
 * @param me		me wrapper
 * @return		NULL on last dereference
 */
rpmme rpmmeUnlink (rpmme me);
#define	rpmmeUnlink(_me)	\
    ((rpmme)rpmioUnlinkPoolItem((rpmioItem)(_me), __FUNCTION__, __FILE__, __LINE__))

/**
 * Reference a me wrapper instance.
 * @param me		me wrapper
 * @return		new me wrapper reference
 */
rpmme rpmmeLink (rpmme me);
#define	rpmmeLink(_me)	\
    ((rpmme)rpmioLinkPoolItem((rpmioItem)(_me), __FUNCTION__, __FILE__, __LINE__))

/**
 * Destroy a me wrapper.
 * @param me		me wrapper
 * @return		NULL on last dereference
 */
rpmme rpmmeFree(rpmme me);
#define	rpmmeFree(_me)	\
    ((rpmme)rpmioFreePoolItem((rpmioItem)(_me), __FUNCTION__, __FILE__, __LINE__))

/**
 * Create and load a me wrapper.
 * @param fn		me args (or NULL)
 * @param flags		me flags
 * @return		new me wrapper
 */
rpmme rpmmeNew(char ** av, uint32_t flags);
/**
 * Unreference a mc wrapper instance.
 * @param mc		mc wrapper
 * @return		NULL on last dereference
 */
rpmmc rpmmcUnlink (rpmmc mc);
#define	rpmmcUnlink(_mc)	\
    ((rpmmc)rpmioUnlinkPoolItem((rpmioItem)(_mc), __FUNCTION__, __FILE__, __LINE__))

/**
 * Reference a mc wrapper instance.
 * @param mc		mc wrapper
 * @return		new mc wrapper reference
 */
rpmmc rpmmcLink (rpmmc mc);
#define	rpmmcLink(_mc)	\
    ((rpmmc)rpmioLinkPoolItem((rpmioItem)(_mc), __FUNCTION__, __FILE__, __LINE__))

/**
 * Destroy a mc wrapper.
 * @param mc		mc wrapper
 * @return		NULL on last dereference
 */
rpmmc rpmmcFree(rpmmc mc);
#define	rpmmcFree(_mc)	\
    ((rpmmc)rpmioFreePoolItem((rpmioItem)(_mc), __FUNCTION__, __FILE__, __LINE__))

/**
 * Create and load a mc wrapper.
 * @param fn		mc args (or NULL)
 * @param flags		mc flags
 * @return		new mc wrapper
 */
rpmmc rpmmcNew(char ** av, uint32_t flags);

#ifdef __cplusplus
}
#endif

#endif	/* _H_ MACRO_ */
