/** \ingroup rpmio
 * \file rpmio/rpmmalloc.c
 */

#include "system.h"

#if defined(HAVE_MALLOC_H)
#include <malloc.h>
#endif

#include <rpmiotypes.h>
#include <rpmio.h>
#include <rpmlog.h>
#include <yarn.h>

#include "debug.h"

#ifdef __cplusplus
GENfree(rpmioItem)
#endif	/* __cplusplus */

#if defined(WITH_DMALLOC)
#undef xmalloc
#undef xcalloc
#undef xrealloc
#undef xstrdup
#endif

#if !defined(EXIT_FAILURE)
#define	EXIT_FAILURE	1
#endif

void *vmefail(size_t size)
{
    fprintf(stderr, _("memory alloc (%u bytes) returned NULL.\n"), (unsigned)size);
    exit(EXIT_FAILURE);
#if     !(__GNUC__ > 2 || (__GNUC__ == 2 && __GNUC_MINOR__ > 4))
    return NULL;
#endif
}

/**
 */
struct rpmioPool_s {
    yarnLock have;		/*!< unused items available, lock for list */
    void *pool;
    rpmioItem head;		/*!< linked list of available items */
    rpmioItem * tail;
    size_t size;		/*!< size of items in this pool */
    int limit;			/*!< number of new items allowed, or -1 */
    int flags;
    char * (*dbug) (void *item, char *b, size_t nb);
				/*!< generate string w Unlink/Link debugging */
    void (*init) (void *item);	/*!< create item contents. */
    void (*fini) (void *item);	/*!< destroy item contents. */
    int reused;			/*!< number of items reused */
    int made;			/*!< number of items made */
    const char *name;
    void * zlog;
};

static rpmioPool _rpmioPool;

rpmioPool rpmioFreePool(rpmioPool pool)
{
    if (pool == NULL) {
	pool = _rpmioPool;
	_rpmioPool = NULL;
    }
    if (pool) {
	rpmioItem item;
	int count = 0;
	yarnPossess(pool->have);
VALGRIND_HG_CLEAN_MEMORY(pool, sizeof(*pool));
	while ((item = pool->head) != NULL) {
VALGRIND_HG_CLEAN_MEMORY(item, pool->size);
	    pool->head = (rpmioItem) item->pool;	/* XXX pool == next */
	    if (item->use)
		item->use = yarnFreeLock(item->use);
	    item = _free(item);
	    count++;
	}
	yarnRelease(pool->have);
	pool->have = yarnFreeLock(pool->have);
	rpmlog(RPMLOG_DEBUG, D_("pool %s:\treused %d, alloc'd %d, free'd %d items.\n"), pool->name, pool->reused, pool->made, count);
#ifdef	NOTYET
assert(pool->made == count);
#else
if (pool->made != count)
rpmlog(RPMLOG_WARNING, D_("pool %s: FIXME: made %d, count %d\nNote: This is a harmless memory leak discovered while exiting, relax ...\n"), pool->name, pool->made, count);
#endif
	(void) _free(pool);
	VALGRIND_DESTROY_MEMPOOL(pool);
    }
    return NULL;
}

rpmioPool rpmioNewPool(const char * name, size_t size, int limit, int flags,
		char * (*dbug) (void *item, char *b, size_t nb),
		void (*init) (void *item),
		void (*fini) (void *item))
{
    rpmioPool pool = (rpmioPool) xcalloc(1, sizeof(*pool));
#if defined(WITH_VALGRIND)
    static int rzB = 0;		/* size of red-zones (if any) */
    static int is_zeroed = 0;	/* does pool return zero'd allocations? */
    (void)rzB;			/* XXX CentOS5 valgrind doesn't use. */
    (void)is_zeroed;		/* XXX CentOS5 valgrind doesn't use. */
#endif
    VALGRIND_CREATE_MEMPOOL(pool, rzB, is_zeroed);
    pool->have = yarnNewLock(0);
    pool->pool = NULL;
    pool->head = NULL;
    pool->tail = &pool->head;
    pool->size = size;
    pool->limit = limit;
    pool->flags = flags;
    pool->dbug = dbug;
    pool->init = init;
    pool->fini = fini;
    pool->reused = 0;
    pool->made = 0;
    pool->name = name;
    pool->zlog = NULL;
    rpmlog(RPMLOG_DEBUG, D_("pool %s:\tcreated size %u limit %d flags %d\n"), pool->name, (unsigned)pool->size, pool->limit, pool->flags);
    return pool;
}

rpmioItem rpmioUnlinkPoolItem(rpmioItem item,
		const char * msg, const char * fn, unsigned ln)
{
    rpmioPool pool;
    if (item == NULL) return NULL;
    yarnPossess(item->use);
ANNOTATE_HAPPENS_AFTER(item);
    if ((pool = (rpmioPool) item->pool) != NULL && pool->flags && msg) {
	char * imsg;
	char b[BUFSIZ];
	size_t nb = sizeof(b);
	(void) snprintf(b, nb, "--> %s %p -- %ld %s at %s:%u", pool->name,
			item, yarnPeekLock(item->use), msg, fn, ln);
	imsg = (pool->dbug ? (*pool->dbug)((void *)item, b, nb) : b);
	if (imsg) {
	    fprintf(stderr, "%s\n", imsg);
	    if (imsg != b) free(imsg);
	}
    }
    yarnTwist(item->use, BY, -1);
    return item;
}

rpmioItem rpmioLinkPoolItem(rpmioItem item,
		const char * msg, const char * fn, unsigned ln)
{
    rpmioPool pool;
    if (item == NULL) return NULL;
    yarnPossess(item->use);
    if ((pool = (rpmioPool) item->pool) != NULL && pool->flags && msg != NULL) {
	char * imsg;
	char b[BUFSIZ];
	size_t nb = sizeof(b);
	(void) snprintf(b, nb, "--> %s %p ++ %ld %s at %s:%u", pool->name,
			item, yarnPeekLock(item->use)+1, msg, fn, ln);
	imsg = (pool->dbug ? (*pool->dbug)((void *)item, b, nb) : b);
	if (imsg) {
	    fprintf(stderr, "%s\n", imsg);
	    if (imsg != b) free(imsg);
	}
    }
ANNOTATE_HAPPENS_BEFORE(item);
    yarnTwist(item->use, BY, 1);
    return item;
}

void * rpmioFreePoolItem(rpmioItem item,
                const char * msg, const char * fn, unsigned ln)
{
    rpmioPool pool;
    if (item == NULL) return NULL;

#ifdef	NOTYET
assert(item->pool != NULL);	/* XXX (*pool->fini) is likely necessary */
#endif
    yarnPossess(item->use);
ANNOTATE_HAPPENS_AFTER(item);
    if ((pool = (rpmioPool) item->pool) != NULL && pool->flags && msg != NULL) {
	char * imsg;
	char b[BUFSIZ];
	size_t nb = sizeof(b);
	(void) snprintf(b, nb, "--> %s %p -- %ld %s at %s:%u", pool->name,
			item, yarnPeekLock(item->use), msg, fn, ln);
	imsg = (pool->dbug ? (*pool->dbug)((void *)item, b, nb) : b);
	if (imsg) {
	    fprintf(stderr, "%s\n", imsg);
	    if (imsg != b) free(imsg);
	}
    }
    if (yarnPeekLock(item->use) <= 1L) {
VALGRIND_HG_CLEAN_MEMORY(item, pool->size);
	if (pool->fini)
	    (*pool->fini) ((void *)item);
	VALGRIND_MEMPOOL_FREE(pool, item + 1);
	item = rpmioPutPool(item);
    } else
	yarnTwist(item->use, BY, -1);
    return (void *) item;
}

rpmioItem rpmioGetPool(rpmioPool pool, size_t size)
{
    rpmioItem item;

    if (pool) {
	/* if can't create any more, wait for a space to show up */
	yarnPossess(pool->have);
	if (pool->limit == 0)
	    yarnWaitFor(pool->have, NOT_TO_BE, 0);

	/* if a space is available, pull it from the list and return it */
	if (pool->head) {
	    item = pool->head;
	    pool->head = (rpmioItem) item->pool;	/* XXX pool == next */
	    if (pool->head == NULL)
		pool->tail = &pool->head;
	    pool->reused++;
	    item->pool = pool;		/* remember the pool this belongs to */
	    yarnTwist(pool->have, BY, -1);      /* one less in pool */
	    VALGRIND_MEMPOOL_ALLOC(pool,
		item + 1,
		size - sizeof(struct rpmioItem_s));
	    if (pool->init)
		(*pool->init) ((void *)item);
	    return item;
	}

	/* nothing available, don't want to wait, make a new item */
assert(pool->limit != 0);
	if (pool->limit > 0)
	    pool->limit--;
	pool->made++;
	yarnRelease(pool->have);
    }

    item = (rpmioItem) xcalloc(1, size);
    item->use = yarnNewLock(0);		/* XXX newref? */
    item->pool = pool;
    VALGRIND_MEMPOOL_ALLOC(pool,
	item + 1,
	size - sizeof(struct rpmioItem_s));
    if (pool->init)
	(*pool->init) ((void *)item);
    return item;
}

rpmioItem rpmioPutPool(rpmioItem item)
{
    rpmioPool pool;

    if ((pool = (rpmioPool) item->pool) != NULL) {
	yarnPossess(pool->have);
	item->pool = NULL;		/* XXX pool == next */
	*pool->tail = item;
	pool->tail = (rpmioItem *)&item->pool;/* XXX pool == next */
	yarnTwist(pool->have, BY, 1);
	if (item->use)
	    yarnTwist(item->use, TO, 0);
	return NULL;
    }

    if (item->use) {
	yarnTwist(item->use, TO, 0);
	item->use = yarnFreeLock(item->use);
    }
    (void) _free(item);
    return NULL;
}

int rpmioMcheck(void (*abortfunc)())
{
    int rc = -2;
#if defined(HAVE_MCHECK)
    rc = mcheck(abortfunc);
#endif	/* HAVE_MCHECK */
    return rc;
}

int rpmioMcheckPedantic(void (*abortfunc)())
{
    int rc = -2;
#if defined(HAVE_MCHECK_PEDANTIC)
    rc = mcheck_pedantic(abortfunc);
#endif	/* HAVE_MCHECK_PEDANTIC */
    return rc;
}

void rpmioMcheckCheckAll(void)
{
#if defined(HAVE_MCHECK_CHECK_ALL)
    mcheck_check_all();
#endif	/* HAVE_MCHECK_CHECK_ALL */
}

int rpmioMprobe(void *ptr)
{
    int rc = -2;
#if defined(HAVE_MPROBE)
    rc = mprobe(ptr);
#endif	/* HAVE_MPROBE */
    return rc;
}

int rpmioMallopt(int param, int val)
{
    int rc = -2;

    (void)val;
    switch (param) {
    default:
	break;
#if defined(HAVE_MALLOC_INFO)
    case 0:
	rc = malloc_info(param, stderr);
	break;
#endif	/* HAVE_MALLOC_INFO */
#if defined(HAVE_MALLOPT)
    case M_MXFAST:
    case M_NLBLKS:
    case M_GRAIN:
    case M_KEEP:
    case M_TRIM_THRESHOLD:
    case M_TOP_PAD:
    case M_MMAP_THRESHOLD:
    case M_MMAP_MAX:
    case M_CHECK_ACTION:
    case M_PERTURB:
    case M_ARENA_TEST:
    case M_ARENA_MAX:
	rc = mallopt(param, val);
	break;
#endif	/* HAVE_MALLOPT */
    }
    return rc;
}

void rpmioMallocInfo(const char *msg)
{
#if defined(HAVE_MALLINFO)
#define	_MLOG(_item, _desc) \
    rpmlog(RPMLOG_DEBUG, "%10s: %10d | %s\n", #_item, minfo._item, _desc)
    struct mallinfo minfo = mallinfo();	/* structure assignment */
    if (msg == NULL)
	msg = "malloc info:";
    rpmlog(RPMLOG_DEBUG, "======================== %s\n", msg);
    _MLOG(arena,	"non-mmap'ed space allocated from system sbrk()");
    _MLOG(ordblks,	"no. free chunks not in use");
    _MLOG(smblks,	"no. fastbin blocks (unused)");
    _MLOG(hblks,	"no. mmap'ed regions");
    _MLOG(hblkhd,	"space in mmap'ed regions");
    _MLOG(usmblks,	"always 0, backward compatibility (unused)");
    _MLOG(fsmblks,	"space available in free'd fastbin blocks");
    _MLOG(uordblks,	"total allocated space");
    _MLOG(fordblks,	"total free space");
    _MLOG(keepcost,	"top-most, releasable (via malloc_trim) space");
#undef	_MLOG
#endif	/* HAVE_MALLINFO */
}

void rpmioMallocStats(void)
{
#if defined(HAVE_MALLOC_STATS)
    malloc_stats();
#endif
}

int rpmioMallocTrim(size_t pad)
{
    int rc = -2;
    (void)pad;
#if defined(HAVE_MALLOC_TRIM)
    rc = malloc_trim(pad);
#endif
    return rc;
}

size_t rpmioMallocUsableSize(void *ptr)
{
    size_t rc = 0;
    (void)ptr;
#if defined(HAVE_MALLOC_USABLE_SIZE)
    rc = malloc_usable_size(ptr);
#endif
    return rc;
}

void * rpmioMallocGetState(void)
{
    void *ptr = NULL;
#if defined(HAVE_MALLOC_GET_STATE)
    ptr = malloc_get_state();
#endif
    return ptr;
}

int rpmioMallocSetState(void *ptr)
{
    int rc = -2;
    (void)ptr;
#if defined(HAVE_MALLOC_SET_STATE)
    if (ptr)
	rc = malloc_set_state(ptr);
#endif
    return rc;
}

void rpmioMtrace(void)
{
#if defined(HAVE_MTRACE)
    mtrace();	/* Trace malloc only if MALLOC_TRACE=mtrace-output-file. */
#endif
}

void rpmioMuntrace(void)
{
#if defined(HAVE_MUNTRACE)
    muntrace();	/* Trace malloc only if MALLOC_TRACE=mtrace-output-file. */
#endif
}

#if !(HAVE_MCHECK_H && defined(__GNUC__))

void * xmalloc (size_t size)
{
    register void *value;
    if (size == 0) size++;
    value = malloc (size);
    if (value == 0)
	value = vmefail(size);
    return value;
}

void * xcalloc (size_t nmemb, size_t size)
{
    register void *value;
    if (size == 0) size++;
    if (nmemb == 0) nmemb++;
    value = calloc (nmemb, size);
    if (value == 0)
	value = vmefail(size);
    return value;
}

void * xrealloc (void *ptr, size_t size)
{
    register void *value;
    if (size == 0) size++;
    value = realloc (ptr, size);
    if (value == 0)
	value = vmefail(size);
    return value;
}

char * xstrdup (const char *str)
{
    size_t size = strlen(str) + 1;
    char *newstr = (char *) malloc (size);
    if (newstr == 0)
	newstr = (char *) vmefail(size);
    strcpy (newstr, str);
    return newstr;
}

#endif	/* !(HAVE_MCHECK_H && defined(__GNUC__)) */
