/** \ingroup rpmio
 * \file rpmio/rpmmqtt.c
 */

#include "system.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>

#define	_RPMIOB_INTERNAL	/* for rpmiobSlurp */
#include <rpmiotypes.h>
#include <rpmio.h>		/* for *Pool methods */
#include <rpmlog.h>
#include <rpmcb.h>
#include <rpmdir.h>
#define	_MACRO_INTERNAL		/* for struct MacroContext_s */
#include <rpmmacro.h>
#include <rpmurl.h>
#include <poptIO.h>
#include <argv.h>

#ifdef	WITH_PAHO
#include <MQTTAsync.h>
#endif
#ifdef	WITH_MOSQUITTO
#include <mosquitto.h>
#endif
#ifdef	WITH_RABBITMQ
#include <amqp.h>
#include <amqp_framing.h>
#include <amqp_ssl_socket.h>
#include <amqp_tcp_socket.h>
#endif
#ifdef	WITH_PROTON
#include <proton/message.h>
#include <proton/messenger.h>
#ifdef 	DYING
#include "/usr/share/proton-0.14.0/examples/c/include/pncompat/misc_funcs.inc"
#endif
#endif
#ifdef	WITH_ZEROMQ
#include <zmq.h>
#endif

#define	_RPMMQTT_INTERNAL
#include <rpmmqtt.h>

#include "debug.h"

#if 0
int atexit(void (*function)(void))
{
#if 0
    fprintf(stderr, "*** %s(%p)\n", __FUNCTION__, function);
#endif
    return 0;
}
#endif

int _rpmmqtt_debug;

#define F_ISSET(_f, _FLAG) (((_f) & ((MQTT_FLAGS_##_FLAG) & ~0x40000000)) != MQTT_FLAGS_NONE)
#define MF_ISSET(_FLAG) F_ISSET(mqtt->flags, _FLAG)

/*==============================================================*/
#ifndef	DYING
static void dumpU(const char * msg, urlinfo u)
{
    FILE * fp = stderr;
    const char * url = (u ? u->url : NULL);
    const char *path = NULL;
    urltype ut = urlPath(url, &path);;

    if (msg)
	fprintf(fp, "===================================== %s u %p ut %d\n",
		msg, u, ut);
    if (u) {

	fprintf(fp, "     url: %s\n", u->url);
	fprintf(fp, "  scheme: %s\n", u->scheme);
	fprintf(fp, "    user: %s\n", u->user);
	fprintf(fp, "password: %s\n", u->password);
	fprintf(fp, "    host: %s\n", u->host);
	fprintf(fp, " portstr: %s\n", u->portstr);
	fprintf(fp, "    path: %s\n", path);
	fprintf(fp, "   query: %s\n", u->query);
	fprintf(fp, "fragment: %s\n", u->fragment);
	if (u->query) {
	    ARGV_t av = NULL;
	    (void) argvSplit(&av, u->query, ",");
	    argvPrint(u->query, av, fp);
	    av = argvFree(av);
	}
    }
}
#endif

/*==============================================================*/
typedef struct key_s {
    int         v;
    const char *n;
} KEY;

static const char * tblName(uint32_t v, KEY * tbl, size_t ntbl)
{
    const char * n = NULL;
    static char buf[32];
    size_t i;

    for (i = 0; i < ntbl; i++) {
uint32_t tbl_v = tbl[i].v & 0x3fffffff;
	if (v != tbl_v)
	    continue;
	n = tbl[i].n;
	break;
    }
    if (n == NULL) {
	(void) snprintf(buf, sizeof(buf), "0x%x", (unsigned)v);
	n = buf;
    }
    return n;
}

static const char * fmtBits(uint32_t flags, KEY tbl[], size_t ntbl, char *t)
{
    char pre = '<';
    char * te = t;
    int i;

flags &= 0x3fffffff;
    sprintf(t, "0x%x", (unsigned)flags);
    te = t;
    te += strlen(te);
    for (i = 0; i < 32; i++) {
	uint32_t mask = (1 << i);
	const char * name;

	if (!(flags & mask))
	    continue;

	name = tblName(mask, tbl, ntbl);
	*te++ = pre;
	pre = ',';
	te = stpcpy(te, name);
    }
    if (pre == ',') *te++ = '>';
    *te = '\0';
    return t;
}

static KEY MqttFlags[] = {
#define _ENTRY(_v)      { MQTT_FLAGS_##_v, #_v, }
    _ENTRY(CLEAN),
    _ENTRY(EOL),
    _ENTRY(NOSTALE),
    _ENTRY(LINES),
    _ENTRY(EMPTY),
    _ENTRY(DNSSRV),
    _ENTRY(INSECURE),
    _ENTRY(RETAIN),
    _ENTRY(WILL_RETAIN),
    _ENTRY(BUFFER),
    _ENTRY(DURABLE),
    _ENTRY(EXCLUSIVE),
    _ENTRY(NOACK),
    _ENTRY(IFEMPTY),
    _ENTRY(IFUNUSED),
    _ENTRY(PASSIVE),
    _ENTRY(AUTODELETE),
    _ENTRY(NOLOCAL),
#undef	_ENTRY
};
static size_t nMqttFlags = sizeof(MqttFlags) / sizeof(MqttFlags[0]);

static const char * fmtMqttFlags(uint32_t flags)
{
    static char buf[BUFSIZ];
    char * te = buf;
    (void) fmtBits(flags, MqttFlags, nMqttFlags, te);
    return buf;
}
#define	_MQTTFLAGS(_flags)	fmtMqttFlags(_flags)

static const char * rpmmqttStrerror(rpmmqtt mqtt, int v)
{
    KEY * tbl = (KEY *) (mqtt && mqtt->vec ? mqtt->vec->errs : NULL);
    size_t ntbl = (mqtt && mqtt->vec && mqtt->vec->errs ? mqtt->vec->nerrs : 0);
    const char * n = NULL;
    static char buf[64];

    for (size_t i = 0; i < ntbl; i++) {
        if (v != tbl[i].v)
            continue;
        n = tbl[i].n;
        break;
    }
    if (n == NULL) {
        (void) snprintf(buf, sizeof(buf), "0x%x", v);
        n = buf;
    }
    return n;
}

static rpmRC Xmqchk(rpmmqtt mqtt, const char * msg, int rc,
		int printit, const char * func, const char * fn, unsigned ln)
{
    int all = (mqtt && mqtt->vec ? mqtt->vec->all : 1);

    if (all || rc != 0) {
	int _lvl = RPMLOG_WARNING;
	const char * api = (mqtt && mqtt->vec && mqtt->vec->prefix
		? mqtt->vec->prefix : "");
	rpmlog(_lvl, "%s:%s:%u: %s%s: %s(%d)\n",
		func, fn, ln, api, msg, rpmmqttStrerror(mqtt, rc), rc);
    }
    return (rpmRC) rc;
}
#define mqchk(_o, _m, _rc)  \
    Xmqchk(_o, _m, _rc, _rpmmqtt_debug, __FUNCTION__, __FILE__, __LINE__)

/*==============================================================*/
static struct MacroContext_s _rpmmqttMacroContext;
MacroContext rpmmqttMacroContext = &_rpmmqttMacroContext;

void rpmmqttFreeMacros(rpmmqtt mqtt)
{
    if (mqtt)
	rpmFreeMacros(mqtt->mc);
}

int rpmmqttDefineMacro(rpmmqtt mqtt, const char * macro, int level)
{
    return rpmDefineMacro((mqtt ? mqtt->mc : NULL), macro, level);
}

char * rpmmqttExpand(rpmmqtt mqtt, char **lhs, const char *arg, ...)
{
    char *t = NULL;
    va_list ap;

    if (lhs && *lhs) {
	free(*lhs);
	*lhs = NULL;
    }

    va_start(ap, arg);
    t = vrpmMCExpand((mqtt ? mqtt->mc : NULL), arg, ap);
    va_end(ap);

    if (lhs)
	*lhs = t;
    return t;
}

int rpmmqttExpandNumeric(rpmmqtt mqtt, const char *arg)
{
    return rpmMCExpandNumeric((mqtt ? mqtt->mc : NULL), arg);
}

/*==============================================================*/

static void dumpMQTT(const char * msg, rpmmqtt mqtt)
{
    FILE * fp = stderr;

    if (msg)
	fprintf(fp, "===================================== %s(%p)\n",
		msg, mqtt);
    if (mqtt) {

	if (mqtt->flags != 0x40000003)
	    fprintf(fp, "%19s: %s\n", "flags", _MQTTFLAGS(mqtt->flags));
#define	PRINT(_fmt, _foo) fprintf(fp, "%19s: %"#_fmt"\n", #_foo, mqtt->_foo)
	if (mqtt->msg_input != 0)
	    PRINT(d, msg_input);
	if (mqtt->msg_output != 1)
	    PRINT(d, msg_output);
	if (mqtt->av)
	    argvPrint("mqtt->av", (ARGV_t)mqtt->av, fp);

	if (mqtt->_progname)
	    PRINT(s, _progname);
	if (mqtt->_progmode)
	    PRINT(s, _progmode);

	if (mqtt->_address)
	    PRINT(s, _address);
	if (mqtt->msgs)
	    argvPrint("mqtt->msgs", (ARGV_t)mqtt->msgs, fp);

	if (mqtt->cacertfile)
	    PRINT(s, cacertfile);
	if (mqtt->_capath)
	    PRINT(s, _capath);
	if (mqtt->certfile)
	    PRINT(s, certfile);
	if (mqtt->keyfile)
	    PRINT(s, keyfile);
	if (mqtt->ciphers)
	    PRINT(s, ciphers);

	if (mqtt->_tls_version == NULL || strcmp(mqtt->_tls_version, "1.2"))
	    PRINT(s, _tls_version);
	if (mqtt->_psk_key)
	    PRINT(s, _psk_key);
	if (mqtt->_psk_identity)
	    PRINT(s, _psk_identity);
	if (mqtt->_proxy)
	    PRINT(s, _proxy);

	if (mqtt->keepalive != 60)
	    PRINT(d, keepalive);
	if (mqtt->max_inflight != 20)
	    PRINT(d, max_inflight);
	if (mqtt->protocol_version == NULL
	 || strcmp(mqtt->protocol_version, "auto"))
	    PRINT(s, protocol_version);

	if (mqtt->will_topic) {
	    PRINT(s, will_message);
	    PRINT(d, will_qos);
	    PRINT(s, will_topic);
	}

	if (mqtt->user)
	    PRINT(s, user);
	if (mqtt->password)
	    PRINT(s, password);
	if (mqtt->host == NULL || strcmp(mqtt->host, "localhost"))
	    PRINT(s, host);
	if (mqtt->port != 1883)
	    PRINT(d, port);
	if (mqtt->uri == NULL || strcmp(mqtt->uri, "tcp://localhost:1883"))
	    PRINT(s, uri);

	if (mqtt->idprefix == NULL || strcmp(mqtt->idprefix, "rpm"))
	    PRINT(s, idprefix);
	if (mqtt->_clientid == NULL || strcmp(mqtt->_clientid, "rpm-%{pid}"))
	    PRINT(s, _clientid);
	if (mqtt->clientid == NULL || strncmp(mqtt->_clientid, "rpm-", 4))
	    PRINT(s, clientid);

	if (mqtt->qos != 2)
	    PRINT(u, qos);
	if (mqtt->timeout != 10000)
	    PRINT(u, timeout);

	if (mqtt->topics)
	    argvPrint("mqtt->topics", (ARGV_t)mqtt->topics, fp);
	if (mqtt->subtopics)
	    argvPrint("mqtt->subtopics", (ARGV_t)mqtt->subtopics, fp);
	if (mqtt->_filter_out)
	    argvPrint("mqtt->_filter_out", (ARGV_t)mqtt->_filter_out, fp);
	if (mqtt->topic == NULL || strcmp(mqtt->topic, "rpm/mqtt"))
	    PRINT(s, topic);

	if (mqtt->u) {
	    PRINT(d, ut);
	    dumpU("mqtt->u", mqtt->u);
	}

	/* -- AMQP options: */
	if (mqtt->vhost)
	    PRINT(s, vhost);
	if (mqtt->queue)
	    PRINT(s, queue);
	if (mqtt->exchange)
	    PRINT(s, exchange);
	if (mqtt->_routing_key)
	    PRINT(s, _routing_key);
	if (mqtt->_reply_to)
	    PRINT(s, _reply_to);
	if (mqtt->_content_type)
	    PRINT(s, _content_type);
	if (mqtt->_content_encoding)
	    PRINT(s, _content_encoding);
	if (mqtt->_delivery)
	    PRINT(d, _delivery);
	if (mqtt->_prefetch_count)
	    PRINT(d, _prefetch_count);

	if (mqtt->_verify)
	    PRINT(d, _verify);
	if (mqtt->_fail_if_no_peer_cert)
	    PRINT(d, _fail_if_no_peer_cert);
	if (mqtt->_connection_timeout)
	    PRINT(d, _connection_timeout);
	if (mqtt->_channel_max)
	    PRINT(d, _channel_max);
	if (mqtt->_sasl_mechanism)
	    PRINT(s, _sasl_mechanism);

	if (mqtt->I)
	    PRINT(p, I);
	if (mqtt->persist_type != 2) {
	    PRINT(u, persist_type);
	    PRINT(p, persist_ctx);
	    PRINT(s, persist_path);
	}

	if (mqtt->debug)
		PRINT(d, debug);
	if (mqtt->trace)
		PRINT(d, trace);
	if (mqtt->finished)
		PRINT(d, finished);
	if (mqtt->token)
		PRINT(u, token);

	if (mqtt->msg_count) {
	    PRINT(d, msg_count);
	    PRINT(d, max_msg_count);
	}

	if (mqtt->serverURI) {
	    PRINT(s, serverURI);
	    PRINT(d, MQTTVersion);
	    PRINT(d, sessionPresent);
	}

	if (mqtt->cachedn)
	    PRINT(s, cachedn);

	if (mqtt->ifn)
	    argvPrint("mqtt->ifn", (ARGV_t)mqtt->ifn, fp);
	if (mqtt->ifd)
	    PRINT(p, ifd);
	if (mqtt->ofn == NULL || strcmp(mqtt->ofn, "-")) {
	    PRINT(s, ofn);
	    if (mqtt->ofd)
		PRINT(p, ofd);
	}
	if (mqtt->iob)
	    PRINT(p, iob);

#undef	PRINT
    }
}

/*==============================================================*/
static
rpmRC _rpmmqttSubscribeMany(rpmmqtt mqtt, int ac, char ** av)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (ac <= 0)
	goto exit;
    if (!rpmmqttConnect(mqtt)) {
	int *subqos = (int *) xcalloc(ac, sizeof(*subqos));
	for (int i = 0; i < ac; i++) {
	    char * t = av[i];
#ifdef	NOTYET		/* XXX qos as subtopic ?qos=N? */
	    char * te = strchr(t, '?');
	    if (te) {
		*te++ = '\0';
		if ((te = strchr(te, '='))) {
		    if (!strncmp(t, "qos", (te - t)))
			subqos[i] = strtoul(te+1, NULL, 0);
		}
	    } else
#endif
		subqos[i] = mqtt->qos;	/* XXX */
	    rc = rpmmqttSubscribe(mqtt, t, subqos[i]);
	    if (rc)
		break;
	}

	if (subqos)
	    free(subqos);
	if (rc)
	    goto exit;
	rc = RPMRC_OK;
    }

exit:
    return rc;
}

static
rpmRC _rpmmqttUnsubscribeMany(rpmmqtt mqtt, int ac, char ** av)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (ac <= 0)
	goto exit;
    if (!rpmmqttConnect(mqtt)) {
	for (int i = 0; i < ac; i++) {
	    rc = rpmmqttUnsubscribe(mqtt, av[i]);
	    if (rc)
		goto exit;
	}
    }

exit:
    return rc;
}

/*==============================================================*/
#ifdef	WITH_PAHO
static int pahoOnMessageArrived(void * _mqtt, char * topic, int topicLen,
		MQTTAsync_message *  message)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
    int rc = 1;

    mqtt->msg_count++;

    const char * s = (const char *) message->payload;
    size_t ns = message->payloadlen;

    rpmlog(RPMLOG_DEBUG, "%s rcvd topic(%s) \"%.*s\"\n", mqtt->vec->name,
			topic, (int)ns, s);

    if (mqtt->iob) {
	mqtt->iob = rpmiobAppend(mqtt->iob, topic, 0);
	mqtt->iob = rpmiobAppend(mqtt->iob, ":\t", 0);
	{   char * t = (char *) memcpy(alloca(ns+1), s, ns);
	    t[ns] = '\0';
	    mqtt->iob = rpmiobAppend(mqtt->iob, t, 1);
	}
    }
    switch (mqtt->msg_output) {
    case MQTT_OUTPUT_UNKNOWN:
    case MQTT_OUTPUT_CALLBACK:
    default:
	break;
    case MQTT_OUTPUT_STDOUT:
    case MQTT_OUTPUT_FILE:
	if (mqtt->ofd) {
	    size_t nw;
	    nw = Fwrite(topic, sizeof(*topic), topicLen, mqtt->ofd);
assert(nw > 0);
	    nw = Fwrite(":\t", 1, sizeof(":\t")-1, mqtt->ofd);
assert(nw > 0);
	    nw = Fwrite(s, sizeof(*s), ns, mqtt->ofd);
assert(nw > 0);
	    nw = Fwrite("\n", 1, sizeof("\n")-1, mqtt->ofd);
assert(nw > 0);
	}
	break;
    }
#ifdef	NEW
    message->qos
    message->retained
    message->dup
    message->msgid
#endif

    MQTTAsync_freeMessage(&message);
    MQTTAsync_free(topic);

    return rc;
}

static void pahoOnDeliveryComplete(void * _mqtt, int token)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
    mqtt->token = token;
}

static void pahoOnConnectionLost(void * _mqtt, char *cause)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;

    rpmlog(RPMLOG_DEBUG,
		"%s disconnect(%s) version(%d) present(%d)\n", mqtt->vec->name,
		mqtt->serverURI, mqtt->MQTTVersion, mqtt->sessionPresent);
    if (cause)
	rpmlog(RPMLOG_DEBUG, "\tcause: %s\n", cause);

    rpmlog(RPMLOG_WARNING, "%s reconnecting(%s) ...\n", mqtt->vec->name,
			mqtt->serverURI);

    (void) rpmmqttConnect(mqtt);
}

static void pahoOnFailure(void * _mqtt, MQTTAsync_failureData * response)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
    rpmlog(RPMLOG_WARNING, "%s cmd failed.\n", mqtt->vec->name);
    mqtt->finished = 1;
}

static void pahoOnSuccess(void * _mqtt, MQTTAsync_successData * response)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
    mqtt->finished = 1;
}

static void pahoOnDisconnectFailure(void * _mqtt, MQTTAsync_failureData * response)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
    rpmlog(RPMLOG_WARNING, "%s disconnect failed\n", mqtt->vec->name);
    mqtt->finished = 1;
}

static void pahoOnDisconnect(void * _mqtt, MQTTAsync_successData * response)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;

    rpmlog(RPMLOG_DEBUG,
		"%s disconnect(%s) version(%d) present(%d)\n", mqtt->vec->name,
		mqtt->serverURI, mqtt->MQTTVersion, mqtt->sessionPresent);
    mqtt->serverURI = _free(mqtt->serverURI);
    mqtt->finished = 1;
}

static void pahoOnConnectFailure(void * _mqtt, MQTTAsync_failureData * response)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
    rpmlog(RPMLOG_WARNING, "%s connect failed\n", mqtt->vec->name);
    mqtt->finished = 1;
}

static void pahoOnConnect(void * _mqtt, MQTTAsync_successData * response)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;

    if (response) {
	mqtt->token = response->token;
	mqtt->serverURI = xstrdup(response->alt.connect.serverURI);
	mqtt->MQTTVersion = response->alt.connect.MQTTVersion;
	mqtt->sessionPresent = response->alt.connect.sessionPresent;
    }

    rpmlog(RPMLOG_DEBUG,
		"%s connect(%s) version(%d) present(%d)\n", mqtt->vec->name,
		mqtt->serverURI, mqtt->MQTTVersion, mqtt->sessionPresent);
    mqtt->finished = 1;
}

static void pahoOnSubscribeFailure(void * _mqtt, MQTTAsync_failureData * response)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
    if (response) {
	const char *s = response->message;
	int token = response->token;
	int code = response->code;
	rpmlog(RPMLOG_WARNING, "%s subscribe(%d) failed: code(%d) msg %s\n",
			mqtt->vec->name, token, code, s);
    } else
	rpmlog(RPMLOG_WARNING, "%s subscribe failed\n", mqtt->vec->name);
    mqtt->finished = 1;
}

static void pahoOnSubscribe(void * _mqtt, MQTTAsync_successData * response)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
    int qos = response->alt.qos;

    rpmlog(RPMLOG_DEBUG, "%s subscribe qos(%d)\n", mqtt->vec->name, qos);
    mqtt->finished = 1;
}

static void pahoOnSubscribeManyFailure(void * _mqtt, MQTTAsync_failureData * response)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;

    rpmlog(RPMLOG_WARNING, "%s subscribeMany failed\n", mqtt->vec->name);
    mqtt->finished = 1;
}

static void pahoOnSubscribeMany(void * _mqtt, MQTTAsync_successData * response)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;

#ifdef	DYING
    int *subqos = response->alt.qosList;
    for (int i = 0; i < mqtt->ac; i++) {
	rpmlog(RPMLOG_DEBUG, "%s subscribe qos(%d)\n", mqtt->vec->name, subqos[i]);
    }
#endif

    mqtt->finished = 1;
}

static void pahoOnUnsubscribeManyFailure(void * _mqtt, MQTTAsync_failureData * response)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;

    rpmlog(RPMLOG_WARNING, "%s unsubscribeMany failed\n", mqtt->vec->name);
    mqtt->finished = 1;
}

static void pahoOnUnsubscribeMany(void * _mqtt, MQTTAsync_successData * response)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;

    mqtt->finished = 1;
}

static void pahoOnSendFailure(void * _mqtt, MQTTAsync_failureData * response)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;

    {
	const char *s = response->message;
	int token = response->token;
	int code = response->code;
	rpmlog(RPMLOG_WARNING, "%s send(%d) failed: code(%d) msg %s\n",
			mqtt->vec->name, token, code, s);
    }
    mqtt->finished = 1;
}

static void pahoOnSend(void * _mqtt, MQTTAsync_successData * response)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;

    {
	const char * s = (const char *) response->alt.pub.message.payload;
	size_t ns = response->alt.pub.message.payloadlen;
	int token = response->token;
	rpmlog(RPMLOG_DEBUG, "%s sent(%d) topic(%s) \"%.*s\"\n",
			mqtt->vec->name, token, mqtt->topic, (int)ns, s);
    }
    mqtt->finished = 1;
}

static void onTrace(enum MQTTASYNC_TRACE_LEVELS level, char *message)
{
    rpmlog(RPMLOG_DEBUG, "%s\n", message);
}
#endif	/* WITH_PAHO */

/*==============================================================*/
#ifdef	WITH_PAHO
static int pahoOpen(void **_mqttp, const char *clientID, const char *serverURI,
		void *_mqtt)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
    char * dn;
    char * te;
    int rc = MQTTCLIENT_PERSISTENCE_ERROR;

    *_mqttp = _mqtt;

    if (mqtt->cachedn) {
	free((void *)mqtt->cachedn);
	mqtt->cachedn = NULL;
    }
    dn = rpmGetPath(mqtt->persist_path, "/", clientID, "-", serverURI, NULL);
    for (te = dn; (te = strchr(te, ':')) != NULL; te++)
	*te = '-';

    if (rpmioMkpath(dn, (mode_t)0740, (uid_t)-1, (gid_t)-1)) {
	if (dn) free(dn);
	goto exit;
    }
    mqtt->cachedn = dn;

    rc = 0;

exit:
SPEW((stderr, "<--    %s(%p,\"%s\",\"%s\",%p) rc %d dn %s\n", __FUNCTION__, _mqttp, clientID, serverURI, _mqtt, rc, mqtt->cachedn));

    return rc;
}

static int pahoClose(void *_mqtt)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
    int rc = MQTTCLIENT_PERSISTENCE_ERROR;


    if (mqtt->cachedn == NULL)
	goto exit;

    if (Rmdir(mqtt->cachedn) && (errno != ENOENT && errno != ENOTEMPTY))
	goto exit;

    rpmlog(RPMLOG_DEBUG, D_("removed directory %s\n"), mqtt->cachedn);

    rc = 0;

exit:
SPEW((stderr, "<--    %s(%p) rc %d\n", __FUNCTION__, _mqtt, rc));
    return rc;
}

static int pahoPut(void *_mqtt, char *key,
		int bufcount, char *buffers[], int buflens[])
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
    char *fn = NULL;
    FD_t fd = NULL;
    size_t nb = 0;
    size_t nw = 0;
    int rc = MQTTCLIENT_PERSISTENCE_ERROR;

    if (mqtt->cachedn == NULL)
	goto exit;

    /* XXX add .msg extension? */
    fn = rpmGetPath(mqtt->cachedn, "/", key, NULL);
    fd = Fopen(fn, "wb");
    if (fd == NULL || Ferror(fd))
	goto exit;

    for (int i = 0; i < bufcount; i++) {
	const char * s = buffers[i];
	int ns = buflens[i];
	int j;
	for (j = 0; j < ns; j++) {
	    if (xisprint(s[j]))
		continue;
	    break;
	}
	if (i > 0 && (j == ns || (j == ns-1 && s[j] == '\0'))) {
SPEW((stderr, "%5d\t%p[%d]\t\"%.*s\"\n", i, s, ns, ns, s));
	} else {
	    if (ns <= 8) {
		unsigned long val = 0;
		j = ns;
		while (--j >= 0) {
		    val <<= 8;
		    val |= s[j];
		}
SPEW((stderr, "%5d\t%p[%d]\t0x%lx\n", i, s, ns, val));
	    } else {
SPEW((stderr, "%5d\t%p[%d]\n", i, s, ns));
	    }
	}
	nb += buflens[i];
	nw += Fwrite(buffers[i], sizeof(*buffers[i]), buflens[i], fd);
    }
    (void) fdatasync(Fileno(fd));
    (void) Fclose(fd);
    fd = NULL;

    if (nb != nw) {
	(void) Unlink(fn);
	goto exit;
    }

    rc = 0;

exit:
    if (fd)
	Fclose(fd);
SPEW((stderr, "<--    %s(%p,\"%s\",%d,%p,%p) rc %d fn %s\n", __FUNCTION__, _mqtt, key, bufcount, buffers, buflens, rc, fn));
    if (fn) free(fn);
    return rc;
}

static int pahoGet(void *_mqtt, char *key, char *buffer[], int *buflen)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
    char *fn = NULL;
    FD_t fd = NULL;
    size_t nr = 0;
    char *b = NULL;
    size_t nb = 0;
    struct stat sb;
    int rc = MQTTCLIENT_PERSISTENCE_ERROR;

    if (mqtt->cachedn == NULL)
	goto exit;

    /* XXX add .msg extension? */
    fn = rpmGetPath(mqtt->cachedn, "/", key, NULL);
    fd = Fopen(fn, "rb");
    if (fd == NULL || Ferror(fd))
	goto exit;

    if (Fstat(fd, &sb) != 0)
	goto exit;
    nb = sb.st_size;
    b = (char *) xmalloc(nb);
    nr = Fread(b, sizeof(*b), nb, fd);
    (void) Fclose(fd);
    fd = NULL;

    if (nb != nr) {
	if (b) free(b);
	nb = 0;
	goto exit;
    }

    rc = 0;

exit:
    if (fd)
	Fclose(fd);
SPEW((stderr, "<--    %s(%p,\"%s\",%p,%p) rc %d fn %s\n", __FUNCTION__, _mqtt, key, buffer, buflen, rc, fn));
    if (fn) free(fn);
    *buffer = b;
    *buflen = nb;
    return rc;
}

static int pahoRemove(void *_mqtt, char *key)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
    char *fn = NULL;
    int rc = MQTTCLIENT_PERSISTENCE_ERROR;

    if (mqtt->cachedn == NULL)
	goto exit;

    /* XXX add .msg extension? */
    fn = rpmGetPath(mqtt->cachedn, "/", key, NULL);

    if (Unlink(fn) && errno != ENOENT)
	goto exit;
    rc = 0;

exit:
SPEW((stderr, "<--    %s(%p,\"%s\") rc %d fn %s\n", __FUNCTION__, _mqtt, key, rc, fn));
    if (fn) free(fn);
    return rc;
}

static int pahoKeys(void *_mqtt, char ***keys, int *nkeys)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
    ARGV_t av = NULL;
    int ac = 0;
    DIR * dir = NULL;
    struct dirent *dp;
    int rc = MQTTCLIENT_PERSISTENCE_ERROR;

    if (mqtt->cachedn == NULL)
	goto exit;

    if ((dir = Opendir(mqtt->cachedn)) == NULL)
	goto exit;

    while ((dp = Readdir(dir)) != NULL) {
	char * fn = rpmGetPath(mqtt->cachedn, "/", dp->d_name, NULL);
	struct stat sb;

	if (Lstat(fn, &sb) == 0 && S_ISREG(sb.st_mode))
	    (void) argvAdd(&av, dp->d_name);
	/* XXX strip .msg extension? */
	if (fn) free(fn);
    }
    ac = argvCount(av);
    if (ac == 0)
	av = argvFree(av);
    rc = 0;

exit:
    if (dir)
	(void) Closedir(dir);
SPEW((stderr, "<--    %s(%p,%p,%p) rc %d keys %p[%u]\n", __FUNCTION__, _mqtt, keys, nkeys, rc, av, ac));
    *keys = (char **) av;
    *nkeys = ac;
    return rc;
}

static int pahoClear(void *_mqtt)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
    DIR * dir = NULL;
    struct dirent *dp;
    int nerrs = 0;
    int rc = MQTTCLIENT_PERSISTENCE_ERROR;

    if (mqtt->cachedn == NULL)
	goto exit;

    if ((dir = Opendir(mqtt->cachedn)) == NULL)
	goto exit;

    while ((dp = Readdir(dir)) != NULL) {
	char * fn = rpmGetPath(mqtt->cachedn, "/", dp->d_name, NULL);
	struct stat sb;

	if (Lstat(fn, &sb) == 0 && S_ISREG(sb.st_mode)) {
	    if (Unlink(fn))
		nerrs++;
	}
	if (fn) free(fn);
    }

    if (nerrs == 0)
	rc = 0;

exit:
    if (dir)
	(void) Closedir(dir);
SPEW((stderr, "<--    %s(%p) rc %d\n", __FUNCTION__, _mqtt, rc));
    return rc;
}

static int pahoContainsKey(void *_mqtt, char *key)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
    DIR * dir = NULL;
    struct dirent *dp;
    int rc = MQTTCLIENT_PERSISTENCE_ERROR;

    if (mqtt->cachedn == NULL)
	goto exit;

    if ((dir = Opendir(mqtt->cachedn)) == NULL)
	goto exit;

    while ((dp = Readdir(dir)) != NULL) {
	char * fn = rpmGetPath(mqtt->cachedn, "/", dp->d_name, NULL);
	struct stat sb;

	if (Lstat(fn, &sb) == 0 && S_ISREG(sb.st_mode)) {
	    /* XXX strip .msg extension? */
	    if (!strcmp(dp->d_name, key)) {
		if (fn) free(fn);
		rc = 0;
		goto exit;
	    }
	}
	if (fn) free(fn);
    }

exit:
    if (dir)
	(void) Closedir(dir);
SPEW((stderr, "<--    %s(%p,\"%s\") rc %d\n", __FUNCTION__, _mqtt, key, rc));
    return rc;
}

static MQTTClient_persistence _mqtt_persistence = {
    NULL,
    pahoOpen,
    pahoClose,
    pahoPut,
    pahoGet,
    pahoRemove,
    pahoKeys,
    pahoClear,
    pahoContainsKey,
};

#endif	/* WITH_PAHO */

/*==============================================================*/

#ifdef	WITH_PAHO
struct pahoState_s {
    MQTTAsync_connectOptions	C;
    MQTTAsync_willOptions	W;
    MQTTAsync_SSLOptions	S;
    MQTTAsync_disconnectOptions	D;
    MQTTAsync_responseOptions	R;
    MQTTAsync_createOptions	O;
    MQTTAsync_message 		M;
};

static struct pahoState_s pahoStateInitial = {
    .C = MQTTAsync_connectOptions_initializer,
    .W = MQTTAsync_willOptions_initializer,
    .S = MQTTAsync_SSLOptions_initializer,
    .D = MQTTAsync_disconnectOptions_initializer,
    .R = MQTTAsync_responseOptions_initializer,
    .O = MQTTAsync_createOptions_initializer,
    .M = MQTTAsync_message_initializer,
};

static
void * AOBJ(rpmmqtt mqtt, char otype)
{
    void * ptr = NULL;
    pahoState_t p = NULL;
    pahoState_t q = &pahoStateInitial;
    MQTTAsync_connectOptions *C;
    MQTTAsync_disconnectOptions *D;
    MQTTAsync_message *M;
    MQTTAsync_createOptions *O;
    MQTTAsync_responseOptions *R;
    MQTTAsync_SSLOptions *S;
    MQTTAsync_willOptions *W;
    urlinfo u;

    if (mqtt->state == NULL)
	mqtt->state = (pahoState_t) xcalloc(1, sizeof(*p));
    p = (pahoState_t) mqtt->state;

    switch (otype) {
    case 'C':
	u = mqtt->u;
	C = &p->C;
	memcpy(C, &q->C, sizeof(*C));
	C->keepAliveInterval = mqtt->keepalive;
	C->cleansession = (MF_ISSET(CLEAN) ? 1 : 0);
	C->maxInflight = mqtt->max_inflight;

	W = (MQTTAsync_willOptions *) AOBJ(mqtt, 'W');		/* XXX LWT */
	C->will = (W && W->topicName ? W : NULL);

	C->username = (u && u->user ? u->user : mqtt->user);
	C->password = (u && u->password ? u->password : mqtt->password);

	C->connectTimeout = 30;	/* XXX secs. configure?*/
	C->retryInterval = 0;	/* XXX secs. configure? */

	S = (MQTTAsync_SSLOptions *) AOBJ(mqtt, 'S');		/* XXX SSL */
	C->ssl = (S && S->keyStore) ? S : NULL;	/* XXX */

	C->onSuccess = pahoOnConnect;
	C->onFailure = pahoOnConnectFailure;
	C->context = mqtt;

	C->serverURIcount = 0;
	C->serverURIs = NULL;

	C->MQTTVersion = 0;
	if (mqtt->protocol_version) {
	    if (!strcmp(mqtt->protocol_version, "auto"))
		C->MQTTVersion = 0;
	    if (!strcmp(mqtt->protocol_version, "31"))
		C->MQTTVersion = 3;
	    if (!strcmp(mqtt->protocol_version, "311"))
		C->MQTTVersion = 4;
	}
	C->automaticReconnect = 1;
	C->minRetryInterval = 1;	/* secs */
	C->maxRetryInterval = 60;	/* secs */
	ptr = (void *) C;
	break;
    case 'D':
	D = &p->D;
	memcpy(D, &q->D, sizeof(*D));
	D->timeout = mqtt->timeout;
	D->onSuccess = pahoOnDisconnect;
	D->onFailure = pahoOnDisconnectFailure;
	D->context = mqtt;
	ptr = (void *) D;
	break;
    case 'M':
	M = &p->M;
	memcpy(M, &q->M, sizeof(*M));
	M->payloadlen = 0;
	M->payload = (void *) "";
	M->qos = mqtt->qos;
	M->retained = (MF_ISSET(RETAIN) ? 1 : 0);
	M->dup = 0;
	M->msgid = 0;
	ptr = (void *) M;
	break;
    case 'O':
	O = &p->O;
	memcpy(O, &q->O, sizeof(*O));
	O->sendWhileDisconnected = 1;		/* XXX = 0 */
	O->maxBufferedMessages = 100;
	ptr = (void *) O;
	break;
    case 'R':
	R = &p->R;
	memcpy(R, &q->R, sizeof(*R));
	R->onSuccess = NULL;			/* XXX multiple uses */
	R->onFailure = NULL;			/* XXX multiple uses */
	R->context = mqtt;
	R->token = 0;
	ptr = (void *) R;
	break;
    case 'S':
	S = &p->S;
	memcpy(S, &q->S, sizeof(*S));
	S->trustStore = mqtt->cacertfile;
	S->keyStore = mqtt->certfile;
	S->privateKey = mqtt->keyfile;
	S->enabledCipherSuites = mqtt->ciphers;
	S->enableServerCertAuth = (MF_ISSET(INSECURE) ? 0 : 1);
	ptr = (void *) S;
	break;
    case 'W':
	W = &p->W;
	memcpy(W, &q->W, sizeof(*W));
	W->topicName = (mqtt->will_topic ? mqtt->will_topic : mqtt->topic);
	W->message = (mqtt->will_message ? mqtt->will_message : "");
	W->retained = (MF_ISSET(WILL_RETAIN) ? 1 : 0);
	W->qos = mqtt->will_qos;	/* XXX mqtt->qos? */
	ptr = (void *) W;
	break;
    default:
assert(0);
	break;
    }
    return ptr;
}

static
rpmRC pahoDestroy(rpmmqtt mqtt)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */
    (void) mqchk(mqtt, "destroy",
		(MQTTAsync_destroy(&mqtt->I), 0));
    mqtt->state = _free(mqtt->state);
    rc = RPMRC_OK;
    mqtt->I = NULL;
    return rc;
}

static
rpmRC pahoCreate(rpmmqtt mqtt)
{
    static int oneshot;
    int _lvl = RPMLOG_DEBUG;
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (mqtt->I == NULL) {
	if (mqtt->trace && rpmIsDebug()) {
	    rc = mqchk(mqtt, "setTraceCallback",
		(MQTTAsync_setTraceCallback(onTrace), 0));
	    if (rc)
		goto exit;
	    rc = mqchk(mqtt, "setTraceLevel",
		(MQTTAsync_setTraceLevel((enum MQTTASYNC_TRACE_LEVELS)mqtt->trace), 0));
	    if (rc)
		goto exit;
	}

	rpmlog(_lvl, "==================== %s\n", mqtt->vec->name);

	if (!oneshot) {
	    if (mqtt->trace == 0) {
		MQTTAsync_nameValue *NV = MQTTAsync_getVersionInfo();
		while (NV->name) {
		    rpmlog(_lvl, "%19s: %s\n", NV->name, NV->value);
		    NV++;
		}
	    }
	    oneshot++;
	}

	{
	    char portstr[32] = "";
	    (void) snprintf(portstr, sizeof(portstr), "%d",
			(mqtt->port ? mqtt->port : 1883));
	    (void) rpmmqttExpand(mqtt, &mqtt->uri,
			"tcp://", mqtt->host, ":", portstr, NULL);
	}

    /* XXX set through uri?query */
static const char _mqtt_persist[] =
	"%{?_mqtt_persist}%{!?_mqtt_persist:2}";
    mqtt->persist_type = (rpmmqttExpandNumeric(mqtt, _mqtt_persist) % 3);
static const char _mqtt_cachedir[] =
	"%{?_mqtt_cachedir}%{!?_mqtt_cachedir:/var/cache/mqtt}";
char *persist_path = rpmGetPath(_mqtt_cachedir, NULL);

	if (strcmp(mqtt->uri, "tcp://localhost:1883"))
	    rpmlog(_lvl, "%19s: %s\n", "uri", mqtt->uri);
	if (strcmp(mqtt->host, "localhost"))
	    rpmlog(_lvl, "%19s: %s\n", "host", mqtt->host);
	if (mqtt->port != 1883)
	    rpmlog(_lvl, "%19s: %d\n", "port", mqtt->port);
	if (mqtt->user)
	    rpmlog(_lvl, "%19s: %s\n", "user", mqtt->user);
	if (mqtt->password)
	    rpmlog(_lvl, "%19s: %s\n", "pass", mqtt->password);

	rpmlog(_lvl, "%19s: %s\n", "clientid", mqtt->clientid);
	rpmlog(_lvl, "%19s: %s qos(%d) timeout(%u msecs)\n",
		"topic", mqtt->topic, mqtt->qos, mqtt->timeout);
	rpmlog(_lvl, "%19s: type(%u) %s\n",
		"persist", mqtt->persist_type,
		(mqtt->persist_type ? persist_path : ""));
	rpmlog(_lvl, "%19s: %s\n", "flags", _MQTTFLAGS(mqtt->flags));
	if (mqtt->trace)
	    rpmlog(_lvl, "%19s: %d\n", "trace", mqtt->trace);

if (mqtt->persist_path) free((void *)mqtt->persist_path);
if (mqtt->persist_ctx) free((void *)mqtt->persist_ctx);
	switch (mqtt->persist_type) {
	default:
	case MQTTCLIENT_PERSISTENCE_NONE:
	    mqtt->persist_ctx = NULL;
	    break;
	case MQTTCLIENT_PERSISTENCE_DEFAULT:
	  {
	    mqtt->persist_path = xstrdup(persist_path);
	    /* XXX rpmmqttFini double free */
	    mqtt->persist_ctx = (void *)xstrdup(mqtt->persist_path);
	  } break;
	case MQTTCLIENT_PERSISTENCE_USER:
	  {
	    mqtt->persist_path = xstrdup(persist_path);
	    MQTTClient_persistence * ctx =
		(MQTTClient_persistence *) xmalloc(sizeof(*ctx));
	    *ctx = _mqtt_persistence;	/* structure assignment */
	    ctx->context = mqtt;
	    mqtt->persist_ctx = ctx;
	  } break;
	}
if (persist_path) free(persist_path);

#ifdef	DYING
mqtt->u = NULL;
dumpMQTT(__FUNCTION__, mqtt);
#endif

	rc = mqchk(mqtt, "createWithOptions",
		MQTTAsync_createWithOptions(&mqtt->I,
			mqtt->uri, mqtt->clientid,
			mqtt->persist_type, mqtt->persist_ctx,
			(MQTTAsync_createOptions *)AOBJ(mqtt, 'O')));
	if (rc)
	    goto exit;

	rc = mqchk(mqtt, "setCallbacks",
		MQTTAsync_setCallbacks(mqtt->I, mqtt,
			pahoOnConnectionLost,
			pahoOnMessageArrived,
			pahoOnDeliveryComplete));
	if (rc)
	    goto exit;
    }

    rc = RPMRC_OK;

exit:
    return rc;
}

static
rpmRC pahoDisconnect(rpmmqtt mqtt)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */
    if (rpmmqttIsConnected(mqtt) == RPMRC_OK) {
	mqtt->finished = 0;
	rc = mqchk(mqtt, "disconnect",
		MQTTAsync_disconnect(mqtt->I,
			(const MQTTAsync_disconnectOptions*)AOBJ(mqtt, 'D')));
	while (!mqtt->finished)
	    usleep(100);
	if (rc)
	    goto exit;
    }
    rc = RPMRC_OK;

exit:
    return rc;
}

static
rpmRC pahoConnect(rpmmqtt mqtt)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (rpmmqttIsConnected(mqtt) == RPMRC_NOTFOUND) {
	mqtt->finished = 0;
	rc = mqchk(mqtt, "connect",
		MQTTAsync_connect(mqtt->I,
			(const MQTTAsync_connectOptions*)AOBJ(mqtt, 'C')));
	while (!mqtt->finished)
	    usleep(1000);
	if (rc)
	    goto exit;
    }
    rc = RPMRC_OK;

exit:
    return rc;
}

static
rpmRC pahoIsConnected(rpmmqtt mqtt)
{
    rpmRC rc = RPMRC_NOTFOUND;

    if (MQTTAsync_isConnected(mqtt->I))
	rc = RPMRC_OK;
    return rc;
}

static
rpmRC pahoSendMessage(rpmmqtt mqtt, const char * topic,
		const char * s, size_t ns)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (topic == NULL)
	topic = mqtt->topic;
    if (s == NULL)
	s = "";
    if (ns == 0)
	ns = strlen(s);

    if (rpmmqttIsConnected(mqtt) == RPMRC_OK) {
	MQTTAsync_message *M =
	    (MQTTAsync_message *) AOBJ(mqtt, 'M');
	M->payloadlen = ns;
	M->payload = (char *) s;

	MQTTAsync_responseOptions *R =
	    (MQTTAsync_responseOptions *) AOBJ(mqtt, 'R');
	R->onSuccess = pahoOnSend;
	R->onFailure = pahoOnSendFailure;

	mqtt->finished = 0;
	rc = mqchk(mqtt, "sendMessage",
		MQTTAsync_sendMessage(mqtt->I, topic, M, R));
	while (!mqtt->finished)
	    usleep(100);
	if (rc)
	    goto exit;
	rc = RPMRC_OK;
    }

exit:
    return rc;
}

static
rpmRC pahoSubscribeMany(rpmmqtt mqtt, int ac, char ** av)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (ac <= 0)
	goto exit;
    if (!rpmmqttConnect(mqtt)) {
	int _lvl = RPMLOG_DEBUG;
	int *subqos = (int *) xcalloc(ac, sizeof(*subqos));
	for (int i = 0; i < ac; i++) {
	    char * t = av[i];
#ifdef	NOTYET
	    char * te = strchr(t, '?');
	    if (te) {
		*te++ = '\0';
		if ((te = strchr(te, '='))) {
		    if (!strncmp(t, "qos", (te - t)))
			subqos[i] = strtoul(te+1, NULL, 0);
		}
	    } else
#endif
		subqos[i] = mqtt->qos;	/* XXX */
	    rpmlog(_lvl, "%19s: %s qos(%u)\n", "subscribe", t, subqos[i]);
	}

	MQTTAsync_responseOptions *R =
		(MQTTAsync_responseOptions *) AOBJ(mqtt, 'R');
	R->onSuccess = pahoOnSubscribeMany;
	R->onFailure = pahoOnSubscribeManyFailure;

	mqtt->finished = 0;
	rc = mqchk(mqtt, "subscribeMany",
		MQTTAsync_subscribeMany(mqtt->I, ac, av, subqos, R));
	while (!mqtt->finished)
	    usleep(100);
	if (subqos)
	    free(subqos);
	if (rc)
	    goto exit;
	rc = RPMRC_OK;
    }

exit:
    return rc;
}

static
rpmRC pahoSubscribe(rpmmqtt mqtt, const char * topic, int qos)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (!rpmmqttConnect(mqtt)) {
	int _lvl = RPMLOG_DEBUG;

	rpmlog(_lvl, "%19s: %s qos(%d)\n", "subscribe", topic, qos);

	MQTTAsync_responseOptions *R =
		(MQTTAsync_responseOptions *) AOBJ(mqtt, 'R');
	R->onSuccess = pahoOnSubscribe;
	R->onFailure = pahoOnSubscribeFailure;

	mqtt->finished = 0;
	rc = mqchk(mqtt, "subscribe",
		MQTTAsync_subscribe(mqtt->I, topic, qos, R));
	while (!mqtt->finished)
	    usleep(100);
	if (rc)
	    goto exit;
	rc = RPMRC_OK;
    }

exit:
    return rc;
}

static
rpmRC pahoUnsubscribe(rpmmqtt mqtt, const char * topic)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (!rpmmqttConnect(mqtt)) {
	int _lvl = RPMLOG_DEBUG;

	rpmlog(_lvl, "%19s: %s\n", "unsubscribe", topic);

	MQTTAsync_responseOptions *R =
		(MQTTAsync_responseOptions *) AOBJ(mqtt, 'R');
	R->onSuccess = pahoOnUnsubscribeMany;
	R->onFailure = pahoOnUnsubscribeManyFailure;

	mqtt->finished = 0;
	rc = mqchk(mqtt, "unsubscribe",
		MQTTAsync_unsubscribe(mqtt->I, topic, R));
	while (!mqtt->finished)
	    usleep(100);
	if (rc)
	    goto exit;
	rc = RPMRC_OK;
    }

exit:
    return rc;
}

static
rpmRC pahoUnsubscribeMany(rpmmqtt mqtt, int ac, char ** av)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (ac <= 0)
	goto exit;
    if (!rpmmqttConnect(mqtt)) {
	int _lvl = RPMLOG_DEBUG;
	for (int i = 0; i < ac; i++) {
	    char * t = av[i];
	    rpmlog(_lvl, "%19s: %s\n", "unsubscribe", t);
	}

	MQTTAsync_responseOptions *R =
		(MQTTAsync_responseOptions *) AOBJ(mqtt, 'R');
	R->onSuccess = pahoOnUnsubscribeMany;
	R->onFailure = pahoOnUnsubscribeManyFailure;

	mqtt->finished = 0;
	rc = mqchk(mqtt, "unsubscribeMany",
		MQTTAsync_unsubscribeMany(mqtt->I, ac, av, R));
	while (!mqtt->finished)
	    usleep(100);
	if (rc)
	    goto exit;
	rc = RPMRC_OK;
    }

exit:
    return rc;
}

static KEY pahoErrs[] = {
#define _ENTRY(_v)      { MQTTASYNC_##_v, #_v, }
    _ENTRY(SUCCESS),
    _ENTRY(FAILURE),
    _ENTRY(PERSISTENCE_ERROR),
    _ENTRY(DISCONNECTED),
    _ENTRY(MAX_MESSAGES_INFLIGHT),
    _ENTRY(BAD_UTF8_STRING),
    _ENTRY(NULL_PARAMETER),
    _ENTRY(TOPICNAME_TRUNCATED),
    _ENTRY(BAD_STRUCTURE),
    _ENTRY(BAD_QOS),
    _ENTRY(NO_MORE_MSGIDS),
    _ENTRY(OPERATION_INCOMPLETE),
    _ENTRY(MAX_BUFFERED_MESSAGES),
#undef	_ENTRY
    { 0, NULL },
};

static
struct mqttVec_s pahoVec = {
    .name		= "paho",
    .uri		= "mqtt://localhost:1883/"
				"rpm/mqtt"	/* XXX topic? */
				"?qos=1"	/* XXX */
				"&timeout=10000"
				"&max_inflight=20"
				"&keepalive=60",
    .port		= 1883,
    .sport		= 8883,
    .prefix		= "MQTTAsync_",
    .all		= 0,
    .errs		= pahoErrs,
    .nerrs		= sizeof(pahoErrs) / sizeof(pahoErrs[0]),
    .destroy		= pahoDestroy,
    .create		= pahoCreate,
    .disconnect		= pahoDisconnect,
    .connect		= pahoConnect,
    .isconnected	= pahoIsConnected,
    .unsubscribe	= pahoUnsubscribe,
    .subscribe		= pahoSubscribe,
    .unsubscribeMany	= pahoUnsubscribeMany,
    .subscribeMany	= pahoSubscribeMany,
    .sendMessage	= pahoSendMessage,
};
#endif	/* WITH_PAHO */

/*==============================================================*/
#ifdef	WITH_MOSQUITTO
static int mosqGetPassword(char *b, int size, int rwflag, void * _mqtt)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
    (void)mqtt;
    int nb = 0;
    /* XXX returns password in b[nb] */
    return nb;
}

static void mosqOnConnect(struct mosquitto * I, void * _mqtt, int rc)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
if (_rpmmqtt_debug)
rpmlog(RPMLOG_DEBUG, "%s: connected rc(%d)\n", mqtt->vec->name, rc);
    mqtt->finished = 1;
}

static void mosqOnDisconnect(struct mosquitto * I, void * _mqtt, int rc)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
    if (mqtt->connected)
        rpmlog(RPMLOG_DEBUG, "%s: disconnected rc(%d)\n", mqtt->vec->name, rc);
mqtt->connected = 0;
    mqtt->finished = 1;
}

static void mosqOnPublish(struct mosquitto * I, void * _mqtt, int mid)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
if (_rpmmqtt_debug)
rpmlog(RPMLOG_DEBUG, "%s: sent(%d)\n", mqtt->vec->name, mid);
    mqtt->token = mid;
    mqtt->finished = 1;
}

static void mosqOnMessage(struct mosquitto * I, void * _mqtt,
		const struct mosquitto_message *msg)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;

    mqtt->msg_count++;
    mqtt->token = msg->mid;

    const char * s = (const char *) msg->payload;
    size_t ns = msg->payloadlen;

    rpmlog(RPMLOG_DEBUG, "%s: rcvd topic(%s) qos(%d) retain(%d) \"%.*s\"\n",
		mqtt->vec->name, msg->topic, msg->qos, (msg->retain ? 1 : 0),
		(int)ns, s);

    if (mqtt->iob) {
	mqtt->iob = rpmiobAppend(mqtt->iob, msg->topic, 0);
	mqtt->iob = rpmiobAppend(mqtt->iob, ":\t", 0);
	{   char * t = (char *) memcpy(alloca(ns+1), s, ns);
	    t[ns] = '\0';
	    mqtt->iob = rpmiobAppend(mqtt->iob, t, 1);
	}
    }
    switch (mqtt->msg_output) {
    case MQTT_OUTPUT_UNKNOWN:
    case MQTT_OUTPUT_CALLBACK:
    default:
	break;
    case MQTT_OUTPUT_STDOUT:
    case MQTT_OUTPUT_FILE:
	if (mqtt->ofd) {
	    size_t nw;
	    nw = Fwrite(msg->topic, sizeof(*msg->topic), strlen(msg->topic),
			mqtt->ofd);
assert(nw > 0);
	    nw = Fwrite(":\t", 1, sizeof(":\t")-1, mqtt->ofd);
assert(nw > 0);
	    nw = Fwrite(s, sizeof(*s), ns, mqtt->ofd);
assert(nw > 0);
	    nw = Fwrite("\n", 1, sizeof("\n")-1, mqtt->ofd);
assert(nw > 0);
	}
	break;
    }
#ifdef	NEW
    message->qos
    message->retained
    message->dup
    message->msgid
#endif

    mqtt->finished = 1;
}

static void mosqOnSubscribe(struct mosquitto * I, void * _mqtt, int mid,
		int qos_counted, const int *granted_qos)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
if (_rpmmqtt_debug)
rpmlog(RPMLOG_DEBUG, "%s: subscribed(%d) qos(%d)\n",
		mqtt->vec->name, mid, granted_qos[0]);
    mqtt->token = mid;
    mqtt->finished = 1;
}

static void mosqOnUnsubscribe(struct mosquitto * I, void * _mqtt, int mid)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
if (_rpmmqtt_debug)
rpmlog(RPMLOG_DEBUG, "%s: unsubscribed(%d)\n", mqtt->vec->name, mid);
    mqtt->token = mid;
    mqtt->finished = 1;
}

static void mosqOnLog(struct mosquitto * I, void * _mqtt,
		int logtype, const char *str)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;
    const char * subsys = "";
    int lvl = RPMLOG_DEBUG;

    if (logtype & MOSQ_LOG_INFO)
	lvl = RPMLOG_INFO;
    if (logtype & MOSQ_LOG_NOTICE)
	lvl = RPMLOG_NOTICE;
    if (logtype & MOSQ_LOG_WARNING)
	lvl = RPMLOG_WARNING;
    if (logtype & MOSQ_LOG_ERR)
	lvl = RPMLOG_ERR;
    if (logtype & MOSQ_LOG_DEBUG)
	lvl = RPMLOG_DEBUG;
    if (logtype & MOSQ_LOG_SUBSCRIBE)
	subsys = "subscribe: ";
    if (logtype & MOSQ_LOG_SUBSCRIBE)
	subsys = "subscribe: ";
    if (logtype & MOSQ_LOG_UNSUBSCRIBE)
	subsys = "unsubscribe: ";
    if (logtype & MOSQ_LOG_WEBSOCKETS)
	subsys = "websockets: ";

    rpmlog(lvl, "%s: %s%s\n", mqtt->vec->name, subsys, str);
}

static
rpmRC mosqDestroy(rpmmqtt mqtt)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */
#ifdef	NOTYET
    mqtt->state = _free(mqtt->state);
#endif
    if (mqtt->I) {
	(void) mqchk(mqtt, "loop_stop",
		mosquitto_loop_stop(mqtt->I, true));
	(void) mqchk(mqtt, "destroy",
		(mosquitto_destroy(mqtt->I), 0));
    }
    rc = RPMRC_OK;
    mqtt->I = NULL;
    mosquitto_lib_cleanup();
    return rc;
}

static
rpmRC mosqCreate(rpmmqtt mqtt)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */
    static int oneshot;
    int _lvl = RPMLOG_DEBUG;

    if (mqtt->I == NULL) {

	rpmlog(_lvl, "==================== %s\n", mqtt->vec->name);

	if (!oneshot) {
	    int major = 0;
	    int minor = 0;
	    int revision = 0;
	    int version = 0;

	    rc = mqchk(mqtt, "lib_init",
		mosquitto_lib_init());
	    if (rc)
		goto exit;
	    version = mosquitto_lib_version(&major, &minor, &revision);
	    rpmlog(_lvl, "%19s: %d.%d.%d (%d)\n", "version",
			major, minor, revision, version);
	    oneshot++;
	}

	rpmlog(_lvl, "%19s: %s\n", "clientid", mqtt->clientid);
	rpmlog(_lvl, "%19s: %s qos(%d) timeout(%u msecs)\n",
		"topic", mqtt->topic, mqtt->qos, mqtt->timeout);
	rpmlog(_lvl, "%19s: %s\n", "flags", _MQTTFLAGS(mqtt->flags));
	if (mqtt->trace)
	    rpmlog(_lvl, "%19s: %d\n", "trace", mqtt->trace);

#ifdef	DYING
mqtt->u = NULL;
dumpMQTT(__FUNCTION__, mqtt);
#endif

	rc = mqchk(mqtt, "new",
		(mqtt->I = mosquitto_new(mqtt->clientid,
			(MF_ISSET(CLEAN) ? true : false),
			mqtt)) == NULL);
	if (rc)
	    goto exit;

	mosquitto_connect_callback_set(mqtt->I, mosqOnConnect);
	mosquitto_disconnect_callback_set(mqtt->I, mosqOnDisconnect);
	mosquitto_publish_callback_set(mqtt->I, mosqOnPublish);
	mosquitto_message_callback_set(mqtt->I, mosqOnMessage);
	mosquitto_subscribe_callback_set(mqtt->I, mosqOnSubscribe);
	mosquitto_unsubscribe_callback_set(mqtt->I, mosqOnUnsubscribe);

#ifdef	NOTYET
	if (mqtt->trace && rpmIsDebug())
#else
	if (mqtt->trace || mqtt->debug)
#endif
	    mosquitto_log_callback_set(mqtt->I, mosqOnLog);

/*
 * Example 1:
 *      delay=2, delay_max=10, exponential_backoff=False
 *      Delays would be: 2, 4, 6, 8, 10, 10, ...
 *
 * Example 2:
 *      delay=3, delay_max=30, exponential_backoff=True
 *      Delays would be: 3, 6, 12, 24, 30, 30, ...
 */
	rc = mqchk(mqtt, "reconnect_delay_set",
		mosquitto_reconnect_delay_set(mqtt->I, 2, 10, false));
	if (rc)
	    goto exit;

	rc = mqchk(mqtt, "user_data_set",
		(mosquitto_user_data_set(mqtt->I, mqtt), 0));
	if (rc)
	    goto exit;

	rc = mqchk(mqtt, "loop_start",
		mosquitto_loop_start(mqtt->I));
	if (rc)
	    goto exit;

    }

    rc = RPMRC_OK;

exit:
    return rc;
}

static
rpmRC mosqDisconnect(rpmmqtt mqtt)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (rpmmqttIsConnected(mqtt) == RPMRC_OK) {

	rc = mqchk(mqtt, "loop",
		mosquitto_loop(mqtt->I, -1, 1));	/* XXX */
	if (rc)
	    goto exit;

	mqtt->finished = 0;
	rc = mqchk(mqtt, "disconnect",
		mosquitto_disconnect(mqtt->I));
mqtt->connected = 0;
	if (rc)
	    goto exit;

	rc = mqchk(mqtt, "loop",
		mosquitto_loop(mqtt->I, -1, 1));

	if (rc)
	    goto exit;
    }
    rc = RPMRC_OK;

exit:
    return rc;
}

static
rpmRC mosqConnect(rpmmqtt mqtt)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (rpmmqttIsConnected(mqtt) == RPMRC_NOTFOUND) {
	urlinfo u = mqtt->u;

	if (mqtt->will_topic) {
	    const char *payload =
		(mqtt->will_message ? mqtt->will_message : "");
	    int payloadlen = strlen(payload);
	    rc = mqchk(mqtt, "will_set",
		mosquitto_will_set(mqtt->I,
		(mqtt->will_topic ? mqtt->will_topic : mqtt->topic),
		payloadlen, payload,
		mqtt->will_qos,	/* XXX mqtt->qos? */
		(MF_ISSET(WILL_RETAIN) ? true : false)));
	} else
	    rc = mqchk(mqtt, "will_clear",
		mosquitto_will_clear(mqtt->I));
	if (rc)
	    goto exit;

	rc = mqchk(mqtt, "username_pw_set",
		mosquitto_username_pw_set(mqtt->I,
			(u && u->user ? u->user : mqtt->user),
			(u && u->password ? u->password : mqtt->password)));
	if (rc)
	    goto exit;

	if (mqtt->cacertfile || mqtt->_capath) {
	    rc = mqchk(mqtt, "tls_set",
		    mosquitto_tls_set(mqtt->I,
			mqtt->cacertfile, mqtt->_capath,
			mqtt->certfile, mqtt->keyfile, mosqGetPassword));
	    if (rc)
		goto exit;
	}
	if (MF_ISSET(INSECURE)) {
	    rc = mqchk(mqtt, "tls_insecure_set",
		    mosquitto_tls_insecure_set(mqtt->I, true));
	    if (rc)
		goto exit;
	}
	if (mqtt->_psk_key) {
	    rc = mqchk(mqtt, "tls_psk_set",
		    mosquitto_tls_psk_set(mqtt->I,
			mqtt->_psk_key, mqtt->_psk_identity, NULL));
	    if (rc)
		goto exit;
	}
	if (mqtt->ciphers && mqtt->_tls_version) {
	    static int _cert_reqs = 1;	/* XXX 0/1: SSL_VERIFY_{NONE,PEER} */
	    char * _tls_version = rpmExpand("tlsv", mqtt->_tls_version, NULL);
	    rc = mqchk(mqtt, "tls_opts_set",
		mosquitto_tls_opts_set(mqtt->I,
			_cert_reqs, _tls_version, mqtt->ciphers));
	    if (_tls_version) free(_tls_version);
	    if (rc)
		goto exit;
	}

        rc = mqchk(mqtt, "max_inflight_messages_set",
		mosquitto_max_inflight_messages_set(mqtt->I,
			(mqtt->max_inflight ? mqtt->max_inflight : 20)));
	if (rc)
	    goto exit;

	if (mqtt->_proxy) {
	    urlinfo u;
	    int ut = (urltype) urlSplit(mqtt->_proxy, &u);
	    (void) ut;
	    rc = mqchk(mqtt, "socks5_set",
		mosquitto_socks5_set(mqtt->I,
			u->host,
			u->port,
			u->user,
			u->password));
	    if (rc)
		goto exit;
	}

	{   int MQTTVersion = MQTT_PROTOCOL_V31;
	    if (!strcmp(mqtt->protocol_version, "31"))
		MQTTVersion = MQTT_PROTOCOL_V31;
	    if (!strcmp(mqtt->protocol_version, "311"))
		MQTTVersion = MQTT_PROTOCOL_V311;
	    rc = mqchk(mqtt, "opts_set",
		mosquitto_opts_set(mqtt->I,
			MOSQ_OPT_PROTOCOL_VERSION, &MQTTVersion));
	    if (rc)
		goto exit;
	}

#ifndef	NOTYET
	mqtt->finished = 0;
	/* XXX mosquitto_connect_srv(..., mqtt->_address? */
	if (mqtt->_address) {
	    rc = mqchk(mqtt, "connect_bind_async",
		    mosquitto_connect_bind_async(mqtt->I,
			(u && u->host ? u->host : mqtt->host),
			(u && u->host ? u->port : mqtt->port),
			mqtt->keepalive, mqtt->_address));
	} else
	    rc = mqchk(mqtt, "connect_async",
		    mosquitto_connect_async(mqtt->I,
			(u && u->host ? u->host : mqtt->host),
			(u && u->host ? u->port : mqtt->port),
			mqtt->keepalive));
	if (rc)
	    goto exit;
	
#else	/* NOTYET */

	if (mqtt->_address) {
	    if (MF_ISSET(DNSSRV)) {
		rc = mqchk(mqtt, "connect_srv",
		    mosquitto_connect_srv(mqtt->I,
			(u && u->host ? u->host : mqtt->host),
			mqtt->keepalive, mqtt->_address));
	    } else {
		rc = mqchk(mqtt, "connect_bind",
		    mosquitto_connect_bind(mqtt->I,
			(u && u->host ? u->host : mqtt->host),
			(u && u->host ? u->port : mqtt->port),
			mqtt->keepalive, mqtt->_address));
	    }
	} else
	    rc = mqchk(mqtt, "connect",
		    mosquitto_connect(mqtt->I,
			(u && u->host ? u->host : mqtt->host),
			(u && u->host ? u->port : mqtt->port),
			mqtt->keepalive));
#endif	/* NOTYET */

	rc = mqchk(mqtt, "loop",
		mosquitto_loop(mqtt->I, -1, 1));

	if (rc)
	    goto exit;
mqtt->connected = 1;
    }
    rc = RPMRC_OK;

exit:
    return rc;
}

static
rpmRC mosqIsConnected(rpmmqtt mqtt)
{
    rpmRC rc = (mqtt->connected ? RPMRC_OK : RPMRC_NOTFOUND);
    return rc;
}

static
rpmRC mosqSendMessage(rpmmqtt mqtt, const char * topic,
		const char * s, size_t ns)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (topic == NULL)
	topic = mqtt->topic;
    if (s == NULL)
	s = "";
    if (ns == 0)
	ns = strlen(s);

    if (!rpmmqttConnect(mqtt)) {

	mqtt->finished = 0;
	rc = mqchk(mqtt, "publish",
		mosquitto_publish(mqtt->I,
			(int *)&mqtt->token,
			(topic ? topic : mqtt->topic),
			ns, s,
			mqtt->qos,
			(MF_ISSET(RETAIN) ? true : false)));
	if (rc)
	    goto exit;

	rpmlog(RPMLOG_DEBUG, "%s: send(%d) topic(%s) \"%.*s\"\n",
			mqtt->vec->name, mqtt->token, mqtt->topic, (int)ns, s);

	rc = mqchk(mqtt, "loop",
		mosquitto_loop(mqtt->I, -1, 1));
	if (rc)
	    goto exit;
	rc = RPMRC_OK;
    }

exit:
    return rc;
}

static
rpmRC mosqSubscribe(rpmmqtt mqtt, const char * topic, int qos)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (!rpmmqttConnect(mqtt)) {

	rpmlog(RPMLOG_DEBUG, "%19s: %s qos(%d)\n", "subscribe", topic, qos);

	mqtt->finished = 0;
	rc = mqchk(mqtt, "subscribe",
		mosquitto_subscribe(mqtt->I,
			(int *)&mqtt->token,
			(topic ? topic : mqtt->topic),
			qos));
	if (rc)
	    goto exit;

	rc = mqchk(mqtt, "loop",
		mosquitto_loop(mqtt->I, -1, 1));

	if (rc)
	    goto exit;
	rc = RPMRC_OK;
    }

exit:
    return rc;
}

static
rpmRC mosqUnsubscribe(rpmmqtt mqtt, const char * topic)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (!rpmmqttConnect(mqtt)) {

	rpmlog(RPMLOG_DEBUG, "%19s: %s\n", "unsubscribe", topic);

	mqtt->finished = 0;
	rc = mqchk(mqtt, "unsubscribe",
		mosquitto_unsubscribe(mqtt->I,
			(int *)&mqtt->token,
			(topic ? topic : mqtt->topic)));
	if (rc)
	    goto exit;

	rc = mqchk(mqtt, "loop",
		mosquitto_loop(mqtt->I, -1, 1));

	if (rc)
	    goto exit;
	rc = RPMRC_OK;
    }

exit:
    return rc;
}

static KEY mosqErrs[] = {
#define _ENTRY(_v)      { MOSQ_ERR_##_v, #_v, }
    _ENTRY(CONN_PENDING),
    _ENTRY(SUCCESS),
    _ENTRY(NOMEM),
    _ENTRY(PROTOCOL),
    _ENTRY(INVAL),
    _ENTRY(NO_CONN),
    _ENTRY(CONN_REFUSED),
    _ENTRY(NOT_FOUND),
    _ENTRY(CONN_LOST),
    _ENTRY(TLS),
    _ENTRY(PAYLOAD_SIZE),
    _ENTRY(NOT_SUPPORTED),
    _ENTRY(AUTH),
    _ENTRY(ACL_DENIED),
    _ENTRY(UNKNOWN),
    _ENTRY(ERRNO),
    _ENTRY(EAI),
    _ENTRY(PROXY),
#undef	_ENTRY
    { 0, NULL },
};

static
struct mqttVec_s mosqVec = {
    .name		= "mosquitto",
    .uri		= "mqtt://localhost:1883/"
				"rpm/mqtt"	/* XXX topic? */
				"?qos=1"	/* XXX */
				"&timeout=10000"
				"&max_inflight=20"
				"&keepalive=60",
    .port		= 1883,
    .sport		= 8883,
    .prefix		= "mosquitto_",
    .all		= 0,
    .errs		= mosqErrs,
    .nerrs		= sizeof(mosqErrs) / sizeof(mosqErrs[0]),
    .destroy		= mosqDestroy,
    .create		= mosqCreate,
    .disconnect		= mosqDisconnect,
    .connect		= mosqConnect,
    .isconnected	= mosqIsConnected,
    .unsubscribe	= mosqUnsubscribe,
    .subscribe		= mosqSubscribe,
    .unsubscribeMany	= _rpmmqttUnsubscribeMany,
    .subscribeMany	= _rpmmqttSubscribeMany,
    .sendMessage	= mosqSendMessage,
};
#endif	/* WITH_MOSQUITTO */

/*==============================================================*/
#ifdef	WITH_RABBITMQ

#ifdef	REFERENCE
static
void die(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    fprintf(stderr, "\n");
    exit(1);
}

static
void die_on_error(int x, char const *context)
{
    if (x < 0) {
	fprintf(stderr, "%s: %s\n", context, amqp_error_string2(x));
	exit(1);
    }
}
#endif	/* REFERENCE */

static
rpmRC Xmqchkrpc(rpmmqtt mqtt, const char * msg, amqp_rpc_reply_t x,
		int printit, const char * func, const char * fn, unsigned ln)
{
    const char * api = (mqtt && mqtt->vec && mqtt->vec->prefix
		? mqtt->vec->prefix : "");
    amqp_connection_close_t *mconn;
    amqp_channel_close_t *mchan;
    rpmRC rc = RPMRC_FAIL;;
    int all = (mqtt && mqtt->vec ? mqtt->vec->all : 1);

    switch (x.reply_type) {
    case AMQP_RESPONSE_NORMAL:
	if (all || _rpmmqtt_debug < 0)
	    fprintf(stderr, "\t%s:%s:%u %s%s: %s\n",
			func, fn, ln, api, msg, "OK");
	rc = RPMRC_OK;
	break;

    case AMQP_RESPONSE_NONE:
	fprintf(stderr, "%s:%s:%u %s%s: missing RPC reply type!\n",
		func, fn, ln, api, msg);
	break;

    case AMQP_RESPONSE_LIBRARY_EXCEPTION:
	fprintf(stderr, "%s:%s:%u %s%s: %s\n",
		func, fn, ln, api, msg,
		amqp_error_string2(x.library_error));
	break;

    case AMQP_RESPONSE_SERVER_EXCEPTION:
	switch (x.reply.id) {
	case AMQP_CONNECTION_CLOSE_METHOD:
	    mconn = (amqp_connection_close_t *) x.reply.decoded;
	    fprintf(stderr, "%s:%s:%u %s%s: error(%d): %.*s\n",
			func, fn, ln, api, msg, mconn->reply_code,
			(int) mconn->reply_text.len,
			(char *) mconn->reply_text.bytes);
	    break;
	case AMQP_CHANNEL_CLOSE_METHOD:
	    mchan = (amqp_channel_close_t *) x.reply.decoded;
	    fprintf(stderr, "%s:%s:%u %s%s: error(%d): %.*s\n",
			func, fn, ln, api, msg, mchan->reply_code,
			(int) mchan->reply_text.len,
			(char *) mchan->reply_text.bytes);
	    break;
	default:
	    fprintf(stderr, "%s:%s:%u %s%s: unknown error: method(%d)\n",
			func, fn, ln, api, msg, x.reply.id);
	    break;
	}
	break;
    }

    return rc;
}
#define mqchkrpc(_o, _m, _rc)  \
    Xmqchkrpc(_o, _m, _rc, _rpmmqtt_debug, __FUNCTION__, __FILE__, __LINE__)

static void dump_row(long count, int numinrow, int *chs)
{
    int i;

    printf("%08lX:", count - numinrow);

    if (numinrow > 0) {
	for (i = 0; i < numinrow; i++) {
	    if (i == 8) {
		printf(" :");
	    }
	    printf(" %02X", chs[i]);
	}
	for (i = numinrow; i < 16; i++) {
	    if (i == 8) {
		printf(" :");
	    }
	    printf("   ");
	}
	printf("  ");
	for (i = 0; i < numinrow; i++) {
	    if (isprint(chs[i])) {
		printf("%c", chs[i]);
	    } else {
		printf(".");
	    }
	}
    }
    printf("\n");
}

static int rows_eq(int *a, int *b)
{
    int i;

    for (i = 0; i < 16; i++)
	if (a[i] != b[i]) {
	    return 0;
	}

    return 1;
}

static
void amqp_dump(void const *buffer, size_t len)
{
    unsigned char *buf = (unsigned char *) buffer;
    long count = 0;
    int numinrow = 0;
    int chs[16];
    int oldchs[16] = { 0 };
    int showed_dots = 0;
    size_t i;

    for (i = 0; i < len; i++) {
	int ch = buf[i];

	if (numinrow == 16) {
	    int j;

	    if (rows_eq(oldchs, chs)) {
		if (!showed_dots) {
		    showed_dots = 1;
		    printf
			("          .. .. .. .. .. .. .. .. : .. .. .. .. .. .. .. ..\n");
		}
	    } else {
		showed_dots = 0;
		dump_row(count, numinrow, chs);
	    }

	    for (j = 0; j < 16; j++) {
		oldchs[j] = chs[j];
	    }

	    numinrow = 0;
	}

	count++;
	chs[numinrow++] = ch;
    }

    dump_row(count, numinrow, chs);

    if (numinrow != 0) {
	printf("%08lX:\n", count);
    }
}

static
rpmRC amqpDestroy(rpmmqtt mqtt)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */
    if (mqtt->I) {
	rc = mqchk(mqtt, "destroy_connection",
		amqp_destroy_connection(mqtt->I));
    }
    mqtt->I = NULL;
    return rc;
}

static
rpmRC amqpCreate(rpmmqtt mqtt)
{
    static int oneshot;
    int _lvl = RPMLOG_DEBUG;
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (mqtt->I == NULL) {
	rpmlog(_lvl, "==================== %s\n", mqtt->vec->name);

	if (!oneshot) {
	    int major = AMQP_VERSION_MAJOR;
	    int minor = AMQP_VERSION_MINOR;
	    int patch = AMQP_VERSION_PATCH;
	    int version = AMQP_VERSION;
	    rpmlog(_lvl, "%19s: %d.%d.%d (0x%x)\n", "version",
			major, minor, patch, version);
	    oneshot++;
	}

	rpmlog(_lvl, "%19s: %s\n", "vhost", mqtt->vhost);
	rpmlog(_lvl, "%19s: %s\n", "queue", mqtt->queue);
	rpmlog(_lvl, "%19s: %s\n", "exchange", mqtt->exchange);
	rpmlog(_lvl, "%19s: %s\n", "flags", _MQTTFLAGS(mqtt->flags));
	if (mqtt->trace)
	    rpmlog(_lvl, "%19s: %d\n", "trace", mqtt->trace);

#ifdef	DYING
mqtt->u = NULL;
dumpMQTT(__FUNCTION__, mqtt);
#endif

	amqp_socket_t *socket = NULL;
	int port;

	rc = mqchk(mqtt, "new_connection",
		((mqtt->I = amqp_new_connection()) == NULL));
	if (rc)
	    goto exit;

	if (mqtt->cacertfile || mqtt->_capath) {
	    port = mqtt->vec->sport;	/* XXX elsewhere */
	    rc = mqchk(mqtt, "ssl_socket_new",
		    ((socket = amqp_ssl_socket_new(mqtt->I)) == NULL));
	    if (rc || socket == NULL)
		goto exit;
#ifdef	NOTYET
  amqp_ssl_socket_set_verify_peer(socket, 1);
  amqp_ssl_socket_set_verify_hostname(socket, 1);
#endif
	    if (mqtt->cacertfile) {
		rc = mqchk(mqtt, "ssl_socket_set_cacert",
			amqp_ssl_socket_set_cacert(mqtt->I, mqtt->cacertfile));
		if (rc)
		    goto exit;
	    }
	    if (mqtt->keyfile) {
		rc = mqchk(mqtt, "ssl_socket_set_key",
			amqp_ssl_socket_set_key(mqtt->I,
				mqtt->certfile, mqtt->keyfile));
		if (rc)
		    goto exit;
	    }
	} else {
	    port = mqtt->vec->port;	/* XXX elsewhere */
	    rc = mqchk(mqtt, "tcp_socket_new",
		((socket = amqp_tcp_socket_new(mqtt->I)) == NULL));
	    if (rc || socket == NULL)
		goto exit;
	}

#ifdef	NOTYET
	{   strict timeval tv = { mqtt->connect_timeout, 0 };
	    rc = mqchk(mqtt, "socket_open_noblock",
		amqp_socket_open(socket, mqtt->host, port, &tv));

	}
#endif
	rc = mqchk(mqtt, "socket_open",
		amqp_socket_open(socket, mqtt->host, port));
	if (rc)
	    goto exit;

    }

    rc = RPMRC_OK;

exit:
    return rc;
}

static
rpmRC amqpDisconnect(rpmmqtt mqtt)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (rpmmqttIsConnected(mqtt) == RPMRC_OK) {
	amqp_channel_t _channel = 1;
	int _code = AMQP_REPLY_SUCCESS;
	rc = mqchkrpc(mqtt, "channel_close",
		amqp_channel_close(mqtt->I, _channel, _code));
	if (rc)
	    goto exit;
	rc = mqchkrpc(mqtt, "connection_close",
		amqp_connection_close(mqtt->I, _code));
	if (rc)
	    goto exit;
mqtt->connected = 0;
    }

exit:
    return rc;
}

static
rpmRC amqpConnect(rpmmqtt mqtt)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (rpmmqttIsConnected(mqtt) == RPMRC_NOTFOUND) {
	int _channel_max = AMQP_DEFAULT_MAX_CHANNELS;
	int _frame_max = AMQP_DEFAULT_FRAME_SIZE;
	const amqp_table_t *_properties = NULL;
	/* XXX AMQP_SASL_METHOD_EXTERNAL uses stdargs */
	amqp_sasl_method_enum _sasl_method =
			(mqtt->_sasl_mechanism == NULL
			|| !strcmp(mqtt->_sasl_mechanism, "plain")
				? AMQP_SASL_METHOD_PLAIN
				: AMQP_SASL_METHOD_PLAIN); /* XXX EXTERNAL */

	rc = mqchkrpc(mqtt, "login_with_properties",
		amqp_login_with_properties(mqtt->I,
			(mqtt->vhost ? mqtt->vhost : "/"),
			_channel_max,
			_frame_max,
			mqtt->keepalive,
			_properties,
			_sasl_method,
			mqtt->user,
			mqtt->password));
	if (rc)
	    goto exit;

	amqp_channel_open_ok_t * open_ok = NULL;
	(void)open_ok;
	amqp_channel_t _channel = 1;
	rc = mqchk(mqtt, "channel_open",
		((open_ok = amqp_channel_open(mqtt->I, _channel)),
		0));
	if (rc)
	    goto exit;
	rc = mqchkrpc(mqtt, "get_rpc_reply",
		amqp_get_rpc_reply(mqtt->I));
	if (rc)
	    goto exit;
mqtt->connected = 1;
    } else
	rc = RPMRC_OK;

exit:
    return rc;
}

static
rpmRC amqpIsConnected(rpmmqtt mqtt)
{
    rpmRC rc = (mqtt->connected ? RPMRC_OK : RPMRC_NOTFOUND);
    return rc;
}

static
rpmRC amqpSendMessage(rpmmqtt mqtt, const char * topic,
		const char * s, size_t ns)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (topic == NULL)
	topic = mqtt->topic;
    if (s == NULL)
	s = "";
    if (ns == 0)
	ns = strlen(s);

    if (!rpmmqttConnect(mqtt)) {
	amqp_channel_t _channel = 1;
	amqp_boolean_t _mandatory = false;
	amqp_boolean_t _immediate = false;
	struct amqp_basic_properties_t_ const *_properties = NULL;

	amqp_bytes_t message_bytes;
	message_bytes.len = ns;
	message_bytes.bytes = (char *) s;

#ifdef	NOTYET
	amqp_basic_properties_t props;
	props._flags = AMQP_BASIC_CONTENT_TYPE_FLAG | AMQP_BASIC_DELIVERY_MODE_FLAG;
	props.content_type = amqp_cstring_bytes("text/plain");
	props.delivery_mode = 2; /* persistent delivery mode */
	_properties = &props;
#endif

if (rpmIsDebug())
fprintf(stderr, "*** %s: %s mqtt host %s topic %s\n", __FUNCTION__, topic, mqtt->host, mqtt->topic);
	rc = mqchk(mqtt, "basic_publish",
		amqp_basic_publish(mqtt->I,
				_channel,
				amqp_cstring_bytes(mqtt->exchange),
				amqp_cstring_bytes(mqtt->queue),
				_mandatory,
				_immediate,
				_properties,
				message_bytes));

	rpmlog(RPMLOG_DEBUG, "%s: send() exchange(%s) queue(%s) \"%.*s\"\n",
			mqtt->vec->name, mqtt->exchange, mqtt->queue, (int)ns, s);

	if (rc)
	    goto exit;
	rc = RPMRC_OK;
    }

exit:
    return rc;
}

static
rpmRC amqpSubscribe(rpmmqtt mqtt, const char * topic, int qos)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (!rpmmqttConnect(mqtt)) {
	int _lvl = RPMLOG_DEBUG;

if (rpmIsDebug())
fprintf(stderr, "XXX %s: FIXME\n", __FUNCTION__);
	rpmlog(_lvl, "%19s: %s qos(%d)\n", "subscribe", topic, qos);

	amqp_bytes_t _queuename;

	amqp_channel_t _channel = 1;
	amqp_bytes_t _queue = amqp_empty_bytes;
	amqp_boolean_t _passive = (MF_ISSET(PASSIVE) ? true : false);
	amqp_boolean_t _durable = (MF_ISSET(DURABLE) ? true : false);
	amqp_boolean_t _exclusive = (MF_ISSET(EXCLUSIVE) ? true : false);
#ifdef	NOTYET
	amqp_boolean_t _auto_delete = (MF_ISSET(AUTODELETE) ? true : false);
#else
	amqp_boolean_t _auto_delete = true;
#endif
	amqp_table_t _arguments = amqp_empty_table;
	{   amqp_queue_declare_ok_t *r =
		amqp_queue_declare(mqtt->I,
			_channel,
			_queue,
			_passive,
			_durable,
			_exclusive,
			_auto_delete,
			_arguments);
	    rc = mqchkrpc(mqtt, "queue_declare",
			amqp_get_rpc_reply(mqtt->I));
	    if (rc)
		goto exit;
	    _queuename = amqp_bytes_malloc_dup(r->queue);
assert(_queuename.bytes);
	}

	const char * _routing_key = "test queue";	/* XXX W2DO? */

	amqp_queue_bind(mqtt->I,
			_channel,
			_queuename,
			amqp_cstring_bytes(mqtt->exchange),
			amqp_cstring_bytes(_routing_key),
			_arguments);
	rc = mqchkrpc(mqtt, "queue_bind",
			amqp_get_rpc_reply(mqtt->I));
	if (rc)
	    goto exit;

	amqp_bytes_t _consumer_tag = amqp_empty_bytes;
	amqp_boolean_t _no_local = (MF_ISSET(NOLOCAL) ? true : false);
#ifdef	NOTYET
	amqp_boolean_t _no_ack = (MF_ISSET(NOACK) ? true : false);
#else
	amqp_boolean_t _no_ack = true;
#endif
	amqp_basic_consume(mqtt->I,
			_channel,
			_queuename,
			_consumer_tag,
			_no_local,
			_no_ack,
			_exclusive,
			_arguments);
	rc = mqchkrpc(mqtt, "basic_consume",
			amqp_get_rpc_reply(mqtt->I));
	if (rc)
	    goto exit;

#ifdef	NOTYET	/* XXX make asynchronous */
	for (;;) {
	    amqp_rpc_reply_t res;
	    amqp_envelope_t envelope;
	    struct timeval *_tvp = NULL;
	    int _flags = 0;

	    rc = mqchk(mqtt, "maybe_release_buffers",
			(amqp_maybe_release_buffers(mqtt->I),0));

	    rc = mqchkrpc(mqtt, "consume_message",
			(res = amqp_consume_message(mqtt->I,
					&envelope, _tvp, _flags)));
	    if (rc)
		break;

	    fprintf(stderr, "Delivery %u, exchange %.*s routingkey %.*s\n",
			(unsigned)envelope.delivery_tag,
			(int)envelope.exchange.len,
			(char *)envelope.exchange.bytes,
			(int)envelope.routing_key.len,
			(char *)envelope.routing_key.bytes);

	    if (envelope.message.properties._flags & AMQP_BASIC_CONTENT_TYPE_FLAG) {
		fprintf(stderr, "Content-type: %.*s\n",
			(int) envelope.message.properties.content_type.len,
			(char *)envelope.message.properties.content_type.bytes);
	    }
	    fprintf(stderr, "----\n");

	    amqp_dump(envelope.message.body.bytes, envelope.message.body.len);

	    amqp_destroy_envelope(&envelope);

	}
	if (rc)
	    goto exit;
#endif

	rc = RPMRC_OK;
    }

exit:
    return rc;
}

static
rpmRC amqpUnsubscribe(rpmmqtt mqtt, const char * topic)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (!rpmmqttConnect(mqtt)) {
	int _lvl = RPMLOG_DEBUG;

if (rpmIsDebug())
fprintf(stderr, "XXX %s: FIXME\n", __FUNCTION__);
	rpmlog(_lvl, "%19s: %s\n", "unsubscribe", topic);

	if (rc)
	    goto exit;
	rc = RPMRC_OK;
    }

exit:
    return rc;
}

static KEY amqpErrs[] = {
#define _ENTRY(_v)      { AMQP_STATUS_##_v, #_v, }
    _ENTRY(OK),
    _ENTRY(NO_MEMORY),
    _ENTRY(BAD_AMQP_DATA),
    _ENTRY(UNKNOWN_CLASS),
    _ENTRY(UNKNOWN_METHOD),
    _ENTRY(HOSTNAME_RESOLUTION_FAILED),
    _ENTRY(INCOMPATIBLE_AMQP_VERSION),
    _ENTRY(CONNECTION_CLOSED),
    _ENTRY(BAD_URL),
    _ENTRY(SOCKET_ERROR),
    _ENTRY(INVALID_PARAMETER),
    _ENTRY(TABLE_TOO_BIG),
    _ENTRY(WRONG_METHOD),
    _ENTRY(TIMEOUT),
    _ENTRY(TIMER_FAILURE),
    _ENTRY(HEARTBEAT_TIMEOUT),
    _ENTRY(UNEXPECTED_STATE),
#if defined(AMQP_SOCKET_CLOSED)
    _ENTRY(SOCKET_CLOSED),
#endif
#if defined(AMQP_SOCKET_INUSE)
    _ENTRY(SOCKET_INUSE),
#endif
#if defined(AMQP_UNSUPPORTED_SASL_METHOD)
    _ENTRY(BROKER_UNSUPPORTED_SASL_METHOD),
#endif
#if defined(AMQP_UNSUPPORTED)
    _ENTRY(UNSUPPORTED),
#endif
    _ENTRY(TCP_ERROR),
    _ENTRY(TCP_SOCKETLIB_INIT_ERROR),
    _ENTRY(SSL_ERROR),
    _ENTRY(SSL_HOSTNAME_VERIFY_FAILED),
    _ENTRY(SSL_PEER_VERIFY_FAILED),
    _ENTRY(SSL_CONNECTION_FAILED),
#undef	_ENTRY
    { 0, NULL },
};

static
struct mqttVec_s amqpVec = {
    .name		= "amqp",
    .uri		= "amqp://guest:guest@localhost:5672/"	/* XXX vhost? */
			        "?vhost=/"		/* XXX vhost? */
			        "&exchange=amq.direct"
			        "&queue=test queue"
			        "&verify=1"
			        "&fail_if_no_peer_cert=1"
			        "&sasl_mechanism=plain"
			        "&keepalive=60"
			        "&connection_timeout=60"
			        "&channel_max=60",
    .port		= 5672,
    .sport		= 5671,
    .prefix		= "amqp_",
    .all		= 0,
    .errs		= amqpErrs,
    .nerrs		= sizeof(amqpErrs) / sizeof(amqpErrs[0]),
    .destroy		= amqpDestroy,
    .create		= amqpCreate,
    .disconnect		= amqpDisconnect,
    .connect		= amqpConnect,
    .isconnected	= amqpIsConnected,	/* XXX */
    .unsubscribe	= amqpUnsubscribe,
    .subscribe		= amqpSubscribe,
    .unsubscribeMany	= _rpmmqttUnsubscribeMany,
    .subscribeMany	= _rpmmqttSubscribeMany,
    .sendMessage	= amqpSendMessage,
};
#endif	/* WITH_RABBITMQ */

/*==============================================================*/
#ifdef	WITH_PROTON
static
rpmRC qpidDestroy(rpmmqtt mqtt)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */
    if (mqtt->I) {
	(void) mqchk(mqtt, "messenger_stop",
		pn_messenger_stop(mqtt->I));
	(void) mqchk(mqtt, "messenger_free",
		(pn_messenger_free(mqtt->I),0));
	(void) mqchk(mqtt, "message_free",
		(pn_message_free(mqtt->M),0));
    }
    rc = RPMRC_OK;	/* XXX */
    mqtt->I = NULL;
    return rc;
}

static
rpmRC qpidCreate(rpmmqtt mqtt)
{
    static int oneshot;
    int _lvl = RPMLOG_DEBUG;
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (mqtt->I == NULL) {
	rpmlog(_lvl, "==================== %s\n", mqtt->vec->name);

	if (!oneshot) {
#ifdef	NOTYET
	    int major = AMQP_VERSION_MAJOR;
	    int minor = AMQP_VERSION_MINOR;
	    int patch = AMQP_VERSION_PATCH;
	    int version = AMQP_VERSION;
	    rpmlog(_lvl, "%19s: %d.%d.%d (%d)\n", "version",
		major, minor, patch, version);
#endif
	    oneshot++;
	}

	rpmlog(_lvl, "%19s: %s\n", "vhost", mqtt->vhost);
	rpmlog(_lvl, "%19s: %s\n", "queue", mqtt->queue);
	rpmlog(_lvl, "%19s: %s\n", "exchange", mqtt->exchange);
	rpmlog(_lvl, "%19s: %s\n", "flags", _MQTTFLAGS(mqtt->flags));
	if (mqtt->trace)
	    rpmlog(_lvl, "%19s: %d\n", "trace", mqtt->trace);

#ifdef	DYING
mqtt->u = NULL;
dumpMQTT(__FUNCTION__, mqtt);
#endif

	    rc = mqchk(mqtt, "messenger",
		((mqtt->I = pn_messenger(NULL)) == NULL));
	    if (rc)
		goto exit;
	    rc = mqchk(mqtt, "messenger_start",
		pn_messenger_start(mqtt->I));
	    if (rc)
		goto exit;
    }

    rc = RPMRC_OK;

exit:
    return rc;
}

static
rpmRC qpidDisconnect(rpmmqtt mqtt)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (rpmmqttIsConnected(mqtt) == RPMRC_OK) {
fprintf(stderr, "XXX %s: FIXME\n", __FUNCTION__);
mqtt->connected = 0;
    }
    rc = RPMRC_OK;

    return rc;
}

static
rpmRC qpidConnect(rpmmqtt mqtt)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (rpmmqttIsConnected(mqtt) == RPMRC_NOTFOUND) {
fprintf(stderr, "XXX %s: FIXME\n", __FUNCTION__);
mqtt->connected = 1;
    }
    rc = RPMRC_OK;

    return rc;
}

static
rpmRC qpidIsConnected(rpmmqtt mqtt)
{
    rpmRC rc = (mqtt->connected ? RPMRC_OK : RPMRC_NOTFOUND);
    return rc;
}

static
rpmRC qpidSendMessage(rpmmqtt mqtt, const char * topic,
		const char * s, size_t ns)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (topic == NULL)
	topic = mqtt->topic;
    if (s == NULL)
	s = "";
    if (ns == 0)
	ns = strlen(s);

    if (!rpmmqttConnect(mqtt)) {
	if (mqtt->M == NULL) {
	    rc = mqchk(mqtt, "message",
		((mqtt->M = pn_message()) == NULL));
	    if (rc)
		goto exit;
	}

fprintf(stderr, "*** %s: %s mqtt host %s topic %s\n", __FUNCTION__, topic, mqtt->host, mqtt->topic);
	const char * _address = "amqp://localhost/test";	/* XXX */
	rc = mqchk(mqtt, "message_set_address",
		pn_message_set_address(mqtt->M, _address));
	if (rc)
	    goto exit;

	{   pn_data_t * body = pn_message_body(mqtt->M);
	    rc = mqchk(mqtt, "data_put_string",
			pn_data_put_string(body, pn_bytes(ns, s)));
	    if (rc)
		goto exit;
	    rc = mqchk(mqtt, "messenger_put",
			pn_messenger_put(mqtt->I, mqtt->M));
	    if (rc)
		goto exit;
#ifdef	NOTYET
	    check(messenger);
#endif
	    rc = mqchk(mqtt, "messenger_send",
			pn_messenger_send(mqtt->I, -1));

#ifdef	NOTYET
	    check(messenger);
#endif
	}

	rpmlog(RPMLOG_DEBUG, "%s: send() exchange(%s) queue(%s) \"%.*s\"\n",
			mqtt->vec->name, mqtt->exchange, mqtt->queue, (int)ns, s);

	if (rc)
	    goto exit;
	rc = RPMRC_OK;
    }

exit:
    return rc;
}

static
rpmRC qpidSubscribe(rpmmqtt mqtt, const char * topic, int qos)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (!rpmmqttConnect(mqtt)) {
	int _lvl = RPMLOG_DEBUG;

fprintf(stderr, "XXX %s: FIXME\n", __FUNCTION__);
	rpmlog(_lvl, "%19s: %s qos(%d)\n", "subscribe", topic, qos);

	if (rc)
	    goto exit;
	rc = RPMRC_OK;
    }

exit:
    return rc;
}

static
rpmRC qpidUnsubscribe(rpmmqtt mqtt, const char * topic)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (!rpmmqttConnect(mqtt)) {
	int _lvl = RPMLOG_DEBUG;

fprintf(stderr, "XXX %s: FIXME\n", __FUNCTION__);
	rpmlog(_lvl, "%19s: %s\n", "unsubscribe", topic);

	if (rc)
	    goto exit;
	rc = RPMRC_OK;
    }

exit:
    return rc;
}

static KEY qpidErrs[] = {
#define _ENTRY(_v)      { PN_##_v, #_v, }
    _ENTRY(OK),
    _ENTRY(EOS),
    _ENTRY(ERR),
    _ENTRY(OVERFLOW),
    _ENTRY(UNDERFLOW),
    _ENTRY(STATE_ERR),
    _ENTRY(ARG_ERR),
    _ENTRY(TIMEOUT),
    _ENTRY(INTR),
    _ENTRY(INPROGRESS),
    _ENTRY(OUT_OF_MEMORY),
#undef	_ENTRY
    { 0, NULL },
};

static
struct mqttVec_s qpidVec = {
    .name		= "qpid",
    .uri		= "amqp://localhost:5672/"	/* XXX vhost? */
			        "?vhost=/"		/* XXX vhost? */
			        "&exchange=amq.direct"	/* XXX */
			        "&queue=test queue"	/* XXX */
			        "&verify=1"		/* XXX */
			        "&fail_if_no_peer_cert=1"/* XXX */
			        "&sasl_mechanism=plain"	/* XXX */
			        "&keepalive=60"
			        "&connection_timeout=60"
			        "&channel_max=60",
    .port		= 5672,
    .sport		= 5671,
    .prefix		= "pn_",
    .all		= 1,
    .errs		= qpidErrs,
    .nerrs		= sizeof(qpidErrs) / sizeof(qpidErrs[0]),
    .destroy		= qpidDestroy,
    .create		= qpidCreate,
    .disconnect		= qpidDisconnect,
    .connect		= qpidConnect,
    .isconnected	= qpidIsConnected,	/* XXX */
    .unsubscribe	= qpidUnsubscribe,
    .subscribe		= qpidSubscribe,
    .unsubscribeMany	= _rpmmqttUnsubscribeMany,
    .subscribeMany	= _rpmmqttSubscribeMany,
    .sendMessage	= qpidSendMessage,
};
#endif	/* WITH_PROTON */
/*==============================================================*/
#ifdef	WITH_ZEROMQ
static
rpmRC zmqDestroy(rpmmqtt mqtt)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */
fprintf(stderr, "XXX %s: FIXME\n", __FUNCTION__);
    mqtt->I = NULL;
    return rc;
}

static
rpmRC zmqCreate(rpmmqtt mqtt)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */
    static int oneshot;
    int _lvl = RPMLOG_DEBUG;

    rpmlog(_lvl, "==================== %s\n", mqtt->vec->name);

    if (!oneshot) {
fprintf(stderr, "XXX %s: FIXME: version oneshot\n", __FUNCTION__);
	oneshot++;
    }

#ifdef	NOTYET
    rpmlog(_lvl, "%19s: %s\n", "clientid", mqtt->clientid);
    rpmlog(_lvl, "%19s: %s qos(%d) timeout(%u msecs)\n",
		"topic", mqtt->topic, mqtt->qos, mqtt->timeout);
    rpmlog(_lvl, "%19s: type(%u) %s\n",
		"persist", mqtt->persist_type,
		(mqtt->persist_type ? persist_path : ""));
    rpmlog(_lvl, "%19s: %s\n", "flags", _MQTTFLAGS(mqtt->flags));
    if (mqtt->trace)
	rpmlog(_lvl, "%19s: %d\n", "trace", mqtt->trace);
#else
fprintf(stderr, "XXX %s: FIXME: values\n", __FUNCTION__);
#endif

mqtt->u = NULL;
dumpMQTT(__FUNCTION__, mqtt);

    if (mqtt->I == NULL) {
fprintf(stderr, "XXX %s: FIXME: handle\n", __FUNCTION__);
    }

    rc = RPMRC_OK;

    return rc;
}

static
rpmRC zmqDisconnect(rpmmqtt mqtt)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */
    if (rpmmqttIsConnected(mqtt) == RPMRC_OK) {
fprintf(stderr, "XXX %s: FIXME: connected = 0\n", __FUNCTION__);
	if (rc)
	    goto exit;
mqtt->connected = 0;
    }
    rc = RPMRC_OK;

exit:
    return rc;
}

static
rpmRC zmqConnect(rpmmqtt mqtt)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (rpmmqttIsConnected(mqtt) == RPMRC_NOTFOUND) {
fprintf(stderr, "XXX %s: FIXME: connected = 1\n", __FUNCTION__);
	if (rc)
	    goto exit;
mqtt->connected = 1;
    }
    rc = RPMRC_OK;

exit:
    return rc;
}

static
rpmRC zmqIsConnected(rpmmqtt mqtt)
{
    rpmRC rc = (mqtt->connected ? RPMRC_OK : RPMRC_NOTFOUND);
    return rc;
}

static
rpmRC zmqSendMessage(rpmmqtt mqtt, const char * topic,
		const char * s, size_t ns)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (topic == NULL)
	topic = mqtt->topic;
    if (s == NULL)
	s = "";
    if (ns == 0)
	ns = strlen(s);

    if (!rpmmqttConnect(mqtt)) {
fprintf(stderr, "XXX %s: FIXME: %s %*.s\n", __FUNCTION__, topic, (int)ns, s);
	if (rc)
	    goto exit;
	rc = RPMRC_OK;
    }

exit:
    return rc;
}

static
rpmRC zmqSubscribe(rpmmqtt mqtt, const char * topic, int qos)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (!rpmmqttConnect(mqtt)) {
	int _lvl = RPMLOG_DEBUG;

fprintf(stderr, "XXX %s: FIXME\n", __FUNCTION__);
	rpmlog(_lvl, "%19s: %s qos(%d)\n", "subscribe", topic, qos);

	if (rc)
	    goto exit;
	rc = RPMRC_OK;
    }

exit:
    return rc;
}

static
rpmRC zmqUnsubscribe(rpmmqtt mqtt, const char * topic)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (!rpmmqttConnect(mqtt)) {
	int _lvl = RPMLOG_DEBUG;

fprintf(stderr, "XXX %s: FIXME\n", __FUNCTION__);
	rpmlog(_lvl, "%19s: %s\n", "unsubscribe", topic);

	if (rc)
	    goto exit;
	rc = RPMRC_OK;
    }

exit:
    return rc;
}

static KEY zmqErrs[] = {
    { 0, NULL },
};

static
struct mqttVec_s zmqVec = {
    .name		= "zmq",
    .uri		= "zmq://localhost/",	/* XXX W2DO? */
    .port		= 0xfffe,		/* XXX W2DO? */
    .sport		= 0xfffe,		/* XXX W2DO? */
    .prefix		= "zmq_",
    .errs		= zmqErrs,		/* XXX W2DO? */
    .nerrs		= sizeof(zmqErrs) / sizeof(zmqErrs[0]),
    .destroy		= zmqDestroy,
    .create		= zmqCreate,
    .disconnect		= zmqDisconnect,
    .connect		= zmqConnect,
    .isconnected	= zmqIsConnected,	/* XXX */
    .unsubscribe	= zmqUnsubscribe,
    .subscribe		= zmqSubscribe,
    .unsubscribeMany	= _rpmmqttUnsubscribeMany,
    .subscribeMany	= _rpmmqttSubscribeMany,
    .sendMessage	= zmqSendMessage,
};
#endif	/* WITH_ZEROMQ */

/*==============================================================*/
rpmRC rpmmqttPub(rpmmqtt mqtt, const char * topic, const char *s, size_t ns)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (topic == NULL)
	topic = mqtt->topic;
    if (s == NULL)
	s = "";
    if (ns == 0)
	ns = strlen(s);

    if (rpmmqttConnect(mqtt))
	goto exit;
    {	static char _mqtt_prefix[] = "%{?_mqtt_prefix}";
	/* XXX extra space: prepend in *sendMessage or *send  */
	char * t = rpmmqttExpand(mqtt, NULL,
			_mqtt_prefix, " ", s, NULL);
	size_t nt = strlen(t);

	rc = rpmmqttSendMessage(mqtt, topic, t, nt);
	if (t) free(t);
	if (rc)
	    goto exit;
	rc = RPMRC_OK;
    }

exit:
SPEW((stderr, "<-- %s(%p,%p[%u]) rc %d\n", __FUNCTION__, mqtt, s, (unsigned)ns, (int)rc));
    return rc;
}

rpmRC rpmmqttSub(rpmmqtt mqtt, const char *s, size_t ns)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    if (ns == 0) ns = strlen(s);

    if (rpmmqttConnect(mqtt))
	goto exit;

    {	char * subtopic = rpmmqttExpand(mqtt, NULL,
		(s ? s : mqtt->topic), NULL);
	unsigned subqos = mqtt->qos;
	char *t, *te;
int _lvl = RPMLOG_DEBUG;

	/* XXX qos appended to subscription topic? */
	if ((t = strchr(subtopic, '?')) != NULL) {
	    ARGV_t qav = NULL;
	    int qac;

	    *t++ = '\0';
	    (void) argvSplit(&qav, t, ",");
	    qac = argvCount(qav);
	    for (int i = 0; i < qac; i++) {
		t = (char *)qav[i];
		if ((te = strchr(t, '=')) == NULL)
		    continue;
		if (!strncmp(t, "qos", (te - t)) && xisdigit(te[1])) {
		    subqos = strtol(te+1, NULL, 0);
		    subqos %= 3;
		    continue;
		}
	    }
	    qav = argvFree(qav);
	}

	rpmlog(_lvl, "%19s: %s qos(%u)\n", "subtopic", subtopic, subqos);

	rc = rpmmqttSubscribe(mqtt, subtopic, subqos);

	if (subtopic) free(subtopic);
	if (rc)
	    goto exit;
	rc = RPMRC_OK;
    }

exit:
SPEW((stderr, "<-- %s(%p,%p[%u]) ret %d\n", __FUNCTION__, mqtt, s, (unsigned)ns, (int)rc));
    return rc;
}

/*==============================================================*/
// *INDENT-OFF*
static rpmRC
rpmmqttInitPopt(rpmmqtt mqtt, int ac, char * const* av)
{
  struct poptOption rpmmqttConnectPoptTable[] = {
	/* AMQP -u,--url */
   { "host", 'h', POPT_ARG_STRING|POPT_ARGFLAG_SHOW_DEFAULT,	&mqtt->host, 0,
	N_("Connect to <HOST>."), N_("<HOST>") },
   { "port", 'p', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT,	&mqtt->port, 0,
	N_("Connect to network <PORT>."), N_("<PORT>") },
   { "user", 'u', POPT_ARG_STRING,	&mqtt->user, 0,
	N_("Remote <USER>."), N_("<USER>") },
   { "pw", 'P', POPT_ARG_STRING,	&mqtt->password, 0,
	N_("Remote user <PASSWORD>."), N_("<PASSWORD>") },
   { NULL, 'A', POPT_ARG_STRING,	&mqtt->_address, 0,
	N_("Connect from local <ADDR>."), N_("<ADDR>") },
   { NULL, 'S', POPT_BIT_SET,		&mqtt->flags, MQTT_FLAGS_DNSSRV,
	N_("Use SRV record to find remote host."), NULL },
   { "cafile", '\0', POPT_ARG_STRING,	&mqtt->cacertfile, 0,
	N_("CA certificate(s) from <FILE>."), N_("<FILE>") },
   { "capath", '\0', POPT_ARG_STRING,	&mqtt->_capath, 0,
	N_("CA certificate(s) in <DIR>."), N_("<DIR>") },
   { "cert", '\0', POPT_ARG_STRING,	&mqtt->certfile, 0,
	N_("Client authentication <CERT>."), N_("<CERT>") },
   { "key", '\0', POPT_ARG_STRING,	&mqtt->keyfile, 0,
	N_("Client private <KEY>."), N_("<KEY>") },
   { "ciphers", '\0', POPT_ARG_STRING,	&mqtt->ciphers, 0,
	N_("TLS ciphers <LIST> (openssl names)."), N_("<LIST>") },
   { "tls-version", '\0', POPT_ARG_STRING|POPT_ARGFLAG_SHOW_DEFAULT,&mqtt->_tls_version, 0,
	N_("TLS protocol <VERSION>."), N_("{1.2|1.1|1}") },
   { "insecure", '\0', POPT_BIT_SET,	&mqtt->flags, MQTT_FLAGS_INSECURE,
	N_("Do not verify server cert matches DNS host."), NULL },
   { "psk", '\0', POPT_ARG_STRING,	&mqtt->_psk_key, 0,
	N_("TLS-PSK mode <KEY> hex (w/o 0x)"), N_("<KEY>") },
   { "psk-identity", '\0', POPT_ARG_STRING, &mqtt->_psk_identity, 0,
	N_("TLS-PSK <IDENTITY>."), N_("<IDENTITY>") },
   { "proxy", '\0', POPT_ARG_STRING,	&mqtt->_proxy, 0,
	N_("SOCKS5 proxy <URL>."), N_("<URL>") },
    POPT_TABLEEND
  };
  struct poptOption rpmmqttProtocolPoptTable[] = {
   { "id", 'i', POPT_ARG_STRING|POPT_ARGFLAG_SHOW_DEFAULT,&mqtt->_clientid, 0,
	N_("MQTT client <ID>."), N_("<ID>") },
   { "id-prefix", 'I', POPT_ARG_STRING,	&mqtt->idprefix, 0,
	N_("Use <IDPREFIX>-%{pid} for clientid."), N_("<IDPREFIX>") },
   { "keepalive", 'k', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT, &mqtt->keepalive, 0,
	N_("Keep alive in <SECS>."), N_("<SECS>") },
   { "max-inflight", 'M', POPT_ARG_INT,	&mqtt->max_inflight, 0,
	N_("Permit <MAX> inflight messages."), N_("<MAX>") },
   { "qos", 'q', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT,	&mqtt->qos, 0,
	N_("MQTT <QOS> level."), N_("<QOS>") },
	/* XXX -r,--root */
   { "retain", 'r', POPT_BIT_SET,		&mqtt->flags, MQTT_FLAGS_RETAIN,
	N_("Retain the message on the host."), NULL },
   { "topic", 't', POPT_ARG_ARGV,		&mqtt->topics, 0,
	N_("MQTT pub/sub <TOPIC>."), N_("<TOPIC>") },
   { "protocol", 'V', POPT_ARG_STRING|POPT_ARGFLAG_SHOW_DEFAULT,&mqtt->protocol_version, 0,
	N_("MQTT protocol <VERSION>"), N_("{auto|311|31}") },
   { "will-payload", '\0', POPT_ARG_STRING|POPT_ARGFLAG_SHOW_DEFAULT,&mqtt->will_message, 0,
	N_("Will payload <MESSAGE> ."), N_("<MESSAGE>") },
   { "will-qos", '\0', POPT_ARG_INT,	&mqtt->will_qos, 0,
	N_("Will <QOS> level."), N_("<QOS>") },
   { "will-topic", '\0', POPT_ARG_STRING,	&mqtt->will_topic, 0,
	N_("Will <TOPIC> to publish to."), N_("<TOPIC>") },
   { "will-retain", '\0', POPT_BIT_SET,	&mqtt->flags, MQTT_FLAGS_WILL_RETAIN,
	N_("Retain the client Will."), NULL },
    POPT_TABLEEND
  };
  struct poptOption mqttSubscribePoptTable[] = {
   { "clean", 'c', POPT_ARG_VAL|POPT_ARGFLAG_XOR,&mqtt->flags, MQTT_FLAGS_CLEAN,
	N_("(sub) Do not clean session."), NULL },
   { "count", 'C', POPT_ARG_INT,	&mqtt->max_msg_count, 0,
	N_("(sub) Disconnect/exit after <MSGCNT>."), N_("<MSGCNT>") },
   { "eol", 'N', POPT_ARG_VAL|POPT_ARGFLAG_XOR,	&mqtt->flags, MQTT_FLAGS_EOL,
	N_("(sub) Do not print NL after messages."), NULL },
   { "nostale", '\0', POPT_BIT_SET,	&mqtt->flags, MQTT_FLAGS_NOSTALE,
	N_("(sub) Do not print stale messages."), NULL },
   { "filter-out", 'T', POPT_ARG_ARGV,	&mqtt->_filter_out, 0,
	N_("(sub) Filter out topics."), NULL },
    POPT_TABLEEND
  };
#ifdef	WITH_RABBITMQ
  struct poptOption amqpPoptTable[] = {
   { "vhost", '\0', POPT_ARG_ARGV,		&mqtt->vhost, 0,
	N_("AMQP <vhost>"), N_("<vhost>") },
    { "queue", 'Q', POPT_ARG_STRING,		&mqtt->queue, 0,
	N_("AMQP <queue>"), N_("<queue>") },
   { "exchange", 'X', POPT_ARG_STRING,		&mqtt->exchange, 0,
	N_("AMQP <exchange>"), N_("<exchange>") },
   { "routing-key", 'R', POPT_ARG_STRING,	&mqtt->_routing_key, 0,
	N_("AMQP <route>"), N_("<route>") },
   { "reply-to", '\0', POPT_ARG_STRING,		&mqtt->_reply_to, 0,
	N_("AMQP message <reply-to>"), N_("<reply-to>") },
   { "content-type", '\0', POPT_ARG_STRING,	&mqtt->_content_type, 0,
	N_("AMQP message content-type"), N_("<T>") },
   { "content-encoding", '\0', POPT_ARG_STRING,	&mqtt->_content_encoding, 0,
	N_("AMQP message content-encoding"), N_("<E>") },
   { "heartbeat", '\0', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT, &mqtt->keepalive, 0,
	N_("AMQP heartbeat in <SECS>."), N_("<SECS>") },
   { "persistent", '\0', POPT_ARG_VAL,		&mqtt->_delivery, 2,
	N_("AMQP persistent delivery mode"), NULL },
   { "prefetch-count", '\0', POPT_ARG_INT,	&mqtt->_prefetch_count, 0,
	N_("AMQP receive batch <limit>"), N_("limit") },

   { "exclusive", '\0', POPT_BIT_SET,	&mqtt->flags, MQTT_FLAGS_EXCLUSIVE,
	N_("AMQP mark queue as exclusive"), NULL },
   { "no-ack", '\0', POPT_BIT_SET,	&mqtt->flags, MQTT_FLAGS_NOACK,
	N_("AMQP consume in no-ack mode"), NULL },
    { "if-unused", '\0', POPT_BIT_SET,	&mqtt->flags, MQTT_FLAGS_IFUNUSED,
	N_("AMQP delete queue only if unused"), NULL },
    { "if-empty", '\0', POPT_BIT_SET,	&mqtt->flags, MQTT_FLAGS_IFEMPTY,
	N_("AMQP delete queue only if empty"), NULL },
    POPT_TABLEEND
  };
#endif	/* WITH_RABBITMQ */
  struct poptOption rpmmqttOptionsTable[] = {
   { "buffer", 'b', POPT_BIT_SET,	&mqtt->flags, MQTT_FLAGS_BUFFER,
	N_("Collect input messages in buffer."), NULL },
   { "debug", 'd', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &mqtt->debug, -1,
	N_("Debug spewage."), NULL },
   { "file", 'f', POPT_ARG_ARGV,	&mqtt->ifn, 0,
	N_("Send <FILE> as message."), N_("<FILE>") },
   { "lines", 'l', POPT_BIT_SET,	&mqtt->flags, MQTT_FLAGS_LINES,
	N_("Send messages line-by-line."), NULL },
   { "message", 'm', POPT_ARG_ARGV,	&mqtt->msgs, 0,
	N_("MQTT payload <MESSAGE> to send."), N_("<MESSAGE>") },
   { "null", 'n', POPT_BIT_SET,		&mqtt->flags, MQTT_FLAGS_EMPTY,
	N_("Send a null (zero length) message."), NULL },
   { "output", 'o', POPT_ARG_STRING,	&mqtt->ofn, 0,
	N_("Write input mssages to <FILE>."), N_("<FILE>") },
	/* MQTT -s */

   { NULL, '\0', POPT_ARG_INCLUDE_TABLE, &rpmmqttConnectPoptTable, 0,
	N_("Connect:"),
	NULL },

   { NULL, '\0', POPT_ARG_INCLUDE_TABLE, &rpmmqttProtocolPoptTable, 0,
	N_("MQTT Protocol:"),
	NULL },

   { NULL, '\0', POPT_ARG_INCLUDE_TABLE, &mqttSubscribePoptTable, 0,
	N_("MQTT Subscribe:"),
	NULL },

#ifdef	WITH_RABBITMQ
   { NULL, '\0', POPT_ARG_INCLUDE_TABLE, &amqpPoptTable, 0,
	N_("AMQP Options:"),
	NULL },
#endif

   { NULL, '\0', POPT_ARG_INCLUDE_TABLE, rpmioAllPoptTable, 0,
	N_("Common options for all rpmio executables:"),
	NULL },

    POPT_AUTOALIAS
    POPT_AUTOHELP

    POPT_TABLEEND
  };
    static int _popt_context_flags = 0; /* XXX POPT_CONTEXT_POSIXMEHARDER */
    poptContext con = NULL;
    rpmRC rc;
    int xx;

SPEW((stderr, "--> %s(%p,%p[%d])\n", __FUNCTION__, mqtt, av, ac));

    con = poptGetContext(av[0], ac, (const char **)av,
		rpmmqttOptionsTable, _popt_context_flags);

    while ((xx = poptGetNextOpt(con)) > 0)
    switch (xx) {
    default:
	fprintf(stderr, _("%s: option table misconfigured (%d)\n"),
                __FUNCTION__, xx);
assert(xx == 0);
	break;
    }

    mqtt->av = NULL;
    (void) argvAppend((ARGV_t *)&mqtt->av, poptGetArgs(con));

    mqtt->ac = argvCount((ARGV_t)mqtt->av);
    rc = RPMRC_OK;

    if (con)
	(void) poptFreeContext(con);

SPEW((stderr, "<-- %s(%p,%p[%d]) rc %d\n", __FUNCTION__, mqtt, av, ac, rc));
    return rc;
}
// *INDENT-ON*

static rpmRC rpmmqttInitMacros(rpmmqtt mqtt)
{
    rpmRC rc = RPMRC_OK;

	/* XXX MQTT_FLAGS_CLEAN */
	/* XXX MQTT_FLAGS_EOL */
	/* XXX MQTT_FLAGS_NOSTALE */
	/* XXX MQTT_FLAGS_LINES */
	/* XXX MQTT_FLAGS_EMPTY */
	/* XXX MQTT_FLAGS_DNSSRV */
	/* XXX MQTT_FLAGS_INSECURE */
	/* XXX MQTT_FLAGS_RETAIN */
	/* XXX MQTT_FLAGS_WILL_RETAIN */
static const char _mqtt_qos[] = "%{?_mqtt_qos}%{!?_mqtt_qos:2}";
    mqtt->qos = (rpmmqttExpandNumeric(mqtt, _mqtt_qos) % 3);
static const char _mqtt_timeout[] = "%{?_mqtt_timeout}%{!?_mqtt_timeout:10000}";
    mqtt->timeout = rpmmqttExpandNumeric(mqtt, _mqtt_timeout);
static const char _mqtt_max_inflight[] = "%{?_mqtt_max_inflight}%{!?_mqtt_max_inflight:20}";
    mqtt->max_inflight = rpmmqttExpandNumeric(mqtt, _mqtt_max_inflight);
static const char _mqtt_keepalive[] = "%{?_mqtt_keepalive}%{!?_mqtt_keepalive:60}";
    mqtt->keepalive = rpmmqttExpandNumeric(mqtt, _mqtt_keepalive);
static const char _mqtt_trace[] = "%{?_mqtt_trace}%{!?_mqtt_trace:0}";
    mqtt->trace = rpmmqttExpandNumeric(mqtt, _mqtt_trace);

#ifdef	DYING
static const char _mqtt_topic[] =
	"%{?_mqtt_topic}%{!?_mqtt_topic:rpm/%{pid}/mqtt}";
    (void) rpmmqttExpand(mqtt, &mqtt->topic,
			_mqtt_topic, NULL);
static const char _mqtt_clientid[] =
	"%{?_mqtt_clientid}%{!?_mqtt_clientid:rpm}";
    (void) rpmmqttExpand(mqtt, &mqtt->clientid,
		_mqtt_clientid, NULL);
#endif

SPEW((stderr, "<-- %s(%p) rc %d\n", __FUNCTION__, mqtt, rc));
    return rc;
}

static rpmRC rpmmqttInitURIQuery(rpmmqtt mqtt, const char *query)
{
    ARGV_t av = NULL;
    int ac;
    const char *t;
    const char *te;
    rpmRC rc = RPMRC_OK;

    /* https://en.wikipedia.org/wiki/Query_string */
    /* XXX ampersand? semi-colon? comma? whitespace? newlines? */
    (void) argvSplit(&av, query, "&;,");
#ifdef	DYING
argvPrint(__FUNCTION__, av, NULL);
#endif
    ac = argvCount(av);
    for (int i = 0; i < ac; i++) {
	t = av[i];
	if (!(t && *t))
	    continue;
	if ((te = strchr(t, '=')) == NULL)
	    te += strlen(t);
	if (!strncmp(t, "clean", (te - t))) {
	    *((int *)&mqtt->flags) |= MQTT_FLAGS_CLEAN;
	    continue;
	}
	/* XXX MQTT_FLAGS_EOL */
	/* XXX MQTT_FLAGS_NOSTALE */
	/* XXX MQTT_FLAGS_LINES */
	/* XXX MQTT_FLAGS_EMPTY */
	/* XXX MQTT_FLAGS_DNSSRV */
	/* XXX MQTT_FLAGS_INSECURE */
	if (!strncmp(t, "retain", (te - t))) {
	    *((int *)&mqtt->flags) |= MQTT_FLAGS_RETAIN;
	    continue;
	}
	/* XXX MQTT_FLAGS_WILL_RETAIN */

#define	URIQUERY_SET_INT(OPT, VAL)				\
    if (!strncmp(t, #OPT, (te - t))) {				\
	mqtt->OPT = (te[0] == '=' && xisdigit(te[1])		\
		? strtoul(te+1, NULL, 0) : VAL);		\
	continue;						\
    }
    /* XXX rpmExpand() ??? */
#define	URIQUERY_SET_STR(OPT, VAL)				\
    if (!strncmp(t, #OPT, (te - t))) {				\
	if (mqtt->OPT) free((void *)mqtt->OPT);				\
	mqtt->OPT = xstrdup(te[0] == '=' ? te+1 : VAL);		\
	continue;						\
    }

	URIQUERY_SET_INT(qos,		2)
	URIQUERY_SET_INT(timeout,	10000)
	URIQUERY_SET_INT(max_inflight,	20)
	URIQUERY_SET_INT(keepalive,	60)	/* XXX amqp uses "heartbeat" */
	/* XXX set global mqtt->trace from URI query string? */
	URIQUERY_SET_INT(trace,		4)

	URIQUERY_SET_STR(vhost,		"/")
	URIQUERY_SET_STR(queue,		"test queue")
	URIQUERY_SET_STR(exchange,	"amq.direct")
	URIQUERY_SET_STR(cacertfile,	"/etc/pki/CA/cacert.pem")
	URIQUERY_SET_STR(certfile,	"/etc/pki/foo/certfile.pem")
	URIQUERY_SET_STR(keyfile,	"/etc/pki/foo/private/keyfile.pem")

#undef	URIQUERY_SET_STR
#undef	URIQUERY_SET_INT

	/* XXX notyet overrides have extra underscore */
#define	URIQUERY_SET_INT(OPT, VAL)				\
    if (!strncmp(t, #OPT, (te - t))) {				\
	mqtt->_##OPT = (te[0] == '=' && xisdigit(te[1])		\
		? strtoul(te+1, NULL, 0) : VAL);		\
	continue;						\
    }
#define	URIQUERY_SET_STR(OPT, VAL)				\
    if (!strncmp(t, #OPT, (te - t))) {				\
	if (mqtt->_##OPT) free((void *)mqtt->_##OPT);			\
	mqtt->_##OPT = xstrdup(te[0] == '=' ? te+1 : VAL);	\
	continue;						\
    }

	URIQUERY_SET_INT(verify,	1)		/* XXX mqtt->insecure */

	URIQUERY_SET_INT(fail_if_no_peer_cert, 1)
	URIQUERY_SET_STR(sasl_mechanism, "plain")	/* XXX argv */
	URIQUERY_SET_INT(connection_timeout, 60)
	URIQUERY_SET_INT(channel_max,	60)

#undef	URIQUERY_SET_STR
#undef	URIQUERY_SET_INT

fprintf(stderr, "*** %s: unknown URI query field: \"%s\"\n", __FUNCTION__, t);
#ifdef	NOTYET	/* XXX */
	rc = RPMRC_FAIL;
#endif
    }
    av = argvFree(av);

SPEW((stderr, "<-- %s(%p,%s) rc %d\n", __FUNCTION__, mqtt, query, rc));
    return rc;
}

static rpmRC rpmmqttInitURIFragment(rpmmqtt mqtt, const char *fragment)
{
    ARGV_t av = NULL;
    int ac;
    const char *t;
#ifdef	UNUSED
    const char *te;
#endif
    rpmRC rc = RPMRC_OK;

    /* https://en.wikipedia.org/wiki/Fragment_identifier */
    /* XXX GOOG AJAX search (deprecated) */
    if (*fragment == '!')
	fragment++;
    /* XXX ampersand? semi-colon? comma? whitespace? newlines? */
    (void) argvSplit(&av, fragment, "&;,");
#ifdef	DYING
argvPrint(__FUNCTION__, av, NULL);
#endif

    ac = argvCount(av);
    for (int i = 0; i < ac; i++) {
	t = av[i];
	if (!(t && *t))
	    continue;
#ifdef	UNUSED
	if ((te = strchr(t, '=')) == NULL)
	    te += strlen(t);
#endif
	(void) argvAdd(&mqtt->subtopics, t);
    }
    av = argvFree(av);

SPEW((stderr, "<-- %s(%p,%s) rc %d\n", __FUNCTION__, mqtt, fragment, rc));
    return rc;
}

static rpmRC rpmmqttInitURI(rpmmqtt mqtt, const char *url)
{
    urlinfo u = NULL;
    rpmRC rc = RPMRC_OK;

    /* -- Set unspecified MQTT options from the URI parameters. */
    mqtt->ut = (urltype) urlSplit(url, &u);
    mqtt->u = u;
#ifdef	DYING
dumpU(__FUNCTION__, u);
#endif

    if (u->scheme == NULL
     || !strcmp(u->scheme, "mqtt")
     || !strcmp(u->scheme, "mqtts"))
    {
	if (u->portstr == NULL) {
	    u->portstr = u->scheme && !strcmp(u->scheme, "mqtts")
		? xstrdup("8883") : xstrdup("1883");
	
	}
    }

    if (u->user != NULL) {
if (mqtt->user) free((void *)mqtt->user);
	mqtt->user = xstrdup(u->user);
    }
    if (u->password != NULL) {
if (mqtt->password) free((void *)mqtt->password);
	mqtt->password = xstrdup(u->password);
    }
    if (u->host != NULL) {
if (mqtt->host) free((void *)mqtt->host);
	mqtt->host = xstrdup(u->host);
    }
    if (u->portstr != 0) {	/* XXX u->port? */
	mqtt->port = strtoul(u->portstr, NULL, 0);
     }

#ifdef	DYING
    (void) rpmmqttExpand(mqtt, &mqtt->uri,
		"tcp://", u->host, ":", u->portstr, NULL);
#endif

    {	const char * topic;
	(void) urlPath(u->url, &topic);
	if (topic && *topic == '/')	/* XXX not for amqp:// vhost */
	    topic++;
	if (topic && *topic) {
	    (void) rpmmqttExpand(mqtt, &mqtt->topic,
		topic, NULL);
	}
    }

    if (u->query && rpmmqttInitURIQuery(mqtt, u->query))
	rc = RPMRC_FAIL;

    if (u->fragment && rpmmqttInitURIFragment(mqtt, u->fragment))
	rc = RPMRC_FAIL;

SPEW((stderr, "<-- %s(%p,%s) rc %d\n", __FUNCTION__, mqtt, url, rc));
    return rc;
}

static rpmRC rpmmqttInit(rpmmqtt mqtt, int ac, const char ** av,
		mqttFlags flags)
{
    const char * uri = NULL;
    rpmRC rc;

SPEW((stderr, "--> %s(%p,%p[%d],0x%x)\n", __FUNCTION__, mqtt, av, ac, flags));

    static mqttFlags _flags = MQTT_FLAGS_DEFAULT;	/* CLEAN|EOL */
assert(mqtt);
    mqtt->flags = (flags ? flags : _flags);

    /* -- Set module vector and program name/mode from av[0] */
assert(av && av[0]);
if (mqtt->_progname) free((void *)mqtt->_progname);
if (mqtt->_progmode) free((void *)mqtt->_progmode);
    {	const char *arg0 = av[0];
	const char * _progname;
	const char * _progmode;

	if ((_progname = strrchr(arg0, '/')) != NULL) _progname++;
	else _progname = arg0;
	if (!strncmp(_progname, "lt-", sizeof("lt-")-1))
	     _progname += sizeof("lt-")-1;

	/* XXX initialized based on %__mqtt_module */
	/* XXX initialized based on av[1] */
	/* XXX initialized based on URI scheme */
#ifdef	WITH_PAHO
	if (mqtt->vec == NULL
	 && (strstr(_progname, "paho") || strstr(_progname, "mqtt")))
	    mqtt->vec = &pahoVec;
#endif
#ifdef	WITH_MOSQUITTO
	if (mqtt->vec == NULL
	 && (strstr(_progname, "mosquitto") || strstr(_progname, "mosq") || strstr(_progname, "mqtt")))
	    mqtt->vec = &mosqVec;
#endif
#ifdef	WITH_RABBITMQ
	if (mqtt->vec == NULL
	 && (strstr(_progname, "rabbitmq") || strstr(_progname, "amqp")))
	    mqtt->vec = &amqpVec;
#endif
#ifdef	WITH_PROTON
	if (mqtt->vec == NULL
	 && (strstr(_progname, "proton") || strstr(_progname, "qpid") || strstr(_progname, "amqp")))
	    mqtt->vec = &qpidVec;
#endif
#ifdef	WITH_ZEROMQ
	if (mqtt->vec == NULL
	 && (strstr(_progname, "zeromq") || strstr(_progname, "zmq")))
	    mqtt->vec = &zmqVec;
#endif
	/* XXX mqtt->vec default? */

	_progmode = (strstr(_progname, "sub") ? "sub" : "pub");
	mqtt->_progname = xstrdup(_progname);
	mqtt->_progmode = xstrdup(_progmode);
    }
assert(mqtt->vec);

    /* -- Initialize oddball values. */
if (_rpmmqtt_debug < 0)
fprintf(stderr, "======= Initialize oddball values.\n");

	/* XXX FIXME: needed for --help POPT_ARGFLAG_DEFAULT spewage only */
    (void) rpmmqttExpand(mqtt, &mqtt->protocol_version,
		"auto", NULL);
    (void) rpmmqttExpand(mqtt, &mqtt->will_message,
		"", NULL);
    (void) rpmmqttExpand(mqtt, &mqtt->_tls_version,
		"1.2", NULL);
	/* XXX FIXME:  __progname? */
    (void) rpmmqttExpand(mqtt, &mqtt->idprefix,
		"rpm", NULL);
    (void) rpmmqttExpand(mqtt, &mqtt->_clientid,
		mqtt->idprefix, "-%%{pid}", NULL);

    /* -- Initialize values from default URI. */
assert(mqtt && mqtt->vec);
    uri = rpmmqttExpand(mqtt, NULL,
		"%{?_mqtt_uri}%{!?_mqtt_uri:",
			(mqtt->vec->uri ? mqtt->vec->uri : ""),
			"}", NULL);
    if (uri && *uri) {
if (_rpmmqtt_debug < 0)
fprintf(stderr, "======= Initialize values from default uri\n\t%s\n", uri);
	rc = rpmmqttInitURI(mqtt, uri);
	if (rc)
	    goto exit;
    }

    /* -- Override with values from macros (if any). */
if (_rpmmqtt_debug < 0)
fprintf(stderr, "======= Override with values from macros.\n");
    rc = rpmmqttInitMacros(mqtt);
    if (rc)
        goto exit;

    /* -- Process the explicit options/arguments. */
    /* XXX don't bother if no options/args? */
if (_rpmmqtt_debug < 0)
fprintf(stderr, "======= Process explicit options/arguments.\n");
    rc = rpmmqttInitPopt(mqtt, ac, (char *const *)av);
    if (rc)
        goto exit;

    /* -- Set message input/output modes from flags/options. */
    if (mqtt->msgs)
	mqtt->msg_input = MQTT_INPUT_OPTIONS;
    else if (mqtt->ifn)
	mqtt->msg_input = MF_ISSET(LINES)
		? MQTT_INPUT_FILES_LINES : MQTT_INPUT_FILES;
    else
	mqtt->msg_input = MQTT_INPUT_UNKNOWN;

    if (mqtt->ofn) {
	mqtt->msg_output = strcmp(mqtt->ofn, "-")
		? MQTT_OUTPUT_FILE : MQTT_OUTPUT_STDOUT;
    } else {
	mqtt->msg_output = MQTT_OUTPUT_STDOUT;
	mqtt->ofn = xstrdup("-");
    }

    /* -- XXX Ensure topic/clientid are set. */
assert(mqtt->_clientid);
    if (mqtt->clientid == NULL) {
if (_rpmmqtt_debug < 0)
fprintf(stderr, "*** FIXME: mqtt->clientid\n");
	mqtt->clientid = rpmmqttExpand(mqtt, NULL,
		mqtt->_clientid, NULL);
    }
    if (mqtt->topic == NULL) {
assert(mqtt->idprefix);
if (_rpmmqtt_debug < 0)
fprintf(stderr, "*** FIXME: mqtt->topic\n");
	mqtt->topic = rpmmqttExpand(mqtt, NULL,
		mqtt->idprefix, "/mqtt", NULL);
    }
	
    rc = RPMRC_OK;

exit:
    if (uri) free((void *)uri);
SPEW((stderr, "<-- %s(%p,%p[%d],0x%x) rc %d\n", __FUNCTION__, mqtt, av, ac, flags, rc));
    return rc;
}

static rpmRC rpmmqttInitSubscribe(rpmmqtt mqtt, const char ** topics)
{
    rpmRC rc = RPMRC_FAIL;	/* assume failure */
    int nsubs = 0;

#ifdef	DYING
argvPrint(__FUNCTION__, (ARGV_t)topics, NULL);
#endif
    if (topics) {
	nsubs = argvCount((ARGV_t)topics);
	rc = rpmmqttSubscribeMany(mqtt, nsubs, (char **)topics);
    } else
	rc = RPMRC_OK;
SPEW((stderr, "<-- %s(%p,%p[%d]) rc %d\n", __FUNCTION__, mqtt, topics, nsubs, rc));
    return rc;
}

static rpmRC rpmmqttInitPublish(rpmmqtt mqtt, const char ** topics)
{
    int ntopics = argvCount((ARGV_t)topics);
    int nmsgs = argvCount((ARGV_t)mqtt->msgs);
    int nifn = argvCount(mqtt->ifn);
    const char * ifn = NULL;
    rpmiob iob = NULL;
    const char * topic = NULL;
    const char * s = NULL;
    int ns = 0;
    ARGV_t lav = NULL;
    int lac = 0;
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

    switch (mqtt->msg_input) {
    case MQTT_INPUT_UNKNOWN:
    default:
	break;
    case MQTT_INPUT_OPTIONS:
	for (int i = 0; i < nmsgs; i++) {
	    s = mqtt->msgs[i];
	    ns = strlen(s);
	    if (ntopics <= 0)
		rc = rpmmqttPub(mqtt, topic, s, ns);
	    else
	    for (int j = 0; j < ntopics; j++) {
		topic = topics[j];
		rc = rpmmqttPub(mqtt, topic, s, ns);
		if (rc)
		    break;
	    }
	    if (rc)
		goto exit;
	}
	break;
    case MQTT_INPUT_FILES:
	for (int i = 0; i < nifn; i++) {
	    ifn = mqtt->ifn[i];
	    if (rpmiobSlurp(ifn, &iob) == 0) {
		s = rpmiobStr(iob);
		ns = rpmiobLen(iob);
		/* XXX trim */
		if (ns == 0)	/* XXX skip empty files? */
		    continue;
		if (ntopics <= 0)
		    rc = rpmmqttPub(mqtt, topic, s, ns);
		else
		for (int j = 0; j < ntopics; j++) {
		    topic = topics[j];
		    rc = rpmmqttPub(mqtt, topic, s, ns);
		    if (rc)
			break;
		}
		if (rc)
		    goto exit;
	    }
	    iob = rpmiobFree(iob);
	}
	break;
    case MQTT_INPUT_FILES_LINES:
	for (int i = 0; i < nifn; i++) {
	    ifn = mqtt->ifn[i];
	    if (rpmiobSlurp(ifn, &iob) == 0) {
		rc = argvSplit(&lav, rpmiobStr(iob), "\n\r");
		lac = argvCount(lav);
		for (int j = 0; j < lac; j++) {
		    s = lav[j];
		    ns = strlen(s);
		    /* XXX trim */
		    if (ns == 0)	/* XXX skip empty lines? */
			continue;
		    if (ntopics <= 0)
			rc = rpmmqttPub(mqtt, topic, s, ns);
		    else
		    for (int k = 0; k < ntopics; k++) {
			topic = topics[k];
			rc = rpmmqttPub(mqtt, topic, s, ns);
			if (rc)
			    break;
		    }
		    if (rc)
			goto exit;
		}
		lav = argvFree(lav);
	    }
	    iob = rpmiobFree(iob);
	}
	break;
    }
    rc = RPMRC_OK;
exit:
    return rc;
}

/*==============================================================*/
static void rpmmqttFini(void * _mqtt)
{
    rpmmqtt mqtt = (rpmmqtt) _mqtt;

    (void) rpmmqttDisconnect(mqtt);
    (void) rpmmqttDestroy(mqtt);

#define	FINI_FREE(VAR)		{ if (mqtt->VAR) free((void *)mqtt->VAR); }
#define	FINI_FREE_ARGV(VAR)	{ mqtt->VAR = argvFree((ARGV_t)mqtt->VAR); }
/* ========== */
    FINI_FREE(_progname)
    FINI_FREE(_progmode)

    FINI_FREE(_address)
    FINI_FREE(host)
    FINI_FREE(_clientid)
    FINI_FREE(idprefix)
    FINI_FREE_ARGV(msgs)
    FINI_FREE(password)
    FINI_FREE_ARGV(topics)
    FINI_FREE_ARGV(subtopics)
    FINI_FREE_ARGV(_filter_out)
    FINI_FREE(user)
    FINI_FREE(protocol_version)	/* XXX malloc? */

    FINI_FREE(will_message)
    FINI_FREE(will_topic)
    FINI_FREE(cacertfile)
    FINI_FREE(_capath)
    FINI_FREE(certfile)
    FINI_FREE(keyfile)
    FINI_FREE(ciphers)
    FINI_FREE(_tls_version)	/* XXX malloc? */
    FINI_FREE(_psk_key)
    FINI_FREE(_psk_identity)
    FINI_FREE(_proxy)
/* ========== */
    FINI_FREE(vhost)
    FINI_FREE(queue)
    FINI_FREE(exchange)
    FINI_FREE(_routing_key)
    FINI_FREE(_reply_to)
    FINI_FREE(_content_type)
    FINI_FREE(_content_encoding)
/* ========== */
    FINI_FREE(_sasl_mechanism)
/* ========== */

    mqtt->I = NULL;
    FINI_FREE(uri)
    FINI_FREE(topic)
    FINI_FREE(clientid)
    FINI_FREE(persist_path)
    FINI_FREE(persist_ctx)
    FINI_FREE(cachedn)

    FINI_FREE_ARGV(ifn)
    if (mqtt->ifd)
	(void) Fclose(mqtt->ifd);
    mqtt->ifd = NULL;
    FINI_FREE(ofn)
    if (mqtt->ofd)
	(void) Fclose(mqtt->ofd);
    mqtt->ofd = NULL;
    mqtt->iob = rpmiobFree(mqtt->iob);

    FINI_FREE(serverURI)
#undef	_FINI_FREE

    if (mqtt->mc)
	rpmmqttFreeMacros(mqtt);

    mqtt->av = (char **)argvFree((ARGV_t)mqtt->av);
    mqtt->flags = MQTT_FLAGS_NONE;
}

#define	rpmmqttDbug	NULL
#define	rpmmqttInit	NULL
RPMIOPOOL_INTERP_MODULE(mqtt)
#undef	rpmmqttInit

rpmmqtt rpmmqttNew(char ** av, uint32_t flags)
{
    static const char * _av[] = { (char *) "mqtt", NULL };
    static char ** _avp;
    rpmmqtt mqtt = NULL;

    /* XXX quick-n-dirty means to instantiate _rpmmqttI with parent argv. */
    if ((flags & 0x80000000) && _rpmmqttI == NULL)
	_avp = av;
    mqtt = (flags & 0x80000000)
	? rpmmqttI() : rpmmqttGetPool(_rpmmqttPool);

    if (av == NULL) av = _avp;
    if (av == NULL) av = (char **) _av;

#ifdef	DYING
argvPrint(__FUNCTION__, (ARGV_t)av, NULL);
#endif
    int ac = argvCount((ARGV_t)av);
    flags &= ~0x80000000;
    rpmRC rc = rpmmqttInit(mqtt, ac, (const char **)av, (mqttFlags)flags);
if (rc) fprintf(stderr, "*** %s:%u rc %d\n", __FUNCTION__, __LINE__, rc);
assert(mqtt->_progname);
assert(mqtt->_progmode);

    {

	rc = rpmmqttCreate(mqtt);
if (rc)
fprintf(stderr, "*** %s:%u rc %d\n", __FUNCTION__, __LINE__, rc);

	/* Prepare for subscription delivery. */
	if (MF_ISSET(BUFFER) && mqtt->iob == NULL)
	    mqtt->iob = rpmiobNew(0);
	if (mqtt->ofd)			/* XXX memleaks on recursion */
	    (void) Fclose(mqtt->ofd);
	switch (mqtt->msg_output) {
	default:
	case MQTT_OUTPUT_UNKNOWN:
	    break;
	case MQTT_OUTPUT_STDOUT:
	case MQTT_OUTPUT_FILE:
	    mqtt->ofd = Fopen(mqtt->ofn, "a");	/* XXX always append? */
	    if (mqtt->ofd == NULL || Ferror(mqtt->ofd)) {
		if (mqtt->ofd)
		    (void) Fclose(mqtt->ofd);
		mqtt->ofd = NULL;
	    }
	    break;
	case MQTT_OUTPUT_CALLBACK:
	    break;
	}

	/* XXX If any topic has a wild card, then switch to "sub" mode. */
	{
	    const char * topic;
	    int bingo = 0;
	
	    topic = mqtt->topic;
	    if (topic && (strstr(topic, "+") || strstr(topic, "#")))
		bingo = 1;
	    else {
		int ntopics = argvCount((ARGV_t)mqtt->topics);
		for (int i = 0; i < ntopics; i++) {
		    topic = mqtt->topics[i];
		    if (!(strstr(topic, "+") || strstr(topic, "#")))
			continue;
		    bingo = 1;
		    break;
		}
	    }
	    if (bingo) {
		if (mqtt->_progmode) free((void *)mqtt->_progmode);
		mqtt->_progmode = xstrdup("sub");
	    }
	}

	/* Execute in either "sub" or "pub" mode. */
	if (mqtt->_progmode && !strcmp(mqtt->_progmode, "sub")) {
	    /* Subscribe to topics (if any). */
	    if (mqtt->topic) {
		rc = rpmmqttSub(mqtt, mqtt->topic, strlen(mqtt->topic));
if (rc) fprintf(stderr, "*** %s:%u rc %d\n", __FUNCTION__, __LINE__, rc);
	    }

	    rc = rpmmqttInitSubscribe(mqtt, mqtt->subtopics);
if (rc) fprintf(stderr, "*** %s:%u rc %d\n", __FUNCTION__, __LINE__, rc);
	    rc = rpmmqttInitSubscribe(mqtt, mqtt->topics);
if (rc) fprintf(stderr, "*** %s:%u rc %d\n", __FUNCTION__, __LINE__, rc);

	    /* Wait for messages ... */
	    while (mqtt->max_msg_count <= 0
		|| mqtt->msg_count < mqtt->max_msg_count)
	    {
		sleep(1);
	    }
	} else {
	    /* Subscribe to initial topics (if any). */
	    rc = rpmmqttInitSubscribe(mqtt, mqtt->subtopics);
if (rc) fprintf(stderr, "*** %s:%u rc %d\n", __FUNCTION__, __LINE__, rc);

	    /* Publish any initial messages (if any). */
	    rc = rpmmqttInitPublish(mqtt, mqtt->topics);
if (rc) fprintf(stderr, "*** %s:%u rc %d\n", __FUNCTION__, __LINE__, rc);
	}
    }

    return rpmmqttLink(mqtt);
}

rpmRC rpmmqttRun(rpmmqtt mqtt, const char * str, const char ** resultp)
{
    const char *msg = NULL;
    rpmRC rc = RPMRC_FAIL;

SPEW((stderr, "==> %s(%p,\"%s\",%p)\n", __FUNCTION__, mqtt, str, resultp));

    if (mqtt == NULL) mqtt = rpmmqttI();

    msg = rpmmqttExpand(mqtt, NULL,
		"%now", " ", str, NULL);
    if (rpmmqttConnect(mqtt))
	goto exit;
    if (msg != NULL && !rpmmqttSendMessage(mqtt, NULL, msg, strlen(msg))) {
	rc = RPMRC_OK;
	if (resultp) {
	    *resultp = (mqtt->iob ? rpmiobStr(mqtt->iob) : "");
	}
    }

exit:
    if (msg) free((void *)msg);
SPEW((stderr, "<== %s(%p,\"%s\",%p) rc %d\n", __FUNCTION__, mqtt, str, resultp, rc));
    return rc;
}
