/** \ingroup rpmio
 * \file rpmio/rpmrpc.c
 */

#include "system.h"

#if defined(HAVE_SYS_SYSCALL_H)
# include <sys/syscall.h>
#endif

#include <rpmio_internal.h>
#include <rpmmacro.h>
#include <rpmurl.h>

#define	_RPMDIR_INTERNAL
#include <rpmdir.h>
#define	_RPMDAV_INTERNAL
#include <rpmdav.h>

#include <rpmhash.h>
#include <ugid.h>

#include "debug.h"

extern int _rpmio_debug;
#define SPEW(_fmt, ...) \
    if (_rpmio_debug) \
	fprintf(stderr, _fmt, __VA_ARGS__)

extern int _ftp_debug;
#define FTPSPEW(_fmt, ...) \
    if (_ftp_debug) \
	fprintf(stderr, _fmt, __VA_ARGS__)

/* =============================================================== */
static int ftpMkdir(const char * path, mode_t mode)
{
    int rc;
    if ((rc = ftpCmd("MKD", path, NULL)) != 0)
	return rc;
#if NOTYET
    {	char buf[20];
	sprintf(buf, " 0%o", mode);
	(void) ftpCmd("SITE CHMOD", path, buf);
    }
#endif
    return rc;
}

static int ftpChdir(const char * path)
{
    return ftpCmd("CWD", path, NULL);
}

static int ftpRmdir(const char * path)
{
    return ftpCmd("RMD", path, NULL);
}

static int ftpRename(const char * oldpath, const char * newpath)
{
    int rc;
    if ((rc = ftpCmd("RNFR", oldpath, NULL)) != 0)
	return rc;
    return ftpCmd("RNTO", newpath, NULL);
}

static int ftpUnlink(const char * path)
{
    return ftpCmd("DELE", path, NULL);
}

/* =============================================================== */
int Mkdir (const char * path, mode_t mode)
{
    int rc = -2;
    const char * lpath;
    int ut = urlPath(path, &lpath);

SPEW("*** Mkdir(%s, 0%o)\n", path, (unsigned)mode);
    switch (ut) {
    case URL_IS_FTP:
	return ftpMkdir(path, mode);
	break;
    case URL_IS_HTTPS:
    case URL_IS_HTTP:
#ifdef WITH_NEON
	return davMkdir(path, mode);
#endif
	break;
    case URL_IS_PATH:
	path = lpath;
	break;
    case URL_IS_UNKNOWN:
	break;
    case URL_IS_DASH:
    case URL_IS_HKP:
    case URL_IS_MONGO:	/* XXX FIXME */
    default:
	return rc;
	break;
    }
    return mkdir(path, mode);
}

int Chdir (const char * path)
{
    int rc = -2;
    const char * lpath;
    int ut = urlPath(path, &lpath);

SPEW("*** Chdir(%s)\n", path);
    switch (ut) {
    case URL_IS_FTP:
	return ftpChdir(path);
	break;
    case URL_IS_HTTPS:
    case URL_IS_HTTP:
#ifdef	NOTYET
	return davChdir(path);
#else
	errno = EINVAL;		/* XXX W2DO? */
	return rc;
#endif
	break;
    case URL_IS_PATH:
	path = lpath;
	break;
    case URL_IS_UNKNOWN:
	break;
    case URL_IS_DASH:
    case URL_IS_HKP:
    case URL_IS_MONGO:	/* XXX FIXME */
    default:
	errno = EINVAL;		/* XXX W2DO? */
	return rc;
	break;
    }
    return chdir(path);
}

int Rmdir (const char * path)
{
    int rc = -2;
    const char * lpath;
    int ut = urlPath(path, &lpath);

SPEW("*** Rmdir(%s)\n", path);
    switch (ut) {
    case URL_IS_FTP:
	return ftpRmdir(path);
	break;
    case URL_IS_HTTPS:
    case URL_IS_HTTP:
#ifdef WITH_NEON
	return davRmdir(path);
#endif
	break;
    case URL_IS_PATH:
	path = lpath;
	break;
    case URL_IS_UNKNOWN:
	break;
    case URL_IS_DASH:
    case URL_IS_HKP:
    case URL_IS_MONGO:	/* XXX FIXME */
    default:
	return rc;
	break;
    }
    return rmdir(path);
}

const char * _chroot_prefix = NULL;

int Chroot(const char * path)
{
    int rc = -2;
    const char * lpath;
    int ut = urlPath(path, &lpath);

    switch (ut) {
    case URL_IS_PATH:
	path = lpath;
	break;
    case URL_IS_UNKNOWN:
	break;
    case URL_IS_DASH:
    case URL_IS_HKP:
    case URL_IS_FTP:		/* XXX TODO: implement. */
    case URL_IS_HTTP:		/* XXX TODO: implement. */
    case URL_IS_HTTPS:		/* XXX TODO: implement. */
    case URL_IS_MONGO:	/* XXX FIXME */
    default:
	errno = EINVAL;		/* XXX W2DO? */
	return rc;
	break;
    }

    _chroot_prefix = _free(_chroot_prefix);
    if (strcmp(path, "."))
	_chroot_prefix = rpmGetPath(path, NULL);

    rc = chroot(path);

SPEW("<-- %s(%s) prefix %s rc %d\n", __FUNCTION__, path, _chroot_prefix, rc);

    return rc;
}

int Open(const char * path, int flags, mode_t mode)
{
    int rc = -2;
    const char * lpath;
    int ut = urlPath(path, &lpath);
    int fdno;

    switch (ut) {
    case URL_IS_PATH:
	path = lpath;
	break;
    case URL_IS_UNKNOWN:
	break;
    case URL_IS_DASH:
    case URL_IS_HKP:
    case URL_IS_FTP:		/* XXX TODO: implement. */
    case URL_IS_HTTP:		/* XXX TODO: implement. */
    case URL_IS_HTTPS:		/* XXX TODO: implement. */
    case URL_IS_MONGO:	/* XXX FIXME */
    default:
	errno = EINVAL;		/* XXX W2DO? */
	return rc;
	break;
    }

    if (_chroot_prefix && _chroot_prefix[0] == '/' && _chroot_prefix[1] != '\0')
    {
	size_t nb = strlen(_chroot_prefix);
	size_t ob = strlen(path);
	while (nb > 0 && _chroot_prefix[nb-1] == '/')
	    nb--;
	if (ob > nb && !strncmp(path, _chroot_prefix, nb) && path[nb] == '/')
	    path += nb;
    }
#ifdef	NOTYET	/* XXX likely sane default. */
    if (mode == 0)
	mode = 0644;
#endif
    fdno = open(path, flags, mode);
    /* XXX if the open(2) fails, try to strip a possible chroot(2) prefix. */
    if (fdno < 0 && errno == ENOENT) {
	const char *dbpath = rpmExpand("%{?_dbpath}/", NULL);
	const char * resolved_dbpath = Realpath(dbpath, NULL);
	const char * fn = strstr(path + 1, resolved_dbpath);
	if (fn)
	    fdno = open(fn, flags, mode);
	resolved_dbpath = _free(resolved_dbpath);
	dbpath = _free(dbpath);
	
    }
    if (fdno >= 0) {
	if (fcntl(fdno, F_SETFD, FD_CLOEXEC) < 0) {
	    (void) close(fdno);
	    fdno = -1;
	}
    }

#ifdef	NOTYET	/* XXX db-5.0.26 is slower using POSIX_FADV_RANDOM. */
#if defined(HAVE_POSIX_FADVISE) && defined(POSIX_FADV_RANDOM)
    (void) posix_fadvise(fdno, 0, 0, POSIX_FADV_RANDOM);
#endif
#endif

SPEW("<-- %s(%s, 0x%x, 0%o) prefix %s fdno %d\n", __FUNCTION__, path, flags, (unsigned)mode, _chroot_prefix, fdno);

    return fdno;
}

/* XXX rpmdb.c: analogue to rename(2). */

int Rename (const char * oldpath, const char * newpath)
{
    int rc = -2;
    const char *oe = NULL;
    const char *ne = NULL;
    int oldut, newut;

SPEW("*** Rename(%s, %s)\n", oldpath, newpath);
    /* XXX lib/install.c used to rely on this behavior. */
    if (!strcmp(oldpath, newpath)) return 0;

    oldut = urlPath(oldpath, &oe);
    switch (oldut) {
    case URL_IS_HTTPS:
    case URL_IS_HTTP:
#ifdef WITH_NEON
	return davRename(oldpath, newpath);
#endif
	break;
    case URL_IS_FTP:		/* XXX WRONG WRONG WRONG */
    case URL_IS_PATH:
    case URL_IS_UNKNOWN:
	break;
    case URL_IS_DASH:
    case URL_IS_HKP:
    case URL_IS_MONGO:	/* XXX FIXME */
    default:
	return rc;
	break;
    }

    newut = urlPath(newpath, &ne);
    switch (newut) {
    case URL_IS_FTP:
SPEW("*** rename old %*s new %*s\n", (int)(oe - oldpath), oldpath, (int)(ne - newpath), newpath);
	if (!(oldut == newut && oe && ne && (oe - oldpath) == (ne - newpath) &&
	    !xstrncasecmp(oldpath, newpath, (oe - oldpath))))
	    return rc;
	return ftpRename(oldpath, newpath);
	break;
    case URL_IS_HTTPS:		/* XXX WRONG WRONG WRONG */
    case URL_IS_HTTP:		/* XXX WRONG WRONG WRONG */
    case URL_IS_PATH:
	oldpath = oe;
	newpath = ne;
	break;
    case URL_IS_UNKNOWN:
	break;
    case URL_IS_DASH:
    case URL_IS_HKP:
    case URL_IS_MONGO:	/* XXX FIXME */
    default:
	return rc;
	break;
    }
    return rename(oldpath, newpath);
}

int Link (const char * oldpath, const char * newpath)
{
    int rc = -2;
    const char *oe = NULL;
    const char *ne = NULL;
    int oldut, newut;

SPEW("*** Link(%s, %s)\n", oldpath, newpath);
    oldut = urlPath(oldpath, &oe);
    switch (oldut) {
    case URL_IS_HTTPS:		/* XXX WRONG WRONG WRONG */
    case URL_IS_HTTP:		/* XXX WRONG WRONG WRONG */
    case URL_IS_FTP:		/* XXX WRONG WRONG WRONG */
    case URL_IS_PATH:
    case URL_IS_UNKNOWN:
	break;
    case URL_IS_DASH:
    case URL_IS_HKP:
    case URL_IS_MONGO:	/* XXX FIXME */
    default:
	return rc;
	break;
    }

    newut = urlPath(newpath, &ne);
    switch (newut) {
    case URL_IS_HTTPS:		/* XXX WRONG WRONG WRONG */
    case URL_IS_HTTP:		/* XXX WRONG WRONG WRONG */
    case URL_IS_FTP:		/* XXX WRONG WRONG WRONG */
    case URL_IS_PATH:
SPEW("*** link old %*s new %*s\n", (int)(oe - oldpath), oldpath, (int)(ne - newpath), newpath);
	if (!(oldut == newut && oe && ne && (oe - oldpath) == (ne - newpath) &&
	    !xstrncasecmp(oldpath, newpath, (oe - oldpath))))
	    return rc;
	oldpath = oe;
	newpath = ne;
	break;
    case URL_IS_UNKNOWN:
	break;
    case URL_IS_DASH:
    case URL_IS_HKP:
    case URL_IS_MONGO:	/* XXX FIXME */
    default:
	return rc;
	break;
    }
    return link(oldpath, newpath);
}

/* XXX build/build.c: analogue to unlink(2). */

int Unlink(const char * path)
{
    int rc = -2;
    const char * lpath;
    int ut = urlPath(path, &lpath);

    switch (ut) {
    case URL_IS_FTP:
	rc = ftpUnlink(path);
	break;
    case URL_IS_HTTPS:
    case URL_IS_HTTP:
#ifdef WITH_NEON
	rc = davUnlink(path);
#endif
	break;
    case URL_IS_PATH:
	path = lpath;
	break;
    case URL_IS_UNKNOWN:
	break;
    default:
	errno = EINVAL;
	goto exit;
	break;
    }
    rc = unlink(path);
exit:
SPEW("<-- %s(%s) rc %d\n", __FUNCTION__, path, rc);
    return rc;
}

/* XXX swiped from mc-4.5.39-pre9 vfs/ftpfs.c */

#define	g_strdup	xstrdup
#define	g_free		free

/*
 * FIXME: this is broken. It depends on mc not crossing border on month!
 */
static int current_mday;
static int current_mon;
static int current_year;

/* Following stuff (parse_ls_lga) is used by ftpfs and extfs */
#define MAXCOLS		30

static char *columns [MAXCOLS];	/* Points to the string in column n */
static int   column_ptr [MAXCOLS]; /* Index from 0 to the starting positions of the columns */

static int
vfs_split_text (char *p)
{
    char *original = p;
    int  numcols;


    for (numcols = 0; *p && numcols < MAXCOLS; numcols++){
	while (*p == ' ' || *p == '\r' || *p == '\n'){
	    *p = '\0';
	    p++;
	}
	columns [numcols] = p;
	column_ptr [numcols] = p - original;
	while (*p && *p != ' ' && *p != '\r' && *p != '\n')
	    p++;
    }
    return numcols;
}

static int
is_num (int idx)
{
    if (!columns [idx] || columns [idx][0] < '0' || columns [idx][0] > '9')
	return 0;
    return 1;
}

static int
is_dos_date(const char *str)
{
    if (str != NULL && strlen(str) == 8 &&
		str[2] == str[5] && strchr("\\-/", (int)str[2]) != NULL)
	return 1;
    return 0;
}

static int
is_week (const char * str, struct tm * tim)
{
    static const char * week = "SunMonTueWedThuFriSat";
    const char * pos;

    if (str != NULL && (pos=strstr(week, str)) != NULL) {
        if (tim != NULL)
	    tim->tm_wday = (pos - week)/3;
	return 1;
    }
    return 0;
}

static int
is_month (const char * str, struct tm * tim)
{
    static const char * month = "JanFebMarAprMayJunJulAugSepOctNovDec";
    const char * pos;

    if (str != NULL && (pos = strstr(month, str)) != NULL) {
        if (tim != NULL)
	    tim->tm_mon = (pos - month)/3;
	return 1;
    }
    return 0;
}

static int
is_time (const char * str, struct tm * tim)
{
    const char * p, * p2;

    if (str != NULL && (p = strchr(str, ':')) && (p2 = strrchr(str, ':'))) {
	if (p != p2) {
    	    if (sscanf (str, "%2d:%2d:%2d", &tim->tm_hour, &tim->tm_min, &tim->tm_sec) != 3)
		return 0;
	} else {
	    if (sscanf (str, "%2d:%2d", &tim->tm_hour, &tim->tm_min) != 2)
		return 0;
	}
    } else
        return 0;

    return 1;
}

static int is_year(const char * str, struct tm * tim)
{
    long year;

    if (str == NULL)
	return 0;

    if (strchr(str,':'))
        return 0;

    if (strlen(str) != 4)
        return 0;

    if (sscanf(str, "%ld", &year) != 1)
        return 0;

    if (year < 1900 || year > 3000)
        return 0;

    tim->tm_year = (int) (year - 1900);

    return 1;
}

/*
 * FIXME: this is broken. Consider following entry:
 * -rwx------   1 root     root            1 Aug 31 10:04 2904 1234
 * where "2904 1234" is filename. Well, this code decodes it as year :-(.
 */

static int
vfs_parse_filetype (char c)
{
    switch (c) {
        case 'd': return (int)S_IFDIR;
        case 'b': return (int)S_IFBLK;
        case 'c': return (int)S_IFCHR;
        case 'l': return (int)S_IFLNK;
        case 's':
#ifdef IS_IFSOCK /* And if not, we fall through to IFIFO, which is pretty close */
	          return (int)S_IFSOCK;
#endif
        case 'p': return (int)S_IFIFO;
        case 'm': case 'n':		/* Don't know what these are :-) */
        case '-': case '?': return (int)S_IFREG;
        default: return -1;
    }
}

static int vfs_parse_filemode (const char *p)
{	/* converts rw-rw-rw- into 0666 */
    int res = 0;
    switch (*(p++)) {
	case 'r': res |= 0400; break;
	case '-': break;
	default: return -1;
    }
    switch (*(p++)) {
	case 'w': res |= 0200; break;
	case '-': break;
	default: return -1;
    }
    switch (*(p++)) {
	case 'x': res |= 0100; break;
	case 's': res |= 0100 | S_ISUID; break;
	case 'S': res |= S_ISUID; break;
	case '-': break;
	default: return -1;
    }
    switch (*(p++)) {
	case 'r': res |= 0040; break;
	case '-': break;
	default: return -1;
    }
    switch (*(p++)) {
	case 'w': res |= 0020; break;
	case '-': break;
	default: return -1;
    }
    switch (*(p++)) {
	case 'x': res |= 0010; break;
	case 's': res |= 0010 | S_ISGID; break;
        case 'l': /* Solaris produces these */
	case 'S': res |= S_ISGID; break;
	case '-': break;
	default: return -1;
    }
    switch (*(p++)) {
	case 'r': res |= 0004; break;
	case '-': break;
	default: return -1;
    }
    switch (*(p++)) {
	case 'w': res |= 0002; break;
	case '-': break;
	default: return -1;
    }
    switch (*(p++)) {
	case 'x': res |= 0001; break;
	case 't': res |= 0001 | S_ISVTX; break;
	case 'T': res |= S_ISVTX; break;
	case '-': break;
	default: return -1;
    }
    return res;
}

static int vfs_parse_filedate(int idx, time_t *t)
{	/* This thing parses from idx in columns[] array */

    char *p;
    struct tm tim;
    int d[3];
    int	got_year = 0;

    /* Let's setup default time values */
    tim.tm_year = current_year;
    tim.tm_mon  = current_mon;
    tim.tm_mday = current_mday;
    tim.tm_hour = 0;
    tim.tm_min  = 0;
    tim.tm_sec  = 0;
    tim.tm_isdst = -1; /* Let mktime() try to guess correct dst offset */

    p = columns [idx++];

    /* We eat weekday name in case of extfs */
    if(is_week(p, &tim))
	p = columns [idx++];

    /* Month name */
    if(is_month(p, &tim)){
	/* And we expect, it followed by day number */
	if (is_num (idx))
    	    tim.tm_mday = (int)atol (columns [idx++]);
	else
	    return 0; /* No day */

    } else {
        /* We usually expect:
           Mon DD hh:mm
           Mon DD  YYYY
	   But in case of extfs we allow these date formats:
           Mon DD YYYY hh:mm
	   Mon DD hh:mm YYYY
	   Wek Mon DD hh:mm:ss YYYY
           MM-DD-YY hh:mm
           where Mon is Jan-Dec, DD, MM, YY two digit day, month, year,
           YYYY four digit year, hh, mm, ss two digit hour, minute or second. */

	/* Here just this special case with MM-DD-YY */
        if (is_dos_date(p)){
            p[2] = p[5] = '-';

	    memset(d, 0, sizeof(d));
	    if (sscanf(p, "%2d-%2d-%2d", &d[0], &d[1], &d[2]) == 3){
	    /*  We expect to get:
		1. MM-DD-YY
		2. DD-MM-YY
		3. YY-MM-DD
		4. YY-DD-MM  */
		
		/* Hmm... maybe, next time :)*/
		
		/* At last, MM-DD-YY */
		d[0]--; /* Months are zerobased */
		/* Y2K madness */
		if(d[2] < 70)
		    d[2] += 100;

        	tim.tm_mon  = d[0];
        	tim.tm_mday = d[1];
        	tim.tm_year = d[2];
		got_year = 1;
	    } else
		return 0; /* sscanf failed */
        } else
            return 0; /* unsupported format */
    }

    /* Here we expect to find time and/or year */

    if (is_num (idx)) {
        if(is_time(columns[idx], &tim) || (got_year = is_year(columns[idx], &tim))) {
	idx++;

	/* This is a special case for ctime() or Mon DD YYYY hh:mm */
	if(is_num (idx) &&
	    ((got_year = is_year(columns[idx], &tim)) || is_time(columns[idx], &tim)))
		idx++; /* time & year or reverse */
	} /* only time or date */
    }
    else
        return 0; /* Nor time or date */

    /*
     * If the date is less than 6 months in the past, it is shown without year
     * other dates in the past or future are shown with year but without time
     * This does not check for years before 1900 ... I don't know, how
     * to represent them at all
     */
    if (!got_year &&
	current_mon < 6 && current_mon < tim.tm_mon &&
	tim.tm_mon - current_mon >= 6)

	tim.tm_year--;

    if ((*t = mktime(&tim)) < 0)
        *t = 0;
    return idx;
}

static int
vfs_parse_ls_lga (char * p, struct stat * st,
		const char ** filename, const char ** linkname)
{
    int idx, idx2, num_cols;
    int i;
    char *p_copy;
    long n;

    if (strncmp (p, "total", 5) == 0)
        return 0;

    p_copy = g_strdup(p);
/* XXX FIXME: parse out inode number from "NLST -lai ." */
/* XXX FIXME: parse out sizein blocks from "NLST -lais ." */

    if ((i = vfs_parse_filetype(*(p++))) == -1)
        goto error;

    st->st_mode = i;
    if (*p == ' ')	/* Notwell 4 */
        p++;
    if (*p == '['){
	if (strlen (p) <= 8 || p [8] != ']')
	    goto error;
	/* Should parse here the Notwell permissions :) */
	if (S_ISDIR (st->st_mode))
	    st->st_mode |= (S_IRUSR | S_IRGRP | S_IROTH | S_IWUSR | S_IXUSR | S_IXGRP | S_IXOTH);
	else
	    st->st_mode |= (S_IRUSR | S_IRGRP | S_IROTH | S_IWUSR);
	p += 9;
    } else {
	if ((i = vfs_parse_filemode(p)) == -1)
	    goto error;
        st->st_mode |= i;
	p += 9;

        /* This is for an extra ACL attribute (HP-UX) */
        if (*p == '+')
            p++;
    }

    g_free(p_copy);
    p_copy = g_strdup(p);
    num_cols = vfs_split_text (p);

    n = atol(columns[0]);
    st->st_nlink = n;
    if (n < 0)
        goto error;

    if (!is_num (1))
#ifdef	HACK
	st->st_uid = finduid (columns [1]);
#else
	(void) unameToUid (columns [1], &st->st_uid);
#endif
    else
        st->st_uid = (uid_t) atol (columns [1]);

    /* Mhm, the ls -lg did not produce a group field */
    for (idx = 3; idx <= 5; idx++)
        if (is_month(columns [idx], NULL) || is_week(columns [idx], NULL) || is_dos_date(columns[idx]))
            break;

    if (idx == 6 || (idx == 5 && !S_ISCHR (st->st_mode) && !S_ISBLK (st->st_mode)))
	goto error;

    /* We don't have gid */	
    if (idx == 3 || (idx == 4 && (S_ISCHR(st->st_mode) || S_ISBLK (st->st_mode))))
        idx2 = 2;
    else {
	/* We have gid field */
	if (is_num (2))
	    st->st_gid = (gid_t) atol (columns [2]);
	else
#ifdef	HACK
	    st->st_gid = findgid (columns [2]);
#else
	    (void) gnameToGid (columns [1], &st->st_gid);
#endif
	idx2 = 3;
    }

    /* This is device */
    if (S_ISCHR (st->st_mode) || S_ISBLK (st->st_mode)){
	unsigned maj, min;
	
	if (!is_num (idx2) || sscanf(columns [idx2], " %u,", &maj) != 1)
	    goto error;
	
	if (!is_num (++idx2) || sscanf(columns [idx2], " %u", &min) != 1)
	    goto error;
	
#ifdef HAVE_ST_RDEV
	st->st_rdev = ((maj & 0x000000ffU) << 8) | (min & 0x000000ffU);
#endif
	st->st_size = 0;
	
    } else {
	/* Common file size */
	if (!is_num (idx2))
	    goto error;
	
	st->st_size = (size_t) atol (columns [idx2]);
#ifdef HAVE_ST_RDEV
	st->st_rdev = 0;
#endif
    }

    idx = vfs_parse_filedate(idx, &st->st_mtime);
    if (!idx)
        goto error;
    /* Use resulting time value */
    st->st_atime = st->st_ctime = st->st_mtime;
    st->st_dev = 0;
    st->st_ino = 0;
#ifdef HAVE_ST_BLKSIZE
    st->st_blksize = 512;
#endif
#ifdef HAVE_ST_BLOCKS
    st->st_blocks = (st->st_size + 511) / 512;
#endif

    for (i = idx + 1, idx2 = 0; i < num_cols; i++ )
	if (strcmp (columns [i], "->") == 0){
	    idx2 = i;
	    break;
	}

    if (((S_ISLNK (st->st_mode) ||
        (num_cols == idx + 3 && st->st_nlink > 1))) /* Maybe a hardlink? (in extfs) */
        && idx2)
    {
	size_t tlen;
	char *t;

	if (filename){
	    size_t nb = column_ptr [idx2] - column_ptr [idx] - 1;
	    t = strncpy((char *)xcalloc(1, nb+1), p_copy + column_ptr [idx], nb);
	    *filename = t;
	}
	if (linkname){
	    t = g_strdup (p_copy + column_ptr [idx2+1]);
	    tlen = strlen (t);
	    if (t [tlen-1] == '\r' || t [tlen-1] == '\n')
		t [tlen-1] = '\0';
	    if (t [tlen-2] == '\r' || t [tlen-2] == '\n')
		t [tlen-2] = '\0';
		
	    *linkname = t;
	}
    } else {
	/* Extract the filename from the string copy, not from the columns
	 * this way we have a chance of entering hidden directories like ". ."
	 */
	if (filename){
	    /*
	    *filename = g_strdup (columns [idx++]);
	    */
	    size_t tlen;
	    char *t;

	    t = g_strdup (p_copy + column_ptr [idx]); idx++;
	    tlen = strlen (t);
	    /* g_strchomp(); */
	    if (t [tlen-1] == '\r' || t [tlen-1] == '\n')
	        t [tlen-1] = '\0';
	    if (t [tlen-2] == '\r' || t [tlen-2] == '\n')
		t [tlen-2] = '\0';

	    *filename = t;
	}
	if (linkname)
	    *linkname = NULL;
    }
    g_free (p_copy);
    return 1;

error:
#ifdef	HACK
    {
      static int errorcount = 0;

      if (++errorcount < 5) {
	message_1s (1, "Could not parse:", p_copy);
      } else if (errorcount == 5)
	message_1s (1, "More parsing errors will be ignored.", "(sorry)" );
    }
#endif

    if (p_copy != p)		/* Carefull! */
	g_free (p_copy);
    return 0;
}

typedef enum {
	DO_FTP_STAT	= 1,
	DO_FTP_LSTAT	= 2,
	DO_FTP_READLINK	= 3,
	DO_FTP_ACCESS	= 4,
	DO_FTP_GLOB	= 5
} ftpSysCall_t;

/**
 */
static size_t ftpBufAlloced;

/**
 */
static char * ftpBuf;
	
#define alloca_strdup(_s)       strcpy((char *)alloca(strlen(_s)+1), (_s))

static int ftpNLST(const char * url, ftpSysCall_t ftpSysCall,
		struct stat * st, char * rlbuf, size_t rlbufsiz)
{
    int rc = -2;
    FD_t fd;
    const char * path;
    int bufLength, moretodo;
    const char *n, *ne, *o, *oe;
    char * s;
    char * se;
    const char * urldn;
    char * bn = NULL;
    size_t nbn = 0;
    urlinfo u;

    n = ne = o = oe = NULL;
    (void) urlPath(url, &path);
    if (*path == '\0')
	return rc;

    switch (ftpSysCall) {
    case DO_FTP_GLOB:
	fd = ftpOpen(url, 0, 0, &u);
	if (fd == NULL || u == NULL)
	    return -1;

	u->openError = ftpReq(fd, "LIST", path);
	break;
    default:
	urldn = alloca_strdup(url);
	if ((bn = (char *) strrchr(urldn, '/')) == NULL)
	    return rc;
	else if (bn == path)
	    bn = (char *) ".";
	else
	    *bn++ = '\0';
	nbn = strlen(bn);

	rc = ftpChdir(urldn);		/* XXX don't care about CWD */
	if (rc < 0)
	    return rc;

	fd = ftpOpen(url, 0, 0, &u);
	if (fd == NULL || u == NULL)
	    return -1;

	/* XXX possibly should do "NLST -lais" to get st_ino/st_blocks also */
	u->openError = ftpReq(fd, "NLST", "-la");

	if (bn == NULL || nbn == 0) {
	    rc = -2;
	    goto exit;
	}
	break;
    }

    if (u->openError < 0) {
	fd = fdLink(fd, "error data (ftpStat)");
	rc = -2;
	goto exit;
    }

    if (ftpBufAlloced == 0 || ftpBuf == NULL) {
        ftpBufAlloced = _url_iobuf_size;
        ftpBuf = (char *) xcalloc(ftpBufAlloced, sizeof(ftpBuf[0]));
    }
    *ftpBuf = '\0';

    bufLength = 0;
    moretodo = 1;

    do {

	/* XXX FIXME: realloc ftpBuf if < ~128 chars remain */
	if ((ftpBufAlloced - bufLength) < (1024+80)) {
	    ftpBufAlloced <<= 2;
	    assert(ftpBufAlloced < (8*1024*1024));
	    ftpBuf = (char *) xrealloc(ftpBuf, ftpBufAlloced);
	}
	s = se = ftpBuf + bufLength;
	*se = '\0';

	rc = fdFgets(fd, se, (ftpBufAlloced - bufLength));
	if (rc <= 0) {
	    moretodo = 0;
	    break;
	}
	if (ftpSysCall == DO_FTP_GLOB) {	/* XXX HACK */
	    bufLength += strlen(se);
	    continue;
	}

	for (s = se; *s != '\0'; s = se) {
	    int bingo;

	    while (*se && *se != '\n') se++;
	    if (se > s && se[-1] == '\r') se[-1] = '\0';
	    if (*se == '\0')
		break;
	    *se++ = '\0';

	    if (!strncmp(s, "total ", sizeof("total ")-1))
		continue;

	    o = NULL;
	    for (bingo = 0, n = se; n >= s; n--) {
		switch (*n) {
		case '\0':
		    oe = ne = n;
		    break;
		case ' ':
		    if (o || !(n[-3] == ' ' && n[-2] == '-' && n[-1] == '>')) {
			while (*(++n) == ' ')
			    {};
			bingo++;
			break;
		    }
		    for (o = n + 1; *o == ' '; o++)
			{};
		    n -= 3;
		    ne = n;
		    break;
		default:
		    break;
		}
		if (bingo)
		    break;
	    }

	    if (nbn != (size_t)(ne - n))	/* Same name length? */
		continue;
	    if (strncmp(n, bn, nbn))	/* Same name? */
		continue;

	    moretodo = 0;
	    break;
	}

        if (moretodo && se > s) {
            bufLength = se - s - 1;
            if (s != ftpBuf)
                memmove(ftpBuf, s, bufLength);
        } else {
	    bufLength = 0;
	}
    } while (moretodo);

    switch (ftpSysCall) {
    case DO_FTP_STAT:
	if (o && oe) {
	    /* XXX FIXME: symlink, replace urldn/bn from [o,oe) and restart */
	}
	/*@fallthrough@*/
    case DO_FTP_LSTAT:
	if (st == NULL || !(n && ne)) {
	    rc = -1;
	} else {
	    rc = ((vfs_parse_ls_lga(s, st, NULL, NULL) > 0) ? 0 : -1);
	}
	break;
    case DO_FTP_READLINK:
	if (rlbuf == NULL || !(o && oe)) {
	    rc = -1;
	} else {
	    rc = oe - o;
assert(rc >= 0);
	    if (rc > (int)rlbufsiz)
		rc = (int)rlbufsiz;
	    memcpy(rlbuf, o, (size_t)rc);
	    if (rc < (int)rlbufsiz)
		rlbuf[rc] = '\0';
	}
	break;
    case DO_FTP_ACCESS:
	rc = 0;		/* XXX WRONG WRONG WRONG */
	break;
    case DO_FTP_GLOB:
	rc = 0;		/* XXX WRONG WRONG WRONG */
	break;
    }

exit:
    (void) ufdClose(fd);
    return rc;
}

static const char * statstr(const struct stat * st, char * buf)
{
    char * t = buf;
    sprintf(t, "*** dev %x", (unsigned int)st->st_dev);
	t += strlen(t);
    sprintf(t, " ino %x", (unsigned int)st->st_ino);
	t += strlen(t);
    sprintf(t, " mode %0o", (unsigned int)st->st_mode);
	t += strlen(t);
    sprintf(t, " nlink %d", (unsigned int)st->st_nlink);
	t += strlen(t);
    sprintf(t, " uid %d", (unsigned int)st->st_uid);
	t += strlen(t);
    sprintf(t, " gid %d", (unsigned int)st->st_gid);
	t += strlen(t);
    sprintf(t, " rdev %x", (unsigned int)st->st_rdev);
	t += strlen(t);
    sprintf(t, " size %x", (unsigned int)st->st_size);
	t += strlen(t);
    sprintf(t, "\n");
    return buf;
}

/* FIXME: borked for path with trailing '/' */
static int ftpStat(const char * path, struct stat *st)
{
    char buf[1024];
    int rc;
    rc = ftpNLST(path, DO_FTP_STAT, st, NULL, 0);

    /* XXX fts(3) needs/uses st_ino. */
    /* Hash the path to generate a st_ino analogue. */
    if (st->st_ino == 0)
	st->st_ino = hashFunctionString(0, path, 0);

FTPSPEW("<-- %s(%s) rc %d\n%s", __FUNCTION__, path, rc, statstr(st, buf));
    return rc;
}

/* FIXME: borked for path with trailing '/' */
static int ftpLstat(const char * path, struct stat *st)
{
    char buf[1024];
    int rc;
    rc = ftpNLST(path, DO_FTP_LSTAT, st, NULL, 0);

    /* XXX fts(3) needs/uses st_ino. */
    /* Hash the path to generate a st_ino analogue. */
    if (st->st_ino == 0)
	st->st_ino = hashFunctionString(0, path, 0);

FTPSPEW("<-- %s(%s) rc %d\n%s\n", __FUNCTION__, path, rc, statstr(st, buf));
    return rc;
}

static int ftpReadlink(const char * path, char * buf, size_t bufsiz)
{
    int rc;
    rc = ftpNLST(path, DO_FTP_READLINK, NULL, buf, bufsiz);
FTPSPEW("<-- %s(%s) rc %d\n", __FUNCTION__, path, rc);
    return rc;
}

DIR * ftpOpendir(const char * path)
{
    AVDIR avdir;
    rpmavx avx;
    struct stat * st = NULL;
    const char * s, * sb, * se;
    int nac;
    int c;
    int rc;

FTPSPEW("*** %s(%s)\n", __FUNCTION__, path);

    /* Load FTP collection into argv. */
    avx = (rpmavx) rpmavxNew(path, st);
    if (avx == NULL) {
        errno = ENOENT;         /* Note: avx is NULL iff urlSplit() fails. */
        return NULL;
    }

    rc = ftpNLST(path, DO_FTP_GLOB, NULL, NULL, 0);
    if (rc)
	return NULL;

    nac = 0;
    sb = NULL;
    s = se = ftpBuf;
    while ((c = (int) *se++) != (int) '\0') {
	switch (c) {
	case '/':
	    sb = se;
	    break;
	case '\r':
	    if (sb == NULL) {
		for (sb = se; sb > s && sb[-1] != ' '; sb--)
		    {};
	    }
	    nac++;

	    if (*se == '\n') se++;
	    sb = NULL;
	    s = se;
	    break;
	default:
	    break;
	}
    }

    avx->av = (const char **) xcalloc(nac+1, sizeof(*avx->av));
    avx->modes = (rpmuint16_t *) xcalloc(nac, sizeof(*avx->modes));

    nac = 0;
    sb = NULL;
    s = se = ftpBuf;
    while ((c = (int) *se) != (int) '\0') {
	se++;
	switch (c) {
	case '/':
	    sb = se;
	    break;
	case '\r':
	    if (sb == NULL) {
		avx->modes[nac] = (*s == 'd' ? 0755 : 0644);
		switch(*s) {
		case 'p': avx->modes[nac] |= S_IFIFO;	break;
		case 'c': avx->modes[nac] |= S_IFCHR;	break;
		case 'd': avx->modes[nac] |= S_IFDIR;	break;
		case 'b': avx->modes[nac] |= S_IFBLK;	break;
		case '-': avx->modes[nac] |= S_IFREG;	break;
		case 'l': avx->modes[nac] |= S_IFLNK;	break;
		case 's': avx->modes[nac] |= S_IFSOCK;	break;
		default:  avx->modes[nac] |= S_IFREG;	break;
		}
		for (sb = se; sb > s && sb[-1] != ' '; sb--)
		    {};
	    }
	    avx->av[nac++] = strncpy((char *)xcalloc(1, (se-sb-1)+1), sb, (se-sb-1));
	    if (*se == '\n') se++;
	    sb = NULL;
	    s = se;
	    break;
	default:
	    break;
	}
    }

    avdir = (AVDIR) avOpendir(path, avx->av, avx->modes);

    avx = rpmavxFree(avx);

    return (DIR *) avdir;
}

static char * ftpRealpath(const char * path, char * resolved_path)
{
assert(resolved_path == NULL);	/* XXX no POSIXly broken realpath(3) here. */
    /* XXX TODO: handle redirects. For now, just dupe the path. */
    return xstrdup(path);
}

int Stat(const char * path, struct stat * st)
{
    int rc = -2;
    const char * lpath;
    int ut = urlPath(path, &lpath);

    switch (ut) {
    case URL_IS_FTP:
	rc = ftpStat(path, st);
	goto exit;
	break;
    case URL_IS_HTTPS:
    case URL_IS_HTTP:
#ifdef WITH_NEON
	rc = davStat(path, st);
#endif
	goto exit;
	break;
    case URL_IS_PATH:
	path = lpath;
	break;
    case URL_IS_UNKNOWN:
	break;
    case URL_IS_DASH:
    case URL_IS_HKP:
    case URL_IS_MONGO:	/* XXX FIXME */
    default:
	errno = ENOENT;	
	goto exit;
	break;
    }
    rc = stat(path, st);
exit:
SPEW("<-- %s(%s,%p) rc %d\n", __FUNCTION__, path, st, rc);
    return rc;
}

int Lstat(const char * path, struct stat * st)
{
    int rc = -2;
    const char * lpath;
    int ut = urlPath(path, &lpath);

    switch (ut) {
    case URL_IS_FTP:
	rc = ftpLstat(path, st);
	goto exit;
	break;
    case URL_IS_HTTPS:
    case URL_IS_HTTP:
#ifdef WITH_NEON
	rc = davLstat(path, st);
#endif
	goto exit;
	break;
    case URL_IS_PATH:
	path = lpath;
	break;
    case URL_IS_UNKNOWN:
	break;
    case URL_IS_DASH:
    case URL_IS_HKP:
    case URL_IS_MONGO:	/* XXX FIXME */
    default:
	errno = ENOENT;	
	goto exit;
	break;
    }
    rc = lstat(path, st);
exit:
SPEW("<-- %s(%s,%p) rc %d\n", __FUNCTION__, path, st, rc);
    return rc;
}

int Fstat(FD_t fd, struct stat * st)
{
    int rc = -2;
    const char * path = fdGetOPath(fd);
    const char * lpath;
    int ut = urlPath(path, &lpath);

    FDSANE(fd);
    if (fd == NULL || path == NULL || *path == '\0' || st == NULL) {
	errno = ENOENT;
	goto exit;
    }

    switch (ut) {
    case URL_IS_DASH:
    case URL_IS_PATH:
    case URL_IS_UNKNOWN:
	break;
    case URL_IS_FTP:
    case URL_IS_HTTPS:
    case URL_IS_HTTP:
    case URL_IS_HKP:
	if (fd->contentLength < 0) {
	    errno = ENOENT;
	    goto exit;
	}
	memset(st, 0, sizeof(*st));
	if (path[strlen(path)-1] == '/') {
	    st->st_nlink = 2;
	    st->st_mode = (S_IFDIR | 0755);
	} else {
	    st->st_nlink = 1;
	    st->st_mode = (S_IFREG | 0644);
	}
	st->st_ino = hashFunctionString(0, path, 0);
	st->st_size = fd->contentLength;
	st->st_mtime = fd->lastModified;

	st->st_atime = st->st_ctime = st->st_mtime;
	st->st_blksize = 4 * 1024;  /* HACK correct for linux ext */
	st->st_blocks = (st->st_size + 511)/512;
	rc = 0;
	goto exit;
	break;
    case URL_IS_MONGO:	/* XXX FIXME */
    default:
	errno = ENOENT;	
	goto exit;
	break;
    }
    rc = fstat(Fileno(fd), st);
exit:
SPEW("<-- %s(%p,%p) path %s rc %d\n", __FUNCTION__, fd, st, path, rc);
    return rc;
}

void *Mmap(void *addr, size_t len, int prot, int flags,
                  FD_t fd, off_t offset)
{
    int fdno = (fd ? Fileno(fd) : -1);
    void * ret = mmap(addr, len, prot, flags, fdno, offset);
    if (ret == NULL || ret == MAP_FAILED)
	rpmlog(RPMLOG_ERR, _("%s(%p[%u],0x%x,0x%x,%p,0x%x) failed: %m\n"),
		__FUNCTION__, addr, (unsigned)len, prot, flags, fd,
		(unsigned)offset);
SPEW("<-- %s(%p[%u],0x%x,0x%x,%p,0x%x) ret %p\n", __FUNCTION__, addr, (unsigned)len, prot, flags, fd, (unsigned)offset, ret);
    return ret;
}

int Munmap(void * addr, size_t len)
{
    int rc = munmap(addr, len);
    if (rc < 0)
	rpmlog(RPMLOG_ERR, _("%s(%p[%u]) failed: %m\n"),
		__FUNCTION__, addr, (unsigned)len);
SPEW("<-- %s(%p[%u]) rc %d\n", __FUNCTION__, addr, (unsigned)len, rc);
    return rc;
}

int Msync(void * addr, size_t len, int flags)
{
    int rc = msync(addr, len, flags);
    if (rc < 0)
	rpmlog(RPMLOG_ERR, _("%s(%p[%u],%d) failed: %m\n"),
		__FUNCTION__, addr, (unsigned)len, flags);
SPEW("<-- %s(%p[%u],%d) rc %d\n", __FUNCTION__, addr, (unsigned)len, flags, rc);
    return rc;
}

void * Mremap(void * oaddr, size_t olen, size_t nlen, int flags, ...)
{
    va_list ap;
    va_start(ap, flags);
    void * naddr = va_arg(ap, void *);
    va_end(ap);
    void * ret = mremap(oaddr, olen, nlen, flags, naddr);
    if (ret == NULL || ret == MAP_FAILED)
	rpmlog(RPMLOG_ERR, _("%s(%p[%u => %u],%d,%p) failed: %m\n"),
		__FUNCTION__, oaddr, (unsigned)olen, (unsigned)nlen, flags, naddr);
SPEW("<-- %s(%p[%u => %u],%d,%p) ret %p\n", __FUNCTION__, oaddr, (unsigned)olen, (unsigned)nlen, flags, naddr, ret);
    return ret;
}

int Mprotect(void * addr, size_t len, int prot)
{
    int rc = mprotect(addr, len, prot);
    if (rc < 0)
	rpmlog(RPMLOG_ERR, _("%s(%p[%u],%d) failed: %m\n"),
		__FUNCTION__, addr, (unsigned)len, prot);
SPEW("<-- %s(%p[%u],%d) rc %d\n", __FUNCTION__, addr, (unsigned)len, prot, rc);
    return rc;
}

int Madvise(void *addr, size_t len, int advice)
{
    int rc = madvise(addr, len, advice);
    if (rc < 0)
	rpmlog(RPMLOG_ERR, _("%s(%p[%u],%d) failed: %m\n"),
		__FUNCTION__, addr, (unsigned)len, advice);
SPEW("<-- %s(%p[%u],%d) rc %d\n", __FUNCTION__, addr, (unsigned)len, advice, rc);
    return rc;
}

int Mincore(void * addr, size_t len, unsigned char * vec)
{
    int rc = mincore(addr, len, vec);
    if (rc < 0)
	rpmlog(RPMLOG_ERR, _("%s(%p[%u],%p) failed: %m\n"),
		__FUNCTION__, addr, (unsigned)len, vec);
SPEW("<-- %s(%p[%u],%p) rc %d\n", __FUNCTION__, addr, (unsigned)len, vec, rc);
    return rc;
}

/* XXX added in kernel >= 3.17, wrapped in glibc >= 2.25 */
#ifdef	REF	/* <sys/random.h> */
#define GRND_NONBLOCK 0x01
#define GRND_RANDOM 0x02
#endif
int Getrandom(void *buf, size_t buflen, unsigned int flags);
int Getrandom(void *buf, size_t buflen, unsigned int flags)
{
    int rc = -2;
#if defined(HAVE_GETRANDOM)
    rc = getrandom(buf, buflen, flags);
#elif defined(__linux__) && defined(SYS_getrandom)
    rc = syscall(SYS_getrandom, buf, buflen, flags);
#else
    errno = ENOSYS;
#endif
    if (rc < 0)
	rpmlog(RPMLOG_ERR, _("%s(%p[%u],0x%x) failed: %m\n"),
		__FUNCTION__, buf, (unsigned)buflen, flags);
SPEW("<-- %s(%p[%u],0x%x) rc %d\n", __FUNCTION__, buf, (unsigned)buflen, flags, rc);
    return rc;
}

#ifdef	NOTYET

#if defined(HAVE_SYS_SENDFILE_H)
#include <sys/sendfile.h>
#endif

/* XXX kernel < 2.6.33, one of {in,out}_fd must be a socket. */
ssize_t Sendfile(int out_fd, int in_fd, off_t *offset, size_t count);
ssize_t Sendfile(int out_fd, int in_fd, off_t *offset, size_t count)
{
    int rc;
#if defined(HAVE_SENDFILE)
    rc = sendfile(out_fd, in_fd, offset, count);
#elif defined(__linux__) && defined(SYS_sendfile)
    rc = syscall(SYS_sendfile, out_fd, in_fd, offset, count);
#endif
    if (rc < 0)
	rpmlog(RPMLOG_ERR, _("%s(%d, %d, %p, %u) failed: %m\n"),
		__FUNCTION__, out_fd, in_fd, offset, (unsigned)count);
SPEW("<-- %s(%d, %d, %p, %u) rc %d\n", __FUNCTION__, out_fd, in_fd, offset, (unsigned)count, rc);
    return rc;
}

/* XXX if off_{in,out} are NULL, fd_{in_out} offset is changed. */
/* XXX if off_{in,out} are not NULL, fd_{in_out} offset are not changed. */
loff_t Copy_file_range(int fd_in, loff_t *off_in, int fd_out, loff_t *off_out,
		size_t len, unsigned int flags);
loff_t Copy_file_range(int fd_in, loff_t *off_in, int fd_out, loff_t *off_out,
		size_t len, unsigned int flags)
{
    loff_t  rc = -2;
#if defined(HAVE_COPY_FILE_RANGE)
    rc = copy_file_range(fd_in, off_in, fd_out, off_out, len, flags);
#elif defined(__linux__) && defined(SYS_copy_file_range)
    rc = syscall(SYS_copy_file_range, fd_in, off_in, fd_out, off_out, len, flags);
#else
    errno = ENOSYS;
#endif
    if (rc < 0)
	rpmlog(RPMLOG_ERR, _("%s(%d,%p, %d,%p, %u, %u) failed: %m\n"),
		__FUNCTION__, fd_in, off_in, fd_out, off_out, (unsigned)len, flags);
SPEW("<-- %s(%d,%p, %d,%p, %u, %u) rc %ld\n",
		__FUNCTION__, fd_in, off_in, fd_out, off_out, (unsigned)len, flags, rc);
    return rc;
}
#endif	/* NOTYET */

/* XXX same note as copy_file_range. one of fd_{in,out} must be a pipe. */
ssize_t Splice(int fd_in, loff_t *off_in, int fd_out, loff_t *off_out,
		size_t len, unsigned int flags);
ssize_t Splice(int fd_in, loff_t *off_in, int fd_out, loff_t *off_out,
		size_t len, unsigned int flags)
{
    int rc = -2;
#if defined(HAVE_SPLICE)
    rc = splice(fd_in, off_in, fd_out, off_out, len, flags);
#elif defined(__linux__) && defined(SYS_splice)
    rc = syscall(SYS_splice, fd_in, off_in, fd_out, off_out, len, flags);
#else
    errno = ENOSYS;
#endif
    if (rc < 0)
	rpmlog(RPMLOG_ERR, _("%s(%d,%p, %d,%p, %u, %u) failed: %m\n"),
		__FUNCTION__, fd_in, off_in, fd_out, off_out, (unsigned)len, flags);
SPEW("<-- %s(%d,%p, %d,%p, %u, %u) rc %d\n", __FUNCTION__, fd_in, off_in, fd_out, off_out, (unsigned)len, flags, rc);
    return rc;
}

/* XXX both fd_{in,out} must be pipes. */
ssize_t Tee(int fd_in, int fd_out, size_t len, unsigned int flags);
ssize_t Tee(int fd_in, int fd_out, size_t len, unsigned int flags)
{
    int rc = -2;
#if defined(HAVE_TEE)
    rc = tee(fd_in, fd_out, len, flags);
#elif defined(__linux__) && defined(SYS_tee)
    rc = syscall(SYS_tee, fd_in, fd_out, len, flags);
#else
    errno = ENOSYS;
#endif
    if (rc < 0)
	rpmlog(RPMLOG_ERR, _("%s(%d, %d, %u, %u) failed: %m\n"),
		__FUNCTION__, fd_in, fd_out, (unsigned)len, flags);
SPEW("<-- %s(%d, %d, %u, %u) rc %d\n", __FUNCTION__, fd_in, fd_out, (unsigned)len, flags, rc);
    return rc;
}

/* XXX fd must be a pipe. */
ssize_t Vmsplice(int fd, const struct iovec *iov,
		unsigned long nr_segs, unsigned int flags);
ssize_t Vmsplice(int fd, const struct iovec *iov,
		unsigned long nr_segs, unsigned int flags)
{
    int rc = -2;
#if defined(HAVE_VMSPLICE)
    rc = vmsplice(fd, iov, nr_segs, flags);
#elif defined(__linux__) && defined(SYS_vmsplice)
    rc = syscall(SYS_vmsplice, fd, iov, nr_segs, flags);
#else
    errno = ENOSYS;
#endif
    if (rc < 0)
	rpmlog(RPMLOG_ERR, _("%s(%d, %p, %u, %u) failed: %m\n"),
		__FUNCTION__, fd, iov, (unsigned)nr_segs, flags);
SPEW("<-- %s(%d, %p, %u, %u) rc %d\n", __FUNCTION__, fd, iov, (unsigned)nr_segs, flags, rc);
    return rc;
}

#if defined(HAVE_SYS_AUXV_H)
#include <sys/auxv.h>
#endif

typedef struct key_s {
    uint32_t v;
    const char *n;
} KEY;

static const char * tblName(uint32_t v, KEY * tbl, size_t ntbl)
{
    static char buf[32];
    const char * n = NULL;
    size_t i;

    for (i = 0; i < ntbl; i++) {
	if (v != tbl[i].v)
	    continue;
	n = tbl[i].n;
	break;
    }
    if (n == NULL) {
	(void) snprintf(buf, sizeof(buf), "0x%x", (unsigned)v);
	n = buf;
    }
    return n;
}

#if defined(HAVE_ELF_H)
#include <elf.h>

#define _ENTRY(_v)      { AT_##_v, #_v, }
static KEY ATtypes[] = {
#ifdef	NOTYET
    _ENTRY(NULL),
    _ENTRY(IGNORE),
    _ENTRY(EXECFD),
#endif
    _ENTRY(PHDR),		/* 0x400040 */
    _ENTRY(PHENT),		/* 0x38 */
    _ENTRY(PHNUM),		/* 0x9 */
    _ENTRY(PAGESZ),		/* 0x1000 */
    _ENTRY(BASE),		/* 0x7fae8c4f3000 */
#ifdef	NOTYET
    _ENTRY(FLAGS),
#endif
    _ENTRY(ENTRY),		/* 0x4046e0 */
#ifdef	NOTYET
    _ENTRY(NOTELF),
#endif
    _ENTRY(UID),
    _ENTRY(EUID),
    _ENTRY(GID),
    _ENTRY(EGID),
    _ENTRY(CLKTCK),		/* 0x64 */
    _ENTRY(PLATFORM),		/* 0x7ffdfa0aa809 */
    _ENTRY(HWCAP),		/* 0xbfebfbff */
#ifdef	NOTYET
    _ENTRY(FPUCW),
    _ENTRY(DCACHEBSIZE),
    _ENTRY(ICACHEBSIZE),
    _ENTRY(UCACHEBSIZE),
    _ENTRY(IGNOREPPC),
    _ENTRY(SECURE),
    _ENTRY(BASE_PLATFORM),
#endif
    _ENTRY(RANDOM),		/* 0x7ffdfa0aa7f9 */
#ifdef	NOTYET
    _ENTRY(HWCAP2),
    _ENTRY(EXECFN),
    _ENTRY(SYSINFO),
#endif
    _ENTRY(SYSINFO_EHDR),	/* 0x7ffdfa1a7000 */
#ifdef	NOTYET
    _ENTRY(L1I_CACHESHAPE),
    _ENTRY(L1D_CACHESHAPE),
    _ENTRY(L2_CACHESHAPE),
    _ENTRY(L3_CACHESHAPE),
#endif
};
static size_t nATtypes = sizeof(ATtypes) / sizeof(ATtypes[0]);
#define fmtATtype(_type) tblName(_type, ATtypes, nATtypes)
#undef	_ENTRY
#else	/* HAVE_ELF_H */
#define fmtATtype(_type) "???"
#endif	/* HAVE_ELF_H */

unsigned long Getauxval(unsigned long type)
{
    unsigned long rc = 0;
#if defined(HAVE_GETAUXVAL)
    rc = getauxval(type);
#else
    errno = ENOSYS;
#endif
    if (rc == 0)
	rpmlog(RPMLOG_ERR, _("%s(AT_%s) failed: rc 0x%lx %m\n"),
		__FUNCTION__, fmtATtype(type), rc);
SPEW("<-- %s(AT_%s) rc 0x%lx\n", __FUNCTION__, fmtATtype(type), rc);
    return rc;
}

#if defined(HAVE_SYS_PERSONALITY_H)
#include <sys/personality.h>

#define _ENTRY(_v)      { PER_##_v, #_v, }
static KEY PERsona[] = {
    _ENTRY(LINUX),
    _ENTRY(LINUX_32BIT),
#if defined(PER_LINUX_FDPIC)
    _ENTRY(LINUX_FDPIC),
#endif
    _ENTRY(SVR4),
    _ENTRY(SVR3),
    _ENTRY(SCOSVR3),
    _ENTRY(OSR5),
    _ENTRY(WYSEV386),
    _ENTRY(ISCR4),
    _ENTRY(BSD),
    _ENTRY(SUNOS),
    _ENTRY(XENIX),
    _ENTRY(LINUX32),
    _ENTRY(LINUX32_3GB),
    _ENTRY(IRIX32),
    _ENTRY(IRIXN32),
    _ENTRY(IRIX64),
    _ENTRY(RISCOS),
    _ENTRY(SOLARIS),
    _ENTRY(UW7),
    _ENTRY(OSF4),
    _ENTRY(HPUX),
    { 0xffffffff, "CURRENT" }
};
static size_t nPERsona = sizeof(PERsona) / sizeof(PERsona[0]);
#define fmtPERsona(_persona) tblName(_persona, PERsona, nPERsona)
#undef	_ENTRY
#else	/* HAVE_SYS_PERSONALITY_H */
#define fmtPERsona(_persona) "???"
#endif	/* HAVE_SYS_PERSONALITY_H */

int Personality(unsigned long persona)
{
    int rc = -1;
#if defined(HAVE_PERSONALITY)
    rc = personality(persona);
#else
    errno = ENOSYS;
#endif
    if (rc == -1)
	rpmlog(RPMLOG_ERR, _("%s(PER_%s) failed: %m\n"),
		__FUNCTION__, fmtPERsona(persona));
SPEW("<-- %s(PER_%s) was PER_%s\n", __FUNCTION__, fmtPERsona(persona), (rc != -1 ? fmtPERsona(rc) : "ERROR"));
    return rc;
}

int Chown(const char * path, uid_t owner, gid_t group)
{
    int rc = -2;
    const char * lpath = fdSyscallPrePath(__FUNCTION__, path, 1);
    rc = lpath != NULL
	? chown(lpath, owner, group)
	: -2;
    return fdSyscallPostPath(__FUNCTION__, lpath, rc);
}

int Lchown(const char * path, uid_t owner, gid_t group)
{
    int rc = -2;
    const char * lpath = fdSyscallPrePath(__FUNCTION__, path, 1);
    rc = lpath != NULL
	? lchown(lpath, owner, group)
	: -2;
    return fdSyscallPostPath(__FUNCTION__, lpath, rc);
}

int Chmod(const char * path, mode_t mode)
{
    int rc = -2;
    const char * lpath = fdSyscallPrePath(__FUNCTION__, path, 1);
    rc = lpath != NULL
	? chmod(lpath, mode)
	: -2;
    return fdSyscallPostPath(__FUNCTION__, lpath, rc);
}

int Lchmod(const char * path, mode_t mode)
{
    int rc = -2;
    const char * lpath = fdSyscallPrePath(__FUNCTION__, path, 1);
#if defined(HAVE_LCHMOD)
    rc = lpath != NULL
	? lchmod(lpath, mode)
	: -2;
#else
    errno = ENOSYS;
#endif
    return fdSyscallPostPath(__FUNCTION__, lpath, rc);
}

int Chflags(const char * path, unsigned int flags)
{
    int rc = -2;
    const char * lpath = fdSyscallPrePath(__FUNCTION__, path, 1);
#if defined(HAVE_CHFLAGS)
    rc = lpath != NULL
	? chflags(lpath, flags)
	: -2;
#else
    errno = ENOSYS;
#endif
    return fdSyscallPostPath(__FUNCTION__, lpath, rc);
}

int Lchflags(const char * path, unsigned int flags)
{
    int rc = -2;
    const char * lpath = fdSyscallPrePath(__FUNCTION__, path, 1);
#if defined(HAVE_LCHFLAGS)
    rc = lpath != NULL
	? lchflags(lpath, flags)
	: -2;
#else
    errno = ENOSYS;
#endif
    return fdSyscallPostPath(__FUNCTION__, lpath, rc);
}

int Mkfifo(const char * path, mode_t mode)
{
    int rc = -2;
    const char * lpath = fdSyscallPrePath(__FUNCTION__, path, 1);
    rc = lpath != NULL
	? mkfifo(lpath, mode)
	: -2;
    return fdSyscallPostPath(__FUNCTION__, lpath, rc);
}

int Mknod(const char * path, mode_t mode, dev_t dev)
{
    int rc = -2;
    const char * lpath = fdSyscallPrePath(__FUNCTION__, path, 1);
    rc = lpath != NULL
	? mknod(lpath, mode, dev)
	: -2;
    return fdSyscallPostPath(__FUNCTION__, lpath, rc);
}

int Utime(const char * path, const struct utimbuf *buf)
{
    int rc = -2;
    const char * lpath = fdSyscallPrePath(__FUNCTION__, path, 1);
    rc = lpath != NULL
	? utime(lpath, buf)
	: -2;
    return fdSyscallPostPath(__FUNCTION__, lpath, rc);
}

int Utimes(const char * path, const struct timeval times[2])
{
    int rc = -2;
    const char * lpath = fdSyscallPrePath(__FUNCTION__, path, 1);
    rc = lpath != NULL
	? utimes(lpath, times)
	: -2;
    return fdSyscallPostPath(__FUNCTION__, lpath, rc);
}

int Lutimes(const char * path, const struct timeval times[2])
{
    int rc = -2;
    const char * lpath = fdSyscallPrePath(__FUNCTION__, path, 1);
#ifdef HAVE_LUTIMES
    rc = lpath != NULL
	? lutimes(lpath, times)
	: -2;
#else
    errno = ENOSYS;
#endif
    return fdSyscallPostPath(__FUNCTION__, lpath, rc);
}

int Symlink(const char * oldpath, const char * newpath)
{
    int rc = -2;
    const char * opath;
    int out = urlPath(oldpath, &opath);
    const char * npath;
    int nut = urlPath(newpath, &npath);

    (void)nut;	/* XXX keep gcc/clang quiet. */
    switch (out) {
    case URL_IS_PATH:
	oldpath = opath;
	newpath = npath;
	break;
    case URL_IS_UNKNOWN:
	break;
    default:
	errno = EINVAL;
	goto exit;
	break;
    }
    rc = symlink(oldpath, newpath);
exit:
SPEW("<-- %s(%s,%s) rc %d\n", __FUNCTION__, oldpath, newpath, rc);
    return rc;
}

int Readlink(const char * path, char * buf, size_t bufsiz)
{
    int rc = -2;
    const char * lpath;
    int ut = urlPath(path, &lpath);

    switch (ut) {
    case URL_IS_FTP:
	rc = ftpReadlink(path, buf, bufsiz);
	goto exit;
	break;
    case URL_IS_HTTPS:
    case URL_IS_HTTP:
#ifdef	NOTYET
	rc = davReadlink(path, buf, bufsiz);
#endif
	goto exit;
	break;
    case URL_IS_PATH:
	path = lpath;
	break;
    case URL_IS_UNKNOWN:
	break;
    case URL_IS_DASH:
    case URL_IS_HKP:
    case URL_IS_MONGO:	/* XXX FIXME */
    default:
	errno = EINVAL;		/* XXX W2DO? */
	goto exit;
	break;
    }
    rc = readlink(path, buf, bufsiz);
exit:
SPEW("<-- %s(%s,%p[%u]) rc %d\n", __FUNCTION__, path, buf, (unsigned)bufsiz, rc);
    return rc;
}

int Access(const char * path, int amode)
{
    int rc = -2;
    const char * lpath = fdSyscallPrePath(__FUNCTION__, path, 1);
    rc = lpath != NULL
	? access(lpath, amode)
	: -2;
    return fdSyscallPostPath(__FUNCTION__, lpath, rc);
}

/* glob_pattern_p() taken from bash
 * Copyright (C) 1985, 1988, 1989 Free Software Foundation, Inc.
 *
 * Return nonzero if PATTERN has any special globbing chars in it.
 */
int Glob_pattern_p (const char * pattern, int quote)
{
    const char *p;
    int ut = urlPath(pattern, &p);
    int open = 0;
    char c;

    while ((c = *p++) != '\0')
	switch (c) {
	case '?':
	    /* Don't treat '?' as a glob char in HTTP URL's */
	    if (ut == URL_IS_HTTPS || ut == URL_IS_HTTP || ut == URL_IS_HKP)
		continue;
	    return (1);
	case '*':
	    return (1);
	case '\\':
	    if (quote && *p != '\0')
		p++;
	    continue;

	case '[':
	    open = 1;
	    continue;
	case ']':
	    if (open)
		return (1);
	    continue;

	case '+':
	case '@':
	case '!':
	    if (*p == '(')
		return (1);
	    continue;
	}

    return (0);
}

int Glob_error(const char * epath, int eerrno)
{
    return 1;
}

int Glob(const char *pattern, int flags,
	int errfunc(const char * epath, int eerrno), void *_pglob)
{
    int rc = GLOB_ABORTED;
    glob_t *pglob = (glob_t *) _pglob;
    const char * lpath;
    int ut = urlPath(pattern, &lpath);
    const char *home = getenv("HOME");

SPEW("*** Glob(%s,0x%x,%p,%p)\n", pattern, (unsigned)flags, (void *)errfunc, pglob);
    switch (ut) {
    case URL_IS_HTTPS:
    case URL_IS_HTTP:
    case URL_IS_FTP:
	pglob->gl_closedir = (void (*)(void *)) Closedir;
	pglob->gl_readdir = (struct dirent * (*)(void *)) Readdir;
	pglob->gl_opendir = (void * (*)(const char *)) Opendir;
	pglob->gl_lstat = Lstat;
	pglob->gl_stat = Stat;
	flags |= GLOB_ALTDIRFUNC;
	flags &= ~GLOB_TILDE;
	break;
    case URL_IS_PATH:
	pattern = lpath;
	/*@fallthrough@*/
    case URL_IS_UNKNOWN:
	if (home && home[0])
	    flags |= GLOB_TILDE;
	else
	    flags &= ~GLOB_TILDE;
	break;
    case URL_IS_DASH:
    case URL_IS_HKP:
    case URL_IS_MONGO:	/* XXX FIXME */
    default:
	return rc;
	break;
    }
    return glob(pattern, flags, errfunc, pglob);
}

void Globfree(void *_pglob)
{
    glob_t *pglob = (glob_t *) _pglob;
SPEW("*** Globfree(%p)\n", pglob);
    globfree(pglob);
}

char * Realpath(const char * path, char * resolved_path)
{
    const char * lpath;
    int ut = urlPath(path, &lpath);
    char * rpath = NULL;

SPEW("*** Realpath(%s, %s)\n", path, (resolved_path ? resolved_path : "NULL"));
    /* XXX if POSIXly broken realpath(3) is desired, do that. */
    /* XXX note: preserves current rpmlib realpath(3) usage cases. */
    if (path == NULL || resolved_path != NULL)
	return realpath(path, resolved_path);

    switch (ut) {
    case URL_IS_FTP:
	return ftpRealpath(path, resolved_path);
	break;
    case URL_IS_HTTPS:	
    case URL_IS_HTTP:
    case URL_IS_HKP:
#ifdef WITH_NEON
	return davRealpath(path, resolved_path);
	break;
#endif
	/*@fallthrough@*/
    case URL_IS_MONGO:	/* XXX FIXME */
    default:
	return xstrdup(path);
	break;
    case URL_IS_DASH:
	/* non-GLIBC systems => EINVAL. non-linux systems => EINVAL */
#if defined(__linux__)
	lpath = "/dev/stdin";
#else
	lpath = NULL;
#endif
	break;
    case URL_IS_PATH:		/* XXX note: file:/// prefix is dropped. */
    case URL_IS_UNKNOWN:
	path = lpath;
	break;
    }

    if (lpath == NULL || *lpath == '/')
	rpath = realpath(lpath, resolved_path);
    else {
	char * t;
#if defined(__GLIBC__)
	char * dn = NULL;
#else
	char dn[PATH_MAX];
	dn[0] = '\0';
#endif
	/*
	 * Using realpath on lpath isn't correct if the lpath is a symlink,
	 * especially if the symlink is a dangling link.  What we
	 * do instead is use realpath() on `.' and then append lpath to
	 * the result.
	 */
	if ((t = realpath(".", dn)) != NULL) {
	    rpath = (char *) rpmGetPath(t, "/", lpath, NULL);
	    /* XXX preserve the pesky trailing '/' */
	    if (lpath[strlen(lpath)-1] == '/') {
		char * s = rpath;
		rpath = rpmExpand(s, "/", NULL);
		s = _free(s);
	    }
	} else
	    rpath = NULL;
#if defined(__GLIBC__)
	t = _free(t);
#endif
    }

    return rpath;
}

off_t Lseek(int fdno, off_t offset, int whence)
{
SPEW("*** Lseek(%d,0x%lx,%d)\n", fdno, (long)offset, whence);
    return lseek(fdno, offset, whence);
}
