#ifndef RPMRUBY_H
#define RPMRUBY_H

/**
 * \file rpmio/rpmruby.h
 * \ingroup rpmio
 *
 * Embedded Ruby interpreter
 *
 * This allows the embedding of the Ruby interpreter into RPM5. It can be
 * used, for example, for expanding in the fashion of
 * <tt>%%expand{ruby: ...}</tt>.
 *
 * Currently (as of ruby-1.9.2), the Ruby interpreter allows only one instance
 * of itself per process. As such, rpmio's pooling mechanism will also always
 * contain only one ::rpmruby instance. Calling rpmrubyNew() will initialize
 * a new interpreter, while rpmrubyFree() beds this instance. Make sure you
 * keep things in order, that is, call rpmrubyNew() exactly once, and don't
 * forget to call rpmrubyFree() when you're done. Repeatedly calling
 * rpmrubyNew will have no effect while an interpreter is allocated.
 *
 * You can use rpmrubyRun() to evaluate Ruby code and get the result back.
 */

#include <rpmiotypes.h>
#include <rpmio.h>

/** Triggers printing of debugging information */
extern int _rpmruby_debug;

typedef struct rpmruby_s* rpmruby;

/**
 * Current (global) interpreter instance
 *
 * At the moment, this variable is merely a safeguard against initializing the
 * Ruby interpreter over and over again. In the future, when there is Ruby
 * support for multiple interpreter instances, a flag given to rpmrubyNew()
 * will use this variable and return the global interpreter instance.
 */
extern rpmruby _rpmrubyI;

#if defined(_RPMRUBY_INTERNAL)
/**
 * Instances of this struct carry an embedded Ruby interpreter and the state
 * associated with it.
 */
struct rpmruby_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    void *I;			/*!< The actual interpreter instance */
    unsigned long state;	/*!< interpreter state (e.g. last return code */

};
#endif /* _RPMRUBY_INTERNAL */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Dereferences a Ruby interpreter instance.
 *
 * @param ruby		Ruby interpreter pool item
 * @return		NULL on last dereference
 * @see rpmUnlinkPoolItem
 */
rpmruby rpmrubyUnlink (rpmruby ruby);
#define	rpmrubyUnlink(_ruby) \
    ((rpmruby)rpmioUnlinkPoolItem((rpmioItem)(_ruby), __FUNCTION__, \
        __FILE__, __LINE__))

/**
 * References a Ruby interpreter instance.
 *
 * @param ruby		Ruby interpreter
 * @return		A new ruby interpreter reference
 * @see rpmioLinkPoolItem
 */
rpmruby rpmrubyLink (rpmruby ruby);
#define	rpmrubyLink(_ruby) \
    ((rpmruby)rpmioLinkPoolItem((rpmioItem)(_ruby), __FUNCTION__, \
        __FILE__, __LINE__))

/**
 * Destroys a Ruby interpreter instance.
 *
 * @param ruby		The Ruby interpreter to be destroyed
 * @return		NULL on last dereference
 * @see rpmioFreePoolItem
 */
rpmruby rpmrubyFree(rpmruby ruby);
#define	rpmrubyFree(_ruby)	\
    ((rpmruby)rpmioFreePoolItem((rpmioItem)(_ruby), __FUNCTION__, \
        __FILE__, __LINE__))

/**
 * Creates and initializes a Ruby interpreter.
 *
 * @param av		Arguments to the Ruby interpreter (may be NULL)
 * @param flags		Ruby interpreter flags: ((1<<31): use global interpreter)
 * @return		A new Ruby interpreter
 */
rpmruby rpmrubyNew(char **av, uint32_t flags);

/**
 * Evaluates Ruby code stored in a string.
 *
 * @param ruby		The Ruby interpreter that is to be used
 *			(NULL uses global interpreter)
 * @param str		Ruby code to evaluate (NULL forces return of RPMRC_FAIL)
 * @param resultp	*resultp Result of the evaluation
 * @return		RPMRC_OK on success
 */
rpmRC rpmrubyRun(rpmruby ruby, const char *str, const char **resultp);

#ifdef __cplusplus
}
#endif

#endif /* RPMRUBY_H */
