#ifndef H_RPMURL
#define H_RPMURL

/** \ingroup rpmio
 * \file rpmio/rpmurl.h
 */

#include <sys/types.h>
#include <assert.h>
#include <rpmdefs.h>
#include <rpmiotypes.h>
#include <rpmio.h>
#include <rpmsw.h>

/**
 * Supported URL types.
 */
typedef enum urltype_e {
    URL_IS_UNKNOWN	= 0,	/*!< unknown (aka a file) */
    URL_IS_DASH		= 1,	/*!< stdin/stdout */
    URL_IS_PATH		= 2,	/*!< file://... */
    URL_IS_FTP		= 3,	/*!< ftp://... */
    URL_IS_HTTP		= 4,	/*!< http://... */
    URL_IS_HTTPS	= 5,	/*!< https://... */
    URL_IS_HKP		= 6,	/*!< hkp://... */
    URL_IS_MONGO	= 7,	/*!< mongo://... and mongodb://... */
} urltype;
#define	URL_IS_GIT	(urltype)16
#define	URL_IS_SVN	(urltype)17
#define	URL_IS_SQLITE	(urltype)32
#define	URL_IS_MYSQL	(urltype)33
#define	URL_IS_POSTGRES	(urltype)34
#define	URL_IS_SQLSERVER (urltype)35
#define	URL_IS_MQTT	(urltype)36
#define	URL_IS_MQTTS	(urltype)37
#define	URL_IS_AMQP	(urltype)38
#define	URL_IS_AMQPS	(urltype)39

#define	URLMAGIC	0xd00b1ed0U
#define	URLSANE(u)	assert(u && u->magic == URLMAGIC)

/**
 */
typedef struct urlinfo_s * urlinfo;
typedef struct urlinfo_s * rpmurl;

/**
 */
extern int (*urlNotify) (const urlinfo u, unsigned status);

/**
 */
extern void * urlNotifyArg;

/**
 * URL control structure.
 */
struct urlinfo_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    const char * url;		/*!< copy of original url */
    const char * scheme;	/*!< URI scheme. */
    const char * user;		/*!< URI user. */
    const char * password;	/*!< URI password. */
    const char * host;		/*!< URI host. */
    const char * portstr;	/*!< URI port string. */
    const char * query;		/*!< URI ?query. */
    const char * fragment;	/*!< URI #fragment. */
    const char * proxyu;	/*!< FTP: proxy user */
    const char * proxyh;	/*!< FTP/HTTP: proxy host */
    int proxyp;			/*!< FTP/HTTP: proxy port */
    int	port;			/*!< URI port. */
    urltype ut;			/*!< URI type. */
    FD_t ctrl;			/*!< control channel */
    FD_t data;			/*!< per-xfer data channel */

    void * capabilities;	/*!< neon: ne_server_capabilities ptr */
    void * lockstore;		/*!< neon: ne_lock_store ptr */
    void * sess;		/*!< neon: ne_session ptr */

    const char * date;		/*!< Date: tag. */
    const char * server;	/*!< Server: tag. */
    const char * allow;		/*!< Allow: tag. */

    const char * location;	/*!< Location: tag. */

    const char * etag;		/*!< ETag: tag. */
    int (*notify) (const urlinfo u, unsigned status);
    void * arg;
    struct fdNotify_s {
	unsigned status;
	const char * hostname;
	const char * address;
	int64_t progress;
	int64_t total;
    } info;
    rpmop rop;			/*!< Receive accumulator. */
    rpmop sop;			/*!< Send accumulator. */
    rpmop top;			/*!< Total accumulator. */

    int bufAlloced;		/*!< sizeof I/O buffer */
    char * buf;			/*!< I/O buffer */
    int openError;		/*!< Type of open failure */
    int httpVersion;
    unsigned caps;		/*!< neon: ne_options2 caps bits */
#define	RPMURL_SERVER_DAV_CLASS1	(1 <<  0)
#define	RPMURL_SERVER_DAV_CLASS2	(1 <<  1)
#define	RPMURL_SERVER_DAV_CLASS3	(1 <<  2)
#define	RPMURL_SERVER_MODDAV_EXEC	(1 <<  3)
#define	RPMURL_SERVER_DAV_ACL		(1 <<  4)
#define	RPMURL_SERVER_VER_CONTROL	(1 <<  5)
#define	RPMURL_SERVER_CO_IN_PLACE	(1 <<  6)
#define	RPMURL_SERVER_VER_HISTORY	(1 <<  7)
#define	RPMURL_SERVER_WORKSPACE		(1 <<  8)
#define	RPMURL_SERVER_UPDATE		(1 <<  9)
#define	RPMURL_SERVER_LABEL		(1 << 10)
#define	RPMURL_SERVER_WORK_RESOURCE	(1 << 11)
#define	RPMURL_SERVER_MERGE		(1 << 12)
#define	RPMURL_SERVER_BASELINE		(1 << 13)
#define	RPMURL_SERVER_ACTIVITY		(1 << 14)
#define	RPMURL_SERVER_VC_COLLECTION	(1 << 15)
#define	RPMURL_SERVER_EXT_MKCOL		(1 << 16)
	/* 17 - 23 unused */
#define	RPMURL_SERVER_OPTIONSDONE	(1 << 24)
#define	RPMURL_SERVER_HASRANGE		(1 << 25)

#define	RPMURL_SERVER_HASDAV	(RPMURL_SERVER_DAV_CLASS1|RPMURL_SERVER_DAV_CLASS2|RPMURL_SERVER_DAV_CLASS3|RPMURL_SERVER_MODDAV_EXEC)
    unsigned magic;
};

#ifdef __cplusplus
extern "C" {
#endif

extern int _url_count;		/*!< No. of cached URL's. */

extern urlinfo * _url_cache;	/*!< URL cache. */

extern int _url_iobuf_size;	/*!< Initial size of URL I/O buffer. */
#define RPMURL_IOBUF_SIZE	4096

extern int _rpmurl_debug;		/*!< URL debugging? */
#define RPMURL_DEBUG_IO		0x40000000
#define RPMURL_DEBUG_REFS	0x20000000

/**
 * Create a URL control structure instance.
 * @param msg		debugging identifier (unused)
 * @return		new instance
 */
urlinfo	urlNew(const char * msg);

/** @todo Remove debugging entry from the ABI. */
urlinfo	XurlNew(const char * msg, const char * fn, unsigned ln);
#define	urlNew(_msg) XurlNew(_msg, __FILE__, __LINE__)

/**
 * Reference a URL control structure instance.
 * @param u		URL control structure
 * @param msg		debugging identifier (unused)
 * @return		referenced instance
 */
urlinfo	urlLink(urlinfo u, const char * msg);
#define	urlLink(_u, _msg)	\
	(urlinfo) rpmioLinkPoolItem((rpmioItem)(_u), _msg, __FILE__, __LINE__)

/**
 * Dereference a URL control structure instance.
 * @param u		URL control structure
 * @param msg		debugging identifier (unused)
 * @return		dereferenced instance (NULL if freed)
 */
urlinfo	urlFree(urlinfo u, const char * msg);
#define	urlFree(_u, _msg)	\
	((urlinfo)rpmioFreePoolItem((rpmioItem)(_u), _msg, __FILE__, __LINE__))

/**
 * Free cached URL control structures.
 */
void urlFreeCache(void);

/**
 * Return type of URL.
 * @param url		url string
 * @return		type of url
 */
urltype	urlIsURL(const char * url)
	RPM_GNUC_PURE;
urltype	urlType(void * _u)
	RPM_GNUC_PURE;

/**
 * Return path component of URL.
 * @param url		url string
 * @retval pathp	pointer to path component of url
 * @return		type of url
 */
urltype	urlPath(const char * url, const char ** pathp);

/**
 * Parse URL string into a control structure.
 * @param url		url string
 * @retval uret		address of new control instance pointer
 * @return		0 on success, -1 on error
 */
int urlSplit(const char * url, urlinfo * uret);

/**
 * Copy data from URL to local file.
 * @param url		url string of source
 * @param dest		file name of destination
 * @return		0 on success, otherwise FTPERR_* code
 */
int urlGetFile(const char * url, const char * dest);

#ifdef __cplusplus
}
#endif

#endif	/* H_RPMURL */
