
/**
 * \file rpmio/tmq.c
 */

#include "system.h"

#include <rpmio_internal.h>
#include <rpmlog.h>
#include <rpmmacro.h>
#include <poptIO.h>
#include <argv.h>

# define _RPMEIO_INTERNAL
# include <rpmeio.h>

#if defined(HAVE_EV_H)
# include <ev.h>
#endif

#define	_RPMEV_INTERNAL
#include <rpmev.h>

#define	_RPMW_INTERNAL
#include <rpmw.h>

#include "debug.h"

static int _debug = 0;

#define SPEW(_fmt, ...) \
    if (_debug || _rpmeio_debug || _rpmio_debug) \
	fprintf(stderr, _fmt, ##__VA_ARGS__)

#define Z(_rc)	assert((_rc) == 0)

/*==============================================================*/
static int doEV(ARGV_t av, int ac)
{
    int rc = 0;

SPEW("    version: %u.%u\n", ev_version_major(), ev_version_minor());

char *rstr = NULL;
unsigned mask = 0;
mask = ev_supported_backends();
rstr = rpmioB2N(rpmioEVBACKENDTable, mask);
SPEW("  supported: %s\n", rstr);
rstr = _free(rstr);

mask = ev_recommended_backends();
rstr = rpmioB2N(rpmioEVBACKENDTable, mask);
SPEW("recommended: %s\n", rstr);
rstr = _free(rstr);

mask = ev_embeddable_backends();
rstr = rpmioB2N(rpmioEVBACKENDTable, mask);
SPEW(" embeddable: %s\n", rstr);
rstr = _free(rstr);

SPEW("       time: %f\n", ev_time());

    /* use the default event loop unless you have special needs */
    struct ev_loop *loop = ev_default_loop(EVFLAG_AUTO);
    rpmev ev = NULL;

    /* initialise an io watcher, then start it */
    /* this one will watch for stdin to become readable */
    ev = rpmevNew(RPMEV_TYPE_IO,	NULL, STDIN_FILENO, EV_READ);
#ifdef	DYING
    {	ev_io * w = &ev->evw.io;
	w->fd = STDIN_FILENO;
	w->events = EV_READ;
	ev_io_start(loop, w);
    }
#endif

    /* initialise a timer watcher, then start it */
    /* simple non-repeating 5.5 second timeout */
    ev = rpmevNew(RPMEV_TYPE_TIMER,	NULL, 0,0, 5.5);
#ifdef	DYING
    {	ev_timer * w = &ev->evw.timer;
	w->at = 0.0;
	w->repeat = 5.5;
	ev_timer_start(loop, w);
    }
#endif

    ev = rpmevNew(RPMEV_TYPE_PERIODIC,	NULL, 0.0, 11.0, NULL );
#ifdef	DYING
    {	ev_periodic * w = &ev->evw.periodic;
	w->offset = 0.0;
	w->interval = 11.0;
	w->reschedule_cb = NULL;
	ev_periodic_start(loop, w);
    }
#endif

    ev = rpmevNew(RPMEV_TYPE_SIGNAL,	NULL, SIGUSR1);
#ifdef	DYING
    {	ev_signal * w = &ev->evw.signal;
	w->signum = SIGUSR1;
	ev_signal_start(loop, w);
    }
#endif

#ifdef	DYING
    ev = rpmevNew(RPMEV_TYPE_CHILD,	NULL, 0, 0);
    {	ev_child * w = &ev->evw.child;
	w->pid = 0;
	w->flags = 0;	/** XXX 0=terminate, 1=stop/continue */
    }
#endif

    const char * _path = "/X/src/wdj54/rpmio/tmq";
    ev = rpmevNew(RPMEV_TYPE_STAT,	NULL, _path, 0.0);
#ifdef	DYING
    {	ev_stat * w = &ev->evw.stat;
	w->path = _free(w->path);
	w->path = xstrdup(_path);
	w->interval = 0;	/* XXX 0=recommended ~5s, must be >0.1 */
	w->wd = -2;
	ev_stat_start(loop, w);
    }
#endif

    ev = rpmevNew(RPMEV_TYPE_IDLE,	NULL);
#ifdef	DYING
    {	ev_idle * w = &ev->evw.idle;
	ev_idle_start(loop, w);
    }
#endif

#ifdef	NOTYET
    ev = rpmevNew(RPMEV_TYPE_PREPARE,	NULL);
    ev = rpmevNew(RPMEV_TYPE_CHECK,	NULL);
    ev = rpmevNew(RPMEV_TYPE_FORK,	NULL);
    ev = rpmevNew(RPMEV_TYPE_CLEANUP,	NULL);
    ev = rpmevNew(RPMEV_TYPE_EMBED,	NULL, NULL);
    ev = rpmevNew(RPMEV_TYPE_ASYNC,	NULL);
#endif

    /* now wait for events to arrive */
    ev_run(loop, 0);

    /* unloop was called, so exit */

    /* Deallocate all events. */
    while ((ev = rpmevqGetHead()) != NULL)
	ev = rpmevFree(ev);

    ev_loop_destroy(loop);

    return rc;
}

/*==============================================================*/
static int doEIO(ARGV_t av, int ac)
{
    int rc = 0;
    double max_poll_time = 0.0;
    unsigned max_poll_reqs = 0;
    unsigned min_parallel = 0;
    unsigned max_parallel = 0;
    unsigned max_idle = 0;
    unsigned idle_timeout = 0;
    unsigned flags = 0;	/* RPMEVL_PIPE */

    rpmevl evl = rpmeioStart(max_poll_time, max_poll_reqs,
			min_parallel, max_parallel,
			max_idle, idle_timeout, flags);

    do {
	/* avoid relative paths yourself(!) */
	Eio_mkdir("eio-test-dir", 0777, 0, rpmeioCB, "mkdir");
	Eio_nop(0, rpmeioCB, "nop");
	rpmeioLoop(evl);

	Eio_stat("eio-test-dir", 0, rpmeioCB_stat, "stat");
	Eio_lstat("eio-test-dir", 0, rpmeioCB_stat, "stat");
	Eio_open("eio-test-dir/eio-test-file", O_RDWR | O_CREAT, 0777, 0,
		 rpmeioCB_open, "open");
	Eio_symlink("test", "eio-test-dir/eio-symlink", 0, rpmeioCB,
		    "symlink");
	Eio_mknod("eio-test-dir/eio-fifo", S_IFIFO, 0, 0, rpmeioCB, "mknod");
	rpmeioLoop(evl);

	Eio_utime("eio-test-dir", 12345.678, 23456.789, 0, rpmeioCB,
		  "utime");
	Eio_futime(rpmeioLastFdno, 92345.678, 93456.789, 0, rpmeioCB, "futime");
	Eio_chown("eio-test-dir", getuid(), getgid(), 0, rpmeioCB, "chown");
	Eio_fchown(rpmeioLastFdno, getuid(), getgid(), 0, rpmeioCB, "fchown");
	Eio_fchmod(rpmeioLastFdno, 0723, 0, rpmeioCB, "fchmod");
	Eio_readdir("eio-test-dir", 0, 0, rpmeioCB_readdir, "readdir");
	Eio_readdir("/nonexistant", 0, 0, rpmeioCB_readdir, "readdir");
	Eio_fstat(rpmeioLastFdno, 0, rpmeioCB_stat, "stat");
	Eio_write(rpmeioLastFdno, "test\nfail\n", 10, 4, 0, rpmeioCB, "write");
	rpmeioLoop(evl);

	Eio_read(rpmeioLastFdno, 0, 8, 0, EIO_PRI_DEFAULT, rpmeioCB_read, "read");
	Eio_readlink("eio-test-dir/eio-symlink", 0, rpmeioCB, "readlink");
	rpmeioLoop(evl);

	Eio_dup2(1, 2, EIO_PRI_DEFAULT, rpmeioCB, "dup");	// dup stdout to stderr
	Eio_chmod("eio-test-dir", 0765, 0, rpmeioCB, "chmod");
	Eio_ftruncate(rpmeioLastFdno, 9, 0, rpmeioCB, "ftruncate");
	Eio_fdatasync(rpmeioLastFdno, 0, rpmeioCB, "fdatasync");
	Eio_fsync(rpmeioLastFdno, 0, rpmeioCB, "fsync");
	Eio_sync(0, rpmeioCB, "sync");
	Eio_busy(0.5, 0, rpmeioCB, "busy");
	rpmeioLoop(evl);

	Eio_sendfile(1, rpmeioLastFdno, 4, 5, 0, rpmeioCB, "sendfile");	// write "test\n" to stdout
	Eio_fstat(rpmeioLastFdno, 0, rpmeioCB_stat, "stat");
	rpmeioLoop(evl);

	Eio_truncate("eio-test-dir/eio-test-file", 6, 0, rpmeioCB,
		     "truncate");
	Eio_readahead(rpmeioLastFdno, 0, 64, 0, rpmeioCB, "readahead");
	rpmeioLoop(evl);

	Eio_close(rpmeioLastFdno, 0, rpmeioCB, "close");
	Eio_link("eio-test-dir/eio-test-file",
		 "eio-test-dir/eio-test-file-2", 0, rpmeioCB, "link");
	rpmeioLoop(evl);

	Eio_rename("eio-test-dir/eio-test-file",
		   "eio-test-dir/eio-test-file-renamed", 0, rpmeioCB,
		   "rename");
	rpmeioLoop(evl);

	Eio_unlink("eio-test-dir/eio-fifo", 0, rpmeioCB, "unlink");
	Eio_unlink("eio-test-dir/eio-symlink", 0, rpmeioCB, "unlink");
	Eio_unlink("eio-test-dir/eio-test-file-2", 0, rpmeioCB, "unlink");
	Eio_unlink("eio-test-dir/eio-test-file-renamed", 0, rpmeioCB,
		   "unlink");
	rpmeioLoop(evl);

	Eio_rmdir("eio-test-dir", 0, rpmeioCB, "rmdir");
	rpmeioLoop(evl);

    } while (0);

    rpmeioDrain(evl);

    rc = 0;	/* XXX */

    return rc;
}

/*==============================================================*/

static char rpmwSedArgs[] =
    " -e s/[^,]*\\\\*[^,]*/%p/g"
    " -e s/[^,]*\\\\][^,]*/%p/g"
    " -e s/[^,]*enum\\\\s[^,]*,/0x%x/g"
    " -e s/[^,]*void//g"
    " -e s/[^,]*(caddr_t|timer_t|cap_user_data_t|cap_user_header_t|eio_wd|eio_req)/%p/g"
    " -e s/[^,]*off64_t/0x%llx/g"
    " -e s/[^,]*nfds_t/%lu/g"
    " -e s/[^,]*off_t/0x%lx/g"
    " -e s/[^,]*aio_context_t/0x%lx/g"
    " -e s/[^,]*(mqd_t|key_serial_t|key_t|dev_t|socklen_t|clockid_t)/0x%x/g"
    " -e s/[^,]*(uid_t|gid_t|pid_t|id_t)/%u/g"
    " -e s/[^,]*mode_t/0%o/g"
    " -e s/[^,]*ssize_t/%zd/g"
    " -e s/[^,]*size_t/%zu/g"
    " -e s/[^,]*double/%f/g"
    " -e s/[^,]*struct\\\\s(timeval|timespec)/%f/g"

    " -e s/[^,]*(u64|uint64_t|unsigned\\\\slong\\\\slong)/%lu/g"
    " -e s/[^,]*(i64|int64_t|long\\\\slong)/%ld/g"
    " -e s/[^,]*(unsigned\\\\slong)/%lu/g"
    " -e s/[^,]*(long)/%ld/g"
    " -e s/[^,]*(u32|u16|u8|uint32_t|uint16_t|uint8_t|unsigned)/%u/g"
    " -e s/[^,]*(i32|i16|i8|int32_t|int16_t|int8_t|int)/%d/g"
    ;

/*==============================================================*/

static int rpmwExpand(rpmw w, FILE * fp)
{
    int rc = 0;
    char *xstr;
    char * fmt;
    char * msg;
    int xx;

    if (fp == NULL)	fp = stderr;

    xstr = rpmMCExpand(w->mc, "%{?Subtype}", NULL);
    if (xstr && *xstr) {
	fmt = rpmMCExpand(w->mc,
		"%{sed", rpmwSedArgs, ":", xstr, "}",
		NULL);
	if (fmt && *fmt)
	   addMacro(w->mc, "Subtypefmt", NULL, fmt, 0);
	fmt = _free(fmt);
    }
    xstr = _free(xstr);

    xstr = rpmMCExpand(w->mc, "%{@argtypes:,}", NULL);
ARGV_t argtypes = NULL;
    xx = argvSplit(&argtypes, xstr, ",");
    fmt = rpmMCExpand(w->mc,
		"%{sed", rpmwSedArgs, ":", xstr, "}",
		NULL);
addMacro(w->mc, "argsfmt", NULL, fmt, 0);
    xstr = _free(xstr);

ARGV_t argfmts = NULL;
    xx = argvSplit(&argfmts, fmt, ",");
    fmt = _free(fmt);

/*===================================*/

    char b[BUFSIZ];
    char *te = b;

    char * subname = rpmMCExpand(w->mc, "%{subname}", NULL);

    *te++ = '\n';
    te = stpcpy(stpcpy(te, "/* ========= "), subname);
    *te++ = '\n';

    te += sprintf(te, "    case EIO_%s:\n", subname);

    for (int i = 0; argtypes[i]; i++) {
	const char * argtype = argtypes[i];
	const char * argfmt = argfmts[i];

	if (argfmt == NULL || *argfmt == '\0') {
	    continue;
	} else
	if (!strcmp(argfmt, "%p")) {
	    w->nptrs++;
	    te += sprintf(te, "\treq->ptr%u = va_arg(ap, %s);\n", w->nptrs, argtype);
	} else
	if (!strcmp(argfmt, "%d")
	 || !strcmp(argfmt, "%u")
	 || !strcmp(argfmt, "%x")
	 || !strcmp(argfmt, "0x%x")
	 || !strcmp(argfmt, "%o")
	 || !strcmp(argfmt, "0%o")
	) {
	    w->nints++;
	    te += sprintf(te, "\treq->int%u = va_arg(ap, %s);\n", w->nints, argtype);
	} else
	if (!strcmp(argfmt, "%ld")
	 || !strcmp(argfmt, "%lu")
	 || !strcmp(argfmt, "%lx")
	 || !strcmp(argfmt, "0x%lx")
	) {
	    w->noffs++;
	    te += sprintf(te, "\treq->off%u = va_arg(ap, %s);\n", w->noffs, argtype);
	} else
	if (!strcmp(argfmt, "%zd")
	 || !strcmp(argfmt, "%zu")
	) {
	    w->nsizes++;
	    te += sprintf(te, "\treq->size%u = va_arg(ap, %s);\n", w->nsizes, argtype);
	} else
	if (!strcmp(argfmt, "%f")) {
	    w->ndoubles++;
	    te += sprintf(te, "\treq->nav%u = va_arg(ap, %s);\n", w->ndoubles, argtype);
	} else
	if (!strcmp(argfmt, "%lld")
	 || !strcmp(argfmt, "%llu")
	 || !strcmp(argfmt, "%llx")
	 || !strcmp(argfmt, "0x%llx")
	) {
	    w->nlonglongs++;
	    te += sprintf(te, "\treq->longlong%u = va_arg(ap, %s);\n", w->nlonglongs, argtype);
	} else
	{
	    w->nother++;
	    te += sprintf(te, "\treq->other%u = va_arg(ap, %s);\n", w->nother, argtype);
	}
    }
    te += sprintf(te, "\tbreak;\n");

    te = stpcpy(te, "*/");
    *te++ = '\n';

    *te = '\0';

fprintf(fp, "%s", b);

    subname = _free(subname);
    argtypes = argvFree(argtypes);
    argfmts = argvFree(argfmts);

/*===================================*/

addMacro(w->mc, "init",  NULL, "%{subtype} %{Subname}(%{arglist})", 0);
addMacro(w->mc, "enter", NULL, "%{?Subtype:%{subtype} rc;}", 0);
addMacro(w->mc, "call",  NULL, "%{?Subtype:rc = }%{subname}(%{@argnames:, });", 0);
addMacro(w->mc, "spew",  NULL, "SPEW(\"<-- %s(%{argsfmt})%{?Subtype: rc %{Subtypefmt}}\\n\", __FUNCTION__, %{@argnames:, }%{?Subtype:, rc});", 0);
addMacro(w->mc, "exit",  NULL, "%{?Subtype:return rc;}",0);
addMacro(w->mc, "wrapper",  NULL,
"\n\
%{init};\n\
%{init}\n\
{\n\
    %{?enter}\n\
    %{?pre}\n\
    %{?call}\n\
    %{?post}\n\
%{?spew}\n\
    %{?exit}\n\
}", 0);

    msg = rpmMCExpand(w->mc, "%{wrapper}\n", NULL);
fprintf(fp, "%s", msg);
    msg = _free(msg);

    return rc;
}

/*==============================================================*/
static int rpmwParse(rpmw w, char * proto)
{
    int rc = 0;
    char *t;
    char *te;
    int xx;

addMacro(w->mc, "proto", NULL, proto, 0);
    t = strchr(proto, '(');
    *t++ = '\0';
    char *arglist = t;
    t = strrchr(arglist, ')');
    *t++ = '\0';
addMacro(w->mc, "arglist", NULL, arglist, 0);

    char * subtype = proto;

    /* XXX need to handle '**' */
    if ((t = strrchr(subtype, '*')) != NULL) {
	if (t[-1] == ' ' && t[1] != ' ') {
	    t[-1] = '*';
	    t[0] = ' ';
	}
    }
    if ((t = strrchr(subtype, ' ')) != NULL)
	*t++ = '\0';
    else
	t = proto;

    char * subname = t;
addMacro(w->mc, "subname", NULL, subname, 0);
char savec = *subname;
*subname = toupper(*subname);
addMacro(w->mc, "Subname", NULL, subname, 0);
*subname = savec;

addMacro(w->mc, "subtype", NULL, subtype, 0);
if (strcmp(subtype, "void")) {
addMacro(w->mc, "Subtype", NULL, subtype, 0);
}

    ARGV_t items = NULL;
    xx = argvSplit(&items, arglist, ",");

    int nitems = argvCount(items);
    for (int i = 0; i < nitems; i++) {
	char * item = (char *) items[i];

	/* Trim white space. */
	for (t = item; *t == ' ';) t++;
	for (te = t + strlen(t); te > t && te[-1] == ' ';) te--;
	item = strndup(t, (te-t));
	items[i] = _free(items[i]);
	items[i] = item;
addMacro(w->mc, "items", NULL, item, 0);
	char * argtype = item;

	/* XXX need to handle '**' */
	if ((t = strrchr(item, '*')) != NULL) {
	    if (t[-1] == ' ' && t[1] != ' ') {
		t[-1] = '*';
		t[0] = ' ';
	    }
	}

	if ((t = strrchr(item, ' ')) != NULL)
	    *t++ = '\0';
	else
	    t = "";
	char * argname = t;
addMacro(w->mc, "argtypes", NULL, argtype, 0);
addMacro(w->mc, "argnames", NULL, argname, 0);
    }

    items = argvFree(items);

    return rc;
}

/*==============================================================*/
static int spewEIO(ARGV_t av, int ac)
{
const char * _fn = "rpmw.prototypes";
    FD_t fd = Fopen(_fn, "r.fpio");
ARGV_t protos = NULL;
    int rc = argvFgets(&protos, fd, 1);
    (void) Fclose(fd);

    if (rc == 0 && protos)
    for (int i = 0; protos[i]; i++) {

rpmw w = rpmwNew();
	char * proto = xstrdup(protos[i]);

	rc = rpmwParse(w, proto);
	rc = rpmwExpand(w, NULL);

	proto = _free(proto);

if (_debug)
rpmwDump(__FUNCTION__, w, NULL);

w = rpmwFree(w);

    }

protos = argvFree(protos);

    return rc;
}

/*==============================================================*/
#if !defined(POPT_BIT_XOR)
#define	POPT_BIT_XOR	(POPT_ARG_VAL|POPT_ARGFLAG_XOR)
#endif

static struct poptOption rpmmqOptionsTable[] = {
 { "debug", 'd', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN,	&_debug, 1,
	NULL, NULL },

#ifdef	DYING
#define	_MSQBIT(_v)	(RPMMSQ_FLAGS_##_v)
 { "qname", 'Q', POPT_ARG_STRING,		    &qname, 0,
	N_("Use message queue QNAME"), N_("QNAME") },
 { "posix", 'P', POPT_BIT_SET,                      &flags, _MSQBIT(POSIX),
	N_("Use POSIX message queues"), NULL },
 { "sysv", 'S', POPT_BIT_SET,                       &flags, _MSQBIT(SYSV),
	N_("Use SysV message queues"), NULL },

 { "reset", 'R', POPT_BIT_SET|POPT_ARGFLAG_TOGGLE,  &flags, _MSQBIT(RESET),
	N_("Remove queue before opening"), NULL },
 { "info", 'I', POPT_BIT_SET|POPT_ARGFLAG_TOGGLE,   &flags, _MSQBIT(INFO),
	N_("Display queue info on close"), NULL },
 { "delete", 'D', POPT_BIT_SET|POPT_ARGFLAG_TOGGLE, &flags, _MSQBIT(DELETE),
	N_("Remove queue after closing"), NULL },

 { "count", 'c', POPT_ARG_INT,			    &count, 0,
	N_("Send COUNT messages"), N_("COUNT") },
 { "priority", 'p', POPT_ARG_LONG,		    &priority, 0,
	N_("Send message PRIORITY"), N_("PRIORITY") },
#undef	_MSQBIT
#endif	/* DYING */

 { NULL, '\0', POPT_ARG_INCLUDE_TABLE, rpmioOpenFlagsTable, 0,
	N_("Open(2) flags:"), NULL },

 { NULL, '\0', POPT_ARG_INCLUDE_TABLE, rpmioAllPoptTable, 0,
	N_("Common options for all rpmio executables:"), NULL },

  POPT_AUTOALIAS
  POPT_AUTOHELP
  POPT_TABLEEND
};

int
main(int argc, char *argv[])
{
    poptContext con = rpmioInit(argc, argv, rpmmqOptionsTable);
    ARGV_t av = poptGetArgs(con);
    int ac = argvCount(av);;
    int ec = 0;

#ifdef	DYING
    _rpmmsqOFlags = rpmioOpenFlags;
    if (qname == NULL)
	qname = xstrdup("rpm");
#endif

    switch (6) {
    case 5:	ec = doEV(av, ac);	break;
    case 6:	ec = doEIO(av, ac);	break;
    case 7:	ec = spewEIO(av, ac);	break;
    }

#ifdef	DYING
    qname = _free(qname);
#endif
    con = rpmioFini(con);

    return ec;
}
