/**
 * Copyright 2016-present, Yann Collet, Facebook, Inc.
 * All rights reserved.
 *
 * This source code is licensed under the license found in the
 * LICENSE-examples file in the root directory of this source tree.
 */

#include "system.h"

#include <sys/mount.h>
#include <mntent.h>

#include "rpmio.h"
#include <rpmlog.h>
#include <rpmmacro.h>
#include <poptIO.h>
#include <argv.h>

#include "debug.h"

typedef	struct rpmmu_s * rpmmu;
struct rpmmu_s {
    ARGV_t av;
    int ac;
    unsigned flags;

    const char ** options;

    const char * source;
    const char * target;
    const char * fstype;
    unsigned long mflags;
    const void * data;

    int uflags;
};

/*==============================================================*/
static int rpmmuList(const char *fn, FILE *ofp)
{
    FILE * ifp = NULL;
    struct mntent *mnt;
    int rc = -2;
    int ix = 0;

    if (ofp == NULL)
	ofp = stderr;

    ifp = setmntent(fn, "r");
    if (ifp == NULL)
	goto exit;

#ifdef	NOTYET
    struct mntent mntbuf;
    char b[BUFSIZ];
    size_t nb = sizeof(nb);
    while ((mnt = getmntent_r(ifp, &mntbuf, b, nb)) != NULL)
#else
    while ((mnt = getmntent(ifp)) != NULL)
#endif
    {
	fprintf(ofp, "========= %s #%d\n", fn, ix++);
	fprintf(ofp, "%8s: %s\n", "fsname", mnt->mnt_fsname);
	fprintf(ofp, "%8s: %s\n", "dir", mnt->mnt_dir);
	fprintf(ofp, "%8s: %s\n", "type", mnt->mnt_type);
	fprintf(ofp, "%8s: %s\n", "opts", mnt->mnt_opts);
	fprintf(ofp, "%8s: %d\n", "freq", mnt->mnt_freq);
	fprintf(ofp, "%8s: %d\n", "passno", mnt->mnt_passno);
    }

    rc = 0;

exit:
    if (ifp)
	(void) endmntent(ifp);
    return rc;
}

static int rpmmuGetFSTypes(ARGV_t *avp, const char *fnav[])
{
    const char *fn = NULL;
    FD_t fd = NULL;
    ARGV_t av = NULL;
    int ac = 0;
    int rc = 0;
    int xx;

    if (fnav)
    while ((fn = *fnav++) != NULL) {
fprintf(stderr, "*** Fopen(%s)\n", fn);
	fd = Fopen(fn, "rbm.fpio");
	if (fd && !Ferror(fd)) {
	    ARGV_t nav = NULL;
	    int nac;
	    xx = argvFgets(&nav, fd, 0);
	    xx = Fclose(fd);
	    nac = argvCount(nav);
	    for (int i = 0; i < nac; i++) {
		ARGV_t xav = NULL;
		xx = argvSplit(&xav, nav[i], NULL);
		if (xav[1]) {
		    const char * s = xav[0];
		    xav[0] = xav[1];
		    xav[1] = s;
		}
		nav[i] = _free(nav[i]);
		nav[i] = argvJoin(xav, ':');
		xav = argvFree(xav);
	    }
	    xx = argvAppend(&av, nav);
	    nav = argvFree(nav);
	}
	int ac = argvCount(av);
	if (!strcmp(av[ac-1], "*")) {
	    av[ac-1] = _free(av[ac-1]);
	    continue;
	}
	break;
    }

    xx = argvSort(av, NULL);
    ac = argvCount(av);
    for (int i = 1, j = 0; i < ac; i++) {
	if (!strcmp(av[i], av[j]))
	    av[i] = _free(av[i]);
	else if (++j < i) {
	    av[j] = av[i];
	    av[i] = NULL;
	}
    }

argvPrint(__FUNCTION__, av, NULL);

    if (avp)
	*avp = av;
    else
	av = argvFree(av);

    return rc;
}

/*==============================================================*/

#if !defined(POPT_BIT_XOR)
#define	POPT_BIT_XOR	(POPT_ARG_VAL|POPT_ARGFLAG_XOR)
#endif

#ifdef	REFERENCE
       -a, --all
              Mount all filesystems (of the given types)  mentioned  in  fstab
              (except  for those whose line contains the noauto keyword).  The
              filesystems are mounted following their order in fstab.

       -B, --bind
              Remount a subtree somewhere  else  (so  that  its  contents  are
              available in both places).  See above, under Bind mounts.

       -c, --no-canonicalize
              Dont  canonicalize  paths.  The mount command canonicalizes all
              paths (from command line or fstab) by default.  This option  can
              be  used  together  with  the  -f flag for already canonicalized
              absolute paths.  The option is designed for mount helpers  which
              call  mount -i.  It is strongly recommended to not use this com‐
              mand-line option for normal mount operations.

              Note  that  mount(8)  does  not  pass   this   option   to   the
              /sbin/mount.type helpers.

       -F, --fork
              (Used  in  conjunction  with -a.)  Fork off a new incarnation of
              mount for each device.  This will do  the  mounts  on  different
              devices  or  different  NFS  servers  in parallel.  This has the
              advantage that it is faster; also NFS timeouts go  in  parallel.
              A  disadvantage  is that the mounts are done in undefined order.
              Thus, you cannot use this option if you want to mount both  /usr
              and /usr/spool.

       -f, --fake
              Causes  everything to be done except for the actual system call;
              if its not obvious, this  "fakes"  mounting  the  filesystem.
              This  option is useful in conjunction with the -v flag to deter‐
              mine what the mount command is trying to do.   It  can  also  be
              used  to  add entries for devices that were mounted earlier with
              the -n option.  The -f option checks for an existing  record  in
              /etc/mtab and fails when the record already exists (with a regu‐
              lar non-fake mount, this check is done by the kernel).

       -i, --internal-only
              Dont call the /sbin/mount.filesystem helper even if it exists.

       -L, --label label
              Mount the partition that has the specified label.

       -l, --show-labels
              Add the labels in the mount output.  mount must have  permission
              to  read  the  disk device (e.g. be suid root) for this to work.
              One can set such a label  for  ext2,  ext3  or  ext4  using  the
              e2label(8)  utility, or for XFS using xfs_admin(8), or for reis‐
              erfs using reiserfstune(8).

       -M, --move
              Move a subtree to some other place.  See above,  the  subsection
              The move operation.

       -n, --no-mtab
              Mount without writing in /etc/mtab.  This is necessary for exam‐
              ple when /etc is on a read-only filesystem.

       -O, --test-opts opts
              Limit the set of filesystems to which the -a option applies.  In
              this  regard  it is like the -t option except that -O is useless
              without -a.  For example, the command:

                     mount -a -O no_netdev

              mounts all filesystems except those which have the option  _net‐
              dev specified in the options field in the /etc/fstab file.

              It  is different from -t in that each option is matched exactly;
              a leading no at the beginning of one option does not negate  the
              rest.

              The  -t  and  -O  options are cumulative in effect; that is, the
              command

                     mount -a -t ext2 -O _netdev

              mounts all ext2 filesystems with the  _netdev  option,  not  all
              filesystems  that  are  either  ext2  or have the _netdev option
              specified.

       -o, --options opts
              Use the specified mount options.  The opts argument is a  comma-
              separated list.  For example:

                     mount LABEL=mydisk -o noatime,nodev,nosuid

              For  more  details, see the FILESYSTEM-INDEPENDENT MOUNT OPTIONS
              and FILESYSTEM-SPECIFIC MOUNT OPTIONS sections.

       -R, --rbind
              Remount a subtree and all possible submounts somewhere else  (so
              that its contents are available in both places).  See above, the
              subsection Bind mounts.

       -r, --read-only
              Mount the filesystem read-only.  A synonym is -o ro.

              Note that, depending on the filesystem type,  state  and  kernel
              behavior,  the  system may still write to the device.  For exam‐
              ple, ext3 and ext4 will replay the journal if the filesystem  is
              dirty.   To  prevent  this kind of write access, you may want to
              mount an ext3  or  ext4  filesystem  with  the  ro,noload  mount
              options  or  set  the block device itself to read-only mode, see
              the blockdev(8) command.

       -s     Tolerate sloppy mount options rather than  failing.   This  will
              ignore  mount  options  not supported by a filesystem type.  Not
              all filesystems support this option.  Currently  its  supported
              by the mount.nfs mount helper only.

X      --source device
              If  only  one  argument  for the mount command is given then the
              argument might be interpreted as target (mountpoint)  or  source
              (device).   This  option  allows  to  explicitly define that the
              argument is the mount source.

X      --target directory
              If only one argument for the mount command  is  given  then  the
              argument  might  be interpreted as target (mountpoint) or source
              (device).  This option allows  to  explicitly  define  that  the
              argument is the mount target.

       -T, --fstab path
              Specifies  an  alternative  fstab  file.  If path is a directory
              then the files in the directory  are  sorted  by  strverscmp(3);
              files  that  start  with  "." or without an .fstab extension are
              ignored.  The option can be  specified  more  than  once.   This
              option  is mostly designed for initramfs or chroot scripts where
              additional configuration is  specified  beyond  standard  system
              configuration.

              Note  that  mount(8)  does  not  pass  the option --fstab to the
              /sbin/mount.type helpers, meaning  that  the  alternative  fstab
              files will be invisible for the helpers.  This is no problem for
              normal mounts, but user (non-root) mounts always  require  fstab
              to verify the users rights.

       -t, --types fstype
              The argument following the -t is used to indicate the filesystem
              type.  The filesystem types which are currently supported depend
              on  the  running  kernel.   See  /proc/filesystems and /lib/mod‐
              ules/$(uname -r)/kernel/fs for a complete list of  the  filesys‐
              tems.   The  most common are ext2, ext3, ext4, xfs, btrfs, vfat,
              sysfs, proc, nfs and cifs.

              The programs mount and umount support filesystem subtypes.   The
              subtype   is  defined  by  a  '.subtype'  suffix.   For  example
              'fuse.sshfs'.  Its recommended to use subtype  notation  rather
              than   add   any   prefix  to  the  mount  source  (for  example
              'sshfs#example.com' is deprecated).

              If no -t option is given, or if  the  auto  type  is  specified,
              mount  will try to guess the desired type.  Mount uses the blkid
              library for guessing the filesystem type; if that does not  turn
              up anything that looks familiar, mount will try to read the file
              /etc/filesystems, or, if that does not exist, /proc/filesystems.
              All  of  the filesystem types listed there will be tried, except
              for those that are labeled "nodev" (e.g. devpts, proc and  nfs).
              If  /etc/filesystems  ends in a line with a single *, mount will
              read /proc/filesystems afterwards.  While trying, all filesystem
              types will be mounted with the mount option silent.

              The auto type may be useful for user-mounted floppies.  Creating
              a file /etc/filesystems can be useful to change the probe  order
              (e.g.,  to  try vfat before msdos or ext3 before ext2) or if you
              use a kernel module autoloader.

              More than one type may be specified in a  comma-separated  list,
              for  option  -t  as well as in an /etc/fstab entry.  The list of
              filesystem types for option -t can be prefixed with no to  spec‐
              ify  the  filesystem  types  on which no action should be taken.
              The prefix no has no effect  when  specified  in  an  /etc/fstab
              entry.

              The  prefix  no can be meaningful with the -a option.  For exam‐
              ple, the command

                     mount -a -t nomsdos,smbfs

              mounts all filesystems except those of type msdos and smbfs.

              For most types all the mount program has to do is issue a simple
              mount(2)  system call, and no detailed knowledge of the filesys‐
              tem type is required.  For a few types however (like nfs,  nfs4,
              cifs, smbfs, ncpfs) an ad hoc code is necessary.  The nfs, nfs4,
              cifs, smbfs, and ncpfs filesystems have a  separate  mount  pro‐
              gram.  In order to make it possible to treat all types in a uni‐
              form way, mount will execute the  program  /sbin/mount.type  (if
              that  exists)  when called with type type.  Since different ver‐
              sions of the smbmount program  have  different  calling  conven‐
              tions, /sbin/mount.smbfs may have to be a shell script that sets
              up the desired call.

       -U, --uuid uuid
              Mount the partition that has the specified uuid.

       -v, --verbose
              Verbose mode.

       -w, --rw, --read-write
              Mount the filesystem read/write.  This is the default.   A  syn‐
              onym is -o rw.

       -V, --version
              Display version information and exit.

       -h, --help
              Display help text and exit.

#endif

int
main(int argc, char *argv[])
{
    struct rpmmu_s _mu = {
	.av = NULL,
	.ac = 0,
	.flags = 0,

	.options = NULL,

	.source = xstrdup("sysfs.img"),		/* XXX */
	.target = xstrdup("/mnt/initrd"),	/* XXX */
	.fstype = xstrdup("sysfs"),		/* XXX */
	.mflags = 0,
	.data = NULL,

	.uflags = 0,
    };
    struct rpmmu_s *mu = &_mu;

    struct poptOption rpmmountOptionsTable[] = {
     { "debug", 'd', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmio_debug, 1,
	NULL, NULL },

     { "source", '\0', POPT_ARG_STRING,	&mu->source,	0,
	N_("file system SOURCE"), N_("SOURCE") },
     { "target", '\0', POPT_ARG_STRING,	&mu->target,	0,
	N_("file system TARGET"), N_("TARGET") },
     { "type", 't', POPT_ARG_STRING,	&mu->fstype,	0,
	N_("file system TYPE"), N_("TYPE") },
     { "options", 'o', POPT_ARG_ARGV,	&mu->options,	0,
	N_("file system OPTIONS"), N_("OPTIONS") },

     { NULL, '\0', POPT_ARG_INCLUDE_TABLE, rpmioMountFlagsTable, 0,
	N_("Mount(2) flags:"), NULL },

     { NULL, '\0', POPT_ARG_INCLUDE_TABLE, rpmioUmountFlagsTable, 0,
	N_("Umount(2) flags:"), NULL },

#ifdef	DYING
     { NULL, '\0', POPT_ARG_INCLUDE_TABLE, rpmioOpenFlagsTable, 0,
	N_("Open(2) flags:"), NULL },

     { NULL, '\0', POPT_ARG_INCLUDE_TABLE, rpmioFtsFlagsTable, 0,
	N_("Fts(3) traversal flags:"), NULL },
#endif

     { NULL, '\0', POPT_ARG_INCLUDE_TABLE, rpmioAllPoptTable, 0,
	N_("Common options for all rpmio executables:"), NULL },

      POPT_AUTOALIAS
      POPT_AUTOHELP
      POPT_TABLEEND
    };

    poptContext con = rpmioInit(argc, argv, rpmmountOptionsTable);

    int ec = 0;
    int rc;

    mu->av = (ARGV_t) poptGetArgs(con);
    mu->ac = argvCount(mu->av);

    const char *fnav[] = { "/etc/filesystems", "/proc/filesystems", NULL };
    ARGV_t fstypes = NULL;
    rc = rpmmuGetFSTypes(&fstypes, fnav);
    fstypes = argvFree(fstypes);

    mu->mflags = rpmioMountFlags;
fprintf(stderr, "*** flags: 0x%lx\n", mu->mflags);

    if (mu->options) {
	char * optStr = argvJoin(mu->options, ',');
	unsigned _optFlags = rpmioN2B(rpmioMountFlagsTable, optStr);
fprintf(stderr, "*** option flags: 0x%x\n", _optFlags);
	mu->mflags |= _optFlags;
	optStr = _free(optStr);
    }

char * mflags = rpmioB2N(rpmioMountFlagsTable, mu->mflags);
fprintf(stderr, "--> %s(%s, %s, %s, %s, %p)\n", "Mount", mu->source, mu->target, mu->fstype, mflags, mu->data);
mflags = _free(mflags);
#if defined(__linux__)
    mu->mflags |= MS_MGC_VAL;
#endif
#if defined(__APPLE__) || defined(MNT_VISFLAGMASK)
    mu->mflags &= MNT_VISFLAGMASK;
#endif

    rc = Mount(mu->source, mu->target, mu->fstype, mu->mflags, mu->data);

#ifdef	NOISY
    rc = rpmmuList("/etc/mtab", NULL);
#endif

    mu->uflags = rpmioUmountFlags;

char * uflags = rpmioB2N(rpmioUmountFlagsTable, mu->uflags);
fprintf(stderr, "--> %s(%s, %s)\n", "Umount2", mu->target, uflags);
uflags = _free(uflags);

    rc = Umount2(mu->target, mu->uflags);

    mu->source = _free(mu->source);
    mu->target = _free(mu->target);
    mu->fstype = _free(mu->fstype);
    mu->options = argvFree(mu->options);

    mu->av = argvFree(mu->av);
    mu->ac = 0;
    mu->flags = 0;

    con = rpmioFini(con);

    return ec;
}
