
/**
 * \file rpmio/tmq.c
 */

#include "system.h"

#if defined(HAVE_SYS_SYSCALL_H)
# include <sys/syscall.h>
#endif

#if defined(HAVE_LINUX_MEMFD_H)
# include <linux/memfd.h>
#endif

#if defined(HAVE_SYS_IPC_H)
# include <sys/ipc.h>
#endif

#if defined(HAVE_SYS_MSG_H)
# include <sys/msg.h>
#endif

#if defined(HAVE_SYS_SEM_H)
# include <sys/sem.h>
# if _SEM_SEMUN_UNDEFINED
#  if defined(HAVE_LINUX_SEM_H_XXX)
#   include <linux/sem.h>
#  else
    union semun
    {
      int val;
      struct semid_ds *buf;
      unsigned short *array;
      struct seminfo *__buf;
      void *__pad;
    };
#  endif
# endif
#endif

#if defined(HAVE_SYS_SHM_H)
# include <sys/shm.h>
#endif

#if defined(HAVE_SEMAPHORE_H)
# include <semaphore.h>
#endif

#if defined(HAVE_AIO_H)
# include <aio.h>
#endif

#if defined(HAVE_LIBAIO_H)
# include <libaio.h>
#endif

#if defined(HAVE_MQUEUE_H)
# include <mqueue.h>
#endif

#define _RPMIOB_INTERNAL
#include <rpmio_internal.h>

#ifdef	NOTYET
#define _RPMZQ_INTERNAL
#define _RPMZ_INTERNAL
#define _RPMZ_INTERNAL_PIGZ
#include <rpmz.h>
#endif

#include <rpmlog.h>
#include <rpmmacro.h>
#include <poptIO.h>
#include <argv.h>

#include <rpmmsq.h>

#include "debug.h"

static int _debug = -1;

static const char *qname;
static rpmmsqFlags flags =
#ifdef	DYING
	RPMMSQ_FLAGS_DEBUG |
	RPMMSQ_FLAGS_DELETE |
	RPMMSQ_FLAGS_RESET |
	RPMMSQ_FLAGS_INFO |
	RPMMSQ_FLAGS_SYSV |
	RPMMSQ_FLAGS_POSIX |
#endif
	0;

static long priority;
static int count;
static int _rpmmsqOFlags;

#define SPEW(_fmt, ...) \
    if (_debug || _rpmmsq_debug || _rpmio_debug) \
	fprintf(stderr, _fmt, ##__VA_ARGS__)

#define Z(_rc)	assert((_rc) == 0)

/*==============================================================*/
static int Shm_open(const char *path, int oflags, mode_t mode)
{
    int fdno = -2;
    const char * lpath = fdSyscallPrePath(__FUNCTION__, path, 1);
    fdno = lpath != NULL
        ? shm_open(lpath, oflags, mode)
        : -2;
    return fdSyscallPostPath(__FUNCTION__, lpath, fdno);
}

static int Shm_unlink(const char *path)
{
    int rc = -2;
    const char * lpath = fdSyscallPrePath(__FUNCTION__, path, 1);
    rc = lpath != NULL
        ? shm_unlink(lpath)
        : -2;
    return fdSyscallPostPath(__FUNCTION__, lpath, rc);
}

/*==============================================================*/
static int Shmget(key_t key, size_t size, int shmflg)
{
    char kstr[32];
    int rc = -1;

    switch (key) {
    case IPC_PRIVATE:
	(void) stpcpy(kstr, "IPC_PRIVATE");
	break;
    default:
	(void) snprintf(kstr, sizeof(kstr), "0x%x", key);
	break;
    }
    rc = shmget(key, size, shmflg);
SPEW("<-- %s(%s, %zu, 0%o) rc %d\n", __FUNCTION__, kstr, size, shmflg, rc);
    return rc;
}

static void *Shmat(int shmid, const void * shmaddr, int shmflg)
{
    void * ptr = shmat(shmid, shmaddr, shmflg);
SPEW("<-- %s(0x%x, %p, %d) ptr %p\n", __FUNCTION__, shmid, shmaddr, shmflg, ptr);
    return ptr;
}

static int Shmctl(int shmid, int cmd, struct shmid_ds *buf)
{
    char cstr[32];
    int rc = -1;

    switch (cmd) {
    case IPC_RMID:
	(void) stpcpy(cstr, "IPC_RMID");
	break;
    case IPC_SET:
	(void) stpcpy(cstr, "IPC_SET");
	break;
    case IPC_STAT:
	(void) stpcpy(cstr, "IPC_STAT");
	break;
    case IPC_INFO:
	(void) stpcpy(cstr, "IPC_INFO");
	break;
    case SHM_INFO:
	(void) stpcpy(cstr, "SHM_INFO");
	break;
    case SHM_STAT:
	(void) stpcpy(cstr, "SHM_STAT");
	break;
    case SHM_LOCK:
	(void) stpcpy(cstr, "SHM_LOCK");
	break;
    case SHM_UNLOCK:
	(void) stpcpy(cstr, "SHM_UNLOCK");
	break;
    default:
	(void) snprintf(cstr, sizeof(cstr), "0x%x", cmd);
	break;
    }
    rc = shmctl(shmid, cmd, buf);
SPEW("<-- %s(0x%x, %s, %p) rc %d\n", __FUNCTION__, shmid, cstr, buf, rc);
    return rc;
}

static int Shmdt(const void * shmaddr)
{
    int rc = shmdt(shmaddr);
SPEW("<-- %s(%p) rc %d\n", __FUNCTION__, shmaddr, rc);
    return rc;
}

/*==============================================================*/
static int Memfd_create(const char *name, unsigned int flags)
{
    int rc = -2;
#if defined(HAVE_MEMFD_CREATE)
    rc = memfd_create(name, flags);
#elif defined(SYS_memfd_create)
    rc = syscall(SYS_memfd_create, name, flags);
#else
    errno = ENOSYS;
#endif
SPEW("<-- %s(%s, 0x%x) rc %d\n", __FUNCTION__, name, flags, rc);
    return rc;
}

/*==============================================================*/
static sem_t * Sem_open(const char *path, int oflag, mode_t mode, unsigned int value)
{
    sem_t *sem = SEM_FAILED;
    const char * lpath = fdSyscallPrePath(__FUNCTION__, path, 1);
    sem = lpath != NULL
        ? sem_open(lpath, oflag, mode, value)
        : SEM_FAILED;
    (void) fdSyscallPostPath(__FUNCTION__, lpath, (sem == SEM_FAILED ? -1 : 0));
    return sem;
}

static int Sem_close(sem_t *sem)
{
    int rc = sem_close(sem);
SPEW("<-- %s(%p) rc %d\n", __FUNCTION__, sem, rc);
    return rc;
}

static int Sem_destroy(sem_t *sem)
{
    int rc = sem_destroy(sem);
SPEW("<-- %s(%p) rc %d\n", __FUNCTION__, sem, rc);
    return rc;
}

static int Sem_getvalue(sem_t *sem, int *sval)
{
    int rc = sem_getvalue(sem, sval);
SPEW("<-- %s(%p,%p) rc %d\n", __FUNCTION__, sem, sval, rc);
    return rc;
}

static int Sem_init(sem_t *sem, int pshared, unsigned int value)
{
    int rc = sem_init(sem, pshared, value);
SPEW("<-- %s(%p,%d,%d) rc %d\n", __FUNCTION__, sem, pshared, value, rc);
    return rc;
}

static int Sem_post(sem_t *sem)
{
    int rc = sem_post(sem);
SPEW("<-- %s(%p) rc %d\n", __FUNCTION__, sem, rc);
    return rc;
}

static int Sem_unlink(const char *name)
{
    int rc = sem_unlink(name);
SPEW("<-- %s(%s) rc %d\n", __FUNCTION__, name, rc);
    return rc;
}

static int Sem_wait(sem_t *sem)
{
    int rc = sem_wait(sem);
SPEW("<-- %s(%p) rc %d\n", __FUNCTION__, sem, rc);
    return rc;
}

static int Sem_trywait(sem_t *sem)
{
    int rc = sem_trywait(sem);
SPEW("<-- %s(%p) rc %d\n", __FUNCTION__, sem, rc);
    return rc;
}

static int Sem_timedwait(sem_t *sem, const struct timespec *abs_timeout)
{
    int rc = sem_timedwait(sem, abs_timeout);
SPEW("<-- %s(%p,%p) rc %d\n", __FUNCTION__, sem, abs_timeout, rc);
    return rc;
}

/*==============================================================*/
static int Semget(key_t key, int nsems, int semflg)
{
    char kstr[32];
    int rc = -1;

    switch (key) {
    case IPC_PRIVATE:
	(void) stpcpy(kstr, "IPC_PRIVATE");
	break;
    default:
	(void) snprintf(kstr, sizeof(kstr), "0x%x", key);
	break;
    }
    rc = semget(key, nsems, semflg);
SPEW("<-- %s(%s, %u, 0%o) rc %d\n", __FUNCTION__, kstr, nsems, semflg, rc);
    return rc;
}

static int Semop(int semid, struct sembuf *sops, size_t nsops)
{
    int rc = semop(semid, sops, nsops);
SPEW("<-- %s(0x%x, %p, %zu) rc %d\n", __FUNCTION__, semid, sops, nsops, rc);
    return rc;
}

static int Semctl(int semid, int semnum, int cmd, union semun un)
{
    char cstr[32];
    int rc = -1;

    switch (cmd) {
    case IPC_RMID:
	(void) stpcpy(cstr, "IPC_RMID");
	break;
    case IPC_SET:
	(void) stpcpy(cstr, "IPC_SET");
	break;
    case IPC_STAT:
	(void) stpcpy(cstr, "IPC_STAT");
	break;
    case IPC_INFO:
	(void) stpcpy(cstr, "IPC_INFO");
	break;
    case SEM_INFO:
	(void) stpcpy(cstr, "SEM_INFO");
	break;
    case SEM_STAT:
	(void) stpcpy(cstr, "SEM_STAT");
	break;
    case GETALL:
	(void) stpcpy(cstr, "GETALL");
	break;
    case GETNCNT:
	(void) stpcpy(cstr, "GETNCNT");
	break;
    case GETPID:
	(void) stpcpy(cstr, "GETPID");
	break;
    case GETVAL:
	(void) stpcpy(cstr, "GETVAL");
	break;
    case GETZCNT:
	(void) stpcpy(cstr, "GETZCNT");
	break;
    case SETALL:
	(void) stpcpy(cstr, "SETALL");
	break;
    case SETVAL:
	(void) stpcpy(cstr, "SETVAL");
	break;
    default:
	(void) snprintf(cstr, sizeof(cstr), "0x%x", cmd);
	break;
    }
    rc = semctl(semid, semnum, cmd, un);
SPEW("<-- %s(0x%x, %u, %s, %p) rc %d\n", __FUNCTION__, semid, semnum, cstr, un.buf, rc);
    return rc;
}

/*==============================================================*/

typedef enum rpmshmType_e {
    RPMSHM_TYPE_DEFAULT		= 0,
    RPMSHM_TYPE_POSIX		= 1,
    RPMSHM_TYPE_SYSV		= 2,
    RPMSHM_TYPE_MEMFD		= 3,
    RPMSHM_TYPE_POSIX_SEM	= 4,
    RPMSHM_TYPE_SYSV_SEM	= 5,
} rpmshmType;
#define RPMSHM_TYPE_MASK        0x7     /* XXX grow to 0x0F possible. */

rpmshmType shmType = RPMSHM_TYPE_SYSV_SEM;

static int doSHM(ARGV_t av, int ac)
{
    FD_t fd = NULL;
    int fdno = -1;
    int shmid = -1;
    sem_t *sem = SEM_FAILED;
    int semid = -1;
    char *b = NULL;
    size_t nb = BUFSIZ;
    int rc = -2;	/* assume failure */

    char *fn = rpmGetPath("rpm.%pid", NULL);
    mode_t _omode = 0640;

    switch (shmType) {
    default:
    case RPMSHM_TYPE_DEFAULT:
    case RPMSHM_TYPE_POSIX:
    {
	int _oflag = O_CREAT|O_RDWR;
	fdno = Shm_open(fn, _oflag, _omode);
	if (fdno < 0) {
	    perror("Shm_open");
	    goto exit;
	}
	fd = fdDup(fdno);
	if (fd == NULL) {
	    perror("fdDup");
	    goto exit;
	}

	rc = Ftruncate(fd, nb);
	if (rc < 0) {
	    perror("Ftruncate");
	    goto exit;
	}

	b = Mmap(NULL, nb, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
	if (b == NULL) {
	    perror("Mmap");
	    goto exit;
	}
    }   break;
    case RPMSHM_TYPE_SYSV:
    {
	key_t _key = IPC_PRIVATE;	/* XXX ftok(fn); */
	size_t _size = nb;
	int _shmget_flag = IPC_CREAT | _omode;
	shmid = Shmget(_key, _size, _shmget_flag);
	if (shmid == -1) {
	    perror("Shmget");
	    goto exit;
	}
	void * _shmaddr = NULL;
	int _shmat_flag = 0;	/* XXX SHM_{EXEC,RDONLY,REMAP} */
	b = Shmat(shmid, _shmaddr, _shmat_flag);
	if (b == (void *)-1) {
	    perror("Shmat");
	    goto exit;
	}

	int _cmd = IPC_STAT;
	struct shmid_ds ds;
	rc = Shmctl(shmid, _cmd, &ds);
	if (rc < 0) {
	    perror("Shmctl(IPC_STAT)");
	    goto exit;
	}
    }   break;
    case RPMSHM_TYPE_MEMFD:
    {
	int _oflag = MFD_CLOEXEC|MFD_ALLOW_SEALING;
	fdno = Memfd_create(fn, _oflag);
	if (fdno < 0) {
	    perror("Memfd_create");
	    goto exit;
	}
	fd = fdDup(fdno);
	if (fd == NULL) {
	    perror("fdDup");
	    goto exit;
	}

	rc = Ftruncate(fd, nb);
	if (rc < 0) {
	    perror("Ftruncate");
	    goto exit;
	}

	b = Mmap(NULL, nb, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
	if (b == NULL) {
	    perror("Mmap");
	    goto exit;
	}
    }   break;
    case RPMSHM_TYPE_POSIX_SEM:
    {
	int _oflag = O_CREAT;
	int _value = 0;
	sem = Sem_open(fn, _oflag, _omode, _value);
	if (sem == SEM_FAILED) {
	    perror("Sem_open");
	    goto exit;
	}

	int _pshared = 0;
	rc = Sem_init(sem, _pshared, _value);
	if (rc < 0) {
	    perror("Sem_init");
	    goto exit;
	}
	int _getvalue;
	rc = Sem_getvalue(sem, &_getvalue);
	if (rc < 0 || _getvalue != _value) {
	    perror("Sem_getvalue(init)");
	    goto exit;
	}

	rc = Sem_post(sem);
	if (rc < 0) {
	    perror("Sem_post");
	    goto exit;
	}
	_value++;
	rc = Sem_getvalue(sem, &_getvalue);
	if (rc < 0 || _getvalue != _value) {
	    perror("Sem_getvalue(post)");
	    goto exit;
	}

	rc = Sem_wait(sem);
	if (rc < 0) {
	    perror("Sem_wait");
	    goto exit;
	}
	_value--;
	rc = Sem_getvalue(sem, &_getvalue);
	if (rc < 0 || _getvalue != _value) {
	    perror("Sem_getvalue(wait)");
	    goto exit;
	}
	sleep(60);
	goto exit;	/* XXX skip snprintf */
    }   break;
    case RPMSHM_TYPE_SYSV_SEM:
    {
	key_t _key = IPC_PRIVATE;	/* XXX ftok(fn); */
	int _nsems = 2;
	int _semget_flag = IPC_CREAT | _omode;
	semid = Semget(_key, _nsems, _semget_flag);
	if (semid == -1) {
	    perror("Semget");
	    goto exit;
	}

	struct sembuf sops[2] = {};
	sops[0].sem_num = 0;
	sops[0].sem_op = 0;
	sops[0].sem_flg = SEM_UNDO;
	sops[1].sem_num = 1;
	sops[1].sem_op = 0;
	sops[1].sem_flg = SEM_UNDO;

	rc = Semop(semid, sops, sizeof(sops)/sizeof(sops[0]));
	if (rc < 0) {
	    perror("Semop");
	    goto exit;
	}

	int _cmd = IPC_STAT;
	struct semid_ds ds;
	union semun un;
	un.buf = &ds;
	rc = Semctl(semid, 0, _cmd, un);
	if (rc < 0) {
	    perror("Semctl(IPC_STAT)");
	    goto exit;
	}
	goto exit;	/* XXX skip snprintf */
    }   break;
    }

    (void) snprintf(b, nb, "%s\n", fn);
    sleep(1);

exit:
    switch (shmType) {
    default:
    case RPMSHM_TYPE_DEFAULT:
    case RPMSHM_TYPE_POSIX:
    {
	if (b && (rc = Munmap(b, nb)) < 0)
	    perror("Munmap");

	if (fd && (rc = Fclose(fd)) < 0)
	    perror("Fclose");

	if (fn && (rc = Shm_unlink(fn)) < 0)
	    perror("Shm_unlink");
    }   break;
    case RPMSHM_TYPE_SYSV:
    {
	if (shmid == -1)
	    break;

	if (b && (rc = Shmdt(b)) < 0)
	    perror("Shmdt");

	int _cmd = IPC_RMID;
	rc = Shmctl(shmid, _cmd, NULL);
	if (rc < 0) {
	    perror("Shmctl(IPC_RMID)");
	    goto exit;
	}
    }   break;
    case RPMSHM_TYPE_MEMFD:
    {
	if (b && (rc = Munmap(b, nb)) < 0)
	    perror("Munmap");

	if (fd && (rc = Fclose(fd)) < 0)
	    perror("Fclose");

	if (fdno >= 0 && (rc = close(fdno)) < 0)
	    perror("close");

    }   break;
    case RPMSHM_TYPE_POSIX_SEM:
    {
	if (sem == SEM_FAILED)
	    break;

	if ((rc = Sem_close(sem)) < 0)
	    perror("Sem_close");

	if ((rc = Sem_destroy(sem)) < 0)
	    perror("Sem_destroy");

	if ((rc = Sem_unlink(fn)) < 0)
	    perror("Sem_unlink");

    }   break;
    case RPMSHM_TYPE_SYSV_SEM:
    {
	if (semid == -1)
	    break;

	int _cmd = IPC_RMID;
	union semun un;
	un.buf = NULL;
	rc = Semctl(semid, 0, _cmd, un);
	if (rc < 0) {
	    perror("Semctl(IPC_RMID)");
	    goto exit;
	}
    }   break;
    }

    fn = _free(fn);
    return rc;
}

/*==============================================================*/

#define	SECS	1
#define	TICKS	5
typedef struct AT_s * AT_t;
struct AT_s {
    int secs;
    int tick;
    int nticks;
    clockid_t clockid;
    timer_t timerid;
    int flags;
    struct sigevent sigev;
    struct itimerspec nit;
    struct itimerspec oit;
    pthread_cond_t cond;
    pthread_mutex_t mutex;
};

#define	Z(_rc)	assert((_rc) == 0)
#define	LOCK(_mutex)		Z(pthread_mutex_lock(&_mutex))
#define	UNLOCK(_mutex)		Z(pthread_mutex_unlock(&_mutex))
#define	SIGNAL(_cond)		Z(pthread_cond_signal(&_cond))
#define	WAIT(_cond, _mutex)	Z(pthread_cond_wait(&_cond, &_mutex))

static void
timer_thread (union sigval sv)
{
    AT_t at = (AT_t) sv.sival_ptr;
    LOCK(at->mutex);
    if (++at->tick >= at->nticks)
	SIGNAL(at->cond);
    UNLOCK(at->mutex);

    fprintf(stderr,"\ttick %d\n", at->tick);
}

static int
doTIMER(ARGV_t av, int ac)
{
    struct AT_s _at = {
	.secs = SECS,
	.tick = 0,
	.nticks = TICKS,
	.clockid = CLOCK_REALTIME,
	.flags = 0,
	.sigev = {
	    .sigev_notify = SIGEV_THREAD,
	    .sigev_notify_function = timer_thread,
	    .sigev_notify_attributes = NULL,
	    .sigev_value.sival_ptr = &_at,
	},
	.nit = {
	    .it_value.tv_sec = _at.secs,
	    .it_value.tv_nsec = 0,
	    .it_interval.tv_sec = _at.secs,
	    .it_interval.tv_nsec = 0,
	},
	.cond = PTHREAD_COND_INITIALIZER,
	.mutex = PTHREAD_MUTEX_INITIALIZER,
    };
    AT_t at = &_at;

    Z(timer_create(at->clockid, &at->sigev, &at->timerid));

    Z(timer_settime(at->timerid, at->flags, &at->nit, &at->oit));

    LOCK(at->mutex);
    while (at->tick < at->nticks)
	WAIT(at->cond, at->mutex);
    UNLOCK(at->mutex);

    return 0;
}

/*==============================================================*/
static int doMSQ(ARGV_t av, int ac)
{
    int rc = -2;	/* assume failure */
    char wb[BUFSIZ];
    size_t nwb = sizeof(wb);
    size_t nw;
    char rb[BUFSIZ];
    size_t nrb = sizeof(rb);
    size_t nr;
    size_t blen;
    int xx;

    (void)nrb;

#define F_ISSET(_f, _FLAG) ((_f) & RPMMSQ_FLAGS_##_FLAG)
fprintf(stderr, "*** %s: flags 0x%x\n", __FUNCTION__, flags);
    char fmode[64];
    char * te = fmode;
    te = stpcpy(te, "w+");
    switch (flags & RPMMSQ_TYPE_MASK) {
    default:
    case RPMMSQ_TYPE_DEFAULT:			break;
    case RPMMSQ_TYPE_POSIX:	*te++ = 'P';	break;
    case RPMMSQ_TYPE_SYSV:	*te++ = 'S';    break;
    }
    if (F_ISSET(flags, RESET))	*te++ = 'R';
    if (F_ISSET(flags, INFO))	*te++ = 'I';
    if (F_ISSET(flags, DELETE))	*te++ = 'D';
    if (_rpmmsqOFlags & O_CLOEXEC)	*te++ = 'e';
    if (_rpmmsqOFlags & O_NONBLOCK)	*te++ = 'n';
    if (_rpmmsqOFlags & O_TRUNC)	*te++ = 't';
    if (_rpmmsqOFlags & O_EXCL)	*te++ = 'x';
#ifdef	NOTYET
    if (_rpmmsqOFlags & O_DIRECT)	*te++ = 'D';
#endif
    if (_rpmmsqOFlags & O_SYNC)	*te++ = 'S';
    if (_rpmmsqOFlags & O_TMPFILE)	*te++ = 'T';
    if (_debug)			*te++ = '?';
    te = stpcpy(te, ".msqio");
    *te = '\0';
#undef	F_ISSET

#ifdef	DISABLE
    rpmmsq msq = rpmmsqNew(qname, fmode, -1, flags);

    memset(wb, 0, nwb);
    if (msq) {
	const char msg1[] = "yadda yadda";
	size_t nmsg1 = sizeof(msg1) - 1;
	const char msg2[] = "bing bang boom";
	size_t nmsg2 = sizeof(msg2) - 1;

	strcpy(wb, msg1);
	blen = strlen(wb);
	nw = rpmmsqSend(msq, wb, blen, priority);
	assert(nw == blen);

	strcpy(wb, msg2);
	blen = strlen(wb);
	nw = rpmmsqSend(msq, wb, blen, priority);
	assert(nw == blen);

	memset(rb, 0, nrb);
	nr = rpmmsqRecv(msq, rb, nrb, NULL);
	assert(nr == nmsg1);
	memset(rb, 0, nrb);
	nr = rpmmsqRecv(msq, rb, nrb, NULL);
	assert(nr == nmsg2);

	for (int i = 0; i < count; i++) {
	    blen = snprintf(wb, nwb, "%d", i);
	    nw = rpmmsqSend(msq, wb, blen, priority);
	    memset(rb, 0, blen);
	    nr = rpmmsqRecv(msq, rb, nrb, NULL);
	}
    }

    xx = rpmmsqClose(msq, 0);
    msq = rpmmsqFree(msq);
#endif	/* DISABLE */

    FD_t fd = Fopen(qname, fmode);
    if (fd) {

#ifdef	DISABLE
	xx = Ftell(fd);
#endif	/* DISABLE */

	strcpy(wb, "foo bar baz");
	blen = strlen(wb);
	nw = Fwrite(wb, 1, blen, fd);

#ifdef	DISABLE
	xx = Ferror(fd);
	xx = Fflush(fd);
	xx = Ftell(fd);
#endif	/* DISABLE */

	memset(rb, 0, blen);
	nr = Fread(rb, 1, blen, fd);
	assert(nr == nw && !strncmp(rb, wb, blen));

#ifdef	DISABLE
	xx = Ftell(fd);

	xx = Fileno(fd);
if (_rpmio_debug)
fprintf(stderr, "<== Fileno(%p) rc %d\n", fd, xx);
#endif	/* DISABLE */

	strcpy(wb, "blah blah blah");
	blen = strlen(wb);
	nw = Fwrite(wb, 1, blen, fd);
	memset(rb, 0, blen);
	nr = Fread(rb, 1, blen, fd);
	assert(nr == nw && !strncmp(rb, wb, blen));

	for (int i = 0; i < count; i++) {
	    blen = snprintf(wb, nwb, "%d", i);
	    nw = Fwrite(wb, 1, blen, fd);
	    memset(rb, 0, blen);
	    nr = Fread(rb, 1, blen, fd);
	    assert(nr == nw && !strncmp(rb, wb, blen));
	}

	for (int i = 0; i < 8; i++) {
	    blen = snprintf(wb, nwb, "%d", i);
	    nw = Fwrite(wb, 1, blen, fd);
	}

	for (int i = 0; i < 8; i++) {
	    memset(rb, 0, blen);
	    nr = Fread(rb, 1, blen, fd);
	}

#ifdef	DISABLE
#ifdef	NOTYET
	xx = Feof(fd);
	xx = Clearerr(fd);
#endif
	Rewind(fd);
	fpos_t pos = {};
	xx = Fgetpos(fd, &pos);
	xx = Fsetpos(fd, &pos);

	FD_t nfd = Fdopen(fd, fmode);
	if (nfd) {
	    strcpy(wb, "foo bar baz");
	    blen = strlen(wb);
	    nw = Fwrite(wb, 1, blen, nfd);
	    memset(rb, 0, blen);
	    nr = Fread(rb, 1, blen, nfd);
	    xx = Fclose(nfd);
	}
#endif	/* DISABLE */

	xx = Fclose(fd);
    } else
	perror("Fopen");

    rc = 0;

    return rc;
}
/*==============================================================*/
static
int IO_queue_init(int maxevents, io_context_t *ctxp)
{
    int rc = io_queue_init(maxevents, ctxp);
SPEW("<-- %s(%d,%p) rc %d\n", __FUNCTION__, maxevents, ctxp, rc);
    return rc;
}

static
int IO_queue_release(io_context_t ctx)
{
    int rc = io_queue_release(ctx);
SPEW("<-- %s(%p) rc %d\n", __FUNCTION__, ctx, rc);
    return rc;
}

static
int IO_queue_run(io_context_t ctx)
{
    int rc = io_queue_run(ctx);
SPEW("<-- %s(%p) rc %d\n", __FUNCTION__, ctx, rc);
    return rc;
}

static
int IO_setup(int maxevents, io_context_t *ctxp)
{
    int rc = io_setup(maxevents, ctxp);
SPEW("<-- %s(%d, %p) rc %d\n", __FUNCTION__, maxevents, ctxp, rc);
    return rc;
}

static
int IO_destroy(io_context_t ctx)
{
    int rc = io_destroy(ctx);
SPEW("<-- %s(%p) rc %d\n", __FUNCTION__, ctx, rc);
    return rc;
}

static
int IO_submit(io_context_t ctx, long nr, struct iocb *iocbs[])
{
    int rc = io_submit(ctx, nr, iocbs);
SPEW("<-- %s(%p, %ld, %p) rc %d\n", __FUNCTION__, ctx, nr, iocbs, rc);
    return rc;
}

static
int IO_cancel(io_context_t ctx, struct iocb *iocb, struct io_event *evt)
{
    int rc = io_cancel(ctx, iocb, evt);
SPEW("<-- %s(%p, %p, %p) rc %d\n", __FUNCTION__, ctx, iocb, evt, rc);
    return rc;
}

static
int IO_getevents(io_context_t ctx, long min_nr, long nr,
		struct io_event *events, struct timespec *timeout)
{
    int rc = io_getevents(ctx, min_nr, nr, events, timeout);
SPEW("<-- %s(%p, %ld, %ld, %p, %p) rc %d\n", __FUNCTION__, ctx, min_nr, nr, events, timeout, rc);
    return rc;
}

static
void IO_set_callback(struct iocb *iocb, io_callback_t cb)
{
SPEW("--> %s(%p, %p)\n", __FUNCTION__, iocb, cb);
    io_set_callback(iocb, cb);
}

static
void IO_prep_pread(struct iocb *iocb,
		int fd, void *buf, size_t count, long long offset)
{
SPEW("--> %s(%p, %d, %p[%zu], %lld)\n", __FUNCTION__, iocb, fd, buf, count, offset);
    io_prep_pread(iocb, fd, buf, count, offset);
}

static
void IO_prep_pwrite(struct iocb *iocb,
		int fd, void *buf, size_t count, long long offset)
{
SPEW("--> %s(%p, %d, %p[%zu], %lld)\n", __FUNCTION__, iocb, fd, buf, count, offset);
    io_prep_pwrite(iocb, fd, buf, count, offset);
}

static
void IO_prep_preadv(struct iocb *iocb,
		int fd, const struct iovec *iov, int iovcnt, long long offset)
{
SPEW("--> %s(%p, %d, %p[%u], %lld)\n", __FUNCTION__, iocb, fd, iov, iovcnt, offset);
}

static
void IO_prep_pwritev(struct iocb *iocb,
		int fd, const struct iovec *iov, int iovcnt, long long offset)
{
SPEW("--> %s(%p, %d, %p[%u], %lld)\n", __FUNCTION__, iocb, fd, iov, iovcnt, offset);
    io_prep_pwritev(iocb, fd, iov, iovcnt, offset);
}

static
void IO_prep_poll(struct iocb *iocb, int fd, int events)
{
SPEW("--> %s(%p, %d, %d)\n", __FUNCTION__, iocb, fd, events);
    io_prep_poll(iocb, fd, events);
}

static
int IO_poll(io_context_t ctx, struct iocb *iocb, io_callback_t cb,
		int fd, int events)
{
    int rc = io_poll(ctx, iocb, cb, fd, events);
SPEW("<-- %s(%p, %p, %p, %d, %d) rc %d\n", __FUNCTION__, ctx, iocb, cb, fd, events, rc);
    return rc;
}

static
void IO_prep_fsync(struct iocb *iocb, int fd)
{
SPEW("--> %s(%p, %d)\n", __FUNCTION__, iocb, fd);
    io_prep_fsync(iocb, fd);
}

static
int IO_fsync(io_context_t ctx, struct iocb *iocb, io_callback_t cb, int fd)
{
    int rc = io_fsync(ctx, iocb, cb, fd);
SPEW("<-- %s(%p, %p, %p, %d) rc %d\n", __FUNCTION__, ctx, iocb, cb, fd, rc);
    return rc;
}

static
void IO_prep_fdsync(struct iocb *iocb, int fd)
{
SPEW("--> %s(%p, %d)\n", __FUNCTION__, iocb, fd);
    io_prep_fdsync(iocb, fd);
}

static
int IO_fdsync(io_context_t ctx, struct iocb *iocb, io_callback_t cb, int fd)
{
    int rc = io_fdsync(ctx, iocb, cb, fd);
SPEW("<-- %s(%p, %p, %p, %d) rc %d\n", __FUNCTION__, ctx, iocb, cb, fd, rc);
    return rc;
}

static
void IO_set_eventfd(struct iocb *iocb, int eventfd)
{
SPEW("--> %s(%p, %d)\n", __FUNCTION__, iocb, eventfd);
    io_set_eventfd(iocb, eventfd);
}

/*==============================================================*/
static
void LAIO_done(io_context_t ctx, struct iocb *iocb, long res, long res2)
{
SPEW("--> %s(%p, %p, %ld, %ld)\n", __FUNCTION__, ctx, iocb, res, res2);
}

static int doLAIO(ARGV_t av, int ac)
{
    int rc = -2;	/* assume failure */
    io_context_t ctx = NULL;
    struct iocb cb = {};
    FD_t fd = Fopen("foo", "rbD");
assert(fd);
    int fdno = Fileno(fd);
    size_t nb = BUFSIZ;
    char *b = NULL;

    rc = posix_memalign((void **) &b, 512, BUFSIZ);
    if (rc < 0) {
	perror("posix_memalign");
	goto exit;
    }

    unsigned _nr_events = 10;
    rc = IO_setup(_nr_events, &ctx);
    if (rc < 0) {
	perror("IO_setup");
	goto exit;
    }

    long long _off = 0;
    io_prep_pread(&cb, fdno, b, nb, _off);
    io_set_callback(&cb, LAIO_done);

    struct iocb * iocbs[] = { &cb };
    size_t niocbs = (sizeof(iocbs)/sizeof(iocbs[0]));
    rc = IO_submit(ctx, niocbs, iocbs);
    if (rc != (int)niocbs) {
	perror("IO_submit");
	goto exit;
    }

    long _min_nr = 0;
    long _nr = 1;
    struct io_event _events[1] = {};
    int nwaits = 0;
    do {
	nwaits++;
	struct timespec ts = { 0, 100*1000*1000 };
	nanosleep(&ts, NULL);
	sched_yield();

	struct timespec * _timeout = NULL;
	rc = IO_getevents(ctx, _min_nr, _nr, _events, _timeout);
	if (rc < 0) {
	    perror("IO_getevents");
	    goto exit;
	}
	for (int i = 0; i < rc; i++) {
	    struct io_event * ev = &_events[i];
	    io_callback_t cb = (io_callback_t)ev->data;
	    if (cb == NULL)
		continue;
	    struct iocb *iocb = ev->obj;
	    cb(ctx, iocb, ev->res, ev->res2);
	}
	    
    } while (rc == 0);

    struct io_event * ev = &_events[0];
    int nread = ev->res;
fprintf(stderr, "*** nwaits %d data %p obj %p res %lu res2 %lu\n",  nwaits, ev->data, ev->obj, ev->res, ev->res2);
fprintf(stderr, "%.*s", nread, b); 

    rc = 0;

exit:
    if (ctx) {
	rc = IO_destroy(ctx);
	if (rc < 0)
	    perror("IO_destroy");
    }
    if (fd) {
	rc = Fclose(fd);
	if (rc < 0)
	    perror("Fclose");
    }
    if (b)
	free(b);

    return rc;
}

/*==============================================================*/
#if !defined(POPT_BIT_XOR)
#define	POPT_BIT_XOR	(POPT_ARG_VAL|POPT_ARGFLAG_XOR)
#endif

#define	_MSQBIT(_v)	(RPMMSQ_FLAGS_##_v)
static struct poptOption rpmmqOptionsTable[] = {
 { "debug", 'd', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN,	&_debug, 1,
	NULL, NULL },

 { "qname", 'Q', POPT_ARG_STRING,		    &qname, 0,
	N_("Use message queue QNAME"), N_("QNAME") },
 { "posix", 'P', POPT_BIT_SET,                      &flags, _MSQBIT(POSIX),
	N_("Use POSIX message queues"), NULL },
 { "sysv", 'S', POPT_BIT_SET,                       &flags, _MSQBIT(SYSV),
	N_("Use SysV message queues"), NULL },

 { "reset", 'R', POPT_BIT_SET|POPT_ARGFLAG_TOGGLE,  &flags, _MSQBIT(RESET),
	N_("Remove queue before opening"), NULL },
 { "info", 'I', POPT_BIT_SET|POPT_ARGFLAG_TOGGLE,   &flags, _MSQBIT(INFO),
	N_("Display queue info on close"), NULL },
 { "delete", 'D', POPT_BIT_SET|POPT_ARGFLAG_TOGGLE, &flags, _MSQBIT(DELETE),
	N_("Remove queue after closing"), NULL },

 { "count", 'c', POPT_ARG_INT,			    &count, 0,
	N_("Send COUNT messages"), N_("COUNT") },
 { "priority", 'p', POPT_ARG_LONG,		    &priority, 0,
	N_("Send message PRIORITY"), N_("PRIORITY") },

 { NULL, '\0', POPT_ARG_INCLUDE_TABLE, rpmioOpenFlagsTable, 0,
	N_("Open(2) flags:"), NULL },

 { NULL, '\0', POPT_ARG_INCLUDE_TABLE, rpmioAllPoptTable, 0,
	N_("Common options for all rpmio executables:"), NULL },

  POPT_AUTOALIAS
  POPT_AUTOHELP
  POPT_TABLEEND
};
#undef	_MSQBIT

int
main(int argc, char *argv[])
{
    poptContext con = rpmioInit(argc, argv, rpmmqOptionsTable);
    ARGV_t av = poptGetArgs(con);
    int ac = argvCount(av);;
    int ec = 0;

    _rpmmsqOFlags = rpmioOpenFlags;
    if (qname == NULL)
	qname = xstrdup("rpm");

    switch (4) {
    case 1:	ec = doSHM(av, ac);	break;
    case 2:	ec = doMSQ(av, ac);	break;
    case 3:	ec = doTIMER(av, ac);	break;
    case 4:	ec = doLAIO(av, ac);	break;
    }

    qname = _free(qname);
    con = rpmioFini(con);

    return ec;
}
