/**
 * Copyright 2016-present, Yann Collet, Facebook, Inc.
 * All rights reserved.
 *
 * This source code is licensed under the license found in the
 * LICENSE-examples file in the root directory of this source tree.
 */

#include "system.h"

#include "rpmio_internal.h"
#include <rpmlog.h>
#include <rpmmacro.h>
#include <poptIO.h>
#include <argv.h>
#include <yarn.h>
#define PEEK(_bolt)	yarnPeekLock(_bolt)

#if defined(WITH_ZSTD)
#include <zstd.h>		// presumes zstd library is installed
#endif

#define	_RPMZSTD_INTERNAL
#include "rpmzstd.h"

#include "debug.h"

#define	ZSTDF_ISSET(_foo)		(zstd->flags & RPMZSTD_FLAGS_##_foo)

#define SPEW(_fmt, ...) \
    if ((zstd && ZSTDF_ISSET(DEBUGIO)) || _rpmzstd_debug || _rpmio_debug) \
	fprintf(stderr, _fmt, __VA_ARGS__)

extern int _rpmzstdOFlags;

/*==============================================================*/
static ssize_t rpmzstdCopyAll(rpmzstd zstd)
{
    ssize_t rc = -2;
    char b[16*BUFSIZ];
    size_t nb = sizeof(b);
    size_t nr;
    size_t nw;

SPEW("==> %s(%p) oflags 0x%x\n", __FUNCTION__, zstd, zstd->oflags);

assert(zstd->_stream);
    if (ZSTDF_ISSET(DECOMPRESS)) {	/* decompress */
	FILE * fp = stdout;

SPEW("\tDecompressing ... eof %d\n", feof(zstd->fp));
	do {
	    nr = rpmzstdRead(zstd, b, nb);
	    if (nr <= 0) {
		rc = nr;
		break;
	    }
	    nw = fwrite(b, 1, nr, fp);
assert(nw == nr);
SPEW("<--\tfwrite(%p, 1, %zu, %p) nw %zd\n", b, nr, fp, nw);
	    if (nw <= 0) {
		rc = nw;
		break;
	    }
	} while (zstd->zib.pos < zstd->zib.size);
    } else {				/* compress */
SPEW("\tCompressing ... eof %d\n", feof(zstd->fp));
	rc = 0;
	while (!feof(zstd->fp)) {
	    FILE * fp = stdin;
	    nr = fread(b, 1, nb, fp);
SPEW("<--\tfread(%p, 1, %zu, %p) nr %zd eof %d\n", b, nb, fp, nr, feof(fp));
	    if (nr <= 0) {
		rc = nr;
		break;
	    }
	    nw = rpmzstdWrite(zstd, b, nr);
	    if (nw <= 0) {
		rc = nw;
		break;
	    }
	}
	if (!rc)
	    rc = rpmzstdClose(zstd);
    }

rpmzstdDump(__FUNCTION__, zstd, NULL);
SPEW("<== %s(%p) rc %ld\n", __FUNCTION__, zstd, rc);
    return rc;
}

static int compressFile(const char *ifn, const char *fmode)
{
    rpmzstd zstd = rpmzstdOpen(ifn, fmode, -1, RPMZSTD_FLAGS_NONE);
    int rc = rpmzstdCopyAll(zstd);
    rc = rpmzstdClose(zstd);		/* XXX elsewhere */
    zstd = rpmzstdFree(zstd);
    return rc;
}

static int decompressFile(const char *ifn, const char *fmode)
{
    rpmzstd zstd = rpmzstdOpen(ifn, fmode, -1, RPMZSTD_FLAGS_DECOMPRESS);
    int rc = rpmzstdCopyAll(zstd);
    zstd = rpmzstdFree(zstd);
    return rc;
}

/*==============================================================*/
static int copyZFile(FD_t ifd, FD_t ofd)
{
    char b[BUFSIZ];
    size_t nb = sizeof(b);
    size_t pos = 0;
    int rc = -2;

    if (ifd && ofd)
    while (1) {
	size_t nr = Fread(b, 1, nb, ifd);
fprintf(stderr, "================= nr %zu\n", nr);
	if (Ferror(ifd) || nr == 0) {
	    rc = 0;
	    break;
	}
	pos += nr;
	size_t nw = Fwrite(b, 1, nr, ofd);
fprintf(stderr, "================= nw %zu\n", nw);
	if (nr != nw) {
	    rc = -2;
	    break;
	}
	(void) Fflush(ofd);
    }

    return rc;
}

static int readZFile(const char *ifn, const char *ofn)
{
    FD_t ifd = Fopen(ifn, "r?.zstdio");
    FD_t ofd = Fopen(ofn, "wb");
    int rc = copyZFile(ifd, ofd);

    if (ofd)
	(void) Fclose(ofd);
    if (ifd)
	(void) Fclose(ifd);
    return rc;
}

static int writeZFile(const char *ifn, const char *ofn)
{
    FD_t ifd = Fopen(ifn, "rb");
    FD_t ofd = Fopen(ofn, "w3b?.zstdio");
    int rc = copyZFile(ifd, ofd);

    if (ofd)
	(void) Fclose(ofd);
    if (ifd)
	(void) Fclose(ifd);
    return rc;
}

/*==============================================================*/
static int climode = 'R';

#if !defined(POPT_BIT_XOR)
#define	POPT_BIT_XOR	(POPT_ARG_VAL|POPT_ARGFLAG_XOR)
#endif

static struct poptOption rpmzstdOptionsTable[] = {
 { "debug", 'd', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN,	&_rpmzstd_debug, 1,
	NULL, NULL },
 { "read", 'R', POPT_ARG_VAL,		&climode, 'R',
	NULL, NULL },
 { "write", 'W', POPT_ARG_VAL,		&climode, 'W',
	NULL, NULL },

 { NULL, '\0', POPT_ARG_INCLUDE_TABLE, rpmioOpenFlagsTable, 0,
	N_("Open(2) flags:"), NULL },

 { NULL, '\0', POPT_ARG_INCLUDE_TABLE, rpmioAllPoptTable, 0,
	N_("Common options for all rpmio executables:"), NULL },

  POPT_AUTOALIAS
  POPT_AUTOHELP
  POPT_TABLEEND
};

int
main(int argc, char *argv[])
{
    poptContext con = rpmioInit(argc, argv, rpmzstdOptionsTable);
    ARGV_t av = poptGetArgs(con);
    int ac = argvCount(av);
    int ec = 0;

    if (ac != 1) {
	printf("wrong arguments\n");
	printf("usage:\n");
	printf("%s FILE\n", argv[0]);
	return 1;
    }
    const char * fn = rpmGetPath(av[0], NULL);
    const char * zfn = rpmGetPath(av[0], ".zst", NULL);

    switch (climode) {
    case 1:	ec = compressFile(zfn, "w3?.zstdio");	break;
    case 2:	ec = decompressFile(zfn, "r?.zstdio");	break;
    case 'R':	ec = readZFile(zfn, "-");	break;
    case 'W':	ec = writeZFile("-", zfn);	break;
    }

    fn = _free(fn);
    zfn = _free(zfn);

    con = rpmioFini(con);

    return ec;
}
