/** \ingroup rpmio
 * \file rpmio/zstdio.c
 * Support for ZSTD compression library.
 */

#include "system.h"

#include <stddef.h>        /* size_t */
#include <stdint.h>

#if	defined(HAVE_STDBOOL_H)
# include <stdbool.h>
#else
typedef	enum { true = 1, false = 0 } bool;
#endif

#if defined(WITH_ZSTD)
#include <zstd.h>
#endif

#include "rpmio_internal.h"
#include <rpmmacro.h>
#include <rpmcb.h>
#include <yarn.h>
#define PEEK(_bolt)	yarnPeekLock(_bolt)

#if defined(WITH_ZSTD)
#include <zstd.h>
#endif

#define	_RPMZSTD_INTERNAL
#include "rpmzstd.h"

#include "debug.h"

int _rpmzstd_debug;
#define SPEW(_fmt, ...) \
    if ((zstd && ZSTDF_ISSET(DEBUGIO)) || _rpmzstd_debug || _rpmio_debug) \
	fprintf(stderr, _fmt, __VA_ARGS__)

int _rpmzstdOFlags;

#define	ZSTDF_ISSET(_foo)		(zstd->flags & RPMZSTD_FLAGS_##_foo)

#define	ZSTDONLY(fd)	assert(fdGetIo(fd) == zstdio)

/* =============================================================== */
int rpmzstdDump(const char * msg, void * _zstd, FILE *fp)
{
    int rc = 0;

    if (fp == NULL)	fp = stderr;
    if (msg)	fprintf(fp, "========================== %s\n", msg);
#if defined(WITH_ZSTD)
    rpmzstd zstd = (rpmzstd) _zstd;
    if (zstd) {

#define PRINT(_fmt, _foo) \
    {   fprintf(fp, "%25s: %"#_fmt"\n", #_foo, zstd->_foo); }
	PRINT(s,  path);
	PRINT(s,  fmode);
	PRINT(#o, flags);
	PRINT(d,  fdno);
	PRINT(#o, oflags);
	PRINT(#o, omode);

	PRINT(d,  level);
	PRINT(zu, nextnb);

	PRINT(p,  fp);
	PRINT(p,  _stream);
	PRINT(p,  b);

	if (zstd->zob.dst) {
	    PRINT(p,  zob.dst);
	    PRINT(zu, nb);
	    PRINT(zu, nob);
	    PRINT(zu, zob.size);
	    PRINT(zu, zob.pos);
	}

	if (zstd->zib.src) {
	    PRINT(p,  zib.src);
	    PRINT(zu, nib);
	    PRINT(zu, zib.size);
	    PRINT(zu, zib.pos);
	}
#undef	PRINT

    }
#endif	/* WITH_ZSTD */
    return rc;
}

/* =============================================================== */
static
const char * rpmzstdError(rpmzstd zstd, ssize_t rc)
{
#if defined(WITH_ZSTD)
    return ZSTD_getErrorName(rc);
#else
    return "ZSTD not configured.";
#endif	/* WITH_ZSTD */
}

#define ANSI_BRIGHT_BLUE	"\x1b[34;1m"
#define ANSI_BOLD_ON		"\x1b[1m"
#define ANSI_INVERSE_ON		"\x1b[7m"
#define ANSI_RESET		"\x1b[0m"

#define ANSI_INVERSE_BOLD		"\x1b[7;1m"

static ssize_t rpmzstdCheck(rpmzstd zstd, const char * msg, ssize_t rc)
{
#if defined(WITH_ZSTD)
SPEW("<--\t    %s: rc %6zd\t%s%s%s\n", msg, rc, ANSI_INVERSE_BOLD, ZSTD_getErrorName(rc), ANSI_RESET);
    if (ZSTD_isError(rc))
	fprintf(stderr, "%s error: %s\n", msg, rpmzstdError(zstd, rc));
#endif	/* WITH_ZSTD */
    return rc;
}

static char * rpmzstdDbug(void *_zstd, char *b, size_t nb)
{
    rpmzstd zstd = (rpmzstd) _zstd;
    size_t len = strlen(b);
    char * be = b + len;
    rpmioItem item = (rpmioItem) zstd;
    long use;

    if (zstd && (use = PEEK(item->use)) > 0) {
	int colorize = isatty(fileno(stderr));
	*be++ = '\n';
	if (colorize) be = stpcpy(be, ANSI_BRIGHT_BLUE);
	be += sprintf(be, "========================== zstd(%p) use %ld\n",
		zstd, use);

#if defined(WITH_ZSTD)
#define PRINT(_fmt, _foo) \
    {	be += sprintf(be, "%25s: %"#_fmt"\n", #_foo, zstd->_foo); }
	PRINT(s,  path);
	PRINT(s,  fmode);
	PRINT(#o, flags);
	PRINT(d,  fdno);
	PRINT(#o, oflags);
	PRINT(#o, omode);

	PRINT(d,  level);
	PRINT(zu, nextnb);

	PRINT(p,  fp);
	PRINT(p,  _stream);
	PRINT(p,  b);

	if (zstd->zob.dst) {
	    PRINT(p,  zob.dst);
	    PRINT(zu, nb);
	    PRINT(zu, nob);
	    PRINT(zu, zob.size);
	    PRINT(zu, zob.pos);
	}

	if (zstd->zib.src) {
	    PRINT(p,  zib.src);
	    PRINT(zu, nb);
	    PRINT(zu, nib);
	    PRINT(zu, zib.size);
	    PRINT(zu, zib.pos);
	}

#undef	PRINT
#endif	/* WITH_ZSTD */

	be--;
	if (colorize) be = stpcpy(be, ANSI_RESET);
	*be = '\0';
    }
    return b;
}

static void rpmzstdInit(void *_zstd)
{
}


static void rpmzstdFini(void *_zstd)
{
    rpmzstd zstd = (rpmzstd) _zstd;
    if (zstd) {
#if defined(WITH_ZSTD)
	size_t rc;
	if (zstd->_stream) {
	    if (ZSTDF_ISSET(DECOMPRESS))
		rc = rpmzstdCheck(zstd, "ZSTD_freeDStream",
			ZSTD_freeDStream(zstd->_stream));
	    else
		rc = rpmzstdCheck(zstd, "ZSTD_freeCStream",
			ZSTD_freeCStream(zstd->_stream));
	}
	zstd->_stream = NULL;

	if (zstd->fp && fileno(zstd->fp) > 2)
	    fclose(zstd->fp);
	zstd->fp = NULL;

	if (zstd->b)
	    zstd->b = _free(zstd->b);
	if (zstd->path)
	    zstd->path = _free(zstd->path);
	if (zstd->fmode)
	    zstd->fmode = _free(zstd->fmode);
	zstd->flags = 0;
	zstd->fdno = -1;
	zstd->oflags = 0;
	zstd->omode = 0;

	zstd->nextnb = 0;
	zstd->level = 0;

	zstd->nb = 0;
	zstd->nib = 0;
	zstd->nob = 0;

	zstd->zib.src = NULL;
	zstd->zib.pos = 0;
	zstd->zib.size = 0;
	zstd->zob.dst = NULL;
	zstd->zob.pos = 0;
	zstd->zob.size = 0;
#endif	/* WITH_ZSTD */
    }
}

RPMIOPOOL_MODULE(zstd)

rpmzstd rpmzstdNew(const char *path, const char *fmode, int fdno, unsigned flags)
{
    rpmzstd zstd = rpmzstdGetPool(_rpmzstdPool);
assert(fmode != NULL);		/* XXX return NULL instead? */

    const char * s = fmode;
    int omode = 0644;
    int level = 3;	/* XXX zstd permits 1-19 (default: 3) */

    char stdio[32];
    char *t = stdio;
    char *te = t + sizeof(stdio) - 2;
    int c;

SPEW("--> %s(%s,%s,%d,0x%x)\n", __FUNCTION__, path, fmode, fdno, flags);
    switch ((c = *s++)) {
    case 'a':
	*t++ = (char)c;
	flags &= ~O_ACCMODE;
	flags |= O_WRONLY | O_CREAT | O_APPEND;
	flags &= ~RPMZSTD_FLAGS_DECOMPRESS;	/* XXX */
	break;
    case 'w':
	*t++ = (char)c;
	flags &= ~O_ACCMODE;
	flags |= O_WRONLY | O_CREAT | O_TRUNC;
	flags &= ~RPMZSTD_FLAGS_DECOMPRESS;	/* XXX */
	break;
    case 'r':
	*t++ = (char)c;
	flags &= ~O_ACCMODE;
	flags |= O_RDONLY;
	flags |= RPMZSTD_FLAGS_DECOMPRESS;	/* XXX */
	break;
    }
	
    while ((c = *s++) != 0) {
	switch (c) {
	case '.':
	    break;
	case '+':
	    if (t < te) *t++ = c;
	    flags &= ~O_ACCMODE;
	    flags |= O_RDWR;
	    continue;
	    break;
	case 'b':	/* XXX binary */
	    if (t < te) *t++ = c;
	    continue;
	    break;
	case 'c':	/* XXX no cancel */
	    if (t < te) *t++ = c;
	    continue;
	    break;
	case 'm':	/* XXX mmap */
	    if ((flags & O_ACCMODE) == O_RDONLY)	/* XXX input only */
	        if (t < te) *t++ = c;
	    continue;
	    break;
	case 'e':	/* O_CLOEXEC */
	    flags |= O_CLOEXEC;
	    continue;
	    break;
	case 'n':	/* XXX O_NONBLOCK */
	    flags |= O_NONBLOCK;
	    continue;
	    break;
	case 't':	/* XXX O_TRUNC */
	    continue;
	    break;
	case 'x':
	    flags |= O_EXCL;
	    if (t < te) *t++ = c;
	    continue;
	    break;
	case '?':
	    flags |= RPMZSTD_FLAGS_DEBUGREFS;
	    flags |= RPMZSTD_FLAGS_DEBUGIO;
	    continue;
	    break;
	default:
	    if (c >= (int)'0' && c <= (int)'9') {
		level = strtol(s-1, (char **)&s, 10);
assert(level >= 1 && level <= 19);
	    }
	    continue;
	    break;
	}
	break;
    }
    *t = '\0';

    zstd->fmode = xstrdup(fmode);
    zstd->flags = flags;

    if (fdno >= 0) {
	zstd->path == NULL;
	zstd->fdno = fdno;
	if (ZSTDF_ISSET(DECOMPRESS)) {
assert((zstd->fp = fdopen(fdno, stdio)) != NULL);
	} else {
assert((zstd->fp = fdopen(fdno, stdio)) != NULL);
	}
    } else {
	zstd->path = xstrdup(path);
	zstd->fdno = -1;
    }
    zstd->oflags = (flags & _rpmioOMask);
    zstd->omode = omode;

    zstd->level = level;

#if defined(WITH_ZSTD)
    ssize_t rc;
    if (ZSTDF_ISSET(DECOMPRESS)) {	/* decompress */
	zstd->_stream = (void *) ZSTD_createDStream();
assert(zstd->_stream);

	/*
	 * In more complex scenarios, a file may consist of multiple
	 * appended frames (ex : pzstd).
	 * The following example decompresses only the first frame.
	 * It is compatible with other provided streaming examples.
	 */
	rc = rpmzstdCheck(zstd, "ZSTD_initDStream",
				ZSTD_initDStream(zstd->_stream));
	zstd->nib = rpmzstdCheck(zstd, "ZSTD_DStreamInSize",
				ZSTD_DStreamInSize());
	zstd->nob = rpmzstdCheck(zstd, "ZSTD_DStreamOutSize",
				ZSTD_DStreamOutSize());
	zstd->nb = zstd->nib;
	zstd->b = xmalloc(zstd->nb);
    } else {				/* compress */
	zstd->_stream = ZSTD_createCStream();
assert(zstd->_stream);

	rc = rpmzstdCheck(zstd, "ZSTD_initCStream",
				ZSTD_initCStream(zstd->_stream, level));
	zstd->nib = rpmzstdCheck(zstd, "ZSTD_CStreamInSize",
				ZSTD_CStreamInSize());
	zstd->nob = rpmzstdCheck(zstd, "ZSTD_CStreamOutSize",
				ZSTD_CStreamOutSize());
	zstd->nb = zstd->nob;
	zstd->b = xmalloc(zstd->nb);
    }
#endif	/* WITH_ZSTD */

    return rpmzstdLink(zstd);
}

int rpmzstdFlush(void * _zstd)
{
    int rc = -2;

#if defined(WITH_ZSTD)
    rpmzstd zstd = (rpmzstd) _zstd;

assert(zstd->_stream);
    if (ZSTDF_ISSET(DECOMPRESS)) {	/* decompress */
    } else {				/* compress */
	/* close frame */
	zstd->zob.dst  = zstd->b;
	zstd->zob.size = zstd->nb;
	zstd->zob.pos  = 0;
	/* XXX nremain != 0 on small output buffers. Loop. */
	ssize_t nremain = rpmzstdCheck(zstd, "ZSTD_flushStream",
				ZSTD_flushStream(zstd->_stream, &zstd->zob));
	if (nremain) {
fprintf(stderr, "not fully flushed");
	    rc = -1;
	} else {
	    size_t nw = fwrite(zstd->b, 1, zstd->zob.pos, zstd->fp);
assert(nw == zstd->zob.pos);
SPEW("<--\t    fwrite(%p, 1, %zu, %p) nw %zd\n", zstd->b, zstd->zob.pos, zstd->fp, nw);
	    rc = 0;
	}
    }

SPEW("<--\t%s(%p) rc %d\n", __FUNCTION__, zstd, rc);
#endif	/* WITH_ZSTD */

    return rc;
}

int rpmzstdClose(rpmzstd zstd)
{
    int rc = -2;

#if defined(WITH_ZSTD)
SPEW("-->\t%s(%p)\n", __FUNCTION__, zstd);
assert(zstd->_stream);
    if (ZSTDF_ISSET(DECOMPRESS)) {	/* decompress */
	rc = 0;
    } else {				/* compress */
	/* close frame */
	zstd->zob.dst  = zstd->b;
	zstd->zob.size = zstd->nb;
	zstd->zob.pos  = 0;
	/* XXX nremain != 0 on small output buffers. Loop. */
	ssize_t nremain = rpmzstdCheck(zstd, "ZSTD_endStream",
				ZSTD_endStream(zstd->_stream, &zstd->zob));
	if (nremain) {
fprintf(stderr, "not fully flushed");
	    rc = -1;
	} else {
	    if (zstd->fp) {
		size_t nw = fwrite(zstd->b, 1, zstd->zob.pos, zstd->fp);
assert(nw == zstd->zob.pos);
SPEW("<--\t    fwrite(%p, 1, %zu, %p) nw %zd\n", zstd->b, zstd->zob.pos, zstd->fp, nw);
	    }
	    rc = 0;
	}
	rpmzstdFini(zstd);	/* XXX rpmzstdFree? */
    }

SPEW("<--\t%s(%p) rc %d\n", __FUNCTION__, zstd, rc);
#endif	/* WITH_ZSTD */
    return rc;
}

ssize_t rpmzstdRead(rpmzstd zstd, void *b, size_t nb)
{
    ssize_t rc = -2;
#if defined(WITH_ZSTD)
    ssize_t nr = 0;
    size_t ipos = 0;

SPEW("-->\t%s(%p,%p[%zd]) fp %p\n", __FUNCTION__, zstd, b, nb, (zstd ? zstd->fp : NULL));
    if (zstd->fp == NULL)
	goto exit;

    ZSTD_outBuffer zob = { b, nb, 0 };

    while (zob.pos < zob.size) {

SPEW("\t%s:   top %8s %8zd\tzob %zu:%zu\tzib %zu:%zu\n", __FUNCTION__, "ipos", ipos, zob.pos, zob.size, zstd->zib.pos, zstd->zib.size);
	/* Re-fill compressed data buffer. */
	if (zstd->zib.pos >= zstd->zib.size) {
	    nr = fread(zstd->b, 1, zstd->nb, zstd->fp);
SPEW("<--\t    fread(%p, 1, %zu, %p) nr %zd\n", zstd->b, zstd->nb, zstd->fp, nr);
	    if (nr == 0)
		break;
	    ipos += nr;

	    /* Reset to beginning of compressed data buffer. */
	    zstd->zib.src  = zstd->b;
	    zstd->zib.size = nr;
	    zstd->zib.pos  = 0;
SPEW("\t%s: reset %8s %8zd\tzob %zu:%zu\tzib %zu:%zu\n", __FUNCTION__, "ipos", ipos, zob.pos, zob.size, zstd->zib.pos, zstd->zib.size);
	}

	/* zstd->nextnb : size of next compressed block */
	zstd->nextnb = rpmzstdCheck(zstd, "ZSTD_decompressStream",
			ZSTD_decompressStream(zstd->_stream, &zob, &zstd->zib));
SPEW("\t%s: after %8s %8zd\tzob %zu:%zu\tzib %zu:%zu\n", __FUNCTION__, "nextnb", zstd->nextnb, zob.pos, zob.size, zstd->zib.pos, zstd->zib.size);
	if (ZSTD_isError(zstd->nextnb)) {
	    rc = -1;	/* XXX */
	    goto exit;
	}

    }
SPEW("\t%s: final %8s %8zd\tzob %zu:%zu\tzib %zu:%zu\n", __FUNCTION__, "ipos", ipos, zob.pos, zob.size, zstd->zib.pos, zstd->zib.size);
    rc = zob.pos;
 
exit:
SPEW("<--\t%s(%p,%p[%zd]) rc %ld\n", __FUNCTION__, zstd, b, nb, rc);
#endif	/* WITH_ZSTD */
    return rc;
}

ssize_t rpmzstdWrite(rpmzstd zstd, const void *b, size_t nb)
{
    ssize_t rc = -2;
#if defined(WITH_ZSTD)
    ssize_t nw = 0;
    size_t opos = 0;

SPEW("-->\t%s(%p,%p[%zd])\n", __FUNCTION__, zstd, b, nb);
    if (zstd->fp == NULL)
	goto exit;

    rc = 0;
    ZSTD_inBuffer zib = { b, nb, 0 };

    while (zib.pos < zib.size) {

SPEW("\t%s:   top %8s %8zd\tzob %zu:%zu\tzib %zu:%zu\n", __FUNCTION__, "opos", opos, zstd->zob.pos, zstd->zob.size, zib.pos, zib.size);

	/* Reset to beginning of compressed data buffer. */
	zstd->zob.dst  = zstd->b;
	zstd->zob.size = zstd->nb;
	zstd->zob.pos  = 0;
SPEW("\t%s: reset %8s %8zd\tzob %zu:%zu\tzib %zu:%zu\n", __FUNCTION__, "opos", opos, zstd->zob.pos, zstd->zob.size, zib.pos, zib.size);

	/* zstd->nb is guaranteed to be <= ZSTD_CStreamInSize() */
	zstd->nextnb = rpmzstdCheck(zstd, "ZSTD_compressStream",
			ZSTD_compressStream(zstd->_stream, &zstd->zob, &zib));
SPEW("\t%s: after %8s %8zd\tzob %zu:%zu\tzib %zu:%zu\n", __FUNCTION__, "nextnb", zstd->nextnb, zstd->zob.pos, zstd->zob.size, zib.pos, zib.size);
	if (ZSTD_isError(zstd->nextnb)) {
	    rc = -1;	/* XXX */
	    goto exit;
	}

	/* Write compressed data buffer. */
	if (zstd->zob.pos > 0) {
	    nw = fwrite(zstd->b, 1, zstd->zob.pos, zstd->fp);
assert((size_t)nw == zstd->zob.pos);
SPEW("<--\t    fwrite(%p, 1, %zu, %p) nw %zd\n", zstd->b, zstd->zob.pos, zstd->fp, nw);
	    if (nw > 0)
		opos += nw;
	}

    }
SPEW("\t%s: final %8s %8zd\tzob %zu:%zu\tzib %zu:%zu\n", __FUNCTION__, "opos", opos, zstd->zob.pos, zstd->zob.size, zib.pos, zib.size);
    rc = zib.pos;

exit:
SPEW("<--\t%s(%p,%p[%zd]) rc %ld\n", __FUNCTION__, zstd, b, nb, rc);
#endif	/* WITH_ZSTD */

    return rc;
}

RPM_GNUC_CONST
int rpmzstdSeek(void * _zstd, _libio_pos_t pos, int whence)
{
    int rc = -2;
#ifdef USE_COOKIE_SEEK_POINTER
    _IO_off64_t p = *pos;
#else
    off_t p = pos;
#endif
    (void)p;
    return rc;
}

rpmzstd rpmzstdOpen(const char * path, const char * fmode, int fdno, unsigned flags)
{
    return  rpmzstdNew(path, fmode, fdno, flags);
}

RPM_GNUC_CONST
rpmzstd rpmzstdFdopen(void * _fdno, const char * fmode)
{
#ifdef	NOTYET
    int fdno = (int)((long)_fdno);	/* XXX hack */
    return  rpmzstdNew(NULL, fmode, fdno, 0);
#else
    return NULL;
#endif
}

/* =============================================================== */
typedef struct cpio_state_s {
    uint32_t n;				/* byte progress in cpio header */
    uint32_t mode;			/* file attributes */
    uint32_t nlnk;
    uint32_t size;
} * cpio_state;

#define RSYNC_WIN 4096

typedef struct rsync_state_s {
    uint32_t n;				/* number of elements in the window */
    uint32_t sum;			/* current sum */
    unsigned char win[RSYNC_WIN];	/* window elements */
} * rsync_state;

/* Should zstdflush be called only after RSYNC_WIN boundaries? */
static int enable_rsync = 1;

typedef struct zstdFILE_s * zstdFILE;
struct zstdFILE_s {
    rpmzstd zstd;			/* zstd is a pointer */
    struct rsync_state_s rs;
    struct cpio_state_s cs;
    uint32_t nb;			/* bytes pending for sync */
};

#ifdef __cplusplus
GENfree(zstdFILE)
#endif	/* __cplusplus */

/* from ../lib/cpio.h */
#define CPIO_NEWC_MAGIC "070701"
#define PHYS_HDR_SIZE 110

#define OFFSET_MODE (sizeof(CPIO_NEWC_MAGIC)-1 + 1*8)
#define OFFSET_NLNK (sizeof(CPIO_NEWC_MAGIC)-1 + 4*8)
#define OFFSET_SIZE (sizeof(CPIO_NEWC_MAGIC)-1 + 6*8)

static inline
int hex(char c)
{
    if (c >= '0' && c <= '9')
	return (int)(c - '0');
    else if (c >= 'a' && c <= 'f')
	return (int)(c - 'a') + 10;
    else if (c >= 'A' && c <= 'F')
	return (int)(c - 'A') + 10;
    return -1;
}

static inline
bool cpio_next(cpio_state s, unsigned char c)
{
    if (s->n >= sizeof(CPIO_NEWC_MAGIC)-1) {
	int d = hex(c);
	if (d < 0) {
	    s->n = 0;
	    return false;
	}
	if (0){} /* indent */
	else if (s->n >= OFFSET_MODE && s->n < OFFSET_MODE+8) {
	    if (s->n == OFFSET_MODE)
		s->mode = 0;
	    else
		s->mode <<= 4;
	    s->mode |= d;
	}
	else if (s->n >= OFFSET_NLNK && s->n < OFFSET_NLNK+8) {
	    if (s->n == OFFSET_NLNK)
		s->nlnk = 0;
	    else
		s->nlnk <<= 4;
	    s->nlnk |= d;
	}
	else if (s->n >= OFFSET_SIZE && s->n < OFFSET_SIZE+8) {
	    if (s->n == OFFSET_SIZE)
		s->size = 0;
	    else
		s->size <<= 4;
	    s->size |= d;
	}
	s->n++;
	if (s->n >= PHYS_HDR_SIZE) {
	    s->n = 0;
	    if (!S_ISREG(s->mode) || s->nlnk != 1)
		/* no file data */
		s->size = 0;
	    return true;
	}
    }
    else if (CPIO_NEWC_MAGIC[s->n] == c) {
	s->n++;
    }
    else {
	s->n = 0;
    }
    return false;
}

/* =============================================================== */
static inline
bool rsync_next(rsync_state s, unsigned char c)
{
    uint32_t i;

    if (s->n < RSYNC_WIN) {		/* not enough elements */
	s->sum += (uint32_t)c;		/* update the sum */
	s->win[s->n++] = c;		/* remember the element */
	return false;			/* no match */
    }
    i = s->n++ % RSYNC_WIN;		/* wrap up */
    s->sum -= (uint32_t)s->win[i];	/* move the window on */
    s->sum += (uint32_t)c;
    s->win[i] = c;
    if (s->sum % RSYNC_WIN == 0) {	/* match */
	s->n = 0;			/* reset */
	s->sum = 0;
	return true;
    }
    return false;
}

#define CHUNK 4096

static inline
bool sync_hint(zstdFILE zsf, unsigned char c)
{
    bool cpio_hint;
    bool rsync_hint;

    zsf->nb++;
    cpio_hint = cpio_next(&zsf->cs, c);
    if (cpio_hint) {
	/* cpio header/data boundary */
	zsf->rs.n = zsf->rs.sum = 0;
	if (zsf->nb >= 2*CHUNK)
	    /* better sync here */
	    goto cpio_sync;
	if (zsf->cs.size < CHUNK)
	    /* file is too small */
	    return false;
	if (zsf->nb < CHUNK/2)
	    /* not enough pending bytes */
	    return false;
    cpio_sync:
	zsf->nb = 0;
	return true;
    }
    rsync_hint = rsync_next(&zsf->rs, c);
    if (rsync_hint) {
	/* rolling checksum match */
	assert(zsf->nb >= RSYNC_WIN);
	zsf->nb = 0;
	return true;
    }
    return false;
}

static ssize_t
rsyncable_write(zstdFILE zsf, const unsigned char *const buf, const size_t len)
{
    ssize_t rc;
    size_t n;
    ssize_t n_written = 0;
    const unsigned char *begin = buf;
    size_t i;

    for (i = 0; i < len; i++) {
	if (!sync_hint(zsf, buf[i]))
	    continue;
	n = i + 1 - (begin - buf);
	rc = rpmzstdWrite(zsf->zstd, (void *)begin, (unsigned)n);
	if (rc < 0)
	    return (n_written ? n_written : rc);
	n_written += rc;
	if (rc < (ssize_t)n)
	    return n_written;
	begin += n;
	rc = rpmzstdFlush(zsf->zstd);
	if (rc < 0)
	    return (n_written ? n_written : rc);
    }
    if (begin < buf + len) {
	n = len - (begin - buf);
	rc = rpmzstdWrite(zsf->zstd, (void *)begin, (unsigned)n);
	if (rc < 0)
	    return (n_written ? n_written : rc);
	n_written += rc;
    }
    return n_written;
}

/* =============================================================== */

RPM_GNUC_PURE
static inline
void * zstdFileno(FD_t fd)
{
    void * rc = NULL;
    int i;

    FDSANE(fd);
    for (i = fd->nfps; i >= 0; i--) {
	FDSTACK_t * fps = &fd->fps[i];
	if (fps->io != zstdio)
	    continue;
	rc = fps->fp;
	break;
    }

    return rc;
}

static
FD_t zstdOpen(const char * path, const char * fmode)
{
    if (path == NULL) return NULL;
    if (fmode == NULL) return NULL;

    /* Create the private FILE handle. */
    rpmzstd zstd = rpmzstdOpen(path, fmode, -1, _rpmzstdOFlags);
    if (zstd == NULL)
	return NULL;

    /* Layer on the cpio header flush wrapper. */
    zstdFILE zsf = (zstdFILE) xcalloc(1, sizeof(*zsf));
    zsf->zstd = zstd;

    /* Create the new FD_t */
    FD_t fd = fdNew("open (zstdOpen)");
#ifdef  NOTYET		/* XXX persistent URI cache prevents fileno(bz->fp) */
    fdPop(fd); fdPush(fd, zstdio, zsf, fileno(bz->fp));
#else
    fdPop(fd); fdPush(fd, zstdio, zsf, -1);
#endif
    fdSetOpen(fd, path, zstd->oflags, zstd->omode);

DBGIO(fd, (stderr, "==>\tzstdOpen(\"%s\", \"%s\") fd %p %s\n", path, fmode, (fd ? fd : NULL), fdbg(fd)));
    return fdLink(fd, __FUNCTION__);
}

static FD_t zstdFdopen(void * cookie, const char *fmode)
{
    FD_t fd = c2f(cookie);

    if (fmode == NULL) return NULL;

    int fdno = fdFileno(fd);
    if (fdno < 0) return NULL;

    /* Create the private FILE handle. */
    rpmzstd zstd = rpmzstdOpen(NULL, fmode, fdno, _rpmzstdOFlags);
    if (zstd == NULL)
	return NULL;

    /* Layer on the cpio header flush wrapper. */
    zstdFILE zsf = (zstdFILE) xcalloc(1, sizeof(*zsf));
    zsf->zstd = zstd;

    /* Pull forward the ufdio path. */
    const char *fn = fdGetOPath(fd);
    if (fn) {
	if (zstd->path)
	    zstd->path = _free(zstd->path);
	zstd->path = xstrdup(fn);
    }
	
    /* Replace ufdio with zstdio. */
#ifdef  NOTYET          /* XXX persistent URI cache prevents pop */
    fdPop(fd); fdPush(fd, zstdio, zsf, fileno(zstd->fp));
#else
    (void) fdSetFdno(fd, -1);		/* XXX skip the fdio close */
    fdPush(fd, zstdio, zsf, fdno);		/* Push zstdio onto stack */
#endif
    fdSetOpen(fd, zstd->path, zstd->oflags, zstd->omode);

    return fdLink(fd, __FUNCTION__);
}

static int zstdFlush(void * cookie)
{
    FD_t fd = c2f(cookie);
    zstdFILE zsf;
    zsf = (zstdFILE) zstdFileno(fd);
    if (zsf == NULL) return -2;
    return rpmzstdFlush(zsf->zstd);
}

/* =============================================================== */
static ssize_t zstdRead(void * cookie, char * buf, size_t count)
{
    FD_t fd = c2f(cookie);
    zstdFILE zsf;
    ssize_t rc;

    if (fd == NULL || fd->bytesRemain == 0) return 0;	/* XXX simulate EOF */

    zsf = (zstdFILE) zstdFileno(fd);
    if (zsf == NULL) return -2;	/* XXX can't happen */

    fdstat_enter(fd, FDSTAT_READ);
    rc = rpmzstdRead(zsf->zstd, buf, (unsigned)count);
DBGIO(fd, (stderr, "==>\tzstdRead(%p,%p,%u) rc %lx %s\n", cookie, buf, (unsigned)count, (unsigned long)rc, fdbg(fd)));
    if (rc < 0) {
#ifdef	NOTYET
	int zerror = 0;
	fd->errcookie = rpmzstdError(zsf->zstd, &zerror);
	if (zerror == Z_ERRNO) {
	    fd->syserrno = errno;
	    fd->errcookie = strerror(fd->syserrno);
	}
#else
	fd->syserrno = errno;
	fd->errcookie = "zstdWrite error";
#endif
    } else {
	fdstat_exit(fd, FDSTAT_READ, (rc > 0 ? rc : 0));
	if (fd->ndigests > 0 && rc > 0) fdUpdateDigests(fd, (const unsigned char *)buf, rc);
    }
    return rc;
}

static ssize_t zstdWrite(void * cookie, const char * buf, size_t count)
{
    FD_t fd = c2f(cookie);
    zstdFILE zsf;
    ssize_t rc;

    if (fd == NULL || fd->bytesRemain == 0) return 0;	/* XXX simulate EOF */

    if (fd->ndigests > 0 && count > 0) fdUpdateDigests(fd, (const unsigned char *)buf, count);

    zsf = (zstdFILE) zstdFileno(fd);
    if (zsf == NULL) return -2;	/* XXX can't happen */

    fdstat_enter(fd, FDSTAT_WRITE);
    if (enable_rsync)
	rc = rsyncable_write(zsf, (const unsigned char *)buf, (unsigned)count);
    else
	rc = rpmzstdWrite(zsf->zstd, (void *)buf, (unsigned)count);
DBGIO(fd, (stderr, "==>\tzstdWrite(%p,%p,%u) rc %lx %s\n", cookie, buf, (unsigned)count, (unsigned long)rc, fdbg(fd)));
    if (rc < (ssize_t)count) {
#ifdef	NOTYET
	int zerror = 0;
	fd->errcookie = rpmzstdError(zsf->zstd, &zerror);
	if (zerror == Z_ERRNO) {
	    fd->syserrno = errno;
	    fd->errcookie = strerror(fd->syserrno);
	}
@else
	fd->syserrno = errno;
	fd->errcookie = "zstdWrite error";
#endif
    }
    if (rc > 0)
	fdstat_exit(fd, FDSTAT_WRITE, rc);
    return rc;
}

static int zstdSeek(void * cookie, _libio_pos_t pos, int whence)
{
    int rc;
    FD_t fd = c2f(cookie);
    zstdFILE zsf;

    if (fd == NULL) return -2;
    assert(fd->bytesRemain == -1);	/* XXX FIXME */

    zsf = (zstdFILE) zstdFileno(fd);
    if (zsf == NULL) return -2;	/* XXX can't happen */

    fdstat_enter(fd, FDSTAT_SEEK);
    rc = rpmzstdSeek(zsf->zstd, pos, whence);
DBGIO(fd, (stderr, "==>\tzstdSeek(%p,%p,%d) rc %lx %s\n", cookie, pos, whence, (unsigned long)rc, fdbg(fd)));
    if (rc < 0) {
#ifdef	NOTYET
	int zerror = 0;
	fd->errcookie = rpmzstdError(zsf->zstd, &zerror);
	if (zerror == Z_ERRNO) {
	    fd->syserrno = errno;
	    fd->errcookie = strerror(fd->syserrno);
	}
#else
	fd->syserrno = errno;
	fd->errcookie = "zstdSeek error";
#endif
    }
    if (rc > 0)
	fdstat_exit(fd, FDSTAT_SEEK, rc);
    return rc;
}

static int zstdClose(void * cookie)
{
    FD_t fd = c2f(cookie);
    zstdFILE zsf;
    int rc;

    zsf = (zstdFILE) zstdFileno(fd);
    if (zsf == NULL) return -2;	/* XXX can't happen */

    fdstat_enter(fd, FDSTAT_CLOSE);
    rc = rpmzstdClose(zsf->zstd);
    zsf->zstd = rpmzstdFree(zsf->zstd);
    zsf = _free(zsf);

    /* XXX TODO: preserve fd if errors */

    if (fd) {
DBGIO(fd, (stderr, "==>\tzstdClose(%p) zerror %d %s\n", cookie, rc, fdbg(fd)));
	if (rc < 0) {
#ifdef	NOTYET
	    if (rc == Z_ERRNO) {
		fd->syserrno = errno;
		fd->errcookie = strerror(fd->syserrno);
	    }
#else
	    fd->syserrno = errno;
	    fd->errcookie = "zstdClose error";
#endif
	} else if (rc >= 0) {
	    fdstat_exit(fd, FDSTAT_CLOSE, rc);
	}
    }

DBGIO(fd, (stderr, "==>\tzstdClose(%p) rc %lx %s\n", cookie, (unsigned long)rc, fdbg(fd)));

    if (_rpmio_debug || rpmIsDebug()) fdstat_print(fd, "ZSTIO", stderr);
    if (rc == 0)
	fd = fdFree(fd, "open (zstdClose)");
    return rc;
}

static struct FDIO_s _zstdio = {
  zstdRead, zstdWrite, zstdSeek, zstdClose,	zstdOpen, zstdFdopen, zstdFlush,
};

FDIO_t zstdio = &_zstdio;

